/*
  gm_arch.c
  architecture specific code for AIX 4.3 and 5.1 GM driver
*/

/* NOTE: The include of gm_arch_debug.h MUST be first! */
#include "gm_arch_debug.h"
#include "gm_arch_types.h"
#include "gm_internal.h"
#include "gm_page_hash.h"
#include "gm_lanai.h"
#include "gm_instance.h"
#include "gm_impl.h"
#include "gm_arch.h"
#include "gm_pio.h"
#include "gm_debug_mem_register.h"

#include <sys/ioacc.h>
#include <sys/pin.h>
#include <sys/sleep.h>
#include <sys/syspest.h>
#include <sys/time.h>
#include <sys/timer.h>
#include <sys/uio.h>
#include <sys/xmem.h>

#ifndef PAGESIZE
#define PAGESIZE 4096
#ifdef	PGSHIFT
#undef	PGSHIFT
#endif	/* PGSHIFT */
#define	PGSHIFT	12
#endif	/* PAGESIZE */

#ifdef	SUPPORT_64BIT_OS
#define	DO_MAP_PAGE 1
#else
#define	DO_MAP_PAGE 0
#endif	/* SUPPORT_64BIT_OS */


/***********************************************************************
 * Global data
 ***********************************************************************/

static int sync_count = 0;
static char error_hex[36];

static gm_instance_state_t* gm_aix_instances [GM_ARCH_MAX_INSTANCE];

/*********************************************************************/

/*
 * The global device driver control area. Initialize the lockl lock 
 * (cfg_lock) in the beginning of the structure to LOCK_AVAIL for 
 * synchronizing the config commands.
 */
dd_ctrl_t dd_ctrl = {LOCK_AVAIL, NULL, 0, 0, 0};



/***********************************************************************
 * "Forward" function declarations
 ***********************************************************************/

GM_STATIC int gm_aix_init (dev_ctrl_t *p_dev);
GM_STATIC int gm_aix_clean (dev_ctrl_t *p_dev);
GM_STATIC int gm_aix_rw_pci_cfg (uint bid, uint slot, ulong offset,
				 void *p_buf, int size, int read_write);
GM_STATIC int gm_aix_mmap (dev_ctrl_t *p_dev, gm_port_state_t *ps, void *arg);
GM_STATIC int gm_aix_localize_status (gm_status_t status);

int gm_aix_intr (struct intr *p_ihs);

GM_STATIC int cfg_init (dds_t *dds);
GM_STATIC int cfg_term (dds_t *dds);
GM_STATIC int cfg_dev (dev_ctrl_t *p_dev);
GM_STATIC void ucfg_dev (dev_ctrl_t *p_dev);
GM_STATIC dev_ctrl_t * gm_aix_device (dev_t devno);
GM_STATIC void add_cdt (char *name, char *addr, int len);
GM_STATIC void del_cdt (char *name, char *addr, int len);



/**********************************************************************
 *
 * Global utility functions
 *
 * WARNING!  We can't use gm_aix_malloc() or gm_aix_zalloc() until
 *           our heap has been allocated (see gm_aix_config()).
 *
 **********************************************************************/

void *
gm_aix_malloc (
	       unsigned long len,
	       unsigned int align
	       )
{
  /* We use our own heap */
  gm_assert (dd_ctrl.initialized);
  gm_assert (dd_ctrl.my_heap);
  return (xmalloc (len, align, dd_ctrl.my_heap));

} /* gm_aix_malloc () */



void *
gm_aix_zalloc (
	       unsigned long len
	       )
{
  void *p;

  p = gm_aix_malloc (len, 3);

  bzero(p, len);

  GM_PRINT(GM_AIX_DEBUG_MALLOC,
	   ("gm_aix_zalloc called with length %d, result was 0x%x\n", len, p));
  return p;

} /* gm_aix_zalloc () */


char *
gm_aix_stringify_error(int error_val)
{
  switch (error_val) {
  case EEXIST:
    return "EEXIST";
    break;

  case ENODEV:
    return "ENODEV";
    break;

  case EINVAL:
    return "EINVAL";
    break;

  default:
    sprintf(error_hex, "0x%x", error_val);
    return error_hex;
  }
}



/**********************************************************************
 *
 * Driver hooks called by AIX kernel
 *
 **********************************************************************/

/*****************************************************************************/
/*
 * NAME:                  gm_aix_config
 *
 * FUNCTION:              config entry point -- declared to 'ld' as driver
 *                        entry point (see ../make-os).
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:                 Applies lockl to serialize INIT, TERM and QVPD
 *                        operations.
 *
 *                        This function is too big, and should be decomposed
 *                        along functional lines.
 *
 * DATA:                  dd_ctrl: the global driver control structure
 *                        p_dev:   -> instance-specific dev_ctrl_t struct
 *
 * CAUTION:               The AIX man page for the lockl Kernel Service says,
 *                        "The lockl kernel service is provided for
 *                        compatibility only and should not be used in new
 *                        code, which should instead use simple locks or
 *                        complex locks."  For whatever reason, Manoj chose to
 *                        use lockl().  We should fix this as soon as the
 *                        driver has stabilized.   - Max, 2002/05/21
 *
 * RETURNS:               errno values
 *	          
 */
/*****************************************************************************/
int 
gm_aix_config(
	      int cmd,		/* Desired op (INIT, TERM, QVPD) */
	      struct uio *p_uio	/* pointer to dds_t */
	      )
{
  int	rc = 0;			/* Function return code */
  dds_t	gm_aix_dds;		/* config struct received in uio */
  dev_ctrl_t *p_dev;		/* Device control struct pointer */

  GM_PRINT(GM_AIX_DEBUG,
	   ("gm_aix_config entered; dd_ctrl_t is at 0x%x\n", &dd_ctrl));

#if GM_AIX_DEBUG
  if (cmd == CFG_INIT)
    GM_PRINT(1, ("- entered with CFG_INIT command\n"));
  else if (cmd == CFG_TERM)
    GM_PRINT(1, ("- entered with CFG_TERM command\n"));
  else if (cmd == CFG_QVPD)
    GM_PRINT(1, ("- entered with CFG_QVPD command\n"));
  else
    GM_WARN(("- entered with unknown command %d\n", cmd));
#endif
  /* Use lockl operation to serialize the execution of the
   * config commands.  Can be used without doing any lock alloc or init.
   */
  lockl(&dd_ctrl.cfg_lock, LOCK_SHORT);

  /* Get the gm_aix_dds structure passed by the method */
  if ((rc = uiomove((caddr_t) &gm_aix_dds, sizeof(dds_t),
		    UIO_WRITE, p_uio)) != 0) {
    unlockl(&dd_ctrl.cfg_lock);
    return(rc);
  }

  /* If first time around....  */
  if (!dd_ctrl.initialized) {

    GM_PRINT(GM_AIX_DEBUG, ("- pinning all code and data\n"));

    /* Pin ALL code and data */
    if (rc = pincode(gm_aix_config)) {
      unlockl(&dd_ctrl.cfg_lock);
      return(rc);	/* Returns EINVAL or ENOMEM. */
    }

    /* Allocate and pin our own heap for memory allocations */
    /* WARNING!  We can't use __gm_arch_kernel_malloc(), gm_aix_malloc(),
     *           or gm_aix_zalloc() until this has been done!
     */
    {
      void	*p;
      if ((p = (void *)xmalloc (MY_HEAP_SIZE, 3, kernel_heap)) == NULL) {
	GM_WARN(("- heap alloc failed\n"));
	unlockl(&dd_ctrl.cfg_lock);
	return(ENOMEM);
      }

      if (pin (p, MY_HEAP_SIZE)) {
	GM_WARN(("- heap pin failed\n"));
	xmfree (p, kernel_heap);
	unlockl(&dd_ctrl.cfg_lock);
	return(ENOMEM);
      }

      if ((dd_ctrl.my_heap = init_heap (p, MY_HEAP_SIZE, NULL)) == NULL) {
	GM_WARN(("- heap init failed\n"));
	unpin (p, MY_HEAP_SIZE);
	xmfree (p, kernel_heap);
	unlockl(&dd_ctrl.cfg_lock);
	return(ENOMEM);
      }

      dd_ctrl.my_heap_addr = p;
    }

    /* Initialize the global device driver locks.  */
    lock_alloc(&dd_ctrl.i_lock, LOCK_ALLOC_PIN, GM_AIX_LOCK_CLASS,
	       GM_AIX_LOCK_INSTANCE_DD);
    simple_lock_init(&dd_ctrl.i_lock);
    lock_alloc(&dd_ctrl.o_lock, LOCK_ALLOC_PIN, GM_AIX_LOCK_CLASS,
	       GM_AIX_LOCK_INSTANCE_OPEN);
    simple_lock_init(&dd_ctrl.o_lock);
    GM_PRINT(GM_AIX_DEBUG_LOCKS,
	     ("- created global driver locks at 0x%8x, 0x%8x\n",
	      &dd_ctrl.i_lock, &dd_ctrl.o_lock));

    /* Setup variables in the device control structure. */
    dd_ctrl.num_devs = 0;
    dd_ctrl.num_ip_opens = 0;

    /* Initialize component dump table header */
    /* NOTE: lint may object to this assignment, but it's really OK: */
    dd_ctrl.cdt.head._cdt_magic = DMP_MAGIC;
    strncpy(dd_ctrl.cdt.head._cdt_name, DD_NAME_STR,
	    sizeof(dd_ctrl.cdt.head._cdt_name));
    dd_ctrl.cdt.head._cdt_len = sizeof(dd_ctrl.cdt.head);

    /* Add the driver-specific structure to the component dump table */
    add_cdt("dd_ctrl",(char *)(&dd_ctrl), (int)sizeof(dd_ctrl));

    /* Indicate that we have done this once. */
    dd_ctrl.initialized = TRUE;

  } /* if (!dd_ctrl.initialized) */

  switch(cmd) {

  case CFG_INIT:

    /* Perform the device initialization */
    if ((rc = cfg_init(&gm_aix_dds)) != 0) {
      GM_WARN(("- cfg_init failed 0x%x\n", rc));
      cfg_term(&gm_aix_dds); 
      break;
    }

    break;

  case CFG_TERM:

    /* Perform device termination. */
    rc = cfg_term(&gm_aix_dds);

    break;

  default:
    rc = EINVAL;

  } /* switch(cmd) */


  /* If last device terminated, release driver specific resources */ 
  if (dd_ctrl.num_devs == 0) {
    del_cdt("dd_ctrl", (char *)(&dd_ctrl), (int)sizeof(dd_ctrl)); 
    lock_free(&dd_ctrl.o_lock);
    lock_free(&dd_ctrl.i_lock);
    dd_ctrl.initialized = FALSE;
    GM_PRINT(GM_AIX_DEBUG_LOCKS,
	   ("*** - deleted global driver locks at 0x%8x, 0x%8x\n",
	    &dd_ctrl.i_lock, dd_ctrl.o_lock));

    if (dd_ctrl.my_heap) {
      unpin (dd_ctrl.my_heap_addr, MY_HEAP_SIZE);
      xmfree (dd_ctrl.my_heap_addr, kernel_heap);
      dd_ctrl.my_heap_addr = NULL;
      dd_ctrl.my_heap = NULL;
    }
    else {
      GM_WARN(("- NO my_heap ?????\n"));
    }

    /* Finally unpin the driver. */
    unpincode(gm_aix_config);
  }

  GM_PRINT(GM_AIX_DEBUG, ("- returning %d\n", rc));
  unlockl(&dd_ctrl.cfg_lock);

  return (rc);

} /* gm_aix_config() */


/*****************************************************************************/
/*
 * NAME:                  cfg_init
 *
 * FUNCTION:              Perform CFG_INIT function.
 *			  * Initialize the device-specific structures
 *			  * Init hardware
 *			  * devswadd()
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:		  Assumes the caller has already ensured serialization
 *
 * RETURNS:               0 - OK
 *                        EEXIST - device name in use (from ns_attach)
 *	                  EINVAL - invalid parameter was passed
 *	                  EIO - permanent I/O error
 */
/*****************************************************************************/
GM_STATIC int 
cfg_init(
	 dds_t *dds
	 )
{
  int rc;			/* Function return code */
  dev_ctrl_t *p_dev;
  int dev_size;			/* Size of dev_ctrl memory to alloc */
  struct devsw devsw;

  /* get driver lock for configuring the device */
  gm_aix_lock_driver();

  GM_PRINT(GM_AIX_DEBUG,
	   ("cfg_init - seq_number = %d\n", dds->seq_number));

  /* If there is alread an entry for this device, return EBUSY */
  for (p_dev = dd_ctrl.p_dev_list; p_dev != NULL; p_dev = p_dev->next) {
    if (p_dev->dds.seq_number == dds->seq_number) {
      GM_PRINT(GM_AIX_DEBUG, ("- p_dev already exists!\n"));
      rc = EBUSY;
      goto abort_with_lock;
    }
  }

  /* If at the limit of the number of adapters, error */
  if (dd_ctrl.num_devs >= PCI_MAX_ADAPTERS) {
    GM_WARN(("- EBUSY; too many adapters (%d)\n", dd_ctrl.num_devs));
    rc = EBUSY;
    goto abort_with_lock;
  }

  /* Calculate the size needed for the dev_ctrl plus the
   * transmit wait queue and allocate it
   */
  dev_size = sizeof(dev_ctrl_t);
  if ((p_dev = (dev_ctrl_t *) xmalloc( dev_size, 3, pinned_heap )) == NULL) {
    GM_WARN(("- ENOMEM; xmalloc of DDS and TX queue failed\n"));
    rc = ENOMEM;
    goto abort_with_lock;
  }

  bzero (p_dev, dev_size);
  WRK.alloc_size = dev_size;
  p_dev->usage_count = 0;	/* Make this explicit to expose dependency */

  bcopy (dds, &(DDS), sizeof (DDS));
  GM_PRINT(GM_AIX_DEBUG, ("- DDS.seq_number %d\n", DDS.seq_number));

  /* Add the new dev_ctrl struct into the driver's linked list */
  p_dev->next = dd_ctrl.p_dev_list;
  p_dev->unit = dds->minor / 2;
  dd_ctrl.p_dev_list = p_dev;
  dd_ctrl.num_devs++;

  /* Setup the PCI I/O space map structure */
  if ((IO_MAP = xmalloc(sizeof(struct io_map), 3, pinned_heap)) == NULL) {
    rc = ENOMEM;
    GM_WARN(("- ENOMEM; xmalloc of IO_MAP failed\n"));

    /* Free the device's allocated memory */
    if (xmfree(p_dev, pinned_heap) != 0) {
      GM_WARN(("- xmfree of DDS/TX/RX failed\n"));
    }

    goto abort_with_lock;
  }

  DEV_BID = BID_ALTREG(DDS.bus_id, PCI_BUSMEM);

  bzero (IO_MAP, sizeof IO_MAP);
  IO_MAP->key = IO_MEM_MAP;
  IO_MAP->flags = 0;
  IO_MAP->size = 0x1000000;
  IO_MAP->bid = DEV_BID;
  IO_MAP->busaddr = DDS.bus_mem_addr;

  GM_PRINT(GM_AIX_DEBUG, ("- DDS.bus_id = 0x%X, p_dev->work.bid = 0x%X\n",
			  DDS.bus_id, DEV_BID));

#if DO_MAP_PAGE
  p_dev->d_handle = d_map_init (BID_ALTREG (DDS.bus_id, PCI_BUSMEM),
				DMA_MASTER | DMA_MAXMIN_32M |       \
				DMA_ADDRESS_64_BIT|DMA_ENABLE_64,
				0, 0);
  if ((p_dev->d_handle == (d_handle_t) DMA_FAIL) ||
      (p_dev->d_handle == (d_handle_t) NULL)) {
    rc = ENODEV;
    GM_WARN(("- d_map_init failed\n"));
    /* Free the device's allocated memory */
    if (xmfree(p_dev, pinned_heap) != 0) {
      GM_WARN(("- xmfree of DDS/TX/RX failed\n"));
    }
    goto abort_with_lock;
  }
#endif /* DO_MAP_PAGE */

  GM_PRINT(GM_AIX_DEBUG,
	   ("lname = %s, bus_mem_addr = 0x%8x, intr_priority = 0x%x\n", 
	    DDS.lname, DDS.bus_mem_addr, DDS.intr_priority));

  /* Initialize the device-specific lock */
  lock_alloc(&(p_dev->wrk.dev_lock), LOCK_ALLOC_PIN, MON_DD_LOCK,
	     GM_AIX_LOCK_INSTANCE_DV);
  simple_lock_init(&(p_dev->wrk.dev_lock));
  GM_PRINT(GM_AIX_DEBUG_LOCKS,
	   ("*** - created device-specific lock at 0x%8x\n",
	    &(p_dev->wrk.dev_lock)));

  /* Configure the adapter */
  rc = cfg_dev(p_dev);
  if (rc) {
    GM_WARN(("- cfg_dev failed 0x%x\n", rc));
    goto abort_with_lock;
  }

  /* Add the dev_ctrl to the component dump table */
  add_cdt("dev_ctrl",(char *)p_dev, WRK.alloc_size);

  /* First get the major device number */
  devsw.d_open = gm_aix_open;
  devsw.d_close = gm_aix_close;
  devsw.d_read = nodev;
  devsw.d_write = nodev;
  devsw.d_ioctl = gm_aix_ioctl;	/* Can't find AIX documentation for this */
  devsw.d_strategy = nodev;
  devsw.d_ttys = NULL;
  devsw.d_select = nodev;
  devsw.d_config = gm_aix_cfg;
  devsw.d_print = nodev;
  devsw.d_dump = nodev;
  devsw.d_mpx = gm_aix_mpx;
  devsw.d_revoke = nodev;
  devsw.d_dsdptr = NULL;
  devsw.d_opts = DEV_MPSAFE;

  GM_PRINT(GM_AIX_DEBUG, ("- devswadd to %d:%d\n", DDS.major, DDS.minor));
  rc = devswadd(makedev (DDS.major, DDS.minor), &devsw);
  if (rc) {
    char *error_name;

    error_name = gm_aix_stringify_error(rc);
    GM_WARN(("- devswadd unsuccessful, error = %s\n", error_name));
    goto abort_with_lock;
  }

  GM_PRINT(GM_AIX_DEBUG, ("- devswadd succsessful\n"));

  /* Set up the interrupt control structure section */
  IHS.next     = (struct intr *) NULL;
  IHS.handler  = gm_aix_intr;
  IHS.bus_type = BUS_BID;		/* Says refer to IHS.bid */
  IHS.flags    = INTR_MPSAFE | INTR_LEVEL;
  IHS.level    = DDS.bus_intr_lvl;
  IHS.priority = DDS.intr_priority;
  IHS.bid = DEV_BID;

  GM_PRINT(GM_AIX_DEBUG,
	   ("- IHS.handler = 0x%X, .bus_type = %d, .level = %d,\n",
	    IHS.handler, IHS.bus_type, IHS.level));
  _GM_PRINT(GM_AIX_DEBUG,
	    ("          .priority = %d, .bid = 0x%X\n",
	     IHS.priority, IHS.bid));

  if (gm_aix_init (p_dev) == 0) {
    GM_WARN(("- gm_aix_init failed\n"));
    rc = 1;	/* Need a real return code! */
    goto abort_with_lock;
  }

  /* done with part of init that needs driver lock */
  gm_aix_unlock_driver();

  /* Initialize IP, must be done without driver lock */
  rc = gm_aix_ip_init (p_dev);
  if (rc) {
    GM_WARN(("- gm_aix_ip_init failed\n"));
  }

  /* Detach I/O if attached */
  if (p_dev->mapped_addr != NULL) {
    iomem_det ((void *) p_dev->mapped_addr);
  }
  return (rc);

 abort_with_lock:
  if (p_dev->mapped_addr != NULL) {
    iomem_det ((void *) p_dev->mapped_addr);
  }
  gm_aix_unlock_driver();


  return(rc);

} /* cfg_init() */

/*****************************************************************************/
/*
 * NAME:                  cfg_term
 *
 * FUNCTION:              process gm_aix_config entry with cmd of TERM
 *
 * EXECUTION ENVIRONMENT: process level
 *
 * NOTES:
 *
 * RETURNS:               none.
 */
/*****************************************************************************/
GM_STATIC int
cfg_term(
	 dds_t *dds
	 )
{
  dev_ctrl_t *p_dev;
  dev_ctrl_t *p_this, *p_prev;	/* Work pointers */
  ushort major, minor;
  int rc = 0;

  GM_PRINT(GM_AIX_DEBUG, ("cfg_term enter\n"));

  /* get the driver lock */
  gm_aix_lock_driver();

  /* Locate the dev_ctrl for this device by searching for this device's
   * sequence number (numeric part of device name, e.g., 0 for ent0)
   * in the dev_ctrl linked list.
   */
  GM_PRINT(GM_AIX_DEBUG,
	   ("- seq_number received %d\n", gm_aix_dds.seq_number));
  for (p_dev = dd_ctrl.p_dev_list; p_dev != NULL; p_dev = p_dev->next) {
    if (p_dev->dds.seq_number == dds->seq_number) {
      GM_PRINT(GM_AIX_DEBUG, ("- found matching p_dev\n"));
      break;
    }
  }

  /* If not inited, error */
  if (!dd_ctrl.initialized || p_dev == NULL) {
    GM_WARN(("- device not initialized\n"));
    rc = ENODEV;
    goto abort_with_lock;
  }

  if (WRK.adap_state != CLOSED_STATE) { /* Still in use by eth emulation */
    GM_WARN(("- IP using unit %d\n", p_dev->unit));
    rc = EBUSY;
    goto abort_with_lock;
  }

  if (p_dev->usage_count > 0) {	/* Still open by somebody */
    GM_WARN(("- unit %d is busy\n", p_dev->unit));
    rc = EBUSY;
    goto abort_with_lock;
  }

  major = DDS.major;
  minor = DDS.minor;

  /* Check early for ability to remove from devsw.  If devswdel fails, we
   * need to keep the driver loaded and configured.
   */
  if (dd_ctrl.num_devs == 1) {
    GM_PRINT(GM_AIX_DEBUG, ("- devswdel to %d:%d\n", major, minor));
    rc = devswdel (makedev (major, minor));
    if (rc) {
      char *error_name;

      error_name = gm_aix_stringify_error(rc);
      GM_WARN(("- devswdel unsuccessful, error = %s\n", error_name));

      rc = EBUSY;	/* return EBUSY to keep driver from getting unloaded */
      goto abort_with_lock;
    }
    else {
      GM_PRINT(GM_AIX_DEBUG, ("- devswdel succsessful\n"));
    }
  }

  /* from this point forward, we are committed to disabling this device.  Any
   * errors we print warnings about, but the device is irrevoocably getting
   * unloaded.
   */

  gm_aix_clean (p_dev);

  GM_PRINT(GM_AIX_DEBUG, ("- gm_aix_clean done\n"));

  /* Disable the card's IO Space / DMA / Timer */

  ucfg_dev (p_dev);

  GM_PRINT(GM_AIX_DEBUG, ("- disabled device\n"));

  /* Remove the device control struct from the list */
  p_this = dd_ctrl.p_dev_list;
  p_prev = NULL;
  while (p_this) {
    if (p_this == p_dev) {
      if (p_prev) {
	p_prev->next = p_this->next;
      } else {
	dd_ctrl.p_dev_list = p_this->next;
      }
      break;
    }

    p_prev = p_this;
    p_this = p_this->next;
  }
  dd_ctrl.num_devs--;

  /* Delete dev from component dump table */
  del_cdt("dev_ctrl",(char *)p_dev, WRK.alloc_size);

  /* Free the device lock */
  lock_free(&(p_dev->wrk.dev_lock));
  GM_PRINT(GM_AIX_DEBUG_LOCKS, ("*** - freed device-specific lock at 0x%8x\n",
				&(p_dev->wrk.dev_lock)));

  GM_PRINT(GM_AIX_DEBUG, ("- freeing memory\n"));

  /* Free the I/O map structure */
  if (xmfree(IO_MAP, pinned_heap) != 0) {
    GM_WARN(("- xmfree for IO_MAP failed\n"));
  }

#if DO_MAP_PAGE
  if (p_dev->d_handle) {
    D_MAP_CLEAR (p_dev->d_handle);
  }
#endif /* DO_MAP_PAGE */


  /* we have removed all references to the device, OK to unlock now */
  gm_aix_unlock_driver();

  /* IP termination must be called without the driver lock */
  gm_aix_ip_term (p_dev);
  GM_PRINT(GM_AIX_DEBUG, ("- gm_aix_ip_term done\n"));

  /* Free the device's allocated memory */
  if (xmfree(p_dev, pinned_heap) != 0) {
    GM_WARN(("- xmfree for p_dev failed\n"));
  }

  return (0);

 abort_with_lock:
  gm_aix_unlock_driver();
  return (rc);

} /* cfg_term() */


/*****************************************************************************/
/*
 * NAME:                  cfg_dev
 *
 * FUNCTION:              Writes the PCI configuration registers of the
 *			  current adapter
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:
 *
 * RETURNS:               0 - OK
 *	                  other errno values
 */
/*****************************************************************************/

GM_STATIC int 
cfg_dev(
	dev_ctrl_t  *p_dev	/* Device control structure ptr */
)
{
  ushort vendor_id;		/* PCI Config reg */
  ushort dev_id;		/* PCI Config reg */
  uchar intr_line;		/* PCI Config reg */
  ushort pci_command;		/* PCI Config reg */
  int rc;

  intr_line = DDS.bus_intr_lvl;

  pci_command = PCI_CMD_MEM_SPACE | PCI_CMD_BUS_MASTER | PCI_CMD_SYSTEM_ERROR;

  GM_PRINT(GM_AIX_DEBUG,
	   ("cfg_dev configuring %s at slot %d\n", DDS.lname, DDS.bus_slot));

  /* If Vendor Id and Device Id are not as expected, quit */
  rc = gm_aix_rw_pci_cfg(DEV_BID, DDS.bus_slot, PCI_VENDOR_ID, &vendor_id,
			 sizeof vendor_id, PCI_CFG_READ);
  rc |= gm_aix_rw_pci_cfg(DEV_BID, DDS.bus_slot, PCI_DEVICE_ID, &dev_id,
			  sizeof dev_id, PCI_CFG_READ);
  if (rc != 0) {
    return rc;
  }

  if (vendor_id != 0x14C1 || (dev_id != 0x8043)) {
    GM_WARN(("Bad vendor (0x%X) or product (0x%X) id\n", vendor_id, dev_id));
    return EINVAL;
  }

  /* Set MEM base addr and interrupt line */
  GM_PRINT(GM_AIX_DEBUG, ("- mem addr = 0x%08X, interrupt line = %d\n",
			  DDS.bus_mem_addr, intr_line));
  rc = gm_aix_rw_pci_cfg(DEV_BID, DDS.bus_slot, PCI_BASE_0,
			 &DDS.bus_mem_addr, sizeof DDS.bus_mem_addr,
			 PCI_CFG_WRITE);
  rc |= gm_aix_rw_pci_cfg(DEV_BID, DDS.bus_slot, PCI_INTR_LINE,
			  &intr_line, sizeof intr_line, PCI_CFG_WRITE);
  if (rc != 0) {
    return rc;
  }

  /* Enable MEM/DMA in command register */
  rc |= gm_aix_rw_pci_cfg(DEV_BID, DDS.bus_slot, PCI_COMMAND,
			  &pci_command, sizeof pci_command, PCI_CFG_WRITE);

  return rc;

} /* cfg_dev() */

/*****************************************************************************/
/*
 * NAME:                  ucfg_dev
 *
 * FUNCTION:              Disables the PCI configuration registers of the
 *			  current adapter
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:
 *
 * RETURNS:               0 - OK
 *	                  other errno values
 */
/*****************************************************************************/

GM_STATIC void 
ucfg_dev(
	 dev_ctrl_t  *p_dev	/* Device control structure ptr */
	 )
{
  ushort pci_command = 0;	/* PCI Config reg */

  /* Disable IO in command register */
  gm_aix_rw_pci_cfg (DEV_BID, DDS.bus_slot, 0x04,
		     &pci_command, sizeof pci_command, PCI_CFG_WRITE);

} /* ucfg_dev() */

/*********************************************************************
 * Init and Cleanup routines called from driver INIT and TERM
 *********************************************************************/

GM_STATIC int
gm_aix_init (
	     dev_ctrl_t *p_dev
	     )
{
  gm_instance_state_t	*is; 
  gm_port_state_t	*ps;
  unsigned char		val;
  int			rc;
  
  GM_PRINT(GM_AIX_DEBUG, ("gm_aix_init(0x%x) entered:\n", p_dev));

  /* Set page len */

  gm_arch_get_page_len (&GM_PAGE_LEN);

  GM_PRINT(GM_AIX_DEBUG, ("- GM_PAGE_LEN = %d\n", GM_PAGE_LEN));

  /* Initialize GM layer */

  gm_init ();

#if GM_AIX_DEBUG_LOCKS && GM_DEBUG
  GM_NOTE (("Offsets in dd_ctrl: cfg_lock 0x%x, i_lock 0x%x, o_lock 0x%x\n",
	    offsetof(dd_ctrl_t, cfg_lock),
	    offsetof(dd_ctrl_t, i_lock),
	    offsetof(dd_ctrl_t, o_lock)));
  GM_NOTE (("Offsets in p_dev: dev_lock 0x%x\n",
	    offsetof(dev_ctrl_t, wrk.dev_lock)));
  GM_NOTE (("Offsets in is: pause_sync 0x%x, page_hash.sync: 0x%x\n",
	    offsetof(gm_instance_state_t, pause_sync),
	    offsetof(gm_instance_state_t, page_hash.sync)));
  GM_NOTE (("Offsets in ps: sync 0x%x, sleep_sync 0x%x, wait_sync: 0x%x\n",
	    offsetof(gm_port_state_t, sync),
	    offsetof(gm_port_state_t, sleep_sync),
	    offsetof(gm_port_state_t, wait_sync)));
#endif

  if (!(is = gm_aix_zalloc (sizeof(*is)))) {
    GM_WARN(("- failed to alloc instance\n"));
    return 0;
  }

  gm_aix_instances [p_dev->unit] = is;
  is->arch.p_dev = p_dev;
  p_dev->is = is;

  if (gm_instance_init (is, p_dev->unit, GM_MYRINET_BUS_PCI) != GM_SUCCESS) {
    GM_WARN(("- gm_instance_init failed\n"));
    goto abort_with_malloc;
  }  

  /* Add our interrupt routine to kernel's interrupt chain */

  IHS.next = NULL;
  if (rc = i_init((struct intr *)&IHS)) {
    GM_WARN(("gm_ax_init: i_init error\n"));
    gm_instance_finalize (is);
    goto abort_with_malloc;
  }

  GM_PRINT(GM_AIX_DEBUG,
	   ("- %s interrupt level set to %d\n", DDS.lname, IHS.level));

  if (gm_initialize_interrupts (is) != GM_SUCCESS) {
    GM_WARN(("- gm_initialize_interrupts failed\n"));
    gm_instance_finalize (is);
    i_clear(&(IHS));
    goto abort_with_malloc;
  }

  return 1;

abort_with_malloc:
  __gm_arch_kernel_free (is);
  p_dev->is = NULL;
  gm_aix_instances [p_dev->unit] = 0;
  return 0;

} /* gm_aix_init () */


GM_STATIC int
gm_aix_clean (
	      dev_ctrl_t *p_dev
	      )
{
  gm_instance_state_t	*is = p_dev->is; 
  ulong	mapped_addr;

  if (!is) {
    GM_WARN(("- nothing to cleanup\n"));
    return (0);
  }

  mapped_addr = (int) iomem_att (IO_MAP);
  GM_PRINT(GM_AIX_DEBUG, ("- mapped_addr 0x%8x\n", mapped_addr));
  if (mapped_addr == p_dev->mapped_addr) {

    gm_finalize_interrupts (is);
    i_clear(&(IHS));
    gm_instance_finalize (is);
  }
  else {
    GM_WARN(("- new mapped addr not same\n"));
    i_clear(&(IHS));
  }

  gm_aix_instances [is->id] = 0;
  __gm_arch_kernel_free (is);
  p_dev->is = NULL;

  if (p_dev->mapped_addr)
    iomem_det ((void *) mapped_addr);

  return 0;

} /* gm_aix_clean () */

/*********************************************************************
 * Driver entry points registered via the devswadd ()
 *********************************************************************/

int 
gm_aix_intr (
	     struct intr *p_ihs	/* At start of dev control structure */
	     )
{
  dev_ctrl_t *p_dev = (dev_ctrl_t *)p_ihs; /* Dev control structure ptr */
  int mapped_addr;

  GM_PRINT(GM_AIX_DEBUG, ("- got an interrupt - p_dev 0x%8x\n", p_dev));
  _GM_PRINT(GM_AIX_DEBUG, ("- is 0x%8x\n", p_dev->is));

  mapped_addr = (int) iomem_att (IO_MAP);
  GM_PRINT(GM_AIX_DEBUG, ("- mapped_addr 0x%8x\n", mapped_addr));
  if (mapped_addr != p_dev->mapped_addr) {
    GM_WARN(("- new mapped addr not same\n"));
    iomem_det ((void *) mapped_addr);
    return (INTR_FAIL);
  }

  while (gm_interrupting (p_dev->is) != GM_ARCH_INTR_CLAIMED) {
    GM_PRINT(GM_AIX_DEBUG_INTR, 
	("Board %d: spurrious irq. host: %d, lanai %d\n",
	p_dev->is->id, gm_ntoh_u8(p_dev->is->interrupt->last_byte.ready),
	gm_read_lanai_global_u8(p_dev->is, interrupt.last_byte.ready)));
    iomem_det ((void *) mapped_addr);
    return (INTR_FAIL);
  }
  
  GM_PRINT(GM_AIX_DEBUG, ("- gm is interrupting\n"));

  gm_handle_claimed_interrupt (p_dev->is);

  p_ihs->i_count++;

  iomem_det ((void *)mapped_addr);

  return INTR_SUCC;

} /* gm_aix_intr () */


/*
 * Functions to serialize access to driver-global data
 */

void
gm_aix_lock_driver ( void )
{
  GM_PRINT(GM_AIX_DEBUG, ("*** locking driver\n"));
  GM_AIX_LOCK_ENTER(&(dd_ctrl.o_lock));

} /* gm_aix_lock_driver() */


void
gm_aix_unlock_driver ( void )
{
  GM_PRINT(GM_AIX_DEBUG, ("*** unlocking driver\n"));
  GM_AIX_LOCK_EXIT(&(dd_ctrl.o_lock));

} /* gm_aix_unlock_driver() */



/*****************************************************************************/
/*
 * NAME:     gm_aix_mpx (generically, "ddmpx")
 *
 * FUNCTION: allocate/deallocate AIX "channel", corresponding to GM minor num.
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:    Dual-use function.  If parameter 'channame' is non-null, the
 *           call is for channel allocation; otherwise, the call is for
 *           de-allocation.
 *
 * PARAMETERS for allocation:
 *	devno	 IN	- device number, comprising major and minor parts.
 *	chanp	 OUT	- channel number assigned by this function.
 *	channame IN	- points to (possibly null) character string containing
 *			  the path name extension.  Not used by GM.
 *
 * PARAMETERS for deallocation:
 *	devno	 IN	- device number, comprising major and minor parts.
 *	chanp	 IN	- channel number to be deallocated.
 *	channame IN	- null pointer.
 *
 * RETURNS:             0 ==> OK
 *	                errno value if error
 */
/*****************************************************************************/
int
gm_aix_mpx (
	    dev_t devno,
	    chan_t *chanp,
	    char *channame
	    )
{
  int rc = 0;
  int minor;
  dev_ctrl_t *p_dev;		/* Device control struct pointer */
  gm_status_t status;
  ulong	mapped_addr;

  gm_aix_lock_driver();		/* Serialize access to driver-global data */

  GM_PRINT(GM_AIX_DEBUG,
	   ("gm_aix_mpx devno 0x%x, channame %s\n", devno,
	    ((channame) ?
	     (strlen(channame) ? channame : "<null>") : "[void]")));

  p_dev = gm_aix_device (devno);
  if (p_dev == NULL) {
    GM_WARN(("- bad devno 0x%x\n", devno));
    rc = EINVAL;
    goto abort_with_lock;
  }

  if (p_dev->is == NULL) {
    GM_WARN(("- IS not set\n"));
    rc = EINVAL;
    goto abort_with_lock;
  }

  /* Just a sanity check */

  if (p_dev->unit != (minor (devno) / 2)) {
    GM_WARN(("- unit numbers don't match %d %d\n",
	     p_dev->unit, (minor (devno) / 2)));
    rc = EINVAL;
    goto abort_with_lock;
  }

  if (channame) {

    /* Need to allocate a new channel */

    status = gm_minor_alloc (p_dev->is, &minor);
    if (status != GM_SUCCESS) {
      GM_WARN(("- gm_minor_alloc failed 0x%x!\n", status));
      rc = EINVAL;
      goto abort_with_lock;
    }
    GM_PRINT(GM_AIX_DEBUG, ("- allocated minor number %d\n", minor));

    p_dev->usage_count++;

    *chanp = (chan_t) minor;
    rc = 0;
  }
  else {

    /* Need to de-allocate channel */

    p_dev->usage_count--;

    minor = (int) *chanp;
    GM_PRINT(GM_AIX_DEBUG, ("- freeing %d minor!\n", minor));
    mapped_addr = (int) iomem_att (IO_MAP);
    GM_PRINT(GM_AIX_DEBUG, ("- mapped_addr 0x%8x\n", mapped_addr));
    if (mapped_addr != p_dev->mapped_addr) {
      GM_WARN(("- mapped addr differ 0x%8x!\n", p_dev->mapped_addr));
      iomem_det ((void *) mapped_addr);
      rc = 0;
      goto abort_with_lock;
    }
    gm_minor_free (minor);

    iomem_det ((void *) mapped_addr);
  }

abort_with_lock:
  gm_aix_unlock_driver();

  GM_PRINT(GM_AIX_DEBUG, ("- devno 0x%x done, rc = %d!\n", devno,rc));

  return (rc);

} /* gm_aix_mpx () */

int
gm_aix_open (
	     dev_t devno,
	     int devflag,
	     int chan,
	     int ext
	     )
{
  dev_ctrl_t *p_dev;		/* Device control struct pointer */
  gm_port_state_t *ps;

  gm_aix_lock_driver();		/* Serialize access to driver-global data */

  GM_PRINT(GM_AIX_DEBUG, ("open devno 0x%x chan %d\n", devno, chan));

  p_dev = gm_aix_device (devno);
  if (p_dev == NULL) {
    GM_WARN(("- bad devno 0x%x\n", devno));
    gm_aix_unlock_driver();
    return EINVAL;
  }

  if (p_dev->is == NULL) {
    GM_WARN(("- IS not set\n"));
    gm_aix_unlock_driver();
    return (EINVAL);
  }

  /* Just a sanity check */

  if (p_dev->unit != (minor (devno) / 2)) {
    GM_WARN(("- units don't match %d %d\n", p_dev->unit, (minor (devno) / 2)));
    gm_aix_unlock_driver();
    return (EINVAL);
  }

  if ((ps = gm_minor_get_port_state (chan)) == NULL) {
    GM_WARN(("- could not get ps\n"));
    gm_aix_unlock_driver();
    return (EINVAL);
  }

  GM_PRINT(GM_AIX_DEBUG, ("- ps 0x%8x\n", ps));

  p_dev->usage_count++;

  /* The device is privileged if the device number is odd     */
  ps->privileged = minor (devno) & 1;

  gm_aix_unlock_driver();

  GM_PRINT(GM_AIX_DEBUG, ("- devno 0x%x chan %d returning 0\n", devno, chan));
  _GM_PRINT(GM_AIX_DEBUG, ("- usage count = %d\n", p_dev->usage_count));

  return (0);

} /* gm_aix_open () */


int
gm_aix_close (
	      dev_t devno,
	      int chan,
	      int ext
	      )
{
  dev_ctrl_t *p_dev;		/* Device control struct pointer */
  gm_port_state_t *ps;
  ushort minor;

  gm_aix_lock_driver();		/* Serialize access to driver-global data */

  GM_PRINT(GM_AIX_DEBUG, ("close devno 0x%x chan %d\n", devno, chan));

  p_dev = gm_aix_device (devno);
  if (p_dev == NULL) {
    GM_WARN(("- bad devno 0x%x\n", devno));
    gm_aix_unlock_driver();
    return EINVAL;
  }

  minor = (ushort) chan;

  GM_PRINT(GM_AIX_DEBUG, ("gm_aix_close with minor number %d\n", minor));

  if (minor > GM_ARCH_MAX_MINOR) {
    GM_WARN(("- bad minor 0x%x!\n", minor));
    gm_aix_unlock_driver();
    return ENODEV;
  }
    
  if (!(ps = gm_minor_get_port_state (minor))) {
    GM_WARN(("- couldn't get port state for minor %d\n", minor));
    gm_aix_unlock_driver();
    return ENODEV;
  }

  GM_PRINT(GM_AIX_DEBUG, ("- ps 0x%8x\n", ps));

  p_dev->usage_count--;

  /* There's nothing to do here, no need to lock the device */

  /* Actual free operation will be done in the mpx routine */

  GM_PRINT(GM_AIX_DEBUG, ("- devno 0x%x chan %d LEAVING\n", devno, chan));
  _GM_PRINT(GM_AIX_DEBUG, ("- usage count = %d\n", p_dev->usage_count));

  gm_aix_unlock_driver();

  return (0);

} /* gm_aix_close () */


/*
 *  WARNING!  I can't find documentation of this interface in the AIX 4.3 or
 *            5.1 docs.  There's a problem with the 'arg' parameter.  We want
 *            to coerce it to be a pointer, but that assumes that an int
 *            and a void * are the same size.
 */
int
gm_aix_ioctl (
	      dev_t devno,
	      int op,
	      int arg,
	      ulong dev_flag,
	      int chan,
	      int ext
	      )
{
  dev_ctrl_t *p_dev;		/* Device control struct pointer */
  int func_ret, minor_id;
  gm_instance_state_t *is;
  gm_port_state_t *cps, *ps;
  ulong mapped_addr;

  gm_always_assert (sizeof(arg) == sizeof(void *)); /* See WARNING above */

  gm_aix_lock_driver();		/* Serialize access to driver-global data */

  GM_PRINT(GM_AIX_DEBUG, 
	   ("ioctl devno 0x%x, op 0x%x, arg 0x%x\n", devno, op, arg));
  _GM_PRINT(GM_AIX_DEBUG, 
	    ("      dev_flag 0x%x, chan %d, ext %d\n", dev_flag, chan, ext));

  p_dev = gm_aix_device (devno);
  if (p_dev == NULL) {
    GM_WARN(("- bad devno 0x%x\n", devno));
    gm_aix_unlock_driver();
    return EINVAL;
  }

  if (!(ps = gm_minor_get_port_state (chan))) {
    GM_WARN(("- can't get ps for chan %d\n", chan));
    gm_aix_unlock_driver();
    return ENXIO;
  }

  mapped_addr = (ulong) iomem_att (IO_MAP);
  GM_PRINT(GM_AIX_DEBUG, ("- mapped_addr 0x%8x\n", mapped_addr));
  if (mapped_addr != p_dev->mapped_addr) {
    GM_WARN(("- invalid mapped addr: 0x%x != 0x%x\n",
	     mapped_addr, p_dev->mapped_addr));
    func_ret = GM_UNSUPPORTED_DEVICE;
    iomem_det ((void *) mapped_addr);
    gm_aix_unlock_driver();
    goto abort;
  }

  GM_PRINT(GM_AIX_DEBUG, ("- ps 0x%8x opened %d port_num %d\n",
			  ps, ps->opened, ps->id));
  _GM_PRINT(GM_AIX_DEBUG, ("- op = 0x%x\n", op));

  /* We enter this switch statement holding the device lock;
   * each case should detach from I/O memory and release the device lock
   * according to its own needs, but before the end of each case.
   */
  switch (op)
    {
    case GM_MMAP:
      func_ret = gm_aix_mmap (p_dev, ps, (void *)arg);

      iomem_det ((void *) mapped_addr);
      gm_aix_unlock_driver();
      break;

    case GM_SLEEP:
      /* acquire this just because gm_kernel_sleep unlocks it... */
      gm_arch_mutex_enter(&ps->sync);
      
      /* Do not sleep with the device locked */
      iomem_det ((void *) mapped_addr);
      gm_aix_unlock_driver();

      /* perform the ioctl */
      func_ret = gm_ioctl (ps, op, (char*) arg,
			   INT_MAX, (char*)arg, INT_MAX, 0);

      /* now release the port lock */
      gm_arch_mutex_exit(&ps->sync);
      break;

    default:
      /* lock the port for ioctls */
      gm_arch_mutex_enter(&ps->sync);

      func_ret = gm_ioctl (ps, op, (char*) arg,
			   INT_MAX, (char*)arg, INT_MAX, 0);

      /* now release the port lock */
      gm_arch_mutex_exit(&ps->sync);

      /* detach from memory and release the lock */
      iomem_det ((void *) mapped_addr);
      gm_aix_unlock_driver();
      break;
    }

  GM_PRINT(GM_AIX_DEBUG,
	   ("- devno 0x%x, op = 0x%x, chan %d LEAVING\n", devno, op, chan));

abort:
  GM_PRINT(GM_AIX_DEBUG, ("- returning with status 0x%x\n", func_ret));

  return (gm_aix_localize_status (func_ret));

} /* gm_aix_ioctl () */


int
gm_aix_cfg (
	    dev_t devno,
	    int op,
	    struct uio *uiop
	    )
{
  GM_PRINT(GM_AIX_DEBUG, ("cfg devno 0x%x, op = 0x%x\n", devno, op));
  return (0);

} /* gm_aix_cfg () */


GM_STATIC int
gm_aix_mmap (
	     dev_ctrl_t *p_dev,
	     gm_port_state_t *ps,
	     void *arg
	     )
{
  gm_mmap_info_t mmap;
  struct io_map	io_map;
  int rc = 0;
  int flags;

  GM_PRINT(GM_AIX_DEBUG, ("gm_aix_mmap ...\n"));
  if (copyin (arg, &mmap, sizeof (gm_mmap_info_t)) != 0) {
    GM_WARN(("- error in copying-in data\n"));
    return (ENXIO);
  }

  GM_PRINT(GM_AIX_DEBUG, ("- offset 0x%x\n", mmap.offset));
  _GM_PRINT(GM_AIX_DEBUG, ("- len 0x%x\n", mmap.len));
  _GM_PRINT(GM_AIX_DEBUG, ("- perms 0x%x\n", mmap.requested_permissions));
  _GM_PRINT(GM_AIX_DEBUG, ("- va 0x%8x\n", mmap.va));

  if (mmap.va != (void *) -1) {
    //
    // Destroy the mapping
    //
    rc = rmmap_remove (&mmap.va);
    GM_PRINT(GM_AIX_DEBUG, ("- status 0x%x\n", rc));
    return (rc);
  }

  //
  // Create a mapping
  //
  io_map.key = IO_MEM_MAP;
  io_map.flags = 0;
  io_map.size = mmap.len;
  io_map.bid = DEV_BID;
  io_map.busaddr = DDS.bus_mem_addr;

  flags = RMMAP_PAGE_I | RMMAP_PAGE_M | RMMAP_PAGE_G | RMMAP_PRELOAD;
  switch (mmap.requested_permissions) {
  case GM_MAP_READ:
    flags |= RMMAP_RDONLY;
    break;

  case GM_MAP_WRITE:
  case GM_MAP_RDWR:
    flags |= RMMAP_RDWR;
    break;

  default:
    GM_WARN(("- invalid requested permissions 0x%x\n",
	     mmap.requested_permissions));
    return (ENXIO);
  }

  mmap.va = 0;
  //
  // Validate mapping
  //
  rc = gm_prepare_to_mmap (ps, mmap.offset,
			   mmap.len, mmap.requested_permissions);
  if (rc != GM_SUCCESS) {
    GM_WARN(("- prepare mmap failed 0x%x\n",rc));
    return (ENXIO);
  }

  //
  // Map the offset into valid address in adapter memory. This
  // routine will return the ptr in the kernel mapped memory
  //
  rc = gm_mmap (ps, mmap.offset, (void **) &mmap.va);
  if (rc != GM_SUCCESS) {
    GM_WARN(("- gm_mmap failed 0x%x\n", rc));
    return (ENXIO);
  }
  GM_PRINT(GM_AIX_DEBUG, ("- mmaped va 0x%8x\n", mmap.va));
  _GM_PRINT(GM_AIX_DEBUG, ("- flags 0x%x\n", flags));

  //
  // Xlate the address into offset from start of adapter memory
  //
  io_map.busaddr += ((long long)mmap.va - (p_dev->mapped_addr));
  GM_PRINT(GM_AIX_DEBUG, ("- new busaddr 0x%llx\n", io_map.busaddr));

  mmap.va = 0;
  rc = rmmap_create (&mmap.va, &io_map, flags);
  GM_PRINT(GM_AIX_DEBUG, ("- new va 0x%8x\n", mmap.va));
  if (copyout (&mmap, arg, sizeof (gm_mmap_info_t)) != 0) {
    GM_WARN(("- error in copying-out data\n"));
    if (!rc) {
      rmmap_remove (&mmap.va);
    }
    return (ENXIO);
  }

  GM_PRINT(GM_AIX_DEBUG, ("- returning with status 0x%x\n", rc));

  return (rc);

} /* gm_aix_mmap () */



/**********************************************************************
 * GM OS-abstraction required functions.
 *
 * Description:
 *
 * These are the functions that the platform-independent part of the
 * GM driver uses.  The interface and semantics of these functions are
 * predetermined, but the implementation is platform-dependent.
 *
 * E.g., gm_arch_read_pci_config_32() is declared in gm_impl.h, and it
 * is called by gm_instance.c:gm_pci_board_span() (but only in driver
 * versions that -- like this one -- support the PCI bus).
 *
 * NOTE: Functions below with names like gm_aix_* are local to the AIX-
 * specific part of the driver.  Functions with names like gm_arch_* are
 * exported to the platform_independent part of the driver.
 *
 **********************************************************************/


/* Convert GM specific error code to AIX's closest equivalent.
 */

GM_STATIC int
gm_aix_localize_status (
			gm_status_t status
			)
{
#define CASE(from,to) case from : return to

  switch (status) {
    CASE(GM_SUCCESS, 0);
    CASE(GM_INPUT_BUFFER_TOO_SMALL, EFAULT);
    CASE(GM_OUTPUT_BUFFER_TOO_SMALL, EFAULT);
    CASE(GM_TRY_AGAIN, EAGAIN);
    CASE(GM_BUSY, EBUSY);
    CASE(GM_MEMORY_FAULT, EFAULT);
    CASE(GM_INTERRUPTED, EINTR);
    CASE(GM_INVALID_PARAMETER, EINVAL);
    CASE(GM_OUT_OF_MEMORY, ENOMEM);
    CASE(GM_INVALID_COMMAND, EINVAL);
    CASE(GM_PERMISSION_DENIED, EPERM);
    CASE(GM_INTERNAL_ERROR, EPFNOSUPPORT);
    CASE(GM_UNATTACHED, -1);	/* ??? */
    CASE(GM_UNSUPPORTED_DEVICE, ENXIO);

  default:
    return gm_aix_localize_status (GM_INTERNAL_ERROR);
	}
} /* gm_aix_localize_status () */



/**********************************************************************
 * Kernel memory management
 **********************************************************************/

/************
 * Kernel Memory Allocation for GM's use
 ************/

void *
__gm_arch_kernel_malloc (
			 unsigned long len,
			 int flags
			 )
{
  void *p;

  p = gm_aix_malloc (len, 3);

  if (!p) {
    GM_WARN(("- failed for len 0x%x\n", len));
    return (NULL);
  }

  GM_PRINT(GM_AIX_DEBUG_MALLOC,
	   ("- called with length %d, result was 0x%x\n", len, p));
  return p;

} /* __gm_arch_kernel_malloc () */


/* __gm_arch_kernel_free()   (free() analogue for the driver)
 */

void
__gm_arch_kernel_free (
		       void *ptr
		       )
{
  int rc = 0;
  int len = 0;

  GM_PRINT(GM_AIX_DEBUG_MALLOC,
	   ("__gm_arch_kernel_free called with addr 0x%8x\n", ptr));
  gm_assert (dd_ctrl.my_heap);
  xmfree (ptr, dd_ctrl.my_heap);

} /* __gm_arch_kernel_free () */



/************
 * GM synchronization functions
 ************/

void
gm_arch_sync_init (
		   gm_arch_sync_t *s,
		   gm_instance_state_t *is
		   )
{
  dev_ctrl_t *p_dev;

  GM_PRINT(GM_AIX_DEBUG_LOCKS, ("init s = 0x%8x is = 0x%8x\n", s, is));

  if (!s) {
    GM_WARN(("- bad sync ptr\n"));
    return;
  }

  if (is) {
    p_dev = is->arch.p_dev;
    s->intr_priority = DDS.intr_priority;
  }
  else {
    s->intr_priority = INTCLASS2;
  }
  s->event_word = EVENT_NULL;
  s->p_dev = p_dev;
  lock_alloc(&(s->sync_lock), LOCK_ALLOC_PIN, MON_DD_LOCK,
	     GM_AIX_LOCK_INSTANCE_SYNC_BASE + sync_count++);
  simple_lock_init(&(s->sync_lock));
  lock_alloc(&(s->wakeup_lock), LOCK_ALLOC_PIN, MON_DD_LOCK,
	     GM_AIX_LOCK_INSTANCE_SYNC_BASE + sync_count++);
  simple_lock_init(&(s->wakeup_lock));
  /* The wakeup lock will remain held until we sleep on this sync */
  simple_lock(&(s->wakeup_lock));
  GM_PRINT(GM_AIX_DEBUG_LOCKS,
	   ("*** created lock at 0x%8x\n", &(s->sync_lock)));

} /* gm_arch_sync_init () */


void
gm_arch_mutex_enter (
		     gm_arch_sync_t *s
		     )
{
  GM_PRINT(GM_AIX_DEBUG_LOCKS, ("enter 0x%8x\n", s));
  // s->old_pri = disable_lock (s->intr_priority, &(s->sync_lock));
  GM_AIX_LOCK_ENTER(&(s->sync_lock));

} /* gm_arch_mutex_enter () */

void
gm_arch_mutex_exit (
		    gm_arch_sync_t *s
		    )
{
  GM_PRINT(GM_AIX_DEBUG_LOCKS, ("exit 0x%8x\n", s));
  // unlock_enable (s->old_pri, &(s->sync_lock));
  GM_AIX_LOCK_EXIT(&(s->sync_lock));

} /* gm_arch_mutex_exit () */

void
gm_arch_sync_destroy (
		      gm_arch_sync_t *s
		      )
{
  GM_PRINT(GM_AIX_DEBUG_LOCKS, ("destroy 0x%8x\n", s));
  lock_free (&(s->sync_lock));
  GM_PRINT(GM_AIX_DEBUG_LOCKS,
	   ("*** deleted lock at 0x%8x\n", &(s->sync_lock)));

} /* gm_arch_sync_destroy () */

void
gm_arch_sync_reset (
		    gm_arch_sync_t *s
		    )
{
  GM_PRINT(GM_AIX_DEBUG_LOCKS, ("reset 0x%8x\n", s));
  simple_lock_init(&(s->sync_lock));
  GM_PRINT(GM_AIX_DEBUG_LOCKS, ("*** reset lock at 0x%8x\n", &(s->sync_lock)));
  return;

} /* gm_arch_sync_reset () */



/* Sleep/wake functions. */

void
gm_arch_wake (
	      gm_arch_sync_t * s
	      )
{
  volatile int prior_cnt;

  prior_cnt = fetch_and_add ((int *)&(s->wake_cnt), 1);
  GM_PRINT(GM_AIX_DEBUG, ("- new value of wake_cnt is %d\n", prior_cnt + 1));
  /* Only wake up a thread which is sleeping; AIX kernel might get upset */
  if (simple_lock_try(&(s->wakeup_lock)))
    {
      /* We were able to acquire the lock, therefore the thread is sleeping */
      GM_PRINT(GM_AIX_DEBUG, ("Calling e_wakeup_one()\n"));
      e_wakeup_one (&s->event_word);
      simple_unlock(&(s->wakeup_lock));
    }

} /* gm_arch_wake () */

gm_arch_sleep_status_t
gm_arch_signal_sleep (
		      gm_arch_sync_t *s
		      )
{
  int rc_from_os, rc_to_caller;

  GM_PRINT(GM_AIX_DEBUG, ("sync at 0x%8x\n", s));

  rc_to_caller = GM_SLEEP_WOKE; /* In case wake has already happened */
  while (s->wake_cnt <= 0)	/* While "wake" has not arrived */
    {
      GM_PRINT(GM_AIX_DEBUG, ("gm_arch_signal_sleep\n"));

      /* Put thread on wait queue (interruptible) */
      e_assert_wait (&s->event_word, TRUE);

      /* unlock the wakeup lock, as it is now safe to wake the thread */
      simple_unlock (&s->wakeup_lock);

      /* check wake_cnt again, an interrupt can occur after the above
	 check, but before we are on the wait queue with the
	 wakeup_lock unlocked 
	*/

      if (s->wake_cnt > 0) 
	{
	  GM_PRINT(GM_AIX_DEBUG, ("Not sleeping\n"));
	  /* lock the wakeup_lock & remove ourselves from the wait queue */
	  simple_lock (&s->wakeup_lock);
	  e_clear_wait (thread_self (), THREAD_AWAKENED);
	  break;		/* Exit the 'while' */
	}

      GM_PRINT(GM_AIX_DEBUG, ("Sleeping\n"));

      /* Ready to sleep until awakened or interrupted */
      rc_from_os = e_block_thread ();

      /* re-lock the wakeup_lock, we are no longer on the wait queue */
      simple_lock (&s->wakeup_lock);
      GM_PRINT(GM_AIX_DEBUG, ("- woke up thread \n"));
      if (rc_from_os != THREAD_AWAKENED)
	{
	  rc_to_caller = GM_SLEEP_INTERRUPTED;
	  goto return_to_caller;
	}
      break;			/* Exit the 'while' */

    } /* while */


 return_to_caller:

  if (rc_to_caller == GM_SLEEP_WOKE)
    {
      /* only decrement wake_cnt if gm_arch_wake incremented it */
      fetch_and_add ((int *)&(s->wake_cnt), -1);
    }
  GM_PRINT(GM_AIX_DEBUG, ("- rc 0x%x\n", rc_to_caller));
  return (rc_to_caller);

} /* gm_arch_signal_sleep () */



void
gm_aix_timed_out(struct trb *timer_rb)
{
  GM_PRINT(GM_AIX_DEBUG, 
	   ("Entered with TRB at 0x%x\n", timer_rb));

  e_clear_wait(timer_rb->func_data, THREAD_TIMED_OUT);
}


gm_arch_sleep_status_t
gm_arch_timed_sleep (
		     gm_arch_sync_t *s,
		     int seconds
		     )
{
  int rc_from_os, rc_to_caller;
  struct timestruc_t ticks;
  struct trb *timer_rb = NULL;

  GM_PRINT(GM_AIX_DEBUG, ("timed_sleep 0x%8x secs %d\n", s, seconds));

  timer_rb = talloc();		/* alloc timeout struct */

  rc_to_caller = GM_SLEEP_WOKE; /* In case wake has already happened */
  while (s->wake_cnt <= 0)	/* While "wake" has not arrived */
    {
      gm_always_assert(timer_rb != NULL);

      GM_PRINT(GM_AIX_DEBUG, ("event_word = 0x%x\n", s->event_word));

      /* Put thread on wait queue (UNinterruptible) */
      e_assert_wait (&s->event_word, FALSE);

      /* unlock the wakeup lock, as it is now safe to wake the thread */
      simple_unlock (&s->wakeup_lock);


      /* check wake_cnt again, an interrupt can occur after the above
	 check, but before we are on the wait queue with the
	 wakeup_lock unlocked 
	*/
      if (s->wake_cnt > 0)
	{
	  GM_PRINT(GM_AIX_DEBUG, ("Not sleeping\n"));
	  /* lock the wakeup_lock & remove ourselves from the wait queue */
	  simple_lock (&s->wakeup_lock);
	  e_clear_wait (thread_self (), THREAD_AWAKENED);
	  break;		/* Exit the 'while' */
	}

      timer_rb->func = gm_aix_timed_out; /* fill in handler */
      timer_rb->eventlist = EVENT_NULL;
      timer_rb->ipri = INTTIMER;
      timer_rb->t_func_addr = (char *)thread_self();
      ticks.tv_sec = seconds;		/* max time to wait */
      ticks.tv_nsec = 0;
      timer_rb->timeout.it_value = ticks;	
      timer_rb->flags = T_INCINTERVAL |	/* Timeout is relative, */
			T_LOWRES;	/* and 10ms resolution is adequate */
      GM_PRINT(GM_AIX_DEBUG, ("- starting timer\n"));
      tstart(timer_rb);			/* Start the timer */

      /* Ready to sleep until awakened */
      rc_from_os = e_block_thread ();

      /* re-lock the wakeup_lock, we are no longer on the wait queue */
      simple_lock (&s->wakeup_lock);

      while(tstop(timer_rb)) {		/* cancel timer */
	/* AIX's tstop() documentation sez we should "release_then_reacquire"
	   our DD lock here.  Is the following what they intend? */
#if 0   /* I don't think this is required, I think the comment was aimed
	 at drivers with a global lock */
	if (GM_AIX_LOCK_HELD(&(s->sync_lock))) {
	  GM_AIX_LOCK_EXIT(&(s->sync_lock));
	  GM_AIX_LOCK_ENTER(&(s->sync_lock));
	}
#endif
      }

      GM_PRINT(GM_AIX_DEBUG,
	       ("- woke up %d\n"));

      switch (rc_from_os) {
      case THREAD_INTERRUPTED:
	gm_always_assert(0);	/* should never get here */

	rc_to_caller = GM_SLEEP_INTERRUPTED;
	/* We need to break out of the case _AND_ the while */
	goto return_to_caller;
	break;

      case THREAD_TIMED_OUT:
	rc_to_caller = GM_SLEEP_TIMED_OUT;
	/* We need to break out of the case _AND_ the while */
	goto return_to_caller;
	break;

      default:
      case THREAD_OTHER:
      case THREAD_AWAKENED:
	rc_to_caller = GM_SLEEP_WOKE;
	break;
      }
    } /* while */

 return_to_caller:

  tfree(timer_rb);			/* free timer struct */

  if (rc_to_caller == GM_SLEEP_WOKE)
    {
      /* only decrement wake_cnt if gm_arch_wake incremented it */
      fetch_and_add ((int *)&(s->wake_cnt), -1);
    }

  GM_PRINT(GM_AIX_DEBUG, ("- rc 0x%x\n", rc_to_caller));
  return (rc_to_caller);

} /* gm_arch_timed_sleep () */


/**********************************************************************
 * DMA region manipulation
 **********************************************************************/

#if DO_MAP_PAGE
int
gm_aix_unmap_pages (
		    gm_arch_dma_region_t *r
		    )
{
  int i;

  for (i = 0; i < r->pages; i++) {
    D_UNMAP_PAGE (r->d_handle, &(r->busaddr [i]));
  }
  r->pages = 0;

} /* gm_aix_unmap_pages () */
#endif	/* DO_MAP_PAGE */


/* Allocate LEN bytes of DMA memory for instance 'is', recording all necessary
 * information in structure 'r'.
 *
 * If register_page_func is non-null, call it with (arg, dma address, page num)
 * for each page of the region we have allocated.
 */

gm_status_t
gm_arch_dma_region_alloc (
			  gm_instance_state_t * is, 
			  gm_arch_dma_region_t * r,
			  gm_size_t len,
			  gm_u32_t flags,
			  gm_register_page_function_t register_page_function,
			  void *user_arg
			  )
{
  dev_ctrl_t *p_dev = is->arch.p_dev;
  int i, num_pages;
  int rc;

  GM_PRINT(GM_AIX_DEBUG,
	   ("Allocating region defined at 0x%8x len %d\n", r, len));

  num_pages = (len + GM_PAGE_LEN - 1) / GM_PAGE_LEN;

#if DO_MAP_PAGE
  if (num_pages > MAX_MAPPED_PAGES) {
    GM_WARN(("- too many pages %d\n", num_pages));
    return GM_FAILURE;
  }
  r->length = len;
#endif /* DO_MAP_PAGE */

  r->vaddr = gm_aix_malloc (len, PGSHIFT);
  if (r->vaddr == NULL) {
    GM_WARN(("- malloc failed\n"));
    return GM_FAILURE;
  }

  GM_PRINT(GM_AIX_DEBUG, ("- pages %d\n", num_pages));

#if DO_MAP_PAGE
  r->d_handle = d_map_init (BID_ALTREG (DDS.bus_id, PCI_BUSMEM),
			    DMA_MASTER | DMA_MAXMIN_2M |       \
			    DMA_ADDRESS_64_BIT|DMA_ENABLE_64,
			    0, 0);
  if ((r->d_handle == (d_handle_t) DMA_FAIL) ||
      (r->d_handle == (d_handle_t) NULL)) {
    r->d_handle = NULL;
    GM_WARN(("- bad d_handle\n"));
    xmfree (r->vaddr, dd_ctrl.my_heap);
    return (GM_FAILURE);
  }

  rc = D_MAP_PAGE (r->d_handle, DMA_BYPASS | DMA_READ, 
		   (unsigned long long) r->vaddr,
		   &(r->paddr), xmem_global);
  if (rc != DMA_SUCC) {
    GM_WARN(("- d_map_page failed 0x%x\n", rc));
    xmfree (r->vaddr, dd_ctrl.my_heap);
    return (GM_FAILURE);
  }

#else /* DO_MAP_PAGE */
  r->paddr = (char *) xmemdma (xmem_global, r->vaddr,
			       XMEM_HIDE | XMEM_ACC_CHK);
  r->paddr = (char *) ((ulong) r->paddr | 0x80000000);
#endif /* DO_MAP_PAGE */

  GM_PRINT(GM_AIX_DEBUG, ("- paddr 0x%8x\n", r->paddr));

  r->vaddr_advance = r->vaddr;

  if (register_page_function) {
    GM_PRINT(GM_AIX_DEBUG, ("- registering pages\n"));
#if DO_MAP_PAGE
    r->pages = 0;
#endif /* DO_MAP_PAGE */
    for (i = 0; i < num_pages; i++) {
      gm_dp_t	p;
#if DO_MAP_PAGE
      rc = D_MAP_PAGE (r->d_handle, DMA_BYPASS | DMA_READ,
		       (unsigned long long)(r->vaddr + GM_PAGE_LEN * i),
		       &(r->busaddr [r->pages]), xmem_global);
      if (rc != DMA_SUCC) {
	GM_WARN(("- d_ma_page failed 0x%x\n", rc));
	gm_aix_unmap_pages (r);
	xmfree (r->vaddr, dd_ctrl.my_heap);
	return (GM_FAILURE);
      }

      p = (gm_dp_t) r->busaddr [r->pages];
      r->pages++;
#else /* DO_MAP_PAGE */
      p = (gm_dp_t) xmemdma (xmem_global,
			     (r->vaddr + GM_PAGE_LEN * i),
			     XMEM_HIDE|XMEM_ACC_CHK);
      p |= 0x80000000;
#endif /* DO_MAP_PAGE */
      register_page_function (user_arg, p, i);
    }
  }
  else {
#if DO_MAP_PAGE
    r->pages = 1;
    r->busaddr [0] = r->paddr;
    GM_PRINT(GM_AIX_DEBUG, ("- not registering pages %d\n", num_pages));
#endif /* DO_MAP_PAGE */
  }
  return GM_SUCCESS;

} /* gm_arch_dma_region_alloc () */


void
gm_arch_dma_region_free (
			 gm_arch_dma_region_t * r
			 )
{
  GM_PRINT(GM_AIX_DEBUG, ("Freeing region defined at 0x%8x\n", r));
#if DO_MAP_PAGE
  gm_aix_unmap_pages (r);
#endif /* DO_MAP_PAGE */
  gm_assert (dd_ctrl.my_heap);
  xmfree (r->vaddr, dd_ctrl.my_heap);
} /* gm_arch_dma_region_free () */


gm_dp_t
gm_arch_dma_region_dma_addr (
			     gm_arch_dma_region_t * r
			     )
{
  gm_dp_t p;

  GM_PRINT(GM_AIX_DEBUG, ("gm_arch_dma_region_dma_addr ...\n"));
  p = (gm_dp_t) r->paddr;
  return p;

} /* gm_arch_dma_region_dma_addr () */

gm_dp_t
gm_arch_dma_region_dma_addr_advance (
				     gm_arch_dma_region_t * r
				     )
{
  gm_dp_t p;
  int rc;

  GM_PRINT(GM_AIX_DEBUG, ("gm_arch_dma_region_dma_addr_advance ...\n"));

#if DO_MAP_PAGE
  rc = D_MAP_PAGE (r->d_handle, DMA_BYPASS | DMA_READ, 
		   (unsigned long long)r->vaddr_advance,
		   &(r->busaddr [r->pages]), xmem_global);
  if (rc != DMA_SUCC) {
    GM_WARN(("- failed\n"));
    return ((gm_dp_t) -1);
  }

  p = (gm_dp_t) r->busaddr [r->pages];
  r->pages++;
#else /* DO_MAP_PAGE */
  p = (gm_dp_t) xmemdma (xmem_global, r->vaddr_advance,
			 XMEM_HIDE | XMEM_ACC_CHK);
  p |= 0x80000000;
#endif /* DO_MAP_PAGE */
  r->vaddr_advance += GM_PAGE_LEN;
  return (p);

} /* gm_arch_dma_region_dma_addr_advance () */


/* Return the kernel virtual address for a DMA region */

void *
gm_arch_dma_region_kernel_addr (
				gm_arch_dma_region_t * r
				)
{
  GM_PRINT(GM_AIX_DEBUG, ("gm_arch_dma_region_kernel_addr ...\n"));
  return (r->vaddr);

} /* gm_arch_dma_region_kernel_addr () */

gm_status_t
gm_arch_dma_region_sync (
			 gm_arch_dma_region_t * r,
			 int command
			 )
{
  GM_PRINT(GM_AIX_DEBUG, ("gm_arch_dma_region_sync ...\n"));
  return GM_SUCCESS;

} /* gm_arch_dma_region_sync () */


gm_status_t
gm_arch_map_io_space (
		      gm_instance_state_t *is,
		      gm_u32_t offset,
		      gm_u32_t len,
		      void **kaddr
		      )
{
  dev_ctrl_t *p_dev = is->arch.p_dev;

  GM_PRINT(GM_AIX_DEBUG, ("gm_arch_map_io_space ...\n"));

  if (offset) {
    GM_WARN(("- non-zero offset %d\n", offset));
    return (GM_FAILURE);
  }

  if (p_dev->mapped_addr) {
    GM_WARN(("- some addr already mapped 0x%8x\n", p_dev->mapped_addr));
    return (GM_FAILURE);
  }

  *kaddr = (void *) ((uint)(iomem_att(IO_MAP)) + offset);
  p_dev->mapped_addr = (ulong) *kaddr;
  GM_PRINT(GM_AIX_DEBUG, ("- mapped address 0x%8x\n", *kaddr));
  return (GM_SUCCESS);

} /* gm_arch_map_io_space () */

void
gm_arch_unmap_io_space (
			gm_instance_state_t *is,
			gm_u32_t offset,
			gm_u32_t len,
			void **kaddr
			)
{
  GM_PRINT(GM_AIX_DEBUG, ("Unmap io_space at 0x%8x\n", *kaddr));

  if (offset) {
    GM_WARN(("- non-zero offset %d\n", offset));
    return;
  }

  if (is->arch.p_dev->mapped_addr == 0) {
    GM_WARN(("- addr not mapped 0x%8x\n", *kaddr));
    return;
  }

  if (is->arch.p_dev->mapped_addr != (ulong) *kaddr) {
    GM_WARN(("- addr not same 0x%8x-0x%8x\n",
	     is->arch.p_dev->mapped_addr, *kaddr));
    iomem_det ((void *)*kaddr);
    return;
  }

  iomem_det ((void *)*kaddr);
  is->arch.p_dev->mapped_addr = 0;

} /* gm_arch_unmap_io_space () */

/* extern	int kgethostname (char *, int *); */


gm_status_t
gm_arch_gethostname (
		     char *ptr,
		     int len
		     )
{
#if 1
  if (kgethostname (ptr, &len)) {
    GM_WARN(("kgethostname() failed\n"));
    return GM_FAILURE;
  }
  else {
    GM_PRINT(GM_AIX_DEBUG, ("- hostname %s\n", ptr));
    return GM_SUCCESS;
  }
#else
  GM_WARN(("gm_arch_gethostname: DISABLED\n"));
  /* feldy disabled this function... why? */
  return GM_FAILURE;
#endif

} /* gm_arch_gethostname () */

void
gm_arch_spin (
	      gm_instance_state_t *is,
	      gm_u32_t usecs
	      )
{
  struct timestruc_t dtime;	/* delta time */
  struct timestruc_t ctime;	/* current time */
  struct timestruc_t etime;	/* ending time */

  GM_PRINT(GM_AIX_DEBUG, ("Spin %d microsecs\n", usecs));

  /* Get the current time */
  curtime (&ctime);

  if (usecs < uS_PER_SECOND) {
    dtime.tv_sec = 0;
    dtime.tv_nsec = usecs * NS_PER_uS;
  }
  else {
    dtime.tv_sec = usecs / uS_PER_SECOND;
    usecs -= dtime.tv_sec * uS_PER_SECOND;
    dtime.tv_nsec = usecs * NS_PER_uS;
  }

  /* Generate an ending time */
  ntimeradd (dtime, ctime, etime) 

    /* Now wait for time to pass */
    while (1) {
      curtime (&ctime);
      if (ntimercmp (ctime, etime, >))
	break;
    }

} /* gm_arch_spin () */


gm_status_t
gm_arch_get_page_len (
		      unsigned long *page_len
		      )
{
  GM_PRINT(GM_AIX_DEBUG, ("gm_arch_get_page_len ...\n"));
  *page_len = PAGESIZE;
  return GM_SUCCESS;

} /* gm_arch_get_page_len () */


/* This is called just after the port state is created (in gm_minor.c)
   to perform architecture-specific initialization. */

gm_status_t
gm_arch_port_state_init (
			 gm_port_state_t *ps
			 )
{
  GM_PRINT(GM_AIX_DEBUG, ("Init ps at 0x%x\n", ps));
  return GM_SUCCESS;

} /* gm_arch_port_state_init () */



/* This is called just after the port state is initialized (in
   gm_port_state_open()) to perform architecture-specific initialization. */

gm_status_t
gm_arch_port_state_open (
			 gm_port_state_t *ps
			 )
{
  GM_PRINT(GM_AIX_DEBUG, ("Open ps at 0x%x\n", ps));
  return GM_SUCCESS;

} /* gm_arch_port_state_open () */



/* This is called just before the port state is finalized (in
   gm_port_state_close()) to perform architecture-specific finalization. */

void
gm_arch_port_state_close (
			  gm_port_state_t *ps
			  )
{
  GM_PRINT(GM_AIX_DEBUG, ("Close ps at 0x%x\n", ps));
  return;

} /* gm_arch_port_state_close () */



/* This is called just before the port state is destroyed (in
   gm_minor.c) to perform architecture-specific finalization. */

void
gm_arch_port_state_fini (
			 gm_port_state_t *ps
			 )
{
  GM_PRINT(GM_AIX_DEBUG, ("Finalize ps at 0x%x\n", ps));
  return;

} /* gm_arch_port_state_fini () */


gm_status_t
gm_arch_read_pci_config_32 (
			    gm_instance_state_t *is,
			    gm_offset_t offset,
			    gm_u32_t *value
			    )
{
  dev_ctrl_t *p_dev = is->arch.p_dev;
  int rc;

  GM_PRINT(GM_AIX_DEBUG, ("Read 32 bits at offset 0x%x\n", offset));

  rc = gm_aix_rw_pci_cfg (DEV_BID, DDS.bus_slot, offset,
			  value, 4, PCI_CFG_READ);

  return (rc ? GM_FAILURE : GM_SUCCESS);

} /* gm_arch_read_pci_config_32 () */


gm_status_t
gm_arch_write_pci_config_32 (
			     gm_instance_state_t *is,
			     gm_offset_t offset,
			     gm_u32_t value
			     )
{
  dev_ctrl_t *p_dev = is->arch.p_dev;
  int rc;

  GM_PRINT(GM_AIX_DEBUG,
	   ("Write 32 bits at offset 0x%x value 0x%x\n", offset, value));

  rc = gm_aix_rw_pci_cfg (DEV_BID, DDS.bus_slot, offset,
			  &value, 4, PCI_CFG_WRITE);

  return (rc ? GM_FAILURE : GM_SUCCESS);

} /* gm_arch_write_pci_config_32 () */


gm_status_t
gm_arch_read_pci_config_16 (
			    gm_instance_state_t *is,
			    gm_offset_t offset,
			    gm_u16_t *value
			    )
{
  dev_ctrl_t *p_dev = is->arch.p_dev;
  int rc;

  GM_PRINT(GM_AIX_DEBUG, ("Read 16 bits at offset 0x%x\n", offset));

  rc = gm_aix_rw_pci_cfg (DEV_BID, DDS.bus_slot, offset,
			  value, 2, PCI_CFG_READ);

  return (rc ? GM_FAILURE : GM_SUCCESS);

} /* gm_arch_read_pci_config_16 () */


gm_status_t
gm_arch_write_pci_config_16 (
			     gm_instance_state_t *is,
			     gm_offset_t offset,
			     gm_u16_t value
			     )
{
  dev_ctrl_t *p_dev = is->arch.p_dev;
  int rc;

  GM_PRINT(GM_AIX_DEBUG,
	   ("Write 16 bits at offset 0x%x value 0x%x\n", offset, value));

  rc = gm_aix_rw_pci_cfg (DEV_BID, DDS.bus_slot, offset,
			  &value, 2, PCI_CFG_WRITE);

  return (rc ? GM_FAILURE : GM_SUCCESS);

} /* gm_arch_write_pci_config_16 () */

gm_status_t
gm_arch_read_pci_config_8 (
			   gm_instance_state_t *is,
			   gm_offset_t offset,
			   gm_u8_t *value
			   )
{
  dev_ctrl_t *p_dev = is->arch.p_dev;
  int rc;

  GM_PRINT(GM_AIX_DEBUG, ("Read 8 bits at offset 0x%x\n", offset));

  rc = gm_aix_rw_pci_cfg (DEV_BID, DDS.bus_slot, offset,
			  value, 1, PCI_CFG_READ);

  return (rc ? GM_FAILURE : GM_SUCCESS);

} /* gm_arch_read_pci_config_8 () */


gm_status_t
gm_arch_write_pci_config_8 (
			    gm_instance_state_t *is,
			    gm_offset_t offset,
			    gm_u8_t value
			    )
{
  dev_ctrl_t *p_dev = is->arch.p_dev;
  int rc;

  GM_PRINT(GM_AIX_DEBUG,
	   ("Write 8 bits at offset 0x%x value 0x%x\n", offset, value));

  rc = gm_aix_rw_pci_cfg (DEV_BID, DDS.bus_slot, offset,
			  &value, 1, PCI_CFG_WRITE);

  return (rc ? GM_FAILURE : GM_SUCCESS);

} /* gm_arch_write_pci_config_8 () */


gm_port_state_t
*gm_arch_port_for_id (
		      gm_instance_state_t * is,
		      unsigned port
		      )
{
  GM_PRINT(GM_AIX_DEBUG, ("gm_arch_port_for_id: ...\n"));
  return ((gm_port_state_t *) NULL);

} /* *gm_arch_port_for_id () */


gm_status_t
gm_arch_copyout (
		 const void *from,
		 void *to,
		 gm_size_t len
		 )
{
  int rc;

  rc = copyout ((void *)from, to, len);
  if (rc) {
    GM_WARN(("copyout (0x%8x, 0x%8x, 0x%x) failed 0x%x\n", from, to, len, rc));
  }
  else {
    GM_PRINT(GM_AIX_DEBUG,
	     ("copyout (0x%8x, 0x%8x, 0x%x) OK\n", from, to, len));
  }
  return (rc == 0 ? GM_SUCCESS : GM_FAILURE);

} /* gm_arch_copyout () */


gm_status_t
gm_arch_copyin (
		const void *from,
		void *to,
		gm_size_t len
)
{
  int rc;

  rc = copyin ((void *)from, to, len);
  if (rc) {
    GM_WARN(("copyin (0x%8x, 0x%8x, 0x%x) failed 0x%x\n", from, to, len, rc));
  }
  return (rc == 0 ? GM_SUCCESS : GM_FAILURE);

} /* gm_arch_copyin () */

void
gm_arch_abort (void)
{

  panic ("gm_arch_abort() called!");

  return;

} /* gm_arch_abort () */

#ifdef	junk

void
memcpy (char*a, char*b, int len)
{
  bcopy (b, a, len);

} /* memcpy () */

#endif	/* junk */


/* don't care */


gm_status_t
gm_arch_mmap_contiguous_segment (
				 gm_port_state_t *ps,
				 void * kaddr,
				 unsigned long blockSize,
				 gm_up_t *vaddr
				 )
{
  GM_PRINT(GM_AIX_DEBUG, ("gm_arch_mmap_contiguous_segment: NOT SUPPORTED\n"));
  return GM_FAILURE;

} /* gm_arch_mmap_contiguous_segment () */

void
gm_arch_munmap_contiguous_segments (
				    gm_port_state_t *ps
				    )
{
  return;

} /* gm_arch_munmap_contiguous_segments () */



/* Return the STS bits for a DMA region. */

gm_s32_t
gm_arch_dma_region_status (
			   gm_arch_dma_region_t *r
			   )
{
  return (0x0F); 

} /* gm_arch_dma_region_status () */


/*****************************************************************************/
/*
 * NAME:                  gm_aix_rw_pci_cfg
 *
 * FUNCTION:              Provides access to PCI Configuration address
 *			  space
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:		  
 *
 * RETURNS:               0 - OK
 *	                  other errno values
 */
/*****************************************************************************/

GM_STATIC int
gm_aix_rw_pci_cfg(
		  uint bid,
		  uint slot,
		  ulong offset,       	/* Register offset from base */
		  void *p_buf,      	/* Buffer to read/write */
		  int size,       	/* Size of access [number of bytes] */
		  int read_write  	/* Read or write */
		  )
{

  /* !!!! FIX -- this is inaccessible in ioacc.h */ 
#define PCI_CFGMEM	2		/* PCI configuration region */

  int		rc;
  struct mdio	mdd;		/* Machine dd access struct */
  uint cfg_bid; 			/* BID for PCI Config address space */
  char buf[4], *argbuf = p_buf;		/* For byte-swapping input buffer */

  cfg_bid = BID_ALTREG(bid, PCI_CFGMEM);

  /* Setup structure for pci_cfgrw */
  mdd.md_addr		= offset;
  mdd.md_data		= buf;
  mdd.md_sla	 	= slot; 

  switch (size)
    {
    case 4:
      mdd.md_incr = MV_WORD; 
      mdd.md_size = size/4;

      /* Byte swap */
      if (read_write == PCI_CFG_WRITE) {
#if GM_AIX_DEBUG_PCI_CFG_RW
	GM_PRINT(GM_AIX_DEBUG,
		 ("- bid=0x%X, slot=%d, offset=0x%X, writing 0x%08X\n", 
		  cfg_bid, slot, offset, *(uint*)p_buf));
#endif
	buf[0] = argbuf[3];
	buf[1] = argbuf[2];
	buf[2] = argbuf[1];
	buf[3] = argbuf[0];
      }
      break;

    case 2:
      mdd.md_incr = MV_SHORT; 
      mdd.md_size = size/2;

      /* Byte swap */
      if (read_write == PCI_CFG_WRITE) {
#if GM_AIX_DEBUG_PCI_CFG_RW
	GM_PRINT(GM_AIX_DEBUG,
		 ("- bid=0x%X, slot=%d, offset=0x%X, writing 0x%04X\n",
		  cfg_bid, slot, offset, *(ushort*)p_buf));
#endif
	buf[0] = argbuf[1];
	buf[1] = argbuf[0];
      }
      break;

    case 1:
    default:
      mdd.md_incr = MV_BYTE; 
      mdd.md_size = size;

      buf[0] = argbuf[0];
#if GM_AIX_DEBUG_PCI_CFG_RW
      GM_PRINT(GM_AIX_DEBUG,
	       ("- bid=0x%X, slot=%d, offset=0x%X, writing 0x%02X\n", 
		cfg_bid, slot, offset, *(uchar*)p_buf));
#endif
      break;
    }

  if ((rc = pci_cfgrw(cfg_bid, &mdd, read_write)) != 0)
    GM_WARN(("- pci_cfgrw() failed. rc = %d\n", rc));
	
  if (read_write == PCI_CFG_READ) {
    switch (size)
      {
      case 4:
	/* Byte swap */
	argbuf[3] = buf[0];
	argbuf[2] = buf[1];
	argbuf[1] = buf[2];
	argbuf[0] = buf[3];
#if GM_AIX_DEBUG_PCI_CFG_RW
	GM_PRINT(GM_AIX_DEBUG,
		 ("- bid=0x%X, slot=%d, offset=0x%X, read 0x%08X\n", 
		  cfg_bid, slot, offset, *(uint*)p_buf));
#endif
	break;

      case 2:
	/* Byte swap */
	argbuf[1] = buf[0];
	argbuf[0] = buf[1];
#if GM_AIX_DEBUG_PCI_CFG_RW
	GM_PRINT(GM_AIX_DEBUG,
		 ("- bid=0x%X, slot=%d, offset=0x%X, read 0x%04X\n", 
		  cfg_bid, slot, offset, *(ushort*)p_buf));
#endif
	break;

      case 1:
      default:
	argbuf[0] = buf[0];
#if GM_AIX_DEBUG_PCI_CFG_RW
	GM_PRINT(GM_AIX_DEBUG,
		 ("- bid=0x%X, slot=%d, offset=0x%X, read 0x%02X\n", 
		  cfg_bid, slot, offset, *(uchar*)p_buf));
#endif
	break;
      }

  }
  return (rc);

} /* gm_aix_rw_pci_cfg() */


gm_status_t
gm_arch_lock_user_buffer_page (
			       gm_port_state_t *ps,
			       gm_up_t in,
			       gm_dp_t *dma_addr,
			       gm_arch_page_lock_t *lock
			       )
{
  /* >>>> Review the C spec to make sure order of initializations is assured */
  gm_instance_state_t *is = ps->instance;
  dev_ctrl_t *p_dev = is->arch.p_dev;
  int		rc = 0;
  struct xmem	*dp = &(lock->dp);
  ulong tmp;

  GM_PRINT(GM_DEBUG_MEM_REGISTER, ("- vaddr 0x%8x\n", in));
  if (in & (GM_PAGE_LEN - 1)) {
    GM_WARN(("- unaligned vaddr 0x%8x\n", in));
    return (GM_FAILURE);
  }

#if DO_MAP_PAGE
  lock->p_dev = p_dev;
#endif /* DO_MAP_PAGE */
  lock->in = 0;
  dp->aspace_id = XMEM_INVAL;

  rc = xmattach ((char *)in, GM_PAGE_LEN, dp, USER_ADSPACE);
  if (rc != XMEM_SUCC) {
    GM_WARN(("- xmattach failed 0x%x\n", rc));
    return (GM_FAILURE);
  }

  rc = xmempin ((caddr_t)in, GM_PAGE_LEN, dp);
  if (rc) {
    GM_WARN(("- xmempin failed 0x%x\n", rc));
    xmdetach (dp);
    return (GM_FAILURE);
  }
	
#if DO_MAP_PAGE
  rc = D_MAP_PAGE (p_dev->d_handle, DMA_BYPASS | DMA_READ,
		   (unsigned long long)in, &(lock->busaddr), dp);

  if (rc != DMA_SUCC) {
    GM_WARN(("- d_map_page failed 0x%x\n", rc));
    xmemunpin ((caddr_t)in, GM_PAGE_LEN, dp);
    xmdetach (dp);
    return (GM_FAILURE);
  }

  *dma_addr = (gm_dp_t) lock->busaddr;
#else /* DO_MAP_PAGE */
  *dma_addr = (gm_dp_t) xmemdma (dp, in, XMEM_ACC_CHK | XMEM_HIDE);
  if (tmp == (gm_dp_t) XMEM_FAIL) {
    GM_WARN(("- xmemdma failed\n"));
    xmemunpin ((caddr_t)in, GM_PAGE_LEN, dp);
    xmdetach (dp);
    return (GM_FAILURE);
  }

  *dma_addr |= (gm_dp_t) 0x80000000;
#endif /* DO_MAP_PAGE */

#if DO_MAP_PAGE1		/* N.B. NOT the same symbol as the others! */
  if (((int) *dma_addr) != lock->busaddr) {
    GM_WARN(("- bad addrs - 0x%8x .ne. 0x%llx\n", lock->busaddr, *dma_addr));
    D_UNMAP_PAGE (p_dev->d_handle, &(lock->busaddr));
    xmemunpin ((caddr_t)in, GM_PAGE_LEN, dp);
    xmdetach (dp);
    return (GM_FAILURE);
  }

  *dma_addr = (gm_dp_t) lock->busaddr;
#endif /* DO_MAP_PAGE1 */
  lock->in = in;
  GM_PRINT(GM_DEBUG_MEM_REGISTER, ("- paddr 0x%llx\n", *dma_addr));

  return (GM_SUCCESS);

} /* gm_arch_lock_user_buffer_page () */


void
gm_arch_unlock_user_buffer_page (
				 gm_arch_page_lock_t * lock
				 )
{
  struct xmem *dp = &(lock->dp);

  GM_PRINT(GM_DEBUG_MEM_REGISTER, ("- 0x%11x\n", lock->busaddr));

#if DO_MAP_PAGE
  D_UNMAP_PAGE (lock->p_dev->d_handle, &(lock->busaddr));
#endif /* DO_MAP_PAGE */

  if (xmemunpin ((caddr_t)lock->in, GM_PAGE_LEN, dp)) {
    GM_WARN(("- xmemunpin failed\n"));
    return;
  }

  xmdetach (dp);
	
  return;

} /* gm_arch_unlock_user_buffer_page () */


GM_STATIC dev_ctrl_t *
gm_aix_device (
	       dev_t devno
	       )
{
  dev_ctrl_t *p_dev;
  
  for (p_dev = dd_ctrl.p_dev_list; p_dev != NULL; p_dev = p_dev->next) {
    if (p_dev->unit == (minor (devno) / 2)) {
      GM_PRINT(GM_AIX_DEBUG, ("Found matching p_dev\n"));
      return (p_dev);
    }
  }
  GM_WARN(("No matching p_dev\n"));
  return (NULL);

} /* gm_aix_device */



/*****************************************************************************/
/*
 * NAME:     add_cdt
 *
 * FUNCTION: add an entry to the component dump table
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:
 *
 * INPUT:
 *	name		- character string of the name of the data structure
 *	addr		- address of the data structure
 *	len		- length of the data structure
 *
 * RETURNS:  
 *	none.
 */
/*****************************************************************************/
GM_STATIC void 
add_cdt(
	char *name,  	/* label string for area dumped */
	char *addr,  	/* address of the area to be dumped */
	int   len   	/* amount of data to be dumped */
	)
{
  struct cdt_entry *p_entry;

  p_entry = &dd_ctrl.cdt.entry[dd_ctrl.cdt.count];
  strcpy(p_entry->d_name, name);
  p_entry->d_len = len;
  p_entry->d_ptr = addr;
  p_entry->d_xmemdp = NULL;

  dd_ctrl.cdt.count++;
  dd_ctrl.cdt.head._cdt_len += sizeof(struct cdt_entry);

}


/*****************************************************************************/
/*
 * NAME:     cdt_del
 *
 * FUNCTION: delete an entry to the component dump table
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:
 *
 * INPUT:
 *	name		- character string of the name of the data structure
 *	addr		- address of the data structure
 *	len		- length of the data structure
 *
 * RETURNS:  
 *	none.
 */
/*****************************************************************************/
GM_STATIC void
del_cdt(
	char *name,  	/* label string for area dumped */
	char *addr, 	/* address of the area to be dumped */
	int  len   	/* amount of data to be dumped */
	)
{
  struct cdt_entry *p_entry;
  int i;

  /* find the entry in the table (match only the memory pointer) */
  for (p_entry = &dd_ctrl.cdt.entry[0], i = 0;
       i < dd_ctrl.cdt.count; p_entry++, i++) {
    if (p_entry->d_ptr == addr) 
      break;
  }

  /* if found the entry, remove the entry by re-arrange the table */
  if (i < dd_ctrl.cdt.count) {
    for (; i < dd_ctrl.cdt.count; p_entry++, i++) {
      strcpy(p_entry->d_name, dd_ctrl.cdt.entry[i+1].d_name);
      p_entry->d_len = dd_ctrl.cdt.entry[i+1].d_len;
      p_entry->d_ptr = dd_ctrl.cdt.entry[i+1].d_ptr;
    }
    dd_ctrl.cdt.count--;
    dd_ctrl.cdt.head._cdt_len -= sizeof(struct cdt_entry);
  }

}


/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/

