#ifndef _gm_arch_types_h_
#define _gm_arch_types_h_

/* NOTE: The include of gm_arch_debug.h MUST be first! */
#include "gm_arch_debug.h"

#include <sys/limits.h>
#include <sys/types.h>
#include <sys/mdio.h>
#include <sys/buf.h>
#include <sys/device.h>
#include <sys/devinfo.h>
#include <sys/dma.h>
#include <sys/dump.h>
#include <sys/errno.h>
#include <sys/err_rec.h>
#include <sys/file.h>
#include <sys/intr.h>
#include <sys/ioacc.h>
#include <sys/iocc.h>
#include <sys/atomic_op.h>
#include <sys/lock_alloc.h>
#include <sys/lockl.h>
#include <sys/lockname.h>
#include <sys/malloc.h>
#include <sys/param.h>
#include <sys/sleep.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/timer.h>		/* We use trb timers */
#include <sys/xmem.h>
#include <sys/trchkid.h>
#include <sys/adspace.h>
#include <sys/except.h>
#include <sys/ethernet_mibs.h>
#include <stddef.h>
#include <net/spl.h>
#include <sys/cdli_entuser.h>
#include <sys/cfgdb.h>

#include "gm_dds.h"

#ifdef ISR
#undef ISR
#endif

#ifndef __KERNEL__
#define __KERNEL__ 1
#endif

#ifndef NULL
#define NULL ((void *)0)
#endif

/* 
 * NOTE: AIX seems to want us to use a lock class of 0, once debugging is done.
 *       See /usr/include/sys/lockname.h.
 */
#define GM_AIX_LOCK_CLASS       99
#define GM_AIX_LOCK_INSTANCE_DD 1
#define GM_AIX_LOCK_INSTANCE_DV 2
#define GM_AIX_LOCK_INSTANCE_PS 3
#define GM_AIX_LOCK_INSTANCE_OPEN 4
#define GM_AIX_LOCK_INSTANCE_SYNC_BASE 5

#define	MY_HEAP_SIZE		0x800000L
#define	MAX_MAPPED_PAGES	256

#define	MAX_RX_FRAGS	8
#define	MAX_TX_FRAGS	8

/*****************************************************************************
 * Component Dump Table
 * 
 * The Component Dump Table contains the dd_ctrl structure and a dev_ctrl 
 * structure for each open adapter.  Leave one extra entry always empty for
 * the table management.
 *
 *****************************************************************************/

/* The magic number below is based on a statement in the IBM sales document,
 * "IBM eserver pSeries 660 Model 6H0 and 6H1 Express Configurations" dated
 * October 8, 2002.
 *
 * We should bump the limit if/when we find that some AIX host supports more
 * PCI slots than this.
*/
#define PCI_MAX_ADAPTERS 28
#define GM_AIX_CDT_SIZE (2 + PCI_MAX_ADAPTERS)
#define DD_NAME_STR "gm"	/* Driver name for dump func */

#define	ELAPSED_TIME(s)	((lbolt - s)/HZ)

typedef struct {
   int	            count;			/* number of cdt entries used */
   struct cdt_head  head;			/* cdt head */
   struct cdt_entry entry[GM_AIX_CDT_SIZE];	/* cdt entries */
} cdt_t;


/*****************************************************************************
 * Ethernet packet stuff
 *****************************************************************************/

typedef struct {
	uchar dest_addr[ENT_NADR_LENGTH];
	uchar src_addr[ENT_NADR_LENGTH];
	uchar type[2];			/* For Std Ethernet only */
	uchar data[1];			/* Variable length */
} ENT_PKT;

#define ENT_HDR_LEN		(2*ENT_NADR_LENGTH + 2) /* Ethernet header length */
#define ENT_GROUP_BIT 		0x01	/* Indicates group address */
#define MAX_PACKET_TIME_USEC 	1220

/* flags for WRK.filter */
#define	ENABLE_MULTICAST	0x01
#define	ENABLE_PROMISCUOUS	0x02

typedef enum {
	INDIVID_PACKET,
	MCAST_PACKET,
	BCAST_PACKET
} ENT_ADDR_TYPE;

/* Copy a Ethernet network address
 */
#define COPY_NADR(from, to)	{ \
          *((ulong *)(to)) = *((ulong *)(from)); \
          *((ushort *)((char *)(to) + 4)) = *((ushort *)((char *)(from) + 4)); \
        }

/* Compare Ethernet network addresses
 */
#define SAME_NADR(a, b)	( \
	   *((ulong *)(a)) == *((ulong *)(b)) && \
	   *((ushort *)((char *)(a) + 4)) == *((ushort *)((char *)(b) + 4)) \
			)

/*****************************************************************************
 * Multicast address table
 *****************************************************************************/
#define MULTI_TABLE_SLOT 	16

typedef struct multi_slot {
  char m_addr[ENT_NADR_LENGTH];
  int  ref_count;
} multi_slot_t;

typedef struct ent_multi {
  struct ent_multi *next;			/* Table extension */
  short  in_use;				/* Number of slots in use */
  multi_slot_t m_slot[MULTI_TABLE_SLOT];	/* Slots */
} ent_multi_t;

/*****************************************************************************
 * Adapter states
 *****************************************************************************/

/* Adapter states (easily read when in the kernel debugger.) */
#define DEAD_STATE      0xdead
#define NULL_STATE      0xffff
#define OPEN_PENDING    0xcafe
#define OPEN_STATE      0xaaaa
#define CLOSED_STATE    0xbbbb
#define CLOSE_PENDING   0xcccc

#define ENT_GM_PCI	0x00000008	/* for the entgm_stats extension */

/*****************************************************************************
 * Adapter stats (Same as adapter registers only wider)
 *****************************************************************************/

typedef struct {
	ulong carrier_lost;
	ulong no_cd_beat;
	ulong multiple_coll;
	ulong one_coll;
	ulong late_coll;
	ulong rx_overrun;
	ulong tx_pkts_ok;
	ulong rx_pkts_ok;
	ulong tx_defer;
	ulong upper_pkts_ok;
	ulong rx_bytes_ok;
	ulong tx_bytes_ok;
	ulong bad_ssd;
} WIDE_ADAP_STATS;


/*****************************************************************************
 * Device Work section
 *****************************************************************************/

typedef struct {
	Simple_lock dev_lock;		/* Device lock (serialize access to
					 * device and/or data structure) */
	int adap_state;			/* Current adapter state */
	int alloc_size;			/* Size of alloced memory for the dev
					 * control struct and Wait Q */
	struct mbuf *rx_head;		/* Head of recv'd mbuf queue */
	struct mbuf *rx_tail;		/* Tail of recv'd mbuf queue */
	struct mbuf *mbufs_to_send;	/* List of all tx-mbufs to xmit */
	WIDE_ADAP_STATS adap_stats;	/* Cumulative stats from adapters stat
					 * regs */
	int ndd_stime;			/* Start time of ndd stats */
	int dev_stime;			/* Start time of device stats */
	int send_bad_rx_pkts;		/* save bad pkt ref. count */
	int otherstatus;		/* all other async status ref count */
	int enable_promiscuous;		/* enable promiscuous mode */
	int enable_multicast;		/* enable multicast reference count. */
	ent_multi_t multi_table;	/* Multicast address table */
	int multi_addr_count;		/* Number of entries in multicast addr
					 * table */
	int filter;			/* to store MCAST/PROMISCUOUS flags */
	struct gm_port *port;		/* port used to send/recv ip pkts */

	caddr_t iovec_mem;		/* Memory allocated for iovec structures
					   used for DMA mapin */

	uchar net_addr[ENT_NADR_LENGTH]; /* actual network address in use */
	uint adap_num;			/* Index of adapter in detected array */
	uint bid;			/* BID for I/O access */
	struct io_map *io_map_ptr;	/* For doing PCI bus I/O */
	struct d_handle *d_handle;	/* Pointer returned by d_map_init */
} ent_wrk_t;

/*****************************************************************************
 * Device Control structure
 *****************************************************************************/
typedef struct dev_ctrl {
	struct intr ihs;		/* Interrupt handler ctl struct */
	struct intr dpc_ihs;		/* Must be the second element after
					   ihs */
	struct dev_ctrl *next;		/* Next device control struct in list */
	ndd_t ndd;			/* Ndd for NS ndd chain */
	dds_t dds;			/* Cfg info assembled by cfg method */
	ent_wrk_t wrk;			/* Device work area */
	struct d_handle	*d_handle;	/* For d_map_page() */

	// pci_vpd_t vpd;			/* Adapter VPD (EEPROM) */
	struct gm_port *port;		/* port used to send/recv ip pkts */
	ent_genstats_t entstats;	/* Ethernet generic statistics */
	ethernet_all_mib_t mibs;	/* Ethernet MIB's */

	struct gm_instance_state *is;
	ulong		mapped_addr;
	int	unit;			/* GM unit number (same as is->id) */
	int usage_count;		/* Number of users of this device */
} dev_ctrl_t;

/*
 * Macros for accessing device control area. The pointer to this area has to 
 * be named p_dev for using these macros.
 *
 * NOTE: I definitely don't like these macros, which were devised by Softex.
 *       They obscure the code significantly.  Let's remove them ASAP.
 *                                                            - Max  02/11/26
 */

#define IHS        p_dev->ihs
#define NDD        p_dev->ndd
#define DDS        p_dev->dds
#define VPD        p_dev->vpd
#define WRK        p_dev->wrk
#define IO_MAP	   p_dev->wrk.io_map_ptr
#define DEV_BID    p_dev->wrk.bid
#define ENTSTATS   p_dev->entstats
#define GENSTATS   p_dev->ndd.ndd_genstats
#define DEVSTATS   p_dev->devstats
#define MIB        p_dev->mibs
#define MIB_IFEXTNS	   p_dev->mibs.Generic_mib.ifExtnsEntry
#define MIB_DOT3STATS	   p_dev->mibs.Ethernet_mib.Dot3StatsEntry
#define ADAPSTATS  p_dev->wrk.adap_stats

/***************************************************************************
 *  Driver control structure (not per device)
 ***************************************************************************/

typedef struct {
	lock_t cfg_lock;		/* ent_config lockl use only */
	Simple_lock i_lock;		/* Initialization lock */
	Simple_lock o_lock;		/* Serializes open, close, etc. */
	dev_ctrl_t *p_dev_list;	/* Linked-list of device controls */
	int num_devs;			/* Number devices configured  */
	int initialized;		/* Driver initialized [TRUE/FALSE] */
	int num_ip_opens;		/* Nmmber of ethernet-emulation devices opened */
	cdt_t cdt;				/* Dump table */
	heapaddr_t my_heap;		/* pointer to local pinned heap */
	void *my_heap_addr;		/* pointer to local pinned heap mem */
} dd_ctrl_t;


typedef struct gm_arch_instance_info  
{
	dev_ctrl_t	*p_dev;
} gm_arch_instance_info_t;

typedef struct gm_arch_dma_region
{ 
	gm_arch_instance_info_t	*ai;
	struct d_handle		*d_handle;
	int	length;
	unsigned long long paddr;
	char	*vaddr_advance;
	char	*vaddr;
	int	pages;
	unsigned long long	busaddr [MAX_MAPPED_PAGES];
} gm_arch_dma_region_t;

typedef struct gm_arch_port_info
{
	int	nothing;
} gm_arch_port_info_t;

typedef struct gm_arch_sync
{
	dev_ctrl_t	*p_dev;
  tid_t		event_word;	/* data referred to by this pointer is managed
				   by the core AIX internals and must NOT be
				   altered by the driver! */
	Simple_lock	sync_lock;	/* sync lock (serialize access to
				   device and/or data structure) */
  Simple_lock	wakeup_lock;	/* So we only wake thread which is sleeping */
	uint		intr_priority;
  /*  // uint	old_pri; */
  volatile int	wake_cnt;	/* used to handle premature wake */
} gm_arch_sync_t;

typedef struct gm_arch_page_lock
{
	dev_ctrl_t		*p_dev;
	unsigned long long	in;		/* holds the user's vaddr */
	unsigned long long	busaddr;	/* holds bus addr */
	struct xmem		dp;		/* cross mem descriptor */
} gm_arch_page_lock_t;

typedef int gm_arch_minor_t;
typedef int gm_arch_ioctl_context_t;


/***********************************************************************
 * Function prototype declarations for shared functions
 ***********************************************************************/

int gm_aix_ip_init (dev_ctrl_t *p_dev);
int gm_aix_ip_term (dev_ctrl_t *p_dev);

void gm_aix_lock_driver (void);
void gm_aix_unlock_driver (void);

#define GM_ARCH_LOCK_ETHER_SEND_QUEUE(port)\
int s;\
s = disable_lock((port)->kernel_port_state->instance->arch.p_dev->dds.intr_priority, \
&((port)->kernel_port_state->instance->arch.p_dev->wrk.dev_lock))

#define GM_ARCH_UNLOCK_ETHER_SEND_QUEUE(port)\
unlock_enable(s,\
&((port)->kernel_port_state->instance->arch.p_dev->wrk.dev_lock))


#endif /* _gm_arch_types_h_ */

