/*
 * PRODUCT: gmpci - Myrinet PCI Device Driver
 *
 * CONTENTS: Change Method
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <cf.h>
#include <locale.h>
#include <sys/types.h>
#include <sys/cfgdb.h>
#include <sys/cfgodm.h>
#include <sys/errno.h>


/*==========================================================================
 * To get debugging output, modify CFLAGS in ./Makefile to define DEBUG and
 * (optionally) DBG_FILE.  The output will be written to the file identified
 * by DBG_FILE, else to stderr.
 *==========================================================================*/
   
#ifdef DEBUG
#define DBG(x) x
#else
#define DBG(x)
#endif

/*==========================================================================
 * Local prototypes
 *==========================================================================*/
void err_exit(int exitcode);

/*==========================================================================
 * Externals
 *==========================================================================*/

static char *prog;
struct attr {
	char *name;		/* Attr name (ptr into attr_list) */
	char *new_val;		/* Attr value (ptr into attr_list) */
	char old_val[ATTRVALSIZE];
	struct CuAt CuAt;
} *attr = NULL;

#ifdef DEBUG
static FILE *dbg_fp;
#endif

/*===========================================================================
 *
 * NAME:	main
 *
 * FUNCTION:	Change method for Myrinet PCI Adapter
 *
 * SYNTAX:	chggmpci -l <logical name> [-p <parent>] [-w <location>]
 *                      [-a "attr=val [attr=val] ..."] [-P | -T]
 *
 *		Note: the documented syntax for the -a flag for this method
 *		      is incorrect. If repeated attributes are passed, the
 *		      argv array will contain a single attribute string
 *		      with the attr=val pairs separated by blanks.
 *
 * EXIT CODE:   from cf.h: E_OK if successful, otherwise specific error code
 *
 *===========================================================================*/

main(
int argc,
char *argv[]
)
{
	char *lname;			/* Logical name of this device */
	char *parent;			/* Parent's name */
	char *connection;		/* parent connection */
	char *attr_list;		/* sting of attr=value pairs */
	char *badattr;			/* Attr list arg to/from attrval() */
	char criteria[256];		/* criteria search string */
	char *odmdir_env;		/* Environment ODM directory */
	char lock_file[128];		/* Name of lock file */
 
	int errflg, c;			/* Used in parsing command line */
	int Tflag;                  	/* Flag for -T parameter */
	int Pflag;			/* Flag for -P param (i.e, "only change db")*/
	int new_location;		/* TRUE = generate new location code */
	char *pchar;			/* Ptr into attr_list */
	int num_attr = 0;		/* Number of attr in attr_list */
	int i;				/* Loop index */
	int getattr_how_many;		/* Number of attr from getattr() */
	int rc;				/* Return code */
 
	int oldstatus;			/* Saved status of device */
	char oldparent[NAMESIZE];    	/* Parent name before change */
	char oldconnect[NAMESIZE];   	/* Connection before change */
	char *out_p;                 	/* Message from config method */
	char *err_p;                 	/* Message from config method */
 
	struct Class *cusdev;		/* Class ptr from odm_open_class() */

	struct CuDv CuDv;		/* CuDv for this device */
	struct CuDv oldpar_CuDv;	/* CuDv for old parent */
	struct CuDv par_CuDv;		/* CuDv for new parent */
	struct CuDv other_CuDv;		/* CuDv for other device */

	struct PdDv PdDv;		/* PdDv for this device */
	struct PdCn PdCn;         	/* PDCn for parent */
	struct CuAt *pCuAt;       	/* CuAt (ptr) for this device */
 
	time_t now;			/* Updated by time() */

	extern int optind;		/* defined by getopt() */
	extern char *optarg;		/* defined by getopt() */

#ifdef DEBUG
	time(&now);
#ifdef DBG_FILE
 	dbg_fp = fopen(DBG_FILE, "a+");
	if (dbg_fp == NULL) {
	  printf("open of %s failed with error %d; using stderr\n",
			 DBG_FILE, errno);
	  dbg_fp = stderr;	/* Don't freopen because that affects all */
	}
#else
	dbg_fp = stderr;	/* Don't freopen because that affects all */
#endif
#endif

	DBG(fprintf(dbg_fp, "\n%s: main %s", argv[0], ctime(&now)));
	prog = argv[0];

	setlocale(LC_MESSAGES, "En_US");

	/* Set up initial variable values
	 */
	lname = parent = connection = NULL;
	attr_list = NULL;
	errflg = 0;
	Pflag = 0;
	Tflag = 0;
 
	/* Get command line options
	 */
	while ((c = getopt(argc, argv, "l:a:p:w:PT")) != EOF) {
		switch (c) {
		case 'l':
			if (lname != NULL)
				errflg++;
			lname = optarg;
			break;
		case 'a':
			if (attr_list != NULL)
				errflg++;
			attr_list = optarg;
			break;
		case 'p':
			if (parent != NULL)
				errflg++;
			parent = optarg;
			break;
		case 'w':
			if (connection != NULL)
				errflg++;
			connection = optarg;
			break;
		case 'P':
			Pflag++;
			break;
		case 'T':
			Tflag++;
			break;
		default:
			errflg++;
		}
	}
 
	/* If error parsing options, exit with error
	 */
	if (errflg)
		exit(E_ARGS);
 
	/* If logical name not specified, exit with error
	 */
	if (lname == NULL)
		exit(E_LNAME);
 
	/* If  -T flag set, exit with error (not supported by gmpci)
	 */
	if (Tflag)
		exit(E_TFLAG);
 
	/*
	 * Get ODM objects for this device
	 */

	/* Start up ODM
	 */
	if (odm_initialize() == -1) {
		DBG(fprintf(dbg_fp, "%s: odm_initialize() failed\n", argv[0]));
		exit(E_ODMINIT);
	}
 
	/* Lock the database
	 */
	if ((odmdir_env = getenv("ODMDIR")) == NULL)
		sprintf(lock_file, "/etc/objrepos/config_lock");
	else
		sprintf(lock_file, "%s/%s", odmdir_env, "config_lock");

	if (odm_lock(lock_file, 60) == -1) {
		DBG(fprintf(dbg_fp, "%s: odm_lock() failed\n", argv[0]));
		err_exit(E_ODMLOCK);
	}
	DBG(fprintf(dbg_fp, "%s: ODM initialized and locked\n", argv[0]));

	/* Open CuDv class
	 */
	if ((int)(cusdev = odm_open_class(CuDv_CLASS)) == -1) {
		DBG(fprintf(dbg_fp, "%s: open class CuDv failed\n", argv[0]));
		err_exit(E_ODMOPEN);
	}
 
	/* Find CuDv for this device
	 */
	sprintf(criteria, "name = '%s'", lname);
	rc = (int)odm_get_first(cusdev, criteria, &CuDv);
	if (rc == 0) {
		DBG(fprintf(dbg_fp, "%s: failed to find CuDv object for %s\n",
			    				      argv[0], lname));
		err_exit(E_NOCuDv);
	}
	if (rc == -1) {
		DBG(fprintf(dbg_fp, "%s: ODM failure getting CuDv object",
			    					    argv[0]));
		err_exit(E_ODMGET);
	}

	/* Get PdDv for this device
	 */
	sprintf(criteria, "uniquetype = '%s'", CuDv.PdDvLn_Lvalue);
	rc = (int)odm_get_first(PdDv_CLASS, criteria, &PdDv);
	if (rc == 0) {
		DBG(fprintf(dbg_fp, "%s: failed to find PdDv object for this device\n",
			    					    argv[0]));
		err_exit(E_NOPdDv);
	}
	if (rc == -1) {
		DBG(fprintf(dbg_fp, "%s: ODM failure getting PdDv object",
			    					   argv[0]));
		err_exit(E_ODMGET);
	}

	/* Get the CuDv for old parent
	 */
	sprintf(criteria, "name = '%s'", CuDv.parent);
	rc = (int)odm_get_first(cusdev, criteria, &oldpar_CuDv);
	if (rc == 0) {
		DBG(fprintf(dbg_fp, "%s: no parent CuDv object\n",
			    argv[0]));
		err_exit(E_NOCuDvPARENT);
	}
	if (rc == -1) {
		DBG(fprintf(dbg_fp, "%s: ODM failure getting parent CuDv object\n",
			    argv[0]));
		err_exit(E_ODMGET);
	}

	/*
	 * Validate and parse new attributes
	 */
	if (attr_list != NULL) {
		
		/* If any new attributes fail validation, display them
		 * and exit with error
		 */
		if (attrval(CuDv.PdDvLn_Lvalue, attr_list, &badattr) > 0) {
			list_badattr(CuDv.PdDvLn_Lvalue, badattr);
			err_exit(E_ATTRVAL);
		}

		/* Determine number attribute-name=attribute-value pairs by
		 * counting number of ='s
		 */
		num_attr = 0;
		for (pchar = attr_list; *pchar != 0; pchar++) {
			if (*pchar == '=')
				num_attr++;
		}
		if (num_attr == 0) {
		DBG(fprintf(dbg_fp, "%s: failure parsing attribute list \"%s\"\n",
				    			  argv[0], attr_list));
			err_exit(E_NOATTR);
		}

		/* Allocate space for number of attributes in attr_list
		 */
		attr = (struct attr*)calloc(num_attr, sizeof(struct attr));
		

		/* For each attribute in attr_list, parse name and new value
		 * Note: this modifies attr_list; replaces all '=' and ' '
		 *       with nulls
		 */
		attr[0].name = (char *)strtok(attr_list, "=");
		attr[0].new_val = (char *)strtok(NULL, " ");
		for (i = 1; i < num_attr; i++) {

			attr[i].name = (char *)strtok(NULL, "=");
			attr[i].new_val = (char *)strtok(NULL, " ");
		}

		/* For each attribute in attr_list, get current attribute
		 * in case its value must be restored
		 */
		for (i = 0; i < num_attr; i++) {
			if ((pCuAt = getattr(lname, attr[i].name, FALSE,
					     	 &getattr_how_many)) == NULL) {
				DBG(fprintf(dbg_fp, "%s: failure getting attribute %s\n",
					    argv[0], attr[i].name));
				err_exit(E_BADATTR);
			}
			bcopy(pCuAt, &attr[i].CuAt, sizeof (struct CuAt));
			strcpy(attr[i].old_val, attr[i].CuAt.value);
		}
	}

	/* If parent or connection is NULL, it has not been changed
	 */
	if (parent == NULL)
		parent = CuDv.parent;
 
	if (connection == NULL)
		connection = CuDv.connwhere;

	/* If either parent or connection have changed, validate new values
	 */
	if (strcmp(parent, CuDv.parent) || strcmp(connection, CuDv.connwhere)) {

		/* Set Generate New Location Flag */
		new_location = 1;

		/* Save current parent and connection in case they need
		 * to be restored
		 */
		strcpy(oldparent, CuDv.parent);
		strcpy(oldconnect, CuDv.connwhere);

		/* Get the CuDv for new parent
		 */
		sprintf(criteria, "name = '%s'", parent);
		rc = (int)odm_get_first(cusdev, criteria, &par_CuDv);
		if (rc == 0) {
			DBG(fprintf(dbg_fp, "%s: no parent CuDv object\n",
				    				   argv[0]));
			err_exit(E_NOCuDvPARENT);
		}
		if (rc == -1) {
			DBG(fprintf(dbg_fp, "%s: ODM failure getting parent CuDv object\n",
				    				    argv[0]));
			err_exit(E_ODMGET);
		}

		/* If appropriate PdCn does not exist for new parent, exit
		 */
		sprintf(criteria,
		"uniquetype = '%s' AND connkey = '%s' AND connwhere = '%s'",
			par_CuDv.PdDvLn_Lvalue, PdDv.subclass, connection);
		rc = (int)odm_get_first(PdCn_CLASS, criteria, &PdCn);
		if (rc == 0) {
			DBG(fprintf(dbg_fp, "%s: Invalid connection\n", argv[0]));
			err_exit(E_INVCONNECT);
		}
		if (rc == -1) {
			DBG(fprintf(dbg_fp, "%s: ODM error getting PdCn object\n",
				    				     argv[0]));
			err_exit(E_ODMGET);
		}

		/* If this device is AVAILABLE and another device is configured
		 * at the new location, exit
		 */
		if (CuDv.status != DEFINED) {
			sprintf(criteria,
			  "parent = '%s' AND connwhere = '%s' AND status != %d",
					parent, connection, AVAILABLE);
			rc = (int)odm_get_first(cusdev, criteria, &other_CuDv);
			if (rc == -1) {
				DBG(fprintf(dbg_fp, "%s: ODM failure getting other CuDv object\n",
					    argv[0]));
				err_exit(E_ODMGET);
			}
			if (rc != 0) {
				DBG(fprintf(dbg_fp, "%s: device already AVAILABLE at this connection\n",
				    				    argv[0]));
				err_exit(E_AVAILCONNECT);
			}
		}
	}

	/* No parent or connection change */
	else {
		new_location = 0;

		/* If also no attribute change, just exit */
		if (attr_list == NULL)
			exit (E_OK);
	}

	/* Save device's current status */
	oldstatus = CuDv.status;
 
	/* If the device is configured and the -P flag is FALSE, unconfigure
	 * the device before updating database
	 */
	if (CuDv.status != DEFINED && ! Pflag) {
 
#ifdef DEBUG
#ifdef DBG_FILE
		DBG(fclose(dbg_fp));
#endif
#endif
		sprintf(criteria, "-l %s", CuDv.name);
		rc = odm_run_method(PdDv.Unconfigure, criteria, &out_p, &err_p);
#ifdef DEBUG
#ifdef DBG_FILE
		dbg_fp = fopen(DBG_FILE, "a+");
#endif
#endif
		if (rc == -1) {
			DBG(fprintf(dbg_fp, "%s: error running device unconfigure method\n",
				    				    argv[0]));
			err_exit(E_ODMRUNMETHOD);
		}
		if (rc != 0) {
			DBG(fprintf(dbg_fp,
			    "%s: device unconfigure method failed;\n   stdout: \"%s\";   stderr: \"%s\"\n",
				    			argv[0], out_p, err_p));
			err_exit(rc);
		}
	}

	/* If necessary, update CuDv to show new parent, connection,
	 * and location
	 */
	if (new_location) {
		strcpy(CuDv.parent, parent);
		strcpy(CuDv.connwhere, connection);
		sprintf(CuDv.location, "%s-%02d", par_CuDv.location, atoi(connection));

		if (odm_change_obj(CuDv_CLASS, &CuDv) == -1) {
			DBG(fprintf(dbg_fp, "%s: failure changing CuDv\n",
				    				    argv[0]));
			return(E_ODMUPDATE);
		}
	}
 
	/* Update each attribute in attr_list with the new value
	 */
	for (i = 0; i < num_attr; i++) {
		strcpy(attr[i].CuAt.value, attr[i].new_val);
		if (putattr(&attr[i].CuAt) == -1) {
			DBG(fprintf(dbg_fp, "%s: failure changing CuAt %s\n",
				    	     argv[0], attr[i].CuAt.attribute));
			err_exit(E_ODMUPDATE);
		}
	}
 
	/* If device was unconfigured, re-configure the device
	 */
	if (oldstatus != DEFINED && ! Pflag) {

#ifdef DEBUG
#ifdef DBG_FILE
		DBG(fclose(dbg_fp));
#endif
#endif
		sprintf(criteria, "-l %s", CuDv.name);
		rc = odm_run_method(PdDv.Configure, criteria, &out_p, &err_p);
#ifdef DEBUG
#ifdef DBG_FILE
		dbg_fp = fopen(DBG_FILE, "a+");
#endif
#endif
		if (rc == -1) {
			DBG(fprintf(dbg_fp, "%s: error running device configure method\n",
				    				    argv[0]));
			err_exit(E_ODMRUNMETHOD);
		}

		/* If configure method failed, restore CuDv and CuAt's
		 */
		if (rc != 0) {

			DBG(fprintf(dbg_fp,
			    "%s: device configure method failed;\n   stdout: \"%s\";   stderr: \"%s\"\n",
				    			argv[0], out_p, err_p));

			if (new_location) {
				strcpy(CuDv.parent, oldparent);
				strcpy(CuDv.connwhere, oldconnect);
				sprintf(CuDv.location, "%s-%02d",
					oldpar_CuDv.location, atoi(oldconnect));

				if (odm_change_obj(CuDv_CLASS, &CuDv) == -1) {
					DBG(fprintf(dbg_fp, "%s: failure changing CuDv\n",
						    argv[0]));
					return(E_ODMUPDATE);
				}
			}

			for (i = 0; i < num_attr; i++) {
				strcpy(attr[i].CuAt.value, attr[i].old_val);
				if (putattr(&attr[i].CuAt) == -1) {
					DBG(fprintf(dbg_fp, "%s: failure changing CuAt %s\n",
						    argv[0], attr[i].CuAt.attribute));
					err_exit(E_ODMUPDATE);
				}
			}
 
			/* Exit with configure method's error */
			err_exit(rc);
		}
	}
 
	/* Clean up and return successful
	 */
	if (odm_close_class(cusdev) == -1) {
		DBG(fprintf(dbg_fp, "%s: failure closing CuDv object class\n", argv[0]));
		err_exit(E_ODMCLOSE);
	}
 
	odm_terminate();
	DBG(fprintf(dbg_fp, "%s: successful exit\n", argv[0]));
	DBG(fflush(dbg_fp));
	exit(E_OK);
}
 
 
/*===========================================================================
 *
 * NAME:	err_exit
 *
 * FUNCTION:	Closes any open object classes and terminates ODM.  Used to
 *           	back out of an error.  Does not return.
 *
 * ARGUMENTS:	Exit code
 *
 * EXIT CODE:   from cf.h: E_OK if successful, otherwise specific error code
 *
 *===========================================================================*/

void
err_exit(
int     exit_code              /* Error exit code */
)
{
	if (attr != NULL) free (attr);
	odm_close_class(CuDv_CLASS);
	odm_close_class(CuAt_CLASS);
	odm_close_class(PdAt_CLASS);
	odm_terminate();
	exit(exit_code);
}

/*===========================================================================
 *
 * NAME:	list_badattr
 *
 * FUNCTION:	
 *           	
 *
 * ARGUMENTS:	uniquetype of device
 *		List of attributes failing validation
 *
 * RETURNS:   	from cf.h: E_OK if successful, otherwise specific error code
 *
 *===========================================================================*/

int
list_badattr(
char    *uniquetype,
char    *badattr
)
{
	char *attrname;
	char criteria[128];
	struct PdDv PdDv;
	struct PdAt PdAt;
	int rc;

	/* Get PdDv for this device for message set number
	 */
	sprintf(criteria, "uniquetype = '%s'", uniquetype);
	rc = (int)odm_get_first(PdDv_CLASS, criteria, &PdDv);
	if (rc == 0) {
		DBG(fprintf(dbg_fp, "%s: failed to find PdDv object for this device\n",
			    					    prog));
		return E_NOPdDv;
	}
	if (rc == -1) {
		DBG(fprintf(dbg_fp, "%s: ODM failure getting PdDv object",
			    					   prog));
		return E_ODMGET;
	}

	rc = E_OK;

	/* Loop through bad attribute names
	 */
	attrname = (char *)strtok(badattr, ", ");
	while (attrname != NULL) {

		/* Find the PdAt entry for the attribute
		 */
		sprintf(criteria, "uniquetype = %s AND attribute = %s",
			uniquetype, attrname);
		rc = (int)odm_get_first(PdAt_CLASS, criteria, &PdAt);
		if (rc == -1) {
			return E_ODMGET;
		}

		/* If unable to locate it, just identify it by name
		 */
		if (rc == 0) {
			fprintf(stderr, "           %s\n", attrname);
			rc = E_INVATTR;
		}

		/* Found it; identify it by name and message
		 */
		NLfprintf(stderr, "           %-20.20s%s\n", attrname,
			NLgetamsg(PdDv.catalog,
			PdDv.setno, PdAt.nls_index, ""));

		/* Get next attribute name
		 */
		attrname = (char *)strtok(NULL, ", ");
	}

	return rc;
}




 
