/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1999 by Myricom, Inc.                                   *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/*
 * This file includes the OS-specific driver code for FreeBSD
 *
 * Author: Andrew Gallatin <gallatin@freebsd.org>, <gallatin@cs.duke.edu>
 *
 * This driver was written to be a loadable module using the FreeBSD
 * 4.0 "newbus" api.  As such, it will require backporting to work under 
 * older versions of FreeBSD which do not support loadable PCI drivers
 *
 * For more information about FreeBSD device drivers, look at the
 * section 9 man pages included on the system (eg, man 9 tsleep).  The
 * FreeBSD Developers Handbook may also be useful
 * http://www.freebsd.org/doc/en_US.ISO8859-1/books/developers-handbook/index.html
 * It goes into detail on how to use the ddb (live) and gdb
 * (postmortem) debugging facilities in FreeBSD.
 * 
 * If all else fails, freebsd-hackers@FreeBSD.ORG is a mailing list
 * you can turn to.  Searching the archives at
 * http://www.freebsd.org/search/search.html  can also be quite helpful.
 */

/*
 * FreeBSD specific device driver defines 
 */


#include "gm_arch.h"

#include <sys/systm.h>
#if GM_PREVENT_NATIVE_MALLOC_USE
#undef malloc
#undef free
#include <sys/malloc.h>
#define malloc GM_DO_NOT_USE (malloc)
#define free GM_DO_NOT_USE (free)
#else 
#include <sys/malloc.h>
#endif /* GM_PREVENT_MALLOC_USE */
#include <sys/proc.h>
#include <machine/stdarg.h>
#include <machine/limits.h>
#include <vm/vm_pageout.h>		/* for vm_page_max_wired */
#include <sys/lock.h>
#include <vm/vm_map.h>			/* for vm_map type */
#include <sys/proc.h>			/* for process struct */
#include <sys/vnode.h>			/* for VCHR */
#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_pio.h"

#include <sys/sysctl.h>
#ifdef SYSCTL_NODE
extern int gm_bsd_ether_xmit_copies;
static int gm_freebsd_page_lock_failed = 0;

 SYSCTL_NODE(_net, OID_AUTO, gm, CTLFLAG_RW, 0, "GM");
 SYSCTL_INT(_net_gm, OID_AUTO, ethernet_xmit_copies, CTLFLAG_RW, 
		   &gm_bsd_ether_xmit_copies, 0, 
		   "Mbuf chain copies caused by misaligned or over length mbuf chains");
 SYSCTL_INT(_net_gm, OID_AUTO, ethernet_max_scatter, CTLFLAG_RD, 
		   NULL, GM_MAX_ETHERNET_SCATTER_CNT,
		   "Max mbufs that can be mapped for recv.");
 SYSCTL_INT(_net_gm, OID_AUTO, ethernet_max_gather, CTLFLAG_RD, 
		   NULL, GM_MAX_ETHERNET_GATHER_CNT, 
		   "Max mbufs that can be mapped for xmit.");
 SYSCTL_INT(_net_gm, OID_AUTO, page_lock_failed, CTLFLAG_RW, 
		   &gm_freebsd_page_lock_failed, 0, "failed attempts to lock memory");
#endif
int gm_freebsd_ether_attach (gm_instance_state_t *is);
int gm_freebsd_ether_detach (gm_instance_state_t *is);

/*
 * The GM includes
 */

#include "gm_page_hash.h"
#include "gm_lanai.h"
#include "gm_instance.h"
#include "gm_malloc_debug.h"
#include "gm_debug_lanai_dma.h"
#include "gm_debug_malloc.h"

/***********************************************************************
 * Globals
 ***********************************************************************/

/*
 * GM devices
 */

/* un-staticized for debugging.. */
/*static */ gm_instance_state_t *gm_instances[GM_ARCH_MAX_INSTANCE];
/*static */ int gm_num_instance = 0;
/*static */ int gm_freebsd_max_user_locked_pages;


#if (__FreeBSD_version >= 500000)
#include <sys/mutex.h>
static struct mtx wake_mtx;
#endif

dev_t gm_freebsd_specials[GM_ARCH_MAX_MINOR];

int num_gm = 0;

#define GET_MINOR(minor_id) (minor_id - (GM_ARCH_MAX_INSTANCE*2 - gm_num_instance*2))
#define EXTERN_MINOR(minor_id) (minor_id + (GM_ARCH_MAX_INSTANCE*2 - gm_num_instance*2))
#define INVALID_MINOR(minor_id) ((minor_id +(GM_ARCH_MAX_INSTANCE*2 - gm_num_instance*2)) >= GM_ARCH_MAX_MINOR)

/*
 * FreeBSD Device Driver routines.
 * This section is coded to the FreeBSD 4.x "newbus" API and will
 * *not* work on earlier version of FreeBSD.  FreeBSD 4.x was chosen
 * because it allows loadable PCI device drivers.
 */

static int gm_freebsd_probe(device_t dev);
static int gm_freebsd_attach(device_t dev);
static int gm_freebsd_detach(device_t dev);
static int gm_freebsd_shutdown(device_t dev);
static void gm_freebsd_intr(void *arg);

static device_method_t gm_freebsd_methods[] =
{
		/* Device interface */
	DEVMETHOD(device_probe, gm_freebsd_probe),
	DEVMETHOD(device_attach, gm_freebsd_attach),
	DEVMETHOD(device_detach, gm_freebsd_detach),
	DEVMETHOD(device_shutdown, gm_freebsd_shutdown),

	{0, 0}
};

static driver_t gm_freebsd_driver =
{
	"gm",
	gm_freebsd_methods,
	sizeof(gm_instance_state_t),
};

static devclass_t gm_freebsd_devclass;

DRIVER_MODULE(gm, pci, gm_freebsd_driver, gm_freebsd_devclass, 0, 0);

static d_open_t gm_freebsd_open;
static d_close_t gm_freebsd_close;
static d_ioctl_t gm_freebsd_ioctl;
static d_mmap_t gm_freebsd_mmap;

#define CDEV_MAJOR 201
static struct cdevsw gm_cdevsw =
{
		.d_open = gm_freebsd_open,
		.d_close =  gm_freebsd_close,
		.d_read = noread,
		.d_write = nowrite,
		.d_ioctl =  gm_freebsd_ioctl,
		.d_poll = nopoll,
		.d_mmap =  gm_freebsd_mmap,
		.d_strategy =  nostrategy,
		.d_name =  "gm",
		.d_maj = CDEV_MAJOR,
		.d_dump = nodump,
		.d_flags = 0,
#if (__FreeBSD_version < 500000)
		.d_bmaj = -1,
#endif
#if (__FreeBSD_version >= 430000)
		.d_kqfilter = nokqfilter,
#endif
};


/* debugging toggles */
#define GM_DEBUG_ATTACH 0
#define GM_DEBUG_IOCTL 0
#define GM_DEBUG_SYNC 0
#define GM_DEBUG_DMA_REGION 0
#define GM_DEBUG_MAPIO 0
#define GM_DEBUG_OPEN 0
#define GM_DEBUG_CLOSE 0
#define GM_DEBUG_MMAP 0
#define GM_DEBUG_INTR 0
#define GM_DEBUG_MEMLOCK 0

#ifdef SYSCTL_NODE

static int gm_wake_credit_incr;
static int gm_wake_max_credits;
static int gm_wake_max_sdmas_until_wake;
static char gm_wake_credits_str[10*GM_ARCH_MAX_INSTANCE+1];


static int
sysctl_gm_wake_credit_incr (SYSCTL_HANDLER_ARGS)
{
		static int first_call = 1;
		int error, i;
		gm_instance_state_t *is;

		is = gm_instances[0]; /* XXX HACK */

		if (first_call)	{
				first_call = 0;
				gm_wake_credit_incr = 
						gm_read_lanai_global_u8 (is, 
												 ethernet.wake.credit_incr);
				gm_wake_max_credits = 
						gm_read_lanai_global_u8 (is, 
												 ethernet.wake.max_credits);
				gm_wake_max_sdmas_until_wake =
						gm_read_lanai_global_u16 (is, 
												 ethernet.wake.max_sdmas_until_wake);
		}

		error = sysctl_handle_int(oidp, arg1, arg2, req);
		if (error)
                return (error);

		for (i = 0; i < GM_ARCH_MAX_INSTANCE; i++) {
				is = gm_instances[i];
				if (is == NULL)
						continue;
				gm_pause_lanai(is);
				gm_write_lanai_global_u8 (is, ethernet.wake.credit_incr,
										  (gm_u8_t)gm_wake_credit_incr);
				gm_unpause_lanai(is);
		}
		return 0;
}

SYSCTL_OID(_net_gm, OID_AUTO, ethernet_wake_credit_incr,		\
		   CTLTYPE_INT | CTLFLAG_RW, &gm_wake_credit_incr, 0,	\
		   sysctl_gm_wake_credit_incr, "I", "change wake credit incrment");

static int
sysctl_gm_wake_max_credits (SYSCTL_HANDLER_ARGS)
{
		int error, i;
		gm_instance_state_t *is;

		error = sysctl_handle_int(oidp, arg1, arg2, req);
		if (error)
                return (error);
		for (i = 0; i < GM_ARCH_MAX_INSTANCE; i++) {
				is = gm_instances[i];
				if (is == NULL)
						continue;
				gm_pause_lanai(is);
				gm_write_lanai_global_u8 (is, ethernet.wake.max_credits,
										  (gm_u8_t)gm_wake_max_credits);
		
				gm_unpause_lanai(is);
		}
		return 0;
}

SYSCTL_OID(_net_gm, OID_AUTO, ethernet_wake_max_credits, 		\
		   CTLTYPE_INT | CTLFLAG_RW, &gm_wake_max_credits, 0,	\
		   sysctl_gm_wake_max_credits, "I", "change max wake credits");

static int
sysctl_gm_wake_max_sdmas_until_wake (SYSCTL_HANDLER_ARGS)
{
		int error, i;
		gm_instance_state_t *is;

		error = sysctl_handle_int(oidp, arg1, arg2, req);
		if (error)
                return (error);

		for (i = 0; i < GM_ARCH_MAX_INSTANCE; i++) {
				is = gm_instances[i];
				if (is == NULL)
						continue;
				gm_pause_lanai(is);
				gm_write_lanai_global_u16 (is, 
										   ethernet.wake.max_sdmas_until_wake,
										   (gm_u8_t)
										   gm_wake_max_sdmas_until_wake);
				gm_unpause_lanai(is);
		}
		return 0;
}

SYSCTL_OID(_net_gm, OID_AUTO, ethernet_wake_max_sdmas_until_wake, 		\
		   CTLTYPE_INT | CTLFLAG_RW, &gm_wake_max_sdmas_until_wake, 0,	\
		   sysctl_gm_wake_max_sdmas_until_wake, "I", "change max wake credits");


static int
sysctl_gm_wake_credits (SYSCTL_HANDLER_ARGS)
{
		int error, i;
		gm_instance_state_t *is;
		char buf[11];

		bzero (gm_wake_credits_str, sizeof gm_wake_credits_str);
		for (i = 0; i < GM_ARCH_MAX_INSTANCE; i++) {
				is = gm_instances[i];
				if (is == NULL) {
						strcat (gm_wake_credits_str, "--- ");
						continue;
				}
				sprintf(buf, "%d ",
						gm_read_lanai_global_u8 (is, ethernet.wake.credits)); 
				strcat (gm_wake_credits_str, buf);
		}

		error = sysctl_handle_string(oidp, arg1, arg2, req);
		return error;
}		

SYSCTL_OID(_net_gm, OID_AUTO, ethernet_wake_credits, 				\
		   CTLTYPE_STRING | CTLFLAG_RD, &gm_wake_credits_str[0], 0,	\
		   sysctl_gm_wake_credits, "A", "fetch wake credits");


#endif


static void
gm_freebsd_stop_dma(device_t dev)
{
	/*
	 * disable bus mastering.
	 */
	u_long val;
	val = pci_read_config(dev, PCIR_COMMAND, 2);
	val &= ~(PCIM_CMD_MEMEN | PCIM_CMD_BUSMASTEREN);
	pci_write_config(dev, PCIR_COMMAND, val, 2);
}


/*
 * the boot-time or load-time probe routine.
 */
/*
 * Return identification string if this is device is ours.
 */
static int
gm_freebsd_probe(device_t dev)
{
	if (((pci_get_vendor(dev) == GM_PCI_VENDOR_MYRICOM) &&
			(pci_get_device(dev) == GM_PCI_DEVICE_MYRINET)) ||
	    ((pci_get_vendor(dev) == GM_PCI_VENDOR_MYRICOM2) &&
			(pci_get_device(dev) == GM_PCI_DEVICE_MYRINET))) {
		device_set_desc(dev, "Myrinet PCI interface");
		return 0;
	}

	return ENXIO;
}

gm_status_t gm_arch_page_len(unsigned long *result);

/*
 * The device attach function.  This function returns 0 on success,
 * nonzero on failure. Note that error return codes are not propagated
 * back to the user via kldload, so carefully translating them is
 * mostly useless -- a printf is much more useful.
 */

static int
gm_freebsd_attach(device_t dev)
{
	int error = ENXIO;
	gm_instance_state_t *is = device_get_softc(dev);
	struct ifnet *ifp;
	int i, s;
	u_long val;
	int perms, rid;
	gm_status_t status;
	char name[16];

	s = splimp();

	is->arch.dev = dev;

	/*
	 * disable DMA
	 */

	gm_freebsd_stop_dma(dev);

	/*
	 * Make sure we don't attach more than GM_ARCH_MAX_INSTANCE
	 * devices
	 */

	if (gm_num_instance >= GM_ARCH_MAX_INSTANCE) {
			device_printf(dev, "Cannot attach more than %d boards\n",
						  GM_ARCH_MAX_INSTANCE);
			device_printf(dev, "Increase  GM_ARCH_MAX_INSTANCE\n");
			error = ENXIO;
			goto fail_with_nothing;
	}
			

	if (gm_num_instance == 0) {
			status = gm_init();
			GM_INFO (("GM: driver version %s\n", _gm_build_id));
			if (status != GM_SUCCESS) {
					device_printf(dev, "could not init gm\n");
					error = ENXIO;
					goto fail_with_nothing;
			}
	}
	gm_num_instance++;

	/*
	 * Map control/status registers.
	 */

#define GM_PCI_SHMEM 0x10
	rid = GM_PCI_SHMEM;			/* shared memory base address */
	is->arch.mem = bus_alloc_resource(dev, SYS_RES_MEMORY, &rid,
									  0, ~0, 1, RF_ACTIVE|PCI_RF_DENSE);
	if (!is->arch.mem) {
		device_printf(dev, "could not map memory\n");
		error = ENXIO;
		goto fail_with_gm;
	}

	/* 
	   *  map the device into the kernel virtual address space.  FreeBSD
	   *  would rather we used bus_space access, but the MI portions of GM
	   *  don't support that. 
	 */
	is->arch.csr = rman_get_virtual(is->arch.mem);

	GM_PRINT(GM_DEBUG_ATTACH, ("GM: board mapped at KVA: %p\n", is->arch.csr));


	/*
	 * Allocate our interrupt.
	 */
	rid = 0;
	is->arch.irq = bus_alloc_resource(dev, SYS_RES_IRQ, &rid, 0, ~0, 1,
									  RF_SHAREABLE | RF_ACTIVE);
	if (is->arch.irq == NULL) {
		device_printf(dev, "could not map interrupt\n");
		goto fail_with_mem;
	}

	/* FreeBSD not quite ready for INTR_MPSAFE network drivers yet, 
	   giant is needed by VM system inside of mclget(), uncomment
	   INTR_MPSAFE when this is fixed */

	error = bus_setup_intr(dev, is->arch.irq, 
						   INTR_TYPE_NET /*| INTR_MPSAFE */,
						   gm_freebsd_intr, is, &is->arch.ih);
	if (error) {
		device_printf(dev, "could not setup irq\n");
		goto fail_with_intr;
	}

	/*
	 *  Some GM setup
	 */

	(void) gm_arch_page_len(&GM_PAGE_LEN);	/* can't fail, really.. */


	if (gm_instance_init(is, gm_num_instance - 1, 
						 GM_MYRINET_BUS_PCI) != GM_SUCCESS) {
		GM_NOTE (("gm_instance_init failed\n"));
		goto fail_with_irq;
	}

	/*
	 * enable interrupts
	 */
	if (gm_initialize_interrupts(is) != GM_SUCCESS) {
		GM_NOTE (("Cannot enable interrupts.\n"));
		goto fail_with_is;
	}

	gm_instances[gm_num_instance - 1] = is;
	/*
	 * Enable bus mastering.
	 */
	val = pci_read_config(dev, PCIR_COMMAND, 2);
	val |= (PCIM_CMD_MEMEN | PCIM_CMD_BUSMASTEREN);
	pci_write_config(dev, PCIR_COMMAND, val, 2);

	/*
	 *  success!!
	 */


  /************
   * Create minor devices.  /dev/gm%d is the clone device for normal users,
   * and /dev/gmp%d is the clone device for privileged users.
   ************/

	/* Determine minor node number to use for /dev/gm? . */

	GM_PRINT (GM_DEBUG_ATTACH, ("Allocating minor number.\n"));
	if (gm_minor_alloc(is, &is->clone_minor) != GM_SUCCESS) {
		GM_NOTE (("Failed to allocate minor number.\n"));
		goto fail_with_is;
	}
	GM_PRINT (GM_DEBUG_ATTACH, 
			  ("using minor number %ld.\n", (long) is->clone_minor));
	gm_minor_get_port_state(is->clone_minor)->instance = is;
	gm_minor_get_port_state(is->clone_minor)->privileged = 0;

	/* Determine minor node number to use for /dev/gmp? . */

	GM_PRINT (GM_DEBUG_ATTACH, ("Allocating minor number.\n"));
	if (gm_minor_alloc(is, &is->privileged_clone_minor) != GM_SUCCESS) {
		GM_NOTE (("Failed to allocate minor number.\n"));
		goto fail_with_minor;
	}
	GM_PRINT (GM_DEBUG_ATTACH, ("using minor number %ld.\n",
				 (long) is->privileged_clone_minor));
	gm_minor_get_port_state(is->privileged_clone_minor)->instance = is;
	gm_minor_get_port_state(is->privileged_clone_minor)->privileged = 1;


	splx(s);

	if (gm_num_instance == 1) {

#if (__FreeBSD_version >= 500000)
		mtx_init (&wake_mtx, "GM Wake Mutex", NULL, MTX_DEF);
#endif
	/*  
	 *  this is an arbitrary limit to prevent the user from
	 *  bringing the system to a complete stop 
	 */

		gm_freebsd_max_user_locked_pages = 2 * vm_page_max_wired;

		GM_INFO (("GM: gm_register_memory will be able to lock "
				  "%d pages (%d MBytes)\n", gm_freebsd_max_user_locked_pages,
				  (gm_freebsd_max_user_locked_pages * PAGE_SIZE) / (1024 * 1024)));

	/*
	 * Create character special devices
	 */
		for (i = 0; i < GM_ARCH_MAX_MINOR; i++){
				if (i < 2 * GM_ARCH_MAX_INSTANCE) {
						if (i & 1) {
								sprintf(name, "gmp%d", i/2);
								perms = 0600;
						} else {
								sprintf(name, "gm%d", i/2);
								perms = 0666;
						}
				}	 else {
						sprintf(name, "gm%%d");
						perms = 0666;
				}
				/* make_dev cannot fail, we don't need to check return
				   values 	*/
				gm_freebsd_specials[i] = make_dev(&gm_cdevsw, i, 
										  UID_ROOT, GID_WHEEL, perms, name, i);
		}
	}

	GM_PRINT (GM_DEBUG_ATTACH, ("Attaching the IP interface\n"));
	if (gm_freebsd_ether_attach(is)) {
		GM_NOTE (("GM: IP interface attach FAILED\n"));
	} else {
		GM_INFO (("GM: IP interface attach ok\n"));
	}

	return 0;


fail_with_minor:
	gm_minor_free(is->clone_minor);
  fail_with_is:
	gm_instance_finalize(is);
#ifdef  notyet
	gm_destroy_hash(is->arch.page_lock_hash);
#endif

  fail_with_irq:
	bus_release_resource(dev, SYS_RES_IRQ, 0, is->arch.irq);
  fail_with_intr:
	bus_teardown_intr(dev, is->arch.irq, is->arch.ih);
	is->arch.irq = NULL;
  fail_with_mem:
	bus_release_resource(dev, SYS_RES_MEMORY, GM_PCI_SHMEM, is->arch.mem);
  fail_with_gm:
	gm_num_instance--;
	if (gm_num_instance == 0)
		gm_finalize();
  fail_with_nothing:
	gm_freebsd_stop_dma(dev);
	splx(s);
	return error;
}

/*
 * Destroy character special device files.  If any file is open, fail.
 * Must be called at high ipl, so we don't race with an open on a gm port
 */

static int
gm_freebsd_destroy_specials(void)
{
	int i, ref_count;
	struct vnode *vp;

	/* check for open device files */

	for (i = 0; i < GM_ARCH_MAX_MINOR; i++) {
		if(vfinddev(gm_freebsd_specials[i], VCHR, &vp)){
			if (ref_count = vcount(vp)) {
					GM_WARN (("GM: %d refs to minor unit %d. "
							  "module cannot be unloaded until all ports are closed\n",
							  ref_count, i));
					return EBUSY;
				}
			vgone(vp);
		}
	}

	/* destroy device files */

	for (i = 0; i < GM_ARCH_MAX_MINOR; i++) {
		destroy_dev(gm_freebsd_specials[i]);
	}

	return 0;
}

static int
gm_freebsd_detach(device_t dev)
{
	gm_instance_state_t *is = device_get_softc(dev);
	int s, i;
	extern struct cdevsw *cdevsw[];
	int dev_index;
	static int specials_destroyed  = 0;
   
	GM_CALLED ();

	s = splimp();

	if (is->arch.irq == NULL) {
		GM_RETURN_INT (0);
	}

	if (specials_destroyed == 0) {
		if (gm_freebsd_destroy_specials() == EBUSY) {
			splx(s);
			return EBUSY;
		}
		specials_destroyed = 1;

	}

    dev_index = device_get_unit(dev);

	gm_minor_free (is->privileged_clone_minor);
	gm_minor_free (is->clone_minor);
	/* must allow interrupts while ethernet port is closing, otherwise
	   lanai commands might not work properly */
	gm_freebsd_ether_detach(is);
	gm_finalize_interrupts(is);
	gm_instance_finalize(is);

	/*
	 * Stop DMA 
	 */
	gm_freebsd_stop_dma(dev);

	gm_num_instance--;

	if (gm_num_instance == 0) {
		gm_finalize();
	}
	splx(s);

	if ((dev_index >=0) && (dev_index < GM_ARCH_MAX_INSTANCE)) {
		gm_instances[dev_index] = NULL;
	}

	/*
	 * Deallocate resources.
	 */
	bus_teardown_intr(dev, is->arch.irq, is->arch.ih);
	bus_release_resource(dev, SYS_RES_IRQ, 0, is->arch.irq);
	bus_release_resource(dev, SYS_RES_MEMORY, GM_PCI_SHMEM, is->arch.mem);

#if (__FreeBSD_version >= 500000)
	if (gm_num_instance == 0) {
		mtx_destroy (&wake_mtx);
	}
#endif

	GM_RETURN_INT (0);
}


/*
 * Device shutdown routine. Called at system shutdown after sync. The
 * main purpose of this routine is to shut off receiver DMA so that
 * kernel memory doesn't get clobbered during warmboot.
 */
static int
gm_freebsd_shutdown(device_t dev)
{
	/*
	 * Make sure that DMA is disabled prior to reboot. Not doing
	 * do could allow DMA to corrupt kernel memory during the
	 * reboot before the driver initializes.
	 */
	gm_freebsd_stop_dma(dev);
	return 0;
}


/*
 * Misc required functions
 */

/* 
 *  This is called just after the port state is created (in gm_minor.c)
 *  to perform architecture-specific initialization. 
 */

gm_status_t
gm_arch_port_state_init(gm_port_state_t * ps)
{
	return GM_SUCCESS;
}

/* 
 *  This is called just before the port state is destroyed (in
 *  gm_minor.c) to perform architecture-specific finalization. 
 */

void
gm_arch_port_state_fini(gm_port_state_t * ps)
{
	return;
}

/* 
 * This is called at the end of gm_port_state_open() to perform architecture-
 * specific initialization. 
 */

gm_status_t
gm_arch_port_state_open(gm_port_state_t * ps)
{
	return GM_SUCCESS;
}

/* 
 * This is called at the start of gm_port_state_close to perform
 *  architecture-specific finalization. 
 */

void
gm_arch_port_state_close(gm_port_state_t * ps)
{
	return;
}


/*
 * Utility functions for gm_ioctl( ) 
 */

gm_status_t
gm_arch_copyin(const void *what,	/* copy from this user address  */
	       void *where,		/* to this kernel address       */
	       gm_size_t amount)	/* for this length              */
{				
	int errno;

	GM_PRINT (GM_DEBUG_IOCTL, 
		("copyin: from %p to %p len %ld\n", what, where, amount));
	errno = copyin(what, where, amount);
	return errno ? GM_FAILURE : GM_SUCCESS;
}

gm_status_t
gm_arch_copyout(const void *what,	/* copy from this kernel address */
		void *where,		/* to this user address          */
		gm_size_t amount) 	/* for this length               */
{					
	int errno;

	GM_PRINT (GM_DEBUG_IOCTL, 
		("copyout: from %p to %p len %ld\n", what, where, amount));

	errno = copyout(what, where, amount);
	return errno ? GM_FAILURE : GM_SUCCESS;
}

/*
 *  Synchronization functions
 */

#if (__FreeBSD_version < 500000)

/* versions of FreeBSD prior to 5.0 don't have native mutexs and 
   use interrupt masking for critical sections */

void
gm_arch_sync_init(gm_arch_sync_t * s, gm_instance_state_t * is)
{
	s->held = s->waiters = 0;
}

void
gm_arch_sync_reset(gm_arch_sync_t * s)
{
	s->held = s->waiters = 0;
}

void
gm_arch_sync_destroy(gm_arch_sync_t * s)
{
}


static void
lock_acquire(gm_arch_sync_t * lock)
{
	int s = splimp();

#if !defined(MAX_PERF)
	if (curproc && lock->held == curproc->p_pid)
		panic("locking against gm_self");
#endif

	while (lock->held) {
		lock->waiters++;
		tsleep(lock, PWAIT, "gm_arch_sync_t", 0);
		lock->waiters--;
	}
	lock->held = (curproc ? curproc->p_pid : -1);
	splx(s);
}

static void
lock_release(gm_arch_sync_t * lock)
{
	int s = splimp();
#if !defined(MAX_PERF)
	if (lock->held == 0)
		panic("releasing a free lock");
	if (curproc && lock->held != curproc->p_pid)
		panic("I wasn't the lockholder");
#endif
	if (lock->waiters)
		wakeup(lock);
	lock->held = 0;
	splx(s);
}

void
gm_arch_mutex_enter(gm_arch_sync_t * s)
{
	lock_acquire(s);
}

void
gm_arch_mutex_exit(gm_arch_sync_t * s)
{
	lock_release(s);
}

void
gm_arch_wake(gm_arch_sync_t * s)
{
	int ipl;
	GM_PRINT (GM_DEBUG_SYNC, 
			  ("gm_arch_wake() on chan. 0x%lx.\n", (vm_offset_t) s));
	ipl = splimp();
	wakeup(s);
	s->wake_cnt++;
	splx(ipl);
}


gm_arch_sleep_status_t
gm_arch_timed_sleep(gm_arch_sync_t * s, int seconds)
{
	int ipl;
	ipl = splimp();

	GM_PRINT (GM_DEBUG_SYNC, ("gm_arch_timed_sleep() on chan. 0x%lx\n", 
							  (vm_offset_t)s)); 

	while (s->wake_cnt <= 0) {
		if (tsleep(s, PZERO, "gm_arch_timed_sleep", hz * seconds)) {
			GM_PRINT (GM_DEBUG_SYNC, 
					  ("gm_arch_timed_sleep returning GM_SLEEP_TIMED_OUT\n"));
			splx(ipl);
			return GM_SLEEP_TIMED_OUT;
		}
	}
	s->wake_cnt--;
	splx(ipl);
	GM_PRINT (GM_DEBUG_SYNC, ("gm_arch_timed_sleep() returning.\n"));
	return GM_SLEEP_WOKE;
}

gm_arch_sleep_status_t
gm_arch_signal_sleep(gm_arch_sync_t * s)
{
	int ipl;
	ipl = splimp();

	GM_PRINT (GM_DEBUG_SYNC,
			  ("gm_arch_signal_sleep() on chan. 0x%lx\n", (vm_offset_t)s)); 

	while (s->wake_cnt <= 0) {
		if (tsleep(s, PZERO | PCATCH, "gm_arch_signal_sleep", 0)) {
			GM_PRINT (GM_DEBUG_SYNC, 
					  ("gm_arch_signal_sleep ret GM_SLEEP_INTERRUPTED\n"));
			splx(ipl);
			return GM_SLEEP_INTERRUPTED;
		}
	}
	s->wake_cnt--;
	splx(ipl);
	GM_PRINT (GM_DEBUG_SYNC, ("gm_arch_signal_sleep() returning.\n"));
	return GM_SLEEP_WOKE;
}

#else
/* FreeBSD 5.0 and later have native mutexes and an SMP aware kernel.
   However, it also does not allow a process or thread holding 
   a mutex to go to sleep */

void
gm_arch_sync_init(gm_arch_sync_t * s, gm_instance_state_t * is)
{
	bzero (s, sizeof (*s));
	mtx_init (&s->sleep_mutex, "GM sleep mutex", NULL, MTX_DEF );
	cv_init (&s->sleep_cv, "GM sleep condvar");
	mtx_init (&s->lock_mutex, "GM lock mutex", NULL, MTX_DEF );
	cv_init (&s->lock_cv, "GM lock condvar");
}

void
gm_arch_sync_destroy(gm_arch_sync_t * s)
{
	mtx_destroy (&s->sleep_mutex);
	cv_destroy (&s->sleep_cv);
	mtx_destroy (&s->lock_mutex);
	cv_destroy (&s->lock_cv);
}

void
gm_arch_sync_reset(gm_arch_sync_t * s)
{
	s->wake_cnt = 0;
}


void
gm_arch_mutex_enter(gm_arch_sync_t * s)
{
	mtx_lock (&s->lock_mutex);
	while (s->held) {
		s->waiters++;
		cv_wait (&s->lock_cv, &s->lock_mutex);
		s->waiters--;
	}
	s->held = 1;
	mtx_unlock (&s->lock_mutex);
}

void
gm_arch_mutex_exit(gm_arch_sync_t * s)
{
	mtx_lock (&s->lock_mutex);
	s->held = 0;
	cv_signal (&s->lock_cv);
	mtx_unlock (&s->lock_mutex);

}

void
gm_arch_wake(gm_arch_sync_t * s)
{
	GM_PRINT (GM_DEBUG_SYNC, 
			  ("gm_arch_wake() on chan. 0x%lx.\n", (vm_offset_t) s));

	mtx_lock (&s->sleep_mutex);
	s->wake_cnt++;
	cv_signal (&s->sleep_cv);
	mtx_unlock (&s->sleep_mutex);
}


gm_arch_sleep_status_t
gm_arch_timed_sleep(gm_arch_sync_t * s, int seconds)
{
	gm_arch_sleep_status_t ret = GM_SLEEP_WOKE;

	GM_PRINT (GM_DEBUG_SYNC, ("gm_arch_timed_sleep() on chan. 0x%lx\n", 
							  (vm_offset_t)s)); 

	mtx_lock (&s->sleep_mutex);
	while (s->wake_cnt <= 0) {
		if (cv_timedwait (&s->sleep_cv, &s->sleep_mutex, seconds * hz) != 0) {
			/* if a process is exiting, FreeBSD always returns
			 * EWOULDBLOCK whether it timed out or not */
			if (s->wake_cnt <= 0) {
				ret = GM_SLEEP_TIMED_OUT;
			}
			break;
		}
	}
	s->wake_cnt--;
	mtx_unlock (&s->sleep_mutex);
	GM_PRINT (GM_DEBUG_SYNC, 
			  ("gm_arch_timed_sleep() returning %d.\n", ret));

	return ret;
}

gm_arch_sleep_status_t
gm_arch_signal_sleep(gm_arch_sync_t * s)
{
	gm_arch_sleep_status_t ret = GM_SLEEP_WOKE;

	GM_PRINT (GM_DEBUG_SYNC, ("gm_arch_signal_sleep() on chan. 0x%lx\n", 
							  (vm_offset_t)s)); 

	mtx_lock (&s->sleep_mutex);
	while (s->wake_cnt <= 0) {
		if (cv_wait_sig (&s->sleep_cv, &s->sleep_mutex) != 0) {
			/* if a process is exiting, FreeBSD always returns
			 * EINTR, whether it was interrupted or not */
			if (s->wake_cnt <= 0) {
				ret = GM_SLEEP_INTERRUPTED;
			}
			break;
		}
	}
	s->wake_cnt--;
	mtx_unlock (&s->sleep_mutex);
	GM_PRINT (GM_DEBUG_SYNC, 
			  ("gm_arch_signal_sleep() returning %d.\n", ret));

	return ret;
}


#endif


void
gm_arch_spin(gm_instance_state_t * is, gm_u32_t usecs)
{
	DELAY(usecs);
}


/*
 * DMA region functions 
 */


static gm_dp_t
region_dma_addr(gm_arch_dma_region_t * r, void *addr)
{

#ifdef __alpha__
	return (gm_dp_t) alpha_XXX_dmamap((vm_offset_t) addr);
#endif

#ifdef __i386__
	return (gm_dp_t) vtophys(addr);
#endif

}

gm_status_t
gm_arch_dma_region_alloc(gm_instance_state_t * is,
						 gm_arch_dma_region_t * r,
						 gm_size_t len, gm_u32_t flags,
						 gm_register_page_function_t register_page_func,
						 void *user_arg)
{
	void *addr;
	int pages, page_num;
	GM_PRINT (GM_DEBUG_DMA_REGION, 
			  ("in gm_arch_dma_region_alloc; len = %d,%s\n", len,
				 ((flags & GM_ARCH_DMA_CONTIGUOUS)
				  ? "contiguous"
				  : "non-contiguous")));

	pages = round_page(len) / GM_PAGE_LEN;
	gm_always_assert(pages >= 1);

	/* 
	 * grab the memory and remember how we allocated it 
	 */
	r->is = is;
	r->flags = flags;
	if (len < GM_PAGE_LEN)
		GM_PRINT (GM_DEBUG_DMA_REGION,
				  ("gm_arch_dma_region_alloc: length(%d) < GM_PAGE_LEN\n", 
				   len));

	r->len = round_page(len);
	if (flags & GM_ARCH_DMA_CONTIGUOUS) {
		/*
		 * use contigmalloc to get contig memory in DMA space
		 */
		r->type = GM_ARCH_CONTIGMALLOC_REGION;
		r->alloc_addr =
			(void *) vm_page_alloc_contig(r->len, 0x100000,
										  0xffffffff, PAGE_SIZE);
		r->addr = r->alloc_addr;
	}
	else {
		/*
		 * use normal malloc
		 */
		r->type = GM_ARCH_MALLOC_REGION;
		r->alloc_addr = gm_malloc(r->len + GM_PAGE_LEN);
		r->addr = (void *) round_page((vm_offset_t) r->alloc_addr);
	}

	if (r->addr == NULL)
	    return (GM_OUT_OF_MEMORY);

	r->addr2 = r->addr;
	if (register_page_func) {
		page_num = 0;
		addr = r->addr;
		while (page_num < pages) {
			gm_dp_t bus_addr = region_dma_addr(r, addr);
			register_page_func(user_arg, bus_addr, page_num);
			page_num++;
			addr += PAGE_SIZE;
		}
	}

	return (GM_SUCCESS);
}

void
gm_arch_dma_region_free(gm_arch_dma_region_t * r)
{
	switch (r->type) {
	 case GM_ARCH_MALLOC_REGION:
		 gm_free(r->alloc_addr);
		 break;
	 case GM_ARCH_CONTIGMALLOC_REGION:
		 kmem_free(kernel_map, (vm_offset_t) r->alloc_addr, r->len);
		 break;
	 default:
		 GM_PANIC (("gm_arch_dma_region_free:bad type\n"));
		 break;
	}
}

void *
gm_arch_dma_region_kernel_addr(gm_arch_dma_region_t * r)
{
	return r->addr;
}

gm_s32_t
gm_arch_dma_region_status(gm_arch_dma_region_t * r)
{
	return 0xf;		/* FreeBSD doesn't run on SBUS equipped hosts */
}

gm_dp_t
gm_arch_dma_region_dma_addr(gm_arch_dma_region_t * r)
{
	gm_dp_t ret;
	ret = region_dma_addr(r, r->addr);
	GM_PRINT (GM_DEBUG_DMA_REGION, 
			  ("gm_arch_dma_region_dma_addr:0x%p->0x%x\n", r->addr, ret));
	return ret;
}

gm_dp_t
gm_arch_dma_region_dma_addr_advance(gm_arch_dma_region_t * r)
{
	void *previous = r->addr2;
	r->addr2 += PAGE_SIZE;
	return region_dma_addr(r, previous);
}

gm_status_t
gm_arch_dma_region_sync(gm_arch_dma_region_t * r, int command)
{
	/* this is a no-op on x86 and alpha */
	return GM_SUCCESS;
}

#if GM_PREVENT_NATIVE_MALLOC_USE
#undef malloc
#undef free
#endif /* GM_PREVENT_MALLOC_USE */

void *
__gm_arch_kernel_malloc(unsigned long len, int flags)
{
	return malloc((unsigned long) len, M_DEVBUF, M_NOWAIT);
}

void
__gm_arch_kernel_free(void *ptr)
{
	free(ptr, M_DEVBUF);
}

#if GM_PREVENT_NATIVE_MALLOC_USE
#define malloc GM_DO_NOT_USE (malloc)
#define free GM_DO_NOT_USE (free)
#endif /* GM_PREVENT_MALLOC_USE */

gm_status_t
gm_arch_map_io_space(gm_instance_state_t * is, gm_u32_t offset,
					 gm_u32_t len, void **kaddr)
{
	GM_PRINT (GM_DEBUG_MAPIO, 
			  ("gm_arch_map_io_space: offset=0x%x\n", offset));
	*kaddr = (void *) ((vm_offset_t) is->arch.csr + (vm_offset_t) offset);
	return GM_SUCCESS;
}

void
gm_arch_unmap_io_space(gm_instance_state_t * is, gm_u32_t offset,
					   gm_u32_t len, void **kaddr)
{
	/* whatever .. */
}

gm_status_t
gm_arch_mmap_contiguous_segment(gm_port_state_t * ps, void *kaddr,
								unsigned long blockSize, gm_up_t *vaddr)
{
	/* implement me */
 	GM_WARN(("gm_arch_mmap_contiguous_segment not implemented\n"));
	return GM_FAILURE;
}

void
gm_arch_munmap_contiguous_segments(gm_port_state_t * ps)
{
	/* implement me */
 	GM_WARN(("gm_arch_munmap_contiguous_segment not implemented\n"));
}

/*
 * Miscellaneous functions
 */

gm_status_t
gm_arch_page_len(unsigned long *result)
{
	*result = cnt.v_page_size;
	return GM_SUCCESS;
}

gm_status_t
gm_arch_physical_pages(unsigned long *result)
{
	*result = cnt.v_page_count;
	return GM_SUCCESS;
}

gm_status_t
gm_arch_gethostname(char *ptr, int len)
{
	strncpy(ptr, hostname, len - 1);
    ptr[len-1]=0;
	return GM_SUCCESS;
}


/*
 * PCI config space functions
 */

gm_status_t
gm_arch_read_pci_config_32(gm_instance_state_t * is,
			   gm_offset_t offset,
			   gm_u32_t * value)
{
	*value = pci_read_config(is->arch.dev, offset, 4);
	return GM_SUCCESS;
}

gm_status_t
gm_arch_read_pci_config_16(gm_instance_state_t * is,
			   gm_offset_t offset,
			   gm_u16_t * value)
{
	*value = pci_read_config(is->arch.dev, offset, 2);
	return GM_SUCCESS;
}

gm_status_t
gm_arch_read_pci_config_8(gm_instance_state_t * is,
			  gm_offset_t offset,
			  gm_u8_t * value)
{
	*value = pci_read_config(is->arch.dev, offset, 1);
	return GM_SUCCESS;
}

gm_status_t
gm_arch_write_pci_config_32(gm_instance_state_t * is,
			    gm_offset_t offset,
			    gm_u32_t value)
{
	pci_write_config(is->arch.dev, offset, value, 4);
	return GM_SUCCESS;
}

gm_status_t
gm_arch_write_pci_config_16(gm_instance_state_t * is,
			    gm_offset_t offset,
			    gm_u16_t value)
{
	pci_write_config(is->arch.dev, offset, value, 2);
	return GM_SUCCESS;
}

gm_status_t
gm_arch_write_pci_config_8(gm_instance_state_t * is,
			   gm_offset_t offset,
			   gm_u8_t value)
{
	pci_write_config(is->arch.dev, offset, value, 1);
	return GM_SUCCESS;
}



/* 
 *  open, close, mmap, etc
 */

int
gm_freebsd_open(dev_t dev, int oflags, int devtype, proc_or_td *p)
{
	gm_status_t status;
	gm_port_state_t *ps;
	gm_instance_state_t *is;
	int unit;
	u_int minor_id;

	GM_CALLED ();
	GM_PRINT (GM_DEBUG_OPEN, 
			  ("gm_freebsd_open: open called gm dev 0x%x minor=%d\n", 
			   dev, minor(dev)));

	minor_id = minor(dev);
	if (minor_id >= (GM_ARCH_MAX_INSTANCE * 2)) {
		GM_PRINT (GM_DEBUG_OPEN, 
				  ("gm_freebsd_open: open for 'dynamic' device\n"));
	}
	else {
		if (minor_id < 0) {
			GM_PRINT (GM_DEBUG_OPEN, 
					  ("gm_freebsd_open: open for bad minor_id=%d\n",minor_id));
			return ENODEV;
		}

		unit = minor_id / 2;
		is = gm_instances[unit];

		if (!is) {
			GM_PRINT (GM_DEBUG_OPEN, 
					  ("gm_freebsd_open: open for bad instance=%d\n",unit));
			return ENODEV;
		}
	}

	/* success, so commit */

	GM_PRINT (GM_DEBUG_OPEN, 
			  ("gm_freebsd_open: User opened dev 0x%x with minor num %d.\n",
				 dev, minor(dev)));
	GM_RETURN_INT (0);
}


int
gm_freebsd_close(dev_t dev, int fflag, int devtype, proc_or_td *p)
{
	gm_port_state_t *ps;
	gm_instance_state_t *is;
	int unit;

	GM_CALLED ();
	GM_PRINT (GM_DEBUG_CLOSE, 
			  ("gm_freebsd_close called  dev = 0x%x  minor = %d\n",
				 dev, minor(dev)));

	if (minor(dev) >= (GM_ARCH_MAX_INSTANCE * 2)) {

		GM_PRINT (GM_DEBUG_CLOSE, 
				  ("closing 'dynamic' minor = %d\n",GET_MINOR(minor(dev)) ));

		ps = gm_minor_get_port_state(GET_MINOR(minor(dev)));
		if (!ps) {
			GM_NOTE ((__GM_FUNCTION__
					  ": could not get port_state for minor = %d\n",
					  GET_MINOR(minor(dev))));
			return ENODEV;
		}

		if (ps->opened) {
			gm_port_state_close(ps);
		}
		gm_minor_free(GET_MINOR(minor(dev)));
	}
	else {
		unit = minor(dev) / 2;

		is = gm_instances[unit];

		if (!is) {
			GM_WARN (("gm_freebsd_close: instance[%d] not found\n", unit));
			return ENODEV;
		}
	}

	GM_RETURN_INT (0);
}

static int
gm_freebsd_localize_status(gm_status_t status)
{
#define CASE(from,to) case from : return to
	switch (status) {
		 CASE(GM_SUCCESS, 0);
		 CASE(GM_INPUT_BUFFER_TOO_SMALL, EFAULT);
		 CASE(GM_OUTPUT_BUFFER_TOO_SMALL, EFAULT);
		 CASE(GM_TRY_AGAIN, EAGAIN);
		 CASE(GM_BUSY, EBUSY);
		 CASE(GM_MEMORY_FAULT, EFAULT);
		 CASE(GM_INTERRUPTED, EINTR);
		 CASE(GM_INVALID_PARAMETER, EINVAL);
		 CASE(GM_OUT_OF_MEMORY, ENOMEM);
		 CASE(GM_INVALID_COMMAND, EINVAL);
		 CASE(GM_PERMISSION_DENIED, EPERM);
	 default:
		 CASE(GM_INTERNAL_ERROR, EPFNOSUPPORT);
		 CASE(GM_UNATTACHED, -1);	/* ??? */
		 CASE(GM_UNSUPPORTED_DEVICE, ENXIO);
	}
}

/*
 * dumb-down the BSD ioctl method (where copyins are done by the
 * generic specfs ioctl hander) & have the GM ioctl function call
 * copyin/copyout as if we were a sysv style system 
 */

int
gm_freebsd_ioctl(dev_t dev, u_long cmd, caddr_t data,
				 int fflag, proc_or_td *p)
{
	gm_port_state_t *ps;
	gm_status_t status = GM_SUCCESS;
	vm_offset_t user_addr;
	int error = ENOTTY;
	unsigned copied_out;


	GM_PRINT (GM_DEBUG_IOCTL, ("ioctl: gm dev=0x%x minor=%d ptr = %p %s\n",
				 dev, minor(dev), data, _gm_ioctl_cmd_name(cmd)));

	if (cmd == GM_GET_DEV) {
		gm_instance_state_t *is;
		gm_port_state_t *clone_ps;
		gm_status_t status = 0;
		u_int minor_id = minor(dev);

		GM_PRINT (GM_DEBUG_IOCTL, 
				  ("gm_freebsd_ioctl: command is GM_GET_DEV\n"));

		if (minor_id >= (GM_ARCH_MAX_INSTANCE * 2)) {
			GM_WARN (("gm_freebsd_ioctl() WRONG %d\n", minor_id));
			error = ENODEV;
			goto abort_with_nothing;
		}

		is = gm_instances[(u_int) minor_id / 2];

		if (!is) {
			error = ENODEV;
			goto abort_with_nothing;
		}

		clone_ps = gm_minor_get_port_state(minor(dev));

		if (!clone_ps) {
			GM_NOTE (("gm_freebsd_ioctl: gm_minor_get_port_state(%d) failed\n",
					  minor(dev)));
			error = ENODEV;
			goto abort_with_nothing;
		}

		GM_PRINT (GM_DEBUG_IOCTL, 
				  ("gm_freebsd_ioctl: Got clone_ps 0x%p %d.\n", 
				   clone_ps,minor(dev)));

		/* Determine an unused minor number for the opened device. */

		if (gm_minor_alloc(clone_ps->instance, &minor_id) != GM_SUCCESS) {
			GM_NOTE (("gm_freebsd_ioctl: Could not create new minor number.\n"));
			error = ENOMEM;
			goto abort_with_nothing;
		}

		if (INVALID_MINOR(minor_id)) {
			gm_minor_free(minor_id);
			goto abort_with_nothing;
		}

		GM_PRINT (GM_DEBUG_IOCTL, 
				  ("gm_freebsd_ioctl: Allocated a new minor number %d.\n", 
				   minor_id));
		ps = gm_minor_get_port_state(minor_id);
		if (!ps) {
			GM_NOTE ((__GM_FUNCTION__
					  ": gm_minor_get_port_state(%d) for new device failed\n",
					  minor_id));
			error = ENODEV;
			goto abort_with_nothing;
		}

		GM_PRINT (GM_DEBUG_IOCTL,
				  ("got new port state = %p for minor=%d\n",ps,minor_id));

		/* Record the instance associated with this minor number. */

		ps->instance = clone_ps->instance;
		gm_assert(ps->instance);

		/* Record the privileges from the clone device */

		ps->privileged = clone_ps->privileged;

		GM_PRINT (GM_DEBUG_IOCTL, 
				  ("gm_freebsd_ioctl: Copyout minor=%d to arg = 0x%lx\n",
					 EXTERN_MINOR(minor_id), data));
		{
			int extern_minor = EXTERN_MINOR(minor_id);
			user_addr = *(vm_offset_t *) data;
			gm_arch_copyout(&extern_minor, (void *)user_addr, sizeof(int));
			GM_PRINT (GM_DEBUG_IOCTL, 
					  ("gm_freebsd_ioctl: new extern minor=%d\n", 
					   extern_minor));
		}
		return 0;
	}
	else {
		u_int minor_id = GET_MINOR(minor(dev));
		GM_PRINT (GM_DEBUG_IOCTL, ("ioctl using new minor = %d\n",minor_id));

		ps = gm_minor_get_port_state(minor_id);
		gm_assert(ps);
		if (!ps) {
			error = ENXIO;
			goto abort_with_nothing;
		}
		gm_arch_mutex_enter(&ps->sync);
		user_addr = *(vm_offset_t *) data;
		GM_PRINT (GM_DEBUG_IOCTL, 
				  ("ioctl: minor=%d  ps=0x%p user_addr=0x%lx\n", 
			minor_id,ps,user_addr));
		status = gm_ioctl(ps, cmd, (void *) user_addr, INT_MAX,
						  (void *) user_addr, INT_MAX, 0);
		gm_arch_mutex_exit(&ps->sync);

		GM_PRINT (GM_DEBUG_IOCTL, ("ioctl returns %d\n",
					 gm_freebsd_localize_status(status)));
		return gm_freebsd_localize_status(status);
	}

  abort_with_nothing:
	return (error);
}



/* Map interface memory into the process.  This can be used to map the
   LANai control register, special registers, SRAM, or copy block into
   the user space.  The offsets and sizes of these various regions can
   be obtained using the GM_GET_MAPPING_SPECS ioctl, which returns
   a description of the mappings in a "gm_mapping_specs"
   structure.

   The pages of the copy block must be mapped in order and without
   skipping any offsets.

   gm_segmap ensures that all offsets passed to gm_mmap are
   legitimate, so there is no need to range check offsets here.
   However, checks must be performed here for "promiscuous" mappings
   of the device registers an SRAM areas other than the "send queue"
   and "activation" pages. 

   Note that this function returns the page frame number of the
   individual page that the user mapped.  If the mapping is invalid,
   you must return 0.  There is no way to propagate a meaningful error
   back to the user.
*/

#if __FreeBSD_version < 500103
#define MMAP_RETURNS_PINDEX 1
#else
#define MMAP_RETURNS_PINDEX 0
#endif

int
gm_freebsd_mmap(dev_t dev, vm_offset_t off, 
#if MMAP_RETURNS_PINDEX == 0
				vm_offset_t *paddr, 
#endif
				int prot)
{
	void *result;
	gm_port_state_t *ps;
	int minor_id;
	gm_instance_state_t *is;
#if MMAP_RETURNS_PINDEX
	vm_offset_t phys;
	vm_offset_t *paddr = &phys;
	int retval = 0;  /* invalid pindex */
#else
	int retval = EINVAL; /* error status */
#endif
	off_t memory_offset = 0;
	gm_status_t status;

	GM_CALLED ();

	minor_id = minor(dev);

	GM_PRINT (GM_DEBUG_MMAP, 
			  ("gm_freebsd_mmap(0x%x, 0x%x, 0x%x) called   minor = %d\n",
			 minor_id, dev, off, prot));

	ps = gm_minor_get_port_state(GET_MINOR(minor_id));
	if (!ps)  {
		GM_NOTE (("gm_freebsd_mmap: can't get port_state for minor = %d\n",
				  GET_MINOR(minor_id)));
		goto abort_with_nothing;
	}

	gm_arch_mutex_enter(&ps->sync);

	/* Verify the mapping is legal and ensure all resources are
	   available, allocating them if needed. */

	status = gm_prepare_to_mmap(ps, off, GM_PAGE_LEN,
		(prot & PROT_WRITE ? GM_MAP_WRITE : 0
		| prot & PROT_READ ? GM_MAP_READ : 0));
	if (status != GM_SUCCESS) {
		GM_NOTE (("Error preparing to mmap.\n"));
		goto abort_with_mutex;
	}

	GM_PRINT (GM_DEBUG_MMAP, ("Mapping offset 0x%x for minor device %d\n", off,
				 GET_MINOR(minor(dev))));

	GM_PRINT (GM_DEBUG_MMAP, ("gm_freebsd_mmap: Calling gm_mmap().\n"));
	status = gm_mmap(ps, off, &result);
	if (status != GM_SUCCESS) {
		GM_NOTE (("gm_freebsd_mmap: gm_mmap returns -1\n"));
		goto abort_with_nothing;
	}
	gm_assert(result);

	if (!gm_mapping_in_io_space(ps, off)) {
		*paddr = vtophys(result);
	}
	else {
#ifdef __alpha__
		*paddr = (vm_offset_t) result;
#endif
#ifdef __i386__
		*paddr = vtophys(result);
#endif
	}
	GM_PRINT (GM_DEBUG_MMAP, ("gm_freebsd_mmap: phys = 0x%lx\n", 
							  (unsigned long)*paddr));

#if MMAP_RETURNS_PINDEX
	retval = (int)btoc(*paddr);   /* get the page index */
#else
	retval = 0;                   /* indicate that paddr is valid */
#endif

	GM_PRINT (GM_DEBUG_MMAP, 
			  ("gm_freebsd_mmap: kseg 0x%lx frame 0x%lx\n", result, retval));
  abort_with_mutex:
	gm_arch_mutex_exit(&ps->sync);
	GM_PRINT (GM_DEBUG_MMAP, ("gm_freebsd_mmap: returning 0x%lx.\n", retval));

  abort_with_nothing:
	GM_RETURN_INT (retval);
}

void
gm_freebsd_intr(void *arg)
{
	gm_instance_state_t *is;
	is = (gm_instance_state_t *) arg;
	if (gm_intr(is) == GM_ARCH_INTR_CLAIMED)
		GM_PRINT (GM_DEBUG_INTR, ("the interrupt was claimed\n"));
	else
		GM_PRINT (GM_DEBUG_INTR, ("the interrupt was not claimed\n"));
	return;
}

gm_status_t
gm_arch_lock_user_buffer_page(gm_port_state_t *ps, gm_up_t uaddr,
							  gm_dp_t * dma_addr, gm_arch_page_lock_t * lock)
{
	int error = 0;
	struct proc *p = curproc;
	vm_map_t map;
	vm_offset_t addr;
	GM_PRINT (GM_DEBUG_MEMLOCK, ("enter locking memory\n"));
	if (!p) {
		GM_NOTE (("gm locking memory: curproc was null\n"));
		return GM_OUT_OF_MEMORY;
	}
	map = &p->p_vmspace->vm_map;
	/* must be page aligned */
	addr = trunc_page((vm_offset_t) uaddr);
	if (gm_freebsd_max_user_locked_pages <= 0) {
		gm_freebsd_page_lock_failed++;
		return GM_OUT_OF_MEMORY;
	}
#if (__FreeBSD_version < 500000)
	error = vm_map_user_pageable(map, addr, addr + PAGE_SIZE, FALSE);
#else
	error = vm_map_wire(map, addr, addr + PAGE_SIZE, TRUE);
#endif
	if (error) {
		GM_NOTE (("gm locking memory: user attempted to lock invalid page\n"));
		return GM_FAILURE;
	}
	atomic_subtract_int(&gm_freebsd_max_user_locked_pages, 1);
	lock->proc = p;
	lock->addr = addr;
#ifdef __alpha__
	*dma_addr = alpha_XXX_dmamap_or | pmap_extract(vm_map_pmap(map), addr);
#endif
#ifdef __i386__
	*dma_addr = pmap_extract(vm_map_pmap(map), addr);
#endif
	GM_PRINT (GM_DEBUG_MEMLOCK, ("locking memory: success\n"));

	return GM_SUCCESS;
}


void
gm_arch_unlock_user_buffer_page(gm_arch_page_lock_t * lock)
{
	vm_map_t map;
	vm_offset_t addr;

	map = &lock->proc->p_vmspace->vm_map;
	addr = lock->addr;
#if (__FreeBSD_version < 500000)
	vm_map_user_pageable(map, addr, addr + PAGE_SIZE, TRUE);
#else
	vm_map_unwire(map, addr, addr + PAGE_SIZE, TRUE);
#endif
	atomic_add_int(&gm_freebsd_max_user_locked_pages, 1);
}

void gm_arch_abort (void)
{
	printf ("don't panic, but gm_arch_abort () was called.");
}



/*
  This file uses feldy indentation:

  Local Variables:
  tab-width:4
  c-file-style:"bsd"
  End:
*/
