
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: gallatin@myri.com */

#include "gm_bsd_ether.h"

/*
 * For more information about FreeBSD network drivers, look at a few
 * of the better drivers in the FreeBSD tree.  In particular,
 * sys/dev/bge/* and sys/dev/fxp/* are good examples.  Questions
 * regarding freebsd netoworking are best addressed to
 * the freebsd-net@FreeBSD.ORG mailing list.
 */

struct gm_bsd_ether_sc *gm_bsd_ether_sc[GM_ARCH_MAX_INSTANCE] = { 0 };
static int gm_mbufs_loaned = 0;
static int gm_mbuf_wakeup = 0;

struct mbuf *bogus_m = NULL;
struct gm_bsd_ether_arch_dma_info bogus_dma;

struct gm_jbuf;

struct gm_jpool_entry
{
  struct gm_jbuf *jbuf;
  unsigned int ref_count;
  SLIST_ENTRY (gm_jpool_entry) jpool_entries;
};


SLIST_HEAD (__gm_jfreehead, gm_jpool_entry) gm_jfree_listhead;
SLIST_HEAD (__gm_jinusehead, gm_jpool_entry) gm_jinuse_listhead;

/* make sure each buffer is aligned on a cache-line boundary */
#define GM_JLEN GM_ROUNDUP(u32, (GM_ETHERNET_MTU + 2), GM_RDMA_GRANULARITY)

struct gm_jbuf
{
  struct gm_jpool_entry *entry;
  char	buf[GM_JLEN];
};

#define JSLOTS 512
#define GM_DEBUG_JUMBO_BUFFS 0

char *gm_jumbo_buf = NULL;
char *gm_jumbo_buf_alloc = NULL;

#if (__FreeBSD_version >= 500000)
#include <sys/mutex.h>
static struct mtx jumbo_mutex;
#define gm_bsd_ether_arch_lock_jumbo() mtx_lock (&jumbo_mutex);
#define gm_bsd_ether_arch_unlock_jumbo() mtx_unlock (&jumbo_mutex);
#else
/* FreeBSD 4.x has a giant kernel lock, no need for us to do any
   locking */
#define gm_bsd_ether_arch_lock_jumbo() 
#define gm_bsd_ether_arch_unlock_jumbo() 

#endif


/*
 * Free the pool of jumbo buffers
 */

static void
gm_freebsd_ether_free_jumbo_mem (void)
{
  struct gm_jpool_entry *entry;

  gm_bsd_ether_arch_lock_jumbo();
  while ((entry = SLIST_FIRST (&gm_jfree_listhead)) != NULL)
    {
      SLIST_REMOVE_HEAD (&gm_jfree_listhead, jpool_entries);
      gm_free (entry);
    }
  gm_free (gm_jumbo_buf_alloc);
  gm_jumbo_buf = NULL;
  gm_bsd_ether_arch_unlock_jumbo();
#if (__FreeBSD_version >= 500000)
  mtx_destroy (&jumbo_mutex);
#endif   
}


/*
 *
 * Allocate a pool of jumbo buffers shared among adapters.  This pool
 * is malloced from virtually contigous memory.  Each 9K frame is
 * carved out of the pool in order, and put onto a LIFO freelist until
 * needed.
 *
 * Using contigous 9K sized buffers provides better performance
 * (1.8Gb/sec vs 1.3Gb/sec) than using system (2k) mbufs.  Glenn
 * suspects contigous buffers perform better because when copying,
 * the system is are able to work through the entire 9K, rather than
 * stalling every 2K on a cache miss
 */
static
int gm_freebsd_ether_alloc_jumbo_frames (void)
{
  struct gm_jpool_entry *entry;
  struct gm_jbuf *jbuf;
  int i;

#if (__FreeBSD_version >= 500000)
  mtx_init (&jumbo_mutex, "GM Jumbo freelist mutex", NULL, MTX_DEF);
#endif 
  SLIST_INIT (&gm_jfree_listhead);
  SLIST_INIT (&gm_jinuse_listhead);

  gm_jumbo_buf_alloc = gm_malloc (sizeof (*jbuf) * JSLOTS + PAGE_SIZE);
  if (gm_jumbo_buf_alloc == NULL)
    {
      return GM_OUT_OF_MEMORY;
    }

  gm_jumbo_buf = (char *)round_page ((vm_offset_t)gm_jumbo_buf_alloc);
  jbuf = (struct gm_jbuf *) gm_jumbo_buf;
  for (i = 0; i < JSLOTS; i++)
    {
      entry = gm_malloc (sizeof (*entry));
      if (entry == NULL)
	{
	  gm_freebsd_ether_free_jumbo_mem ();
	  return ENOMEM;
	}
      entry->jbuf = jbuf;
      entry->ref_count = 0;
      jbuf->entry = entry;
      SLIST_INSERT_HEAD (&gm_jfree_listhead, entry, jpool_entries);
      jbuf++;
    }
  return GM_SUCCESS;
}

/*
 * allocate a jumbo buffer for network reception
 * called by gm_bsd_ether_arch_rxalloc()
 */

struct gm_jpool_entry *
gm_freebsd_ether_jalloc (void)
{
  struct gm_jpool_entry *entry;

  gm_bsd_ether_arch_lock_jumbo ();
  entry = SLIST_FIRST (&gm_jfree_listhead);
  if (entry == NULL)
    {
      gm_bsd_ether_arch_unlock_jumbo ();
      return NULL;
    }

  SLIST_REMOVE_HEAD (&gm_jfree_listhead, jpool_entries);
  gm_mbufs_loaned++;
  gm_bsd_ether_arch_unlock_jumbo ();
  return entry;
}

/*
 * free (or drop a reference to) a jumbo buffer
 * Typically called from the network stack via m_freem().
 */

static void
gm_freebsd_ether_jfree (
#if (__FreeBSD_version < 500000)
			 caddr_t ptr, u_int ingnored
#else
			 void *ptr, void *args
#endif
			 )
{
  struct gm_jbuf *buf;
  struct gm_jpool_entry *entry;

  buf = (struct gm_jbuf *)((char *)ptr - GM_OFFSETOF(struct gm_jbuf, buf));
  entry = buf->entry;
  if (__FreeBSD_version >= 500000 || entry->ref_count == 0)
    {
      gm_bsd_ether_arch_lock_jumbo ();
      SLIST_INSERT_HEAD (&gm_jfree_listhead, entry, jpool_entries);
      gm_mbufs_loaned--;
      gm_bsd_ether_arch_unlock_jumbo ();
    }
  else
    entry->ref_count--;
}

#if (__FreeBSD_version < 500000)
/*
 * add a reference to a jumbo buffer
 * Called via m_copym from the network stack
 */

static void
gm_freebsd_ether_jref (caddr_t ptr, u_int ignored)
{
  struct gm_jbuf *buf;

  buf = (struct gm_jbuf *)ptr;
  buf->entry->ref_count++;
}

#endif /* __FreeBSD_version < 500000 */

/* 
 *  allocates a new "softc" struct for the gm ether device
 */

static struct gm_bsd_ether_sc *
gm_freebsd_ether_alloc_sc (int unit)
{
  struct gm_bsd_ether_sc *sc;

  if (gm_bsd_ether_sc[unit] != NULL)
    {
      GM_WARN (("attempt to attach myri%d twice failed\n", unit));
      return NULL;
    }

  sc = gm_malloc(sizeof (*sc));
  if (sc != NULL)
    {
      GM_PRINT (GM_DEBUG_JUMBO_BUFFS, ("GM_JLEN = %d\n", GM_JLEN));
      bzero(sc, sizeof (*sc));
      if (gm_jumbo_buf == NULL && gm_freebsd_ether_alloc_jumbo_frames ())
	{
	  GM_WARN (("Jumbo memory allocation failed\n"));
	}
      gm_bsd_ether_sc[unit] = sc;
    }
  return sc;
}

/* 
 * free a softc
 */

static void
gm_freebsd_ether_dealloc_sc (int unit)
{
  struct gm_bsd_ether_sc *sc;

  sc = gm_bsd_ether_sc[unit];
  gm_bsd_ether_sc[unit] = NULL;
  gm_free (sc);
}

/*
 * called by ether_ioctl to "initialize" an interface.  We must
 * supply the function pointer, but we do not use this feature.
 */

static void
gm_freebsd_ether_init (void *sc_ptr)
{
}

#if (__FreeBSD_version >= 500000 )
/*
 * This is the transmit routine.  Obtain the sleep lock
 * for the driver and call the common transmit routine.
 */
static void
gm_freebsd_ether_start (struct ifnet *ifp)
{
  gm_bsd_ether_arch_sc_lock (ifp->if_softc);
  gm_bsd_ether_start (ifp);
  gm_bsd_ether_arch_sc_unlock (ifp->if_softc);
}

#endif

/*
 * creates a new ethernet driver, initialzes it, and attaches it to the
 * network stack
 */
int
gm_freebsd_ether_attach (gm_instance_state_t * is)
{
  int i, unit;
  struct ifnet *ifp;
  struct gm_bsd_ether_sc *sc;

  if (GM_MAX_ETHERNET_GATHER_CNT * MCLBYTES <= GM_ETHERNET_MTU)
    {
      GM_WARN (("GM_MAX_ETHERNET_GATHER_CNT (%d) is too small to hold a\n"
		"%d byte long mbuf chain consisting of %d byte mbufs\n"
		"Each transmit of %d bytes or more will have to be copied\n"
		"Consider increasing MCLSHIFT and rebuilding your kernel & GM\n"
		"or rebuilding GM using a 32-bit gm_dp_t\n",
		GM_MAX_ETHERNET_GATHER_CNT, GM_ETHERNET_MTU, MCLBYTES,
		(GM_MAX_ETHERNET_GATHER_CNT - 1) * MCLBYTES));
    }
  if (MCLBYTES != mbstat.m_mclbytes)
    {
      GM_WARN (("module MCLBYTES (%d) != kernel MCLBYTES (%ld)\n", MCLBYTES,
		(unsigned long)mbstat.m_mclbytes));
      return EINVAL;
    }
  unit = device_get_unit (is->arch.dev);
  sc = gm_freebsd_ether_alloc_sc (unit);
  if (sc == NULL)
    {
      return ENOMEM;
    }
  sc->is = is;
  if (bogus_m == NULL)
    {
      bogus_m = gm_bsd_ether_arch_rxalloc (M_NOWAIT, GM_MAX_ETHERNET_SCATTER_CNT);
      if (bogus_m == NULL)
	{
	  gm_freebsd_ether_dealloc_sc (unit);
	  return ENOMEM;
	}
      if (!gm_bsd_ether_arch_encap (sc, bogus_m, &bogus_dma))
	{
	  m_freem (bogus_m);
	  bogus_m = NULL;
	  gm_freebsd_ether_dealloc_sc (unit);
	  return ENOMEM;
	}
    }  
  ifp = &sc->is_if;
  ifp->if_softc = sc;
  ifp->if_addrlen = 6;
  for (i = 0; i < 6; i++)
    {
      sc->is_addr[i] = is->lanai.eeprom.lanai_board_id[i];
      sc->current_mac[i] = sc->is_addr[i];
    }

  ifp->if_name = "myri";
  ifp->if_unit = unit;

  /* leave extra space to prepend myritype */
  ifp->if_hdrlen = sizeof (struct ether_header) + 8;
  ifp->if_mtu = GM_IP_MTU;
  ifp->if_timer = 0;
  ifp->if_ioctl = gm_freebsd_ether_ioctl;
  ifp->if_watchdog = NULL;
  ifp->if_init = gm_freebsd_ether_init;
#if (__FreeBSD_version < 500000)
  ifp->if_start = gm_bsd_ether_start;
#else
  mtx_init (&sc->arch.mutex, "myri", NULL, MTX_DEF);
  ifp->if_start = gm_freebsd_ether_start;
#endif
  ifp->if_output = ether_output;
  ifp->if_flags = IFF_BROADCAST | IFF_MULTICAST | IFF_SIMPLEX;
  ifp->if_baudrate = 1000000000;
  ifp->if_snd.ifq_maxlen = GM_SEND_RING_CNT;
#if (__FreeBSD_version >= 500000)
  ether_ifattach (ifp, sc->is_addr);
#elif (__FreeBSD_version >= 410000)
  ether_ifattach (ifp, ETHER_BPF_SUPPORTED);
#else
  if_attach (ifp);
  ether_ifattach (ifp);
#endif
  /*ether_ifattach resets mtu */
  ifp->if_mtu = GM_IP_MTU;
  return 0;

}

static int
gm_freebsd_ether_ioctl (struct ifnet *ifp, unsigned long cmd, caddr_t data)
{
  struct gm_bsd_ether_sc *sc = ifp->if_softc;
  struct ifreq *ifr = (struct ifreq *) data;
  int s, error;

  s = splimp ();

  switch (cmd)
    {

    case SIOCSIFADDR:
    case SIOCGIFADDR:
      error = ether_ioctl (ifp, cmd, data);
      break;

    case SIOCSIFMTU:
      if (ifr->ifr_mtu > GM_IP_MTU)
	{
	  error = EINVAL;
	}
      else
	{
	  ifp->if_mtu = ifr->ifr_mtu;
	  error = 0;
	}
      break;
    case SIOCSIFFLAGS:
      if (ifp->if_flags & IFF_UP)
	{
	  if (!(ifp->if_flags & IFF_RUNNING))
	    {
	      error = gm_bsd_ether_open (sc);
	      ifp->if_flags |= IFF_RUNNING;
	    }
	  gm_bsd_ether_set_promisc (ifp, sc);
	  /* if_setlladdr brings interface down, then up
	     again to set mac address.  So we always try setting it */
	  gm_bsd_ether_set_mac (ifp, sc);
	}
      else
	{
	  if (ifp->if_flags & IFF_RUNNING)
	    {
	      ifp->if_flags &= ~IFF_RUNNING;
	      error = gm_bsd_ether_close (sc);
	    }
	}
      break;
    case SIOCADDMULTI:
      /* fall through */
    case SIOCDELMULTI:
      error = 0;
      break;
    default:
      error = ENOTTY;
    }

  splx (s);
  return error;
}

struct mbuf *
gm_bsd_ether_arch_rxalloc (int how, int max_segs)
{
  struct mbuf *m_head, *mp, *m_tail;
  int len, type;

  GM_VAR_MAY_BE_UNUSED (type);
  len = GM_ETHERNET_MTU;

  /* get the first header mbuf */
  MGETHDR (mp, how, MT_DATA);
  if (mp == NULL)
    {
      return NULL;
    }

  /* if we managed to allocate jumbo mbufs, use them for
     performance reasons.
  */
  if (gm_jumbo_buf)
    {
      /* make it into a contigous 9K "jumbo" external mbuf */
      struct gm_jpool_entry *entry;

      m_head = m_tail = mp;
      entry = gm_freebsd_ether_jalloc ();
      if (entry != NULL)
	{
#if (__FreeBSD_version < 500000)
	  entry->ref_count = 0;
	  mp->m_ext.ext_free = gm_freebsd_ether_jfree;
	  mp->m_ext.ext_ref = gm_freebsd_ether_jref;
	  mp->m_ext.ext_buf = (caddr_t)entry->jbuf->buf;
	  mp->m_data = entry->jbuf->buf;
	  mp->m_flags |= M_EXT;
	  mp->m_ext.ext_size = GM_JLEN;
#else	
#ifdef EXT_EXTREF
	  /* use our own refcounter to avoid malloc() in
	     critical path */
	  mp->m_ext.ref_cnt =  &entry->ref_count;
	  type = EXT_EXTREF;
#else
	  type = EXT_NET_DRV;
#endif /* EXT_EXTREF */
	  m_extadd(mp, (void *)entry->jbuf->buf, GM_JLEN,
		  gm_freebsd_ether_jfree, (void *)entry->jbuf, 
		  0, type);
#endif /* __FreeBSD_version < 500000 */
	  
	  /* misalign 14 byte ethernet header so that IP header ends
	     up on a 4 byte boundary */
	  mp->m_len = GM_JLEN - 2;
	  mp->m_data += 2; 
			      
	  return m_head;
	}
    }

  /* we're out of jumbos, fall back to system mbufs clusters */
  MCLGET (mp, how);
  if ((mp->m_flags & M_EXT) == 0)
    {
      m_free (mp);
      return NULL;
    }

  /* start building the chain of cluster mbufs */
  m_head = m_tail = mp;

  /* misalign 14 byte ethernet header so that IP header ends up on a 4
     byte boundary */ 
  mp->m_data += 2; 
  mp->m_len = MCLBYTES - 2;
  len -= MCLBYTES - 2;

  while (len > 0)
    {
#if (__FreeBSD_version < 470000)
      /* get the next header mbuf */
      MGET (mp, how, MT_DATA);
      if (mp == NULL)
	{
	  goto abort_with_cluster;
	}
      /* attach to chain */
      m_tail->m_next = mp;
      m_tail = mp;
      /* use a 2K "cluster" */
      MCLGET (mp, how);
      if ((mp->m_flags & M_EXT) == 0)
	{
	  goto abort_with_cluster;
	}
#else	
      mp = m_getcl(how, MT_DATA, 0);
      if (mp == NULL)
	{
	  goto abort_with_cluster;
	}
      m_tail->m_next = mp;
      m_tail = mp;
#endif
      mp->m_len = MCLBYTES;
      len -= MCLBYTES;
    }

  return m_head;

 abort_with_cluster:
  m_freem (m_head);
  return (NULL);
}

/* 
 * Encapsulate an mbuf chain into an ethernet segment descriptor.
 * Since FreeBSD doesn't have real DMA mapping support, this is easy.
 * This routine returns the number of segments in the ethernet segment
 * descriptor.
 */

#define GM_FREEBSD_ETHER_DMA_CUT 1

int
gm_bsd_ether_arch_encap (struct gm_bsd_ether_sc *sc, struct mbuf *m_head,
		struct gm_bsd_ether_arch_dma_info *dma)
{
  int count;
  struct mbuf *m;

  count = 0;
  for (m = m_head; m != NULL; m = m->m_next)
    {
      if (m->m_len == 0)
	continue;


      if (GM_FREEBSD_ETHER_DMA_CUT)
	{
	  vm_offset_t start = mtod(m, vm_offset_t);
	  vm_offset_t len = m->m_len;
	  vm_offset_t boundary =  trunc_page (start + PAGE_SIZE);
	  vm_offset_t bytes;

	  while (boundary < start + len)
	    {
	      bytes = boundary - start;
	      dma->desc[count].ptr = vtophys ((caddr_t) start);
	      dma->desc[count].len = bytes;
	      len -= bytes;
	      start += bytes;
	      boundary = trunc_page (start + PAGE_SIZE);
	      count++;
	    }
	  if (len)
	    {
	      dma->desc[count].ptr = vtophys ((caddr_t) start);
	      dma->desc[count].len = len;
	    }
	}
      else 
	{
	  dma->desc[count].ptr = vtophys (mtod (m, caddr_t));
	  dma->desc[count].len = m->m_len;
	}
	count++;
    }
  dma->nsegs = count;
  return count;
}
extern int gm_num_instance;
void
gm_freebsd_ether_detach (gm_instance_state_t * is)
{
  struct gm_bsd_ether_sc *sc;
  struct ifnet *ifp, *gm_ifp;
  int s, ifp_valid;

  sc = gm_bsd_ether_sc[device_get_unit (is->arch.dev)];
  /* if we failed to attach, then there's nothing to detach */
  if (sc == NULL)
    {
      return;
    }

  s = splnet ();

  /* search for our ifnet in the global list */
  ifp_valid = 0;
  gm_ifp = &sc->is_if;
  TAILQ_FOREACH (ifp, &ifnet, if_link)
    {
      if (ifp == gm_ifp)
	{
	  ifp_valid = 1;
	  break;
	}
    }

  if (ifp_valid)
    {
#if (__FreeBSD_version >= 500000)
      ether_ifdetach(ifp);
#elif (__FreeBSD_version >= 410000)
      ether_ifdetach (gm_ifp, ETHER_BPF_SUPPORTED);
#else /* no ether_ifdetach() available, so do it by hand */
      struct ifaddr *ia = 0;
      struct in_ifaddr *in_ia = 0;

      TAILQ_FOREACH (ia, &ifp->if_addrhead, ifa_link)
	{
	  if (ia->ifa_addr->sa_family != AF_INET)
	    continue;
	  in_ia = (struct in_ifaddr *) ia;
	  in_ifscrub (gm_ifp, in_ia);
	  TAILQ_REMOVE (&in_ifaddrhead, in_ia, ia_link);
	}
      if_detach (gm_ifp);
#endif /* (__FreeBSD_version >= 500000) */
    }

  splx (s);

  /* 
   * the system no longer knows about the interface, close it &
   * deallocate it 
   */

  if (sc->open)
    {
      gm_bsd_ether_close (sc);
    }

  if (bogus_m)
    {
      gm_bsd_ether_arch_unencap (sc, &bogus_dma);
      m_freem (bogus_m);
      bogus_m = 0;
    }

#if (__FreeBSD_version >= 500000)  
  mtx_destroy (&sc->arch.mutex);
#endif

  if ((gm_num_instance == 1) && gm_jumbo_buf)
    {
      while (gm_mbufs_loaned != 0)
	{
	  gm_mbuf_wakeup = 1;
	  GM_NOTE (("%d mbufs still remain to be freed, sleeping\n",
		    gm_mbufs_loaned));
	  tsleep (&gm_mbufs_loaned, PUSER, "gm_freebsd_ether_detach", 10 * hz);
	}
      gm_freebsd_ether_free_jumbo_mem ();
    }
    
  gm_freebsd_ether_dealloc_sc (device_get_unit (is->arch.dev));
}




/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
