/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#if !GM_KERNEL
#error GM_KERNEL is not defined for kernel compilations
#endif

#include "gm_call_trace.h"
#include "gm_cpp.h"
#include "gm_debug.h"
#include "gm_debug_board_init.h"
#include "gm_debug_lanai_dma.h"
#include "gm_debug_host_dma.h"
#include "gm_debug_mem_register.h"
#include "gm_debug_page_table.h"
#include "gm_debug_port_state.h"
#include "gm_debug_reg_recv_queue.h"
#include "gm_debug_sync.h"
#include "gm_dma_chain.h"
#include "gm_enable_security.h"
#include "gm_impl.h"
#include "gm_internal.h"
#include "gm_lanai_command.h"
#include "gm_malloc_debug.h"
#include "gm_page_hash.h"
#include "gm_pio.h"
#include "gm_pte.h"
#include "gm_ptr_hash.h"
#include "gm_register_recv_queue.h"
#include "gm_trace.h"
#include "gm_board_mapping.h"

struct gm_page_lock
{
  struct gm_page_lock *next;
  gm_arch_page_lock_t arch;
};


/************
 * GCC compatibility functions.
 ************/

#if GM_NEED_MEMCPY
/* Needed to allow gcc structure copy to work. */
void *
memcpy (void *to, const void *from, size_t bytes)
{
  bcopy ((void *) from, to, bytes);
  return to;
}
#endif

#if GM_NEED_MEMSET
/* Needed for gcc struct constructors. */
void *
memset (void *s, int c, size_t n)
{
  while (n--)
    ((char *) s)[n] = c;
  return s;
}
#endif

/************
 * easy-to-use copyin/copyout.
 ************/

#define GM_DEBUG_COPYOUT 0

gm_status_t
gm_copyout (gm_port_state_t * ps, void *from, void *to, gm_size_t len)
{
  gm_instance_state_t *is;
  gm_status_t ret;
  void *buf = 0;

  GM_CALLED_WITH_ARGS_EX
    (GM_DEBUG_COPYOUT,
     ("%p,%p,%p,%ld", ps, from, to, len));

  GM_PRINT (GM_PRINT_LEVEL >= 11,
	    ("gm_copyOUT(%p, from=%p, to=%p, len=%d)\n", ps, from, to, len));

  /* Avoid partword I/O, if needed. */

  if (GM_NO_PARTWORD_PIO_READ)
    {
      /* If copying data from the LANai SRAM, use special PIO call to copy
         the data. */

      is = ps->instance;
      gm_assert (is);
      if ((char *) from >= (char *) is->lanai.sram
	  && (char *) from < ((char *) is->lanai.sram
			      + is->lanai.eeprom.lanai_sram_size))
	{
	  buf = gm_malloc (len);
	  if (!buf)
	    {
	      GM_RETURN_STATUS (GM_OUT_OF_MEMORY);
	    }
	  /*printf("Doing special copyout for no_pio_read\n"); */
	  __gm_ntok_bcopy (is, from, buf, len);
	  from = buf;
	}
    }

  if (to == ps->out_buff)
    {
      ret = gm_arch_copyout (from, to, len);
    }
  else
    {
      gm_bcopy (from, to, len);
      ret = GM_SUCCESS;
    }

  /* Free any buffer used to avoid partword PIO. */

  if (buf)
    {
      gm_free (buf);
    }

  GM_RETURN_STATUS (ret);
}

#define GM_DEBUG_COPYIN 0

gm_status_t
gm_copyin (gm_port_state_t * ps, const void *from, void *to, gm_size_t len)
{
  GM_CALLED_WITH_ARGS_EX
    (GM_DEBUG_COPYIN,
     ("%p,%p,%p,%ld", ps, from, to, len));
  
  gm_assert ((char *) to < (char *) ps->instance->lanai.sram
	     || (char *) to >= ((char *) ps->instance->lanai.sram
				+
				ps->instance->lanai.eeprom.lanai_sram_size));

  if (from == ps->in_buff)
    {
      GM_RETURN_STATUS (gm_arch_copyin (from, to, len));
    }
  gm_bcopy (from, to, len);
  GM_RETURN_STATUS (GM_SUCCESS);
}

/***********************************************************************
 * Utility functions
 ***********************************************************************/

/* Note: May be called by interrupt handler. */

#define GM_DEBUG_DISABLE_LANAI 0

void
gm_disable_lanai (gm_instance_state_t * is)
{
  GM_CALLED ();

  gm_assert (is);
  GM_PRINT (GM_DEBUG_DISABLE_LANAI, ("Disabling the LANai\n"));
  if (GM_DEBUG_DISABLE_LANAI)
    {
      GM_PRINT
	(1, ("0x%x\n", gm_read_first_sram_page (is, init_progress, u32)));
      GM_PRINT (1, ("*************************\n"));
      gm_log_call_trace ();
      GM_PRINT (1, ("*************************\n"));
    }
  if (is->lanai.running)
    {
      gm_assert (is->lanai.special_regs);

      gm_finalize_interrupts (is);

      /* Reset the LANai, unless an assertion failed (so the LANai will
         flash SOS */

      if (is->lanai.globals == 0
	  || (gm_read_lanai_global_lp (is, interrupt.failed_assertion.file)
	      == 0)
	  || (gm_read_lanai_global_u32 (is, interrupt.failed_assertion.line)
	      == 0))
	{
	  gm_lanai_reset_on (is);
	}

      gm_set_EIMR (is, 0);
      is->lanai.running = 0;
      GM_STBAR ();
    }

  GM_RETURN_NOTHING ();
}

/*----------------------------------------------------------------------
 - Pausing the LANai MCP
 ----------------------------------------------------------------------*/

#define GM_DEBUG_PAUSE 0

/* Obtain exclusive access to LANai memory by requesting that the
   LANai control program pause and acknowlege that it is paused via an
   interrupt.  4-phase acks are used for the pause/unpause cycle. */
void
gm_pause_lanai (gm_instance_state_t * is)
{
  gm_status_t status;

  GM_CALLED_WITH_ARGS_EX (GM_DEBUG_PAUSE, ("%p", is));

  /* Only one thread may command the LANai at a time. */

  GM_PRINT (GM_DEBUG_PAUSE,
	    ("Entering command mutex in sync %p.\n", &is->command_sync));

  gm_arch_mutex_enter (&is->command_sync);
  GM_PRINT (GM_DEBUG_PAUSE, ("Entered command mutex.\n"));

  /* If LANai not running, no need to pause. */

  if (!is->pausible  || !is->lanai.running)
    {
      GM_PRINT (GM_DEBUG_PAUSE, ("LANai not running.  Using shortcut.\n"));
      is->pause_rqst = 1;
      is->pause_ack = 1;
      GM_RETURN_NOTHING ();
    }

  /* Verify not violating protocol. */

  gm_assert (is->lanai.running);
  gm_assert (is->lanai.globals);
  gm_assert (is->pause_rqst == 0);
  gm_assert (is->pause_ack == 0);
  gm_assert (gm_read_lanai_global_u32 (is, magic) == 0xcafebabe);
  /* Tell lanai to pause, caching pause value. */

  GM_PRINT (GM_DEBUG_PAUSE, ("Telling LANai to pause.\n"));
  is->pause_rqst = 1;
  status = gm_lanai_command (is, GM_LANAI_COMMAND_PAUSE);
  if (status != GM_SUCCESS)
    {
      goto abort;		/* Timed out. */
    }
  is->pause_ack = 1;
  GM_PRINT (GM_DEBUG_PAUSE, ("LANai paused.\n"));

  GM_RETURN_NOTHING ();

abort:
  gm_disable_lanai (is);	/* Please talk to Glenn before disabling */
  GM_RETURN_NOTHING ();
}

/* Release exclusive access to LANai memory. */

void
gm_unpause_lanai (gm_instance_state_t * is)
{
  gm_status_t status;

  GM_CALLED_WITH_ARGS_EX (GM_DEBUG_PAUSE, ("%p", is));

  /* Make sure all writes have been flushed. */

  GM_STBAR ();

  if (!is->pausible || !is->lanai.running)
    {
      GM_PRINT (GM_DEBUG_PAUSE, ("LANai not running.  Using shortcut.\n"));
      is->pause_rqst = 0;
      is->pause_ack = 0;
      goto done;
    }

  /* Verify not violating protocol. */

  gm_assert (is->lanai.running);
  gm_assert (is->lanai.globals);
  gm_assert (gm_read_lanai_global_u32 (is, magic) == 0xcafebabe);
  gm_assert (is->pause_rqst != 0);
  gm_assert (is->pause_ack != 0);

  /* Signal LANai to resume. */

  gm_assert (is->pause_rqst == is->pause_ack);
  is->pause_rqst = 0;
  status = gm_lanai_command (is, GM_LANAI_COMMAND_UNPAUSE);
  if (status != GM_SUCCESS)
    {
      goto abort;		/* Timed out. */
    }
  is->pause_ack = 0;
  GM_PRINT (GM_DEBUG_PAUSE, ("LANai resumed.\n"));

done:
  GM_PRINT (GM_DEBUG_PAUSE, ("Exiting command mutex %p.\n", &is->command_sync));
  gm_arch_mutex_exit (&is->command_sync);
  GM_PRINT (GM_DEBUG_PAUSE, ("Exited command mutex %p.\n", &is->command_sync));
  GM_RETURN_NOTHING ();

abort:
  gm_disable_lanai (is);	/* Please talk to Glenn before disabling */
  GM_PRINT (GM_DEBUG_PAUSE, ("Exiting command mutex.\n"));
  gm_arch_mutex_exit (&is->command_sync);
  GM_RETURN_NOTHING ();
}

/*----------------------------------------------------------------------
 - Mapping spec initialization
 ----------------------------------------------------------------------*/

/* Initialize the "gm_mapping_specs_t mappings" field of a
   gm_port_state_t structure.  This structure describes the what
   offsets should be mapped to map each device resource.  All entries
   in the mapping spec are specified, even if the user will not be
   allowed to map the region due to lack of access privileges.

   Called from PASSIVE_LEVEL, but may be called from any IRQL. */

void
gm_init_mapping_spec (gm_port_state_t * ps)
{
  gm_instance_state_t *is;
  gm_mapping_specs_t *m;
  int port_id;

  GM_CALLED ();

  gm_assert (GM_PAGE_LEN);

  is = ps->instance;
  m = &ps->mappings;
  port_id = ps->id;

  /*** The Device is mapped as follows

    +------------------+ <- offset 0
    | Control regs     | (1 page in device)
    +------------------+
    | Special regs     | (1 page in device)
    +------------------+
    | LANai SRAM       | (many pages in device)
    +------------------+
    | Recv queue       | (several pages host DMAable memory)
    +------------------+
    | copy block       | (many pages host DMAable memory)
    +------------------+
    */

  m->control_regs.offset = 0;
  m->control_regs.len = (gm_u32_t) GM_PAGE_LEN;
  m->control_regs.permissions = (ps->privileged ? GM_MAP_RDWR : 0);

  m->special_regs.offset = (m->control_regs.offset
			    + (gm_offset_t) m->control_regs.len);
  m->special_regs.len = (gm_u32_t) GM_PAGE_LEN;
  m->special_regs.permissions = (ps->privileged ? GM_MAP_RDWR : 0);

  m->sram.offset = (m->special_regs.offset
		    + (gm_offset_t) m->special_regs.len);
  m->sram.len = is->lanai.eeprom.lanai_sram_size;
  m->sram.permissions = (ps->privileged ? GM_MAP_RDWR : 0);
  gm_assert (m->sram.len);

  m->copy_block.offset = (m->sram.offset
			  + (gm_offset_t) m->sram.len);
  m->copy_block.len =
#if GM_CAN_REGISTER_MEMORY
    0
#else
    ps->max_dma_mem
#endif
    ;
  m->copy_block.permissions = GM_MAP_RDWR;

/*** Pointers to the pieces of LANai SRAM region of the device */

  /*** The SRAM is divided as follows.

    +------------------+ <- address 0
    | control program  |
    ...
    |                  |
    +------------------+
    | Send queue array | (1 page per port)
    +------------------+
    | Hash piece table | (several pages)
    +------------------+ <- top of SRAM
    */

  m->hash_piece_ptrs.len = (gm_u32_t) GM_PAGE_HASH_PIECE_REF_TABLE_LEN;
  m->hash_piece_ptrs.offset = (m->sram.offset + (gm_offset_t) m->sram.len
			       - (gm_offset_t) m->hash_piece_ptrs.len);
  m->hash_piece_ptrs.permissions = ps->privileged ? GM_MAP_RDWR : 0;

  /* Send queue pages are below the hash piece table page.  Report the
     offsets for the one page associated with this port. */

  m->send_queue.len = (gm_u32_t) GM_PORT_MAPPED_SRAM_LEN;
  m->send_queue.offset = (gm_u32_t) (m->hash_piece_ptrs.offset
				     - (GM_PORT_MAPPED_SRAM_LEN
					* (GM_NUM_PORTS - port_id)));

  GM_PRINT (GM_PRINT_LEVEL >= 5,
	    ("****** send_queue.offset for port = %d   = 0x%x\n", port_id,
	     m->send_queue.offset));
  _GM_PRINT (GM_PRINT_LEVEL >= 5,
	     ("****** calculation = 0x%x - (%d * (%d - %d))\n",
	      m->hash_piece_ptrs.offset, GM_PAGE_LEN, GM_NUM_PORTS, port_id));

  m->send_queue.permissions = GM_MAP_RDWR;

  /* The RTC page is the first page is SRAM, and may be mapped read-only */

  m->RTC.offset = m->sram.offset;
  m->RTC.len = (gm_u32_t) GM_PAGE_LEN;
  m->RTC.permissions = GM_MAP_READ;

  GM_RETURN_NOTHING ();
}

/*----------------------------------------------------------------------
 - Hash table
 ----------------------------------------------------------------------*/

/* Register a page-sized piece of the page hash table.  This function
   is used as a gm_dma_region_alloc callback, with arg being a pointer
   to the instance state. */

#if GM_ENABLE_VM
static gm_status_t
gm_register_page_hash_piece (void *arg, gm_dp_t dma_addr, gm_u32_t page_num)
{
  gm_dp_n_t *slot;
  gm_instance_state_t *is = arg;
  gm_u32_t offset;

  GM_PRINT (GM_DEBUG_PAGE_HASH,
	    ("gm_register_page_hash_piece (%p, 0x%x, %d) called \n",
	     arg, dma_addr, page_num));

  gm_assert (page_num < GM_MAX_NUM_HOST_HASH_TABLE_PIECES);
  gm_assert (is);
  gm_assert (is->lanai.sram);

  /* Compute pointer to LANai-resident array of pointers to hash table pieces
     at end of LANai SRAM. */

  slot = (gm_dp_n_t *)
    ((char *) is->lanai.sram + (is->lanai.eeprom.lanai_sram_size
				- GM_PAGE_HASH_PIECE_REF_TABLE_LEN));

  /* determine offset in table for records so we don't overwrite
     earlier records.  (We convert the number of PTEs to the number of
     pages.) */

  offset = (gm_u32_t) ((is->page_hash.max_index + 1)
		       * sizeof (gm_pte_t) / GM_PAGE_LEN);

  /* Record page dma address. */

  GM_PRINT (GM_PRINT_LEVEL >= 6,
	    ("Registering a page hash piece #0x%x at %p.\n", page_num,
	     &slot[page_num + offset]));
  gm_always_assert (page_num + offset < GM_MAX_NUM_HOST_HASH_TABLE_PIECES);
  slot[page_num + offset] = gm_hton_dp (dma_addr);
  gm_notice_dma_addr (is->dma_page_bitmap, dma_addr);
  GM_PRINT (GM_PRINT_LEVEL >= 6, ("Registered a page hash piece\n"));
  return GM_SUCCESS;
}
#endif /* GM_ENABLE_VM */

/* Fill the unfilled slots in LANai page table piece table with
   copies of the legitimate filled entries at the bottom of the table.

   Pointers to the page hash table pieces are stored in an array in
   the LANai SRAM, but this array may be not full.  Rather then tell
   the LANai the actual fullness of the table, fill the table
   cyclically with references to the pieces.  This way, the LANai
   control program need not know the actual number of pieces, but only
   the constant size of its worst-case piece table. This works because
   the size of the LANai table is guaranteed to be a multiple of the
   number of pieces, because each is a power of two.

   For example, if the page hash table is two pages long, and
   GM_PAGE_HASH_PIECE_REF_TABLE_LEN is 8, the DMA pointers in the
   GM_PAGE_HASH_PIECE table in the LANai would be initialized as
   follows:
   
     +---+---+---+---+---+---+---+---+
     |   |   |   |   |   |   |   |   | Page hash piece table in LANai SRAM.
     +---+---+---+---+---+---+---+---+
       |   |   |   |   |   |   |   |
   .---'---|---'---|---'---|---'   |
   |       |       |       |       | DMA pointers
   |       |-------'-------'-------'
   V       |
   +-----+ V
   |     | +-----+  Page hash table in two locked pages of host DMAable 
   +-----+ |     |  memory. Note that the pages need not be contiguous 
	   +-----+  in host DMA space.

   Note how the first two references in the page hash piece table are
   mirrored in the remaining slots in this table.  This allows the
   LANai to ignore the fact that the host hash table is dynamically
   sized and assume that there are GM_PAGE_HASH_PIECE_REF_TABLE_LEN
   pages in the table. */

static void
gm_mirror_page_hash_pieces (gm_instance_state_t * is, int pte_cnt)
{
  unsigned int legit_slots;
  gm_dp_t *slot;

  legit_slots = (unsigned int) (pte_cnt * sizeof (gm_pte_t) / GM_PAGE_LEN);
  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("There are 0x%x pages of non-redundant pieces.\n", legit_slots));

  /* Compute the location of the page hash piece table in the LANai. */

  gm_assert (is->lanai.sram);
  gm_assert (is->lanai.eeprom.lanai_sram_size);
  slot = (gm_dp_t *) ((char *) is->lanai.sram
		      + (is->lanai.eeprom.lanai_sram_size
			 - GM_PAGE_HASH_PIECE_REF_TABLE_LEN));

  /* Mirror the legitimate slots in the table. */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("mirroring page hash pieces.\n"));
  gm_assert (legit_slots > 0);
  gm_always_assert (legit_slots <= GM_MAX_NUM_HOST_HASH_TABLE_PIECES);
  gm_assert (GM_POWER_OF_TWO (legit_slots));
  gm_assert (GM_POWER_OF_TWO (GM_PAGE_HASH_PIECE_REF_TABLE_LEN));
  while (legit_slots < GM_MAX_NUM_HOST_HASH_TABLE_PIECES)
    {
      /* Double the number of legitimate entries. */
      __gm_nton_bcopy (is,
		       (gm_u8_n_t *) & slot[0],
		       (gm_u8_n_t *) & slot[legit_slots],
		       ((gm_u8_n_t *) & slot[legit_slots]
			- (gm_u8_n_t *) & slot[0]));
      legit_slots *= 2;
    }
  gm_assert (legit_slots == GM_MAX_NUM_HOST_HASH_TABLE_PIECES);

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Done mirroring page hash piece table.\n"));
}

/* Utility function for following functions. */

static gm_inline void
_gm_page_hash_segment_bin_for_bin (gm_page_hash_table_t * hash, int *segment,
				   unsigned int *bin)
{
  int seg, nib;

  GM_PRINT (GM_PRINT_LEVEL >= 10,
	    ("_gm_page_hash_segment_bin_for_bin (%p, %p, %p )\n", hash,
	     segment, bin));

  gm_assert (hash->max_index > 0);
  /* Since *bin is unsigned, if it has been decremented from 0 it will be
     greater than max_index.                                              */
  gm_assert (*bin <= (gm_u32_t) hash->max_index);

  /* Scan backwards (largest segment first) for constant, rather than
     log(table_size), average time. */

  nib = hash->max_index - *bin;	/* bins from end */
  seg = hash->segment_cnt - 1;
  while ((unsigned int) nib >= hash->segment[seg].entry_cnt)
    {
      nib -= hash->segment[seg].entry_cnt;
      seg--;
      gm_assert (seg >= 0);
    }
  gm_assert (nib >= 0);
  gm_assert (nib < (int) hash->segment[seg].entry_cnt);
  *bin = hash->segment[seg].entry_cnt - 1 - nib;
  gm_assert (*bin < hash->segment[seg].entry_cnt);
  *segment = seg;
}

/* Return the gm_pte_t structure corresponding to page
   hash table bin number BIN. */

static gm_inline gm_pte_t *
gm_pte_for_bin (gm_page_hash_table_t * hash, unsigned int bin)
{
  unsigned int segment;

  GM_CALLED_WITH_ARGS (("%p,%d", hash, bin));

  gm_assert ((int) bin <= hash->max_index);
  _gm_page_hash_segment_bin_for_bin (hash, (int *) &segment, &bin);

  GM_PRINT (GM_PRINT_LEVEL >= 10,
	    ("gm_pte_for_bin: segment = 0x%x  bin = 0x%x\n", segment, bin));

  GM_RETURN (&hash->segment[segment].entry[bin]);
}

/* Return the bin corresponding to USER_VMA for port PORT.  If
   there is an existing entry for the mapping, return the bin
   containing it.  If there is no existing entry, return the empty
   entry where the info should be stored. */

static gm_u32_t
gm_bin_for_mapping (gm_page_hash_table_t * hash, gm_up_t page_port)
{
  gm_pte_t *e;
  unsigned int bin;

  GM_CALLED_WITH_ARGS (("%p,"GM_U64_TMPL, hash, GM_U64_ARG (page_port)));

  gm_assert (GM_POWER_OF_TWO (sizeof (gm_pte_t)));
  gm_assert (GM_POWER_OF_TWO (hash->max_index + 1));
  gm_assert (hash->max_index > 0);

  gm_assert (page_port);

  /* Scan for the mapping bin. */

  bin = GM_HASH_PAGE_PORT (page_port) & hash->max_index;
  GM_PRINT (GM_PRINT_LEVEL >= 6, ("Hashed to bin 0x%x.\n", bin));

  e = gm_pte_for_bin (hash, bin);
  while (gm_pte_get_user_page (e))
    {
      if (gm_ntoh_up (e->page_port) == page_port)
	{
	  GM_RETURN (bin);
	}

      /* Try the next bin */

      gm_assert (GM_POWER_OF_TWO (hash->max_index + 1));
      bin = (bin + 1) & hash->max_index;
      e = gm_pte_for_bin (hash, bin);
    }
  GM_RETURN (bin);
}

/* return the DMA address corresponding a user virtual memory addres
   registered on a GM port. */

#define GM_DEBUG_DMA_ADDR_FOR_PORT_ADDR 0
gm_status_t
gm_dma_addr_for_port_addr (gm_page_hash_table_t *hash, unsigned int port,
			   gm_up_t addr, gm_dp_t *ret)
{
  gm_pte_t *e;
  gm_up_t page_port;
  unsigned int page_offset = GM_PAGE_OFFSET (addr);
  gm_dp_t dp;
  
  GM_CALLED_WITH_ARGS (("%p, %u, "GM_U64_TMPL", %p",
			hash, port, GM_U64_ARG (addr), ret));
  
  if (port >= GM_PAGE_LEN)
    {
      GM_RETURN_STATUS (GM_INVALID_PARAMETER);
    }
  page_port = addr - page_offset + port;
  e = gm_pte_for_bin (hash, gm_bin_for_mapping (hash, page_port));
  if (gm_pte_get_user_page (e) + page_offset != addr
      || gm_pte_get_port_id (e) != port)
    {
      GM_RETURN_STATUS (GM_INVALID_PARAMETER);
    }

  /* report success */

  dp = gm_pte_get_dma_page (e) + page_offset;
  GM_PRINT (GM_DEBUG_DMA_ADDR_FOR_PORT_ADDR,
	    (""GM_U64_TMPL" -> "GM_U64_TMPL"\n", GM_U64_ARG (addr),
	     GM_U64_ARG (dp)));
  if (ret)
    {
      *ret = dp;
    }
  GM_RETURN_STATUS (GM_SUCCESS);
}

#if GM_PRINT_LEVEL >= 5
void
gm_print_page_hash_table (gm_page_hash_table_t * hash)
{
  int bin;
  gm_pte_t *e;

  GM_PRINT (GM_PRINT_LEVEL >= 5,
	    ("\n\ngm_print_page_hash_table: hash->max_index = 0x%x.\n",
	     hash->max_index));
  for (bin = 0; bin <= hash->max_index; bin++)
    {
      gm_up_t page;

      e = gm_pte_for_bin (hash, bin);
      page = gm_pte_get_user_page (e);
      if (!page)
	continue;
      GM_PRINT (GM_PRINT_LEVEL >= 5,
		("bin = 0x%x, virt_page = 0x%x, port = 0x%x;\n",
		 bin, page, gm_pte_get_port_id (e)));
    }
}
#endif

/* Double the size of the page hash table (or initialize it if the
   size is zero).  Note that this does not invalidate the cached PTEs
   in the LANai. */

static gm_status_t
gm_grow_page_hash_table (gm_instance_state_t * is)
{
  gm_u32_t N, seg;
  gm_page_hash_table_t *hash;

  GM_PRINT ((GM_DEBUG_PAGE_TABLE || GM_DEBUG_PAGE_HASH), 
            ("Growing page hash table.\n"));

  hash = &is->page_hash;

  /* Until now, we have not modified the page table from the point of
     view of the LANai, but we are now about to, so pause the LANai
     during the changes. Specifically, the gm_arch_dma_region_alloc
     callback function will modify the LANai tables.
  */

  gm_pause_lanai (is);

  /* Determine new page hash table segment number. */

  seg = hash->segment_cnt;

  /* Verify we can handle another page table segment. */

  if (hash->max_index == 0x3fffffff 
      || seg == GM_NUM_PAGE_TABLE_SEGMENTS
      || hash->max_index == GM_PAGE_HASH_MAX_INDEX)
    {
      GM_NOTE (("Hash table reached addressing limit.\n"));
      goto abort_with_nothing;
    }

  /* compute number of entries to allocate. */

  if (hash->max_index == -1)
    {
      /* This is the first allocation, so create 1 page worth of entries. */
      N = (gm_u32_t) (GM_PAGE_LEN / sizeof (gm_pte_t));
    }
  else
    {
      /* Double the number of entries. */
      N = hash->max_index + 1;
    }


  gm_assert (GM_POWER_OF_TWO (N));
  gm_assert (GM_PAGE_ALIGNED (N * sizeof (gm_pte_t)));

  GM_PRINT (GM_DEBUG_HOST_DMA,
	    ("Adding 0x%x entries to page hash table.\n", N));

  if (gm_arch_dma_region_alloc (is,
				&hash->segment[seg].dma_region,
				N * sizeof (gm_pte_t),
				GM_ARCH_DMA_READ | GM_ARCH_DMA_WRITE
				| GM_ARCH_DMA_CONSISTENT,
#if GM_ENABLE_VM
				gm_register_page_hash_piece,
#else
				NULL,
#endif
				is) != GM_SUCCESS)
    {
      GM_NOTE (("Failed to allocate segment of page hash table.\n"));
      goto abort_with_nothing;
    }

  hash->segment[seg].entry
    = (gm_pte_t *) (gm_arch_dma_region_kernel_addr
		    (&hash->segment[seg].dma_region));
  gm_bzero (hash->segment[seg].entry, N * sizeof (gm_pte_t));

  /* Commit to using this segment now that it has been fully allocated. */

  hash->max_index += N;
  gm_assert (GM_POWER_OF_TWO (hash->max_index + 1));
  hash->segment_cnt = seg + 1;
  hash->segment[seg].entry_cnt = N;

  /* Remove and reinsert each entry in the page hash table to restore
     hash table consistency.  This is done in reverse order to
     minimize the need to rehash clusters of PTEs. */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Rehashing all PTEs.\n"));
  {
    gm_s32_t bin;

    for (bin = N - 1; bin >= 0; bin--)
      {
	gm_pte_t *e, *new_e;
	int new_bin;

	e = gm_pte_for_bin (hash, bin);
	gm_assert (e);
	if (!gm_pte_get_user_page (e))
	  continue;

	/* find new location for entry if needed. */
	new_bin = gm_bin_for_mapping (hash, gm_pte_get_page_port (e));
	if (new_bin == bin)
	  continue;

	GM_PRINT (GM_PRINT_LEVEL >= 5,
		  ("*** Relocating a page hash mapping from bin "
		   "0x%x to bin 0x%x\n", bin, new_bin));
	_GM_PRINT (GM_PRINT_LEVEL >= 5,
		   ("virtual_addr = %p, port = 0x%x\n",
		    (unsigned int) gm_pte_get_user_page (e),
		    gm_pte_get_port_id (e)));

	/* copy entry to new location */
	GM_PRINT (GM_PRINT_LEVEL >= 5,
		  ("Calling gm_pte_for_bin(?,%d)\n", new_bin));
	new_e = gm_pte_for_bin (hash, new_bin);
	*new_e = *e;
#if GM_CAN_REGISTER_MEMORY
	gm_hash_rekey (hash->page_lock_hash, e, new_e);
#endif
	gm_pte_clear (e);

	/* Arrange to reinsert all entries after the moved entry.
	   This while loop will not overrun the end of the page table
	   because either the table is empty, or there is guaranteed
	   to be an empty slot between the original BIN and the end of
	   the table because the table is less than half full and
	   entries in the upper half of the table never need
	   relocating, and BIN starts in the first half of the
	   table. */
	GM_PRINT (GM_PRINT_LEVEL >= 5, ("Scanning \n"));
	do
	  {
	    bin++;
	    gm_assert (bin <= hash->max_index);
	  }
	while (gm_pte_get_user_page (gm_pte_for_bin (hash, bin)));
      }
  }

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Mirroring PTE piece table.\n"));
  /* Update the page hash piece table in the LANai. */
  gm_mirror_page_hash_pieces (is, hash->max_index + 1);

#if GM_PRINT_LEVEL >= 5
  gm_print_page_hash_table (hash);
#endif

  gm_unpause_lanai (is);

  return GM_SUCCESS;

abort_with_nothing:
  gm_unpause_lanai (is);

  return GM_FAILURE;
}

/****************************************************************
 * Page table
 ****************************************************************/

/* Allocate the page hash table.

   The page hash table keeps track of the mappings of virtual
   addresses for each port to dma addresses.  It is big, so we store
   it in host memory and the MCP caches entries in LANai SRAM.

   There is one page hash table per device.  We would prefer to have
   just one shared table, but there appears to be no clean way to do
   this in Solaris.  It might be possible, but doesn't seem worth the
   effort.

   The hash table potentially needs enough storage to store an entry
   for every page of physical memory.  It is not necessary to allocate
   more capacity than this, because it is impossible to allocate all
   physical pages for DMA, but we dynamically allocate enough capacity
   to keep the hash table at least 50% empty to minimize hash
   collisions.

   The hash table uses a linear-scan-on-miss strategy.  That is, if
   the location for an entry is occupied, the entry is stored in the
   next empty bin.

   NOTE: No need to mutex protect the table as it is being created. */

gm_status_t
gm_alloc_page_hash_table (gm_instance_state_t * is)
{
  gm_status_t status;
  gm_page_hash_table_t *hash;

  GM_CALLED ();

  hash = &is->page_hash;

  GM_PRINT (GM_DEBUG_PAGE_TABLE, ("**************************\n"));
  GM_PRINT (GM_DEBUG_PAGE_TABLE, ("allocating page hash table\n"));
  GM_PRINT (GM_DEBUG_PAGE_TABLE, ("**************************\n"));

  /* Initialize hash table */

  gm_bzero (hash, sizeof (*hash));

  GM_PRINT (GM_DEBUG_SYNC, ("Initializing page_hash sync for unit %d\n",
			    is->id));
  gm_arch_sync_init (&hash->sync, is);
  hash->max_index = -1;
  gm_assert (GM_POWER_OF_TWO (hash->max_index + 1));

  /* Allocate pages to be used for send and receive DMA any time the user
     specifies a send or receive using nonDMAable memory.  Such receives are
     discarded by DMAing them into the bogus rdma page, and such sends send
     all zeros from the bogus send page. The pointers are passed to the LANai
     eventually, where the MCP uses them for any DMA to or from a non-DMA
     virtual memory address page. */

  /* Alloc page for bogus receives. */
  /* mark page for dma reads and writes for DMA testing */

  GM_PRINT ((GM_DEBUG_PAGE_TABLE | GM_DEBUG_HOST_DMA),
            ("Allocating bogus receive page.\n"));
  status 
    = gm_arch_dma_region_alloc (is, &hash->bogus_rdma_region,
				GM_PAGE_LEN,
				GM_ARCH_DMA_READ | GM_ARCH_DMA_WRITE | 
				GM_ARCH_DMA_STREAMING,
				NULL, NULL);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not allocate dma page for bogus receives.\n"));
      goto abort_with_sync;
    }

  /* Alloc page for bogus sends. */
  /* mark page for dma reads and writes for DMA testing */

  GM_PRINT ((GM_DEBUG_PAGE_TABLE | GM_DEBUG_HOST_DMA),
            ("Allocating bogus send page.\n"));
  status
    = gm_arch_dma_region_alloc (is, &hash->bogus_sdma_region,
				GM_PAGE_LEN,
				GM_ARCH_DMA_READ | GM_ARCH_DMA_WRITE |
				GM_ARCH_DMA_STREAMING,
				NULL, NULL);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not allocate dma page for bogus sends.\n"));
      goto abort_with_bogus_rdma_page;
    }

  /* Initialize page for bogus sends to be filled with 0xAAAAAAAAs.
     We used to use 0xdeadbeef, but that was endian dependent. */

  {
    gm_u32_t *p, *p_limit;

    gm_assert (GM_PAGE_LEN % 4 == 0);
    p = gm_arch_dma_region_kernel_addr (&hash->bogus_sdma_region);
    p_limit = (gm_u32_t *) ((char *) p + GM_PAGE_LEN);

    for (; p < p_limit; p++)
      *p = 0xAAAAAAAA;
  }

  /* Create a lookaside list for page locks. */
  hash->page_lock_lookaside = gm_create_lookaside (sizeof (struct
                                                           gm_page_lock),
                                                   0);
  if (!hash->page_lock_lookaside)
    {
      GM_NOTE (("Could not create page lock lookaside.\n"));
      goto abort_with_bogus_sdma_page;
    }
  

  /* Create hash table for memory page locks.  This is done even if we
     don't need the locks, since it doesn't waste much memory and is
     simpler. */

  hash->page_lock_hash = gm_create_hash (gm_hash_compare_ptrs,
					 gm_hash_hash_ptr,
					 0, 0, 0, 0);
  if (!hash->page_lock_hash)
    {
      GM_NOTE (("Could not create page lock hash table.\n"));
      goto abort_with_page_lock_lookaside;
    }

  /* Allocate the minimal page hash table, since the LANai needs to be
     able to assume that it exists. */

  {
    unsigned long maxsize = 1;

#if 0 && GM_OS_LINUX
      maxsize = 2*gm_arch_max_locked_pages();
#endif

    if (maxsize<=0)
      maxsize=1;

    do 
      {
        status = gm_grow_page_hash_table (is);
        if (status != GM_SUCCESS)
          {
            GM_NOTE (("Could not allocate the page hash table maxsize=%d.\n",maxsize));
            goto abort_with_page_hash_lock;
          }
      } while(hash->max_index < maxsize);
    GM_PRINT (GM_DEBUG_BOARD_INIT,
	      ("page hash table max_index = %d\n", hash->max_index));
  }

  hash->to_unlock = 0;

  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_page_hash_lock:
  gm_destroy_hash (hash->page_lock_hash);
abort_with_page_lock_lookaside:
  gm_destroy_lookaside (hash->page_lock_lookaside);
abort_with_bogus_sdma_page:
  gm_arch_dma_region_free (&is->page_hash.bogus_sdma_region);
abort_with_bogus_rdma_page:
  gm_arch_dma_region_free (&is->page_hash.bogus_rdma_region);
abort_with_sync:
  gm_arch_sync_destroy (&is->page_hash.sync);
  gm_bzero (&is->page_hash, sizeof (is->page_hash));
  GM_RETURN_STATUS (GM_FAILURE);
}

void
gm_free_page_hash_table (gm_instance_state_t * is)
{
  unsigned int seg;

  gm_assert (is);
  gm_assert (!is->lanai.running);
#if 0
  if (is->lanai.running)
    gm_disable_lanai (is);
#endif

  seg = is->page_hash.segment_cnt;
  GM_PRINT ((GM_DEBUG_PAGE_TABLE | GM_DEBUG_HOST_DMA),
	    ("Freeing page hash table regions (there are %d).\n", seg));
  gm_assert (seg < 64);
  while (seg--)
    {
      GM_PRINT (GM_DEBUG_PAGE_TABLE,
		("Freeing page hash table segment %d.\n", seg));
      gm_arch_dma_region_free (&is->page_hash.segment[seg].dma_region);
    }
  is->page_hash.segment_cnt = seg;

  GM_PRINT (GM_DEBUG_PAGE_TABLE, ("Destroying page lock hash\n"));
  gm_destroy_hash (is->page_hash.page_lock_hash);

  GM_PRINT (GM_DEBUG_PAGE_TABLE, ("Destroying page lock lookaside\n"));
  gm_destroy_lookaside (is->page_hash.page_lock_lookaside);

  GM_PRINT ((GM_DEBUG_PAGE_TABLE | GM_DEBUG_HOST_DMA),
            ("Freeing bogus SDMA region.\n"));
  gm_arch_dma_region_free (&is->page_hash.bogus_sdma_region);

  GM_PRINT ((GM_DEBUG_PAGE_TABLE | GM_DEBUG_HOST_DMA),
            ("Freeing bogus RDMA region.\n"));
  gm_arch_dma_region_free (&is->page_hash.bogus_rdma_region);
  gm_arch_sync_destroy (&is->page_hash.sync);
  gm_bzero (&is->page_hash, sizeof (is->page_hash));
}

/* Return the DMA pointer for the mapping, or (gm_dp_t)0 if no mapping
   exists. */

gm_dp_t
gm_dma_addr_for_mapping (gm_instance_state_t * is,
			 gm_up_t user_vma, gm_u32_t port)
{
  gm_dp_t ret;
  unsigned int offset;
  
  GM_CALLED_WITH_ARGS (("%p, %lx, %u", is, (long) user_vma, port));

  offset = GM_PAGE_OFFSET (user_vma);
  user_vma -= offset;
  
#if !GM_ENABLE_VM
  {
    gm_dp_t set, clear;
    
    set = gm_read_lanai_global_dp (is, no_VM.set);
    clear = gm_read_lanai_global_dp (is, no_VM.clear);
    ret = (gm_dp_t) ((gm_up_t) user_vma);
    gm_assert ((set & clear) == 0);
    ret |= set;
    ret &= ~clear;
  }
#else  /* GM_ENABLE_VM */
  {
    unsigned int bin;
    gm_pte_t *e;
    gm_page_hash_table_t *hash;

    gm_assert (GM_PAGE_ALIGNED (user_vma));
    gm_assert (port <= GM_NUM_PORTS);

    GM_PRINT (GM_DEBUG_PAGE_TABLE, (__GM_FUNCTION__"(%p, %lx, %d)\n",
				    is, (unsigned long) user_vma, port));

    hash = &is->page_hash;

    gm_assert (GM_POWER_OF_TWO (sizeof (gm_pte_t)));
    gm_assert (user_vma);
    GM_PRINT (GM_DEBUG_PAGE_TABLE,
	      ("hash->max_index = 0x%x\n", hash->max_index));
    gm_assert (GM_POWER_OF_TWO (hash->max_index + 1));

    GM_PRINT (GM_PRINT_LEVEL >= 10, ("finding refs for mapping (%p,%x)->XXX\n",
				     user_vma, port));

    bin = gm_bin_for_mapping (hash, user_vma + port);

    GM_PRINT (GM_PRINT_LEVEL >= 10,
	      ("%s: bin = 0x%x\n", __GM_FUNCTION__, bin));

    e = gm_pte_for_bin (hash, bin);

    /* verify entry is nonempty and return the DMA address. */

    if (gm_pte_get_user_page (e))
      {
	ret = gm_pte_get_dma_page (e);
      }
    else
      {
	ret = 0;
      }
  }
#endif /* GM_ENABLE_VM */

  ret += offset;
  GM_RETURN_DP (ret);
}

/* Increments the reference count of a mapping already in the page
   table.

   Return values:
   GM_SUCCESS on success
   GM_INVALID_PARAMETER if their is no extant mapping to reference
   GM_FAILURE otherwise

   Returns GM_SUCCESS unless the reference is not in the table
   or the mapping has been referenced too many times. */

static gm_status_t
gm_reference_mapping (gm_instance_state_t * is,
		      gm_up_t user_vma, gm_u32_t port, gm_u32_t *bin,
		      gm_pte_t **e)
{
  gm_page_hash_table_t *hash;

  gm_assert (GM_PAGE_ALIGNED (user_vma));
  gm_assert (port <= GM_NUM_PORTS);

  GM_PRINT (GM_DEBUG_PAGE_TABLE,
	    ("gm_reference_mapping(%p, %lx, %d)\n",
	     is, (unsigned long) user_vma, port));

  hash = &is->page_hash;

  gm_assert (GM_POWER_OF_TWO (sizeof (gm_pte_t)));
  gm_assert (user_vma);
  GM_PRINT (GM_DEBUG_PAGE_TABLE,
	    ("hash->max_index = 0x%x\n", hash->max_index));
  gm_assert (GM_POWER_OF_TWO (hash->max_index + 1));

  GM_PRINT (GM_PRINT_LEVEL >= 10, ("finding refs for mapping (%p,%x)->XXX\n",
				   user_vma, port));

  *bin = gm_bin_for_mapping (hash, user_vma + port);

  GM_PRINT (GM_PRINT_LEVEL >= 10,
	    ("gm_reference_mapping: bin = 0x%x\n", *bin));

  *e = gm_pte_for_bin (hash, *bin);

  /* verify entry is in the table */

  if (!gm_pte_get_user_page (*e))
    {
      return GM_INVALID_PARAMETER;
    }

  /* increment reference count, checking for overflow */

  if (gm_pte_incr_ref_cnt (*e) != GM_SUCCESS)
    {
      GM_PRINT (GM_DEBUG_PAGE_TABLE,
		("*** gm_reference_mapping: overflow?? \n"));
      return GM_FAILURE;
    }

  /* increment reference count */

  hash->total_refs++;

  return GM_SUCCESS;
}

/* Add the translation (PORT_VMA,PORT)->DMA_ADDR to the page
   translation hash table.  If DMA_ADDR is not supplied, pin USER_VMA
   and use its DMA address.

   GM_RETURNS: the number of earlier references to this mapping.

   WARNING: This function assumes it has exclusive access to the page
   hash table. Consequently, the ps->instance->page_hash.sync mutex must
   have been allocated by a caller. */

gm_status_t
gm_add_mapping_to_page_table (gm_port_state_t *ps,
			      gm_up_t user_vma,
			      gm_up_t port_vma,
			      gm_dp_t dma_addr)
{
  gm_instance_state_t *is;
  gm_u32_t bin;
  gm_pte_t *e;
  gm_page_hash_table_t *hash;
  gm_status_t status;
#if GM_CAN_REGISTER_MEMORY
  struct gm_page_lock *lock;
#endif
  unsigned int port;

  GM_CALLED_WITH_ARGS (("%p,"GM_U64_TMPL","GM_U64_TMPL","GM_U64_TMPL,
			ps, GM_U64_ARG (user_vma), GM_U64_ARG (port_vma),
			GM_U64_ARG (dma_addr)));
  
  port = ps->id;
  gm_assert (ps);
  is = ps->instance;
  gm_assert (is);
  port = ps->id;
  
  hash = &is->page_hash;

  gm_assert (GM_POWER_OF_TWO (sizeof (gm_pte_t)));
  gm_assert (user_vma);
  gm_assert (port_vma);
  gm_assert (GM_POWER_OF_TWO (hash->max_index + 1));
  gm_assert (hash->max_index > 0);
  gm_assert (GM_PAGE_ALIGNED (port_vma));
  gm_assert (GM_PAGE_ALIGNED (user_vma));
  gm_assert (GM_PAGE_ALIGNED (dma_addr));

  /* If the mapping is already in the table, reference it and return. */

  status = gm_reference_mapping (is, port_vma, port, &bin, &e);

  GM_PRINT (GM_DEBUG_PAGE_TABLE,
	    ("gm_add_mapping_to_page_table: "
	     "gm_reference_mapping returns %d\n", status));

  if (status == GM_SUCCESS)
    {
      gm_assert (gm_noticed_dma_addr (is->dma_page_bitmap, dma_addr)
		 == GM_SUCCESS);
      GM_PRINT (GM_DEBUG_PAGE_TABLE, ("%p already in the table\n",
				      (char *) port_vma));
      GM_RETURN_STATUS (GM_SUCCESS);
    }

  if (status == GM_FAILURE)
    {
      GM_RETURN_STATUS (GM_FAILURE);
    }
  
  gm_assert (status == GM_INVALID_PARAMETER);

  GM_PRINT (GM_DEBUG_PAGE_TABLE,
	    ("gm_add_mapping_to_page_table: invalid_parameter?\n"));

  /* grow the hash table if adding a new entry would make the table
     overfull. */

  if (hash->filled_entries >= (gm_u32_t) (hash->max_index / 2))
    {
      if (gm_grow_page_hash_table (is) != GM_SUCCESS)
	goto abort_with_nothing;
      gm_assert (hash->filled_entries < (gm_u32_t) (hash->max_index >> 1));
      bin = gm_bin_for_mapping (hash, (gm_up_t) port_vma + port);
      e = gm_pte_for_bin (hash, bin);
    }

  /* Compute the bin in which to record the mapping */

  GM_PRINT (GM_DEBUG_PAGE_TABLE,
	    ("Using bin 0x%x (0 earlier references).\n", bin));

#if GM_CAN_REGISTER_MEMORY

  /* If needed, lock the user buffer page and get its DMA addr. */
  
  if (dma_addr == 0)
    {
      /* Allocate storage for the page lock */

      gm_assert (!gm_hash_find (hash->page_lock_hash, e));
      lock = gm_lookaside_zalloc (hash->page_lock_lookaside);
      if (!lock)
        {
          GM_NOTE (("Could not allocate storage for page lock.\n"));
          goto abort_with_nothing;
        }
      if (gm_hash_insert (hash->page_lock_hash, e, lock) != GM_SUCCESS)
	{
	  GM_NOTE (("Could not allocate storage for page lock.\n"));
	  goto abort_with_lock;
	}

      /* Lock the page and get its dma address */
      if (gm_arch_lock_user_buffer_page (ps, user_vma, &dma_addr, &lock->arch)
	  != GM_SUCCESS)
	{
	  GM_PRINT (GM_PRINT_LEVEL > 0, ("Could not lock user buffer.\n"));
	  GM_PRINT (GM_PRINT_LEVEL > 0, ("Failed to add mapping (0x%lx,0x%x)\n",
					 (unsigned long) port_vma, port));
	  gm_assert (dma_addr == 0);
	  goto abort_with_hash_insert;
	}
      GM_PRINT (GM_DEBUG_PAGE_TABLE,
		("adding mapping (%lx,%x)->%p\n",
		 (unsigned long) port_vma, port, dma_addr));
    }
#endif
  
  /* Record the DMA address of the mapping.  This should be done before
     the PTE is marked valid. */

  gm_notice_dma_addr (is->dma_page_bitmap, dma_addr);

  /* Fill PTE */
  
  gm_assert (dma_addr);
  gm_pte_set_dma_page (e, dma_addr);
  gm_pte_set_ref_cnt (e, 1);
  GM_STBAR();
  gm_assert (gm_pte_get_user_page (e) == 0);
  gm_pte_set_page_port (e, (gm_up_t) port_vma + port); /* mark PTE valid */
  
  hash->filled_entries++;
  GM_PRINT (GM_DEBUG_PAGE_TABLE,
	    ("There are now %d filled hash bins. (gm_add_mapping)\n",
	     hash->filled_entries));
  hash->total_refs++;
  GM_RETURN_STATUS (GM_SUCCESS);

#if GM_CAN_REGISTER_MEMORY
 abort_with_hash_insert:
  gm_hash_remove (hash->page_lock_hash, e);
 abort_with_lock:
  gm_lookaside_free (lock);
#endif /* GM_CAN_REGISTER_MEMORY */
 abort_with_nothing:
  GM_RETURN_STATUS (GM_FAILURE);
}

static gm_status_t
gm_mapping_in_page_table (gm_instance_state_t * is,
			  gm_up_t user_vma, gm_u32_t port)
{
  unsigned int bin;
  gm_pte_t *e;
  gm_page_hash_table_t *hash;

  gm_assert (GM_PAGE_ALIGNED (user_vma));
  gm_assert (port <= GM_NUM_PORTS);

  GM_PRINT (GM_DEBUG_PAGE_TABLE,
	    ("gm_mapping_in_page_table(%p, %lx, %d)\n",
	     is, (unsigned long) user_vma, port));

  hash = &is->page_hash;

  gm_assert (GM_POWER_OF_TWO (sizeof (gm_pte_t)));
  gm_assert (user_vma);
  GM_PRINT (GM_DEBUG_PAGE_TABLE,
	    ("hash->max_index = 0x%x\n", hash->max_index));
  gm_assert (GM_POWER_OF_TWO (hash->max_index + 1));

  GM_PRINT (GM_PRINT_LEVEL >= 10, ("finding refs for mapping (%p,%x)->XXX\n",
				   user_vma, port));

  bin = gm_bin_for_mapping (hash, user_vma + port);

  GM_PRINT (GM_PRINT_LEVEL >= 10,
	    ("gm_mapping_in_page_table: bin = 0x%x\n", bin));

  e = gm_pte_for_bin (hash, bin);

  /* verify entry is in the table */

  if (!gm_pte_get_user_page (e))
    {
      return GM_INVALID_PARAMETER;
    }

  return GM_SUCCESS;
}

/****************************************************************
 * Page Deregistration
 ****************************************************************/

/* Returns the number of references to a mapping in the page table,
   or 0 if the mapping is not in the table or if the DMA address of
   the mapping does not match the DMA address specified. */

gm_u32_t
gm_refs_to_mapping_in_page_table (gm_port_state_t * ps,
				  gm_up_t user_vma, gm_u32_t port)
{
  unsigned bin;
  gm_pte_t *e;
  gm_page_hash_table_t *hash;
  gm_instance_state_t *is = ps->instance;

  hash = &is->page_hash;

  gm_assert (GM_PAGE_ALIGNED (user_vma));
  gm_assert (GM_POWER_OF_TWO (sizeof (gm_pte_t)));
  gm_assert (user_vma);
  gm_assert (hash->max_index > 0);
  gm_assert (GM_POWER_OF_TWO (hash->max_index + 1));

  GM_PRINT (GM_DEBUG_PAGE_TABLE,
	    ("finding refs for mapping (%lx,%x)->XXX\n",
	     (unsigned long) user_vma, port));

  bin = gm_bin_for_mapping (hash, user_vma + port);
  e = gm_pte_for_bin (hash, bin);
  gm_assert (gm_pte_get_user_page (e) != 0 || gm_pte_get_ref_cnt (e) == 0);
  return gm_pte_get_ref_cnt (e);
}

/* Arrange to later unpin the page locked with lock. */

void
gm_schedule_unlock (gm_instance_state_t *is, struct gm_page_lock *lock)
{
  if (!lock)
    {
      /* This mapping must not require an unlock. */
      return;
    }
  
  lock->next = is->page_hash.to_unlock;
  is->page_hash.to_unlock = lock;
}

/* Perform the page unlocks scheduled by gm_schedule_unlock() */

void
gm_dereference_mapping_commit (gm_instance_state_t *is, gm_port_state_t *ps)
{
  struct gm_page_lock *lock, *next;

  GM_CALLED_WITH_ARGS (("%p,%p", is, ps));

  for (lock=is->page_hash.to_unlock; lock; lock=next)
    {
      next = lock->next;
#if GM_CAN_REGISTER_MEMORY
      gm_arch_unlock_user_buffer_page (&lock->arch);
#endif
      gm_lookaside_free (lock);
    }
  is->page_hash.to_unlock = 0;
  GM_RETURN_NOTHING ();
}


/* Remove a reference to the translation (USER_ADDR,PORT)->DMA_ADDR
   from the page translation hash table.  Returns the number of
   remaining references to this mapping.

   WARNING: It is an error to call this function to remove a non-existant
   mapping from the page table.

   WARNING: This function assumes it has exclusive access to the page
   hash table. Consequently, the ps->instance->page_hash.sync mutex must
   have been allocated by a caller. */

gm_u32_t
gm_dereference_mapping (gm_port_state_t *ps, gm_up_t user_vma)
{
  gm_instance_state_t *is;
  gm_page_hash_table_t *hash;
  gm_pte_t *e;
  unsigned int bin;
  unsigned int port;
  
  GM_CALLED ();

  gm_assert (ps);
  is = ps->instance;
  gm_assert (is);
  
  port = ps->id;
  
  hash = &is->page_hash;

  gm_assert (GM_POWER_OF_TWO (sizeof (gm_pte_t)));
  gm_assert (user_vma);
  gm_assert (GM_PAGE_ALIGNED (user_vma));
  gm_assert (hash->max_index > 0);
  gm_assert (GM_POWER_OF_TWO (hash->max_index + 1));

  GM_PRINT (GM_DEBUG_PAGE_TABLE,
	    ("removing mapping (%lx,%x)->??\n",
	     (unsigned long) user_vma, port));

  /* Find the mapping. */

  bin = gm_bin_for_mapping (hash, (gm_up_t) user_vma + port);
  e = gm_pte_for_bin (hash, bin);
  gm_assert (gm_pte_get_user_page (e));

  /* Dereference the bin */

  {
    gm_u32_t remaining_refs;

    remaining_refs = gm_pte_decr_ref_cnt (e);
    hash->total_refs--;
    GM_PRINT (GM_DEBUG_PAGE_TABLE,
	      ("%d refs remaining.\n", (int) hash->total_refs));
    if (remaining_refs != 0)
      GM_RETURN (remaining_refs);
  }

  /****
   * The mapping is no longer referenced
   ****/

  /* Catch deregistration of recv queue pages and invalidate the receive
     queue if we catch any. */

  gm_catch_recv_queue_deregistration (ps, user_vma);
  
#if GM_CAN_REGISTER_MEMORY

  /* Unlock the page and free the lock */

  {
    struct gm_page_lock *lock;

    lock = (struct gm_page_lock *) gm_hash_remove (hash->page_lock_hash, e);
    gm_schedule_unlock (is, lock);
  }

#endif

  /* remove the entry from the table */
  gm_assert (gm_noticed_dma_addr (is->dma_page_bitmap, gm_pte_get_dma_page(e)) == GM_SUCCESS);
  gm_forget_dma_addr (is->dma_page_bitmap, gm_pte_get_dma_page(e));
  gm_pte_clear (e);
  hash->filled_entries--;
  GM_PRINT (GM_DEBUG_PAGE_TABLE,
	    ("There are now %d filled hash bins. "
	     "(gm_dereference_mapping)\n", hash->filled_entries));

  /* Remove and reinsert each entry following the removed entry. */

  bin = (bin + 1) & hash->max_index;
  e = gm_pte_for_bin (hash, bin);
  while (gm_pte_get_user_page (e))
    {
      gm_u32_t new_bin;

      new_bin = gm_bin_for_mapping (hash, gm_pte_get_page_port (e));
      if (new_bin != bin)
	{
	  gm_pte_t *new_e;

	  new_e = gm_pte_for_bin (hash, new_bin);

	  /* Move entry to new location. */
	  *new_e = *e;
	  gm_pte_clear (e);
#if GM_CAN_REGISTER_MEMORY
	  /* update the lock hash table key, since it has changed */
	  gm_hash_rekey (hash->page_lock_hash, e, new_e);
#endif
	}

      bin = (bin + 1) & hash->max_index;
      e = gm_pte_for_bin (hash, bin);
    }

  GM_RETURN (0);
}

/* Remove all page table entries for port PORT from the page table for
   the interface described by IS.  This prevents the LANai from
   performing DMAs into any of the memory associated with that
   port. */

void
gm_disable_port_DMAs (gm_port_state_t * ps)
{
  unsigned int i, port;
  gm_instance_state_t *is;

  GM_CALLED ();

  GM_PRINT (GM_DEBUG_PORT_STATE, ("Disabling port DMAs.\n"));

  port = ps->id;
  is = ps->instance;
  gm_assert (is);

  gm_arch_mutex_enter (&is->page_hash.sync);

  /* Remove the page table entries if the page hash table exists. */

  gm_pause_lanai (is);
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Removing page hash table entries.\n"));
  if (is->page_hash.max_index != -1)	/* if table exists */
    {
      /* scan for PTEs for this port. */

      for (i = 0; i <= (unsigned int) is->page_hash.max_index; i++)
	{
	  gm_u32_t port_num;
	  gm_up_t user_vma;
	  gm_pte_t *e;

	  e = gm_pte_for_bin (&is->page_hash, i);
	  user_vma = gm_pte_get_user_page (e);
	  if (!user_vma)
	    continue;

	  port_num = gm_pte_get_port_id (e);
	  if (port_num != port)
	    continue;

	  /* Remove the entry from the page hash table. */

	  gm_pte_set_ref_cnt (e, 1);
	  gm_dereference_mapping (ps, user_vma);
	  /* Arrange to recheck same slot. */
	  i--;
	}
    }
  gm_unpause_lanai (is);
  
  /* Command the LANai to reset the port.  This must be done *after*
     the port mappings have been removed from the page table to
     prevent the cache from recaching entries after the LANai side of
     the port is closed. */

  gm_lanai_command_uncache_port_ptes (is, port);

  /* Unpin the mappings only now that we know the LANai PTE cache is
     consistent with the host side table. */

  gm_dereference_mapping_commit (is, ps);

  if (is->lanai.running)
    {
      /* Deschedule any pending host wakeup. */

      GM_PRINT (GM_PRINT_LEVEL >= 4, ("Descheduling host wakeup.\n"));
      /* gm_kton_8 (is, &is->lanai.globals->port[port].wake_host, 0);
         close, below, will do this, and we can't do partword PIO
         while the LANai is paused. */
      gm_arch_sync_reset (&ps->sleep_sync);

    }

  /* Make changes visible to LANai. */

  GM_PRINT ((GM_DEBUG_HOST_DMA),
            ("Synchronizing dma regions.\n"));
  {
    unsigned int j;
    for (j = 0; j < is->page_hash.segment_cnt; j++)
      {
	gm_arch_dma_region_sync (&is->page_hash.segment[j].dma_region,
				 GM_ARCH_SYNC_FOR_DEVICE);
      }
  }

  gm_arch_mutex_exit (&is->page_hash.sync);

  GM_RETURN_NOTHING ();
}

#if !GM_CAN_REGISTER_MEMORY
/* Attempt to grow the copy block for device to LENGTH */

gm_status_t
gm_expand_copy_block (gm_port_state_t * ps, gm_u32_t length)
{
  gm_instance_state_t *is;
  struct gm_segment_desc *seg;
  gm_status_t status;

  GM_CALLED ();
  GM_PRINT (GM_DEBUG_HOST_DMA, ("*** Expanding the copy block ***\n"));

  is = ps->instance;

  /* Disallow oversized allocations. */

  if (length > ps->max_dma_mem)
    GM_RETURN_STATUS (GM_PERMISSION_DENIED);

  /* Expand the copy block by adding segments until the copy block is large
     enough. */

  while (ps->copy_block.len < length)
    {
      gm_size_t alloc_len;

      /* Should never need more than 16 segments. */

      gm_assert (ps->copy_block.segment_cnt < 32);

      seg = &ps->copy_block.segment[ps->copy_block.segment_cnt];

      /* Alloc segments of length GM_PAGE_LEN, GM_PAGE_LEN, GM_PAGE_LEN*2,
         GM_PAGE_LEN*4, &c */

#if GM_BC_DRIVER
      if (ps->compatibility_mode)
	alloc_len = ps->max_dma_mem;
      else
	alloc_len = (unsigned) (ps->copy_block.len ?
				ps->copy_block.len : GM_PAGE_LEN);
#else
      alloc_len = (unsigned) (ps->copy_block.len ?
			      ps->copy_block.len : GM_PAGE_LEN);
#endif

      gm_assert (GM_POWER_OF_TWO (alloc_len));

      GM_PRINT (GM_PRINT_LEVEL >= 3,
		("[%d]: Allocating a segment of 0x%x bytes for port %d.\n",
		 is->id, alloc_len, ps->id));

      status = gm_arch_dma_region_alloc (is, &seg->dma_region, alloc_len,
					 GM_ARCH_DMA_RDWR
					 | GM_ARCH_DMA_STREAMING,
					 NULL, NULL);
      if (status != GM_SUCCESS)
	{
	  GM_NOTE (("[%d]: Allocation of dma segment failed.\n", is->id));
	  goto abort_with_nothing;
	}
      seg->data = (char *) gm_arch_dma_region_kernel_addr (&seg->dma_region);
      GM_PRINT (GM_DEBUG_HOST_DMA,
		("New copy block segment at KVA %p.\n", seg->data));

#if GM_BC_DRIVER

#error not platform-independent

      /* Require compatibility-mode DMA memory to be contiguous. */

      if (ps->compatibility_mode && seg->dma_region.cookie_cnt != 1)
	{
	  status = GM_OUT_OF_MEMORY;
	  GM_NOTE (("Could not allocate contiguous memory for "
		    "compatibility-mode port.\n"));
	  goto abort_with_dma_region;
	}
#endif

      seg->len = alloc_len;

      ps->copy_block.segment_cnt++;
      ps->copy_block.len += alloc_len;
      gm_assert (GM_POWER_OF_TWO (ps->copy_block.len));
    }
  GM_RETURN_STATUS (GM_SUCCESS);

  /* Here we abort only the failed segment allocation, although we may have
     allocated other segments in this call. */

#if GM_BC_DRIVER
abort_with_dma_region:
  gm_arch_dma_region_free (&seg->dma_region);
#endif
abort_with_nothing:
  GM_RETURN_STATUS (status);
}
#endif

#if !GM_CAN_REGISTER_MEMORY
/* Free all the copy block for a device. Called when the device is closed. */

void
gm_free_copy_block (gm_port_state_t * ps)
{
  unsigned cnt;

  GM_CALLED ();

  gm_assert (ps);

  /* BAD: Should sync with LANai and inform it that the mappings have
     changed. (this is currently done by the caller) */

  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("Freeing the copy block with %d segments.\n",
	     ps->copy_block.segment_cnt));
  cnt = ps->copy_block.segment_cnt;
  while (cnt--)
    {
      GM_PRINT (GM_DEBUG_HOST_DMA, ("Freeing a copy block segment.\n"));
      gm_arch_dma_region_free (&ps->copy_block.segment[cnt].dma_region);
    }
  ps->copy_block.segment_cnt = 0;

  ps->copy_block.len = 0;
  GM_RETURN_NOTHING ();
}
#endif

gm_status_t
gm_kernel_sleep (gm_port_state_t * ps)
{
  unsigned port_id;
  gm_status_t status;

  GM_CALLED_WITH_ARGS (("%p", ps));

  gm_assert (ps);

  if (!ps->opened)
    {
#ifdef __GNUC__
      GM_NOTE (("User tried API function "
		__GM_FUNCTION__ " " "on port before setting port number.\n"));
#else
      GM_NOTE (("User tried API function on port before setting port "
		"number.\n"));
#endif
      return GM_PERMISSION_DENIED;
    }

  port_id = ps->id;

  /* GM daemon MAY NOT use this ioctl */

  if (port_id == GM_DAEMON_PORT_ID)
    return (GM_PERMISSION_DENIED);

  /* Sleep until a message or signal is received. */

  GM_PRINT (GM_PRINT_LEVEL >= 5, ("Port %d sleeping.\n", port_id));
  gm_arch_mutex_exit (&ps->sync);
  status = gm_arch_signal_sleep (&ps->sleep_sync);
  gm_arch_mutex_enter (&ps->sync);
  switch (status) {
    case GM_SLEEP_INTERRUPTED:

      /* Got a signal */

      GM_PRINT (GM_PRINT_LEVEL >= 5, ("Port %d woke by signal.\n", port_id));
      GM_RETURN_STATUS (GM_INTERRUPTED);
    case GM_SLEEP_TIMED_OUT:

      /* Timeout */

      GM_PRINT (GM_PRINT_LEVEL >= 5, ("Port %d woke by timeout.\n", port_id));
      break;

      /* success */

    default:
      GM_PRINT (GM_PRINT_LEVEL >= 5, ("Port %d woke.\n", port_id));
      break;
    }
  GM_RETURN_STATUS (GM_SUCCESS);
}

/****************************************************************
 * Interrupt handing
 ****************************************************************/

#define GM_INTR_PRINT 0
#define GM_DEBUG_WAKE_INTR 0

/* Should we disable interrupts during interrupt processing?

   We should if gm_interrupting() is called in a top-half interrupt
   handler but gm_handle_claimed_interrupt() is called in a
   bottom-half (WIN2K), or if there is a race between
   is->claim_interrupt(is) causing the LANai to deassert the
   interrupt, and the OS assuming the signal is deasserted
   (Solaris). */

#define GM_DISABLE_INTERRUPTS_WHILE_HANDLING (GM_OS_SOLARIS || GM_OS_WIN2K)

/* Determine if IS generated an interrupt.

   Returns GM_ARCH_INTR_CLAIMED if the board is generating an interrupt.
   Otherwise, returns GM_ARCH_INTR_UNCLAIMED. */

gm_s32_t
gm_interrupting (gm_instance_state_t * is)
{
  union gm_interrupt_descriptor *interrupt;

  interrupt = is->interrupt;
  if (!interrupt || gm_ntoh_u8 (interrupt->last_byte.ready) == 0)
    {
      GM_PRINT (GM_INTR_PRINT,
                ("gm_intrrupting: unclaimed interrupt isr=0x%x eimr=0x%x\n",
		 is->get_ISR (is), is->get_EIMR (is)));
      return GM_ARCH_INTR_UNCLAIMED;
    }

  if (GM_DISABLE_INTERRUPTS_WHILE_HANDLING)
    {
      /* Stop the interrupt by masking it. */
      is->disable_interrupts (is);
      GM_STBAR ();
    }
  
  return GM_ARCH_INTR_CLAIMED;
}

#if GM_USE_DEFAULT_INTERRUPT_HANDLER
#if !defined GM_ARCH_INTR_CLAIMED || !defined GM_ARCH_INTR_UNCLAIMED
#error GM_ARCH_INTR_CLAIMED and GM_ARCH_INTR_UNCLAIMED must be defined
#endif

/* New-style interrupt handler.

   Like the old-style interrupt handler gm_intr(), but assumes the
   caller has already called gm_interrupting() to verify that IS is
   interrupting. */

void
gm_handle_claimed_interrupt (gm_instance_state_t * is)
{
  enum gm_interrupt_type type;

  gm_assert (is);
  gm_assert (is->lanai.running);
  gm_assert (is->lanai.special_regs);

  if (GM_OS_SOLARIS)
    {
      GM_STBAR ();
      is->disable_interrupts (is); /* Solaris bug workaround */
      GM_STBAR ();
    }

  type = (enum gm_interrupt_type) gm_ntoh_u32 (is->interrupt->type);

  GM_PRINT (GM_INTR_PRINT,
	    ("*** got an interrupt to handler %d  type = %d\n",
	     is->id, type));

  switch (type)
    {

    case GM_COMMAND_COMPLETE_INTERRUPT:
      is->command_done = 1;
      gm_arch_wake (&is->command_sync);
      break;

      /* Determine which port caused the interrupt, and wake it. */
    case GM_WAKE_INTERRUPT:
    case GM_YP_WAKE_INTERRUPT:
      {
	unsigned int port;
	gm_port_state_t *ps;

	port = gm_ntoh_u32 (is->interrupt->wake.port);
	GM_PRINT
	  (GM_INTR_PRINT | GM_DEBUG_WAKE_INTR,
	   ("Waking port %d if needed.\n", port));
	gm_assert (port < GM_NUM_PORTS);

	ps = __gm_port_state_for_id (is, port);
	if (!ps)
	  GM_PANIC (("Could not wake port %d, which is not open.\n", port));
	if (port == GM_MAPPER_PORT_ID)
	  {
	    GM_PRINT
	      (GM_INTR_PRINT | GM_DEBUG_WAKE_INTR,
	       ("Waking mapper port.\n"));
	  }
	if (type == GM_WAKE_INTERRUPT)
	  {
	    gm_assert (ps->wake.handler);
	    ps->wake.handler (ps->wake.context);
	  }
	else if (type == GM_YP_WAKE_INTERRUPT)
	  {
	    gm_arch_wake (&ps->yp_sync);
	  }
      }
      break;

    case GM_NODE_ID_TRANSLATION_INTERRUPT:
      {
	void *hashed_node_id;
	gm_u32_t node_id;
	gm_u32_t global_id;

	node_id = gm_ntoh_u32
	  (is->interrupt->node_id_translation.node_id);
	gm_assert (node_id != 0);
	global_id = gm_ntoh_u32 
	  (is->interrupt->node_id_translation.global_id);

	/* insert the entry if it is not already present */
	hashed_node_id = gm_ptr_hash_find (is->node_id_hash, 
					   (void *)(gm_size_t)global_id);
	if (hashed_node_id == 0) 
	  {
	    gm_ptr_hash_insert (is->node_id_hash, 
				(void *)(gm_size_t)global_id, 
				(void *)(gm_size_t)node_id);
	  }

	/* insert the reverse translation. */
	is->global_id_array[node_id] = global_id;
      }
      break;

      /* Print a debugging message from the LANai. */

    case GM_PRINT_INTERRUPT:
      {
	gm_lp_t print_string;
	char *s;

	print_string = gm_read_lanai_global_lp (is, interrupt.print.string);
	gm_assert (print_string);
	s = (char *) is->lanai.sram + print_string;
	/* NOTE: don't ever turn this off here.  Turn it off in the MCP. */
#if !GM_OS_VXWORKS
	GM_INFO (("@@@@@@@@@@@@@ lanai sez: %s\n", s));
#endif /* GM_OS_VXWORKS */
      }
      break;

      /* Print the details of a failed assertion in the LANai. */

    case GM_FAILED_ASSERTION_INTERRUPT:
      if (gm_read_lanai_global_lp (is, interrupt.failed_assertion.text))
	{
	  gm_lp_t gm_file;
	  gm_lp_t gm_text;
	  gm_u32_t gm_line;

	  gm_file
	    = gm_read_lanai_global_lp (is, interrupt.failed_assertion.file);
	  gm_line
	    = gm_read_lanai_global_u32 (is, interrupt.failed_assertion.line);
	  gm_text
	    = gm_read_lanai_global_lp (is, interrupt.failed_assertion.text);

	  GM_NOTE (("*********** Failed assertion in LANai MCP (%s:%d %s)\n",
		    (char *) is->lanai.sram + gm_file, gm_line,
		    ((char *) is->lanai.sram + gm_text)));
	}
      else
	{
	  GM_NOTE (("Failed assertion in LANai MCP (%s:%d)\n",
		    ((char *) is->lanai.sram
		     + gm_read_lanai_global_lp (is,
						interrupt.failed_assertion.
						file)),
		    gm_read_lanai_global_u32 (is,
					      interrupt.failed_assertion.
					      line)));
	}

      GM_NOTE
	(("current handler is %s()\n",
	  (char *) is->lanai.sram
	  + gm_read_lanai_global_lp (is, current_handler)));
      
      GM_PRINT
	(GM_TRACE_LANAI_DMA,
	 ("L7 DMA len=%d lar=0x%x ear=0x%x%08x\n",
	  gm_read_lanai_global_u32 (is, sync_dma.descriptor[0].format.pre_lX.len),
	  gm_read_lanai_global_lp (is, sync_dma.descriptor[0].format.pre_lX.lanai_addr),
	  (gm_read_lanai_global_u32
	   (is, sync_dma.descriptor[0].format.pre_lX.ebus_addr_high)),
	  (gm_read_lanai_global_u32
	   (is, sync_dma.descriptor[0].format.pre_lX.ebus_addr_low))));
      GM_PRINT
	(GM_TRACE_LANAI_DMA,
	 ("L4 DMA specials EAR=0x%x LAR=0x%x DMA_CTR=0x%x\n",
	  gm_read_lanai_special_reg_u32 (is, l4.gmEAR),
	  gm_read_lanai_special_reg_u32 (is, l4.LAR),
	  gm_read_lanai_special_reg_u32 (is, l4.DMA_CTR)));

      is->pausible = 0;
      /* gm_disable_lanai (is); */
      break;

    case GM_WRITE_INTERRUPT:

      /* Perform writes using "GM_PRINT (GM_PRINT_LEVEL >= 0, ("%s",
         buf))".  We don't use "GM_PRINT (GM_PRINT_LEVEL >= 0, (buf))"
         to prevent printf-like interpretation of the data in buf.  It
         is safe to null-terminate the buffer in LANai sram because
         the MCP blocks until this interrupt is handled. */

      /* HACK: Here we use __gm_pio_read_u8() and __gm_pio_write_u8() directly,
         since we cannot use gm_kton_u8() and gm_ntok_u8() in the interrupt
         handler, because we must do a partword PIO and cannot make the
         target 32 bits, and because it happens to be safe because
         the write() implementation in the LANai blocks until the interrupt is
         completed. */

      {
	gm_u32_t len;
	gm_lp_t from_lp;
	gm_u8_n_t save;
	gm_u8_n_t *from;

	from_lp = gm_ntoh_lp (is->interrupt->write.buffer);
	from = (gm_u8_n_t *) is->lanai.sram + from_lp;
	len = gm_ntoh_u32 (is->interrupt->write.length);
	save = __gm_pio_read_u8 (is, &from[len]);
	/* HACK (see above) */
	___gm_pio_write_u8 (&from[len], gm_hton_u8 (0));
/*
	GM_INFO (("LANAI[%d,0x%x,%p,%d]: GM_WRITE_INTERRUPT\n    %s\n",
                   is->id, from_lp, from, len, (char *) from));
*/
	GM_INFO (("LANAI[%d]: %s", is->id, (char *) from));
	___gm_pio_write_u8 (&from[len], save);	/* HACK (see above) */
	break;
      }

    case GM_BOGUS_SEND_INTERRUPT:
    case GM_BOGUS_RECV_INTERRUPT:
      {
	gm_page_hash_table_t *hash;
	gm_up_t host_ptr;
	gm_up_t host_page;
	gm_u32_t myport;
	gm_u32_t mybin;
	gm_status_t mystatus;
	gm_u32_t repetitions;
        gm_port_state_t *ps;
	gm_pte_t *my_e;

	host_ptr = gm_ntoh_up (is->interrupt->bogus.pointer);
	host_page = (gm_up_t)((unsigned long)host_ptr) & ~((unsigned long)(GM_PAGE_LEN-1));
	myport   = gm_ntoh_u32(is->interrupt->bogus.port);
	repetitions = gm_ntoh_u32 (is->interrupt->bogus.repetitions);

        /* if the port is closing or already closed, get out of there */
        ps = __gm_port_state_for_id (is, myport);
        if (!ps || (ps->closing == 1) || (ps->opened == 0)) 
          {
            break;
          }

        hash = &is->page_hash;
	mystatus = gm_reference_mapping (is, host_page, myport, &mybin, &my_e);
	if (repetitions != 0)
	  {
	    GM_INFO (("[Skipped %u bogus %s reports]\n", repetitions,
		      (type == GM_BOGUS_SEND_INTERRUPT) ? "SEND" : "RECV"));
	  }
	GM_INFO(("LANAI[%d]: port=%d bogus page %s ptr = 0x%lx %s  (maxidx=%d) %s\n",
		is->id, myport, 
		(type==GM_BOGUS_SEND_INTERRUPT)?"SEND":"RECV",
		(unsigned long)host_ptr,
		(mystatus == GM_SUCCESS)?"SUCCESS":"FAILED",
      		hash->max_index,
		(mystatus == GM_SUCCESS)?"**BUG!**":" "));
      }
      break;

    case GM_NO_INTERRUPT:
      GM_NOTE (("spurious interrupt detected\n"));
      gm_disable_lanai (is);
      /* return, because disabling the lanai will cause is->interrupt
	 to become null */
      return;

    default:
      GM_WARN (("gm_intr: unit=%d got an unrecognized interrupt of type 0x%x\n",
		is->id, type));
      gm_disable_lanai (is);
      break;
    }

  if (is->msi_enabled && !GM_OS_SOLARIS && !GM_NEED_READBACK_TO_FLUSH_WRITES)
    {
      /* For MSIs, we're certain that the interrupt has been generated,
	 And we want to exit the isr as quickly as
	 possible, to avoid the MCP triggering another MSI before the
	 host's chipset knows that this interrupt has been handled.
	 So we simply skip the rest of the function and return here. */
      gm_write_lanai_global_u32 (is, interrupt.type, GM_NO_INTERRUPT);
      return;
    }
  

  /* Spin until we are absolutely certain that the interrupt has been
     generated, by spinning until a post-interrupt DMA has completed. */
  
  while (gm_ntoh_u8 (is->interrupt->last_byte.ready) != 2)
    {
      GM_STBAR ();
    }

  /* clear last_byte.ready before we allow the MCP to begin to
     interrupt again by DMA'ing last_byte.ready to the host */

  is->interrupt->last_byte.ready = gm_hton_u8 (0);
  GM_STBAR ();

  /* Claim the interrupt.  (Write ISR in the card to clear the bit
     that generated the interrupt.) */
  
  is->claim_interrupt (is);
  GM_STBAR ();

  /* Tell the LANai we are ready for the next interrupt. */
  
  GM_PRINT (GM_INTR_PRINT, ("gm_intr: done with handling \n"));
  gm_write_lanai_global_u32 (is, interrupt.type, GM_NO_INTERRUPT);
  
  GM_STBAR ();
  
  if (GM_DISABLE_INTERRUPTS_WHILE_HANDLING)
    {
      is->enable_interrupts (is); /* Solaris bug workaround */
      GM_STBAR ();
    }

  /* read back the ISR to force the write to land in the LANai */
  /* before returning from the interrupt routine */
  if (GM_NEED_READBACK_TO_FLUSH_WRITES)
    {
      is->get_ISR (is);
    }

  GM_PRINT (GM_INTR_PRINT, ("gm_intr: return from interrupt \n"));
}

/* Interrupt handler.

   This function is invoked when the LANai generates an interrupt.
   The function determines the cause of the interrupt (pause, resume,
   or port wake), and acts accordingly.  Wake events cause the
   corresponding sleeping port to be awakened.

   The MCP and the interrupt handler interract such that each
   interrupt corresponds to exactly one interrupt event.  This is done
   by having the LANai block until any earlier posted interrupt is
   acknowleged (by clearing the GM_HOST_SIG_BIT of ISR) before posting
   the next interrupt (by setting GM_HOST_SIG_BIT). The host clears
   GM_HOST_SIG_BIT after processing each interrupt. */

int gm_in_intr;

gm_s32_t gm_intr (gm_instance_state_t * is)
{
  gm_s32_t ret;

  gm_in_intr = 1;
  ret = gm_interrupting (is);
  if (ret == GM_ARCH_INTR_CLAIMED)
    {
      gm_handle_claimed_interrupt (is);
    }
  gm_in_intr = 0;
  return ret;
}
#endif

/*
 * gm_ioctl() perform most IOCTLs in a platform-independent
 * fashion.  System that do not know the input and output buffer
 * lengths (most unix drivers) should pass INT_MAX for IN_LEN and
 * OUT_LEN, effectively disabling the length checks, and should
 * implement gm_copyin() and gm_copyout() to return an
 * error on buffer overflow.
 */

#if GM_CPU_sparc64 && GM_OS_LINUX && 0
/* This is disabled because the new support for 32 bit apps using 64 bit
   drivers should eliminate the problem it is meant to resolve. */
struct gm_off_len_uvma32
{
  gm_u32_t offset;
  gm_u32_t len;
  gm_up_t uvma;
};
#endif

/* "if(1)" suppresses the "end-of-loop code not reached" Sun compiler
   warning*/
#define GM_BREAK_WITH_STATUS(s) do{status=(s);if(1)goto done;}while(0)

/* This GM_REQUIRE_OPEN_PORT macro should not be used for any ioctl
   that covers a kernel version of a GM API function, because the
   check should be inside the API function in this case, to ensure
   that the check is peformed when the kernel calls the function. */

#define GM_REQUIRE_OPEN_PORT() do {					\
  if (!ps->opened)							\
    {									\
      GM_NOTE (("User attempted port IOCTL on uninitialized port.\n"));	\
      GM_RETURN_STATUS (GM_PERMISSION_DENIED);				\
    }									\
} while (0)

#define GM_DEBUG_IOCTL 0
gm_status_t
gm_ioctl (gm_port_state_t * ps, unsigned int cmd,
	  const void *in_buff, unsigned int in_buff_len,
	  void *out_buff, unsigned int out_buff_len, unsigned int *copied_out)
{
  gm_status_t status;
  gm_port_t *gm_fake_port;	/* HACK */
  gm_u32_t output_bytes = 0;
  gm_u32_t register_by_struct = 0;
  gm_u32_t can_copy_status = 1;

  GM_CALLED_WITH_ARGS (("%p,%u,%p,%u,%p,%u,%p", ps, cmd, in_buff, in_buff_len,
			out_buff, out_buff_len, copied_out));

  GM_PRINT (GM_DEBUG_IOCTL, ("ioctl on %sprivilaged port\n",
			      ps->privileged ? "" : "un"));
  
  /* record the input and output buffers for gm_copyin() and gm_copyout() */

  ps->in_buff = in_buff;
  ps->out_buff = out_buff;

  /* Compute a pointer to a fake port such that the "kernel_port_state" field
     of the port is valid. */

  gm_fake_port
    = (gm_port_t *) ((char *) &ps
		     - GM_OFFSETOF (struct gm_port, kernel_port_state));

  GM_PRINT
    (GM_DEBUG_IOCTL,
     ("ioctl command is 0x%x (%s)\n", cmd, _gm_ioctl_cmd_name (cmd)));
  switch (cmd)
    {
    case GM_SET_REGISTER_MEMORY_LENGTH_AND_PVMA:
      GM_PRINT (GM_DEBUG_IOCTL, ("gm_ioctl: GM_SET_REGISTER_MEMORY_LENGTH\n"));
      GM_REQUIRE_OPEN_PORT ();
      GM_PRINT (GM_PRINT_LEVEL >= 4,
		("Setting memory (de)registration length.\n"));
      if (in_buff_len < sizeof (gm_u32_t))
	{
	  GM_PRINT (GM_PRINT_LEVEL >= 4,
		    ("gm_ioctl: GM_SET_REGISTER_MEMORY_LENGTH %d < %d \n",
		     in_buff_len, sizeof (gm_u32_t)));
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	}
      {
	struct gm_len_pvma s;
	
	status = gm_arch_copyin (in_buff, &s, sizeof (s));
	ps->mem_reg_len = s.len;
	ps->mem_reg_pvma = s.pvma;
      }
      if (status != GM_SUCCESS)
	{
	  GM_NOTE (("could not read memory registration length.\n"));
	  GM_BREAK_WITH_STATUS (status);
	}
      else
	{
	  GM_PRINT
	    (GM_PRINT_LEVEL >= 4,
	     ("gm_ioctl: GM_SET_REGISTER_MEMORY_LENGTH OK so far = %d\n",
	      ps->mem_reg_len));
	}
      GM_BREAK_WITH_STATUS (status);
      
      /* Memory registration/deregistration */
      
    case GM_REGISTER_MEMORY_BY_STRUCT:
    case GM_DEREGISTER_MEMORY_BY_STRUCT:
      register_by_struct = 1;
    case GM_REGISTER_MEMORY:
    case GM_DEREGISTER_MEMORY:
      if (GM_CAN_REGISTER_MEMORY)
	{
	  gm_instance_state_t *is;
	  gm_up_t pvma;		/* port virtual memory address */
          gm_up_t user_vma;
          gm_size_t total_len;
          unsigned int port_id;

#if !GM_CPU_powerpc64 && !GM_CPU_x86_64
          gm_assert (sizeof (void *) == sizeof (gm_up_t));
#endif /* !GM_CPU_powerpc64 && !GM_CPU_powerpc64 */
          is = ps->instance;
          gm_assert (is);
          port_id = ps->id;

	  GM_REQUIRE_OPEN_PORT ();
	  GM_PRINT (GM_DEBUG_MEM_REGISTER,
		    ("%s: in_buff = %p\n",
		     _gm_ioctl_cmd_name (cmd), in_buff));

	  /* determine ptr, len, and handle to register. */

	  if (register_by_struct)
	    {
	      struct gm_off_len_uvma_pvma s;
	      
	      if (in_buff_len < sizeof (s))
		{
		  GM_PRINT
		    (GM_DEBUG_IOCTL,
		     ("gm_ioctl: %s %d < %d \n",
		      _gm_ioctl_cmd_name (cmd), in_buff_len, sizeof (s)));
		  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
		}
	      status
		= gm_arch_copyin (in_buff, &s, sizeof (s));
	      if (status != GM_SUCCESS)
		{
		  GM_NOTE
		    (("could not read memory (de)registration struct.\n"));
		  GM_BREAK_WITH_STATUS (status);
		}
	      else
		{
		  GM_PRINT
		    (GM_PRINT_LEVEL >= 4,
		     ("Copied register_memory_struct information\n"));
		}
	      user_vma = s.uvma;
	      total_len = s.len;
	      pvma = s.pvma;
	      can_copy_status = 1;
	    }
          else                  /* !register_by_struct */
            {
              user_vma = (gm_up_t) in_buff;
              total_len = ps->mem_reg_len;
	      pvma = ps->mem_reg_pvma;
	      
              can_copy_status = 0;
              
              if (total_len > in_buff_len)
                {
                  GM_NOTE
		    (("User memory (de)registration buffer too small.\n"));
		  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
                }
            }

	  /* Register the memory. */

	  if (cmd == GM_REGISTER_MEMORY_BY_STRUCT
	      || cmd == GM_REGISTER_MEMORY)
	    {
	      GM_PRINT (GM_DEBUG_MEM_REGISTER,
			("registering 0x%x bytes at user addr 0x%lx->0x%lx\n",
			 total_len, (unsigned long) user_vma,
			 (unsigned long) pvma));
	      status = gm_register_memory_ex (gm_fake_port,
					      (void *) (gm_size_t) user_vma,
					      total_len,
					      (void *) (gm_size_t) pvma);
	    }
	  else
	    {
	      GM_PRINT (GM_DEBUG_MEM_REGISTER,
			("deregistering 0x%x bytes at user addr 0x%lx\n",
			 total_len, (unsigned long) pvma));
	      status = gm_deregister_memory (gm_fake_port,
					     (void *) (gm_size_t) pvma,
					     total_len);
	    }
	  GM_BREAK_WITH_STATUS (status);
	}
      else
	{
	  GM_BREAK_WITH_STATUS (GM_INVALID_COMMAND);
	}

      /************
       * Set the mapper state summary */

    case GM_SET_MAPPER_STATE:
      if (in_buff_len < sizeof (struct gm_mapper_state))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      status = _gm_set_mapper_state (gm_fake_port, in_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = 0;
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Get the mapper state summary */

    case GM_GET_MAPPER_STATE:
	if (out_buff_len < sizeof (struct gm_mapper_state))
	  GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
	status = _gm_get_mapper_state (gm_fake_port, out_buff);
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	output_bytes = sizeof (struct gm_mapper_state);
	GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Get the page_hash_cache_size  */

    case GM_GET_PAGE_HASH_CACHE_SIZE:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_PAGE_HASH_CACHE_SIZE\n"));
      if (out_buff_len < sizeof (unsigned int))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      status = _gm_get_page_hash_cache_size(gm_fake_port, out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = sizeof (unsigned int);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);


      /************
       * Get the max node id for this device. */

    case GM_GET_MAX_NODE_ID:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_MAX_NODE_ID\n"));
      if (out_buff_len < sizeof (unsigned int))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      status = gm_max_node_id (gm_fake_port, out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = sizeof (unsigned int);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Get the max node id in use for this device. */

    case GM_GET_MAX_NODE_ID_IN_USE:
      GM_PRINT (GM_PRINT_LEVEL >= 3,
		("gm_ioctl:: GM_GET_MAX_NODE_ID_INUSE\n"));
      if (out_buff_len < sizeof (unsigned int))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      status = gm_max_node_id_in_use (gm_fake_port, out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = sizeof (unsigned int);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Set the process_id of the opener of this port */

    case GM_SET_OPENER_PID:
      {
	int pid;
	GM_PRINT (GM_PRINT_LEVEL >= 5, ("gm_ioctl:: GM_SET_OPENER_PID\n"));
	if (in_buff_len < sizeof (gm_u32_t))
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	status = gm_copyin (ps, in_buff, &pid, sizeof (gm_u32_t));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = _gm_set_opener_pid (gm_fake_port, pid);
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	GM_BREAK_WITH_STATUS (GM_SUCCESS);
      }

      /************
       * Get the process_ids of the openers of each port */

    case GM_GET_OPENER_PIDS:
      GM_PRINT (GM_PRINT_LEVEL >= 5, ("gm_ioctl:: GM_GET_OPENER_PIDS\n"));
      if (out_buff_len < (GM_NUM_PORTS * sizeof (gm_u32_t)))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      status = _gm_get_opener_pids (gm_fake_port, out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = (GM_NUM_PORTS * sizeof (gm_u32_t));
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Get the host id for this device. */

    case GM_GET_NODE_ID:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_NODE_ID\n"));
      if (out_buff_len < sizeof (unsigned int))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      status = gm_get_node_id (gm_fake_port, (unsigned int *) out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = sizeof (unsigned int);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Get the host name. */

    case GM_GET_HOST_NAME:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_HOST_NAME\n"));
      if (in_buff_len < GM_MAX_HOST_NAME_LEN)
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      status = gm_get_host_name (gm_fake_port, out_buff);
      output_bytes = GM_MAX_HOST_NAME_LEN;
      GM_BREAK_WITH_STATUS (status);

      /************
       * Get a string from the LANai. */

    case GM_GET_FIRMWARE_STRING:
      if (!ps->privileged)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      if (in_buff_len < GM_MAX_FIRMWARE_STRING_LEN)
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      {
	gm_lp_t lanai_ptr;
	gm_size_t ob;
	
	status = gm_copyin (ps, in_buff, &lanai_ptr, sizeof (lanai_ptr));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = _gm_get_firmware_string (gm_fake_port, lanai_ptr, out_buff, &ob);
	output_bytes = ob + 1;
	GM_BREAK_WITH_STATUS (status);
      }

      /************
       * convert a node id to a unique id */

    case GM_NODE_ID_TO_UNIQUE_ID:
      /* Might be called once per node, so has higher print level */
      GM_PRINT (GM_PRINT_LEVEL >= 11,
		("gm_ioctl:: GM_NODE_ID_TO_UNIQUE_ID\n"));
      if (out_buff_len < 6)
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      if (in_buff_len < sizeof (unsigned int))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      {
	unsigned int node_id;
	status = gm_copyin (ps, in_buff, &node_id, sizeof (node_id));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	if (node_id > ps->instance->max_node_id)
	  GM_BREAK_WITH_STATUS (GM_INVALID_PARAMETER);
	status = gm_node_id_to_unique_id (gm_fake_port, node_id,
					  (char *) out_buff);

	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
      }
      output_bytes = 6;
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * convert a unique id to a node id */

    case GM_UNIQUE_ID_TO_NODE_ID:
      /* Might be called once per node, so has higher print level */
      GM_PRINT (GM_PRINT_LEVEL >= 11,
		("gm_ioctl:: GM_UNIQUE_ID_TO_NODE_ID\n"));
      if (out_buff_len < sizeof (unsigned int))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      if (in_buff_len < 6)
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      status = gm_unique_id_to_node_id (gm_fake_port,
					(char *) in_buff,
					(unsigned int *) out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = 6;
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Return the page size, as computed by the driver. */

    case GM_GET_PAGE_LEN:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_PAGE_LEN\n"));
      if (out_buff_len < sizeof (unsigned int))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      if (!GM_PAGE_LEN)
	GM_BREAK_WITH_STATUS (GM_INTERNAL_ERROR);
      status = gm_copyout (ps, &GM_PAGE_LEN, out_buff, sizeof (GM_PAGE_LEN));
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = sizeof (gm_u32_t);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Return the gm_mapping_specs structure for the port.  This
       * structure specifies the locations of the various
       * memory-mapped regions that the user may map. */

    case GM_GET_MAPPING_SPECS:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_MAPPING_SPECS\n"));
      if (out_buff_len < sizeof (gm_mapping_specs_t))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      status = _gm_get_mapping_specs (gm_fake_port,
				      (gm_mapping_specs_t *) out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = sizeof (gm_mapping_specs_t);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Get board "ethernet" ID. */

    case GM_GET_UNIQUE_BOARD_ID:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_UNIQUE_BOARD_ID\n"));
      if (in_buff_len < 6)
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      status = gm_unique_id (gm_fake_port, (char *) out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = 6;
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Get "ethernet" ID of the mapper node. */

    case GM_GET_MAPPER_UNIQUE_ID:
      GM_PRINT (GM_PRINT_LEVEL >= 3,
		("gm_ioctl:: GM_GET_MAPPER_UNIQUE_ID\n"));
      if (in_buff_len < 6)
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      status = gm_get_mapper_unique_id (gm_fake_port, (char *) out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = 6;
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Set the mask of sizes that are not acceptable to the receiving port */

    case GM_SET_ACCEPTABLE_SIZES_LOW:
      GM_PRINT (GM_PRINT_LEVEL >= 3,
		("gm_ioctl:: GM_SET_ACCEPTABLE_SIZES_LOW\n"));
      if (in_buff_len < sizeof (gm_u32_t))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      if (ps->id == GM_DAEMON_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      {
	gm_u32_t sizes;
	status = gm_copyin (ps, in_buff, &sizes, sizeof (sizes));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status =
	  gm_set_acceptable_sizes (gm_fake_port, GM_LOW_PRIORITY, sizes);
	GM_BREAK_WITH_STATUS (status);
      }

    case GM_SET_ACCEPTABLE_SIZES_HIGH:
      GM_PRINT (GM_PRINT_LEVEL >= 3,
		("gm_ioctl:: GM_SET_ACCEPTABLE_SIZES_HIGH\n"));
      if (in_buff_len < sizeof (gm_u32_t))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      if (ps->id == GM_DAEMON_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      {
	gm_u32_t sizes;
	status = gm_copyin (ps, in_buff, &sizes, sizeof (sizes));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status =
	  gm_set_acceptable_sizes (gm_fake_port, GM_HIGH_PRIORITY, sizes);
	GM_BREAK_WITH_STATUS (status);
      }

      /************
       * Convert a host name to a node ID */

    case GM_HOST_NAME_TO_NODE_ID:
      GM_PRINT (GM_PRINT_LEVEL >= 3,
		("gm_ioctl:: GM_HOST_NAME_TO_NODE_ID\n"));
      {
	struct
	{
	  unsigned int timeout_usecs;
	  char hostname[GM_MAX_HOST_NAME_LEN+1];
	} in;
	unsigned int node_id;
	
	if (in_buff_len < sizeof (in))
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	if (out_buff_len < sizeof (node_id))
	  GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);

	status = gm_copyin (ps, in_buff, &in, sizeof (in));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status
	  = gm_host_name_to_node_id_ex (gm_fake_port, in.timeout_usecs,
					in.hostname, &node_id);
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = gm_copyout (ps, &node_id, out_buff, sizeof (node_id));
      }
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = sizeof (unsigned int);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Convert a node ID to a host name */

    case GM_NODE_ID_TO_HOST_NAME:
      GM_PRINT (GM_PRINT_LEVEL >= 11,
		("gm_ioctl:: GM_NODE_ID_TO_HOST_NAME\n"));
      {
	struct 
	{
	  unsigned int timeout_usecs;
	  unsigned int node_id;
	} in;
	char host_name[GM_MAX_HOST_NAME_LEN+1];

	if (in_buff_len < sizeof (in))
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	if (out_buff_len < sizeof (host_name))
	  GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      
	status = gm_copyin (ps, in_buff, &in, sizeof (in));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = gm_node_id_to_host_name_ex (gm_fake_port, in.timeout_usecs,
					     in.node_id, &host_name);
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = gm_copyout (ps, host_name, out_buff, sizeof (host_name));
      }
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = GM_MAX_HOST_NAME_LEN+1;
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /*******************************
      ********************************
      **** Trivial ioctls
      ********************************
      ********************************/

      /************
       * Copy the board's ROM config strings to a user buffer */

    case GM_GET_BOARD_CONFIG_STRINGS:
      GM_PRINT
	(GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_BOARD_CONFIG_STRINGS\n"));
      if (out_buff_len < 256)
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      status = _gm_get_board_config_strings (gm_fake_port, out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = 256;
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Copy the LANai eeprom to a user buffer */

    case GM_GET_EEPROM:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_EEPROM\n"));
      if (out_buff_len < sizeof (gm_myrinet_eeprom_t))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      status = _gm_get_eeprom (gm_fake_port, out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = sizeof (gm_myrinet_eeprom_t);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Copy the LANai globals to a user buffer */

    case GM_GET_GLOBALS:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_GLOBALS\n"));
      if (out_buff_len < sizeof (struct gm_lanai_globals))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      status = _gm_get_globals (gm_fake_port, out_buff,
				(sizeof (struct gm_lanai_globals)
				 + (gm_read_lanai_global_u32
				    (ps->instance,
				     connection.first_free_id)
				    * sizeof (struct gm_connection))));
      /* (ps->instance->max_node_id+1) was 1024 */

      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = out_buff_len;
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

    case GM_GET_GLOBALS_BY_REQUEST:
      GM_PRINT (GM_PRINT_LEVEL >= 4,
		("gm_ioctl:: GM_GET_GLOBALS_BY_REQUEST\n"));
      {
	gm_u32_t space[2];
	gm_globals_request_t *request = (gm_globals_request_t *) & space[0];

	if (in_buff_len < 2 * sizeof (gm_u32_t))
	  {
	    GM_PRINT (GM_PRINT_LEVEL >= 0,
		      ("gm_ioctl: GM_GET_GLOBALS_BY_REQUEST %d < %d \n",
		       in_buff_len, 2 * sizeof (gm_u32_t)));
	    GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	  }
	status = gm_arch_copyin (in_buff, request, 2 * sizeof (gm_u32_t));
	if (status != GM_SUCCESS)
	  {
	    GM_NOTE (("could not read globals request struct.\n"));
	    break;
	  }
	else
	  {
	    GM_PRINT (GM_PRINT_LEVEL >= 4,
		      ("Copied globals request struct. off=%d len=%d\n",
		       request->offset, request->len));
	  }

	status = _gm_get_globals_by_request (gm_fake_port, out_buff,
					     request->offset, request->len);
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);

	output_bytes = request->len;
      }
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

    case GM_GET_GLOBALS_OFFSET:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_GLOBALS_OFFSET\n"));
      if (out_buff_len < sizeof (gm_u32_t))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      /* note the 1024 is a HACK */
      {
	gm_u32_t offset;

	offset = (gm_u32_t) ((char *) ps->instance->lanai.globals
			     - (char *) ps->instance->lanai.sram);
	status = gm_arch_copyout (&offset, out_buff, sizeof (offset));
      }
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = sizeof (gm_u32_t);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /*******************************
      ********************************
      **** Mapper-specific IOCTLS ****
      ********************************
      ********************************/

      /************
       * Allow the caller to perform raw receives */

    case GM_ENABLE_RAW_RECEIVES:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_ENABLE_RAW_RECEIVES\n"));
      if (ps->id != GM_MAPPER_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      if (_gm_enable_raw_receives (gm_fake_port) != GM_SUCCESS)
	status = GM_PERMISSION_DENIED;
      if (ps->id == GM_DAEMON_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

      /************
       * Set a route. */

    case GM_SET_ROUTE:
      /* Might be called once per node, so has higher print level */
      GM_PRINT (GM_PRINT_LEVEL >= 11, ("gm_ioctl:: GM_SET_ROUTE\n"));
      if (ps->id != GM_MAPPER_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      if (in_buff_len < sizeof (gm_set_route_info_t))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      if (ps->id == GM_DAEMON_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      {
	gm_set_route_info_t ri;
	status = gm_copyin (ps, in_buff, &ri, sizeof (gm_set_route_info_t));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = _gm_set_route (gm_fake_port,
				ri.mac_addr, ri.length,
				(char *) &(ri.route[0]));
	GM_BREAK_WITH_STATUS (status);
      }

	/***********
	 * Clear all routes. */
    case GM_CLEAR_ALL_ROUTES:
      GM_PRINT (GM_PRINT_LEVEL >= 7, ("gm_ioctl:: GM_CLEAR_ALL_ROUTES\n"));
      if (ps->id != GM_MAPPER_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      if (ps->id == GM_DAEMON_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      {
	status = _gm_clear_all_routes (gm_fake_port);
	GM_BREAK_WITH_STATUS (status);
      }

      /************
       * Get a route. */

    case GM_GET_ROUTE:
      GM_PRINT (GM_PRINT_LEVEL >= 11, ("gm_ioctl:: GM_GET_ROUTE\n"));
      if (in_buff_len < sizeof (gm_route_info_t))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      if (out_buff_len < sizeof (gm_route_info_t))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      {
	gm_route_info_t ri;
	unsigned int route_length;

	status = gm_copyin (ps, in_buff, &ri, sizeof (gm_route_info_t));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	route_length = ri.length;
	status = _gm_get_route (gm_fake_port,
				ri.target_node_id, (char *) ri.route,
				&route_length);
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	ri.length = (gm_u8_t) route_length;
	status = gm_copyout (ps, &ri, out_buff, sizeof (gm_route_info_t));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	output_bytes = sizeof (gm_route_info_t);
	GM_BREAK_WITH_STATUS (GM_SUCCESS);
      }

      /************
       * Set the node ID.  For mapper only! */

    case GM_SET_NODE_ID:
      GM_PRINT (GM_DEBUG_IOCTL, ("case GM_SET_NODE_ID\n"));
      gm_assert (ps);
      if (!ps->privileged)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      if (in_buff_len < sizeof (unsigned int))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      if (ps->id == GM_DAEMON_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      {
	unsigned int node_id;
	status = gm_copyin (ps, in_buff, &node_id, sizeof (unsigned int));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = _gm_set_node_id (gm_fake_port, node_id);
	GM_BREAK_WITH_STATUS (status);
      }

      /************
       * Set enable_nack_down flag for this port */

    case GM_SET_ENABLE_NACK_DOWN:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_SET_ENABLE_NACK_DOWN\n"));
      if (in_buff_len < sizeof (gm_u32_t))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      {
	gm_u32_t flag;
	status = gm_copyin (ps, in_buff, &flag, sizeof (gm_u32_t));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = gm_set_enable_nack_down(gm_fake_port, flag);
	GM_BREAK_WITH_STATUS (status);
      }

      /*cop support */
    case GM_COP_WAKEUP:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_COP_WAKEUP\n"));
      if (GM_ENABLE_SECURITY
	  && (!ps->privileged || ps->id != GM_MAPPER_PORT_ID))
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      status = _gm_cop_wakeup (gm_fake_port);
      GM_BREAK_WITH_STATUS (status);

    case GM_COP_END:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_COP_END\n"));
      if (GM_ENABLE_SECURITY
	  && (!ps->privileged || ps->id != GM_MAPPER_PORT_ID))
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      status = _gm_cop_end (gm_fake_port);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

    case GM_COP_SEND:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_COP_SEND\n"));
      if (GM_ENABLE_SECURITY
	  && (!ps->privileged || ps->id != GM_MAPPER_PORT_ID))
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      if (in_buff_len < sizeof (gm_s32_t))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      {
	gm_s32_t d;
	status = gm_copyin (ps, in_buff, &d, sizeof (gm_s32_t));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = _gm_cop_send (gm_fake_port, d);
	GM_BREAK_WITH_STATUS (status);
      }

    case GM_COP_RECEIVE:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_COP_SEND\n"));
      if (GM_ENABLE_SECURITY
	  && (!ps->privileged || ps->id != GM_MAPPER_PORT_ID))
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      if (out_buff_len < sizeof (unsigned int))
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      status = _gm_cop_receive (gm_fake_port, (int *) out_buff);
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = sizeof (int);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

#if 0
      /************
       * Set the number of physical pages.  For mapper only! */

    case GM_SET_PHYSICAL_PAGES:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_SET_PHYSICAL_PAGES\n"));
      if (ps->id != GM_MAPPER_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      if (in_buff_len < sizeof (gm_u32_t))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      if (ps->id == GM_DAEMON_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      status = gm_copyin (ps, in_buff, &gm_physical_pages,
			  sizeof (gm_physical_pages));
      GM_BREAK_WITH_STATUS (status);
#endif

      /************
       * Set the host name.  For mapper only! */

    case GM_SET_HOST_NAME:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_SET_HOST_NAME\n"));
      if (ps->id != GM_MAPPER_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      if (ps->id == GM_DAEMON_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      {
	struct
	{
	  char host_name[GM_MAX_HOST_NAME_LEN+1];
	}
	in;

	GM_PRINT (0, ("*** input host name between %p and %p.\n",
		      &in, &in + 1));

	if (in_buff_len < sizeof (in))
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	status = gm_copyin (ps, in_buff, &in, sizeof (in));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = _gm_set_host_name (gm_fake_port, &in.host_name);
	GM_BREAK_WITH_STATUS (status);
      }

    case GM_SET_UNIQUE_ID:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_SET_UNIQUE_ID\n"));
      if (ps->id == GM_DAEMON_PORT_ID)
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
      {
	struct
	{
	  unsigned int node_id;
	  char unique[6];
	}
	in;

	if (in_buff_len < sizeof (in))
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	status = gm_copyin (ps, in_buff, &in, sizeof (in));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = _gm_set_unique_id (gm_fake_port,
				    in.node_id, (char *) &in.unique[0]);
	GM_BREAK_WITH_STATUS (status);
      }

    case GM_SLEEP:
      GM_PRINT (GM_PRINT_LEVEL >= 5, ("gm_ioctl:: GM_SLEEP\n"));
      status = gm_kernel_sleep (ps);
      GM_BREAK_WITH_STATUS (status);

    case GM_SET_PORT_NUM:
      /* NOTE: This ioctl must not cause any GM sleep, because the
               platform-specific code is permitted to hold a global lock
               while processing it.
      */
      if (ps->opened)
	{
	  GM_NOTE (("User attempted to re-set port number.\n"));
	  GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
	}
      GM_PRINT (GM_DEBUG_PORT_STATE, ("handling GM_SET_PORT_NUM.\n"));
      {
	gm_u32_t port_num;

	status = gm_copyin (ps, in_buff, &port_num, sizeof (port_num));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	GM_PRINT (GM_PRINT_LEVEL >= 5,
		  ("Calling gm_port_state_open (%p, %d)\n", ps, port_num));
	status = gm_port_state_open (ps, port_num);
      }

      if (status != GM_SUCCESS)
	{
	  gm_assert (!ps->opened);
	  GM_PRINT(GM_DEBUG_PORT_STATE, ("Could not open port state.\n"));
	  break;
	}
      gm_assert (ps->opened);
      GM_PRINT (GM_DEBUG_PORT_STATE, ("opened the port state.\n"));
      break;

    case GM_GET_KERNEL_BUILD_ID_LEN:
      GM_PRINT (GM_PRINT_LEVEL >= 3,
		("gm_ioctl:: GM_GET_KERNEL_BUILD_ID_LEN\n"));
      {
	int kernel_build_id_len;

	kernel_build_id_len = (int) (gm_strlen (_gm_build_id) + 1);
	status =
	  gm_copyout (ps, &kernel_build_id_len, out_buff, sizeof (int));
      }
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = sizeof (int);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

    case GM_GET_KERNEL_BUILD_ID:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_GET_KERNEL_BUILD_ID\n"));
      if (out_buff_len < gm_strlen (_gm_build_id) + 1)
	GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
      status = gm_copyout (ps, (void *) _gm_build_id, out_buff,
			   (unsigned int) (gm_strlen (_gm_build_id) + 1));
      if (status != GM_SUCCESS)
	GM_BREAK_WITH_STATUS (status);
      output_bytes = (gm_u32_t) (gm_strlen (_gm_build_id) + 1);
      GM_BREAK_WITH_STATUS (GM_SUCCESS);

    case GM_FINISH_MMAP:
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl: GM_FINISH_MMAP\n"));
      {
	struct gm_off_len_uvma s;
#if GM_CPU_sparc64 && GM_OS_LINUX && 0
	struct gm_off_len_uvma32 s_32;
	status = gm_copyin (ps, in_buff, &s_32, sizeof (s_32));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	s.offset = (unsigned long) s_32.offset;
	s.len = (unsigned long) s_32.len;
	s.uvma = s_32.uvma;
#else
	status = gm_copyin (ps, in_buff, &s, sizeof (s));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
#endif
	status = _gm_finish_mmap (gm_fake_port, s.offset, s.len, s.uvma);
	GM_BREAK_WITH_STATUS (status);
      }


#if GM_ENABLE_TRACE
      /************
       * Get the kernel traces */

    case GM_GET_KTRACE:
      GM_PRINT (GM_PRINT_LEVEL >= 11, ("gm_ioctl:: GM_GET_KTRACE\n"));
      if (out_buff_len < sizeof (gm_file_trace_t) * GM_HOST_NUMTRACE)
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      {
	status = gm_get_ktrace (gm_fake_port, out_buff);
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	output_bytes = sizeof (gm_file_trace_t) * GM_HOST_NUMTRACE;
	GM_BREAK_WITH_STATUS (GM_SUCCESS);
      }
#endif

    case GM_WRITE_LANAI_REGISTER:
      if (GM_ENABLE_SECURITY
	  && (!ps->privileged || ps->id != GM_MAPPER_PORT_ID))
	GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);

      if (in_buff_len < sizeof (gm_lanai_register_access_t))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      {
	gm_lanai_register_access_t a;
	status = gm_copyin (ps, in_buff, &a, sizeof (gm_lanai_register_access_t));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	/*
	if ((((char*)gm_fake_port->kernel_port_state->instance->lanai.special_regs) + a.offset) !=
	    (char*)&((gm_lanai_special_registers_t*)
	    gm_fake_port->kernel_port_state->instance->lanai.special_regs)->l9.MYRINET)
	  gm_linux_vprintk ("problem\n");
	else
	  gm_linux_vprintk ("not a problem\n");
	*/
	/*if (gm_fake_port->kernel_port_state->instance->lanai.eeprom.lanai_cpu_version == 0x900)
	  gm_write_lanai_special_reg_u32 (gm_fake_port->kernel_port_state->instance, l9.MYRINET, a.value);
	*/
	__gm_kton_u32 (gm_fake_port->kernel_port_state->instance,
	(gm_u32_n_t*)(((char*)gm_fake_port->kernel_port_state->instance->lanai.special_regs) + a.offset), a.value);
	
	GM_BREAK_WITH_STATUS (GM_SUCCESS);
      }


      
#if GM_ENABLE_DIRECTCOPY
      /************
       * Provide a memory copy between two separated user-spaces */

    case GM_DIRECTCOPY_GET:
      GM_PRINT (GM_PRINT_LEVEL >= 5, ("gm_ioctl:: GM_DIRECTCOPY_GET\n"));
#if GM_OS_LINUX
      {
	gm_directcopy_t args;
	gm_port_state_t *source_ps;
	gm_u16_t pid_source;

	if (in_buff_len < sizeof (gm_directcopy_t))
	  {
	    GM_PRINT (GM_PRINT_LEVEL >= 4,
		      ("gm_ioctl: GM_DIRECTCOPY_GET %d < %d \n", in_buff_len,
		       sizeof (gm_directcopy_t)));
	    GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	  }
	status =
	  gm_arch_copyin (in_buff, &args, sizeof (gm_directcopy_t));
	if (status != GM_SUCCESS)
	  {
	    GM_NOTE (("could not read GET directcopy parameters struct.\n"));
	    break;
	  }
	else
	  {
	    GM_PRINT (GM_PRINT_LEVEL >= 4,
		      ("Copied GET directcopy parameters information\n"));
	  }

	/* PS is the port state of the port doing the copy. */
	if (ps->id == args.source_port_id
	    && ps->instance->id == args.source_instance_id)
	  {
	    /* User requested copy from its own memory space.  
	       Abort to prevent attempting to acquire the lock for
	       this port state, which we already hold. */
	    goto abort_with_nothing;
	  }
	source_ps = gm_acquire_port_state_for_ids (args.source_instance_id,
						   args.source_port_id);
	if (!source_ps)
	  goto abort_with_nothing;
	if (!source_ps->opened)
	  goto abort_with_lock;
	/* we need the pid of the other process to acces its virtual
	   memory space via the kernel */
	pid_source = source_ps->opener_pid;

	/* we ask the kernel to copy the data by a GET (copy by the
	   receiver process) */
	status = gm_arch_directcopy_get ((void *) args.source_addr,
					 (void *) args.target_addr,
					 args.length, pid_source);
	/* we unlock the port state */
	gm_release_port_state (source_ps);
	break;

	/* something goes wrong ? */
      abort_with_lock:
	gm_release_port_state (source_ps);
      abort_with_nothing:
	status = GM_FAILURE;
      }
#else
      status = GM_FAILURE;
#endif
#endif

    case GM_REGISTER_RECV_QUEUE:
      GM_REQUIRE_OPEN_PORT ();
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_REGISTER_RECV_QUEUE\n"));
      if (in_buff_len < sizeof (gm_up_t))
	GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
      {
	gm_up_t up;
	status = gm_copyin (ps, in_buff, &up, sizeof (gm_up_t));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = gm_register_recv_queue (gm_fake_port, up);
	GM_BREAK_WITH_STATUS (status);
      }

    case GM_DMA_ADDR:
      {
	gm_up_t in;
	gm_dp_t out;
	
	GM_PRINT (GM_DEBUG_IOCTL, ("gm_ioctl(): GM_DMA_ADDR\n"));
	if (out_buff_len < sizeof (gm_dp_t))
	  GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
	if (in_buff_len < sizeof (gm_up_t))
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	status = gm_copyin (ps, in_buff, &in, sizeof (in));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = gm_dma_addr (gm_fake_port, in, &out);
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = gm_copyout (ps, &out, out_buff, sizeof (out));
	GM_BREAK_WITH_STATUS (status);
      }
      
      /* Allocate a single contiguous memory region for a port. */
      
    case GM_CONTIGUOUS_DMA_MALLOC:
      {
	gm_size_t len;
	gm_dp_t dma_ptr;
	
	GM_PRINT
	  (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_CONTIGUOUS_DMA_MALLOC\n"));

	/* The usual checks for open port and buffer lengths. */
	
	GM_REQUIRE_OPEN_PORT ();
	if (in_buff_len < sizeof (len))
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	if (out_buff_len < sizeof (dma_ptr))
	  GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);

	/* Only privileged users may request contiguous allocations. */
	
	if (GM_ENABLE_SECURITY && !ps->privileged)
	  GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);
	
	/* Copyin the length of the allocation */
	
	status = gm_copyin (ps, in_buff, &len, sizeof (len));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);

	/* Allocate the contiguous memory */

	status = gm_contiguous_dma_malloc (gm_fake_port, len, &dma_ptr);
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);

	/* Copy the result back to user space. */

	dma_ptr = gm_arch_dma_region_dma_addr (&ps->contiguous_dma_region);
	status = gm_copyout (ps, &dma_ptr, out_buff, sizeof (dma_ptr));
	if (status != GM_SUCCESS)
	  {
	    gm_arch_dma_region_free (&ps->contiguous_dma_region);
	    GM_BREAK_WITH_STATUS (status);
	  }

	/* Return successfully */
	
	ps->contiguous_dma_region_allocated_len = len;
	GM_BREAK_WITH_STATUS (GM_SUCCESS);
      }

      /* Free the previously allocated contiguous DMA region for the port,
         if any. */
      
    case GM_CONTIGUOUS_DMA_FREE:
      {
	GM_PRINT
	  (GM_PRINT_LEVEL >= 3, ("gm_ioctl:: GM_CONTIGUOUS_DMA_FREE\n"));
	status = gm_contiguous_dma_free (gm_fake_port);
	GM_BREAK_WITH_STATUS (status);
      }

    case GM_NODE_ID_TO_GLOBAL_ID:
      {
	gm_u32_t in, out;
	
	if (in_buff_len < sizeof (in))
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	if (out_buff_len < sizeof (out))
	  GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);
	status = gm_copyin (ps, in_buff, &in, sizeof (in));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = gm_node_id_to_global_id (gm_fake_port, in, &out);
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);
	status = gm_copyout (ps, &out, out_buff, sizeof (out));
	output_bytes = sizeof(out);
	GM_BREAK_WITH_STATUS (status);
      }

    case GM_ETHERNET_SET_CREDIT_INCR:
      {
	gm_u32_t in, max_credits;
	gm_instance_state_t *is;
	

	is = ps->instance;
	gm_assert(is);

	/* Only privileged users may change ethernet port settings. */
	
	if (GM_ENABLE_SECURITY && !ps->privileged)
	  GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);

	if (in_buff_len < sizeof (in))
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	status = gm_copyin (ps, in_buff, &in, sizeof (in));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);

	max_credits = gm_read_lanai_global_u8 (is, ethernet.wake.max_credits);
	
	/* An increment greater than the max value makes no sense */
	if (in > max_credits )
	  GM_BREAK_WITH_STATUS (GM_INVALID_PARAMETER);
	gm_pause_lanai (is);
	gm_write_lanai_global_u8 (is, ethernet.wake.credit_incr,
				  (gm_u8_t)in);
	gm_unpause_lanai (is);
	GM_BREAK_WITH_STATUS (GM_SUCCESS);
      }
      
    case GM_ETHERNET_SET_MAX_CREDITS:
      {
	gm_u32_t in, credit_incr;
	gm_instance_state_t *is;
	

	is = ps->instance;
	gm_assert(is);

	/* Only privileged users may change ethernet port settings. */
	
	if (GM_ENABLE_SECURITY && !ps->privileged)
	  GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);

	if (in_buff_len < sizeof (in))
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	status = gm_copyin (ps, in_buff, &in, sizeof (in));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);

	/* Ensure that the number of credits doesn't get too large */
	if (in > GM_MAX_ETHERNET_WAKE)
	  GM_BREAK_WITH_STATUS (GM_INVALID_PARAMETER);

	credit_incr = gm_read_lanai_global_u8 (is, ethernet.wake.credit_incr);
	if (credit_incr > in )
	  credit_incr = in;
	gm_pause_lanai (is);
	gm_write_lanai_global_u8 (is, ethernet.wake.max_credits,
				  (gm_u8_t)in);
	gm_write_lanai_global_u8 (is, ethernet.wake.credit_incr,
				  (gm_u8_t)credit_incr);
	gm_unpause_lanai (is);
	GM_BREAK_WITH_STATUS (GM_SUCCESS);
      }
      
    case GM_ETHERNET_SET_MAX_SDMAS:
      {
	gm_u32_t in;
	gm_instance_state_t *is;
	

	is = ps->instance;
	gm_assert(is);

	/* Only privileged users may change ethernet port settings. */
	
	if (GM_ENABLE_SECURITY && !ps->privileged)
	  GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);

	if (in_buff_len < sizeof (in))
	  GM_BREAK_WITH_STATUS (GM_INPUT_BUFFER_TOO_SMALL);
	status = gm_copyin (ps, in_buff, &in, sizeof (in));
	if (status != GM_SUCCESS)
	  GM_BREAK_WITH_STATUS (status);

	/* Ensure that the sender will not run out of tokens before 
	   a trasmit complete interrupt is generated */
	if (in > GM_NUM_SEND_TOKENS - 1)
	  GM_BREAK_WITH_STATUS (GM_INVALID_PARAMETER);
	gm_pause_lanai (is);
	gm_write_lanai_global_u16 (is, ethernet.wake.max_sdmas_until_wake,
				  (gm_u16_t)in);
	gm_write_lanai_global_u16 (is, ethernet.wake.sdmas_until_wake,
				  (gm_u16_t)in);
	gm_unpause_lanai (is);
	GM_BREAK_WITH_STATUS (GM_SUCCESS);
      }

    case GM_ETHERNET_GET_PARAMS:
      {
	gm_size_t ethernet_len;
	gm_instance_state_t *is;
	void *scratch;

	is = ps->instance;
	gm_assert(is);

	ethernet_len = 
	  sizeof (((gm_lanai_globals_t *) is->lanai.globals)->ethernet.wake);

	if (out_buff_len < ethernet_len)
	  GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);

	scratch = gm_malloc (ethernet_len);
	if (scratch == NULL)
	  GM_BREAK_WITH_STATUS (GM_OUT_OF_MEMORY);

	__gm_ntok_bcopy 
	  (is, 
	   (volatile gm_u8_n_t *)
	   &(((gm_lanai_globals_t *) is->lanai.globals)->ethernet.wake),
	   scratch, ethernet_len);

	output_bytes = ethernet_len;
	status = gm_copyout (ps, scratch,  out_buff, ethernet_len);
	gm_free (scratch);
	GM_BREAK_WITH_STATUS (status);
      } 
    case GM_DUMP_MCP:
      {
	gm_instance_state_t *is;
	char *sram_copy;

	is = ps->instance;
	gm_assert(is);

	/* Only privileged users may examine the firmware contents. */
	
	if (GM_ENABLE_SECURITY && !ps->privileged)
	  GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);


	if (out_buff_len < is->lanai.eeprom.lanai_sram_size)
	  GM_BREAK_WITH_STATUS (GM_OUTPUT_BUFFER_TOO_SMALL);

	output_bytes = is->lanai.eeprom.lanai_sram_size;
	sram_copy = gm_malloc (output_bytes);
	if (sram_copy == NULL)
	  GM_BREAK_WITH_STATUS (GM_OUT_OF_MEMORY);

	/* pause the lanai to get a consistant view */
	gm_pause_lanai (is);
	__gm_ntok_bcopy (is, 
			 (volatile gm_u8_n_t *)is->lanai.sram,
			 sram_copy, output_bytes);
	gm_unpause_lanai (is);
	status = gm_copyout (ps, sram_copy, out_buff, output_bytes);
	gm_free (sram_copy);
	GM_BREAK_WITH_STATUS (status);
      }

    /* Clear all the counters on the Lanai */
    case GM_CLEAR_COUNTERS:
      {
	/* Make sure user has privilege */
	if (!ps->privileged)
	  GM_BREAK_WITH_STATUS (GM_PERMISSION_DENIED);

	status = _gm_clear_counters(gm_fake_port);
	GM_BREAK_WITH_STATUS (status);
      }

    default:
      GM_PRINT (GM_PRINT_LEVEL >= 1,
		("gm_ioctl: *** unhandled ioctl 0x%x ***\n", cmd));
      GM_BREAK_WITH_STATUS (GM_INVALID_COMMAND);
    }
done:

  GM_PRINT (GM_DEBUG_IOCTL, ("Done with switch\n"));

  /* If there was an error, inform the user via the copy buffer, ensuring
     that the OS does not get a chance to mangle the error code. */
  if (can_copy_status && (status != GM_INVALID_COMMAND) 
      && (status != GM_SUCCESS) && out_buff && out_buff_len >= sizeof (status))
    {
      GM_PRINT (GM_DEBUG_IOCTL,
		("ioctl (0x%x) failed, copying status = %d (0x%x)\n", cmd,
		 status, status));

      /* restore ps->out_buff since gm_copyout uses it and it may
	 have been changed while/if we slept. */
      ps->out_buff = out_buff;

      output_bytes = sizeof (status);
      gm_copyout (ps, (void *) &status, out_buff, output_bytes);
    }
  else if (!can_copy_status)
    {
      GM_PRINT (GM_DEBUG_IOCTL,
		("gm_ioctl NOT copying status = %d\n", status));
    }

  GM_PRINT (GM_DEBUG_IOCTL, ("copied_out = %p\n", copied_out));

  if (copied_out)
    *copied_out = output_bytes;

#if 0
  /* This NT HACK is no longer used.  Instead, it is the caller's
     responsibility to localize GM error codes if needed. */
  if (status == GM_FAILURE)
    status = GM_OUT_OF_MEMORY;
#endif

  if (status != GM_SUCCESS)
    {
      GM_PRINT (GM_DEBUG_IOCTL,
		("ioctl %s returning with error status 0x%x (%s).\n",
		 _gm_ioctl_cmd_name (cmd), status, gm_strerror (status)));
    }

  GM_RETURN_STATUS (status);
}

#undef GM_REQUIRE_OPEN_PORT
#undef GM_BREAK_WITH_STATUS

/************************************************************************
 * Unix mmap functions.
 ************************************************************************/

static void
gm_print_port_state (struct gm_port_state *ps)
{
  GM_PARAMETER_MAY_BE_UNUSED (ps);
  
  GM_PRINT (1, ("id = %d\n", (int) ps->id));
  _GM_PRINT (1, ("instance = %p\n", ps->instance));
  _GM_PRINT (1, ("compatibility_mode = %d\n", (int) ps->compatibility_mode));
  _GM_PRINT (1, ("privileged = %d\n", (int) ps->privileged));
  _GM_PRINT (1, ("min_dma_mem = 0x%x\n", (int) ps->min_dma_mem));
  _GM_PRINT (1, ("max_dma_mem = 0x%x\n", (int) ps->max_dma_mem));
  _GM_PRINT (1, ("mapping_specs = {...}\n"));
  _GM_PRINT (1, ("sleep_sync = {...}\n"));
  _GM_PRINT (1, ("copy_block = {...}\n"));
  _GM_PRINT (1, ("recv_queue_region = {...}\n"));
  _GM_PRINT (1, ("arch = {...}\n"));
}

/* The Unix Mmap functions here are split into 3 phases for compatibility
   across Unix platforms:

   (1) Verify that the MMAP will succeed (gm_prepare_to_mmap()).
   (2) Map the individual pages (gm_mmap()).
   (3) Update data internal data structures depending on the
   successful mapping.b (gm_finish_mmap()).

   Phase 2 is platform-dependent, so we just provide a function to return
   a pointer to the kernel page to be mapped. */

#define GM_DEBUG_PREPARE_TO_MMAP 0

gm_status_t
gm_prepare_to_mmap (gm_port_state_t * ps, gm_offset_t off, gm_size_t len,
		    unsigned int permissions)
{
  gm_instance_state_t *is;

  GM_CALLED ();

  is = ps->instance;

#if 1
  if (!is)
    gm_print_port_state (ps);
#endif

  gm_assert (is);

  /* Verify alignments */

  if (!GM_PAGE_ALIGNED (len) || !GM_PAGE_ALIGNED (off))
    {
      GM_NOTE (("alignment problem in gm_prepare_to_mmap\n"));
      _GM_NOTE (("len = 0x%lx, off = 0x%lx\n", len, off));
      GM_RETURN_STATUS (GM_INVALID_PARAMETER);
    }


  /* Verify that the user has set the port number before trying to mmap. */

  if (!ps->opened)
    {
      if (ps->privileged)
	{
	  /* Allow the hardware to be mapped in "hardware developer mode",
	     with the following layout:

	     +------------------------------+ 0
	     | entire PCI card memory image |
	     +------------------------------+ is->board_span
	     | contiguous DMA region	    |
	     | ...                          | */

	  /* Make sure the mapping is legit. */

	  if (off > is->board_span
	      || (off + len) < off)
	    {
	      GM_RETURN_STATUS (GM_INVALID_PARAMETER);
	    }

	  /* If needed, allocate a contiguous DMAable memory region */

	  if (off + len > is->board_span)
	    {
	      gm_status_t status;
	      gm_size_t alloc_len;
	      
	      if (ps->contiguous_dma_region_allocated_len != 0)
		{
		  /* The user is only allowed one contiguous
		     DMA mapping per port. */
		  GM_RETURN_STATUS (GM_BUSY);
		}

	      alloc_len = off + len - is->board_span;
	      status
		= (gm_arch_dma_region_alloc
		   (is,
		    &ps->contiguous_dma_region,
		    alloc_len,
		    GM_ARCH_DMA_CONTIGUOUS
		    + GM_ARCH_DMA_STREAMING
		    + GM_ARCH_DMA_RDWR,
		    0,
		    0));
	      if (status != GM_SUCCESS)
		{
		  GM_RETURN_STATUS (status);
		}

	      /* Put the DMA address of the region in the first 8 bytes
		 of the DMA region. */
	       
	      *(gm_u64_t *)(gm_arch_dma_region_kernel_addr
			    (&ps->contiguous_dma_region))
		= ((gm_u64_t)
		   gm_arch_dma_region_dma_addr (&ps->contiguous_dma_region));
	      
	      ps->contiguous_dma_region_allocated_len = alloc_len;
	    }

	  /* Allow the HW dev mapping. */
	  
	  GM_NOTE (("Allowing board mapping for HW development.\n"));
	  GM_RETURN_STATUS (GM_SUCCESS);
	}
      
      GM_NOTE (("User failed to set port number before mmap().\n"));
      GM_RETURN_STATUS (GM_INVALID_PARAMETER);
    }

  GM_PRINT (GM_DEBUG_PREPARE_TO_MMAP,
	    ("gm_prepare_to_mmap called for instance 0x%x, port 0x%x.\n",
	     is->id, ps->id));
  _GM_PRINT (GM_DEBUG_PREPARE_TO_MMAP,
	     ("  offset=0x%lx len=0x%lx\n", off, len));

#if GM_BC_DRIVER
  if (ps->compatibility_mode)
    {
      status = gm_expand_copy_block (ps, ps->max_dma_mem);
      GM_RETURN_STATUS (status);
    }
#endif

  /* Allow individual pages to be mapped */

#define GM_ARCH_MAPS_PAGE_BY_PAGE 1

#if GM_ARCH_MAPS_PAGE_BY_PAGE
#define LE <=
#else
#define LE ==
#endif

#define LEGIT_MAPPING(type)						\
  (ps->mappings.type.offset LE off					\
   && (off+len) LE (ps->mappings.type.offset + ps->mappings.type.len) 	\
   && (ps->mappings.type.permissions & permissions) == permissions)

  if (LEGIT_MAPPING (control_regs)
      || LEGIT_MAPPING (special_regs)
      || LEGIT_MAPPING (sram)
      || LEGIT_MAPPING (send_queue) || LEGIT_MAPPING (RTC))
    {
      GM_PRINT (GM_DEBUG_PREPARE_TO_MMAP, ("mapping is legit\n"));
      GM_RETURN_STATUS (GM_SUCCESS);
    }

#undef LEGIT_MAPPING
#undef LE

#if !GM_CAN_REGISTER_MEMORY
  /* Allow mappings within the copy block, but only if the copy block
     can be expanded to satisfy the request.

     The Copy block is needed only on machines that cannot perform
     memory registration. */

  if (ps->mappings.copy_block.offset <= off
      && off + len <= (ps->mappings.copy_block.offset
		       + ps->mappings.copy_block.len))
    {
      /* If needed, expand the copy block to satisfy the request. */

      if (off + len > ps->mappings.copy_block.offset + ps->copy_block.len)
	{
	  gm_status_t status;
	  status
	    = gm_expand_copy_block (ps,
				    (gm_u32_t) (off + len -
						ps->mappings.copy_block.
						offset));
	  if (status != GM_SUCCESS)
	    {
	      GM_PRINT (GM_DEBUG_PREPARE_TO_MMAP,
			("copy_block %d %d %d", off, len,
			 ps->mappings.copy_block.offset));
	      GM_RETURN_STATUS (status);
	    }
	}

      /* Verify that the mapping falls completely within a copy block
         segment */

      {
	struct gm_segment_desc *d;

	/* determine which segment holds the start of the mapping */

	off -= ps->mappings.copy_block.offset;
	d = &ps->copy_block.segment[0];
	while (off >= d->len)
	  {
	    off -= d->len;
	    d++;
	  }

	/* make sure the mapping does not run past the end of the segment */

	if (off + len > d->len)
	  {
	    GM_PRINT (GM_DEBUG_PREPARE_TO_MMAP,
		      ("seg running copy_block %d %d %d", off, len, d->len));
	    GM_RETURN_STATUS (GM_INVALID_PARAMETER);
	  }
      }

      GM_RETURN_STATUS (GM_SUCCESS);
    }
#endif

  /* Did not recognize the requested mapping, so fail. */

  GM_RETURN_STATUS (GM_INVALID_PARAMETER);
}

gm_status_t
gm_finish_mmap (gm_port_state_t * ps, gm_offset_t off, gm_size_t len,
		gm_up_t uaddr)
{
  gm_instance_state_t *is;
  GM_PARAMETER_MAY_BE_UNUSED (off);
  GM_PARAMETER_MAY_BE_UNUSED (len);
  GM_PARAMETER_MAY_BE_UNUSED (uaddr);
  
  GM_CALLED ();

  GM_PRINT ((GM_PRINT_LEVEL >= 5),
	    ("gm_finish_mmap (0x%p, 0x%lx, 0x%lx, 0x%p) called\n", ps, off,
	     len, (void*) uaddr));
  is = ps->instance;
  gm_assert (is);

#if !GM_CAN_REGISTER_MEMORY
  /* If needed, add the mapped pages to hash table */

  {
    unsigned long start, limit;

    start = GM_MAX (off, ps->mappings.copy_block.offset);
    limit = GM_MIN (off + len,
		    ps->mappings.copy_block.offset + ps->copy_block.len);
    if (limit > start)
      {
	gm_status_t status;

	gm_assert (GM_PAGE_ALIGNED (start));
	gm_assert (GM_PAGE_ALIGNED (limit));
	status = gm_add_segment_to_page_table
	  (ps, uaddr,
	   (gm_u32_t) (start - ps->mappings.copy_block.offset),
	   (gm_u32_t) (limit - start));
	if (status != GM_SUCCESS)
	  GM_RETURN_STATUS (status);
      }
  }

#endif

  GM_RETURN_STATUS (GM_SUCCESS);
}

#if !GM_CAN_REGISTER_MEMORY
/* Add the LEN/GM_PAGE_LEN pages at OFFSET in the copy_block for the
   PORT into the page hash table for the device. */
gm_status_t
gm_add_segment_to_page_table (gm_port_state_t * ps, gm_up_t user_addr,
			      gm_u32_t offset, gm_u32_t len)
{
  int i;
  gm_instance_state_t *is;
  gm_status_t status;

  GM_CALLED ();

  GM_PRINT (GM_PRINT_LEVEL >= 2,
	    ("gm_add_segment_to_page_table(uaddr=%p off=0x%x len=0x%x) called\n",
	     user_addr, offset, len));

  gm_assert (GM_PAGE_ALIGNED (user_addr));
  gm_assert (GM_PAGE_ALIGNED (offset));
  gm_assert (GM_PAGE_ALIGNED (len));

  is = ps->instance;

  gm_assert (ps);
  gm_assert (is);

  gm_arch_mutex_enter (&is->page_hash.sync);

  /* No need to pause the LANai when ADDING entries to the page table. */

#if GM_BC_DRIVER
  /* For backwards compatibility, reproduce the mlanai driver HACK of placing
     some info in the first page of the copy block.  We also harmlessly place
     it in the first page of every segment. */

  if (ps->compatibility_mode && offset == 0)
    {
      ((int *) ps->copy_block.segment[0].data)[0]
	= (int) (gm_arch_dma_region_dma_addr
		 (&ps->copy_block.segment[0].dma_region));

      ((int *) ps->copy_block.segment[0].data)[1] = ps->max_dma_mem;

      ((int *) ps->copy_block.segment[0].data)[2]
	= ps->copy_block.segment[0].dma_region.sts;
    }
#endif

  for (i = 0; len && i < 32; i++)
    {
      while (len && offset < ps->copy_block.segment[i].len)
	{
	  gm_dp_t dma_addr;

	  GM_PRINT (GM_PRINT_LEVEL >= 5, ("len = 0x%lx.\n", len));

	  dma_addr
	    =
	    gm_arch_dma_region_dma_addr_advance (&ps->copy_block.
						 segment[i].dma_region);

          GM_PRINT ((GM_DEBUG_PAGE_HASH | GM_DEBUG_HOST_DMA),
                    ("gm_add_segment_to_page_table():\n"
                     "    i=%d, len=0x%x, offset=0x%p, dma_addr=0x%p\n",
		     i, len, offset, dma_addr));

          if (!dma_addr)
	    {
	      GM_WARN (("DMA addr is 0; not adding mapping\n"));
	      break;
	    }

	  GM_PRINT (GM_PRINT_LEVEL >= 5,
		    ("calling gm_add_mapping_to_page_table\n"));
	  status = gm_add_mapping_to_page_table (ps, user_addr, user_addr,
						 dma_addr);
	  if (status != GM_SUCCESS)
	    {
	      GM_NOTE (("Failed to add a mapping to the page table offset"
			" = 0x%x\n", offset));
	      gm_pause_lanai (is);
	      while (offset)
		{
		  GM_PRINT (GM_PRINT_LEVEL >= 5,
			    ("dereferncing mapping for user addr %p\n",
			     user_addr));
		  gm_dereference_mapping (ps, user_addr);
		  user_addr -= GM_PAGE_LEN;
		  offset -= GM_PAGE_LEN;
		  len += GM_PAGE_LEN;
		}
	      gm_unpause_lanai (is);
	      gm_lanai_command_deregister (is, ps->id, user_addr, len);
	      gm_dereference_mapping_commit (is, ps);
	      goto abort_with_mutex;
	    }

	  user_addr += GM_PAGE_LEN;
	  offset += GM_PAGE_LEN;
	  len -= GM_PAGE_LEN;
	}
      offset -= ps->copy_block.segment[i].len;
    }

  GM_PRINT (GM_PRINT_LEVEL >= 6, ("Freeing mutex.\n"));
  gm_arch_mutex_exit (&is->page_hash.sync);

  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_mutex:
  gm_arch_mutex_exit (&is->page_hash.sync);
  GM_RETURN_STATUS (status);
}
#endif

/* Return a pointer to the resource in kernel memory that is mapped to
   offset OFF in the device.  The resource can be the LANai control
   registers, special registers, SRAM, or copy block.  The offsets
   and sizes of these various regions can be obtained using the
   GM_GET_MAPPING_SPECS ioctl, which returns a description of the
   mappings in a "gm_mapping_specs" structure.

   The pages of the copy block must be obtained in order and without
   skipping any offsets.

   gm_prepare_to_mmap ensures that all offsets passed to gm_mmap are
   legitimate, so there is no need to range check offsets here.
   However, checks must be performed here for "promiscuous" mappings
   of the device registers an SRAM areas other than the "send queue".

   Returns pointer to resource, or (void *)0 on failure 
   (was returning (-1) before, changed for consistency with _gm_mmap
*/

#define GM_DEBUG_MMAP 0

#if GM_DEBUG_MMAP
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

gm_status_t
gm_mmap (gm_port_state_t * ps, gm_offset_t orig_off, void **kptr)
{
  gm_instance_state_t *is;
  gm_offset_t off = orig_off;

  GM_CALLED_WITH_ARGS (("%p,0x%lx,%p", ps, orig_off, kptr));

  GM_PRINT (GM_DEBUG_MMAP, ("Mapping offset 0x%lx\n", orig_off));

  if (!GM_PAGE_ALIGNED (off))
    {
      GM_NOTE (("Could not map non-page-aligned offset 0x%lx.\n", off));
      GM_RETURN_STATUS (GM_INVALID_PARAMETER);
    }

  is = ps->instance;
  gm_assert (is);

  *kptr = 0;
  if (!ps->opened)
    {
      /* Special hack for the HW group to allow them to map the entire
	 board. */
      
      if (ps->privileged)
	{
	  /* Allow the user to map the device in "HW developer mode" */

	  GM_PRINT (GM_DEBUG_MMAP, ("Mapping offset 0x%lx\n", orig_off));
	  if (off < is->board_span)
	    {
	      /* Mapping PCI board page */
	      GM_PRINT (GM_DEBUG_MMAP, ("Mapping PCI interface page.\n"));
	      gm_assert (is->board_base);
	      *kptr = (char*) is->board_base + off;
	      GM_RETURN_STATUS (GM_SUCCESS);
	    }
	  else
	    {
	      void *base;
	      
	      /* Mapping DMAable memory page */
	      GM_PRINT (GM_DEBUG_MMAP, ("Mapping contig. DMA page.\n"));
	      off -= is->board_span;
	      base
		= gm_arch_dma_region_kernel_addr (&ps->contiguous_dma_region);
	      *kptr = (char *) base + off;
	      GM_RETURN_STATUS (GM_SUCCESS);
	    }
	}
      
      GM_NOTE (("User tried to map memory before setting port ID.\n"));
      GM_RETURN_STATUS (GM_INVALID_COMMAND);
    }

#if GM_BC_DRIVER

  /* Special case for /dev/mlanai mappings */

  if (ps->compatibility_mode)
    {
      if (off < GM_BC_PAGE_LEN)
	{
	  *kptr = is->lanai.control_regs + off;
	  GM_RETURN_STATUS (GM_SUCCESS);
	}
      off -= GM_BC_PAGE_LEN;

      if (off < GM_BC_PAGE_LEN)
	{
	  *kptr = is->lanai.special_regs + off;
	  GM_RETURN_STATUS (GM_SUCCESS);
	}
      off -= GM_BC_PAGE_LEN;

      if (off < ps->mappings.sram.len)
	{
	  *kptr = is->lanai.sram + off;
	  GM_RETURN_STATUS (GM_SUCCESS);
	}
      off -= ps->mappings.sram.len;
      if (off < ps->copy_block.segment[0].len)
	{
	  *kptr = ps->copy_block.segment[0].data + off;
	  GM_RETURN_STATUS (GM_SUCCESS);
	}
      GM_WARN (("failed compatibility mode mapping at offset 0x%lx", off));

      GM_RETURN_STATUS (GM_INVALID_PARAMETER);
    }
#endif

#define MAPPING_OF_TYPE(type, off, permissions)				\
  (ps->mappings.type.offset <= off					\
   && off < (ps->mappings.type.offset					\
	     + (gm_offset_t) ps->mappings.type.len))

/*** Is this a control register page? */

  if (MAPPING_OF_TYPE (control_regs, off, permissions))
    {
      off -= ps->mappings.control_regs.offset;
      GM_PRINT (GM_DEBUG_MMAP,
		("mapping control reg page (off = 0x%lx)\n", off));
      *kptr = (char *) is->lanai.control_regs + off;
      GM_RETURN_STATUS (GM_SUCCESS);
    }

/*** Is this a special register page? */

  if (MAPPING_OF_TYPE (special_regs, off, permissions))
    {
      off -= ps->mappings.special_regs.offset;
      GM_PRINT (GM_DEBUG_MMAP,
		("mapping special reg page (off = 0x%lx)\n", off));
      *kptr = (char *) is->lanai.special_regs + off;
      GM_RETURN_STATUS (GM_SUCCESS);
    }

/*** Is this an SRAM page? */

  gm_assert (ps->mappings.sram.len == is->lanai.eeprom.lanai_sram_size);
  if (MAPPING_OF_TYPE (sram, off, permissions))
    {
      off -= ps->mappings.sram.offset;
      GM_PRINT (GM_DEBUG_MMAP,
		("mapping sram page (off = 0x%lx)\n", off));
      *kptr = (char *) is->lanai.sram + off;
      GM_RETURN_STATUS (GM_SUCCESS);
    }

  if (MAPPING_OF_TYPE (send_queue, off, permissions))
    {
      GM_PANIC (("loic: oops: this code looks wrong\n"));
      off -= ps->mappings.send_queue.offset;
      GM_PRINT (GM_DEBUG_MMAP,
		("mapping send_queue page (off = 0x%lx)\n", off));
      *kptr = (char *) is->lanai.sram + off;
      GM_RETURN_STATUS (GM_SUCCESS);
    }

  if (MAPPING_OF_TYPE (RTC, off, permissions))
    {
      GM_PRINT (GM_DEBUG_MMAP, ("User mapped the RTC\n"));
      off -= ps->mappings.RTC.offset;
      GM_PRINT (GM_DEBUG_MMAP,
		("mapping RTC page (off = 0x%lx)\n", off));
      *kptr = (char *) is->lanai.sram + off;
      GM_RETURN_STATUS (GM_SUCCESS);
    }

#if !GM_CAN_REGISTER_MEMORY
/*** Is this a copy block page? */

  if (MAPPING_OF_TYPE (copy_block, off, permissions))
    {
      int i;

      off -= ps->mappings.copy_block.offset;

      gm_assert (off < ps->max_dma_mem);
      gm_assert (off < ps->copy_block.len);

      GM_PRINT (GM_DEBUG_MMAP,
		("mapping copy block page (off = 0x%lx)\n", off));

      /* Compute I and OFF such that the page at offset OFF in segment I of
         the copy block is the page to be mapped. */

      for (i = 0;
	   off >= ps->copy_block.segment[i].len;
	   off -= ps->copy_block.segment[i++].len)
	{
	  gm_assert (i < 32);
	}
      *kptr = ps->copy_block.segment[i].data + off;
      GM_RETURN_STATUS (GM_SUCCESS);
    }
#endif

/*** The offset is invalid. */
  GM_NOTE (("User attempted invalid mapping at offset 0x%lx\n", off));
  GM_RETURN_STATUS (GM_INVALID_PARAMETER);
}

int
gm_mapping_in_io_space (gm_port_state_t * ps, gm_offset_t off)
{
  return (MAPPING_OF_TYPE (control_regs, off, 0)
	  || MAPPING_OF_TYPE (special_regs, off, 0)
	  || MAPPING_OF_TYPE (sram, off, 0));
}

#undef MAPPING_OF_TYPE

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
