/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* Written by Brice Goglin (bgoglin@myri.com) July-August 2001
 *
 * The space is divided into free segments and used segments
 *
 * Contigous free segments are merged.
 * Free segments are stored in OPIOM_NB_LISTS lists.
 * The n-th list contains segments of (n+1) blocks (size OPIOM_PAGE_SIZE).
 * The last list (the (OPIOM_NB_LISTS-1)-th)
 *  contains all big free segments, sorted by their size
 *
 * Used segment are stored in a AVL tree.
 * As equilibrating of the AVL implies switching of nodes,
 *  a pointer to such a used segment may change
 * Each node is associated to a gm_opiom_segment_t
 *  whose address never changes.
 * Users only use gm_opiom_segment_t
 */

#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_opiom_malloc.h"


#if GM_ENABLE_OPIOM

#define GM_DEBUG_OPIOM_MALLOC 0


#define OPIOM_PAGE_SIZE 512
#define OPIOM_PAGE_SIZE_BIT 9

#define OPIOM_NB_LISTS 256

#define OPIOM_NB_INIT_FREE 256
#define OPIOM_NB_INIT_USED 256

#define OPIOM_LIST_IS_EMPTY(x) ( free_lists[x].next == free_lists+x )

#define OPIOM_GET_PID(args...) (current->pid)


/********
 * Types
 ********/

/* type used to store free segments in lists */
typedef struct _free {
  gm_up_t addr;
  gm_u32_t size;
  struct _free *next, *prev;
} free_t;

/* return value of AVL functions */
enum AVLSKEW { AVLSKEW_NONE, AVLSKEW_LEFT, AVLSKEW_RIGHT };
enum AVLRES { AVLRES_ERROR = 0, AVLRES_OK, AVLRES_BALANCE };

/* type used to stores used segments in AVL */
struct avlnode
{
  struct avlnode *left, *right;
  gm_u32_t size;
  gm_up_t addr;
  struct _free * next_free;
  enum AVLSKEW skew;
  gm_opiom_segment_t *seg;
  gm_pid_t pid;
};


/**********
 * Globals
 **********/

/* chained lists for free segment */
GM_STATIC free_t free_lists[OPIOM_NB_LISTS];
/* ptrs to the next non-empty list */
GM_STATIC gm_u32_t ptr_lists[OPIOM_NB_LISTS];
/* ptr to first free segment (it is not the next_free of any used segment) */
GM_STATIC free_t * first_free;

/* AVL tree for used segment */
GM_STATIC struct avlnode * used_tree;

/* available free structures */
GM_STATIC struct gm_lookaside * free_lookaside;
/* available used structures and pointers*/
GM_STATIC struct gm_lookaside * used_lookaside;
GM_STATIC struct gm_lookaside * used_ptr_lookaside;

/* address at the begin of the opiom buffer */
GM_STATIC gm_up_t init_addr;



/***************
 * Rotating AVL
 ***************/

GM_STATIC void
avl_rot_left(struct avlnode **n)
{
  struct avlnode *tmp = *n;
  
  *n = (*n)->right;
  tmp->right = (*n)->left;
  (*n)->left = tmp;
}


GM_STATIC void
avl_rot_right(struct avlnode **n)
{
  struct avlnode *tmp = *n;
  
  *n = (*n)->left;
  tmp->left = (*n)->right;
  (*n)->right = tmp;
}


/**************
 * Growing AVL
 **************/

GM_STATIC enum AVLRES
avl_left_grown(struct avlnode **n)
{
  switch ((*n)->skew)
  {
    case AVLSKEW_LEFT:
      if ((*n)->left->skew == AVLSKEW_LEFT)
      {
	(*n)->skew = (*n)->left->skew = AVLSKEW_NONE;
	avl_rot_right(n);
      }	
      else 
      {
	switch ((*n)->left->right->skew) 
	{
	  case AVLSKEW_LEFT:
	    (*n)->skew = AVLSKEW_RIGHT;
	    (*n)->left->skew = AVLSKEW_NONE;
	    break;
	    
	  case AVLSKEW_RIGHT:
	    (*n)->skew = AVLSKEW_NONE;
	    (*n)->left->skew = AVLSKEW_LEFT;
	    break;
	    
	  default:
	    (*n)->skew = AVLSKEW_NONE;
	    (*n)->left->skew = AVLSKEW_NONE;
	}
	(*n)->left->right->skew = AVLSKEW_NONE;
	avl_rot_left(& (*n)->left);
	avl_rot_right(n);
      }
      return AVLRES_OK;
      
    case AVLSKEW_RIGHT:
      (*n)->skew = AVLSKEW_NONE;
      return AVLRES_OK;
      
    default:
      (*n)->skew = AVLSKEW_LEFT;
      return AVLRES_BALANCE;
  }
}


GM_STATIC enum AVLRES
avl_right_grown(struct avlnode **n)
{
  switch ((*n)->skew) 
  {
    case AVLSKEW_LEFT:					
      (*n)->skew = AVLSKEW_NONE;
      return AVLRES_OK;
      
    case AVLSKEW_RIGHT:
      if ((*n)->right->skew == AVLSKEW_RIGHT) 
      {	
	(*n)->skew = (*n)->right->skew = AVLSKEW_NONE;
	avl_rot_left(n);
      }
      else 
      {
	switch ((*n)->right->left->skew) 
	{
	  case AVLSKEW_RIGHT:
	    (*n)->skew = AVLSKEW_LEFT;
	    (*n)->right->skew = AVLSKEW_NONE;
	    break;
	    
	  case AVLSKEW_LEFT:
	    (*n)->skew = AVLSKEW_NONE;
	    (*n)->right->skew = AVLSKEW_RIGHT;
	    break;
	    
	  default:
	    (*n)->skew = AVLSKEW_NONE;
	    (*n)->right->skew = AVLSKEW_NONE;
	}
	(*n)->right->left->skew = AVLSKEW_NONE;
	avl_rot_right(& (*n)->right);
	avl_rot_left(n);
      }
      return AVLRES_OK;
      
    default:
      (*n)->skew = AVLSKEW_RIGHT;
      return AVLRES_BALANCE;
  }
}


/****************
 * Insert in AVL
 ****************/

GM_STATIC enum AVLRES
avl_real_insert(struct avlnode **n, struct avlnode *new, gm_u8_t *inserted)
{
  enum AVLRES tmp;
  
  if (!(*n))
  {
    new->left = NULL;
    new->right = NULL;
    new->skew = AVLSKEW_NONE;
    *n = new;
    *inserted = 1;
    return AVLRES_BALANCE;
  }

  if ( new->addr < (*n)->addr)
  {
    tmp = avl_real_insert(& (*n)->left, new, inserted);
    if ( *inserted )
    {
      // nothing
      *inserted = 1;
    } 
    if ( tmp == AVLRES_BALANCE )
      return avl_left_grown(n);
    return tmp;
  }

  if ( new->addr > (*n)->addr)
  {
    tmp = avl_real_insert(& (*n)->right, new, inserted);
    if ( *inserted )
    {
      (*n)->next_free = NULL;
      *inserted = 0;
    } 
    if ( tmp == AVLRES_BALANCE )
      return avl_right_grown(n);
    return tmp;
  }

  return AVLRES_ERROR;
}


GM_STATIC enum AVLRES
avl_insert(struct avlnode **n, struct avlnode *new)
{
  gm_u8_t flag = 0;
  return avl_real_insert(n, new, &flag);
}


/*************
 * Shrunk AVL
 *************/

GM_STATIC enum AVLRES
avl_left_shrunk(struct avlnode **n)
{
  switch ((*n)->skew)
  {
    case AVLSKEW_LEFT:
      (*n)->skew = AVLSKEW_NONE;
      return AVLRES_BALANCE;
      
    case AVLSKEW_RIGHT:
      if ((*n)->right->skew == AVLSKEW_RIGHT)
      {
	(*n)->skew = (*n)->right->skew = AVLSKEW_NONE;
	avl_rot_left(n);
	return AVLRES_BALANCE;
      }
      else if ((*n)->right->skew == AVLSKEW_NONE)
      {
	(*n)->skew = AVLSKEW_RIGHT;
	(*n)->right->skew = AVLSKEW_LEFT;
	avl_rot_left(n);
	return AVLRES_OK;
      }
      else
      {
	switch ((*n)->right->left->skew)
	{
	  case AVLSKEW_LEFT:
	    (*n)->skew = AVLSKEW_NONE;
	    (*n)->right->skew = AVLSKEW_RIGHT;
	    break;
	    
	  case AVLSKEW_RIGHT:
	    (*n)->skew = AVLSKEW_LEFT;
	    (*n)->right->skew = AVLSKEW_NONE;
	    break;
	    
	  default:
	    (*n)->skew = AVLSKEW_NONE;
	    (*n)->right->skew = AVLSKEW_NONE;
	}
	(*n)->right->left->skew = AVLSKEW_NONE;
	avl_rot_right(& (*n)->right);
	avl_rot_left(n);
	return AVLRES_BALANCE;
      }
      
    default:
      (*n)->skew = AVLSKEW_RIGHT;
      return AVLRES_OK;
  }
}


GM_STATIC enum AVLRES
avl_right_shrunk(struct avlnode **n)
{
  switch ((*n)->skew)
  {
    case AVLSKEW_RIGHT:
      (*n)->skew = AVLSKEW_NONE;
      return AVLRES_BALANCE;
      
    case AVLSKEW_LEFT:
      if ((*n)->left->skew == AVLSKEW_LEFT)
      {
	(*n)->skew = (*n)->left->skew = AVLSKEW_NONE;
	avl_rot_right(n);
	return AVLRES_BALANCE;
      }
      else if ((*n)->left->skew == AVLSKEW_NONE)
      {
	(*n)->skew = AVLSKEW_LEFT;
	(*n)->left->skew = AVLSKEW_RIGHT;
	avl_rot_right(n);
	return AVLRES_OK;
      }
      else
      {
	switch ((*n)->left->right->skew)
	{
	  case AVLSKEW_LEFT:
	    (*n)->skew = AVLSKEW_RIGHT;
	    (*n)->left->skew = AVLSKEW_NONE;
	    break;
	    
	  case AVLSKEW_RIGHT:
	    (*n)->skew = AVLSKEW_NONE;
	    (*n)->left->skew = AVLSKEW_LEFT;	
	    break;
	    
	  default:
	    (*n)->skew = AVLSKEW_NONE;
	    (*n)->left->skew = AVLSKEW_NONE;
	}
	(*n)->left->right->skew = AVLSKEW_NONE;
	avl_rot_left(& (*n)->left);
	avl_rot_right(n);
	return AVLRES_BALANCE;
      }
      
    default:
      (*n)->skew = AVLSKEW_LEFT;
      return AVLRES_OK;
  }
}


/*********************************
 * Find Lowest and Highest in AVL
 *********************************/

/* This function finds the highest node
 * stores its pointer in "used_to_free"
 * put its coordinates in the "target" node
 * and store its new pointer in "target"
 *
 * it is used in remove: the node to be removed
 * is switch with the highest node is the left child
 * which is easier to remove
 */

GM_STATIC int
avl_find_highest(struct avlnode *target, struct avlnode **n,
		 enum AVLRES *res,
		 struct avlnode **used_to_free,
		 struct avlnode **previous)
{
  struct avlnode *tmp;
  gm_opiom_segment_t *used_ptr_tmp;
  
  *res = AVLRES_BALANCE;
  if (!(*n))
    return 0;

  if ((*n)->right)
  {
    if (!avl_find_highest(target, &(*n)->right, res, used_to_free, previous))
      return 0;
    if (*res == AVLRES_BALANCE)
      *res = avl_right_shrunk(n);
    return 1;
  }

  /* take data */
  target->addr = (*n)->addr;
  target->size = (*n)->size;
  target->next_free = (*n)->next_free;
  target->pid = (*n)->pid;

  /* switch gm_opiom_segments */
  used_ptr_tmp = target->seg;
  target->seg = (*n)->seg;
  (*n)->seg = used_ptr_tmp;
  /* update the good one */
  *(target->seg) = target;

  /* finish tree modification */
  tmp = *n;
  *n = (*n)->left;

  /* store the previous node */
  *previous = target;

  /* prepare deleting */
  *used_to_free = tmp;

  return 1;
}


/* This function finds the lowest node
 * stores its pointer in "used_to_free"
 * put its coordinates in the "target" node
 *
 * it is used in remove: the node to be removed
 * is switch with the lowest node is the right child
 * which is easier to remove
 */

GM_STATIC int
avl_find_lowest(struct avlnode *target, struct avlnode **n,
		enum AVLRES *res,
		struct avlnode **used_to_free)
{
  struct avlnode *tmp;
  gm_opiom_segment_t *used_ptr_tmp;

  *res = AVLRES_BALANCE;
  if (!(*n))
    return 0;
  if ((*n)->left)
  {
    if (!avl_find_lowest(target, &(*n)->left, res, used_to_free))
      return 0;
    if (*res == AVLRES_BALANCE)
      *res =  avl_left_shrunk(n);
    return 1;
  }

  /* take data */
  target->addr = (*n)->addr;
  target->size = (*n)->size;
  target->next_free = (*n)->next_free;
  target->pid = (*n)->pid;

  /* switch gm_opiom_segments */
  used_ptr_tmp = target->seg;
  target->seg = (*n)->seg;
  (*n)->seg = used_ptr_tmp;
  /* update the good one */
  *(target->seg) = target;

  /* finish tree modification */
  tmp = *n;
  *n = (*n)->right;
  
  /* prepare deleting */
  *used_to_free = tmp;

  return 1;
}


/****************
 * Remove in AVL
 ****************/

GM_STATIC enum AVLRES
avl_remove(struct avlnode **n, gm_up_t addr,
	   struct avlnode **used_to_free,
	   struct avlnode **previous)
{
  enum AVLRES tmp = AVLRES_BALANCE;

  /* not the good node */
  if (!(*n))
    return AVLRES_ERROR;
  if ( addr < (*n)->addr )
  {
    if ((tmp = avl_remove(& (*n)->left, addr,
			  used_to_free, previous)) == AVLRES_BALANCE)
      return avl_left_shrunk(n);
    return tmp;
  }
  if ( addr > (*n)->addr )
  {
    *previous = *n;
    if ((tmp = avl_remove(& (*n)->right, addr,
			  used_to_free, previous)) == AVLRES_BALANCE)
      return avl_right_shrunk(n);
    return tmp;
  }

  /* we've got it
   * now, we need the highest previous or the lowest next to switch.
   * if none exists, there are no child so we can delete 
   */
  if ((*n)->left) {
    if (avl_find_highest(*n, &((*n)->left), &tmp, used_to_free, previous))
    {
      if (tmp == AVLRES_BALANCE)
	tmp = avl_left_shrunk(n);
      return tmp;
    }
  }
  if ((*n)->right)
  {
    if (avl_find_lowest(*n, &((*n)->right), &tmp, used_to_free))
    {
      if (tmp == AVLRES_BALANCE)
	tmp = avl_right_shrunk(n);
      return tmp;
    }
  }

  /* prepare deleting */
  *used_to_free = *n;
  
  *n = NULL;

  return AVLRES_BALANCE;
}


/**************************
 * Manipulating free lists 
 **************************/

GM_STATIC void 
add_free_to_list(free_t * new_free)
{
  gm_u32_t j;
  gm_u32_t size = new_free->size;
  gm_u32_t i = ( size >> OPIOM_PAGE_SIZE_BIT ) - 1;
  free_t * begin;

  /* last list keeps all big size */
  if ( i >= OPIOM_NB_LISTS )
    i = OPIOM_NB_LISTS-1;

  /* update ptr */
  if ( OPIOM_LIST_IS_EMPTY(i) )
  {
    for(j=i-1; j>=0 && j<OPIOM_NB_LISTS; j--)
    {
      if ( ! OPIOM_LIST_IS_EMPTY(j) )
	break;
      ptr_lists[j] = i;
    }
    if ( j >= 0 && j<OPIOM_NB_LISTS )
      ptr_lists[j] = i;
  }

  /* add the element */
  begin = free_lists+i;

  if ( i == OPIOM_NB_LISTS-1 )
  {
    while ( begin->next != free_lists+i 
	    && begin->next->size < size )
      begin = begin->next;
  }

  begin->next->prev = new_free;
  new_free->prev = begin;

  new_free->next = begin->next;
  begin->next = new_free;
}


GM_STATIC /*inline*/void
update_pointer_for_empty_next(gm_u32_t i)
{
  int j;
  int k = ptr_lists[i];
  for(j=i-1; j>=0 && j<OPIOM_NB_LISTS ; j--)
  {
    if ( ! OPIOM_LIST_IS_EMPTY(j) )
      break;
    ptr_lists[j] = k;
  }
  if ( j >= 0 && j<OPIOM_NB_LISTS )
    ptr_lists[j] = k;
}


GM_STATIC free_t *
get_free_from_list(gm_u32_t size)
{
  free_t * found;
  int i = ( size >> OPIOM_PAGE_SIZE_BIT ) - 1;

  /* last list keeps all big size */
  if ( i >= OPIOM_NB_LISTS )
    i = OPIOM_NB_LISTS-1;

    /* find the first next non-empty list */
  while ( OPIOM_LIST_IS_EMPTY(i) )
  {
    i = ptr_lists[i];
    if ( i == 0 )
      return NULL;
  }
  found = free_lists[i].next;

  if ( i == OPIOM_NB_LISTS-1 )
  {
    while ( found != free_lists+i
	    && found->size < size )
      found = found->next;
    if ( found == free_lists+i )
      return NULL;
  }

  found->next->prev = found->prev;
  found->prev->next = found->next;
   
  if ( OPIOM_LIST_IS_EMPTY(i) )
    update_pointer_for_empty_next(i);

  return found;
}


GM_STATIC void
remove_free_from_list(free_t *old)
{
  old->next->prev = old->prev;
  old->prev->next = old->next;
  
  if ( old->next == old->prev )
    /* the list is empty */
    update_pointer_for_empty_next(old->next - free_lists);

  return;
}
#endif /* GM_ENABLE_OPIOM */


/******************
 * Malloc Library *
 ******************/

int
_gm_opiom_malloc_init(void)
{
#if GM_ENABLE_OPIOM
  gm_u32_t i;
  free_t *all;

  if ( gm_init() != GM_SUCCESS )
    return -1;

  init_addr = 0;

  free_lookaside = gm_create_lookaside(sizeof(free_t), OPIOM_NB_INIT_FREE);
  used_lookaside = gm_create_lookaside(sizeof(struct avlnode), OPIOM_NB_INIT_USED);
  used_ptr_lookaside = gm_create_lookaside(sizeof(gm_opiom_segment_t), OPIOM_NB_INIT_USED);
  if ( !free_lookaside  ||  !used_lookaside )
    return -1;

  /* all lists are empty except the last */
  for(i=0; i<OPIOM_NB_LISTS-1; i++)
  {
    free_lists[i].next = free_lists+i;
    free_lists[i].prev = free_lists+i;
    ptr_lists[i] = OPIOM_NB_LISTS-1;
  }

  /* create the big free */
  all = (free_t*) gm_lookaside_alloc(free_lookaside);
  if ( !all )
    return -1;

  all->addr = init_addr;
  all->size = GM_OPIOM_DATA_LENGTH;
  all->next = free_lists + OPIOM_NB_LISTS-1;
  all->prev = free_lists + OPIOM_NB_LISTS-1;

  free_lists[OPIOM_NB_LISTS-1].next = all;
  free_lists[OPIOM_NB_LISTS-1].prev = all;
  ptr_lists[OPIOM_NB_LISTS-1] = 0;

  /* the tree is empty */
  used_tree = NULL;
  /* the first free */
  first_free = all;

  return 0;
#else /* GM_ENABLE_OPIOM */
  return -1;
#endif /* GM_ENABLE_OPIOM */
}


void
_gm_opiom_malloc_finalize(void)
{
#if GM_ENABLE_OPIOM
  gm_destroy_lookaside(free_lookaside);
  gm_destroy_lookaside(used_lookaside);
  gm_destroy_lookaside(used_ptr_lookaside);  
#endif /* GM_ENABLE_OPIOM */
}
  

/**********
 * Malloc *
 **********/

gm_status_t
_gm_opiom_malloc(gm_u32_t real_size, gm_opiom_segment_t **node)
{
#if GM_ENABLE_OPIOM
  gm_u32_t size = (real_size + OPIOM_PAGE_SIZE - 1) & ~(OPIOM_PAGE_SIZE - 1);
  struct avlnode * new_used;
  gm_opiom_segment_t * new_used_ptr;
  free_t * old_free;

  GM_PRINT(GM_DEBUG_OPIOM_MALLOC, ("Malloc of size %d\n", size));

  if ( size <= 0 )
    goto error;

  new_used = gm_lookaside_alloc(used_lookaside);
  new_used_ptr = gm_lookaside_alloc(used_ptr_lookaside);
  if ( !new_used  ||  !new_used_ptr )
    goto error;

  old_free = get_free_from_list(size);
  if ( old_free == NULL )
    goto error;

  new_used->addr = old_free->addr;
  new_used->size = size;
  new_used->seg = new_used_ptr;
  new_used->pid = OPIOM_GET_PID();
  *new_used_ptr = new_used;

  if ( old_free->size == size )
  {
    new_used->next_free = NULL;
    gm_lookaside_free(old_free);
   }
  else
  {
    old_free->addr += size;
    old_free->size -= size;
    add_free_to_list(old_free);
    new_used->next_free = old_free;
  }

  if ( old_free == first_free )
    first_free = NULL;

  avl_insert(&used_tree, new_used);

  GM_PRINT(GM_DEBUG_OPIOM_MALLOC, ("Returning %p for %p\n", new_used_ptr, new_used->addr));

  *node = new_used_ptr;
  return GM_SUCCESS;

  error:
  GM_WARN(("Opiom Malloc: Cannot find buffer of size %d\n", size));
  return GM_FAILURE;
#else /* GM_ENABLE_OPIOM */
  return GM_FAILURE;
#endif /* GM_ENABLE_OPIOM */
}


gm_status_t
_gm_opiom_free(gm_opiom_segment_t * seg)
{
#if GM_ENABLE_OPIOM
  struct avlnode * old_used = *seg;

  /* check the pid */
  if ( old_used->pid != OPIOM_GET_PID() )
  {
    GM_WARN(("Opiom Malloc Get Ptr: Bad pid %d instead of %d\n", OPIOM_GET_PID(), old_used->pid));
    return GM_FAILURE;
  }

  {
    /* caracteristics of the node which will be removed */
    gm_up_t addr = old_used->addr;
    gm_u32_t size = old_used->size;
    free_t * next_free = old_used->next_free;
    /* the (updated) pointer to the node which will be removed */
    struct avlnode * used_to_free = NULL;
    /* the (updated) pointer of the old previous node */
    struct avlnode * prev_used = NULL;
    /* the free segment which will get the new free space:
     * the previous one if it exists, or a new one */
    free_t * prev_free;

    GM_PRINT(GM_DEBUG_OPIOM_MALLOC, ("Freeing %p for %p and %p\n", seg, old_used, addr));

    /* remove the node from the tree, but store pointers
     * it and its previous node */
    avl_remove(&used_tree, addr, &used_to_free, &prev_used);

    if ( prev_used )
    { /* there is a used before */
      prev_free = prev_used->next_free;
      if ( prev_free )
      { /* one free just before: increases its size */
	remove_free_from_list(prev_free);
	prev_free->size += size;
      }
      else
      { /* a block just before: add a new free */
	prev_free = (free_t*) gm_lookaside_alloc(free_lookaside);
	if ( prev_free == NULL )
	  return GM_FAILURE;
	prev_free->addr = addr;
	prev_free->size = size;
	/* link to the previous used */
	prev_used->next_free = prev_free;
      } 
    }
    else 

    { /* no used before */
      if ( addr != init_addr )
      { /* a free before: the first free: increases its size */
	remove_free_from_list(first_free);
	prev_free = first_free;
	prev_free->size += size;
      }
      else
      { /* at the begining: add a new free */
	prev_free = (free_t*) gm_lookaside_alloc(free_lookaside);
	if ( prev_free == NULL )
	  return GM_FAILURE;
	prev_free->addr = init_addr;
	prev_free->size = size;
	/* link to the first free */
	first_free = prev_free;	
      }
    }
    if ( next_free )
      prev_free->size += next_free->size;
    add_free_to_list(prev_free);

    /* remove the next free segment from list */
    if ( next_free )
    {
      remove_free_from_list(next_free);
      gm_lookaside_free(next_free);
    }

    /* real memory remove
     * we now are happy that gm_opiom_segment_t had been well switched */
    gm_lookaside_free(used_to_free->seg);
    gm_lookaside_free(used_to_free);
    return GM_SUCCESS;
  }
#else /* GM_ENABLE_OPIOM */
  return GM_FAILURE;
#endif /* GM_ENABLE_OPIOM */
}


gm_status_t
_gm_opiom_malloc_get_ptr(gm_opiom_segment_t node, gm_up_t *addr)
{
#if GM_ENABLE_OPIOM
  /* check the pid */
  if ( node->pid != OPIOM_GET_PID() )
  {
    GM_WARN(("Opiom Malloc Get Ptr: Bad pid %d instead of %d\n", OPIOM_GET_PID(), node->pid));
    return GM_FAILURE;
  }
  *addr = node->addr;
  return GM_SUCCESS;
#else /* GM_ENABLE_OPIOM */
  return GM_FAILURE;
#endif /* GM_ENABLE_OPIOM */
}
