
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

/* This module initializes and finalizes ports */

#include "gm_call_trace.h"
#include "gm_debug_host_dma.h"
#include "gm_debug_port_state.h"
#include "gm_debug_sync.h"
#include "gm_compiler.h"
#include "gm_impl.h"
#include "gm_enable_security.h"
#include "gm_lanai_command.h"
#include "gm_pio.h"

/* Set this to 1 to enable debugging of the local mutex */
#define GM_DEBUG_LOCAL_MUTEX 0

#if GM_DEBUG_PORT_STATE
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

static struct gm_mutex *local_mutex = (struct gm_mutex *)0;
static int local_mutex_initialized=0;

gm_status_t
__gm_init_local_mutex ()
{
  GM_CALLED ();
  gm_mutex_enter (_gm_global_mutex);
  GM_PRINT(GM_DEBUG_LOCAL_MUTEX, ("entry to __gm_init_local_mutex()\n"));

  gm_always_assert(local_mutex_initialized>=0);
  if (!local_mutex_initialized++)
    {
      if (local_mutex)
	GM_PANIC(("local_mutex_initialized = %d, local_mutex = 0x%p\n",
		  local_mutex_initialized, local_mutex));

      GM_PRINT(GM_DEBUG_LOCAL_MUTEX,
               ("__gm_init_local_mutex() calling gm_create_mutex()\n"));
      local_mutex = gm_create_mutex ();
      GM_PRINT(GM_DEBUG_LOCAL_MUTEX, ("gm_create_mutex() %s\n",
               ((local_mutex) ? "succeeded" : "failed") ));
    }
  else {
  }
  GM_PRINT(GM_DEBUG_LOCAL_MUTEX, ("exit from __gm_init_local_mutex()\n"));

  gm_mutex_exit (_gm_global_mutex);

  if (!local_mutex) 
    GM_RETURN_STATUS (GM_FAILURE);
  else
    GM_RETURN_STATUS (GM_SUCCESS);
}

void
__gm_finalize_local_mutex ()
{
  GM_CALLED ();
  gm_mutex_enter (_gm_global_mutex);
  GM_PRINT(GM_DEBUG_LOCAL_MUTEX, ("entry to __gm_finalize_local_mutex()\n"));
  if (!--local_mutex_initialized)
    {
      if (!local_mutex)
	GM_PANIC(("local_mutex_initialized = %d, local_mutex = 0x%p\n",
		  local_mutex_initialized, local_mutex));

      GM_PRINT(GM_DEBUG_LOCAL_MUTEX,
               ("__gm_finalize_local_mutex() calling gm_destroy_mutex()\n"));
      gm_destroy_mutex (local_mutex);
      local_mutex = (struct gm_mutex *)0;
    }
  else {
  }
  gm_always_assert (local_mutex_initialized >= 0);
  GM_PRINT(GM_DEBUG_LOCAL_MUTEX, ("exit from __gm_finalize_local_mutex()\n"));
  gm_mutex_exit (_gm_global_mutex);
  GM_RETURN_NOTHING ();
}

static
void
default_port_state_wake_handler (void *context)
{
  gm_port_state_t *ps;

  ps = (gm_port_state_t *) context;
  gm_assert (ps);
  gm_arch_wake (&ps->sleep_sync);
}

/************
 * open
 ************/

gm_status_t
gm_port_state_open (gm_port_state_t * ps, unsigned int port_num)
{
  /* NOTE: This function must not cause any GM sleep, because it is
           called as part of GM_SET_PORT_NUM ioctl processing, and the
           platform-specific code is permitted to hold a global lock while
           processing that ioctl.
  */
  gm_instance_state_t *is;
  gm_status_t status = GM_INTERNAL_ERROR;

  GM_CALLED ();

  if (!local_mutex)
    GM_PANIC(("local_mutex_initialized = %d, local_mutex = 0x%p\n",
	      local_mutex_initialized, local_mutex));

  gm_mutex_enter (local_mutex);

  GM_PRINT (GM_DEBUG_PORT_STATE,
	    ("gm_port_state_open (%p, %d) called.\n", ps, port_num));

  if (port_num > GM_NUM_PORTS)
    {
      status = GM_INVALID_PARAMETER;
      goto abort_with_mutex;
    }

  gm_assert (ps);
  is = ps->instance;
  gm_assert (is);

  /* Verify that the port is not already in use. */

  GM_PRINT (GM_DEBUG_PORT_STATE,
            ("[%d]: gm_port_state_open: instance = %p\n", is->id, is));

  if (port_num < GM_NUM_ELEM (is->ps_for_id)
      && is->ps_for_id[port_num])
    {
      if (!GM_ENABLE_SECURITY && !is->lanai.running && (port_num == 0))
	{
	  GM_NOTE (("Allowing user to open a port (%d) that is in use for"
		    " debug.\n", port_num));
	}
      else
	{
	  GM_PRINT(GM_PRINT_LEVEL>=1,
		   ("User tried to claim a port number (%d) that is in use.\n",
		    port_num));
	  status = GM_BUSY;
	  goto abort_with_mutex;
	}
    }

  /* Verify the LANai is running. */

  if (!is->lanai.running)
    {
      if (ps->privileged && (port_num == 0))
	{
	  GM_NOTE (("LANai is not running. Allowing port=0 open for"
		    " debugging\n"));
	}
      else
	{
	  GM_NOTE (("Could not open port because LANai is not running.\n"));
	  status = GM_FIRMWARE_NOT_RUNNING;
	  goto abort_with_mutex;
	}
    }

  /* Check permissions */

#if !GM_ENABLE_SECURITY
  ps->privileged = 1;
  GM_PRINT (GM_DEBUG_SECURITY,
	    ("security disabled: port privileges forced\n"));
#endif

  GM_PRINT (GM_DEBUG_PORT_STATE, ("Checked permissions.\n"));
  if (!ps->privileged && (port_num == GM_ETHERNET_PORT_ID
			  || port_num == GM_MAPPER_PORT_ID))
    {
      GM_PRINT (GM_DEBUG_SECURITY,
		("Unprivileged user tried to open privileged port.\n"));
      status = GM_PERMISSION_DENIED;
      goto abort_with_mutex;
    }

  /* Verify any previous port open has finished shutting down. */

  if (gm_read_lanai_global_u32 (is, port[port_num].active_subport_cnt) != 0)
    {
      if (!GM_ENABLE_SECURITY && (port_num == 0))
	{
	  GM_NOTE (("Still shutting down. Allowing port=0 open for"
		    " debugging\n"));
	}
      else
	{
	  GM_NOTE (("Could not open port because it is still shutting"
		    " down active_count=%d.\n",
		    gm_read_lanai_global_u32 (is,
                                         port[port_num].active_subport_cnt)));
	  status = GM_STILL_SHUTTING_DOWN;
	  goto abort_with_mutex;
	}
    }

  /* HACK: Clean up the port state in the lanai, since the process of
     closing down the port can corrupt the reset state. */

  gm_lanai_command_close (is, port_num);

  /* prep the host side of the port */

  ps->opened = 1;
  ps->closing = 0;
  ps->id = port_num;
  ps->compatibility_mode = 0;
  /* ps->privileged; set by arch */
  ps->min_dma_mem = 0;
  /* max amount of DMA-able memory when no registration is supported (Solaris) */
  ps->max_dma_mem = 0x40000000;
  /* GM_NOTE (("DMA memory limits are disabled.\n")); */
  gm_init_mapping_spec (ps);
  GM_PRINT (GM_DEBUG_SYNC, ("Creating sleep sync for unit %d, port %d\n",
			    is->id, ps->id));
  gm_arch_sync_init (&ps->sleep_sync, is);
  GM_PRINT (GM_DEBUG_SYNC, ("Creating wait sync for unit %d, port %d\n",
			    is->id, ps->id));
  gm_arch_sync_init (&ps->wait_sync, is);
  /* ps->copy_block: 0 */
  /* ps->recv_queue_region: initialized by this function call */
  ps->wake.handler = default_port_state_wake_handler;
  ps->wake.context = (void *) ps;
  GM_PRINT (GM_DEBUG_SYNC, ("Creating YP sync for unit %d, port %d\n",
			    is->id, ps->id));
  gm_arch_sync_init (&ps->yp_sync, is);
  
  /* Insert port in instance's port hash table to enable port_num->ps
     translation */

  if (port_num >= GM_NUM_ELEM (is->ps_for_id))
    {
      GM_NOTE (("Could not insert port state in instance's port table.\n"));
      goto abort_with_yp_sync;
    }
  is->ps_for_id[port_num] = ps;
  
  /* Peform architecture-specific port opening actions */

  status = gm_arch_port_state_open (ps);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not perform arch-specific port state open.\n"));
      goto abort_with_hash_insertion;
    }

  /* perform the LANai-specific port opening actions */
  
  status = gm_lanai_command_open (is, port_num);
  if (is->lanai.running && status != GM_SUCCESS)
    {
      goto abort_with_arch_port_state_open;
    }

  gm_write_lanai_global_u32 (is, port[ps->id].privileged, ps->privileged);
  gm_mutex_exit (local_mutex);
  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_arch_port_state_open:
  gm_arch_port_state_close (ps);
abort_with_hash_insertion:
  is->ps_for_id[port_num] = 0;
abort_with_yp_sync:
  gm_arch_sync_destroy (&ps->yp_sync);
  gm_arch_sync_destroy (&ps->wait_sync);
  gm_arch_sync_destroy (&ps->sleep_sync);
  ps->opened = 0;
  ps->closing = 1;
abort_with_mutex:
  gm_mutex_exit (local_mutex);
  GM_RETURN_STATUS (status);
}

/************
 * close
 ************/

#define GM_DEBUG_ADDR_TABLES 0

void
gm_port_state_close (gm_port_state_t * ps)
{
  gm_instance_state_t *is;
  unsigned int port_num;

  GM_CALLED ();

 if (!ps->opened)
   {
    /* this port may have been opened for HW group testing, and may
	have only a DMA region associated with it */

     if (ps->contiguous_dma_region_allocated_len)
       {
	 gm_arch_dma_region_free (&ps->contiguous_dma_region);
	 ps->contiguous_dma_region_allocated_len = 0;
       }
     GM_RETURN_NOTHING ();
   }  

  is = ps->instance;
  gm_assert (is);
  port_num = ps->id;
  is->opener_pids[port_num] = 0;
  ps->closing = 1;

  GM_PRINT (GM_DEBUG_PORT_STATE,
            ("[%d]: Closing port state %p, port_num %d.\n",
             is->id, ps, port_num));

  /* print debugging info */

  {
    unsigned int active;

    active = gm_read_lanai_global_u32 (is, port[port_num].active_subport_cnt);
    if (active)
      {
        GM_NOTE (("There are %d active subports for port %d at close.\n",
		active, port_num));
      }
  }

  /* Release resources allocated since open. */

  gm_disable_port_DMAs (ps);
  if (ps->contiguous_dma_region_allocated_len)
    {
      gm_arch_dma_region_free (&ps->contiguous_dma_region);
      ps->contiguous_dma_region_allocated_len = 0;
    }
  gm_lanai_command_close (ps->instance, ps->id);
#if !GM_CAN_REGISTER_MEMORY
  GM_PRINT (GM_PRINT_LEVEL >= 3, ("about to free copy block.\n"));
  gm_free_copy_block (ps);
#endif

  /* release resources from open. */

  gm_arch_port_state_close (ps);
  is->ps_for_id[ps->id] = 0;
  GM_PRINT (GM_DEBUG_HOST_DMA, ("Freeing receive queue.\n"));
  gm_arch_sync_destroy (&ps->yp_sync);
  gm_arch_sync_destroy (&ps->wait_sync);
  gm_arch_sync_destroy (&ps->sleep_sync);
  if (ps->is_a_kernel_port)
    {
      unsigned int i,active;
      
      for (i=0;i<10;i++)
	{
	  active = 
	    gm_read_lanai_global_u32 (is, port[port_num].active_subport_cnt);
	  if (!active)
	    break;
	  gm_arch_spin(is,100000);
	}
      if (active)
	{
	  GM_NOTE (("Kernel close:  still %d active subports for port %d.\n",
		    active, port_num));
	}
    }
  ps->opened = 0;
  GM_RETURN_NOTHING ();
}

/****************************************************************
 * Thread safe port state acquisition
 ****************************************************************/

/* Return the port state given the instance state and the port number.
   WARNING: This is the unprotected version called from the interrupt
   handler, since mutex's cannot be acquire in the interrupt handler
   of most OS's.  This is safe in the interrupt handler because the
   port closure code ensures that no interrupts will be generated for
   the port before the port is destroyed. */

gm_port_state_t *
__gm_port_state_for_id (gm_instance_state_t * is, unsigned id)
{
  return (id < GM_NUM_ELEM (is->ps_for_id)
	  ? is->ps_for_id[id]
	  : 0);
}

/* Acquire the locked port state identified by IS and PORT.  If you
   get a port state with this call, you must release the port state
   with gm_release_port_state() when you are done with it. */


gm_port_state_t *
gm_acquire_port_state_for_id (gm_instance_state_t * is, unsigned port)
{
  gm_port_state_t *ps;

  gm_always_assert(local_mutex);
  gm_mutex_enter (local_mutex);
  ps = __gm_port_state_for_id (is, port);
  if (!ps)
    goto abort_with_mutex;
  gm_arch_mutex_enter (&ps->sync);
  gm_mutex_exit (local_mutex);
  return ps;

abort_with_mutex:
  gm_mutex_exit (local_mutex);
  return 0;
}

/* Acquire the locked port state identified by INSTANCE and PORT.  If
   you get a port state with this call, you must release the port
   state with gm_release_port_state() when you are done with it. */

gm_port_state_t *
gm_acquire_port_state_for_ids (unsigned int instance, unsigned int port)
{
  gm_instance_state_t *is;
  gm_port_state_t *ps;

  gm_always_assert(local_mutex);
  gm_mutex_enter (local_mutex);
  is = gm_instance_for_id (instance);
  if (!is)
    goto abort_with_mu;
  ps = __gm_port_state_for_id (is, port);
  if (!ps)
    goto abort_with_mu;
  gm_arch_mutex_enter (&ps->sync);
  gm_mutex_exit (local_mutex);
  return ps;

abort_with_mu:
  gm_mutex_exit (local_mutex);
  return 0;
}

/* Release the port state acquired with one of the gm_acquire_port_state*
   functions above. */

void
gm_release_port_state (gm_port_state_t * ps)
{
  gm_arch_mutex_exit (&ps->sync);
}

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
