/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation   			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

#include "gm_internal.h"
#include "gm_ether.h"
#include "gm_debug_lanai_dma.h"

#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/string.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/etherdevice.h>
#include <asm/byteorder.h>
#include <asm/io.h>
#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/inet.h>
#include <net/checksum.h>

#if LINUX_VERSION_CODE >= VERSION_CODE(2,4,6)
#include <linux/mii.h>
#endif

#include "gm_arch_def.h"
#include "gm_arch_pci_map.h"
#include "gm_klog_debug.h"
#include "gm_trace.h"
#include "gm_pio.h"

static int gmip_hw_checksum = 0;
MODULE_PARM (gmip_hw_checksum, "i");

static int gm_net_copy_threshold = 256 + 16;
MODULE_PARM (gm_net_copy_threshold, "i");

void gmip_timeout (struct net_device *dev);
void gmip_set_multicast_list (struct net_device *dev);
int gmip_set_mac_address (struct net_device *dev, void *addr);

#if GM_DEBUG_KLOG
struct gm_klog_struct gm_klog_tab[GM_KLOG_NBEVT];
atomic_t gm_klog_index;
#endif

#define GM_DMA_CUT_AT_PAGE_BOUNDARY 1


#if GM_DEBUG_LANAI_DMA && !GM_DMA_CUT_AT_PAGE_BOUNDARY
#error pagecount maintaince requires a real per page dma_handle
#endif

static void
gm_set_bitmap (gm_instance_state_t * is, struct gm_linux_dma_info *dma_info,
	       int set)
{
  if (GM_DEBUG_LANAI_DMA)
    {
      static unsigned char pagecount[GM_MAX_HOST_DMA_PAGES];
      static spinlock_t bitmap_lock = SPIN_LOCK_UNLOCKED;
      int i;
      long flags;

      /* Here is the trick, there can be several buffer in the same page,
         so we increments the page count, to track this, this work as long
         as the clean becomes before freeing the buffer, and that the
         memory is kmalloced memory so page->count = 1 means not in use
         anymore (except with slab, page->count can be zero) --Loic */

      spin_lock_irqsave(&bitmap_lock, flags);
      for (i = 0; i < dma_info->scatter_cnt; i++)
	{
	  unsigned long phys = virt_to_phys (bus_to_virt(dma_info->dma[i]));
	  struct page *page = GM_LINUX_PHYS_TO_PAGE (is, phys);
	  unsigned long pagenum = page - mem_map;
	  gm_always_assert (pagenum < GM_MAX_HOST_DMA_PAGES);
	  if (set)
	    {
	      GM_BITMAP_WRAPPING_SET (*is->dma_page_bitmap, pagenum);
	      gm_always_assert (atomic_read (&page->count) >= 0);
	      pagecount[pagenum]++;
	    }
	  else
	    {
	      gm_always_assert (atomic_read (&page->count) >= 0);
	      gm_always_assert (pagecount[pagenum] >= 1);
	      pagecount[pagenum]--;
	      if (pagecount[pagenum] == 0)
		{
		  GM_BITMAP_WRAPPING_SET (*is->dma_page_bitmap, pagenum);
		}
	    }
	}
      spin_unlock_irqrestore (&bitmap_lock, flags);
    }
}


/* gmip_build_desc: returns the number of descriptors needed to map
   the buffer */
static int
gmip_build_desc (gm_arch_net_info_t * gmnet, void *data, int len,
		 struct gm_linux_dma_info *dma_info)
{

  int i = 0;
  unsigned long start = (unsigned long) data;

  gm_assert (dma_info->scatter_cnt == 0);

  while (len > 0)
    {
      unsigned long page_end = (start + PAGE_SIZE) & ~(PAGE_SIZE - 1);
      unsigned long offset = start & (PAGE_SIZE -1);
      unsigned long page_start = start - offset;
      unsigned long fraglen;
      gm_status_t map_stat;

      gm_assert (i == 0 || offset == 0);
      gm_always_assert (i < GM_MAX_ETHERNET_SCATTER_CNT);
      gm_always_assert (i < GM_MAX_ETHERNET_GATHER_CNT);

      if (GM_DMA_CUT_AT_PAGE_BOUNDARY)
	  fraglen = page_end > start + len ? len : page_end - start;
      else
	  fraglen = len;

      gm_assert (dma_info->dma[i] == 0);
      map_stat = gm_linux_pci_map (gmnet->is, __pa (page_start), 
				   &(dma_info->dma[i]));
      if ( map_stat == GM_FAILURE)
	{
	  int j;
	  for (j=0;j<i;j++)
	    {
	      gm_linux_pci_unmap( gmnet->is, dma_info->dma[j]);
	      dma_info->dma[j] = 0;
	    }
	  dma_info->scatter_cnt = 0;
	  return 0;
	}
      dma_info->desc[i].len = fraglen;
      /* Moving the following assignment up one line can result in mangled
	 DMA addresses being passed to the firmware on SuSE 8.1 with 
	 gcc 3.2. (ia32)  The host writes 0xabcdefgh and the firmware sees
	 (0xabcdef << 20) | 0xgh  */
      dma_info->desc[i].ptr = dma_info->dma[i] + (gm_dp_t)offset;
      start += fraglen;
      len -= fraglen;
      i += 1;
    }
  gm_assert (len == 0);
  gm_assert (i > 0);
  dma_info->scatter_cnt = i;
  gm_assert (GM_DMA_CUT_AT_PAGE_BOUNDARY || dma_info->scatter_cnt == 1);
  gm_set_bitmap (gmnet->is, dma_info, 1);
  return i;
}

static void
gmip_free_desc (gm_arch_net_info_t * gmnet, 
		struct gm_linux_dma_info *dma_info)
{
  int i;
  gm_set_bitmap (gmnet->is, dma_info, 0);
  for (i=0;i<dma_info->scatter_cnt;i++)
    {
      gm_assert(dma_info->dma[i] != 0);
      gm_linux_pci_unmap (gmnet->is, dma_info->dma[i]);
      dma_info->dma[i] = 0;
    }
  dma_info->scatter_cnt = 0;
}


#if GM_ENABLE_TRACE
unsigned
gm_tcp_window (char *data, int len)
{
#if 0
  if (len >= 56)
    {
      return (ntohs (((unsigned short *) data)[(14 /*eth */  + 20 /*ip */
						+ 14 /*window off */ ) / 2])
	      + (len << 16));
    }
  else
#endif
    return 1 + (len << 16);
}
#endif

/* this initialize the checksum related fields of a skb before passing
   it to net stack */
static void
gmip_skb_checksum(struct sk_buff *skb, gm_u16_t csum, gm_u8_t flags)
{
  if (gmip_hw_checksum)
    {
      skb->ip_summed = CHECKSUM_UNNECESSARY;
      skb->csum = 0;
    }
  else if (flags & GM_IPV4_PARTIAL_CHECKSUM)
    {
      skb->csum = ntohs(csum);
      skb->ip_summed = CHECKSUM_HW;
    }
  else
    {
      skb->ip_summed = CHECKSUM_NONE;
      skb->csum = 0;
    }
}

/* this function pass a sk_buff to the linux network stack */
static void
gmip_netif_rx (struct net_device *dev, struct sk_buff *skb, gm_u16_t csum, 
	       gm_u8_t flags)
{
  gmip_skb_checksum(skb, csum, flags);
  skb->protocol = eth_type_trans (skb, dev);
  skb->dev = dev;
  netif_rx (skb);
  dev->last_rx = jiffies;
}

static void
gmip_recv_interrupt (void *gmnetp, gm_ethernet_recv_t *e)
{
  gm_arch_net_info_t *gmnet = gmnetp;
  struct net_device *dev = gmnet->dev;
  const int len = gm_ntoh_u32 (e->length);
  gm_u16_t csum = gm_ntoh_u16 (e->ip_checksum);
  gm_u8_t index = gm_ntoh_u8 (e->tag);
  gm_u8_t flags = gm_ntoh_u8 (e->flags);
  struct gm_linux_ether_pkt *pkt = gmnet->recv_ring+index;
  struct sk_buff *skb = pkt->skb;
  const unsigned int ulen = gm_ntoh_u32 (e->length);

  gm_always_assert (skb);

  GM_LOG_STAMPED_EVT (GM_ETHER_RCV, gm_tick (), 0,
		      gm_tcp_window ((char *) skb->data, ulen));

  GM_PRINT (GM_PRINT_LEVEL >= 5,
	    ("gmip_recv_interrupt(%p, len=%d)\n", gmnetp, ulen));

  if (len < ETH_HLEN)
    {
      GM_WARN (("Hmm, received invalid packet: len=%d\n", len));
      gmnet->stats.rx_dropped += 1;
      goto end;
    }
  if (len > skb_tailroom (skb))
    {
      GM_PANIC (("Received ethernet packet bigger than buffer: "
		 "len=%d,tailroom=%d\n", len, skb_tailroom (skb)));
    }

  /* we do not modify the pointers inside the skbuff in case we want
     to reuse it as is, do not use skb_pull or skb_put or skb_push
     until we are sure the buffer will not be put again into the recv
     ring */

  if (ulen <= sizeof (struct ethhdr))
    {
      struct ethhdr *eth = (struct ethhdr *) skb->data;
      GM_VAR_MAY_BE_UNUSED (eth);
      GM_WARN (("%s:bad eth pkt rcvd, len= %d, src:%x:%x:%x:%x:%x:%x "
		"dst: %x:%x:%x:%x:%x:%x proto=0x%x\n", gmnet->dev->name, len,
		eth->h_source[0], eth->h_source[1], eth->h_source[2],
		eth->h_source[3], eth->h_source[4], eth->h_source[5],
		eth->h_dest[0], eth->h_dest[1], eth->h_dest[2],
		eth->h_dest[3], eth->h_dest[4], eth->h_dest[5],
		ntohs (eth->h_proto)));
      gmnet->stats.rx_dropped += 1;
      goto end;
    }

  /* 
     after the two cases, pkt->skb should be set to the new pristine buffer 
     to be put in the ring
   */
  if (len < gm_net_copy_threshold)
    {
      /* small packet, better copy that waste our current full-sized buffer */
      struct sk_buff *new = dev_alloc_skb (ulen + 2);
      if (!new)
	{
	  gmnet->stats.rx_dropped += 1;
	  goto end;
	}
      /* misalign 14 byte ethernet header so that IP header ends up on
	 a 4 byte boundary */
      skb_reserve (new, 2);
      memcpy (skb_put (new, ulen), skb->data, ulen);
      gmip_netif_rx (dev, new, csum, flags);
      /* we leave the old skb in the ring */
    }
  else
    {
      /* above threeshold, better waste unused part of the buffer that
         do a memcpy to a smaller one */
      struct gm_linux_dma_info dinfo_new;
      struct sk_buff *new;
      
      memset (&dinfo_new, 0, sizeof(dinfo_new));
      new = dev_alloc_skb (GM_ETHERNET_MTU);
      if (!new)
	{
	  gmnet->stats.rx_dropped += 1;
	  goto end;
	}

      /* misalign 14 byte ethernet header so that IP header ends up on a 4
	 byte boundary */
      skb_reserve (new, 2);

      if (!gmip_build_desc (gmnet, new->data, skb_tailroom (new), &dinfo_new))
	{
	  GM_WARN (("%s:%d  gmip_build_desc failed\n", __FILE__, __LINE__));
	  dev_kfree_skb_irq (new);
	  new = (struct sk_buff *) 0;
	  goto end;
	}
      gmip_free_desc (gmnet, &pkt->dma_info);
      /* set how much of the skb was filled */
      skb_put (skb, ulen);
      gmip_netif_rx (dev, skb, csum, flags);

      memcpy(&pkt->dma_info, &dinfo_new, sizeof(dinfo_new));
      GM_PRINT (GM_PRINT_LEVEL >= 5,
		("%s: received packet with protocol, "
		 "passed to netif_rx:0x%x,type=%d,len=%d\n",
		 dev->name, ntohs (skb->protocol), skb->pkt_type, skb->len));
      GM_KLOG_EVT (GM_KLOG_RECV, skb);
      /* now we put in the ring the newly allocated buffer */
      pkt->skb = new;
    }
  gmnet->stats.rx_packets += 1;
  gmnet->stats.rx_bytes += ulen;

 end:

  GM_KLOG_EVT (GM_KLOG_PROVIDE, pkt->skb);
  GM_PRINT (GM_PRINT_LEVEL >= 5, ("receive dma ptr = 0x%08x\n", 
				  pkt->dma_info.desc[0].ptr));
  gm_ethernet_provide_scatter_list_with_tag (gmnet->port,
					     pkt->dma_info.scatter_cnt,
					     pkt->dma_info.desc,
					     index);
  gmnet->rdone += 1;
}

static void
gmip_sent_interrupt (void *gmnetp)
{
  int index;
  gm_arch_net_info_t *gmnet = gmnetp;
  struct gm_linux_ether_pkt *pkt;
  index = gmnet->sdone_index;

  GM_PRINT (GM_PRINT_LEVEL >= 5,
	    ("gmip_sent_interrupt(%p) index = %d\n", gmnetp, index));

  if (gmnet->sreq - gmnet->sdone == 0)
    {
      GM_WARN (("%s: Internal error: SENT_EVENT but no skb "
		"in send_ring!??,sreq=%u,sdone=%u\n",
		gmnet->dev->name, gmnet->sreq, gmnet->sdone));
      return;
    }
  pkt = gmnet->send_ring + index;
  gm_always_assert (pkt->skb);

  gmip_free_desc (gmnet, &pkt->dma_info);

  /* the original skb has not been touched */
  GM_LOG_STAMPED_EVT (GM_ETHER_SINTR, gm_tick (), 0,
		      gm_tcp_window (skb->data, skb->len));

  gmnet->stats.tx_bytes += pkt->skb->len;
  dev_kfree_skb_irq (pkt->skb);
  pkt->skb = 0;
  gmnet->sdone += 1;
  gmnet->sdone_index += 1;
  if (gmnet->sdone_index > GM_SEND_RING_MAX_INDEX)
    {
      gmnet->sdone_index = 0;
    }

  gmnet->stats.tx_packets += 1;
  if (netif_queue_stopped (gmnet->dev) 
      && gmnet->sreq - gmnet->sdone <= GM_SEND_RING / 4)
    {
      netif_wake_queue (gmnet->dev);
    }
}

static int
gmip_open (struct net_device *dev)
{
  gm_arch_net_info_t *gmnet = dev->priv;
  int status, i;
  int err;

  MOD_INC_USE_COUNT;

  status = gm_open (&gmnet->port, gmnet->is->id, GM_ETHERNET_PORT_ID,
		    "ethernet", (enum gm_api_version) GM_API_VERSION);
  /* myrip->gm_port = gmk_open(unit, GM_ETHERNET_PORT_ID, 
     &myrip->gm_device); */
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("%s:error opening GM port,status=%d\n", dev->name, status));
      err = -EIO;
      goto abort_with_nothing;
    }

  /* request an interupt for every 6 packets we send.  This is
     emperically determined & may need tuning -- gallatin */
  status = gm_enable_ethernet (gmnet->port, GM_RECV_RING, 6);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("%s: error enabling ethernet,status=%d\n", dev->name, 
		status));
      status = -EIO;
      goto abort_with_open_port;
    }
  
  gmnet->sreq = 0;
  gmnet->sdone = 0;
  gmnet->rdone = 0;
  gmnet->dead = 0;
  gmnet->sdone_index = 0;
  gmnet->sreq_index = 0;

  gm_always_assert (GM_NUM_SEND_TOKENS >= GM_SEND_RING);
  gm_always_assert (GM_NUM_ETHERNET_RECV_TOKENS >= GM_RECV_RING);

  dev_start (dev, 1);

  for (i = 0; i < GM_RECV_RING; i++)
    {
      struct sk_buff *skb;
      struct gm_linux_ether_pkt *pkt = gmnet->recv_ring + i;
      gm_always_assert (pkt->skb == 0 &&  pkt->dma_info.scatter_cnt == 0);
      skb = dev_alloc_skb (GM_ETHERNET_MTU);
      if (!skb)
	{
	  err = -ENOMEM;
	  goto abort_with_open_port;
	}

      /* misalign 14 byte ethernet header so that IP header ends up on
	 a 4 byte boundary */
      skb_reserve (skb, 2);

      pkt->skb = skb;
      if (gmip_build_desc (gmnet, skb->data, skb_tailroom (skb), 
			   &pkt->dma_info) == 0)
	{
	  err = -ENOMEM;
	  goto abort_with_open_port;
	}
    }
  
  gm_pause_lanai(gmnet->is);
  /* loic: setting the callback must be done before any interrupt can
     be delivered calls to gm_ethernet_provide_scatter_list must be
     serialized, so we will disable interrupts while filling the ring,
     the important thing is to enable the callback before releasing
     the spinlock.  I think there was a potential race condition at
     open time in my old code.  */
  gm_ethernet_set_recv_intr_callback (gmnet->port, gmip_recv_interrupt,
				      gmnet);
  gm_ethernet_set_sent_intr_callback (gmnet->port, gmip_sent_interrupt,
				      gmnet);

  for (i=0; i < GM_RECV_RING; i++)
    {
      struct gm_linux_ether_pkt *pkt = gmnet->recv_ring + i;
      GM_PRINT (GM_PRINT_LEVEL >= 5,
		("GM-IP: receive dma ptr = 0x%08x\n", pkt->dma_info.desc[0].ptr));
      gm_ethernet_provide_scatter_list_with_tag (gmnet->port, 
						 pkt->dma_info.scatter_cnt,
						 pkt->dma_info.desc, i);
    }
  gm_unpause_lanai(gmnet->is);

  gm_tracing (gmnet->port, 1);

  if (gmip_hw_checksum)
    {
      GM_INFO (("GM IP device disabling linux software IP checksums.\n"));
    }

  netif_start_queue (dev);

  return 0;

 abort_with_open_port:
  for (i = 0; i < GM_RECV_RING; i++)
    {
      struct gm_linux_ether_pkt *pkt = gmnet->recv_ring + i;
      if (pkt->dma_info.scatter_cnt)
	gmip_free_desc(gmnet, &pkt->dma_info);
      if (pkt->skb)
	{
	  dev_kfree_skb (pkt->skb);
	  pkt->skb = 0;
	}
    }
  gm_close (gmnet->port);
 abort_with_nothing:
  dev_start (dev, 0);
  MOD_DEC_USE_COUNT;
  return err;
}

static int
gmip_close (struct net_device *dev)
{
  int i;
  gm_arch_net_info_t *gmnet = (gm_arch_net_info_t *) dev->priv;

  gm_disable_ethernet (gmnet->port);
  gm_close (gmnet->port);
  gmnet->port = 0;
  dev_start (dev, 0);
  /* code looks strange but overflow on gmnet->sreq and 
     gmnet->sdone should handled correctly */
  for (i = 0; i < gmnet->sreq - gmnet->sdone; i++)
    {
      struct gm_linux_ether_pkt *pkt;

      pkt = gmnet->send_ring + gmnet->sdone_index;
      gmnet->sdone_index += 1;
      if (gmnet->sdone_index > GM_SEND_RING_MAX_INDEX)
	{
	  gmnet->sdone_index = 0;
	}
      gm_always_assert (pkt->skb && pkt->dma_info.scatter_cnt);
      gmip_free_desc (gmnet, &pkt->dma_info);
      dev_kfree_skb (pkt->skb);
      pkt->skb = 0;
      gmnet->stats.tx_dropped += 1;
    }
  for (i = 0; i < GM_RECV_RING; i++)
    {
      struct gm_linux_ether_pkt *pkt = gmnet->recv_ring + i;
      gm_always_assert (pkt->skb && pkt->dma_info.scatter_cnt);
      gmip_free_desc (gmnet, &pkt->dma_info);
      dev_kfree_skb (pkt->skb);
      pkt->skb = 0;
    }

  MOD_DEC_USE_COUNT;
  return 0;
}

static int gm_max_diff_ring = 0;

static int
gmip_xmit (struct sk_buff *skb, struct net_device *dev)
{
  gm_arch_net_info_t *gmnet = dev->priv;
  struct ethhdr *eth = (struct ethhdr *) skb->data;
  void *data;
  int len, index;
#if GM_USE_INTR_LOCK
  long flags;
#endif
  unsigned long spin_flags = 0;
  struct gm_linux_ether_pkt *pkt;


  if (!gmnet->is->lanai.running)
    goto drop;

  if (GM_USE_INTR_LOCK)
     {
       spin_lock_irqsave (&gm_linux_intr_lock, spin_flags);
     }
  
#if LINUX_22
  /* before 2.4, timeout notification was done by calling us even
     tough the tbusy field was set, we have to call the gmip_timeout
     procedure here. */
  /* Most 2.2 network drivers are setting tbusy during xmit (to protect
     against reentrancy?), but packet sending is serialized in 2.2.
     (and GM_USE_INTR_LOCK adds a second redundant layer of
     protection), so we just use tbusy to emulate the 2.4 control flow */
  if (dev->tbusy != 0)
    {
      if (jiffies - dev->trans_start > HZ * 10 && !gmnet->dead)
	{
	  gmip_timeout (dev);
	  dev->trans_start = jiffies;
	  gmnet->dead = 1;
	}
      if (GM_USE_INTR_LOCK)
	 {
	    spin_unlock_irqrestore (&gm_linux_intr_lock, spin_flags);
	 }
      return 1;
    }
#endif


  if (gmnet->sreq - gmnet->sdone >= GM_SEND_RING)
    {
      /* should never happen because of the driver/kernel control flow
         interface */
      GM_NOTE(("Send queue overflowed\n"));
      goto drop_busy;
    }
  if (skb->len > dev->mtu + ETH_HLEN)
    {
      /* loic: we do not count the myritype, because it is hidden from
	 upper layers in the net_device struct hard_header_len = 14,
	 myritype is not counted */
      GM_NOTE (("%s,oops , skbuff len (%d) > mtu+eth-headers(%d)!!!??\n",
		gmnet->dev->name, skb->len, dev->mtu + ETH_HLEN));
      goto drop;
    }

  pkt = gmnet->send_ring + gmnet->sreq_index;

  gm_always_assert (pkt->skb == 0);
  data = skb->data;
  len =  skb->len;
  if (gmip_build_desc (gmnet, data, len, &pkt->dma_info) == 0)
    {
      GM_WARN (("%s:%d  gmip_build_desc failed for send packet\n",
		__FILE__, __LINE__));
      goto drop;
    }
  GM_PRINT (GM_PRINT_LEVEL > 5,
	    ("%s:send eth pkt %p, proto=0x%x, len = %d numdesc = %d \n",
	     gmnet->dev->name, skb->data,
	     ntohs (skb->protocol), len,
	     pkt->dma_info.scatter_cnt));
#if (GM_PRINT_LEVEL >= 5)
  gm_hex_dump (data, len);
#endif

  /* update the send_ring before actually sending so the interrupt handler */
  /* is sure to have up-to-date send_ring info */
  dev->trans_start = jiffies;
  pkt->skb = skb;
  /* no need to be atomic, we are the only writer, 
     and there is write to LANAI + interrupt before we are read*/
  gmnet->sreq += 1;
  gmnet->sreq_index += 1;
  if (gmnet->sreq_index > GM_SEND_RING_MAX_INDEX)
    {
      gmnet->sreq_index = 0;
    }

  /* when tracing, get in the aux field the len and the tcp window */
  GM_LOG_STAMPED_EVT (GM_ETHER_SND, gm_tick (), 0,
		      gm_tcp_window (data, len));

  if ((gmnet->sreq - gmnet->sdone) > gm_max_diff_ring)
    {
      gm_max_diff_ring = (gmnet->sreq - gmnet->sdone);
    }
  /* loic : I have moved the check before the real send to avoid any
     race condition (where the sent_interrupt handler would run
     between the check and the netif_stop_queue), necessary now that
     we check in sent_interrupt that the queue was stopped before
     restarting it (rather than relying on the fact the number of
     entries in the ring is > 1, and always restarting it) */
  if ((gmnet->sreq - gmnet->sdone) >= GM_SEND_RING)
    {
      netif_stop_queue (dev);
    }
  if (eth->h_dest[0] & 0x01)
    {
      GM_ARCH_LOCK_ETHER_SEND_QUEUE (gmnet->port);
      gm_ethernet_broadcast (gmnet->port, pkt->dma_info.scatter_cnt, 
			     pkt->dma_info.desc);
      GM_ARCH_UNLOCK_ETHER_SEND_QUEUE (gmnet->port);
    }
  else
    {
      GM_ARCH_LOCK_ETHER_SEND_QUEUE (gmnet->port);
      gm_ethernet_send (gmnet->port, pkt->dma_info.scatter_cnt, 
			pkt->dma_info.desc, eth->h_dest);
      GM_ARCH_UNLOCK_ETHER_SEND_QUEUE (gmnet->port);
    }

  GM_PRINT (GM_PRINT_LEVEL >= 5,
	    ("gmip_xmit - sent a packet sreq=%d sdone=%d\n",
	     gmnet->sreq, gmnet->sdone));


  if (GM_USE_INTR_LOCK)
    {
      spin_unlock_irqrestore (&gm_linux_intr_lock, spin_flags);
    }
  GM_KLOG_EVT (GM_KLOG_SENDEND, skb);
  return 0;

 drop:
  /* the xmit procedure should only return a failed error code, if the
     kernel should requeue the packet (to recompute route for
     non-ethernet devices for instances). The only case here we fail
     is resource shortage, so we record the dropped packet, but we
     still return success which means we took responsability for the
     skb (we sent it we success to a blackhole :-) */
 drop_busy:
  gmnet->stats.tx_dropped += 1;
  dev_kfree_skb_any (skb);
  if (GM_USE_INTR_LOCK)
    {
      spin_unlock_irqrestore (&gm_linux_intr_lock, spin_flags);
    }
  return 0;
}


static struct net_device_stats *
gmip_get_stats (struct net_device *dev)
{
  gm_arch_net_info_t *gmnet = dev->priv;
  return &gmnet->stats;
}

static int
gmip_dev_init (struct net_device *dev)
{
  GM_PRINT (GM_PRINT_LEVEL >= 5, 
	    ("Initialized network driver %s\n", dev->name));
  return 0;
}

static int
gmip_change_mtu (struct net_device *dev, int new_mtu)
{
  if ((new_mtu < 68) || (ETH_HLEN + new_mtu > GM_ETHERNET_MTU))
    {
      GM_NOTE (("new_mtu (%d) is not valid\n", new_mtu));
      return -EINVAL;
    }
  GM_INFO (("changing mtu from %d to %d\n", dev->mtu, new_mtu));
  dev->mtu = new_mtu;
  return 0;
}

static int
gmip_ioctl (struct net_device *dev, struct ifreq *ifr, int cmd)
{
  struct mii_ioctl_data *data = (struct mii_ioctl_data *) &ifr->ifr_data;
  gm_arch_net_info_t *gmnet = dev->priv;
  gm_instance_state_t *is = gmnet->is;

  /* for now we only support a fake MII interface for l9 cards */
  if (!gm_is_l9(is))
    return -EOPNOTSUPP;
  
  switch (cmd)
    {
#if LINUX_VERSION_CODE >= VERSION_CODE(2,4,6)
    case SIOCGMIIPHY: 
    case SIOCDEVPRIVATE:
    case SIOCGMIIREG: 
    case SIOCDEVPRIVATE+1:
      {
	
	/* according to Donald Becker's website, MII devices have 32 management
	   registers, the first 16 are reserved for standard defined uses, and
	   the remaining ones are available for chip-specific features.  Only
	   the first seven registers are currently defined */
	
	unsigned int isr;
	int have_link;
	
	/* get the link status. bit 16 of the ISR is the LINK_DOWN 
	   bit for L9 */
	
	isr = is->get_ISR(is);
	have_link = !(isr & (1<<16));
	
	data->val_out = 0;
	switch (data->reg_num & 0x1f)
	  {
	  case 0: /* basic mode configuration */
	    data->val_out = (BMCR_ANENABLE);
	    break;
	  case 1: /* basic mode status */
	    data->val_out = (BMSR_100FULL | BMSR_ANEGCAPABLE);
	    if (have_link) 
	      data->val_out |= (BMSR_LSTATUS | BMSR_ANEGCOMPLETE);
	    break;
	  case 2: /* phy id 1 */
	  case 3: /* phy id 2 */
	    break;
	  case 4: /* autonegotiation advertisement */
	    data->val_out = ( ADVERTISE_100FULL );
	    break;
	  case 5: /* link partner ability */
	    if (have_link) 
	      data->val_out = ( LPA_100FULL | LPA_LPACK );
	    break;
	  case 6: /* autonegotiation expansion */
	    /* to be filled in when I find info on it */
	    break;
	  default:
	    break;
	  }
      }
      break;
#endif /* LINUX_VERSION >= 2.4.6 */
      
    default:
      return -EOPNOTSUPP;
      break;
    }
      
  return 0;
}

void
gmip_timeout (struct net_device *dev)
{
  gm_arch_net_info_t *gmnet = dev->priv;
  gm_instance_state_t *is = gmnet->is;
  GM_WARN (("myri%d: gmip_timeout called (lanai.running=%d)!\n",
	    is->id, is->lanai.running));
  if (is->lanai.running)
    {
      _GM_WARN (("status info isr=0x%x eimr=0x%x,globals.interrupt.type=%d\n",
		 is->get_ISR (is), is->get_EIMR (is),
		 gm_read_lanai_global_u32 (is, interrupt.type)));
      schedule_task (&is->arch.test_lanai_task);
    }
}

int
gmip_set_mac_address (struct net_device *dev, void *addr)
{
  struct sockaddr *sa = (struct sockaddr *) addr;
  gm_arch_net_info_t *gmnet = dev->priv;

  /* change the dev structure */
  gm_bcopy(sa->sa_data, dev->dev_addr, 6);

  /* set address on the Lanai */
  gm_set_ethernet_mac_address(gmnet->is, sa->sa_data);
  return 0;
}

void
gmip_set_multicast_list (struct net_device *dev)
{
  gm_arch_net_info_t *gmnet = dev->priv;

  if (dev->flags & IFF_PROMISC)
    {
      gmnet->mcp_flags |= GM_ETHER_PROMISC;
    }
  else
    {
      gmnet->mcp_flags &= ~GM_ETHER_PROMISC;
    }

  gm_set_ethernet_flags (gmnet->is, gmnet->mcp_flags);
}

/* we do not count the 2 bytes of
   myrinet type in the dev->hard_header_len, */

int
gmip_init (gm_instance_state_t * is)
{
  struct net_device *dev;
  int i, err;

  dev = gm_linux_dev_alloc ("myri%d", &err, is->id);
  if (!dev)
    {
      return err;
    }

  ether_setup (dev);
  /* currently take the smallest page_size of all linux system 
     to compute the same MTU everywhere */
  dev->mtu = GM_IP_MTU;
  /* loic : not sure if upper layers include the ETH_HLEN in the MTU, 
     but assuming not is always safe */
  gm_always_assert (ETH_HLEN + dev->mtu <= GM_ETHERNET_MTU);
  dev->open = gmip_open;
  dev->stop = gmip_close;
  dev->hard_start_xmit = gmip_xmit;
  dev->get_stats = gmip_get_stats;
  dev->base_addr = is->arch.iomem_base;
  dev->irq = is->arch.irq;
  dev->init = gmip_dev_init;
  dev->change_mtu = gmip_change_mtu;
  dev->set_multicast_list = gmip_set_multicast_list;
  dev->set_mac_address = gmip_set_mac_address;
#if LINUX_24
  dev->tx_timeout = gmip_timeout;
  dev->watchdog_timeo = HZ * 25;
  dev->do_ioctl = gmip_ioctl;
#endif
  gm_read_lanai_global_byte_array(is,
				  ethernet.mac_address,
				  dev->dev_addr);

  gm_always_assert (dev->hard_header_len == 14);

  is->arch.net.dev = dev;
  is->arch.net.is = is;
  (gm_arch_net_info_t *) dev->priv = &is->arch.net;
  if (register_netdev (dev))
    {
      GM_NOTE (("%s:register_netdev failed!!\n", dev->name));
      kfree (dev);
      return -EIO;
    }
  return 0;
}

void
gmip_finalize (gm_instance_state_t * is)
{
  struct net_device *dev = is->arch.net.dev;
  unregister_netdev (dev);
  kfree (dev);
}


/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
