/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

/************************************************************************
 * IRQL level summary
 ************************************************************************/

/* In this driver, as few functions as possible run at elevated IRQ levels.
   Specifically, the functions run at the following levels:

   DIRQL_LEVEL:		gm_nt_Isr() (the interrupt handler)
   DISPATCH_LEVEL:	gm_nt_Isr_dpc() (the interrupt handler delayed
				proceedure call)
			gm_nt_cancel_irp() (the NT IRP cancellation function)
   			_gm_maybe_cancel_sync() (a helper function for
				gm_nt_cancel_irp()
   APC_LEVEL:		(none)
   PASSIVE_LEVEL:	(all others) */

#include <ntddk.h>
#include <ndis.h>
#include <limits.h>
#include <stdarg.h>

#include "gm_arch.h"
#include "gm_call_trace.h"
#include "gm_event_log.h"
#include "gm_pio.h"

#include "gmm.h"

#if !GM_CAN_REGISTER_MEMORY
#  error GM_CAN_REGISTER_MEMORY not defined in NT gm_arch.c
#endif

#if GM_ENABLE_ETHERNET
#  include "gm_ether.h"
#endif

/****************************************************************
 * NDIS layering support
 ****************************************************************/

/* entry points cached from the NDIS driver */

static struct
{
  PDRIVER_DISPATCH create;
  PDRIVER_DISPATCH close;
  PDRIVER_DISPATCH cleanup;
  PDRIVER_DISPATCH device_control;
  PDRIVER_DISPATCH pnp;
  PDRIVER_UNLOAD unload;
}
ndis;

struct gm_hash *gm_device_hash;	/* references all GM devices */
struct gm_hash *gm_nt_phys_hash; /* references all locked memory */

int gm_usage_count = 0;

static gm_inline int
is_gm_device (PDEVICE_OBJECT d)
{
  return gm_hash_find (gm_device_hash, d) == d;
}

/************************************************************************
 * NT Device Extension support
 ************************************************************************/

/* These structures allow clean support for the different types of
   device extensions required in the GM NT driver. */

enum gm_nt_device_extension_type
{
  GM_NT_UNKNOWN_DEVICE_EXTENSION,
  GM_NT_MINOR_DEVICE_EXTENSION,
  GM_NT_RESOURCE_DEVICE_EXTENSION
};

union gm_nt_device_extension
{
  /* used for minor devices */
  struct gm_nt_minor_device_extension
  {
    enum gm_nt_device_type type;
    gm_arch_minor_t minor;
  }
  minor;
};

/************
 * Extension extraction functions
 ************/

static gm_inline union gm_nt_device_extension *
gm_nt_device_extension (PDEVICE_OBJECT dev)
{
  return (union gm_nt_device_extension *) dev->DeviceExtension;
}

/************
 * Extension type extraction function
 ************/

static gm_inline enum gm_nt_device_extension_type
gm_nt_device_extension_type (PDEVICE_OBJECT dev)
{
  return gm_nt_device_extension (dev)->minor.type;
}

/************************************************************************
 * NT utility functions for managing minor nodes.
 ************************************************************************

 These gm_nt_* functions come near the top so they don't need
 forward declarations to be used in gm_arch_* functions below.  The
 gm_arch_* functions below have prototypes in gm_impl.h, so they can
 safely be anywhere in this file. */

/* Return the minor device number corresponding to a device. */

static gm_arch_minor_t
gm_nt_device_to_minor (PDEVICE_OBJECT dev)
{
  gm_assert (gm_nt_device_extension_type (dev)
	     == GM_NT_MINOR_DEVICE_EXTENSION);
  return (gm_nt_device_extension (dev)->minor.minor);
}

/* Return the port state corresponding to a device. */

static gm_port_state_t *
gm_nt_device_to_ps (PDEVICE_OBJECT dev)
{
  return gm_minor_get_port_state (gm_nt_device_to_minor (dev));
}

/************
 * Minor node management
 ************/

/* Create a new minor node, mimicing the Unix functionality as closely
   as possible.  The minor node includes the NT named device and an
   associated gm_port_state.  However, the gm_port_state created is
   only partially initialized.  Clone devices will never be fully
   initialized, and non-clone devices will be later initialized
   with gm_open_port_state(). */

gm_status_t
gm_nt_create_minor_node (gm_instance_state_t * is,
			 PCWSTR _basename,
			 gm_arch_minor_t * minorp,
			 int is_clone_device,
			 int is_priv, unsigned long port_id)
{
  NTSTATUS nt_status;
  gm_arch_minor_t minor;
  gm_port_state_t *ps;

  UNICODE_STRING basename;

  UNICODE_STRING unit;
  WCHAR wchar_unit[32];

  UNICODE_STRING port;
  WCHAR wchar_port[32];

  if (!minorp)
    {
      GM_NOTE (("internal error: minorp==0\n"));
      goto abort_with_nothing;
    }

  if (port_id >= GM_NUM_PORTS)
    {
      GM_NOTE (("User tried to set port number out of range.\n"));
      goto abort_with_nothing;
    }

  /* Allocate a minor number for the new device. */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Allocating a minor node.\n"));
  if (gm_minor_alloc (is, &minor) != GM_SUCCESS)
    {
      GM_NOTE (("Failed to allocate minor number.\n"));
      goto abort_with_nothing;
    }
  ps = gm_minor_get_port_state (minor);
  gm_assert (ps);
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("minor<->ps = %d,%p.\n", minor, ps));
  ps->privileged = is_priv;
  ps->arch.is_clone_device = is_clone_device;
  if (!is_clone_device)
    ps->id = port_id;


  /* init scratch unicode strings */

  unit.Length = 0;
  unit.MaximumLength = 32;
  unit.Buffer = wchar_unit;

  port.Length = 0;
  port.MaximumLength = 32;
  port.Buffer = wchar_port;

  RtlInitUnicodeString (&basename, _basename);

  /* Convert ints to strings */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Converting ints to strings.\n"));
  nt_status = RtlIntegerToUnicodeString (is->id, 10, &unit);
  if (NT_ERROR (nt_status))
    {
      GM_WARN (("Can't convert integer to Unicode\n"));
      goto abort_with_minor;
    }

  nt_status = RtlIntegerToUnicodeString (ps->id, 10, &port);
  if (NT_ERROR (nt_status))
    {
      GM_WARN (("Can't convert integer to Unicode\n"));
      goto abort_with_minor;
    }

  /* Initialize storage for device name with at least enough storage */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Initializing storage for device name.\n"));
  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("basename.Length = %d.\n", basename.Length));
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("unit.Length = %d.\n", unit.Length));
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("port.Length = %d.\n", port.Length));

  RtlInitUnicodeString (&ps->arch.nt_device_name, NULL);
  gm_assert (ps->arch.nt_device_name.Length == 0);
  ps->arch.nt_device_name.MaximumLength
    = ( /*BAD*/ 2 * 15 * sizeof (WCHAR) + basename.Length + unit.Length
       + port.Length);
  ps->arch.nt_device_name.Buffer
    = gm_malloc (ps->arch.nt_device_name.MaximumLength);
  if (ps->arch.nt_device_name.Buffer == NULL)
    {
      GM_WARN (("Can't allocate memory for unicode string\n"));
      nt_status = STATUS_INSUFFICIENT_RESOURCES;
      goto abort_with_minor;
    }

  RtlInitUnicodeString (&ps->arch.win32_device_name, NULL);
  gm_assert (ps->arch.win32_device_name.Length == 0);
  ps->arch.win32_device_name.MaximumLength
    = ( /*BAD*/ 2 * 15 * sizeof (WCHAR) + basename.Length + unit.Length
       + port.Length);
  ps->arch.win32_device_name.Buffer
    = gm_malloc (ps->arch.win32_device_name.MaximumLength);
  if (ps->arch.win32_device_name.Buffer == NULL)
    {
      GM_WARN (("Can't allocate memory for unicode string\n"));
      nt_status = STATUS_INSUFFICIENT_RESOURCES;
      goto abort_with_nt_buffer;
    }

  /* Format the device name into the allocated buffers */

  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("Formatting device names with basename \"%S\".\n",
	     basename.Buffer));
  RtlAppendUnicodeToString (&ps->arch.nt_device_name, L"\\Device\\");
  RtlAppendUnicodeStringToString (&ps->arch.nt_device_name, &basename);
  RtlAppendUnicodeStringToString (&ps->arch.nt_device_name, &unit);
  if (!ps->arch.is_clone_device)
    {
      RtlAppendUnicodeToString (&ps->arch.nt_device_name, L"-");
      RtlAppendUnicodeStringToString (&ps->arch.nt_device_name, &port);
    }
  GM_PRINT (GM_PRINT_LEVEL >= 2, ("NT device name is \"%S.\"\n",
				  ps->arch.nt_device_name.Buffer));

  RtlAppendUnicodeToString (&ps->arch.win32_device_name, L"\\??\\");
  RtlAppendUnicodeStringToString (&ps->arch.win32_device_name, &basename);
  RtlAppendUnicodeStringToString (&ps->arch.win32_device_name, &unit);
  if (!ps->arch.is_clone_device)
    {
      RtlAppendUnicodeToString (&ps->arch.win32_device_name, L"-");
      RtlAppendUnicodeStringToString (&ps->arch.win32_device_name, &port);
    }
  GM_PRINT (GM_PRINT_LEVEL >= 2, ("WIN32 device name is \"%S.\"\n",
				  ps->arch.win32_device_name.Buffer));

  /* Create a device with that name and minor device number. */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Creating the minor device for port %d of "
				  "instance %d.\n", ps->id, is->id));
  GM_PRINT (GM_PRINT_LEVEL >= 5,
	    ("Calling IoCreateDevice (%p, %d, %p, %d, %d, %d, %p)\n",
	     is->arch.driver_object,
	     sizeof (struct gm_nt_minor_device_extension),
	     &ps->arch.nt_device_name, FILE_DEVICE_PHYSICAL_NETCARD, 0, FALSE,
	     &ps->arch.device_object));
  nt_status =
    IoCreateDevice (is->arch.driver_object,
		    sizeof (struct gm_nt_minor_device_extension),
		    &ps->arch.nt_device_name, FILE_DEVICE_PHYSICAL_NETCARD, 0,
		    FALSE, &ps->arch.device_object);
  if (NT_ERROR (nt_status))
    {
      GM_NOTE (("Could not create minor device (error 0x%x).\n", nt_status));
      GM_NOTE (("WIN32 device name was \"%S.\"\n",
		ps->arch.win32_device_name.Buffer));

      goto abort_with_win32_buffer;
    }
  gm_assert (ps->arch.device_object);

  /* record that we created this device */

  if (gm_hash_insert
      (gm_device_hash, ps->arch.device_object,
       ps->arch.device_object) != GM_SUCCESS)
    {
      GM_NOTE (("Could not remember device is ours.\n"));
      goto abort_with_device;
    }

  nt_status = IoCreateSymbolicLink (&ps->arch.win32_device_name,
				    &ps->arch.nt_device_name);
  if (NT_ERROR (nt_status))
    {
      GM_NOTE (("Could not create WIN32 subsystem symbolic link.\n"));
      goto abort_with_hash_entry;
    }

  /* Fill the device extension */

  GM_PRINT (GM_PRINT_LEVEL >= 5, ("Filling device extension.\n"));
  gm_nt_device_extension (ps->arch.device_object)->minor.type
    = GM_NT_MINOR_DEVICE_EXTENSION;
  gm_nt_device_extension (ps->arch.device_object)->minor.minor = minor;
  /* If this is a dynamically created device, clear the initializing
     flag so that client applications can open it. */

  /* Under Win2K, all devices get dynamically created. */
  ps->arch.device_object->Flags &= ~DO_DEVICE_INITIALIZING;

  GM_PRINT (GM_PRINT_LEVEL >= 5, ("Returning the minor device number.\n"));
  *minorp = minor;
  return GM_SUCCESS;

abort_with_hash_entry:
  gm_hash_remove (gm_device_hash, ps->arch.device_object);
abort_with_device:
  IoDeleteDevice (ps->arch.device_object);
abort_with_win32_buffer:
  gm_free (ps->arch.win32_device_name.Buffer);
abort_with_nt_buffer:
  gm_free (ps->arch.nt_device_name.Buffer);
abort_with_minor:
  gm_minor_free (minor);
abort_with_nothing:
  return GM_FAILURE;
}

/* Destroy a minor node, including the NT device and associated
   gm_port_state. */

void
gm_nt_destroy_minor_node (gm_arch_minor_t minor)
{
  gm_port_state_t *ps;

  ps = gm_minor_get_port_state (minor);
  if (!ps)
    {
      GM_WARN (("Tried to destroy nonexistant minor node %d.\n", minor));
      return;
    }

  gm_assert (!ps->opened);

  IoDeleteSymbolicLink (&ps->arch.win32_device_name);
  gm_hash_remove (gm_device_hash, ps->arch.device_object);
  IoDeleteDevice (ps->arch.device_object);
  gm_free (ps->arch.win32_device_name.Buffer);
  gm_free (ps->arch.nt_device_name.Buffer);
  gm_minor_free (minor);
}

#if GM_ENABLE_ETHERNET

/************************************************************************
 * GM ethernet driver function table support 
 ************************************************************************/

#if (GM_SIZEOF_DP_T == 4)
#define HIGH_MASK 0
#elif (GM_SIZEOF_DP_T == 8)
#define HIGH_MASK 0xffffffff
#else
#error sizeof gm_dp_t unknown
#endif

void
gmm_allocate_shared_memory (struct gm_port *p,
			    int size,
			    void **virtualAddress,
			    ULONG * physicalLow, ULONG * physicalHigh)
{
  PHYSICAL_ADDRESS pa;

  pa.HighPart = HIGH_MASK;
#if GM_TRACE_LANAI_DMA
  pa.LowPart = (GM_MAX_HOST_PAGES * GM_PAGE_LEN) - 1;
#else
  pa.LowPart = 0xffffffff;
#endif

#if 1
  *virtualAddress = MmAllocateContiguousMemory (size, pa);
#else
  gm_always_assert ((unsigned) size < GM_PAGE_LEN);
  *virtualAddress = gm_page_alloc ();
#endif
  if (*virtualAddress)
    {
      pa = MmGetPhysicalAddress (*virtualAddress);
      *physicalLow = pa.LowPart;
      *physicalHigh = pa.HighPart;
#if GM_TRACE_LANAI_DMA
      GM_NOTE (("*** Ethernet buffer at DMA addr 0x%x.\n", *physicalLow));
      gm_assert (!*physicalHigh);
      gm_assert (p);
      gm_assert (p->kernel_port_state);
      gm_assert (p->kernel_port_state->instance);
      gm_assert (p->kernel_port_state->instance->dma_pages_bitmap);
      {
	gm_dp_t d;

	for (d = *physicalLow; d < *physicalLow + size; d += GM_PAGE_LEN)
	  {
	    unsigned long page_num;

	    page_num = d / GM_PAGE_LEN;
	    GM_PRINT (GM_PRINT_LEVEL >= 1,
		      ("*** Ethernet buffer at page 0x%x.\n", page_num));
	    p->kernel_port_state->instance->dma_pages_bitmap[page_num / 32] |=
	      gm_htonl (1 << page_num % 32);
	  }
      }
#endif
    }
}

void
gmm_free_shared_memory (struct gm_port *p, void *virtualAddress)
{
#if GM_TRACE_LANAI_DMA
  PHYSICAL_ADDRESS pa;

  pa = MmGetPhysicalAddress (virtualAddress);
  if (GM_SIZEOF_DP_T == 4)
    gm_assert (!pa.HighPart);
  gm_assert (p);
  gm_assert (p->kernel_port_state);
  gm_assert (p->kernel_port_state->instance);
  gm_assert (p->kernel_port_state->instance->dma_pages_bitmap);
  p->kernel_port_state->instance->dma_pages_bitmap[pa.LowPart
						   / GM_PAGE_LEN / 32]
    &= ~gm_htonl (1 << (pa.LowPart / GM_PAGE_LEN % 32));
#endif
  MmFreeContiguousMemory (virtualAddress);
}

void
gmm_disable_interrupts (struct gm_port *port)
{
  gm_assert (port);
  gm_assert (port->kernel_port_state);
  gm_assert (port->kernel_port_state->instance);

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("*** gmm_disable_interrupts called.\n"));
  gm_finalize_interrupts (port->kernel_port_state->instance);
}

void
gmm_enable_interrupts (struct gm_port *port)
{
  gm_assert (port);
  gm_assert (port->kernel_port_state);
  gm_assert (port->kernel_port_state->instance);

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("*** gmm_enable_interrupts called.\n"));
  gm_initialize_interrupts (port->kernel_port_state->instance);
}

#endif /* GM_ENABLE_ETHERNET */

/***********************************************************************
 * Required functions (gm_arch_*)
 ***********************************************************************/

/************
 * Driver entry point serialization.
 ************/

/* This is really simple on NT, which does not block in the driver to
   sleep a process.  We just call gm_nt_enter() when we enter a
   driver entry point, and we call gm_nt_exit() when we return, and
   have DriverEntry and Unload init and destroy the sync var. */

/* initialized by DriverEntry */
gm_arch_sync_t _gm_arch_serialize_sync;

void
gm_nt_enter ()
{
  gm_arch_mutex_enter (&_gm_arch_serialize_sync);
}

void
gm_nt_exit ()
{
  gm_arch_mutex_exit (&_gm_arch_serialize_sync);
}

/************
 * gm_port_state initialization 
 ************/

gm_status_t gm_arch_port_state_init (gm_port_state_t * ps)
{
  gm_assert (ps->arch.device_object == 0);
  gm_assert (ps->arch.nt_device_name.Buffer == 0);
  gm_assert (ps->arch.win32_device_name.Buffer == 0);
  InitializeListHead (&ps->arch.sleep_q_root);
  InitializeListHead (&ps->arch.mapping_list);
  gm_assert (ps->arch.sleep_cnt == 0);
  gm_assert (ps->arch.is_clone_device == 0);
  return GM_SUCCESS;
}

gm_status_t gm_arch_port_state_open (gm_port_state_t * ps)
{
  if (!ps->arch.is_clone_device)
    {
      gm_instance_state_t *is;

      is = ps->instance;
      gm_assert (is);
    }

  return GM_SUCCESS;
}

void
gm_arch_port_state_close (gm_port_state_t * ps)
{
  /*** Free resources allocated since gm_arch_port_state_open() */

  gm_arch_munmap_contiguous_segments (ps);
}

void
gm_arch_port_state_fini (gm_port_state_t * ps)
{
  return;
}

#if GM_SUPPORT_PCI
/************
 * PCI configuration space access.
 ************/

gm_status_t
gm_arch_read_pci_config (gm_instance_state_t* is, gm_offset_t offset,
                         void* value, unsigned long length)
{
  return (NdisReadPciSlotInformation (is->arch.miniportAdapterHandle, 0,
                                      offset, value, length)
          == length ? GM_SUCCESS : GM_FAILURE);
}

gm_status_t
gm_arch_write_pci_config (gm_instance_state_t* is, gm_offset_t offset,
                          void* value, unsigned long length)
{
  return (NdisWritePciSlotInformation (is->arch.miniportAdapterHandle, 0,
                                       offset, value, length)
          == length ? GM_SUCCESS : GM_FAILURE);
}

gm_status_t
gm_arch_read_pci_config_32 (gm_instance_state_t* is,
                            gm_offset_t offset, gm_u32_t* value)
{
  return gm_arch_read_pci_config (is, offset, value, sizeof (*value));
}

gm_status_t
gm_arch_write_pci_config_32 (gm_instance_state_t* is,
                             gm_offset_t offset, gm_u32_t value)
{
  return gm_arch_write_pci_config (is, offset, &value, sizeof (value));
}

gm_status_t
gm_arch_read_pci_config_16 (gm_instance_state_t* is,
			    gm_offset_t offset, gm_u16_t* value)
{
  return gm_arch_read_pci_config (is, offset, value, sizeof (*value));
}

gm_status_t
gm_arch_write_pci_config_16 (gm_instance_state_t* is,
			     gm_offset_t offset, gm_u16_t value)
{
  return gm_arch_write_pci_config (is, offset, &value, sizeof (value));
}

gm_status_t
gm_arch_read_pci_config_8 (gm_instance_state_t* is,
			   gm_offset_t offset, gm_u8_t* value)
{
  return gm_arch_read_pci_config (is, offset, value, sizeof (*value));
}

gm_status_t
gm_arch_write_pci_config_8 (gm_instance_state_t* is,
			    gm_offset_t offset, gm_u8_t value)
{
  return gm_arch_write_pci_config (is, offset, &value, sizeof (value));
}
#endif /* GM_SUPPORT_PCI */

/************
 * Copy data into/out of user buffers.
 ************/

gm_status_t
gm_arch_copyin (void *from, void *to, gm_size_t len)
{
  if (len)
    gm_bcopy (from, to, len);
  return GM_SUCCESS;
}

gm_status_t
gm_arch_copyout (void *from, void *to, gm_size_t len)
{
  if (len)
    gm_bcopy (from, to, len);
  return GM_SUCCESS;
}

/************
 * Introduce delay
 ************/

/* Spin for at least USECS microseconds */

void
gm_arch_spin (gm_instance_state_t * is, gm_u32_t usecs)
{
  LARGE_INTEGER large;

  large = RtlConvertUlongToLargeInteger (usecs);

  GM_PRINT (GM_PRINT_LEVEL >= 7, ("Pausing for %d usecs.\n", usecs));

  /* Multiply large by -10, since Microsoft does not supply a
     RtlLargeIntegerMultiply() function.  The factor of 10 is required
     because KeSetTimer requires times in 100 nanosecond intervals.  The
     negation is required to specify a relative time. */

  large = (RtlLargeIntegerNegate
	   (RtlLargeIntegerAdd (RtlLargeIntegerShiftLeft (large, 1),
				RtlLargeIntegerShiftLeft (large, 3))));

  /* Wait for the specified amount of time (or more if the timer is busy). */

  ExAcquireFastMutex (&is->arch.spin_fast_mutex);
  KeSetTimer (&is->arch.spin_timer, large, NULL);
  KeWaitForSingleObject (&is->arch.spin_timer, Executive, KernelMode, FALSE,
			 NULL);
  ExReleaseFastMutex (&is->arch.spin_fast_mutex);

  GM_PRINT (GM_PRINT_LEVEL >= 7, ("Done pausing.\n"));

}

/************
 * kernel memory allocation
 ************/

static int _gm_arch_alloc_cnt;

void*
__gm_arch_kernel_malloc (unsigned long len, int flags)
{
  void *ret;
  
  flags &= ~GM_KERNEL_MALLOC_PAGEABLE;
  
  if (flags & GM_KERNEL_MALLOC_PAGEABLE)
    ret = ExAllocatePool (PagedPool, len);
  else
    ret = ExAllocatePool (NonPagedPool, len);
  
  if (!ret)
    return ret;
  GM_PRINT (GM_PRINT_LEVEL >= 7,
	    ("%d kernel buffers allocated.\n", ++_gm_arch_alloc_cnt));
  return ret;
}

void
__gm_arch_kernel_free (void* ptr)
{
  ExFreePool (ptr);
  GM_PRINT (GM_PRINT_LEVEL >= 7,
	    ("%d kernel buffers allocated.\n", --_gm_arch_alloc_cnt));
}

/************
 * synchronization
 ************/

void
gm_arch_sync_reset (gm_arch_sync_t* s)
{
  GM_PRINT (GM_PRINT_LEVEL >= 5, ("gm_arch_sync_reset\n"));

  KeInitializeEvent(&s->sleep_event, SynchronizationEvent, FALSE);
}

void
gm_arch_sync_init (gm_arch_sync_t* s, gm_instance_state_t* is)
{
  GM_PRINT (GM_PRINT_LEVEL >= 5, ("gm_arch_sync_init.\n"));
  RtlZeroMemory (s, sizeof (*s));
  ExInitializeFastMutex (&s->mu);
  gm_arch_sync_reset (s);
}

void
gm_arch_mutex_enter (gm_arch_sync_t* s)
{
  GM_PRINT (GM_PRINT_LEVEL >= 10, ("Entering a mutex.\n"));
  ExAcquireFastMutex (&s->mu);
  GM_PRINT (GM_PRINT_LEVEL >= 10, ("Entered a mutex.\n"));
}

void
gm_arch_mutex_exit (gm_arch_sync_t* s)
{
  GM_PRINT (GM_PRINT_LEVEL >= 10, ("Exiting a mutex.\n"));
  ExReleaseFastMutex (&s->mu);
  GM_PRINT (GM_PRINT_LEVEL >= 10, ("Exited a mutex.\n"));
}

void
gm_arch_sync_destroy (gm_arch_sync_t* s)
{
  RtlZeroMemory (s, sizeof (*s));
}

/* This may be called at DISPATCH_LEVEL */

void
gm_arch_wake (gm_arch_sync_t* s)
{
  KeSetEvent(&s->sleep_event, 0, FALSE);
}

gm_arch_sleep_status_t
gm_arch_signal_sleep (gm_arch_sync_t * s)
{
  NTSTATUS status;
  gm_arch_sleep_status_t sleep_status = GM_SLEEP_WOKE;
  
  status = KeWaitForSingleObject(&s->sleep_event, Executive, UserMode,
				 FALSE, NULL);
  switch (status) {
  case STATUS_SUCCESS:
    sleep_status = GM_SLEEP_WOKE;
      break;
  case STATUS_USER_APC:
    sleep_status = GM_SLEEP_INTERRUPTED;
    break;
  default:
    GM_PANIC (("unexpected result in gm_arch_signal_sleep\n"));
    sleep_status = GM_SLEEP_WOKE;
    break;      
  }
  return sleep_status;
}

/* Called by gm_pause_lanai() at PASSIVE_LEVEL to wait for LANai to
   pause. */

gm_arch_sleep_status_t
gm_arch_timed_sleep (gm_arch_sync_t* s, int secs)
{
  NTSTATUS status;
  gm_arch_sleep_status_t sleep_status = GM_SLEEP_WOKE;
  LARGE_INTEGER timeout;
  
  timeout.QuadPart = secs;
  timeout.QuadPart *= 1000; /* to milliseconds */
  timeout.QuadPart *= 1000; /* to microseconds */
  timeout.QuadPart *= 10;   /* to 100 nanoseconds */
  timeout.QuadPart *= -1;   /* to relative time */

  status = KeWaitForSingleObject(&s->sleep_event, Executive, KernelMode,
				 FALSE, &timeout);
  switch (status) {
  case STATUS_SUCCESS:
    sleep_status = GM_SLEEP_WOKE;
    break;
  case STATUS_USER_APC:
    sleep_status = GM_SLEEP_INTERRUPTED;
    break;
  case STATUS_TIMEOUT:
    sleep_status = GM_SLEEP_TIMED_OUT;
    break;
  default:
    GM_PANIC (("unexpected result in gm_arch_timed_sleep\n"));
    sleep_status = GM_SLEEP_WOKE;
    break;      
  }
  return sleep_status;
}

/************
 * Misc required functions
 ************/

gm_status_t gm_arch_get_page_len (unsigned long *result)
{
  gm_u32_t i;

  /* Use the NT BYTES_TO_PAGES macro to infer the page size, since NT does
     not make the page size directly available. */

  i = 4096;
  ASSERT (BYTES_TO_PAGES (i) == 1);
  do
    i <<= 1;
  while (BYTES_TO_PAGES (i) == 1);;
  i >>= 1;

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("There are 0x%x bytes in a page.\n", i));
  *result = i;
  return GM_SUCCESS;
}

/* As indicated by the GM_ARCH_SYNC_FOR_DEVICE and
   GM_ARCH_SYNC_FOR_CPU bits of COMMAND, flush all transactions on the
   DMA region. */

gm_status_t gm_arch_dma_region_sync (gm_arch_dma_region_t * r, int command)
{
  return GM_SUCCESS;
}

/************
 * Memory mapping
 ************/

/* Structure for recording mappings so they can be unmapped */

typedef struct gm_map_record
{
  PVOID base;
  PVOID physicalMemorySection;
  LIST_ENTRY list_entry;
}
gm_map_record_t;

/* Map a segment of kernel memory into user space, and return the user
   virtual address. The segment is known to be contiguous in physical
   memory and in DMAable memory (in the bottom 2**32 bytes)

   Must run at PASSIVE_LEVEL */

gm_status_t
gm_arch_mmap_contiguous_segment (gm_port_state_t * ps,
				 void *kaddr,
				 unsigned long oldBlockSize, gm_up_t * upvaddr)
{
  NTSTATUS nt_status;
  PHYSICAL_ADDRESS physicalAddress;
  UNICODE_STRING physicalMemoryString;
  PVOID physicalMemorySection;
  OBJECT_ATTRIBUTES objectAttributes;
  HANDLE physicalMemoryHandle;
  PHYSICAL_ADDRESS viewBase;
  gm_map_record_t *record;
  void **vaddr;
  SIZE_T blockSize;

  blockSize = oldBlockSize;
  vaddr = (void **) upvaddr;
  ASSERT (kaddr != (PVOID) - 1 && kaddr != 0);

  physicalAddress = MmGetPhysicalAddress (kaddr);
  if (GM_SIZEOF_DP_T == 4)
    {
      if (physicalAddress.HighPart != 0)
	GM_NOTE (("gm_arch_mmap_contiguous_segment: HighPart != 0\n"));
      ASSERT (physicalAddress.HighPart == 0);
    }

  /* Get handle for physical memory */

  RtlInitUnicodeString (&physicalMemoryString, L"\\Device\\PhysicalMemory");
  InitializeObjectAttributes (&objectAttributes,
			      &physicalMemoryString,
			      OBJ_CASE_INSENSITIVE,
			      (HANDLE) NULL, (PSECURITY_DESCRIPTOR) NULL);
  nt_status = ZwOpenSection (&physicalMemoryHandle,
			     SECTION_ALL_ACCESS, &objectAttributes);
  if (NT_ERROR (nt_status))
    {
      GM_WARN (("Open section failed\n"));
      goto abort_with_nothing;
    }

  /* Allocate a record of the mapping */

  record = gm_malloc (sizeof (*record));
  if (!record)
    {
      GM_WARN (("Could not allocate mapping record.\n"));
      goto abort_with_nothing;
    }

  /* Record a new reference to physical memory, and get pointer to
     memory object */

  nt_status = ObReferenceObjectByHandle (physicalMemoryHandle,
					 SECTION_ALL_ACCESS,
					 (POBJECT_TYPE) NULL,
					 KernelMode,
					 &physicalMemorySection,
					 (POBJECT_HANDLE_INFORMATION) NULL);
  if (NT_ERROR (nt_status))
    {
      GM_WARN (("Reference object failed\n"));
      goto abort_with_record;
    }

  /* Map the section. */

  viewBase = physicalAddress;
#if GM_CAN_MAP_PAGES
#  error May be broken
  if (!*vaddr)
    {
#endif
      /* Map contiguous memory to system determined VMA */

      *vaddr = NULL;

      nt_status = ZwMapViewOfSection (physicalMemoryHandle,
				      (HANDLE) (-1),
				      vaddr,
				      0,
				      blockSize,
				      &viewBase,
				      &blockSize,
				      ViewShare,
				      0, PAGE_READWRITE | PAGE_NOCACHE);
      if (NT_ERROR (nt_status))
	{
	  GM_WARN (("Map view of section failed\n"));
	  goto abort_with_object_reference;
	}
#if GM_CAN_MAP_PAGES
#  error
    }
  else
    {
      /* Map non-contiguous or contiguous memory to user-supplied VMA
         one page at a time. */

      ULONG offset;

      if (!GM_PAGE_ALIGNED (*vaddr) || !GM_PAGE_ALIGNED (blockSize))
	{
	  GM_NOTE (("User attempted bad mapping of length 0x%x at %p.",
		    blocksize, *vaddr));
	  goto abort_with_object_reference;
	}

      for (offset = 0; offset < blockSize; offset += GM_PAGE_LEN)
	{
	  LARGE_INTEGER pageBase;
	  ULONG pageSize;
	  PVOID userAddr;

	  pageBase.QuadPart = viewBase.QuadPart + offset;
	  pageSize = GM_PAGE_LEN;
	  userAddr = (char *) *vaddr + offset;

	  GM_PRINT (GM_PRINT_LEVEL >= 4,
		    ("Mapping a page of physical memory.\n"));
	  nt_status =
	    ZwMapViewOfSection (physicalMemoryHandle, (HANDLE) (-1),
				&userAddr, 0, 0, &pageBase, &pageSize,
				ViewShare, 0, PAGE_READWRITE | PAGE_NOCACHE);
	  if (NT_ERROR (nt_status))
	    {
	      while (offset >= GM_PAGE_LEN)
		{
		  offset -= GM_PAGE_LEN;
		  ZwUnmapViewOfSection ((HANDLE) (-1),
					(PVOID) ((char *) *vaddr + offset));
		}
	      *vaddr = (PVOID) (-1);
	      goto abort_with_object_reference;
	    }
	}
    }
#endif

  /* Mapping the section above rounded the physical address down to
     the nearest 64K boundary. Compute the real virtual address by
     adding in the offset from the beginning of the section. */

#if GM_CPU_x86
  {
    gm_u32_t u32;
    u32 = (gm_u32_t) *vaddr;
    u32 += physicalAddress.LowPart - viewBase.LowPart;
    *vaddr = (void*) u32;
  }
#elif GM_CPU_ia64 || GM_CPU_x86_64
  {
    gm_u64_t u64;
    u64 = (gm_u64_t) *vaddr;
    u64 += physicalAddress.QuadPart - viewBase.QuadPart;
    *vaddr = (void*) u64;
  }
#else
#error Unknown cpu
#endif
  /* Record the base of the mapping for later unmapping. */

  record->base = *vaddr;
  record->physicalMemorySection = physicalMemorySection;
  InsertTailList (&ps->arch.mapping_list, &record->list_entry);

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Mapped to address %p.\n", *vaddr));

  return GM_SUCCESS;

abort_with_object_reference:
  ObDereferenceObject (physicalMemoryHandle);
abort_with_record:
  gm_free (record);
abort_with_nothing:
  *vaddr = (void *) -1;
  return GM_FAILURE;
}

void
gm_arch_munmap_contiguous_segments (gm_port_state_t * ps)
{
  gm_map_record_t *record;
  LIST_ENTRY *list_entry;
  NTSTATUS nt_status;

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Unmapping user mappings.\n"));

  while (!IsListEmpty (&ps->arch.mapping_list))
    {
      list_entry = RemoveHeadList (&ps->arch.mapping_list);
      ASSERT (list_entry);
      record = CONTAINING_RECORD (list_entry, gm_map_record_t, list_entry);
      GM_PRINT (GM_PRINT_LEVEL >= 4, ("Unmapping base %p section %p.\n",
				      record->base,
				      record->physicalMemorySection));
      nt_status = ZwUnmapViewOfSection ((HANDLE) - 1, record->base);
      if (NT_ERROR (nt_status))
	GM_WARN (("Error 0x%08x when unmapping memory.\n", nt_status));
      ObDereferenceObject (record->physicalMemorySection);
      gm_free (record);
    }
}

/************
 * DMA regions
 ************/

/* DMA regions functions (required by generic code in gm.c to grow the
   copy block.) */

/* Return the DMA address of the next page in the DMA region.

   Runs at PASSIVE_LEVEL but may run at any level. */

gm_dp_t
gm_arch_dma_region_dma_addr (gm_arch_dma_region_t * r)
{
  PHYSICAL_ADDRESS phys;

  phys = MmGetPhysicalAddress ((char *) r->addr + r->iterator_offset);
  if (GM_SIZEOF_DP_T == 4)
    {
      if (phys.HighPart != 0) {
	GM_NOTE (("gm_arch_dma_region_dma_addr: MmGetPhysicalAddress high"));
      }
      ASSERT (phys.HighPart == 0);
    }
#if (GM_SIZEOF_DP_T == 4)
  return (gm_dp_t) phys.LowPart;
#else
  return (gm_dp_t) phys.QuadPart;
#endif
}

/* Runs at PASSIVE_LEVEL but may be run at any level */

gm_dp_t
gm_arch_dma_region_dma_addr_advance (gm_arch_dma_region_t * r)
{
  PHYSICAL_ADDRESS phys;

  phys = MmGetPhysicalAddress ((char *) r->addr + r->iterator_offset);
  if (GM_SIZEOF_DP_T == 4) {
    if (phys.HighPart != 0) {
      GM_NOTE (("gm_arch_dma_region_dma_addr_advance: MmGetPhysicalAddress high"));
    }
    ASSERT (phys.HighPart == 0);
    /* ASSERT ((ULONG)r->addr == (ULONG)phys.LowPart | 0x80000000); */
  }
  r->iterator_offset += GM_PAGE_LEN;
#if (GM_SIZEOF_DP_T == 4)
  return (gm_dp_t) phys.LowPart;
#else 
  return (gm_dp_t) phys.QuadPart;
#endif
}

/* Allocate LEN bytes of DMA memory that is contiguous in kernal space
   but possibly segmented in DMA space.  HACK: This memory must also be
   contiguous in physical memory in NT.

   If r->register_function is non-null, call r->register_page_function
   (r, dma_addr, page_num) for each page.

   Must be run at PASSIVE_LEVEL */

gm_status_t
gm_arch_dma_region_alloc (gm_instance_state_t * is,
			  gm_arch_dma_region_t * r,
			  gm_size_t len,
			  gm_u32_t flags,
			  gm_status_t
			  (*register_page_func) (void *user_arg,
						 gm_dp_t dma_addr,
						 gm_u32_t page_num),
			  void *user_arg)
{
  PHYSICAL_ADDRESS phys;
  unsigned int page_num = 0;
  unsigned int page_offset;

  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("Allocating a DMA region of length 0x%x with flags 0x%x.\n", len,
	     flags));

  r->flags = flags;
  len = GM_PAGE_ROUNDUP (u32, len);
  gm_assert (len >= GM_PAGE_LEN);
  gm_assert (GM_PAGE_LEN);
  if (r->flags & GM_ARCH_DMA_CONTIGUOUS)
    {
      /* Allocate contiguous page-aligned memory */

      PHYSICAL_ADDRESS highest_acceptable;

      GM_PRINT (GM_PRINT_LEVEL >= 4,
		("Attempting to allocate contiguous memory.\n"));

      highest_acceptable.HighPart = HIGH_MASK;
      highest_acceptable.LowPart = 0xffffffff;
      r->orig_addr = MmAllocateContiguousMemory (len, highest_acceptable);
      if (!r->orig_addr)
	{
	  GM_NOTE (("Could not allocate contiguous DMA region"
		    " of 0x%x bytes.\n", len));
	  goto abort_with_nothing;
	}
      r->addr = GM_PAGE_ROUNDUP (ptr, r->orig_addr);
    }
  else
    {
      /* Allocate noncontiguous page-aligned memory */

      GM_PRINT (GM_PRINT_LEVEL >= 4,
		("Attempting to allocate noncontiguous memory.\n"));

      r->orig_addr = ExAllocatePool (NonPagedPool, len + GM_PAGE_LEN);
      if (!r->orig_addr)
	{
	  GM_NOTE (("Could not allocate noncontiguous DMA region"
		    " of 0x%x bytes.\n", len));
	  goto abort_with_nothing;
	}
      r->addr = r->orig_addr;
    }
  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("Allocated DMA region at system address %p.\n", r->orig_addr));
  r->len = len;
  r->iterator_offset = 0;
  r->is = is;

  /* Verify that NT actually page-aligns allocations larger than a
     page size as stated in the docs. */

  if (!GM_PAGE_ALIGNED (r->addr))
    {
      GM_NOTE (("Address of DMA buffer not page aligned.\n"));
      goto abort_with_memory;
    }

  /* Check that each page is DMAable. */

  for (page_offset = 0; page_offset < len; page_offset += GM_PAGE_LEN)
    {
      phys = MmGetPhysicalAddress ((char *) r->addr + page_offset);
      if (GM_SIZEOF_DP_T == 4)
	{
	  if (phys.HighPart != 0)
	    {
	      GM_NOTE (("At least some of DMA buffer is not DMAable."));
	      goto abort_with_memory;
	    }
	  ASSERT (phys.HighPart == 0);
	}
    }

#if GM_USE_MDLS
  /* Build a NT memory descriptor list to allow the DMA region to be
     flushed efficiently. */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Allocating an MDL.\n"));
  r->mdl = IoAllocateMdl (r->addr, r->len, FALSE, FALSE, NULL);
  if (r->mdl == NULL)
    {
      GM_WARN (("Could not allocate MDL"));
      goto abort_with_memory;
    }
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Building an MDL.\n"));
  MmBuildMdlForNonPagedPool (r->mdl);
#endif

  /* Register each page of the memory unless the user passed a NULL 
     registration function. */

  /* Registering the pages of the DMA region */

  if (register_page_func != 0 || GM_TRACE_LANAI_DMA)
    {
      GM_PRINT (GM_PRINT_LEVEL >= 4, ("Registering the pages.\n"));
      GM_PRINT (GM_PRINT_LEVEL >= 4, ("LANai SRAM at %p-%p.\n",
				      is->lanai.sram,
				      ((char *) is->lanai.sram
				       + is->lanai.eeprom.lanai_sram_size)));
      ASSERT (page_num == 0);
      for (page_offset = 0; page_offset < len; page_offset += GM_PAGE_LEN)
	{
	  GM_PRINT (GM_PRINT_LEVEL >= 6, ("Getting physical address.\n"));
	  phys = MmGetPhysicalAddress ((char *) r->addr + page_offset);
	  if (GM_SIZEOF_DP_T == 4)
	    {
	      if (phys.HighPart != 0) {
		GM_NOTE (("gm_arch_dma_alloc: MmGetPhysicalAddress high != 0\n"));
	      }
	      ASSERT (phys.HighPart == 0);
	    }
	  GM_PRINT (GM_PRINT_LEVEL >= 6, ("registering.\n"));
	  if (register_page_func)
	    {
#if (GM_SIZEOF_DP_T == 4)
	      register_page_func (user_arg, phys.LowPart, page_num++);
#else
	      register_page_func (user_arg, phys.QuadPart, page_num++);
#endif
	    }
#if GM_TRACE_LANAI_DMA
	  GM_NOTE (("*** GM DMA page at 0x%x.\n", phys.LowPart));
	  is->dma_pages_bitmap[phys.LowPart / GM_PAGE_LEN / 32]
	    |= gm_htonl (1 << (phys.LowPart / GM_PAGE_LEN % 32));
#endif
	}
    }
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Done registering.\n"));

  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("DMA region has system addr=%p and starting physaddr=%p.\n",
	     r->addr, gm_arch_dma_region_dma_addr (r)));

  return GM_SUCCESS;

abort_with_memory:
  if (r->flags & GM_ARCH_DMA_CONTIGUOUS)
    MmFreeContiguousMemory (r->orig_addr);
  else
    ExFreePool (r->orig_addr);
abort_with_nothing:
  RtlZeroMemory (r, sizeof (*r));
  return GM_FAILURE;
}

/* Must be run at PASSIVE_LEVEL */

void
gm_arch_dma_region_free (gm_arch_dma_region_t * r)
{
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Freeing %scontiguous DMA region %p.\n",
				  r->flags & GM_ARCH_DMA_CONTIGUOUS ? "" :
				  "non-", r->orig_addr));
  ASSERT (r->orig_addr);

#if GM_TRACE_LANAI_DMA
  /* remove the pages from the bitmap of safe DMA pages. */
  {
    char *addr;
    PHYSICAL_ADDRESS phys;

    for (addr = r->addr; addr < (char *) r->addr + r->len;
	 addr += GM_PAGE_LEN)
      {
	phys = MmGetPhysicalAddress (addr);
	if (GM_SIZEOF_DP_T == 4)
	  {
	    if (phys.HighPart != 0) {
	      GM_NOTE (("gm_arch_dma_region_free: MmGetPhysicalAddress\n"));
	    }
	    ASSERT (phys.HighPart == 0);
	  }
	r->is->dma_pages_bitmap[phys.LowPart / GM_PAGE_LEN / 32]
	  &= ~gm_htonl (1 << (phys.LowPart / GM_PAGE_LEN % 32));
      }
  }
#endif

  if (r->flags & GM_ARCH_DMA_CONTIGUOUS)
    {
      GM_PRINT (GM_PRINT_LEVEL >= 4,
		("Freeing contiguous DMA region at %p.\n", r->orig_addr));
      MmFreeContiguousMemory (r->orig_addr);
    }
  else
    {
      GM_PRINT (GM_PRINT_LEVEL >= 4,
		("Freeing discontiguous DMA region at %p.\n", r->orig_addr));
      ExFreePool (r->orig_addr);
    }
  RtlZeroMemory (r, sizeof (*r));
}

/* Return the kernal virtual address for a DMA region */

void *
gm_arch_dma_region_kernel_addr (gm_arch_dma_region_t * r)
{
  return r->addr;
}

/* Return the STS bits for a DMA region.

   Runs at PASSIVE_LEVEL but may be run at any level */

gm_s32_t gm_arch_dma_region_status (gm_arch_dma_region_t * r)
{
  return 0xf;			/* Invariant for LANai PCI interfaces. */
}


/************
 * Interrupt handling
 ************/

/* Interrupts are handled as follows:

   (1) The LANai generates an interrupt by setting HOST_SIG_BIT in the ISR.

   (2) NT invokes the interrupt handler for each instance that might
   have caused the interrupt.

   (3) The interrupt handler checks to see if it is already handling an
   interrupt. Then it checks to see if it is responsible for handling
   the interrupt.  If so, it sets a flag indicating an interrupt is
   being handled, schedules a delayed proceedure call (DPC) to handle
   the interrupt, and claims the interrupt by returning TRUE.
   Otherwise, it returns FALSE.

   (4) Eventually, the DPC runs.  The DPC handles the interrupt, informs
   the LANai that it has handled the interrupt (by clearing HOST_SIG_BIT
   in the ISR) and clears the flag saying it is being handled.
   
   Only one DPC can be queued at a time. The LANai will not interrupt
   again until HOST_SIG_BIT is written into the ISR, which is done at
   the end of the DPC. */

/* The interrupt handler DPC.

   Runs at DISPATCH_LEVEL */


#define GM_INTR_PRINT 0

VOID
gm_nt_DpcForIsr (IN PKDPC Dpc,
		 IN PVOID DeferredContext,
		 IN PVOID SystemArgument1, IN PVOID SystemArgument2)
{
  gm_instance_state_t *is;

  GM_PRINT (GM_INTR_PRINT, ("Interrupt DPC invoked.\n"));


  /* Perform sanity check */

  ASSERT (DeferredContext == SystemArgument1);

  is = (gm_instance_state_t *) DeferredContext;
  ASSERT (is);
  ASSERT (is->lanai.running);

  gm_handle_claimed_interrupt (is);

  GM_PRINT (GM_INTR_PRINT, ("Interrupt DPC completed.\n"));
  is->arch.in_intr = 0;
  gm_set_EIMR (is, GM_HOST_SIG_BIT);
  GM_STBAR ();
}

/* The interrupt handler.  Return TRUE if interrupt claimed, else FALSE.

   Runs at some DIRQ level > DISPATCH_LEVEL */

BOOLEAN
gm_nt_Isr (IN PKINTERRUPT intr, IN PVOID context)
{
  gm_instance_state_t *is = (gm_instance_state_t *) context;

  ASSERT (is);

  /* Exit if interrupt doesn't belong to us. */
  if (is->arch.in_intr ||
      (gm_interrupting (is) != GM_ARCH_INTR_CLAIMED))
    {
      return FALSE;
    }

  gm_set_EIMR (is, 0);
  GM_STBAR ();

  is->arch.in_intr = 1;
  KeInsertQueueDpc (&is->arch.dpc_for_isr, context, NULL);

  return TRUE;
}

/************************************************************************
 * IRP cancellation functions
 ************************************************************************/

/* Cancel an IRP that is in the work queue for its class */

VOID
gm_nt_cancel_irp (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
  gm_port_state_t *ps;

  GM_PRINT (GM_PRINT_LEVEL >= 5,
	    ("*** gm_nt_cancel_irp:  dev=0x%lx  irp=0x%lx.\n", DeviceObject,
	     Irp));

  /* The only cancellable IRPs are those in port sleep queue. */

  ps = gm_nt_device_to_ps (DeviceObject);
  gm_assert (ps);

  if (!IsListEmpty (&Irp->Tail.Overlay.ListEntry))
    {
      RemoveEntryList (&Irp->Tail.Overlay.ListEntry);
      ps->arch.sleep_cnt--;
    }
  else
    {
      GM_WARN (
	       ("Internal driver error: NT asked to cancel completed IRP.\n"));
    }
  IoReleaseCancelSpinLock (Irp->CancelIrql);
  Irp->IoStatus.Status = STATUS_CANCELLED;
  Irp->IoStatus.Information = 0;
  IoCompleteRequest (Irp, IO_NO_INCREMENT);
}

/***********************************************************************
 * functions to lock/unlock user memory
 **********************************************************************/

typedef struct gm_nt_phys_key
{
  HANDLE pid;
  PHYSICAL_ADDRESS phys_addr;
} gm_nt_phys_key;

typedef struct gm_nt_phys_data
{
  PMDL mdl;
  int count;
} gm_nt_phys_data;

static long gm_nt_phys_key_compare (void * key1, void * key2)
{
  gm_nt_phys_key k1;
  gm_nt_phys_key k2;

  gm_bcopy (key1, &k1, sizeof (k1));
  gm_bcopy (key2, &k2, sizeof (k2));
  if ((long) (k1.phys_addr.QuadPart != k2.phys_addr.QuadPart))
    return (long) (k1.phys_addr.QuadPart - k2.phys_addr.QuadPart);
  return (long) k1.pid - (long) k2.pid;
}

static gm_u32_t gm_nt_phys_key_hash (void * key1)
{
  gm_nt_phys_key k1;

  gm_bcopy (key1, &k1, sizeof (k1));
  return (gm_hash_hash_long (&k1.phys_addr.HighPart)
          ^ gm_hash_hash_long (&k1.phys_addr.LowPart));
}

#if GM_CAN_REGISTER_MEMORY
/* lock down a page and return the dma address for the page and a lock
   that will later be passed to gm_arch_unlock_user_buffer_page. */

#define GM_DEBUG_PIN 0

gm_status_t
gm_arch_lock_user_buffer_page (gm_port_state_t * ps,
			       gm_up_t upin,
			       gm_dp_t * dma_addr, gm_arch_page_lock_t * lock)
{
#if GM_NT_REF_COUNT_LOCK
  void *in;
  gm_nt_phys_key key;
  gm_nt_phys_data data;
  gm_nt_phys_data *p;

  in = (void*) upin;
  /* The physical address is subject to change if the page hasn't been locked
   * yet, but it doesn't matter. In this case, it can't be in the hash table.
   */
  key.phys_addr = MmGetPhysicalAddress (in);
  key.pid = PsGetCurrentProcessId ();
  p = gm_hash_find (gm_nt_phys_hash, &key);
  if (p == NULL)
    {
      data.mdl = IoAllocateMdl (in, GM_PAGE_LEN, FALSE, FALSE, NULL);
      if (data.mdl == NULL)
        goto abort_with_nothing;
      data.count = 1;
      __try
        {
          MmProbeAndLockPages (data.mdl, KernelMode, IoModifyAccess);
          /* Virtual address may be paged up to the point it is pinned. */
          key.phys_addr = MmGetPhysicalAddress (in);
          gm_hash_insert (gm_nt_phys_hash, &key, &data);
          p = gm_hash_find (gm_nt_phys_hash, &key);
          ASSERT (p != NULL);
        }
      __except (EXCEPTION_EXECUTE_HANDLER)
        {
          GM_PRINT (GM_DEBUG_PIN, ("Could not pin page.\n"));
          /* status = STATUS_INVALID_PARAMETER; */
          goto abort_with_mdl;
        }
    }
  else
    {
      p->count++;
    }

#if (GM_SIZEOF_DP_T == 4)
  *dma_addr = GM_PAGE_ROUND_DOWN (dp, key.phys_addr.LowPart);
#else
  *dma_addr = GM_PAGE_ROUND_DOWN (dp, key.phys_addr.QuadPart);
#endif
  *lock = key.phys_addr;
  return GM_SUCCESS;

abort_with_mdl:
  IoFreeMdl (data.mdl);
abort_with_nothing:
  return GM_FAILURE;
#else
  PMDL mdl;
  PHYSICAL_ADDRESS phys;
  void *in;

  in = (void *) upin;
  __try
  {
    mdl = IoAllocateMdl (in, GM_PAGE_LEN, FALSE, FALSE, NULL);
  }
  __except (EXCEPTION_EXECUTE_HANDLER)
  {
    GM_PRINT (GM_DEBUG_PIN, ("Could not allocate MDL to register user page."));
    /* status = STATUS_INVALID_PARAMETER; */
    goto abort_with_nothing;
  }
  if (!mdl)
    {
      GM_PRINT (GM_DEBUG_PIN, ("Could not allocate MDL to register user page."));
      /* status = STATUS_INSUFFICIENT_RESOURCES; */
      goto abort_with_nothing;
    }
  ASSERT (in == MmGetMdlVirtualAddress (mdl));

  /* Lock the page */

  __try
  {
    MmProbeAndLockPages (mdl, KernelMode, IoModifyAccess);
  }
  __except (EXCEPTION_EXECUTE_HANDLER)
  {
    GM_PRINT (GM_DEBUG_PIN, ("Could not pin page."));
    /* status = STATUS_INVALID_PARAMETER; */
    goto abort_with_mdl;
  }

  /* Now that the page is locked, we can get the dma addr */

  __try
  {
    phys = MmGetPhysicalAddress (in);
    if (GM_SIZEOF_DP_T == 4) {
      if (phys.HighPart != 0) {
	GM_PRINT (GM_DEBUG_PIN,
                  ("MmGetPhysicalAddress\n"));
      }
    }
  }  
  __except (EXCEPTION_EXECUTE_HANDLER)
  {
    GM_PRINT (GM_DEBUG_PIN, ("Exception getting phys addr for user buff."));
    /* status = STATUS_INVALID_PARAMETER; */
    goto abort_with_locked_page;
  }
  if (GM_SIZEOF_DP_T == 4)
    {
      ASSERT (phys.HighPart == 0);
      ASSERT (phys.LowPart != 0);
    }
  else
    {
      ASSERT (phys.QuadPart != 0);
    }
  /* Done.  Report results. */
#if (GM_SIZEOF_DP_T == 4)
  *dma_addr = GM_PAGE_ROUND_DOWN (dp, phys.LowPart);
#else
  *dma_addr = GM_PAGE_ROUND_DOWN (dp, phys.QuadPart);
#endif
  *lock = mdl;
  return GM_SUCCESS;

abort_with_locked_page:
  GM_PRINT (GM_DEBUG_PIN, ("******** Unlocking a page.\n"));
  MmUnlockPages (mdl);
abort_with_mdl:
  GM_PRINT (GM_DEBUG_PIN, ("Freeing an MDL.\n"));
  IoFreeMdl (mdl);
abort_with_nothing:
  return GM_FAILURE;
#endif
}

void
gm_arch_unlock_user_buffer_page (gm_arch_page_lock_t * lock)
{
#if GM_NT_REF_COUNT_LOCK

  gm_nt_phys_key key;
  gm_nt_phys_data *data;

  key.phys_addr = *lock;
  key.pid = PsGetCurrentProcessId ();
  data = gm_hash_find (gm_nt_phys_hash, &key);
  ASSERT (data != NULL);
  if (data == NULL)
    {
      GM_PRINT (GM_DEBUG_PIN, ("Could not find page to unlock\n"));
      return;
    }
  data->count--;
  if (data->count < 0)
    {
      GM_PRINT (GM_DEBUG_PIN, ("Lock count < 0.\n"));
      gm_hash_remove (gm_nt_phys_hash, &key);
    }
  if (data->count == 0)
    {
      MmUnlockPages (data->mdl);
      IoFreeMdl (data->mdl);
      gm_hash_remove (gm_nt_phys_hash, &key);
    }
#else
  gm_assert (*lock);
  MmUnlockPages (*lock);
  IoFreeMdl (*lock);
  *lock = 0;
#endif
}
#undef GM_DEBUG_PIN
#endif /* GM_CAN_REGISTER_MEMORY */

/************************************************************************
 * Device I/O Controls (ioctls)
 ************************************************************************/

NTSTATUS gm_nt_localize_status (gm_status_t status)
{
#define CASE(from,to) case from : return to
  switch (status)
    {
      CASE (GM_SUCCESS, STATUS_SUCCESS);
      CASE (GM_INPUT_BUFFER_TOO_SMALL, STATUS_BUFFER_TOO_SMALL);
      CASE (GM_OUTPUT_BUFFER_TOO_SMALL, STATUS_BUFFER_TOO_SMALL);
      CASE (GM_TRY_AGAIN, STATUS_WAIT_0);
      CASE (GM_BUSY, STATUS_ABANDONED_WAIT_0);
      CASE (GM_MEMORY_FAULT, STATUS_SEGMENT_NOTIFICATION);
      CASE (GM_INTERRUPTED, STATUS_CANCELLED);
      CASE (GM_INVALID_PARAMETER, STATUS_INVALID_DISPOSITION);
      CASE (GM_OUT_OF_MEMORY, STATUS_NO_MEMORY);
      CASE (GM_INVALID_COMMAND, STATUS_INVALID_DISPOSITION);
      CASE (GM_PERMISSION_DENIED, STATUS_ACCESS_VIOLATION);
      CASE (GM_INTERNAL_ERROR, STATUS_INTERNAL_ERROR);
      CASE (GM_UNATTACHED, STATUS_NONCONTINUABLE_EXCEPTION);
      CASE (GM_UNSUPPORTED_DEVICE, STATUS_BAD_DEVICE_TYPE);
    default:
      return gm_nt_localize_status (GM_INTERNAL_ERROR);
    }
}

#define GM_REQUIRE_OPEN_PORT() do {					\
  if (!ps->opened)							\
    {									\
      GM_NOTE (("User attempted port IOCTL "				\
		   "on uninitialized port.\n"));			\
      gm_status = GM_PERMISSION_DENIED;					\
      break;								\
    }									\
  } while (0)

/* Fuction to actually perform the requested DeviceControl in the worker
   thread. */

#define GM_DEBUG_NT_IOCTL 0

NTSTATUS
gm_nt_ioctl (PDEVICE_OBJECT dev, PIRP Irp)
{
  PIO_STACK_LOCATION irpStack;
  NTSTATUS nt_status;
  gm_status_t gm_status;
  PVOID ioBuffer;
  ULONG inBufferLen, outBufferLen;
  int instance, port_id;
  gm_instance_state_t *is;
  gm_port_state_t *ps;
  ULONG output_bytes = 0;
  ULONG code;
  int inhibit_status_copy = 0;

  /* Compute values needed by nearly all ioctls */

  irpStack = IoGetCurrentIrpStackLocation (Irp);
  ASSERT (irpStack);
  ioBuffer = Irp->AssociatedIrp.SystemBuffer;
  inBufferLen = irpStack->Parameters.DeviceIoControl.InputBufferLength;
  outBufferLen = irpStack->Parameters.DeviceIoControl.OutputBufferLength;

  ps = gm_nt_device_to_ps (dev);
  gm_assert (ps);
  is = ps->instance;
  gm_assert (is);

  port_id = ps->id;
  instance = is->id;

  /* Prepare for successful return */

  output_bytes = 0;
  gm_status = GM_SUCCESS;

  /* Handle cases of clone devices and non-clone devices */

  code = irpStack->Parameters.DeviceIoControl.IoControlCode;

  GM_PRINT (GM_DEBUG_NT_IOCTL,
	    ("gm_nt_port_ioctl called with control code 0x%x %s irp=0x%lx.\n",
	     code, _gm_ioctl_cmd_name (code), Irp));

  switch (code)
    {
      /*******************************
       * Platform-specific IOCTLs
       ********************************/

    case GM_SET_PORT_NUM:
      {
	unsigned long port_id;
	gm_arch_minor_t minor;
	gm_port_state_t *new_ps;

	GM_PRINT (GM_DEBUG_NT_IOCTL, ("ioctl: GM_SET_PORT_NUM\n"));
	if (inBufferLen < sizeof (gm_u32_t))
	  {
	    gm_status = GM_INPUT_BUFFER_TOO_SMALL;
	    break;
	  }
	if (!ps->arch.is_clone_device)
	  {
	    gm_status = GM_PERMISSION_DENIED;
	    break;
	  }
	port_id = *(gm_u32_t *) ioBuffer;

	gm_status = gm_nt_create_minor_node (is, L"gm", &minor,
					     0 /* not clone */ ,
					     ps->privileged, port_id);
	if (gm_status != GM_SUCCESS)
	  {
	    GM_NOTE (("Failed to create minor node.\n"));
	    break;
	  }
	++gm_usage_count;
	new_ps = gm_minor_get_port_state (minor);
	gm_assert (new_ps);

	gm_status = gm_port_state_open (new_ps, port_id);
	if (gm_status != GM_SUCCESS)
	  {
	    GM_NOTE (("Failed to open port state.\n"));
	    --gm_usage_count;
	    gm_nt_destroy_minor_node (minor);
	    break;
	  }
	break;
      }

      /************
       * Simulate the Unix mmap functionality. */

    case GM_MMAP:
      {
	gm_mmap_info_t *mmi;
	void *kaddr;

	inhibit_status_copy = 1;	/* mustn't touch mapped buffer */

	GM_PRINT (GM_DEBUG_NT_IOCTL, ("ioctl: GM_MMAP\n"));
	GM_REQUIRE_OPEN_PORT ();

	if (0 && (ps->id == GM_DAEMON_PORT_ID))
	  {
	    gm_status = GM_PERMISSION_DENIED;
	    break;
	  }

	if (inBufferLen < sizeof (*mmi))
	  {
	    gm_status = GM_INPUT_BUFFER_TOO_SMALL;
	    break;
	  }
	mmi = ioBuffer;
	mmi->va = (void *) -1;	/* default */
	if (!GM_PAGE_ALIGNED (mmi->offset) || !GM_PAGE_ALIGNED (mmi->len))
	  {
	    gm_status = GM_INVALID_PARAMETER;
	    break;
	  }

	/* Verify that the user has requested a legitimate mapping, and
	   preallocate any required resource. */

	gm_status = gm_prepare_to_mmap (ps, mmi->offset, mmi->len,
					mmi->requested_permissions);
	if (gm_status != GM_SUCCESS)
	  {
	    GM_NOTE (("User attempted forbidden mmap.\n"));
	    break;
	  }

	/* Determine the kernel address of the resource to map.  This
	   should never fail because gm_prepare_to_mmap checks the
	   legitimacy of the offsets. */
	gm_status = gm_mmap (ps, mmi->offset, &kaddr);
	if (gm_status != GM_SUCCESS)
	  {
	    GM_WARN (("internal error: gm_mmap failed.\n"));
	    break;
	  }

	/* Map the requested region.  We know that the mapping is
	   contiguous in physical memory (HACK) */

	gm_assert (kaddr);
	gm_status = gm_arch_mmap_contiguous_segment (ps, kaddr, mmi->len,
						     (gm_up_t *) & mmi->va);
	if (gm_status != GM_SUCCESS)
	  {
	    gm_status = GM_INPUT_BUFFER_TOO_SMALL;
	    break;
	  }
	gm_assert (mmi->va != (void *) 0);
	gm_assert (mmi->va != (void *) -1);

	/* Finish the mapping, adding any mapped page to the page hash
	   table */

	gm_finish_mmap (ps, mmi->offset, mmi->len, (gm_up_t) mmi->va);
	output_bytes = sizeof (*mmi);
	break;
      }

    case GM_FINISH_MMAP:
      GM_PRINT (GM_DEBUG_NT_IOCTL, ("ioctl: GM_FINISH_MMAP\n"));
      break;


#if GM_CAN_REGISTER_MEMORY
    /**************
     * Register user-allocated memory for DMA access. */

    case GM_REGISTER_MEMORY:
    case GM_DEREGISTER_MEMORY:

      GM_PRINT (GM_DEBUG_NT_IOCTL, ("ioctl: GM_(DE)REGISTER_MEMORY\n"));
      GM_REQUIRE_OPEN_PORT ();

      /* fixup the input buffer, which is in an unusual place for
         this IOCTL because it had to use METHOD_NEITHER buffer
         management. */

      ioBuffer = irpStack->Parameters.DeviceIoControl.Type3InputBuffer;

      /* Perform the default memory registration. */

      gm_status = gm_ioctl (ps, code, ioBuffer, inBufferLen, ioBuffer,
			    outBufferLen, &output_bytes);
      inhibit_status_copy = 1;	/* mustn't do for METHOD_NEITHER buffer */
      break;

#endif /* GM_CAN_REGISTER_MEMORY */

    case GM_SLEEP:
      gm_nt_exit();
      gm_arch_mutex_enter(&ps->sync);
      gm_status = gm_ioctl (ps, code, ioBuffer, inBufferLen, ioBuffer,
			    outBufferLen, &output_bytes);
      gm_arch_mutex_exit(&ps->sync);
      gm_nt_enter();
      break;

      /*******************************
       * Platform-independent IOCTLs (also returns error 
       ********************************/

    default:
      GM_PRINT (GM_DEBUG_NT_IOCTL, ("ioctl: default - using gm_ioctl\n"));
      gm_arch_mutex_enter(&ps->sync);
      gm_status = gm_ioctl (ps, code, ioBuffer, inBufferLen, ioBuffer,
			    outBufferLen, &output_bytes);
      gm_arch_mutex_exit(&ps->sync);
      inhibit_status_copy = 1;	/* done already */
      break;
    }

  /* If there was an error, inform the user via the copy buffer, ensuring
     that the OS does not get a chance to mangle the error code. */

  if ((gm_status != GM_SUCCESS)
      && ioBuffer
      && outBufferLen >= sizeof (gm_status) && !inhibit_status_copy)
    {
      GM_PRINT (GM_DEBUG_NT_IOCTL,
		("ioctl: gm_status != GM_SUCCESS, copyout the error\n"));
      output_bytes = sizeof (gm_status);
      gm_copyout (ps, (void *) &gm_status, ioBuffer, output_bytes);
    }

  nt_status = gm_nt_localize_status (gm_status);
  GM_PRINT (GM_DEBUG_NT_IOCTL,
	    ("ioctl: gm_status = 0x%x nt_status = 0x%x\n", gm_status,
	     nt_status));
  if (nt_status != STATUS_SUCCESS)
    {
      GM_PRINT
	(GM_DEBUG_NT_IOCTL,
	 ("Reporting failed ioctl completion with GM status 0x%x;"
	  " code was 0x%x\n", gm_status, code));
    }



/* The Irp status and the return code are supposed to be the same
   according to the NT example code. We've decided to return fail,
   but leave success in the Irp so that the error will get copied to the
   user's buffer.
   
   My testing - feldy
   
   Irp     Function
   Status  Return          Result
   
   Fail    Fail            ioctl fails, but the error return code is not
   copied to user
   Fail    Success         ioctl returns success
   Success Fail            ioctl fails, error code is copied
   Success Success         ioctl suceeds, normal data is copied */


  Irp->IoStatus.Status = /* nt_status */ STATUS_SUCCESS;
  Irp->IoStatus.Information = output_bytes;
  IoCompleteRequest (Irp, IO_NO_INCREMENT);
  GM_PRINT (GM_DEBUG_NT_IOCTL,
	    ("Ioctl complete. irp=0x%lx  status = 0x%x obytes = %d\n", Irp,
	     nt_status, output_bytes));

  return nt_status;
}

/* Simplified interface to NT registry: Fetch the entry from
   KEY_NAME\VALUE_NAME of type DATA_TYPE into a dynamically allocated
   buffer and store a pointer to that buffers at *PTR. User will free
   the buffer using Free SnarfedData. */

NTSTATUS
gm_nt_SnarfFromRegistry (IN PCWSTR key_name,
			 IN PCWSTR value_name,
			 IN ULONG data_type,
			 OUT PKEY_VALUE_FULL_INFORMATION * ptr)
{
  OBJECT_ATTRIBUTES oa;
  NTSTATUS nt_status;
  UNICODE_STRING us, us2;
  ULONG buf_len = 0;
  ULONG trash;
  PKEY_VALUE_FULL_INFORMATION buffer;
  HANDLE h;

  ASSERT (ptr != NULL);
  RtlInitUnicodeString (&us, key_name);
  RtlInitUnicodeString (&us2, value_name);
  InitializeObjectAttributes (&oa, &us, OBJ_CASE_INSENSITIVE, NULL, NULL);
  nt_status = ZwOpenKey (&h, KEY_READ, &oa);
  if (NT_ERROR (nt_status))
    {
      GM_NOTE (("Could not open registry key. (error 0x%x)\n", nt_status));
      goto abort_with_nothing;
    }
  nt_status = ZwQueryValueKey (h, &us2, KeyValueFullInformation, NULL, 0,
			       &buf_len);

  if (!buf_len ||
      (NT_ERROR (nt_status) && nt_status != STATUS_BUFFER_TOO_SMALL))
    {
      GM_NOTE (("Could not determine size of registry key data. "
		"(error 0x%x)\n", nt_status));
      goto abort_with_open_key;
    }
  GM_PRINT (GM_PRINT_LEVEL >= 3,
	    ("Allocating buffer for %d bytes of registry information.\n",
	     buf_len - sizeof (KEY_VALUE_FULL_INFORMATION)));
  buffer = gm_malloc (buf_len);
  if (!buffer)
    {
      GM_NOTE (("Could not allocate buffer for registry data.\n"));
      nt_status = STATUS_INSUFFICIENT_RESOURCES;
      goto abort_with_open_key;
    }
  /* Querying registery. */
  nt_status = ZwQueryValueKey (h, &us2, KeyValueFullInformation, buffer,
			       buf_len, &trash);
  if (NT_ERROR (nt_status))
    {
      GM_NOTE (("Could not read registry data.\n"));
      goto abort_with_buffer;
    }
  if (trash != buf_len)
    {
      GM_NOTE (("read less (%d bytes) than expected (%d bytes)\n",
		trash, buf_len));
      goto abort_with_buffer;
    }
  if (buffer->Type != data_type)
    {
      GM_NOTE (("Read value from registry, but it had wrong type (0x%x).\n",
		buffer->Type));
      nt_status = STATUS_INVALID_PARAMETER;
      goto abort_with_buffer;
    }

  GM_PRINT (GM_PRINT_LEVEL >= 3, ("Closing registry handle.\n"));
  ZwClose (h);
  *ptr = buffer;
  return STATUS_SUCCESS;

abort_with_buffer:
  gm_free (buffer);
abort_with_open_key:
  ZwClose (h);
abort_with_nothing:
  return nt_status;
}

void
gm_nt_FreeSnarfedData (void *ptr)
{
  gm_free (ptr);
}

/* Report the number of pages of physical memory in the machine.

   This is done by fetching the system memory resource list from the
   registry and summing up the lengths of the memory resources found
   there.
   
   Does and must run at PASSIVE_LEVEL. */
gm_status_t
gm_arch_physical_pages (gm_u32_t * result)
{
  NTSTATUS nt_status;
  PKEY_VALUE_FULL_INFORMATION value;
  PCM_RESOURCE_LIST list;
  PCM_FULL_RESOURCE_DESCRIPTOR full;
  PCM_PARTIAL_RESOURCE_LIST partial_list;
  PCM_PARTIAL_RESOURCE_DESCRIPTOR partial;
  static ULONG pages = 0;
  unsigned int i, j;

  /* Return cached page count, if any */
  if (pages != 0)
    {
      *result = pages;
      return GM_SUCCESS;
    }

  /* Get the Internal resource descriptor from the registry. */
  nt_status =
    gm_nt_SnarfFromRegistry
    (L"\\Registry\\Machine\\HARDWARE\\RESOURCEMAP\\System Resources\\Physical Memory",
     L".Translated", REG_RESOURCE_LIST, &value);
  if (NT_ERROR (nt_status))
    {
      GM_NOTE (("Could not snarf resource list from registry.\n"));
      return GM_FAILURE;
    }
  list = (PCM_RESOURCE_LIST) ((char *) value + value->DataOffset);
  GM_PRINT (GM_PRINT_LEVEL >= 3,
	    ("Scanning system resources list of length %d for memory.\n",
	     list->Count));

  for (i = 0; i < list->Count; i++)
    {
      full = &list->List[i];
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("partial descriptor:\n"));
      GM_PRINT (GM_PRINT_LEVEL >= 3,
		("  InterfaceType = %d\n", full->InterfaceType));
      GM_PRINT (GM_PRINT_LEVEL >= 3, ("  BusNumber = %d\n", full->BusNumber));
      if (full->InterfaceType != Internal)
	{
	  GM_WARN (("Interface type is not \"Internal.\"\n"));
	  continue;
	}
      partial_list = &full->PartialResourceList;
      GM_PRINT (GM_PRINT_LEVEL >= 4,
		("Scanning partial resource list of length %d.\n",
		 partial_list->Count));
      for (j = 0; j < partial_list->Count; j++)
	{
	  ULONG new_pages;

	  partial = &partial_list->PartialDescriptors[j];
	  if (partial->Type != CmResourceTypeMemory)
	    {
	      GM_NOTE (("Skipping non-memory resource.\n"));
	      continue;
	    }
	  GM_PRINT (GM_PRINT_LEVEL >= 3,
		    ("Found system memory at 0x%x%08x of length 0x%x.\n",
		     partial->u.Memory.Start.HighPart,
		     partial->u.Memory.Start.LowPart,
		     partial->u.Memory.Length));
	  /* Update the total amount of host memory, handling overflow
	     gracefully. */
	  new_pages
	    = ((partial->u.Memory.Length + GM_PAGE_LEN - 1) / GM_PAGE_LEN);
	  if (pages + new_pages < pages)
	    {
	      GM_WARN (("Overflow in memory computation.\n"));
	      pages = ULONG_MAX;
	    }
	  else
	    pages += new_pages;
	}
    }
  gm_nt_FreeSnarfedData (value);
  GM_PRINT (GM_PRINT_LEVEL >= 3,
	    ("*** total physical memory: 0x%x pages (0x%x MB).\n", pages,
	     (pages * GM_PAGE_LEN) >> 20));
  *result = pages;
  return GM_SUCCESS;
}

gm_status_t
gm_arch_map_io_space (gm_instance_state_t * is, gm_u32_t offset, gm_u32_t len,
		      void **ptr)
{
  PHYSICAL_ADDRESS translatedAddress;
  PHYSICAL_ADDRESS bus_addr;
  ULONG addressSpace;

  bus_addr = RtlLargeIntegerAdd (is->arch.board_phys_addr,
				 RtlConvertUlongToLargeInteger (offset));

  addressSpace = 0;
  gm_assert (is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI);
  if (!HalTranslateBusAddress (PCIBus,
			       is->arch.bus_number,
			       bus_addr, &addressSpace, &translatedAddress))
    {
      GM_WARN (("Can't translate the address\n"));
      return GM_FAILURE;
    }
  if (addressSpace != 0)
    {
      GM_WARN (("Unexpected address space value\n"));
      return GM_FAILURE;
    }

  gm_assert (sizeof (is->ifc.pci.config.Command) == 2);
  gm_arch_read_pci_config_16 (is, GM_OFFSETOF (gm_pci_config_t, Command),
			      &is->ifc.pci.config.Command);
  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("PCI Command = 0x%x.\n", is->ifc.pci.config.Command));

  /* Map the region of LANai space into system memory and return a
     pointer to it, or NULL on failure. */

  *ptr = MmMapIoSpace (translatedAddress, len, FALSE);
  if (!*ptr)
    {
      GM_NOTE (("MmMapIoSpace failed - phys_addr = 0x%x  len = %d\n",
		translatedAddress, len));
      return GM_FAILURE;
    }


  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Mapped 0x%x bytes at offset 0x%x to %p.\n",
				  len, offset, *ptr));

  return GM_SUCCESS;
}

void
gm_arch_unmap_io_space (gm_instance_state_t * is, gm_u32_t offset,
			gm_u32_t len, void **ptr)
{
  MmUnmapIoSpace (*ptr, len);
}

gm_status_t
gm_arch_gethostname (char *ptr, int len)
{
#if 0
  PKEY_VALUE_FULL_INFORMATION full;
  NTSTATUS nt_status;
  UNICODE_STRING us;
  ANSI_STRING as;
  char *name;

  /* Read the hostname from the register as a WCHAR string */
  nt_status =
    gm_nt_SnarfFromRegistry
    (L"\\Registry\\Machine\\SYSTEM\\CurrentControlSet\\Control\\ComputerName\\ActiveComputerName",
     L"ComputerName", REG_SZ, &full);
  if (NT_ERROR (nt_status))
    {
      GM_NOTE (("Could not snarf computer name from registry.\n"));
      goto abort_with_nothing;
    }
  /* Convert the hostname to a character string */
  RtlInitUnicodeString (&us, (PWCHAR) ((char *) full + full->DataOffset));
  nt_status = RtlUnicodeStringToAnsiString (&as, &us, TRUE);
  if (NT_ERROR (nt_status))
    {
      gm_nt_FreeSnarfedData (full);
      goto abort_with_hash;
    }
  /* Convert hostname to lower case. */
  for (name = as.Buffer; *name; name++)
    *name = tolower (*name);
  name = as.Buffer;
  GM_PRINT (GM_PRINT_LEVEL >= 1,
	    ("**** Hostname = \"%s\" (length %d). ****\n", name,
	     strlen (name)));
  /* Store the host name in the configuration packets. */
  strncpy (ptr, as.Buffer, len);
  RtlFreeAnsiString (&as);
  gm_nt_FreeSnarfedData (full);

  return GM_SUCCESS;

abort_with_hash:
  gm_nt_FreeSnarfedData (full);
abort_with_nothing:
  return GM_FAILURE;
#else
  strncpy(ptr, "localhost", len);
  return GM_SUCCESS;
#endif
}

/* Utility routine for DispatchCleanup.

   NOTE: Must hold Cancel spin lock before calling. */

void
gm_nt_cancel_last (IN PDEVICE_OBJECT dev, PLIST_ENTRY root, KIRQL irql)
{
  PIRP irp;
  PDRIVER_CANCEL cancel_routine;

  /* Get last IRP in list. */

  irp = CONTAINING_RECORD (root->Blink, IRP, Tail.Overlay.ListEntry);
  ASSERT (irp);

  /* Cancel the IRP following the NT cancelRoutine conventions.  The
     cancel routine will remove the IRP from the list. */

  cancel_routine = irp->CancelRoutine;
  ASSERT (cancel_routine);
  irp->CancelIrql = irql;
  irp->CancelRoutine = NULL;
  irp->Cancel = TRUE;

  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("gm_nt_cancel_last: Calling cancel routine for dev=0x%lx"
	     "  irp=0x%lx.\n",
	     dev, irp));
  cancel_routine (dev, irp);	/* releases cancel spin lock */

  /* reacquire since the cancel_routine released it */
  IoAcquireCancelSpinLock (&irql);
}

void
gm_nt_print_offsets ()
{
#define gm_print_offset(name) do {					\
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("0x%x: %s\n",				\
		GM_OFFSETOF (gm_lanai_globals_t, name), #name));	\
} while (0)

  gm_print_offset (magic);
  gm_print_offset (event_index_table);
  gm_print_offset (page_hash);
  gm_print_offset (first_port_with_sent_packets);
  gm_print_offset (_subport[0]);
  gm_print_offset (_subport[1]);
  gm_print_offset (timeout_time);
  gm_print_offset (current_rdma_port);
  gm_print_offset (recv_token_bin[0]);
  gm_print_offset (recv_token_bin[1]);
  gm_print_offset (first_active_connection);
  gm_print_offset (_send_record[0]);
  gm_print_offset (_send_record[1]);
  gm_print_offset (recv_token_dma_stage);
  gm_print_offset (failed_send_event_dma_stage);
  gm_print_offset (report_dma_stage);
  gm_print_offset (nack_delay);
  gm_print_offset (backlog_delay);
  gm_print_offset (led);
  gm_print_offset (port[0]);
  gm_print_offset (port[1]);
  gm_print_offset (finishing_rdma_for_port);
  gm_print_offset (record_log);
  gm_print_offset (resume_after_halt);
  gm_print_offset (pause_cnt);
  gm_print_offset (trash[0]);
}

/************************************************************************
 * NT event logging functions
 ************************************************************************/

#define LOG_LEVEL_NONE 0
#define LOG_LEVEL_DEBUG 1
#define GM_MAX_INSERTION_STRINGS 4
#define GM_ERRORLOG 1

PDRIVER_OBJECT the_one_and_only_driver_object;

static ULONG _gm_nt_event_log_level;

static ULONG
gm_nt_event_log_strlen (IN PWSTR string)
{
  UNICODE_STRING tmp;

  RtlInitUnicodeString (&tmp, string);
  return tmp.Length;
}

static BOOLEAN
gm_nt_report_event (IN ULONG message_level, IN NTSTATUS error_code,
		    IN ULONG unique_error_value, IN PVOID io_object,
		    IN PIRP irp, IN ULONG dump_data[],
		    IN ULONG dump_data_count, IN PWSTR strings[],
		    IN ULONG string_count)
{
  PIO_ERROR_LOG_PACKET packet;
  PIO_STACK_LOCATION irp_stack;
  PUCHAR insertion_string;
  UCHAR packet_size;
  UCHAR string_size[GM_MAX_INSERTION_STRINGS];
  ULONG i;

  if ((_gm_nt_event_log_level == LOG_LEVEL_NONE) ||
      (message_level > _gm_nt_event_log_level))
    {
      return TRUE;
    }

  /* calculate size of packet */

  packet_size = sizeof (IO_ERROR_LOG_PACKET);
  if (dump_data_count > 0)
    {
      packet_size += (UCHAR) ((dump_data_count - 1) * sizeof (ULONG));
    }
  if (string_count > 0)
    {
      if (string_count > GM_MAX_INSERTION_STRINGS)
	{
	  string_count = GM_MAX_INSERTION_STRINGS;
	}
      for (i = 0; i < string_count; ++i)
	{
	  string_size[i] = (UCHAR) gm_nt_event_log_strlen (strings[i])
	    + sizeof (WCHAR);
	  packet_size += string_size[i];
	}
    }

  packet = IoAllocateErrorLogEntry (io_object, packet_size);
  if (packet == NULL)
    {
      return FALSE;
    }

  /* fill in packet */

  RtlZeroMemory (packet, packet_size);

  {
    char *pch;
    pch = (char *) packet;
    for (i = 0; i < packet_size; ++i)
      {
	ASSERT (*pch == 0);
      }
    ASSERT (packet->DumpDataSize == 0);
  }

  packet->ErrorCode = error_code;
  packet->UniqueErrorValue = unique_error_value;

  if (irp != NULL)
    {
      irp_stack = IoGetCurrentIrpStackLocation (irp);

      packet->MajorFunctionCode = irp_stack->MajorFunction;
      packet->FinalStatus = irp->IoStatus.Status;

      if ((irp_stack->MajorFunction == IRP_MJ_DEVICE_CONTROL) ||
	  (irp_stack->MajorFunction == IRP_MJ_INTERNAL_DEVICE_CONTROL))
	{
	  packet->IoControlCode =
	    irp_stack->Parameters.DeviceIoControl.IoControlCode;
	}
    }

  /* add dump data */

  packet->DumpDataSize = (USHORT) (dump_data_count * sizeof (ULONG));
  if (dump_data_count > 0)
    {
      packet->DumpDataSize = (USHORT) (dump_data_count * sizeof (ULONG));
      for (i = 0; i < dump_data_count; ++i)
	{
	  packet->DumpData[i] = dump_data[i];
	}
    }

  /* add insertion strings */

  if (string_count > 0)
    {
      packet->NumberOfStrings = (USHORT) string_count;
      packet->StringOffset = (sizeof (IO_ERROR_LOG_PACKET) +
	(dump_data_count - 1) * sizeof (ULONG));
      insertion_string = (PUCHAR) packet + packet->StringOffset;
      for (i = 0; i < string_count; ++i)
	{
	  RtlCopyBytes (insertion_string, strings[i], string_size[i]);
	  insertion_string += string_size[i];
	}
    }

  /* log the message */

  IoWriteErrorLogEntry (packet);
  return TRUE;
}

VOID gm_nt_log (int code, char *fmt, va_list ap)
{
#if 0
  NTSTATUS nt_status;
  char event_log_buffer[256];
  ANSI_STRING astr;
  UNICODE_STRING ustr;
  
  astr.Length = vsprintf (event_log_buffer, fmt, ap);
  astr.MaximumLength = sizeof (event_log_buffer);
  astr.Buffer = event_log_buffer;
  nt_status = RtlAnsiStringToUnicodeString (&ustr, &astr, TRUE);
  if (nt_status == STATUS_SUCCESS)
    {
      gm_nt_report_event (LOG_LEVEL_DEBUG, GM_MSG_INFO,
			  GM_ERRORLOG,
			  (PVOID) the_one_and_only_driver_object,
			  NULL, NULL, 0, &ustr.Buffer, 1);
      RtlFreeUnicodeString (&ustr);
    }
#endif
}

VOID gm_nt_info (char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  gm_nt_log (GM_MSG_INFO, fmt, ap);
  va_end (ap);
}

VOID gm_nt_note (char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  gm_nt_log (GM_MSG_NOTE, fmt, ap);
  va_end (ap);
}

VOID gm_nt_warn (char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  gm_nt_log (GM_MSG_WARN, fmt, ap);
  va_end (ap);
}

VOID gm_nt_panic (char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  gm_nt_log (GM_MSG_PANIC, fmt, ap);
  va_end (ap);
}

static VOID
gm_nt_initialize_event_log (IN PDRIVER_OBJECT DriverObject)
{
  NTSTATUS nt_status;
  PKEY_VALUE_FULL_INFORMATION value;

  nt_status =
    gm_nt_SnarfFromRegistry
    (L"\\Registry\\Machine\\SYSTEM\\CurrentControlSet\\Services\\GM\\Parameters",
     L"EventLogLevel", REG_DWORD, &value);
  if (NT_ERROR (nt_status))
    {
      _gm_nt_event_log_level = LOG_LEVEL_NONE;
      GM_PRINT (GM_PRINT_LEVEL >= 1,
		("Could not snarf event log level from registry.\n"));
    }
  else
    {
      _gm_nt_event_log_level =
	(gm_u32_t) * ((char *) value + value->DataOffset);
      gm_nt_FreeSnarfedData (value);
    }

  GM_NOTE (("GM event logging enabled.\n"));
}

/*
 * More required kernel debugging functions.
 */

void gm_arch_abort (void)
{
  DbgPrint ("gm_arch_abort\n");
  DbgBreakPoint ();
}


/************************************************************************
 * Driver entry points
 ************************************************************************/

static NTSTATUS
gm_nt_DispatchCreate (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
  NTSTATUS status;

  gm_nt_enter ();
  if (!is_gm_device (DeviceObject)) {
    status = ndis.create (DeviceObject, Irp);
  }
  else {
    status = STATUS_SUCCESS;
    Irp->IoStatus.Status = status;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest (Irp, IO_NO_INCREMENT);
  }
  gm_nt_exit ();
  return status;
}

static NTSTATUS
gm_nt_DispatchClose (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
  NTSTATUS status;

  gm_nt_enter ();
  if (!is_gm_device (DeviceObject)) {
    status = ndis.close (DeviceObject, Irp);
  }
  else {
    gm_port_state_t* ps;
   
    ps = gm_nt_device_to_ps (DeviceObject);
    if (ps->arch.is_clone_device) {
    }
    else {
      gm_arch_minor_t minor;
      
      minor = ps->minor;
      gm_port_state_close (ps);
      --gm_usage_count;
      gm_nt_destroy_minor_node (minor);
    }
    status = STATUS_SUCCESS;
    Irp->IoStatus.Status = status;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest (Irp, IO_NO_INCREMENT);
  }
  gm_nt_exit ();
  return status;
}

/* Handle I/O request messages which have been queued for each device.
   This function will be called only when the process is not handling
   any other requests for the same device.  However, the driver may be
   processing I/O for other devices concurrently.

   NOTE: returns only STATUS_SUCCESS or STATUS_PENDING.  The status to
   be returned to the user is set in Irp->IoStatus.Status.

   NOTE NOTE: Returning STATUS_SUCCESS doesn't seem to work correctly.
   Already opened ports are allowed to be re-opened. Even though the Irp
   is marked as STATUS_INSUFFICIENT_RESOURCES. Now gm_nt_ioctl returns
   a status that is what the Irp says. feldy - 990721

   Runs at PASSIVE_LEVEL in user context. */

static NTSTATUS
gm_nt_DispatchDeviceControl (IN PDEVICE_OBJECT DeviceObject,
                             IN PIRP Irp)
{
  NTSTATUS status;

  gm_nt_enter ();
  if (!is_gm_device (DeviceObject)) {
    status = ndis.device_control (DeviceObject, Irp);
  }
  else {
    status = gm_nt_ioctl (DeviceObject, Irp);
  }
  gm_nt_exit ();
  return status;
}

/* The Dispatch Cleanup routine cancels all IRPs associated with a
   device. This is done by cancelling all IRPs in the worker thread
   queue.

   BAD: If this is the daemon port, then this function should cleanup
   all ports. */

static NTSTATUS
gm_nt_DispatchCleanup (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
  NTSTATUS status;

  gm_nt_enter ();
  if (!is_gm_device (DeviceObject)) {
    status = ndis.cleanup (DeviceObject, Irp);
  }
  else {
    gm_port_state_t* ps;
    KIRQL irql;

    ps = gm_nt_device_to_ps (DeviceObject);
    IoAcquireCancelSpinLock (&irql);
    while (!IsListEmpty (&ps->arch.sleep_q_root)) {
      gm_nt_cancel_last (DeviceObject, &ps->arch.sleep_q_root, irql);
    }
    IoReleaseCancelSpinLock (irql);
    status = STATUS_SUCCESS;
    if (Irp) {
      Irp->IoStatus.Status = status;
      Irp->IoStatus.Information = 0;
      IoCompleteRequest (Irp, IO_NO_INCREMENT);
    }
  }
  gm_nt_exit ();
  return status;
}

static NTSTATUS
gm_nt_DispatchPnp (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
  NTSTATUS status;
  PIO_STACK_LOCATION irpStack;

  irpStack = IoGetCurrentIrpStackLocation (Irp);
  if (irpStack->MinorFunction == IRP_MN_QUERY_REMOVE_DEVICE &&
      gm_usage_count >0) {
    status = STATUS_UNSUCCESSFUL;
    Irp->IoStatus.Status = status;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest (Irp, IO_NO_INCREMENT);
  }
  else {
    status = ndis.pnp (DeviceObject, Irp);
  }
  return status;
}

VOID
gm_nt_Unload (IN PDRIVER_OBJECT DriverObject)
{
  ndis.unload (DriverObject);
}


/* The driver entry routine

   Note: the name of the function cannot be changed
   Note: GM_NOTE and GM_WARN cannot be used until after
         gm_nt_initialize_event_log */

gm_u32_t unit_count;

#define GM_DEBUG_DRIVER_ENTRY 0
NTSTATUS
DriverEntry (IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath)
{
  NDIS_HANDLE wrapper;
  NDIS50_MINIPORT_CHARACTERISTICS mp_char;
  NDIS_STATUS status;

  GM_PRINT (GM_DEBUG_DRIVER_ENTRY, ("enter DriverEntry\n"));
  GM_PRINT (GM_DEBUG_DRIVER_ENTRY, ("DriverObject = %p\n", DriverObject));
  GM_PRINT (GM_DEBUG_DRIVER_ENTRY, ("RegistryPath = %ws\n",
                                    RegistryPath->Buffer));

  the_one_and_only_driver_object = DriverObject;
  unit_count = 0;

  if (gm_init () != GM_SUCCESS) {
    status = STATUS_UNSUCCESSFUL;
    goto abort_with_nothing;
  }
  gm_arch_sync_init (&_gm_arch_serialize_sync, 0);
  gm_nt_enter ();
  if (gm_arch_get_page_len (&GM_PAGE_LEN) != GM_SUCCESS) {
    status = STATUS_UNSUCCESSFUL;
    goto abort_with_gm_init;
  }
  gm_device_hash = gm_create_hash (gm_hash_compare_ptrs,
                                   gm_hash_hash_ptr, 0, 0, 2, 0);
  if (gm_device_hash == NULL) {
    status = STATUS_INSUFFICIENT_RESOURCES;
    goto abort_with_gm_init;
  }
  gm_nt_phys_hash = gm_create_hash (gm_nt_phys_key_compare,
                                    gm_nt_phys_key_hash,
                                    sizeof (gm_nt_phys_key),
                                    sizeof (gm_nt_phys_data), 0, 0);
  if (gm_nt_phys_hash == NULL) {
    status = STATUS_INSUFFICIENT_RESOURCES;
    goto abort_with_gm_device_hash;
  }

  NdisMInitializeWrapper (&wrapper, DriverObject, RegistryPath, NULL);

  NdisZeroMemory (&mp_char, sizeof (mp_char));
  mp_char.MajorNdisVersion = 5;
  mp_char.MinorNdisVersion = 0;
  mp_char.CheckForHangHandler = gmm_CheckForHang;
  mp_char.HaltHandler = gmm_Halt;
  mp_char.InitializeHandler = gmm_Initialize;
  mp_char.QueryInformationHandler = gmm_QueryInformation;
  mp_char.ResetHandler = gmm_Reset;
  mp_char.SetInformationHandler = gmm_SetInformation;
  mp_char.ReturnPacketHandler = gmm_ReturnPacket;
  mp_char.SendPacketsHandler = gmm_SendPackets;
  status = NdisMRegisterMiniport (wrapper,
                                  (PNDIS_MINIPORT_CHARACTERISTICS) &mp_char,
                                  sizeof (mp_char));
  if (status != NDIS_STATUS_SUCCESS) {
    goto abort_with_wrapper;
  }

  NdisMRegisterUnloadHandler (wrapper, gmm_Unload);

  ndis.create = DriverObject->MajorFunction[IRP_MJ_CREATE];
  ndis.close = DriverObject->MajorFunction[IRP_MJ_CLOSE];
  ndis.cleanup = DriverObject->MajorFunction[IRP_MJ_CLEANUP];
  ndis.device_control
    = DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL];
  ndis.pnp = DriverObject->MajorFunction[IRP_MJ_PNP];
  ndis.unload = DriverObject->DriverUnload;

  DriverObject->MajorFunction[IRP_MJ_CREATE] = gm_nt_DispatchCreate;
  DriverObject->MajorFunction[IRP_MJ_CLOSE] = gm_nt_DispatchClose;
  DriverObject->MajorFunction[IRP_MJ_CLEANUP] = gm_nt_DispatchCleanup;
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]
    = gm_nt_DispatchDeviceControl;
  DriverObject->MajorFunction[IRP_MJ_PNP] = gm_nt_DispatchPnp;
  DriverObject->DriverUnload = gm_nt_Unload;

  gm_nt_exit ();

  GM_PRINT (GM_DEBUG_DRIVER_ENTRY, ("DriverEntry exiting with status %lx\n",
                                    (unsigned long) status));

  return status;

 abort_with_wrapper:
  NdisTerminateWrapper (wrapper, NULL);
  gm_destroy_hash (gm_nt_phys_hash);
 abort_with_gm_device_hash:
  gm_destroy_hash (gm_device_hash);
 abort_with_gm_init:
  gm_finalize ();
 abort_with_nothing:
  return status;
}


/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
