#include <ndis.h>
#include "gmm.h"

NDIS_STATUS
gmm_allocate_recv_item (NDIS_HANDLE handle, gmm_recv_item * item,
			NDIS_HANDLE packet_pool, NDIS_HANDLE buffer_pool)
{
  NDIS_STATUS status;

  NdisMAllocateSharedMemory (handle, GM_ETHERNET_OVER_MYRINET_MTU, FALSE,
			     &item->va, &item->pa);
  if (item->va == NULL)
    {
      status = NDIS_STATUS_FAILURE;
      goto abort_with_nothing;
    }
  NdisAllocatePacket (&status, &item->packet, packet_pool);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_shmem;
    }
  NDIS_SET_PACKET_HEADER_SIZE (item->packet, GMM_ETHERNET_HEADER_SIZE);
  NdisAllocateBuffer (&status, &item->buffer, buffer_pool,
		      item->va, GM_ETHERNET_OVER_MYRINET_MTU);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_packet;
    }
  NdisChainBufferAtFront (item->packet, item->buffer);
  *(gmm_recv_item **) & item->packet->MiniportReserved = item;
  return NDIS_STATUS_SUCCESS;

 abort_with_packet:
  NdisFreePacket (item->packet);
 abort_with_shmem:
  NdisMFreeSharedMemory (handle, GM_ETHERNET_OVER_MYRINET_MTU, FALSE,
			 item->va, item->pa);
 abort_with_nothing:
  return status;
}

NDIS_STATUS
gmm_free_recv_item (NDIS_HANDLE handle, gmm_recv_item * item,
		    NDIS_HANDLE packet_pool, NDIS_HANDLE buffer_pool)
{
  NdisFreeBuffer (item->buffer);
  NdisFreePacket (item->packet);
  NdisMFreeSharedMemory (handle, GM_ETHERNET_OVER_MYRINET_MTU, FALSE,
			 item->va, item->pa);
  return STATUS_SUCCESS;
}

NDIS_STATUS
gmm_allocate_recv_ring (gmm_Context * context)
{
  int i;
  int j;
  gmm_recv_item *item;
  NDIS_STATUS status;

  NdisAllocatePacketPool (&status, &context->recv_packet_pool,
			  GMM_NUM_RECV_BUFFERS, 4 * sizeof (PVOID));
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_nothing;
    }
  NdisAllocateBufferPool (&status, &context->recv_buffer_pool,
			  GMM_NUM_RECV_BUFFERS);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_packet_pool;
    }
  NdisAllocateSpinLock (&context->recv_lock);

  item = context->recv_ring;
  for (i = 0; i < GMM_NUM_RECV_BUFFERS; ++i)
    {
      status = gmm_allocate_recv_item (context->miniportAdapterHandle, item,
				       context->recv_packet_pool,
				       context->recv_buffer_pool);
      if (status != NDIS_STATUS_SUCCESS)
	{
	  goto abort_with_items;
	}
      item->offset = i;
      ++item;
    }
  return NDIS_STATUS_SUCCESS;

 abort_with_items:
  item = context->recv_ring;
  for (j = 0; j < i; ++j)
    {
      gmm_free_recv_item (context->miniportAdapterHandle, item,
			  context->recv_packet_pool,
			  context->recv_buffer_pool);
      ++item;
    }
  NdisFreeSpinLock (&context->recv_lock);
  NdisFreeBufferPool (context->recv_buffer_pool);
 abort_with_packet_pool:
  NdisFreePacketPool (context->recv_packet_pool);
 abort_with_nothing:
  return NDIS_STATUS_FAILURE;
}

void
gmm_provide_recv_ring (gmm_Context * context)
{
  int i;
  gmm_recv_item *item;
  gm_ethernet_segment_descriptor_t segment;

  item = context->recv_ring;
  for (i = 0; i < GMM_NUM_RECV_BUFFERS; ++i)
    {
      segment.ptr = item->pa.QuadPart;
      segment.len = GM_ETHERNET_OVER_MYRINET_MTU;
      gm_ethernet_provide_scatter_list_with_tag (context->port, 1,
                                                 &segment, i);
      ++item;
    }
}

void
gmm_free_recv_ring (gmm_Context * context)
{
  int i;
  gmm_recv_item *item;

  item = context->recv_ring;
  for (i = 0; i < GMM_NUM_RECV_BUFFERS; ++i)
    {
      gmm_free_recv_item (context->miniportAdapterHandle, item,
			  context->recv_packet_pool,
			  context->recv_buffer_pool);
      ++item;
    }

  NdisFreeSpinLock (&context->recv_lock);
  NdisFreeBufferPool (context->recv_buffer_pool);
  NdisFreePacketPool (context->recv_packet_pool);
}

VOID
gmm_ReturnPacket (NDIS_HANDLE miniportAdapterContext, PNDIS_PACKET packet)
{
  gmm_Context *context = (gmm_Context *) miniportAdapterContext;
  gmm_recv_item *item;
  gm_ethernet_segment_descriptor_t segment;

  NdisAcquireSpinLock (&context->recv_lock);
  item = *((gmm_recv_item **) packet->MiniportReserved);
  context->recv_ring_count--;

  segment.ptr = item->pa.QuadPart;
  segment.len = GM_ETHERNET_OVER_MYRINET_MTU;
  gm_ethernet_provide_scatter_list_with_tag (context->port, 1,
                                             &segment, item->offset);
  NdisReleaseSpinLock (&context->recv_lock);
}

void
gmm_receiveInterrupt (void *p, gm_ethernet_recv_t *e)
{
  gmm_Context *context = (gmm_Context*) p;
  gmm_recv_item *item;
  gm_u32_t length = gm_ntoh_u32 (e->length);
  gm_u16_t csum = gm_ntoh_u16 (e->ip_checksum);
  gm_u8_t index = gm_ntoh_u8 (e->tag);

  NdisAcquireSpinLock (&context->recv_lock);
  item = context->recv_ring+index;
  context->recv_ring_count++;

  NdisAdjustBufferLength (item->buffer, length);
  NDIS_SET_PACKET_STATUS (item->packet, NDIS_STATUS_RESOURCES);
  context->rcv_ok++;
  NdisReleaseSpinLock (&context->recv_lock);
  NdisMIndicateReceivePacket (context->miniportAdapterHandle, &item->packet,
			      1);
  gmm_ReturnPacket (context, item->packet);
}
