/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_alloc_pages.c
*/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_malloc_debug.h"

static
void *
_gm_alloc_pages (gm_size_t alloc_len)
{
  void *ret;

  gm_assert (alloc_len > GM_PAGE_LEN);
  
  /* Handle larger-than-a-page case */

#if GM_KERNEL && GM_KERNEL_MALLOC_ALIGNS_PAGES
  ret = gm_malloc (alloc_len, 0);
#elif GM_KERNEL && defined WIN32
  ret = ExAllocatePool (NonPagedPool, alloc_len);
#elif !GM_KERNEL && defined WIN32
  ret = VirtualAlloc (NULL, alloc_len, MEM_COMMIT, PAGE_READWRITE);
#else
  {
    void *tmp, *page;

    if (!GM_PAGE_LEN)
      return 0;

    tmp = gm_malloc (alloc_len + GM_PAGE_LEN + sizeof (void *));
    if (!tmp)
      return 0;

    /* reserve room for pointer to beginning of buffer. */

    page = (void *) ((char *) tmp + sizeof (void *));

    /* round page to a page boundary. */

    if (((long) page & (GM_PAGE_LEN - 1)) != 0)
      page = (void *) ((char *) page
		       + GM_PAGE_LEN - ((long) page & (GM_PAGE_LEN - 1)));

    GM_PRINT (0, ("Storing pointer at %p.\n", &((void **) page)[-1]));
    ((void **) page)[-1] = tmp;
    ret = page;
  }
#endif
  return ret;
}

static
void
_gm_free_pages (void *addr, gm_size_t alloc_len)
{
  gm_assert (alloc_len > GM_PAGE_LEN);

  /* Handle larger-than-a-page case */

#if GM_KERNEL && GM_KERNEL_MALLOC_ALIGNS_PAGES
  gm_free (addr);
#elif GM_KERNEL && defined WIN32
  ExFreePool (addr);
#elif !GM_KERNEL && defined WIN32
  VirtualFree (addr, 0, MEM_RELEASE);
#else
  GM_PRINT (0,
	    ("gm_free_pages: Retrieving pointer at %p.\n",
	     &((void **) addr)[-1]));
  gm_free (((void **) addr)[-1]);
#endif
}

/**
   gm_alloc_pages() allocates a page-aligned buffer of length ALLOC_LEN,
   where ALLOC_LEN is a multiple of GM_PAGE_LEN.  Any fractional page
   following the buffer is wasted.
 
   @retval ptr Pointer to the allocated buffer.
   @retval 0 Error occurred.
 
   @param alloc_len \b (IN) The length of buffer to be allocated.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT void *
gm_alloc_pages (gm_size_t alloc_len)
{
  void *ret;
  GM_CALLED ();

  /* special case for allocations smaller than a page */
  
  if (alloc_len <= GM_PAGE_LEN)
    {
      GM_RETURN (gm_page_alloc ());
    }

  ret = _gm_alloc_pages (alloc_len);
  if (!ret)
    goto abort_with_nothing;
  if (GM_MDEBUG_RECORD_PTR (ret, gm_free_pages) != GM_SUCCESS)
    goto abort_with_ret;
  
  GM_RETURN (ret);

 abort_with_ret:
  _gm_free_pages (ret, alloc_len);
 abort_with_nothing:
  GM_RETURN (0);
}

/**
   gm_free_pages() frees the pages at \b addr, which were previously
   allocated with gm_alloc_pages().
 
   @param addr \b (IN) The address of the buffer to be freed.
 
   @param alloc_len \b (IN) The length (in bytes) of the buffer to be freed.
 
   @see gm_alloc_pages
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT void
gm_free_pages (void *addr, gm_size_t alloc_len)
{
  GM_CALLED ();

  if (alloc_len <= GM_PAGE_LEN)
    {
      gm_page_free (addr);
      GM_RETURN_NOTHING ();
    }
  
  GM_MDEBUG_REMOVE_PTR (addr, gm_free_pages);
  _gm_free_pages (addr, alloc_len);
  GM_RETURN_NOTHING ();
}

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
