/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_blocking_receive.c
*/

#include "gm_call_trace.h"
#include "gm_cmp64.h"
#include "gm_compiler.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_set_alarm.h"

#define GM_DEBUG_BLOCKING_RECEIVE 0

#if GM_DEBUG_BLOCKING_RECEIVE
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

static void
_never_called (void *x)
{
  GM_PARAMETER_MAY_BE_UNUSED (x);
  GM_PANIC (("should never get here\n"));
}

/**
   gm_blocking_receive() blocks until there is a receive event
   and then returns a pointer to the event.  If no send is
   immediately available, this call suspends the current process
   until a receive event is available.  As an optimization for
   applications with one CPU per CPU-intensive thread, this function
   polls for receives for one millisecond before sleeping the
   process, so it is not suited for machines running more than one
   performance critical process or thread on the machine.
 
   @retval GM_SUCCESS Operation completed successfully.
   @retval GM_NO_RECV_EVENT Handle all flushed alarm events, which the
             user never needs to know about, and which must not cause
             my_alarm to be cancelled or reset.
   @retval GM_FLUSHED_ALARM_EVENT Intercept my_alarm.
   @retval GM_ALARM_EVENT If our alarm went off, block.  Be careful to
             not handle any other alarms at this time since we are not
             passing the GM_ALARM_EVENT to the user and we want to maintain
             the semantics that user alarms are called only inside
             gm_unknown() when it is called by the user.
 
   @param p \b (IN) The GM port in use.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)

*/

GM_ENTRY_POINT
gm_recv_event_t *
gm_blocking_receive (gm_port_t * p)
{
  gm_recv_event_t *ret;
  gm_alarm_t my_alarm, *ap;
  gm_u64_t rtc;

  GM_CALLED_WITH_ARGS (("%p", p));

  /* Check for message already pending */

  ret = gm_receive (p);
  if (gm_ntoh_u8 (ret->recv.type) != GM_NO_RECV_EVENT)
    GM_RETURN_PTR (ret);

  /* No message is pending.  Arrange to go to sleep if no message is
     received in 1ms. */

  gm_initialize_alarm (&my_alarm);
  gm_set_alarm (p, &my_alarm, 1000, _never_called, p);
  
  /* Spin, until an event is received, and pass the event to the user.
     Make special provisions for our timer, which we want to cancel if
     and only if it doesn't get triggered before a user event is received. */

  while (1)
    {
      ret = gm_receive (p);
      switch (gm_ntoh_u8 (ret->recv.type))
	{
	case GM_NO_RECV_EVENT:
	  continue;
	  
	  /* Handle all flushed alarm events, which the user never
	     needs to know about, and which must not cause my_alarm to
	     be cancelled or reset. */

	case _GM_FLUSHED_ALARM_EVENT:
	  _gm_handle_flushed_alarm (p);
	  continue;

	  /* Intercept my_alarm. */

	case GM_ALARM_EVENT:
	  {
	    /* If our alarm went off, block.  Be careful to not handle
	       any other alarms at this time since we are not passing
	       the GM_ALARM_EVENT to the user and we want to maintain
	       the semantics that user alarms are called only inside
	       gm_unknown() when it is called by the user. */

	    rtc = gm_ticks (p);
	    for (ap = p->first_alarm; ap; ap = ap->next)
	      {
		if ((gm_cmp64 (ap->deadline, rtc) <= 0) &&
		    (ap->callback != my_alarm.callback))
		{
		  /*  A user alarm expired, so return it to the user.*/
		  goto return_to_user;
		}
	      }

	    /* Only our alarm went off, so it is safe to consume the
	       event and block */

 	    gm_cancel_alarm (&my_alarm);
	    p->lanai_alarm_state = LANAI_ALARM_IDLE;
	    _reschedule_lanai_alarm_if_needed (p);
	    GM_RETURN_PTR (gm_blocking_receive_no_spin (p));

	  }
	  /* fall through */
	  
	  /* Pass all other events, (including GM_SLEEP events) to the
	     user, and cancel the alarm.  Either this is an event the
	     user cares about, or it is a GM_SLEEP event, which will
	     put the user process to sleep.  Note that we may be
	     cancelling my_alarm after it has fired, but that's
	     harmless. */

	default:
	  GM_PRINT (GM_DEBUG_BLOCKING_RECEIVE,
		    ("Cancelling alarm for event of type %d.\n",
		     (int) gm_ntoh_u8 (ret->recv.type)));
	  goto return_to_user;
	}
    }

 return_to_user:
  gm_cancel_alarm (&my_alarm);
  GM_RETURN_PTR (ret);
}


/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
