/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_close.c
*/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_debug_open.h"
#include "gm_dma_malloc.h"
#include "gm_enable_trace.h"
#include "gm_internal.h"
#include "gm_ptr_hash.h"
#include "gm_trace.h"

#if !GM_KERNEL
#  include <stdio.h>
   
				/* close() */
#  ifdef HAVE_UNISTD_H
#    include <unistd.h>
#  endif
   
				/* munmap() */
#  ifdef HAVE_SYS_TYPES_H
#    include <sys/types.h>
#  endif
#  ifdef HAVE_SYS_MMAN_H
#    include <sys/mman.h>
#  endif
#endif


/**
   gm_close() closes a previously opened port \b p, and frees all
   resources associated with that port.
 
   @param p \b (IN) The GM port to be closed.
 
   @see gm_open gm_init gm_exit gm_finalize gm_abort
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT void
gm_close (gm_port_t *p)
{
  GM_CALLED_WITH_ARGS (("%p", p));
  
  /****
   * Remove the port from the list of open ports.
   ****/

  if (!p)
    {
      GM_RETURN_NOTHING ();
    }
  
  GM_PRINT(GM_DEBUG_OPEN, ("gm_close() entered for unit %d, port %d\n",
			   p->unit, p->id));

  _gm_open_ports_remove (p);

  /****
   * Free resources allocated since gm_open()
   ****/

  /* Free the buffer allocated in _gm_get_kernel_build_id */

  if (p->kernel_build_id)
    gm_free (p->kernel_build_id);

  /* Free the allocated DMA memory zones and unmap the associated
     memory. */
  GM_PRINT(GM_DEBUG_OPEN,
           ("gm_close() calling _gm_dma_finalize() for unit %d, port %d\n",
	    p->unit, p->id));

  /* Perform all DMA frees before finalizaing DMA support. */

  if (p->recv_queue_allocation)
    {
      gm_dma_free (p, p->recv_queue_allocation);
      p->recv_queue_allocation = 0;
    }
  
  /* Finalize DMA support, freeing all DMA blocks. */
  
  _gm_dma_finalize (p);

  /****
   * Free resources allocated in gm_open(). 
   ****/

  gm_destroy_ptr_hash (p->global_id_cache.hash);
  gm_free (p->global_id_cache.array);
  if (p->first_sram_page)
    _gm_munmap (p, (void *)p->first_sram_page,
		(unsigned long) (p->mappings->RTC.len));
  if (p->send_queue_start)
    _gm_munmap (p, (void *)p->send_queue_start,
		(unsigned long) (p->mappings->send_queue.len));
  if (p->mappings)
    gm_free (p->mappings);

#if GM_KERNEL
  gm_port_state_close (p->kernel_port_state);
  gm_minor_free (p->kernel_port_state->minor);
#else 
  GM_PRINT(GM_DEBUG_OPEN,
           ("gm_close() calling system close function for unit %d, port %d\n",
	    p->unit, p->id));
#  ifdef _WIN32
  CloseHandle (p->fd);
#  else
  close (p->fd);
#  endif
#endif
  gm_free (p);
  gm_finalize ();

  GM_RETURN_NOTHING ();
}
