/**
   @file gm_crc.c
   This file contains the GM API functions, gm_crc() and gm_crc_str(),
   which compute 32-bit CRCs on the contents of memory. These functions are
   not guaranteed to perform any particular variant of the CRC-32, but these
   functions are useful for creating robust hashing functions.
*/

#include "gm_internal.h"
#include "gm_crc32.h"

#if GM_CRC32_IN_HARDWARE
#include "gm_bootstrap.h"	/* for get_ISR() */
#endif

/**
   gm_crc() computes a CRC-32 of the indicated range of memory.
 
   @retval long
 
   @param _ptr \b (IN) Pointer to a range of memory.
 
   @param len \b (IN) The length of the indicated range of memory.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT unsigned long
gm_crc (void *_ptr, gm_size_t len)
{
  unsigned long ret;
  
#if GM_CRC32_IN_HARDWARE
  /* We know we are in the single-threaded LANai environment, so it is
     safe to use the non-thread-safe hardware CRC engine and COPY
     engine. */
  {
    GM_CRC32_SET (0xffffffffU);
    while (get_ISR() & COPY_BUSY) ;
    COPY = (gm_u32_t) _ptr;
    COPY = len | COPY_TO_CRC32;
    while (get_ISR() & COPY_BUSY) ;
    ret = ~GM_CRC32_GET ();
  }
#else
  /* We may be in a multi-threaded environment, so use a thread-safe
     software CRC. */
  {
    gm_u8_t *ptr;
    gm_u32_t accum = 0xffffffffU;

    for (ptr = (gm_u8_t *) _ptr;
	 len--;
	 ptr++)
      {
	gm_crc32_u8 (*ptr, &accum);
      }
    ret = (unsigned long) ~accum;
  }
#endif
  
  gm_assert (sizeof (unsigned long) >= sizeof (gm_u32_t));
  return ret;
}

/**
   gm_crc_str() computes a CRC-32 for the indicated string.
 
   @retval long
 
   @param ptr \b (IN) Pointer to a string.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT unsigned long
gm_crc_str (const char *ptr)
{
  unsigned long ret;
  
#if GM_CRC32_IN_HARDWARE
  /* We know we are in the single-threaded LANai environment, so
     it is safe to use the non-thread-safe hardware CRC engine */
  {
    GM_CRC32_SET (0xffffffffU);
    for (; *ptr; ptr++)
      {
	GM_CRC32_BYTE ((gm_u8_t) *ptr);
      }
    ret = ~GM_CRC32_GET ();
  }
#else
  /* We may be in a multi-threaded environment, so use a thread-safe
     software CRC. */
  {
    gm_u32_t accum = 0xffffffffU;
    while (*ptr)
      {
	gm_crc32_u8 ((gm_u8_t) *ptr++, &accum);
      }
    ret = (unsigned long) ~accum;
  }
#endif
  
  gm_assert (sizeof (unsigned long) >= sizeof (gm_u32_t));
  return ret;
}
