/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2002 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_config.h"
#include "gm_crc32.h"
#include "gm_debug.h"
#include "gm_internal_funcs.h"
#if !GM_KERNEL
#include <stdio.h>
#endif

/* Global CRC-32 tables.  The should be accessed only using the
   GM_CRC32_TABLE_ENTRY() macro. */
struct __gm_crc32_tables __gm_crc32_tables;

#if !GM_CRC32_IN_HARDWARE
/* Globals CRC-32 accumulator for LANaiX CRC32 abstraction. */
gm_u32_t GM_CRC32;
#endif

/* Initialize a software CRC table.  This does not need to be fast. */

static
void
gm_init_crc32_table (int N)
{
  int i, j;

  /* For each entry in the table */
  for (i=0; i<(1<<N); i++)
    {
      gm_u32_t accum;
      int in = 0;
      
      /* Put "i" in the top N bits of the input variable. */
      accum = ((gm_u32_t) i) << (8*sizeof(in) - N);

      /* Compute the CRC-32 1 bit at a time. */
      for (j=0; j<N; j++)
	{
	  GM_CRC32_COMPUTE_1_BIT (in, accum);
	}

      /* Record the precomputed table entry. */

      /* This switch HACK is required to allow a preprocessing hack
	 int GM_CRC32_TABLE_ENTRY() to work.  It's worth it for the
	 performance and memory savings, though. */
      switch (N)
	{
#if GM_NO_CACHE
	case 8: GM_CRC32_TABLE_ENTRY(8,i) = accum; break;
#elif GM_CACHE_LINE_LEN == 128
	case 4: GM_CRC32_TABLE_ENTRY(4,i) = accum; break;
#elif GM_CACHE_LINE_LEN == 32
	case 3: GM_CRC32_TABLE_ENTRY(3,i) = accum; break;
#else
#error This should never happen.
#endif
	default:
	  gm_abort ();
	  break;
	}
	  
      /* fprintf (stderr, "%d->0x%08x\n", i, accum); */
    }
}

/* Question: Does anyone have some test cases for confirming that AAL5 CRC-
   32 software works on a given machine?

   Answer: There are four examples of valid AAL-5 CS-PDU in I.363: */

struct gm_crc32_test_vector 
{
  gm_u8_t data[44];
  gm_u32_t crc32;
  unsigned int length;
};

#define U (unsigned char)

static struct gm_crc32_test_vector crc32_test_vector[] = {
  {
    /* 40 Octets filled with "0" */
    /* CPCS-UU = 0, CPI = 0, Length = 40, CRC-32 = 864d7f99 */
    {
      U 0x00, U 0x00, U 0x00, U 0x00,
      U 0x00, U 0x00, U 0x00, U 0x00,
      U 0x00, U 0x00, U 0x00, U 0x00,
      U 0x00, U 0x00, U 0x00, U 0x00,
      U 0x00, U 0x00, U 0x00, U 0x00,
      U 0x00, U 0x00, U 0x00, U 0x00,
      U 0x00, U 0x00, U 0x00, U 0x00,
      U 0x00, U 0x00, U 0x00, U 0x00,
      U 0x00, U 0x00, U 0x00, U 0x00,
      U 0x00, U 0x00, U 0x00, U 0x00,
      U 0x00, U 0x00, U 0x00, U 0x28
    },
    0x864d7f99U, 44
  },
  /* 40 Octets filled with "1" */
  /* CPCS-UU = 0, CPI = 0, Length = 40, CRC-32 = c55e457a */
  {
    {
      U 0xff, U 0xff, U 0xff, U 0xff,
      U 0xff, U 0xff, U 0xff, U 0xff,
      U 0xff, U 0xff, U 0xff, U 0xff,
      U 0xff, U 0xff, U 0xff, U 0xff,
      U 0xff, U 0xff, U 0xff, U 0xff,
      U 0xff, U 0xff, U 0xff, U 0xff,
      U 0xff, U 0xff, U 0xff, U 0xff,
      U 0xff, U 0xff, U 0xff, U 0xff,
      U 0xff, U 0xff, U 0xff, U 0xff,
      U 0xff, U 0xff, U 0xff, U 0xff,
      U 0x00, U 0x00, U 0x00, U 0x28
    },
    0xc55e457aU, 44
  },
  /* 40 Octets counting: 1 to 40 */
  /* CPCS-UU = 0, CPI = 0, Length = 40, CRC-32 = bf671ed0 */
  {
    {
      U 0x01, U 0x02, U 0x03, U 0x04,
      U 0x05, U 0x06, U 0x07, U 0x08,
      U 0x09, U 0x0a, U 0x0b, U 0x0c,
      U 0x0d, U 0x0e, U 0x0f, U 0x10,
      U 0x11, U 0x12, U 0x13, U 0x14,
      U 0x15, U 0x16, U 0x17, U 0x18,
      U 0x19, U 0x1a, U 0x1b, U 0x1c,
      U 0x1d, U 0x1e, U 0x1f, U 0x20,
      U 0x21, U 0x22, U 0x23, U 0x24,
      U 0x25, U 0x26, U 0x27, U 0x28,
      U 0x00, U 0x00, U 0x00, U 0x28
    },
    0xbf671ed0, 44
  },
  /* 40 Octets counting: 1 to 40 */
  /* CPCS-UU = 11, CPI = 22, CRC-32 = acba602a */
  {
    {
      U 0x01, U 0x02, U 0x03, U 0x04,
      U 0x05, U 0x06, U 0x07, U 0x08,
      U 0x09, U 0x0a, U 0x0b, U 0x0c,
      U 0x0d, U 0x0e, U 0x0f, U 0x10,
      U 0x11, U 0x12, U 0x13, U 0x14,
      U 0x15, U 0x16, U 0x17, U 0x18,
      U 0x19, U 0x1a, U 0x1b, U 0x1c,
      U 0x1d, U 0x1e, U 0x1f, U 0x20,
      U 0x21, U 0x22, U 0x23, U 0x24,
      U 0x25, U 0x26, U 0x27, U 0x28,
      U 0x11, U 0x22, U 0x00, U 0x28
    },
    0xacba602a, 44
  }
};

/* Test the various CRC-32 implementations. */

void
gm_crc32_test (void)
{
  int test_number;
  int failed = 0;
  
  /* Test using each test vector. */
  
  for (test_number=0; test_number<4; test_number++)
    {
      struct gm_crc32_test_vector *vector;
      int width;
      
      vector = &crc32_test_vector[test_number];

      /* Test the GM API CRC function. */

      gm_always_assert
	(gm_crc (vector->data, vector->length) == vector->crc32);
      
      /* Test each width of computation. */
      
      for (width=8; width<=32; width*=2)
	{
	  int hw;
	  
	  /* Test both software and hardware (or hardware emulation)
	     implementations. */
	  
	  for (hw=0; hw<=1; hw++)
	    {
	      gm_u32_t accum;
	      unsigned int i;
	      
	      /* Initialize the accumulator */
	  
	      if (hw)
		GM_CRC32_SET (0xffffffffU);
	      else
		accum = 0xffffffffU;

	      /* Iterate */

	      i = 0;
	      while (i<vector->length)
		{
		  unsigned in;

		  in = vector->data[i++];
		  switch (width)
		    {
		    case 8:
		      if (hw)
			GM_CRC32_BYTE ((gm_u8_t) in);
		      else
			gm_crc32_u8 ((gm_u8_t) in, &accum);
		      break;
		  
		    case 16:
		      in = (in << 8) | vector->data[i++];
		      if (hw)
			GM_CRC32_HALF ((gm_u16_t) in);
		      else
			gm_crc32_u16 ((gm_u16_t) in,  &accum);
		      break;

		    case 32:
		      in = (in << 8) | vector->data[i++];
		      in = (in << 8) | vector->data[i++];
		      in = (in << 8) | vector->data[i++];
		      if (hw)
			GM_CRC32_WORD ((gm_u32_t) in);
		      else
			gm_crc32_u32 ((gm_u32_t) in,  &accum);
		      break;
		    }
		}

	      /* Check result. */
	      
	      {
		int this_test_failed;

		if (hw)
		  this_test_failed = (vector->crc32 != ~GM_CRC32_GET ());
		else
		  this_test_failed = (vector->crc32 != ~accum);

		if (this_test_failed)
		  {
		    GM_WARN
		      (("crc32 failed for test_number=%d hw=%d width=%d\n",
			test_number, hw, width));
		    failed = 1;
		  }
	      }
	    }
	}
    }
  gm_always_assert (!failed);
}

/* Initialize the tables used by this package. */

void
gm_init_crc32 (void)
{

  /* Initialize any software CRC tables. */
  
#if GM_NO_CACHE
  gm_init_crc32_table (8);
#elif GM_CACHE_LINE_LEN == 128
  gm_init_crc32_table (4);
#elif GM_CACHE_LINE_LEN == 32
  gm_init_crc32_table (3);
#else
#error
#endif


#if !GM_BUILDING_FIRMWARE
  /* Make sure the CRC implementation is working. */
  
  gm_crc32_test ();
#endif
}
