/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_datagram_send.c
   This file contains the GM API function gm_datagram_send().
*/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_enable_fast_small_send.h"
#include "gm_enable_trace.h"
#include "gm_internal.h"
#include "gm_send_queue.h"

#define GM_DEBUG_DATAGRAM_SEND 0

#if GM_DEBUG_DATAGRAM_SEND
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/**
   gm_datagram_send() queues \b message of length \b length to be
   sent unreliably to a buffer of size \b size at \b target_port_id
   on \b target_node_id.  \b length must be no larger than \b GM_MTU.
   If any network error is encountered while sending the packet, the
   packet is silently and immediately dropped.  After the packet has
   been DMA'ed from host memory, callback(port,context,status) is
   called inside a user invocation of gm_unknown(), reporting the
   \b status of the attempted send.
 
   @param p \b (IN) The GM port on the source/sender GM node from which
                    the communication is being sent.
   @param message \b (IN) The pointer to the data to be communicated.
   @param size \b (IN) The size of the buffer.
   @param len \b (IN) The length in bytes of the array.
   @param priority \b (IN) The priority of the message being communicated.
   @param target_node_id \b (IN) The GM node to which the message is
                       being sent.
   @param target_port_id \b (IN) The GM port on the destination GM node
                       to which the message is being sent.
   @param callback \b (IN) The function called when the send is complete.
   @param context \b (IN) Pointer to an integer or to a structure
                           that is passed to the callback function.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)

*/

GM_ENTRY_POINT
void
gm_datagram_send (gm_port_t *p, void *message, unsigned int size,
		  gm_size_t len, unsigned int priority,
		  unsigned int target_node_id,
		  unsigned int target_port_id,
		  gm_send_completion_callback_t callback,
		  void *context)
{
  GM_CALLED_WITH_ARGS (("%p,%p,%u,%lu,%u,%u,%u,%p,%p",
			p, message, size, len, priority, target_node_id,
			target_port_id, callback, context));

  gm_assert (p);
  gm_assert (message);
  gm_assert (gm_min_size_for_length (len) <= size);
  /* gm_assert (GM_MIN_PRIORITY <= priority); */
  gm_assert (size >= GM_MIN_MESSAGE_SIZE && size <= 32);
  gm_assert (priority == GM_LOW_PRIORITY ||
	       priority == GM_HIGH_PRIORITY);
  gm_assert (len <= GM_MTU);
  
  GM_LOG_EVT (GM_DATAGRAM_SEND_EVENT_0 + size);

#if GM_DEBUG_DATAGRAM_SEND
  gm_printf ("gm_datagram_send_with_callback "
	     "(%p, %p, %d, 0x%lx, %s, %d, %d, %p, %p) called.\n",
	     p, message, size, len,
	     priority ? "GM_HIGH_PRIORITY" : "GM_LOW_PRIORITY",
	     target_node_id, target_port_id, callback, context);
#endif

#if GM_DEBUG_BUFFERS
  {
    struct gm_buf_handle *bh = gm_find_buf (message);
    if (!bh) {
      fprintf (stderr, "gm_datagram_send: got unregistered buffer %p\n",
	       message);
      abort();
    }
    if (bh->size != -1 && bh->size != (int) size) {
      fprintf (stderr, "gm_datagram_send: buffer %p initialized with size %d,"
	       " arrived with size %d\n",
	       message, bh->size, size);
      abort();
    }
    if (bh->status != gm_in_app) {
      fprintf (stderr, "gm_datagram_send: buffer %p should have status %s"
	       " not %s\n",
	       message,
	       gm_get_buf_status_name(gm_in_app),
	       gm_get_buf_status_name(bh->status));
      abort();
    }
    bh->status = gm_in_send;
  }
#endif /* GM_DEBUG_BUFFERS */

  GM_SIMULATE_SIMULATE ( );

  /* Pass send token to LANai.  Care must be taken to flush the message to
     memory before flushing the token, and the token should be flushed to
     memory immediately. */

  {
    struct gm_send_send_event volatile *se;
    struct gm_send_send_event batch_write;

    se = GM_SEND_QUEUE_SLOT (p, callback, context, send);

    if (GM_STRUCT_WRITE_COMBINING)
      {
	batch_write.message = gm_hton_up ((gm_up_t) message);
	batch_write.length = gm_hton_u32 ((gm_u32_t) len);
	batch_write.target_node_id = gm_hton_u16 ((gm_u16_t) target_node_id);
	batch_write.target_subport_id
	  = gm_hton_u8 ((gm_u8_t) GM_SUBPORT (priority, target_port_id));
	batch_write.type
	  = gm_hton_u8 ((gm_u8_t) (GM_DATAGRAM_SEND_EVENT_0 + size));

	if (GM_FAST_SMALL_SEND && len <= GM_FAST_SEND_LEN)
	  {
	    unsigned int i;
	    volatile gm_ptr_t *data;

	    data = (volatile gm_ptr_t *) &se->message;
	    data -= len / sizeof (gm_ptr_t);
	    for (i=0; i<len/GM_SIZEOF_VOID_P; i++)
	      {
		data[i] = ((gm_ptr_t *) message)[i];
	      }
	  }

	GM_COPY_TO_IO_SPACE (*se, batch_write);
      }
    else
      {
	se->message = gm_hton_up ((gm_up_t) message);
	se->target_node_id = gm_hton_u16 ((gm_u16_t) target_node_id);
	se->length = gm_hton_u32 ((gm_u32_t) len);
	se->target_subport_id =
	  gm_hton_u8 ((gm_u8_t) GM_SUBPORT (priority, target_port_id));
	GM_STBAR ();
	se->type = gm_hton_u8 ((gm_u8_t) (GM_DATAGRAM_SEND_EVENT_0 + size));
      }
    GM_FLUSH_SEND_EVENT (p, se);
  }

  GM_RETURN_NOTHING ();
}
