/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_deregister.c
   This file contains the GM API function gm_deregister_memory().
*/

#include "gm_call_trace.h"
#include "gm_compiler.h"
#include "gm_debug.h"
#include "gm_debug_mem_register.h"
#include "gm_internal.h"
#include "gm_io.h"
#include "gm_lanai_command.h"

static gm_status_t
__gm_deregister_memory (gm_port_t *p, void *_pvma, gm_size_t length)
{
  gm_status_t status;
  gm_up_t pvma;

  GM_CALLED_WITH_ARGS (("%p,%p,"GM_U64_TMPL, p, _pvma, GM_U64_ARG (length)));
  
  pvma = (gm_up_t) (gm_size_t) _pvma;
  
#if !GM_KERNEL && !GM_CAN_REGISTER_MEMORY
  {
    GM_PARAMETER_MAY_BE_UNUSED (p);
    GM_PARAMETER_MAY_BE_UNUSED (pvma);
    GM_PARAMETER_MAY_BE_UNUSED (length);
  
    status = GM_NOT_SUPPORTED_ON_ARCH;
  }
#elif GM_KERNEL
  {
    gm_instance_state_t *is;
    gm_page_hash_table_t *hash;
    gm_port_state_t *ps;
    gm_size_t offset;
    gm_size_t remaining_len;
    unsigned int port_id;
    
    gm_assert (p);
    ps = p->kernel_port_state;
    gm_assert (ps);
    is = ps->instance;
    gm_assert (is);
    hash = &is->page_hash;
    
    /* Round start of registration to a page boundary. */
    
    length += GM_PAGE_OFFSET (pvma);
    pvma -= GM_PAGE_OFFSET (pvma);

    /* Round end of registration to a page boundary. */

    length = GM_PAGE_ROUNDUP (size, length);
    
    /* Check input parameters. */

    if (pvma+length < pvma
	|| length == 0
	|| pvma == 0)
      {
	status = GM_INVALID_PARAMETER;
	goto abort_with_nothing;
      }

    port_id = ps->id;
    if (port_id == GM_DAEMON_PORT_ID)	/* port for information only */
      {
	status = GM_PERMISSION_DENIED;
	goto abort_with_nothing;
      }
    
    /* get exclusive access to the page table. */
        
    gm_arch_mutex_enter (&hash->sync);

    /* Verify that each of the pages is in the page table.  If
       not, the user tried to deregister an illegitimate
       region, and we fail. */

    GM_PRINT (GM_DEBUG_MEM_REGISTER,
	      ("deregistering 0x%x bytes at port virtual addr 0x%lx\n",
	       length, (unsigned long) pvma));

    for (offset=0; offset<length; offset+=GM_PAGE_LEN)
      {
	if (!gm_refs_to_mapping_in_page_table (ps, pvma + offset, port_id))
	  {
	    GM_NOTE (("User tried to deregister unregistered page.\n"));
	    status = GM_INVALID_PARAMETER;
	    goto abort_with_mutex;
	  }
      }
    /* remove the pages from the page table.

    However, only do this a few pages at a time because
    gm_lanai_command_deregister() has a timeout that we don't
    want to trigger.  Don't do it a single page at a time
    because gm_lanai_command_deregister() is much more
    efficient on large blocks.*/

    remaining_len = length;
    pvma = pvma + remaining_len;
    while (remaining_len)
      {
	gm_up_t len;

	len = (remaining_len <= GM_MAX_LANAI_DEREG_LEN
	       ? remaining_len : GM_MAX_LANAI_DEREG_LEN);

	pvma -= len;

	/* Dereference each page of the buffer, and unlock the page if
	   it is no longer referenced. */

	GM_PRINT (GM_DEBUG_MEM_REGISTER,
		  ("Deregistering user buffer pages.\n"));
	offset = len;
	gm_pause_lanai (is);
	while (offset != 0)
	  {
	    offset -= GM_PAGE_LEN;
	    gm_dereference_mapping (ps, pvma + offset);
	  }
	gm_unpause_lanai (is);
	
	/* Now have the NIC flush its cache.  This must happen
	   only after the host entries have been removed to
	   prevent re-caching the values by accident. */
        
	gm_lanai_command_deregister (is, port_id, pvma, len);
	
	/* now unpin the memory.  This must be done only once the
	   NIC is guaranteed not to DMA into them. */
        
	gm_dereference_mapping_commit (is, ps);

	/* advance */
            
	remaining_len -= len;
      }

    status = GM_SUCCESS;
    
  abort_with_mutex:
    /* release exclusive access to the page table */
    gm_arch_mutex_exit (&hash->sync);
  abort_with_nothing:
    /* Fall through */
    ;
  }
#else  /* !GM_KERNEL */
  {
    if (sizeof (length) > 4)
      gm_always_assert(length <= 0xffffffffUL);
    _gm_register_memory_mutex_enter();

    /* Tell the driver how much to deregister */

    if (GM_OS_OSF1 || GM_OS_LINUX || GM_OS_FREEBSD || GM_OS_IRIX || GM_OS_HPUX ||
	GM_OS_WIN2K || GM_OS_MACOSX || GM_OS_AIX)
      {
	struct gm_off_len_uvma_pvma s;
	  
	s.offset = 0;
	s.len = length;
	s.uvma = 0;
	s.pvma = pvma;
	  
	status = _gm_user_ioctl (p, GM_DEREGISTER_MEMORY_BY_STRUCT, &s,
				 sizeof(s));
	if (status != GM_SUCCESS)
	  {
	    gm_pid_t this_node, my_pid;
	    my_pid = gm_getpid ();
	    gm_get_node_id (p, &this_node);
	    GM_NOTE (("%ld Could not deregister memory (%p, %ld)"
		      " for gmID %d:\nerror %d: %s\n",
		      (unsigned long) my_pid, _pvma, length, this_node,
		      status, gm_strerror (status)));
	    goto abort_with_mutex;
	  }
      }
    else if (GM_OS_SOLARIS)
      {
	gm_u32_t klen = (gm_u32_t) length;
	
	status = _gm_user_ioctl (p, GM_SET_REGISTER_MEMORY_LENGTH_AND_PVMA,
				 &klen, sizeof (klen));
	if (status != GM_SUCCESS)
	  {
	    GM_NOTE (("Could not set registration length.\n"));
	    goto abort_with_mutex;
	  }

	/* Tell the driver what to deregister. */
	status = _gm_user_ioctl (p, GM_DEREGISTER_MEMORY, (void *) (gm_size_t) pvma, length);
	if (status != GM_SUCCESS)
	  {
	    gm_pid_t my_pid;
	    unsigned int this_node;
	    
	    my_pid = gm_getpid ();
	    gm_get_node_id (p, &this_node);
	    GM_NOTE (("%d Could not deregister memory (%p, %ld)"
		      " for gmID %u\n",
		      my_pid, _pvma, length, this_node));
	    goto abort_with_mutex;
	  }
      }
    else
      {
	status = GM_NOT_SUPPORTED_ON_ARCH;
      }

  abort_with_mutex:
    _gm_register_memory_mutex_exit();
    /* fall through */
  }
#endif /* !GM_KERNEL */

  GM_RETURN_STATUS (status);
}

/**
   gm_deregister_memory() deregisters \b len bytes of user virtual
   that were previously registered for DMA transfers with a matching
   call to gm_register_memory() using \b pvma.
 
   @retval GM_SUCCESS Operation completed successfully.
   @retval GM_FAILURE
 
   @param p \b (IN) The GM port in use.
   @param pvma \b (IN) The pvma used to register the memory to now be
                 deregistered.
   @param length \b (IN) The number of bytes in the array to be
                 deregistered.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT gm_status_t
gm_deregister_memory (gm_port_t *p, void *ptr, gm_size_t length)
{
  gm_status_t status;
  
  GM_CALLED_WITH_ARGS_EX
    (1, ("%p,%p,"GM_U64_TMPL, p, ptr, GM_U64_ARG(length)));
  status = __gm_deregister_memory (p, ptr, length);
  GM_RETURN_STATUS (status);
}

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
