/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_dma_malloc.c
   This file includes source for the user-level API calls
   gm_dma_malloc() and gm_dma_free().
*/

#include "gm_call_trace.h"
#include "gm_compiler.h"
#include "gm_debug.h"
#include "gm_debug_mem_register.h"
#include "gm_internal.h"
#include "gm_malloc_debug.h"
#include "gm_register_recv_queue.h"

#if !GM_KERNEL
				/* mmap() */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif

#else  /* GM_KERNEL */

#include "gm_impl.h"

#endif /* GM_KERNEL */

/****************************************************************
 * Debug switches
 ****************************************************************/

#define GM_DEBUG_DMA_MALLOC 0

#if GM_DEBUG_DMA_MALLOC
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/****************************************************************
 * Support functions
 ****************************************************************/

#if GM_KERNEL

/****************
 * Kernel support for DMA region registration
 *
 * Normally, the gm_dma_region_alloc() page callback is used to
 * register DMAable pages.  However, this interface is insufficient
 * for adding PTEs to the GM user page hash table because of the need
 * to undo page registrations of the DMA region if adding a PTE fails.
 *
 * To allow these registrations to be undone cleanly, we temporarily
 * allocate an array for the DMA addrs of each page, have the
 * register_page callback store the dp's here, and after we know the
 * kernel addr of the DMAable region, we register the pages and
 * cleanup in case of any errors.
 ****************/

struct gm_page_registration_context
{
  gm_dp_t *dma_addr;
  unsigned long num_pages;
};

/* Allocate the resources necessary to guarantee that register_page() will
   succeed and return a pointer to these resources at *context. */

static
gm_status_t
prepare_to_register (gm_size_t len,
		     struct gm_page_registration_context **context)
{
#if !GM_ENABLE_VM
  *context = NULL;
  return GM_SUCCESS;
#else /* GM_ENABLE_VM */
  gm_status_t status;
  unsigned long pages;

  GM_CALLED_WITH_ARGS ((GM_U64_TMPL",%p", GM_U64_ARG (len), context));
  
  *context = ((struct gm_page_registration_context *)
	      gm_malloc (sizeof (struct gm_page_registration_context)));
  if (!*context)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_nothing;
    }
  
  pages = len / GM_PAGE_LEN + 1;
  (*context)->dma_addr = (gm_dp_t *) gm_calloc (pages, sizeof (gm_dp_t));
  if (!(*context)->dma_addr)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_context;
    }
  GM_PRINT (GM_DEBUG_DMA_MALLOC, ("dma addrs start at %p\n",
				  (*context)->dma_addr));
  (*context)->num_pages = pages;
  GM_RETURN_STATUS (GM_SUCCESS);

 abort_with_context:
  gm_free (*context);
 abort_with_nothing:
  GM_RETURN_STATUS (status);
#endif /* GM_ENABLE_VM */
}

/* If we find that we cannot complete the registration after calling
   prepare_to_register(), then we must cancel the registration to free
   the resources. */

static
void
cancel_registration (struct gm_page_registration_context *context)
{
#if !GM_ENABLE_VM
  return;
#else /* GM_ENABLE_VM */
  GM_CALLED_WITH_ARGS (("%p", context));
  gm_free (context->dma_addr);
  gm_free (context);
  GM_RETURN_NOTHING ();
#endif /* GM_ENABLE_VM */
}

/* Callback used by gm_dma_region_alloc() to register the DMA addr of
   each page of the DMA region.  We do not know the kernel addr of the
   region when this callback is called, so we just stick the DMA addrs
   in an array. */

static
gm_status_t
register_page (void *arg, gm_dp_t dp, gm_u32_t page_num)
{
#if !GM_ENABLE_VM
  return GM_SUCCESS;
#else /* GM_ENABLE_VM */
  struct gm_page_registration_context *context;

  GM_CALLED_WITH_ARGS (("%p,"GM_U64_TMPL",%u", arg, GM_U64_ARG(dp),
			(unsigned int) page_num));
  context = (struct gm_page_registration_context *) arg;
  gm_assert (page_num < context->num_pages);
  context->dma_addr[page_num] = dp;
  GM_RETURN_STATUS (GM_SUCCESS);
#endif /* GM_ENABLE_VM */
}

/* Now that gm_dma_malloc() has used register_page() to register each
   DMA page, and now that we know the kaddr of the DMA region, add the
   mappings to the page table.  Also free the registration context,
   even if this function fails. */
   
static
gm_status_t
commit_registration (gm_port_t *port, void *kaddr, gm_size_t len,
		     struct gm_page_registration_context *context)
{
#if !GM_ENABLE_VM
  return GM_SUCCESS;
#else /* GM_ENABLE_VM */
  gm_dp_t *dma_addr;
  unsigned long i;
  gm_status_t status = GM_SUCCESS;
  gm_instance_state_t *is;
  gm_port_state_t *ps;
  unsigned int port_num;

  GM_CALLED_WITH_ARGS (("%p,%p,"GM_U64_TMPL",%p", port, kaddr, GM_U64_ARG(len),
			context));
  
  if (!GM_PAGE_ALIGNED (len))
    {
      status = GM_INVALID_PARAMETER;
      goto abort_with_dma_addr;
    }

  GM_PRINT (GM_DEBUG_DMA_MALLOC, ("1\n"));
  dma_addr = context->dma_addr;
  GM_PRINT (GM_DEBUG_DMA_MALLOC, ("2\n"));
  gm_assert (port);
  GM_PRINT (GM_DEBUG_DMA_MALLOC, ("3\n"));
  ps = port->kernel_port_state;
  gm_assert (ps);
  GM_PRINT (GM_DEBUG_DMA_MALLOC, ("4\n"));
  is = ps->instance;
  gm_assert (ps);
  GM_PRINT (GM_DEBUG_DMA_MALLOC, ("5\n"));
  port_num = ps->id;
  GM_PRINT (GM_DEBUG_DMA_MALLOC, ("6\n"));

  /* Now that we know kaddr, we register the pages for real. */

  i = 0;
  while (len)
    {
      GM_PRINT (GM_DEBUG_DMA_MALLOC, ("7\n"));
#if !GM_CPU_powerpc64 && !GM_CPU_x86_64
      gm_assert (sizeof (gm_up_t) >= sizeof (gm_size_t));
#endif
      status = gm_add_mapping_to_page_table (ps,
					     (gm_up_t) ((gm_size_t) kaddr),
					     (gm_up_t) ((gm_size_t) kaddr),
					     dma_addr[i]);
      GM_PRINT (GM_DEBUG_DMA_MALLOC, ("8\n"));
      if (status != GM_SUCCESS)
	{
	  gm_pause_lanai (is);
	  while (i)
	    {
	      GM_PRINT (GM_DEBUG_DMA_MALLOC, ("9\n"));
	      gm_dereference_mapping (ps, (gm_up_t) ((gm_size_t) kaddr));
	      GM_PRINT (GM_DEBUG_DMA_MALLOC, ("10\n"));
	      /* Funky syntax to placate the IRIX MIPSpro compiler */
              kaddr = (void *) ((char *) kaddr - GM_PAGE_LEN);
	      i--;
	    }
	  gm_unpause_lanai (is);
	  goto abort_with_dma_addr;
	}
      /* Funky syntax to placate the IRIX MIPSpro compiler */
      kaddr = (void *) ((char *) kaddr + GM_PAGE_LEN);
      len -= GM_PAGE_LEN;
      i++;
    }

  GM_PRINT (GM_DEBUG_DMA_MALLOC, ("11\n"));
  cancel_registration (context);
  GM_RETURN_STATUS (status);

 abort_with_dma_addr:
  cancel_registration (context);
  GM_RETURN_STATUS (status);
#endif /* GM_ENABLE_VM */
}
#endif /* GM_KERNEL */

/****************************************************************
 * Environment abstraction
 *
 * DMAable memory allocation is performed very differently on
 * different environments.  We hide these details in alloc_dma_pages()
 * and free_dma_pages().
 ****************************************************************/

static
gm_status_t
alloc_dma_pages (struct gm_port *p, gm_size_t length, void **ret)
{
  GM_CALLED_WITH_ARGS (("%p,"GM_U64_TMPL",%p",
			p, GM_U64_ARG (length), ret));
  gm_assert (GM_PAGE_ALIGNED (length));
  
#if !GM_KERNEL && GM_CAN_REGISTER_MEMORY
  
  /* Simply allocate the pages and register them */
  
  {
    gm_status_t status;
    void *addr;
  
    /* allocate some pages */
    
    addr = gm_alloc_pages (length);
    if (addr == 0)
      {
	status = GM_OUT_OF_MEMORY;
	goto abort_with_nothing;
      }
    GM_PRINT
      (GM_DEBUG_MEM_REGISTER,
       ("gm_dma_malloc() will register "GM_U64_TMPL" new page[s] at 0x%p\n",
	GM_U64_ARG (GM_PAGE_ROUNDUP(u32, length) / GM_PAGE_LEN), addr));
    
    /* bzero the page[s] to fault it [them] into memory */
    
    gm_bzero (addr, length);

    /* and register the pages */
    
    gm_assert (GM_PAGE_ALIGNED (addr));
    status = gm_register_memory (p, addr, length);
    if (status != GM_SUCCESS)
      {
	goto abort_with_pages;
      }

    *ret = addr;
    GM_RETURN_STATUS (GM_SUCCESS);
    
  abort_with_pages:
    gm_free_pages (addr, length);
  abort_with_nothing:
    GM_RETURN_STATUS (status);
  }
#elif !GM_KERNEL && !GM_CAN_REGISTER_MEMORY

  /* mmap DMAable memory from the kernel */
  
  {
    gm_status_t status;
    void *addr;
    
    status = _gm_mmap (p, (p->mappings->copy_block.offset
			   + (gm_offset_t) p->dma.alloced_mem_len),
		       length, GM_MAP_READ | GM_MAP_WRITE,
		       &addr);
    if (status != GM_SUCCESS)
      {
	goto abort_with_nothing;
      }
    if (!GM_PAGE_ALIGNED (addr))
      {
	GM_NOTE (("Zone is not page aligned.  "
		  "Performance will be impacted.\n"));
      }
    *ret = addr;
    GM_RETURN_STATUS (GM_SUCCESS);

  abort_with_nothing:
    GM_RETURN_STATUS (status);
  }
#else  /* GM_KERNEL */
  
  /* In the kernel, allocating dmaable memory is slightly convoluted:
     We must use a DMA region to allocate the DMAable memory, so we
     must keep track of the region until the corresponding free, which
     we do with a hash table. */

  {
    gm_arch_dma_region_t *region;
    gm_instance_state_t *is;
    gm_port_state_t *ps;
    gm_status_t status;
    void *kaddr;

    gm_assert (p);
    ps = p->kernel_port_state;
    gm_assert (ps);
    is = ps->instance;
    gm_assert (is);

    /* auto-initialize */
    
    if (p->kernel_ptr_to_region_hash_population == 0)
      {
	p->kernel_ptr_to_region_hash
	  = gm_create_hash (gm_hash_compare_ptrs, gm_hash_hash_ptr,
			    0, 0, 0, 0);
	if (!p->kernel_ptr_to_region_hash)
	  {
	    status = GM_OUT_OF_MEMORY;
	    goto abort_with_nothing;
	  }
      }

    /* Allocate space for the DMA region */

    region = gm_calloc (1, sizeof (*region));
    if (!region)
      {
	status = GM_OUT_OF_MEMORY;
	goto abort_with_kernel_ptr_to_region_hash;
      }

    /* Allocate the DMAable memory and add the entries to the page table */
    
    {
      struct gm_page_registration_context *registration_context;

      GM_PRINT (GM_DEBUG_DMA_MALLOC, ("p = %p\n", p));
      status = prepare_to_register (length, &registration_context);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_region;
	}
      GM_PRINT (GM_DEBUG_DMA_MALLOC, ("p = %p", p));
      status = gm_arch_dma_region_alloc (is, region, length, 
					 GM_ARCH_DMA_RDWR | GM_ARCH_DMA_STREAMING,
					 register_page, registration_context);
      if (status != GM_SUCCESS)
	{
	  cancel_registration (registration_context);
	  goto abort_with_region;
	}
      GM_PRINT (GM_DEBUG_DMA_MALLOC, ("p = %p", p));
      kaddr = gm_arch_dma_region_kernel_addr (region);
      GM_PRINT (GM_DEBUG_DMA_MALLOC, ("p = %p", p));
      status = commit_registration (p, kaddr, length, registration_context);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_region;
	}
      GM_PRINT (GM_DEBUG_DMA_MALLOC, ("p = %p", p));
    }
    
    /* put the region in the hash table so we can find it when we are asked
       to free the pages. */
    
    status = gm_hash_insert (p->kernel_ptr_to_region_hash, kaddr, region);
    if (status != GM_SUCCESS)
      {
	goto abort_with_dmaable_memory;
      }

    /* commit & return */
    
    p->kernel_ptr_to_region_hash_population++;
    *ret = kaddr;
    GM_RETURN_STATUS (GM_SUCCESS);

  abort_with_dmaable_memory:
    gm_arch_dma_region_free (region);
  abort_with_region:
    gm_free (region);
  abort_with_kernel_ptr_to_region_hash:
    if (p->kernel_ptr_to_region_hash_population == 0)
      {
	gm_destroy_hash (p->kernel_ptr_to_region_hash);
	p->kernel_ptr_to_region_hash = 0;
      }
  abort_with_nothing:
    GM_RETURN_STATUS (status);
  }
#endif /* GM_KERNEL */
}

static
void
free_dma_pages (struct gm_port *p, void *ptr, gm_size_t length)
{
  GM_CALLED_WITH_ARGS (("%p,%p,"GM_U64_TMPL, p, ptr, GM_U64_ARG (length)));
  
  gm_assert (GM_PAGE_ALIGNED (length));
  
#if !GM_KERNEL && GM_CAN_REGISTER_MEMORY
  {
    gm_deregister_memory (p, ptr, length);
    gm_free_pages (ptr, length);
  }
#elif !GM_KERNEL && !GM_CAN_REGISTER_MEMORY
  {
    _gm_munmap (p, ptr, length);
  }
#else  /* GM_KERNEL */
  {
    gm_arch_dma_region_t *region;
    gm_instance_state_t *is;
    gm_port_state_t *ps;
    void *kaddr;
    
    gm_assert (p);
    ps = p->kernel_port_state;
    gm_assert (ps);
    is = ps->instance;
    gm_assert (is);

    /* retrieve the DMA region corresponding the allocation */
    
    region = gm_hash_remove (p->kernel_ptr_to_region_hash, ptr);
    gm_assert (region);

#if GM_ENABLE_VM
    /* deregister the pages */
    
    kaddr = gm_arch_dma_region_kernel_addr (region);
    gm_pause_lanai (is);
    while (length)
      {
	length -= GM_PAGE_LEN;
	gm_dereference_mapping (ps, (gm_up_t) ((char *) kaddr + length));
      }
    gm_unpause_lanai (is);
#endif /* GM_ENABLE_VM */
    
    /* Free the DMA region */
    
    gm_arch_dma_region_free (region);
    gm_free (region);

    /* If there are no more allocated regions, destroy the hash table
       that records them. */
    
    if (--p->kernel_ptr_to_region_hash_population == 0)
      {
	gm_destroy_hash (p->kernel_ptr_to_region_hash);
	p->kernel_ptr_to_region_hash = 0;
      }
  }
#endif /* GM_KERNEL */

  GM_RETURN_NOTHING ();
}

static
void *
__gm_dma_malloc (struct gm_port *p, gm_size_t length)
{
  void *ret;
  int i;

  GM_CALLED_WITH_ARGS (("%p,"GM_U64_TMPL, p, GM_U64_ARG (length)));
  
  if (!length)
    length = 1;

  /* Try to allocate a buffer from the
     existing memory allocation
     zones. */
  for (i = p->dma.num_zones - 1; i >= 0; i--)
    {
      ret = gm_zone_malloc (p->dma.zone[i], length);
      if (ret)
	GM_RETURN_PTR ((ret));
    }
  gm_assert (i == -1);
  
  /* Could not allocate the buffer from the existing zones */

  /* First, check if the requested length is too large to possibly be
     allocated, even if we fully expand the copy block. */

#if !GM_CAN_REGISTER_MEMORY
  if (length > p->mappings->copy_block.len / 2)
    GM_RETURN_PTR (0);
#endif

  /* Try to create bigger zones until the alloc succeeds or we cannot
     create a bigger zone.  */
  
  do
    {
      void *addr;
      gm_size_t alloc_len;

      alloc_len = p->dma.alloced_mem_len;
      if (!alloc_len)
	alloc_len = GM_PAGE_LEN;
      gm_assert (GM_POWER_OF_TWO (alloc_len));

      /* allocate memory for the zone */

      if (alloc_dma_pages (p, alloc_len, &addr) != GM_SUCCESS)
	{
	  GM_RETURN_PTR (0);
	}

      /* Create a new zone */

      p->dma.zone[p->dma.num_zones] = gm_zone_create_zone (addr, alloc_len);
      if (!p->dma.zone[p->dma.num_zones])
	{
	  free_dma_pages (p, addr, alloc_len);
	  GM_RETURN_PTR (0);
	}
      
      p->dma.zone_base[p->dma.num_zones] = addr;
      p->dma.zone_len[p->dma.num_zones] = alloc_len;
      p->dma.num_zones++;
      p->dma.alloced_mem_len += alloc_len;
      ret = gm_zone_malloc (p->dma.zone[p->dma.num_zones - 1], length);
    }
  while (!ret);

  GM_RETURN_PTR (ret);
}

static
void
__gm_dma_free (gm_port_t *p, void *addr)
{
  int i;

  GM_CALLED_WITH_ARGS (("%p,%p", p, addr));
  
  /* Find the zone this pointer belongs to and free it */
  
  for (i = p->dma.num_zones-1; i >= 0; i--)
    {
      if (gm_zone_addr_in_zone (p->dma.zone[i], addr))
	{
	  gm_zone_free (p->dma.zone[i], addr);
	  GM_RETURN_NOTHING ();
	}
    }
  GM_NOTE (("gm_dma_free: pointer does not belong to this port.\n"));

  GM_RETURN_NOTHING ();
}

/****************************************************************
 * Instrumented entry points
 ****************************************************************/

/**
   gm_dma_free() frees \b p, which was allocated by a call to
   gm_dma_calloc() or gm_dma_malloc().  Note that the memory
   is not necessarily unlocked and returned to the operating
   system, but may be reused in future calls to gm_dma_calloc()
   or gm_dma_malloc().
 
   @param p \b (IN) The GM port.
 
   @param addr \b (IN) The address of the memory to be freed.
 
   @see gm_dma_calloc gm_dma_malloc
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
void
gm_dma_free (gm_port_t *p, void *addr)
{
  GM_CALLED ();
  GM_MDEBUG_REMOVE_PTR (addr, gm_dma_free);
  __gm_dma_free (p, addr);
  GM_RETURN_NOTHING ();
}

/**
   gm_dma_malloc() allocates \b length bytes of DMAable memory
   aligned on a 4-byte boundary.  Memory should be freed using
   gm_dma_free().
 
   @param p \b (IN) Handle to the GM port.
 
   @param length \b (IN) The number of bytes to be malloc'ed.
 
   @see gm_dma_calloc gm_dma_free
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
void *
gm_dma_malloc (struct gm_port *p, gm_size_t length)
{
  gm_status_t status;
  void *ret;

  GM_CALLED ();
  ret = __gm_dma_malloc (p, length);
  if (!ret)
    {
      goto abort_with_nothing;
    }

  status = GM_MDEBUG_RECORD_PTR (ret, gm_dma_free);
  if (status != GM_SUCCESS)
    {
      goto abort_with_dma_malloc;
    }

  GM_RETURN_PTR (ret);

 abort_with_dma_malloc:
  __gm_dma_free (p, ret);
 abort_with_nothing:
  GM_RETURN_PTR (0);
}

/* Internal function used by gm_close() to free all of the resources
   allocated by this module. */

void
_gm_dma_finalize (gm_port_t *p)
{
  int i;
  unsigned long len;
  void *addr;

  GM_CALLED_WITH_ARGS (("%p", p));

#if GM_KERNEL
  /* We must call this explicitly in the kernel before cleaning up
     DMAable memory since there is no other hook for catching these frees
     if the user does not call gm_dma_free() on a block. */
  
  gm_deregister_recv_queue (p->kernel_port_state);
#endif
  
  for (i=((int) p->dma.num_zones)-1; i>=0; i--)
    {
      addr = p->dma.zone_base[i];
      gm_assert (addr);
      len = (unsigned long) (p->dma.zone_len[i]);
      gm_assert (len);
      gm_assert (GM_POWER_OF_TWO (len));
      GM_PRINT (GM_PRINT_LEVEL >= 1, ("Freeing zone %p.\n", p->dma.zone[i]));
      gm_zone_destroy_zone (p->dma.zone[i]);
      p->dma.zone[i] = 0;
      free_dma_pages (p, addr, len);
    }

  GM_RETURN_NOTHING ();
}


/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/

