/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_debug_ethernet.h"
#include "gm_enable_ethernet.h"
#include "gm_debug_gateway_cache.h"
#include "gm_ether.h"
#include "gm_ether_inlines.h"
#include "gm_send_queue.h"
#include "gm_trace.h"
#if GM_KERNEL && (GM_OS_FREEBSD || GM_OS_OSF1 || GM_OS_MACOSX)
#include "gm_bsd_ether.h"
#endif
#include "gm_internal.h"
#if GM_KERNEL
#include "gm_pio.h"
#endif

#define GM_DEBUG_ETHERNET_PROVIDE_GATHER_LIST 0

#if GM_DEBUG_ETHERNET
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/* Copy a 48-bit MAC address.  We do this a byte at a time because
   we don't know anything about the alignment of the source addr. */

static
void
gm_copy_ethernet_mac (const gm_u8_t *from, gm_ethernet_mac_addr_t _to)
{
  gm_u8_t a, b, c;
  gm_u8_n_t *to;
  int i;

  to = _to;			/* change type so to++ will work. */
  for (i=2; i; i--)
    {
      a = *from++;
      b = *from++;
      c = *from++;
      *to++ = gm_hton_u8 (a);
      *to++ = gm_hton_u8 (b);
      *to++ = gm_hton_u8 (c);
    }
}

static
void
dummy_handler(void *unused)
{
  /* nothing */
}

void
gm_disable_ethernet (gm_port_t *p)
{
  gm_size_t i;
#if GM_KERNEL
  gm_instance_state_t *is = p->kernel_port_state->instance;
  gm_port_state_t *ps = p->kernel_port_state;

  gm_pause_lanai (is);

  /* Drop all incoming frames */
  gm_write_lanai_global_u32 (is, ethernet.enabled, 0);

  /* Install a dummy handler */
  ps->wake.handler = dummy_handler;
  ps->wake.context = (void *) ps;

  gm_unpause_lanai (is);
#endif

  /* there is no ethernet state, then fail gracefully */
  if (p->ethernet == 0)
    {
      return;
    }

  /* free each list.  Check for null, as we may be called as part of
     the error handling code in gm_enable_ethernet () 	*/
  
  for (i = 0; i < GM_NUM_ELEM (p->ethernet->gather.list); i++)
    {
      if (p->ethernet->gather.list[i].ptr != 0)
	{
	  gm_dma_free (p, (void *) p->ethernet->gather.list[i].ptr);
	}
    }
  for (i = 0; i < GM_NUM_ELEM (p->ethernet->scatter.list); i++)
    {
      if (p->ethernet->scatter.list[i].ptr != 0)
	{
	  gm_dma_free (p, (void *) p->ethernet->scatter.list[i].ptr);
	}
    }

  gm_free ((void *) p->ethernet);
}

gm_status_t
gm_enable_ethernet (gm_port_t *p, int recv_ring_entries, gm_u16_t tx_intr_freq)
{
  gm_ethernet_gather_list_n_t *ptr;
  gm_dp_t dma_addr;
  gm_status_t status;
  gm_size_t i;
#if GM_KERNEL
  gm_instance_state_t *is;
  gm_port_state_t *ps;
  gm_u8_t credit_incr;
  gm_u16_t sdmas_until_wake;

  ps = p->kernel_port_state;
  gm_assert (ps);
  is = p->kernel_port_state->instance;
  sdmas_until_wake = 
    gm_read_lanai_global_u16 (is, ethernet.wake.sdmas_until_wake);
  
  if (sdmas_until_wake == 0)
    {
      /*
       * The user has not tuned the ethernet interrupt coalessing
       * parameters, so we do so now based on info from the MD driver.
       * Credit_incr is how many times the firmware will be allowed
       * to interrupt in a 100us period of time.  A value of 1 yields
       * a max of 10,000 interrupts/sec.  4 = 40,0000, etc.
       *
       * The values below are tuned so that the firmware can receive
       * up to 128,000 packets/sec without starvation.
       *
       */
      credit_incr = 128 / recv_ring_entries;
      if (credit_incr == 0)
	{
	  credit_incr = 1;
	}
      
      gm_write_lanai_global_u8 (is, ethernet.wake.max_credits,
				8 * credit_incr - 1);
      
      gm_write_lanai_global_u8 (is, ethernet.wake.credit_incr, 
				credit_incr);
      gm_write_lanai_global_u16 (is, ethernet.wake.max_sdmas_until_wake, 
				 tx_intr_freq);
      gm_write_lanai_global_u16 (is, ethernet.wake.sdmas_until_wake, 
				 tx_intr_freq);
    }
#endif

  /* Allocate ethernet state */
  p->ethernet = gm_calloc (1, sizeof (*p->ethernet));
  if (p->ethernet == 0)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_nothing;
    }
  
  /* Allocate ethernet gather lists and compute corresponding DMA addrs. */

  /* Ensure that each allocation will be physically contiguous. */
  gm_always_assert (sizeof (gm_ethernet_gather_list_n_t) <= GM_PAGE_LEN);
  
  for (i = 0; i < GM_NUM_ELEM (p->ethernet->gather.list); i++)
    {
      ptr 
	= p->ethernet->gather.list[i].ptr 
	= gm_dma_malloc (p, sizeof (*ptr));
      if (ptr == 0)
	{
	  status = GM_OUT_OF_MEMORY;
	  goto abort_with_ethernet;
	}
      
      status = gm_dma_addr (p, (gm_up_t) ptr, &dma_addr);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_ethernet;
	}
      p->ethernet->gather.list[i].dma_addr = dma_addr;
    }

  for (i = 0; i < GM_NUM_ELEM (p->ethernet->scatter.list); i++)
    {
      ptr 
	= p->ethernet->scatter.list[i].ptr 
	= gm_dma_malloc (p, sizeof (*ptr));
      if (ptr == 0)
	{
	  status = GM_OUT_OF_MEMORY;
	  goto abort_with_ethernet;
	}
      
      status = gm_dma_addr (p, (gm_up_t) ptr, &dma_addr);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_ethernet;
	}
      p->ethernet->scatter.list[i].dma_addr = dma_addr;
    }

#if GM_KERNEL
  gm_pause_lanai (is);
  gm_write_lanai_global_u32 (is, ethernet.enabled, 1);
  gm_unpause_lanai (is);
#endif

  return (GM_SUCCESS);

 abort_with_ethernet:
  gm_disable_ethernet (p);
 abort_with_nothing:
  return status;
}

#if GM_KERNEL
void
gm_set_ethernet_flags (gm_instance_state_t *is, unsigned int flags)
{
  gm_write_lanai_global_u32 (is, ethernet.flags, flags);
}

void
gm_set_ethernet_mac_address (gm_instance_state_t *is, gm_u8_t *addr)
{
  gm_u8_t mac[6];

  /* make a local copy of the new MAC address */
  gm_bcopy(addr, mac, sizeof (mac));

  /* poke it into the Lanai, guarding against false positives by making the
   * address be totally bogus, then correct, but never valid while incorrect. */

  /* write really bogus address */
  gm_write_lanai_global_u32 (is, ethernet.mac_address[0], 0xffffffff);

  /* wait for lanai not to be in middle of loading old addr */
  gm_arch_spin(is, 1);

  /* now write slightly bogus address */
  mac[0] = 0xff;
  gm_write_lanai_global_byte_array (is, mac, ethernet.mac_address);

  /* and finally, make it right... */
  gm_write_lanai_global_u8 (is, ethernet.mac_address[0], *(gm_u8_t *)addr);
}
#endif


static 
gm_dp_t
gm_ethernet_provide_gather_list (gm_port_t * p,
				 gm_u32_t segment_cnt,
				 gm_ethernet_segment_descriptor_t segment[])
{
  gm_ethernet_gather_list_n_t *pos;
  unsigned int i;
  unsigned int total_len = 0;
  gm_dp_t dma_addr;
  
  GM_CALLED ();

  pos = p->ethernet->gather.list[p->ethernet->gather.index].ptr;
  dma_addr = p->ethernet->gather.list[p->ethernet->gather.index].dma_addr;
  if (++p->ethernet->gather.index == GM_NUM_SEND_TOKENS)
    {
      p->ethernet->gather.index = 0;
    }

  GM_PRINT (GM_DEBUG_ETHERNET_PROVIDE_GATHER_LIST,
	      ("user provide ethernet gather list:\n"));
  for (i = 0; i < segment_cnt; i++)
    {
      gm_dp_t ptr;
      unsigned int len;
      
      ptr = segment[i].ptr;
      len = segment[i].len;
      _GM_PRINT (GM_DEBUG_ETHERNET_PROVIDE_GATHER_LIST,
		 ("  "GM_U64_TMPL",%u\n", GM_U64_ARG(ptr), len));
      (*pos)[i].ptr = gm_hton_dp (ptr);
      (*pos)[i].len = gm_hton_u32 (len);
      total_len += len;
    }
  gm_assert (total_len <= GM_ETHERNET_MTU);

  GM_RETURN (dma_addr);
}


/* ethernet_send_with_callback() and
   ethernet_broadcast_with_callback() have a lot of similar code the
   common part is factorized here in gm_ether_cmn() */
static void
gm_ether_cmn (gm_port_t * p,
	      gm_u32_t segment_cnt,
	      gm_ethernet_segment_descriptor_t segment[],
	      const gm_u8_t *ethernet_addr,
	      int evt_type, unsigned target_node_id,
	      gm_send_completion_callback_t callback, void *context)
{
  volatile struct gm_ethernet_send_event *ese;
  struct gm_send_queue_slot *sqs;
  unsigned int i;
  gm_offset_t io_len;
  unsigned char *ptr;
  struct gm_ethernet_send_event batch_write;
  gm_dp_t dp;

  GM_CALLED_WITH_ARGS (("%p,%u,...,%d,%02x:%02x:%02x:%02x:%02x:%02x,%u,%p,%p",
			p, (unsigned int) segment_cnt,
			ethernet_addr[0], ethernet_addr[1], ethernet_addr[2],
			ethernet_addr[3], ethernet_addr[4], ethernet_addr[5],
			evt_type, target_node_id, callback, context));

  GM_PRINT
    (GM_DEBUG_GATEWAY_CACHE
     && (ethernet_addr[0] != 0x00
	 || ethernet_addr[1] != 0x60
	 || ethernet_addr[2] != 0xdd),
     ("Odd Ethernet address %02x:%02x:%02x:%02x:%02x:%02x\n",
      ethernet_addr[0], ethernet_addr[1], ethernet_addr[2],
      ethernet_addr[3], ethernet_addr[4], ethernet_addr[5]));
  
  /* print the segment list */
  
  if (GM_DEBUG_ETHERNET)
    {
      for (i=0; i<segment_cnt; i++)
	{
	  GM_PRINT (GM_DEBUG_ETHERNET,
		    ("{"GM_U64_TMPL",0x%u}\n",
		     GM_U64_ARG (segment[i].ptr),
		     (unsigned int) segment[i].len));
	}
    }
    
  /* BAD: The following two blocks could be combined and pipelined. */

  /* ethernet does not use completion callbacks; it uses interrupts,
     hence the null callback. */

  ese = GM_SEND_QUEUE_SLOT (p, callback, context, ethernet);
  sqs = ((struct gm_send_queue_slot *) (ese + 1) - 1);

#if !GM_STRUCT_WRITE_COMBINING
#error  GM_STRUCT_WRITE_COMBINING must be defined
#endif
  gm_assert (segment_cnt);

  ptr = (batch_write.packed_gather_list
	 + sizeof (batch_write.packed_gather_list));
  
  if (segment_cnt > GM_MAX_ETHERNET_IMMEDIATE_GATHER_CNT)
    {
      gm_assert (segment_cnt <= GM_MAX_ETHERNET_GATHER_CNT);
      dp =    
	gm_ethernet_provide_gather_list (p, segment_cnt, segment);
      ptr -= sizeof (gm_dp_t);
      *(gm_dp_n_t *)ptr = gm_hton_dp (dp);
    }
  else
    {
      /* Pack the gather list into the staged send event. */

      for (i = segment_cnt; i; i--)
	{
#if !GM_KERNEL
	  /* 
	   *  Some drivers use physically contiguous buffers which
	   *  cross page boundaries, so we trust the kernel - gallatin
	   */
	  gm_assert (GM_PAGE_REMAINING (segment[segment_cnt - i].ptr)
		     >= segment[segment_cnt - i].len);
#endif
	  ptr -= sizeof (gm_dp_t);
	  /* Make sure the target location is adequately aligned. */
	  gm_assert (GM_ALIGNED (ptr, sizeof (void *)));
	  *(gm_dp_n_t *)ptr = gm_hton_dp (segment[segment_cnt - i].ptr);
	  gm_assert (*(gm_dp_t *)ptr);
	}
      for (i = segment_cnt; i; i--)
	{
	  ptr -= sizeof (gm_u16_t);
	  gm_assert (GM_ALIGNED (ptr, sizeof (gm_u16_t)));
	  *(gm_u16_n_t *)ptr = gm_hton_u16 ((gm_u16_t)segment[segment_cnt - i].len);
	  gm_assert (*(gm_u16_t *)ptr);
	}
    }
  io_len = (unsigned char *) (&batch_write + 1) - ptr;
  
  /* complete the staged send event */

  if (evt_type == GM_GATEWAY_SEND_EVENT)
    {
      /* The caller did not resolve the gateway node, so have the
	 firmware do it. */
      
      gm_copy_ethernet_mac
	(ethernet_addr, batch_write.via.gateway.target_mac_addr);
    }
  else
    {
      /* The caller has provided the gateway node. */
      
      gm_assert (evt_type == GM_ETHERNET_SEND_EVENT);
      if (GM_DEBUG_GATEWAY_CACHE__ALWAYS_USE
	  && ((ethernet_addr[0] & ethernet_addr[1] & ethernet_addr[2]
	       & ethernet_addr[3] & ethernet_addr[4] & ethernet_addr[4])
	      != (gm_u8_t) 0xff))
	{
	  /* For debugging purposes, tell the firmware to use the
	     gateway cache even though the caller has resolved the
	     destination gateway. */
	  evt_type = GM_GATEWAY_SEND_EVENT;
	  gm_copy_ethernet_mac
	    (ethernet_addr, batch_write.via.gateway.target_mac_addr);
	}
      else
	{
	  /* Tell the firmware the gateway node provided by the caller. */
	  batch_write.via.node.target_node_id
	    = gm_hton_u16 ((gm_u16_t) target_node_id);
	}
    }
  batch_write.gather_cnt = gm_hton_u8 ((gm_u8_t) segment_cnt);
  batch_write.type = gm_hton_u8 ((gm_u8_t) evt_type);
  
  /* pad the start to correct any iobus transfer misalignment */
  
  io_len += GM_MISALIGNMENT (ptr, sizeof (gm_dp_t));
  ptr -= GM_MISALIGNMENT (ptr, sizeof (gm_dp_t));
  
  /* copy the send event to the LANai */
  
  gm_copy_to_io_space ((gm_u8_t *) (ese + 1) - io_len, ptr,
		       (unsigned int)io_len);

  
  GM_FLUSH_SEND_EVENT (p, ese);
  GM_RETURN_NOTHING ();
}

GM_ENTRY_POINT
void
gm_ethernet_send_with_callback (gm_port_t * p,
				gm_u32_t segment_cnt,
				gm_ethernet_segment_descriptor_t segment[],
				gm_u8_t * ethernet_addr,
				gm_send_completion_callback_t callback,
				void *context)
{
  int type;
  unsigned target_node_id = 0;

  GM_CALLED_WITH_ARGS (("%p,%u,...,%p,%p", p, (unsigned int) segment_cnt,
			callback, context));
  
  /* Convert the ethernet address (unique ID) to a GM ID. */
  
  if (gm_unique_id_to_node_id (p, (char *) ethernet_addr, &target_node_id)
      != GM_SUCCESS)
    {
      GM_PRINT (GM_DEBUG_ETHERNET,
		("cannot find target id for ether %x:%x:%x:%x:%x:%x\n",
		 ethernet_addr[0], ethernet_addr[1], ethernet_addr[2],
		 ethernet_addr[3], ethernet_addr[4], ethernet_addr[5]));
      /* We don't know where to send it, perhaps this came from an
         XP.  Ask the firmware to check in its hash table */
      type = GM_GATEWAY_SEND_EVENT;
      target_node_id = GM_NO_SUCH_NODE_ID;
    }
  else
    {
      type = GM_ETHERNET_SEND_EVENT;
    }
  
  gm_ether_cmn (p, segment_cnt, segment, ethernet_addr,
		type, target_node_id, callback, context);
  GM_RETURN_NOTHING ();
}

GM_ENTRY_POINT
void
gm_ethernet_broadcast_with_callback (gm_port_t * p,
				     gm_u32_t segment_cnt,
				     gm_ethernet_segment_descriptor_t
				     /* */ segment[],
				     gm_send_completion_callback_t callback,
				     void *context)
{
  static const gm_u8_t broadcast_mac[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
  
  GM_CALLED_WITH_ARGS (("%p,%u", p, (unsigned int) segment_cnt));
  gm_ether_cmn (p, segment_cnt, segment, broadcast_mac,
		GM_ETHERNET_SEND_EVENT, GM_NO_SUCH_NODE_ID, callback, context);
  GM_RETURN_NOTHING ();
}

#if GM_KERNEL

/* Special handler for GM_WAKE_INTERRUPTs on the ethernet port. */

static
void
__gm_ethernet_wake_callback (void *context)
{
  gm_instance_state_t *is;
  gm_port_state_t *ps;
  gm_port_t *p;
  gm_recv_event_t *e;
  int max_iters = (GM_NUM_SEND_TOKENS + GM_NUM_ETHERNET_RECV_TOKENS);

  GM_CALLED ();
  
  p = (gm_port_t *) context;
  gm_assert (p);
  ps = p->kernel_port_state;
  gm_assert (ps);
  is = p->kernel_port_state->instance;
  gm_assert (is);

  /* try to empty the receive queue, but don't iterate too long, since we
     are in the interrupt handler, after all. */
  
  while (max_iters--)
    {
      e = gm_blocking_receive_no_spin (p);
      switch (gm_ntoh_u8 (e->recv.type))
	{
	case GM_ETHERNET_RECV_EVENT:
	  /* receives cause the ethernet receive callback to be called. */
	  is->ethernet_recv_intr_callback
	    (is->ethernet_recv_intr_context, &e->ethernet_recv);
	  break;
	  
	case _GM_SLEEP_EVENT:
	  /* The _GM_SLEEP_EVENT means that
	     gm_blocking_receive_no_spin() has scheduled the next
	     wake.  Do NOT sleep, but depend on this function to be
	     called again upon the next wake interrupt. */
	  goto done;

	case GM_SENT_TOKENS_EVENT:
	  {
	    GM_ARCH_LOCK_ETHER_SEND_QUEUE (p);
	    gm_handle_sent_tokens (p, e);
	    GM_ARCH_UNLOCK_ETHER_SEND_QUEUE (p);
	  }
	  break;

	default:
	  /* This ensures that the ethernet send completion callback will
	     be called. */
	  gm_unknown (p, e);
	}
    }
 done:
  GM_RETURN_NOTHING ();
}

void
gm_ethernet_set_recv_intr_callback (gm_port_t * p,
				    void (*callback) (void *, 
						      gm_ethernet_recv_t *),
				    void *context)
{
  gm_port_state_t *ps;
  gm_instance_state_t *is;
  
  GM_CALLED_WITH_ARGS (("%p,%p,%p", p, callback, context));
  
  gm_assert (p);
  ps = p->kernel_port_state;
  gm_assert (ps);
  is = p->kernel_port_state->instance;
  gm_assert (is);

  /* record the callback and context */

  is->ethernet_recv_intr_callback = callback;
  is->ethernet_recv_intr_context = context;

  /* Install special wake event handler for this port */
  
  ps->wake.handler = __gm_ethernet_wake_callback;
  ps->wake.context = (void *) p;
  
  GM_RETURN_NOTHING ();
}

void
gm_ethernet_set_sent_intr_callback (gm_port_t * p,
				    void (*callback) (void *), void *context)
{
  GM_CALLED_WITH_ARGS (("%p,%p,%p", p, callback, context));
  
  gm_assert (p);
  gm_assert (p->kernel_port_state);
  gm_assert (p->kernel_port_state->instance);

  p->kernel_port_state->instance->ethernet_sent_intr_callback = callback;
  p->kernel_port_state->instance->ethernet_sent_intr_context = context;

  GM_RETURN_NOTHING ();
}

static
void
gm_ethernet_sent (gm_port_t *p, void *ignored, gm_status_t status)
{
  gm_instance_state_t *is;
  gm_port_state_t *ps;

  GM_CALLED ();
  gm_assert (p);
  ps = p->kernel_port_state;
  gm_assert (ps);
  is = ps->instance;
  gm_assert (is);
  if (is->ethernet_sent_intr_callback)
    {
      is->ethernet_sent_intr_callback (is->ethernet_sent_intr_context);
    }
  else
    {
      GM_WARN (("Ethernet send completion with no completion callback.\n"));
    }
  GM_RETURN_NOTHING ();
}

void
gm_ethernet_send (gm_port_t *p,
		  gm_u32_t segment_cnt,
		  gm_ethernet_segment_descriptor_t segment[],
		  gm_u8_t *ethernet_addr)
{
  GM_CALLED ();
  gm_ethernet_send_with_callback (p, segment_cnt, segment, ethernet_addr,
				  gm_ethernet_sent, (void *) 0);
  GM_RETURN_NOTHING ();
}

void
gm_ethernet_broadcast (gm_port_t *p,
		       gm_u32_t segment_cnt,
		       gm_ethernet_segment_descriptor_t segment[])
{
  GM_CALLED ();
  gm_ethernet_broadcast_with_callback (p, segment_cnt, segment, 
				       gm_ethernet_sent, (void *) 0);
  GM_RETURN_NOTHING ();
}

#endif /* GM_KERNEL */

/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
