/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* authors: feldy@myri.com
            loic@myri.com */

#if GM_KERNEL
#error This file is not supported in the kernel.
#endif

#include "gm_compiler.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_enable_fork_system.h"

#if HAVE_UNISTD_H && !defined WIN32
/* we do not want to (re)define fork on platforms where it is
   irrelevant */
#include <unistd.h>
#include <signal.h>

pid_t fork (void);        /* suppress missing prototype warning */

/* 'ENABLE' means allow users to have fork or system */
/*          means that we don't provide our own versions of fork/system */

#if (GM_ENABLE_FORK_SYSTEM==0)

#if GM_OS_LINUX
/* loic: for a few functions in the libc, __libc_foo() is the real def, 
   __foo and foo are weak aliases to it, 
   foo can be redefined to get a layer between the apps and the libc
   __foo can intercept calls made inside the libc
     (other libc functions using foo will call __foo()
   libpthread redefine __fork(), so we only redefine fork() to avoid
    conflict.
*/
pid_t __fork(void);

#endif

#if GM_OS_MACOSX
#include <mach-o/dyld.h>
typedef pid_t (gm_forkptr_t)(void);

static pid_t
gm_macosx_fork(void)
{
	gm_forkptr_t *fork_address;
	NSSymbol fork_symbol;
	const struct mach_header *libc;

	/* get a pointer to the symbols for libc */
	libc = NSAddImage("/usr/lib/libc.dylib", 0);
	if (libc == 0) {
		GM_WARN (("Couldn't map libc to find fork()\n"));
		goto abort_without_fork;
	}

	/* lookup fork in libc */
	fork_symbol = NSLookupSymbolInImage(libc, "_fork", 0);
	if (fork_symbol == 0) {
		GM_WARN (("Couldn't find fork() in libc\n"));
		goto abort_without_fork;
	}

	/* get a pointer for libc's fork */
	fork_address = NSAddressOfSymbol(fork_symbol);
	if (fork_address == 0) {
		GM_WARN (("Couldn't lookup fork()'s address in libc\n"));
		goto abort_without_fork;
	}

	/* finally, call libc's fork */
	return(fork_address());

 abort_without_fork:
	GM_WARN (("'fork' is NOT being called\n"));
	return -1;
}
#endif

pid_t
fork()
{
	if (_gm_open_ports_test()) {
		GM_WARN (("you can NOT call 'fork()' in a GM program with ports open.\n\n"));
		return -1;
	}
	else {
#if GM_OS_LINUX || GM_OS_AIX
		return(__fork());
#elif GM_OS_FREEBSD || GM_OS_OSF1
		return(_fork());
#elif GM_OS_MACOSX
		return(gm_macosx_fork());
#else
		GM_WARN (("'fork' is NOT being called - don't know how.\n"));
    	return -1;
#endif
	}
}


#endif /* GM_ENABLE_FORK_SYSTEM */
#endif /* HAVE_UNISTD_H */

int
gm_fork ()
{
	if (_gm_open_ports_test()) {
		GM_WARN (("you can NOT call 'fork()' in a GM program with ports open.\n\n"));
		return -1;
	}
	else {
#if HAVE_UNISTD_H && !defined WIN32
		return(fork());
#else
                return -1;
#endif
	}
}

#if GM_OS_LINUX
#include <syscall.h>

int gm_execve (const char *p, char*argv[], char *envp[])
{
  return syscall(__NR_execve,p,argv,envp);
}
#endif

#if GM_OS_LINUX  && !GM_CPU_ia64 && defined __OPTIMIZE__ && defined __GNUC__
/* because vfork can be redefined in libpthread.h as fork(),
   try to create an inline version instead,
   it _must_ be inline so only do this with optimisation */
#include <asm/unistd.h>
#ifdef _syscall0
#undef vfork
#define vfork gm_vfork_inline
#define __NR_gm_vfork_inline __NR_vfork
static inline _syscall0(int,gm_vfork_inline)
#endif /* _syscall0 */
#endif /* __OPTIMIZE__ ... */



/* always use our custom system() for Linux to minimize fork() use */
#if (GM_ENABLE_FORK_SYSTEM == 0) || GM_OS_LINUX
int
system (const char *cmd)
{
#if GM_OS_LINUX
  /* loic: Implementation of system on top on vfork.  I should admit I
     have read the GNU libc implementation of system (which use fork),
     so some credit should be given to whoever wrote that one in
     GLIBC.  This could potentially be used on others OS too, but
     vfork presence is not a sufficient reason to enable it, you need
     to be sure vfork is not defined as a normal fork, which some OS
     do */
  /*  fork () is currently not supported in GM because the duplication
      of the address space with copy-on-write techniques will cause
      inconsistencies betwee the process page table and the GM page
      hash. vfork () does not duplicate any page, does not do any
      address space manipulation, and is then perfectly compatible
      with GM. When fork() is implemented it could be very costly in
      the presence of registered memory, so vfork() is still quite
      useful */

  sigset_t mask, old_mask;
  struct sigaction new_sa, old_intr, old_quit;
  int status, cc;
  pid_t pid;

  cc = 0;
  /* 
     normally sigaction and sigmask never failed except for EFAULT
     moreover they are not critical, let's be paranoid and still
     get a useless boolean just in case */
  new_sa.sa_handler = SIG_IGN;
  new_sa.sa_flags = 0;
  cc |= sigaction (SIGINT, &new_sa, &old_intr);
  cc |= sigaction (SIGQUIT, &new_sa, &old_quit);
  sigemptyset (&mask);
  sigaddset (&mask, SIGCHLD);
  cc |= sigprocmask (SIG_BLOCK, &mask, &old_mask);
  
  pid = vfork ();
  if (pid == 0)
    {
      const char *argv[4];
      argv[0] = "/bin/sh";
      argv[1] = "-c";
      argv[2] = cmd;
      argv[3] = 0;
      __sigaction (SIGINT, &old_intr, NULL);
      __sigaction (SIGQUIT, &old_quit, NULL);
      sigprocmask (SIG_SETMASK, &old_mask, NULL);
      gm_execve (argv[0], (char **)argv, __environ);
      _exit(127);
    }
  else if (pid < 0)
    {
      status = -1;
    }
  else
    {
      /* vfork success */
      pid_t n;
      do
	{
	  pid_t waitpid (pid_t pid, int *status, int options);
	  
	  n = waitpid (pid, &status, 0);
	}
      while (n == -1 && errno == EINTR);
      gm_always_assert (n == -1 || pid == n);
    }
  cc |= sigaction (SIGINT, &old_intr, NULL);
  cc |= sigaction (SIGQUIT, &old_intr, NULL);
  cc |= sigprocmask (SIG_SETMASK, &old_mask, NULL);
  if (cc)
    GM_WARN (("system: warning setting signal handling failed\n"));
  return status;
#else /* GM_OS_LINUX */
  if (_gm_open_ports_test ())
    {
      GM_WARN (("you can NOT call 'system()' with GM ports open\n\n"));
      return -1;
    }
  else
    {
      GM_WARN (("'system' is NOT being called - don't know how.\n"));
      return -1;
    }
#endif
}

#endif /* GM_ENABLE_FORK_SYSTEM == 0 */
