/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_get.c
   This file contains the GM API function
   gm_get().
*/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_enable_get.h"
#include "gm_internal.h"
#include "gm_send_queue.h"

#if GM_DEBUG_GET
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/** gm_get() performs an RDMA Read operation.  

   gm_get() transfers the \b len bytes at
   \b remote_buffer to \b target_port_id on \b target_node_id with
   priority \b priority and stores the data at the local virtual
   memory address \b local_buffer.  Call callback(port,context,status)
   when the receive completes or fails, with \b status indicating the
   status of the receive.

   @param p \b (IN) The GM port on the destination GM node to
               which the communication is being received.
   @param remote__buffer \b (IN) Address of the remote buffer.
   @param local_buffer \b (OUT) Address of the local buffer.
   @param len \b (IN) The length in bytes of the buffer to be received.
   @param priority \b (IN) The priority of the data being received.
   @param target_node_id \b (IN) The GM node to which the data is
                      being received. 
   @param target_port_id \b (IN) The GM port on the destination GM node
                      to which the message is being received.
   @param callback \b (IN) The function called when the receive is complete.

   @param context \b (IN) Pointer to an integer or to a structure
                           that is passed to the callback function.

   @author Glenn Brown

   @version GM_API_VERSION (as defined in gm.h)

*/

GM_ENTRY_POINT
void
gm_get (gm_port_t * p,
	gm_remote_ptr_t remote_buffer,
	void *local_buffer,
	gm_size_t len,
	enum gm_priority priority,
	unsigned int target_node_id,
	unsigned int target_port_id,
	gm_send_completion_callback_t callback,
	void *context)
{
  GM_CALLED_WITH_ARGS (("%p, "GM_U64_TMPL", %p, %lu, %d, %u, %u, %p, %p",
			p, GM_U64_ARG (remote_buffer), local_buffer, len,
			priority, target_node_id, target_port_id, callback,
			context));
  
  gm_assert (p);
  gm_assert (local_buffer);
  gm_assert (remote_buffer);

#if 0
  if (!GM_ALIGNED (local_buffer, GM_DMA_GRANULARITY)
      || !GM_ALIGNED (remote_buffer, GM_DMA_GRANULARITY)
      || !GM_ALIGNED (len, GM_DMA_GRANULARITY))
    {
      GM_WARN (("this function only works for 8-byte-aligned pointers"
		" and lengths\n"));
      _GM_WARN (("    NOT sending message\n\n"));
      GM_RETURN_NOTHING ();
    }
#endif

  GM_SIMULATE_SIMULATE ( );

  /* Pass send token to LANai.  Care must be taken to flush the message to
     memory before flushing the token, and the token should be flushed to
     memory immediately. */

  {
    struct gm_put_get_send_event volatile *se;
    struct gm_put_get_send_event batch_write;

    se = GM_SEND_QUEUE_SLOT (p, callback, context, put_get);
    
    if (GM_STRUCT_WRITE_COMBINING)
      {
	batch_write.remote_buffer = gm_hton_rp (remote_buffer);
	batch_write.local_buffer = gm_hton_up ((gm_up_t) local_buffer);
	batch_write.length = gm_hton_u32 ((gm_u32_t) len);
	batch_write.remote_node_id = gm_hton_u16 ((gm_u16_t) target_node_id);
	batch_write.remote_subport_id
	  = gm_hton_u8 ((gm_u8_t) GM_SUBPORT (priority, target_port_id));
	batch_write.type = gm_hton_u8 ((gm_u8_t) GM_GET_SEND_EVENT);
	GM_COPY_TO_IO_SPACE (*se, batch_write);
      }
    else
      {
	se->remote_buffer = gm_hton_rp (remote_buffer);
	se->local_buffer = gm_hton_up ((gm_up_t) local_buffer);
	se->length = gm_hton_u32 ((gm_u32_t) len);
	se->remote_node_id = gm_hton_u16 ((gm_u16_t) target_node_id);
	se->remote_subport_id
	  = gm_hton_u8 ((gm_u8_t) GM_SUBPORT (priority, target_port_id));
	GM_STBAR ();
	se->type = gm_hton_u8 (GM_GET_SEND_EVENT);
      }
    GM_FLUSH_SEND_EVENT (p, se);
  }
  GM_RETURN_NOTHING ();
}
