/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2002 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_call_trace.h"
#include "gm_internal.h"
#include "gm_ptr_hash.h"

#if GM_KERNEL
#include "gm_lanai_command.h"
#include "gm_undocumented_funcs.h"
#endif

/**
   gm_global_id_to_node_id() converts a GM node ID into a GM
   connection ID.

   A GM connection ID is a small integer uniquely identifying the
   remote host, but which is valid only only for this port.
   Connections are numbered from 0 to gm_max_node_id(port).

   A GM node ID is a global unique ID uniquely identifying a GM board.
   GM node ID's are sparsely packed.  That is, they are not sequentially
   numbered.
   
   @retval GM_SUCCESS Operation completed successfully.
 
   @param port \b (IN) the GM port associated with the specific GM
   connection ID
 
   @param node_id \b (IN) the connection ID to translate

   @param global_id \b (OUT) a pointer to storage for resulting node ID.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

#define GM_DEBUG_GLOBAL_ID_TO_NODE_ID 0

GM_ENTRY_POINT
gm_status_t
gm_global_id_to_node_id (gm_port_t *port,
			     unsigned int global_id,
			     unsigned int *node_id)
{
  GM_CALLED_WITH_ARGS_EX
    (GM_DEBUG_GLOBAL_ID_TO_NODE_ID,
     ("%p,0x%x,%p", port, global_id, node_id));
#if GM_KERNEL
  {
    gm_instance_state_t *is;
    gm_port_state_t *ps;
    void *hashed_node_id;

    gm_assert (port);
    ps = port->kernel_port_state;
    gm_assert (ps);
    is = ps->instance;
    gm_assert (is);

    /* Lookup the answer in the host's per-instance hash table */

    hashed_node_id = gm_ptr_hash_find (is->node_id_hash,
                                       (void *)(gm_size_t)global_id);
    if (hashed_node_id != 0)
      {
	*node_id = (unsigned int)(gm_size_t)hashed_node_id;
	GM_RETURN_STATUS (GM_SUCCESS);
      }
    
    GM_RETURN_STATUS (GM_FAILURE);

  }
#else  /* !GM_KERNEL */
  {
    gm_status_t status;
    unsigned int cached_node_id;

    /* Try to return a translation from the cache. */

    gm_assert (port->global_id_cache.hash);
    cached_node_id
      = ((unsigned int)
	 (gm_size_t)
	 gm_ptr_hash_find (port->global_id_cache.hash,
			   (void *) (gm_size_t) global_id));
    if (cached_node_id)
      {
	*node_id = cached_node_id;
	GM_RETURN_STATUS (GM_SUCCESS);
      }
    
    /* If no cached value, update the cache to contain all values. */

    status = _gm_update_global_id_caches (port);
    if (status != GM_SUCCESS)
      {
	GM_RETURN_STATUS (status);
      }

    /* Return value from the cache or fail. */

    cached_node_id
      = ((unsigned int)
	 (gm_size_t)
	 gm_ptr_hash_find (port->global_id_cache.hash,
			   (void *) (gm_size_t) global_id));
    if (!cached_node_id)
      {
	GM_RETURN_STATUS (GM_INVALID_PARAMETER);
      }
  
    *node_id = cached_node_id;
    GM_RETURN_STATUS (GM_SUCCESS);
  }
#endif
}
