/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_host_name_to_node_id.c
   The file containing the GM API function gm_host_name_to_node_id().
*/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_debug_yp.h"
#include "gm_internal.h"
#include "gm_lanai_command.h"

#define GM_DEBUG_HOSTNAME_TO_NODE_ID 0

#if GM_DEBUG_HOSTNAME_TO_NODE_ID
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/**
   gm_host_name_to_node_id_ex() translates a GM host name to a node ID.
 
   @retval GM_SUCCESS Operation completed successfully.

   @param port \b (IN) The GM port associated with the specific GM ID.
 
   @param timeout_usecs \b (IN) The maximum length of time (in
   		microseconds) to query, or 0 to use the default (long)
   		timeout.
 
   @param host_name \b (IN)

   @param node_id \b (OUT) Where to store the node ID.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT gm_status_t
gm_host_name_to_node_id_ex (gm_port_t *port,
			    unsigned int timeout_usecs,
			    const char *host_name,
			    unsigned int *node_id)
{
  gm_status_t status;
  
  GM_CALLED_WITH_ARGS
    (("%p,%d,\"%s\",%p", port, timeout_usecs, host_name, node_id));

#if GM_KERNEL
  {
    gm_port_state_t *ps;
    
    gm_assert (port);
    ps = port->kernel_port_state;
    gm_assert (ps);
    
    /* Translate the name using the yellow pages.  We are looking for
       a node_id with a "gm.hostname" of hostname. */

    *node_id = 0;		/* broadcast the query */
    status = gm_lanai_command_yp (ps,
				  timeout_usecs,
				  "gm.hostname",
				  host_name,
				  node_id, /* Answer is stored here. */
				  0 /* ignore match value. */
				  );
    if (status == GM_SUCCESS)
      {
	_GM_PRINT (GM_DEBUG_YP, ("node_id is %d\n", *node_id));
      }
    GM_RETURN_STATUS (status);
  }
#else
  {
    union
    {
      struct 
      {
	unsigned int timeout_usecs;
	char name[GM_MAX_HOST_NAME_LEN+1];
      } in;
      unsigned int node_id;
    } in_out;

    in_out.in.timeout_usecs = timeout_usecs;
    gm_strncpy (in_out.in.name, host_name, sizeof (in_out.in.name));
    status
      = (_gm_user_ioctl
	 (port, GM_HOST_NAME_TO_NODE_ID, &in_out, sizeof (in_out)));
    *node_id = in_out.node_id;
    if (status == GM_SUCCESS)
      {
	_GM_PRINT (GM_DEBUG_YP, ("node_id is %d\n", *node_id));
      }
    GM_RETURN_STATUS (status);
  }
#endif
}

/**
   This function is deprecated.  Use gm_host_name_to_node_id_ex()
   instead.

   gm_host_name_to_node_id() returns the GM ID associated \b host_name
   or GM_NO_SUCH_NODE_ID in case of an error.
 
   @retval GM_SUCCESS Operation completed successfully.
 
   @retval GM_NO_SUCH_NODE_ID Error.
 
   @param port \b (IN) The GM port associated with the specific GM ID.
 
   @param host_name \b (OUT) Where to store the host name.  At most
   GM_MAX_HOST_NAME_LEN+1 bytes will be written.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT unsigned int
gm_host_name_to_node_id (gm_port_t *port, char *host_name)
{
  gm_status_t status;
  unsigned int node_id;
  
  GM_CALLED_WITH_ARGS (("%p,\"%s\"", port, host_name));
  status = gm_host_name_to_node_id_ex (port,
				       0, /* use default (long) timeout */
				       host_name,
				       &node_id);
  if (status != GM_SUCCESS)
    {
      GM_RETURN (GM_NO_SUCH_NODE_ID);
    }
  GM_RETURN (node_id);
}
