/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_lookaside.c
   This file contains the GM API functions gm_create_lookaside(),
   gm_destroy_lookaside(), gm_lookaside_alloc(), gm_lookaside_zalloc(),
   and gm_lookaside_free().
 
   This file implements a lookaside list.  It is mainly intended to allow
   efficient memory allocation of small structures in kernels where the
   minimum memory allocation is a page, but is not kernel-specific.
 
   GM implements a lookaside list, which may be used to manage small
   fixed-length blocks more efficiently than gm_malloc() and gm_free().
   Lookaside lists can also be used to ensure that at least a minimum
   number of blocks are available for allocation at all times.

   gm_lookaside_alloc() returns cache-line-aligned buffers, in an attempt
   to minimize the cost of accessing data stored in the buffers. */

#include "gm.h"
#include "gm_cache_line.h"
#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_malloc_debug.h"
#include "gm_struct_lock.h"

#define GM_DEBUG_LOOKASIDE 0

#if GM_DEBUG_LOOKASIDE
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

struct gm_lookaside_free_entry
{
  struct gm_lookaside_free_entry *next;
};

/** Structure holding a group of allocated lookaside entries.  It is
    used primarily when entries are smaller than a page to allocate a
    page worth of entries at a time.  This is important since some
    kernel memory allocators (Windows) round all allocations up to a page
    length, but this implementation allows us to be efficient for
    small allocations despite this. */
struct gm_lookaside_segment
{
  struct gm_lookaside_segment *next;
  struct gm_lookaside_segment *prev;
  struct gm_lookaside_free_entry *free_entries;
  struct gm_lookaside *lookaside;
  gm_size_t alloced_entry_cnt;
  char data[1];
};

GM_TOP_LEVEL_ASSERT (sizeof (gm_size_t) == GM_SIZEOF_SIZE_T);
#define GM_LOOKASIDE_SEGMENT_DATA_OFFSET			\
	    GM_OFFSETOF (struct gm_lookaside_segment, data[0])

/** State of a lookaside list, which is an very fast and space efficient
    memory allocator for fixed-sized buffers. */
struct gm_lookaside
{
  unsigned long entry_len;
  unsigned long entry_cnt;
  unsigned long min_entry_cnt;
  /** List of segments.  Segments with free entries are at the front. */
  struct gm_lookaside_segment_list
  {
    struct gm_lookaside_segment *next;
    struct gm_lookaside_segment *prev;
  }
  segment_list;
  struct gm_lookaside_free_entry *delayed_free_entry;
  GM_STRUCT_LOCK_DECL
};

/****************************************************************
 * Protos
 ****************************************************************/

static void _gm_lookaside_free (struct gm_lookaside_free_entry *free_entry);

/****************************************************************
 * Vars
 ****************************************************************/

/* count of user lookaside tables */

static unsigned long _gm_lookaside_cnt;

/* An internal lookaside, from which lookaside structures are allocated. */

static struct gm_lookaside _gm_internal_lookaside;

/* Return the number of entries that will fit in a page with a little
   room to spare, or 1, whichever is larger. */

static unsigned long
_gm_lookaside_entries_per_seg (struct gm_lookaside *l)
{
  unsigned long ret;

  gm_assert (l->entry_len);
  gm_assert (GM_PAGE_LEN >  (unsigned long) GM_LOOKASIDE_SEGMENT_DATA_OFFSET);
  ret = (unsigned long) ((GM_PAGE_LEN - GM_LOOKASIDE_SEGMENT_DATA_OFFSET)
			 / l->entry_len);

  return (ret ? ret : 1);
}

static unsigned long
_gm_lookaside_segment_alloc_len (struct gm_lookaside *l)
{
  return (unsigned long) ((GM_LOOKASIDE_SEGMENT_DATA_OFFSET
			   + (l->entry_len
			      * _gm_lookaside_entries_per_seg (l))));
}

/************
 * segment list manipulation
 ************/

/* remove segment from segment list. */

static gm_inline void
_gm_lookaside_segment_list_remove (struct gm_lookaside_segment *seg)
{
  struct gm_lookaside_segment *next, *prev;

  next = seg->next;
  prev = seg->prev;
  next->prev = prev;
  prev->next = next;
}

/* Prepend segment to segment list */

static gm_inline void
_gm_lookaside_segment_list_prepend (struct gm_lookaside *l,
				    struct gm_lookaside_segment *seg)
{
  _gm_lookaside_segment_list_remove (seg);

  seg->next = l->segment_list.next;
  seg->prev = (struct gm_lookaside_segment *) &l->segment_list;
  seg->next->prev = seg;
  l->segment_list.next = seg;
}

/* Append segment to segment list */

static gm_inline void
_gm_lookaside_segment_list_append (struct gm_lookaside *l,
				   struct gm_lookaside_segment *seg)
{
  _gm_lookaside_segment_list_remove (seg);

  seg->next = (struct gm_lookaside_segment *) &l->segment_list;
  seg->prev = l->segment_list.prev;
  l->segment_list.prev = seg;
  seg->prev->next = seg;
}

/************
 * Utilities
 ************/

static gm_status_t
_gm_lookaside_grow (struct gm_lookaside *l)
{
  gm_size_t entry_len;
  struct gm_lookaside_free_entry *e;
  struct gm_lookaside_free_entry *free_entries = 0;
  struct gm_lookaside_segment *seg;
  unsigned long alloc_len;
  unsigned long entry_cnt;
  
  /* allocate another segment to create more entries. */

  alloc_len = _gm_lookaside_segment_alloc_len (l);
  seg = (struct gm_lookaside_segment *) gm_alloc_pages (alloc_len);
  if (!seg)
    return GM_OUT_OF_MEMORY;
  gm_assert (GM_PAGE_ALIGNED (seg));

  /* init segment */
  gm_bzero (seg, sizeof (*seg));	/* OK to not zero entries. */
  seg->lookaside = l;
  seg->next = seg->prev = seg;

  /* Add each entry to free list. */

  entry_len = l->entry_len;
  entry_cnt = l->entry_cnt;
  for (e = ((struct gm_lookaside_free_entry *)
	    ((char *) seg + alloc_len - entry_len));
       e >= (struct gm_lookaside_free_entry *) seg->data;
       e = (struct gm_lookaside_free_entry *)((char *)e - entry_len))
    {
      e->next = free_entries;
      free_entries = e;
      entry_cnt++;
    }
  seg->free_entries = free_entries;
  l->entry_cnt = entry_cnt;

  /* Prepend segment to segment list, since it has free entries. */
  _gm_lookaside_segment_list_prepend (l, seg);
  return GM_SUCCESS;
}

/* finalize an initialized lookaside table. */

#define GM_DEBUG_LOOKASIDE_FINALIZE (GM_DEBUG_LOOKASIDE || 0)
static gm_status_t
_gm_lookaside_finalize (struct gm_lookaside *l)
{
  struct gm_lookaside_segment *seg, *next;
  unsigned long residual_cnt = 0;

  GM_CALLED ();

#if GM_DEBUG_MALLOC
  /* Prevent auto-freeing from confusing GM_DEBUG_MALLOC by freeing all
     the unfreed lookaside entries that the client did not free. */
  {
    for (seg = l->segment_list.next;
	 seg != (struct gm_lookaside_segment *) &l->segment_list;
	 seg = next)
      {
	gm_u8_t *buf;

	next = seg->next;
	for (buf = (gm_u8_t *) seg->data;
	     buf < (gm_u8_t *) seg + GM_PAGE_LEN;
	     buf += l->entry_len)
	  {
	    if (_gm_allocated (buf))
	      {
		gm_lookaside_free (buf);
	      }
	  }
      }
  }
#endif

  /* Free all the segments in the segment list.  There is no need to
     free any allocated lookaside entries in each segment. */

  /* Null terminate the cyclic list to simplify the for loop. */
  l->segment_list.prev->next = 0;
  for (seg = l->segment_list.next; seg; seg = next)
    {
      residual_cnt++;
      gm_assert (GM_PAGE_ALIGNED (seg));
      next = seg->next;
      gm_free_pages (seg, _gm_lookaside_segment_alloc_len (l));
    }

  if (residual_cnt)
    {
      GM_PRINT (GM_DEBUG_LOOKASIDE,
		("Automatically freeing %lu residual segments in lookaside "
		 "list.\n", residual_cnt));
    }

  GM_RETURN_STATUS (GM_SUCCESS);
}

/* initialize a preallocated lookaside table. */

static gm_status_t
_gm_lookaside_init (struct gm_lookaside *l,
		    gm_size_t entry_len, gm_size_t min_entry_cnt)
{
  gm_status_t status;
  
  GM_CALLED ();
  GM_PRINT (GM_PRINT_LEVEL >= 10,
	    ("lookaside = 0x%p, len = 0x%lx,cnt =%ld \n", l,
	     entry_len, min_entry_cnt));

  /* Make sure entries are large enough for the pointer used to chain
     them when freed. */

  if (entry_len < sizeof (void *))
    {
      entry_len = sizeof (void *);
    }

  /* Increase entry size to arrange for entries to either start on a
     cache line or not cross a cache line boundary. */

  if (entry_len < GM_CACHE_LINE_LEN)
    {
      entry_len = 1 << gm_log2_roundup ((unsigned long)entry_len);
    }
  else if (entry_len > GM_CACHE_LINE_LEN)
    {
      entry_len = GM_ROUNDUP (size, entry_len, GM_CACHE_LINE_LEN);
    }
  
  /* Initialize fields. */

  GM_PRINT (GM_PRINT_LEVEL >= 10, ("initialize fields \n"));
  gm_bzero (l, sizeof (*l));
  l->entry_len = (unsigned long) entry_len;
  l->min_entry_cnt = (unsigned long) min_entry_cnt;
  l->segment_list.next
    = l->segment_list.prev = (struct gm_lookaside_segment *) &l->segment_list;

  /* Preallocate the minimum number of entries. */

  GM_PRINT (GM_PRINT_LEVEL >= 10, ("growing\n"));
  while (l->entry_cnt < l->min_entry_cnt)
    {
      if (_gm_lookaside_grow (l) != GM_SUCCESS)
	{
	  status = GM_OUT_OF_MEMORY;
	  goto abort_with_alloced_entries;
	}
    }
  GM_PRINT (GM_PRINT_LEVEL >= 10, ("grew\n"));
  GM_STRUCT_LOCK (*l);
  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_alloced_entries:
  _gm_lookaside_finalize (l);
  GM_RETURN_STATUS (status);
}

/************
 * Entry points
 ************/

/* Allocate an entry from the lookaside table. */

static void *
_gm_lookaside_alloc (struct gm_lookaside *l)
{
  struct gm_lookaside_segment *seg;
  struct gm_lookaside_free_entry *free_entry;

  GM_CALLED_WITH_ARGS (("%p", l));

  GM_PRINT (GM_PRINT_LEVEL >= 10, (" lookaside buffer = 0x%p\n", l));

  if (l->delayed_free_entry)
    {
      GM_PRINT (GM_PRINT_LEVEL >= 10, (" free buffer = %p\n", l));
      /* If an entry was schedule to be freed, simply use that entry. */

      free_entry = l->delayed_free_entry;
      l->delayed_free_entry = 0;
    }
  else
    {
      /* Arrange for there to be free entries. */

      GM_PRINT (GM_PRINT_LEVEL >= 10, (" Arrange for free entries\n"));
      seg = l->segment_list.next;
      if (seg == (struct gm_lookaside_segment *) &l->segment_list
	  || !seg->free_entries)
	{
	  if (_gm_lookaside_grow (l) != GM_SUCCESS)
	    GM_RETURN_PTR (0);
	  seg = l->segment_list.next;
	  gm_assert (seg != (struct gm_lookaside_segment *) &l->segment_list
		     && seg->free_entries);
	}

      /* Use a free entry. */

      GM_PRINT (GM_PRINT_LEVEL >= 10, (" set free entries \n"));
      free_entry = seg->free_entries;
      seg->free_entries = free_entry->next;
      seg->alloced_entry_cnt++;
      if (!seg->free_entries)
	_gm_lookaside_segment_list_append (l, seg);
    }
  GM_PRINT (GM_PRINT_LEVEL >= 10, (" return from gm_lookaside_alloc \n"));
  GM_RETURN_PTR ((void *) free_entry);
}

/**
 
   gm_lookaside_alloc() allocates an entry from the lookaside table, with
   debugging.  It returns a buffer of size ENTRY_LEN specified when the
   entry list L was created, or `0' if the buffer could not be allocated.
 
   @retval ptr Buffer of size ENTRY_LEN.
 
   @param l \b (IN) Handle to the gm_lookaside list.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT void *
gm_lookaside_alloc (struct gm_lookaside *l)
{
  void *ret;

  GM_CALLED ();
  GM_STRUCT_UNLOCK (*l);

  ret = _gm_lookaside_alloc (l);
  if (!ret)
    goto abort_with_struct_unlocked;
  if (GM_MDEBUG_RECORD_PTR (ret, gm_lookaside_free) != GM_SUCCESS)
    goto abort_with_ret;
  
  GM_STRUCT_LOCK (*l);
  GM_RETURN_PTR (ret);
  
 abort_with_ret:
  _gm_lookaside_free (ret);
 abort_with_struct_unlocked:
  GM_STRUCT_LOCK (*l);
  GM_RETURN_PTR (0);
}

/**
 
   gm_lookaside_zalloc() allocates and clear an entry from the lookaside
   table.
 
   @retval ptr
 
   @param l \b (IN) Handle to the gm_lookaside list.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT void *
gm_lookaside_zalloc (struct gm_lookaside *l)
{
  void *ret;

  GM_CALLED ();
  GM_STRUCT_UNLOCK (*l);

  ret = gm_lookaside_alloc (l);
  if (ret)
    {
      gm_bzero (ret, l->entry_len);
    }
  
  GM_STRUCT_LOCK (*l);
  GM_RETURN_PTR (ret);
}

/* Immediately free an entry allocated from the lookaside list. */

static void
_gm_lookaside_free (struct gm_lookaside_free_entry *free_entry)
{
  struct gm_lookaside_segment *seg;
  struct gm_lookaside *l;

  GM_CALLED_WITH_ARGS (("%p", free_entry));

  /* Infer the segment and lookaside for the entry. */

  gm_assert (!GM_PAGE_ALIGNED (free_entry));
  seg = (struct gm_lookaside_segment *) GM_PAGE_ROUND_DOWN (ptr, free_entry);
  gm_assert (seg);
  gm_assert (GM_PAGE_ALIGNED (seg));
  l = seg->lookaside;
  gm_assert (l);

  /* free the entry */

  free_entry->next = seg->free_entries;
  seg->free_entries = free_entry;
  gm_assert (seg->alloced_entry_cnt);
  seg->alloced_entry_cnt--;
  _gm_lookaside_segment_list_prepend (l, seg);

  /* free the segment if all the entries in the segment have been
     freed and doing so would not cause there to be less than the
     minimum number of entries. */

  if (!seg->alloced_entry_cnt)
    {
      unsigned long per_seg;

      per_seg = _gm_lookaside_entries_per_seg (l);
      if (l->entry_cnt - per_seg >= l->min_entry_cnt)
	{
	  _gm_lookaside_segment_list_remove (seg);
	  gm_free_pages (seg, _gm_lookaside_segment_alloc_len (l));
	  l->entry_cnt -= per_seg;
	}
    }

  GM_RETURN_NOTHING ();
}

/**
 
   gm_lookaside_free() schedules an allocated entry to be freed, and actually
   performs any scheduled free.
   It frees a block of memory previously allocated by a call to
   gm_lookaside_alloc().  The contents of the block of memory are
   guaranteed to be unchanged until the next operation is performed on the
   lookaside list.
 
   @param ptr \b (IN) Pointer to the entry to be freed.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT void
gm_lookaside_free (void *ptr)
{
  struct gm_lookaside_segment *seg;
  struct gm_lookaside_free_entry *free_entry;
  struct gm_lookaside *l;

  GM_CALLED_WITH_ARGS (("%p", ptr));

  GM_MDEBUG_REMOVE_PTR (ptr, gm_lookaside_free);

  /* infer free entry, segment, and lookaside list for the buffer. */

  free_entry = (struct gm_lookaside_free_entry *) ptr;
  seg = (struct gm_lookaside_segment *) GM_PAGE_ROUND_DOWN (ptr, ptr);
  gm_assert (GM_PAGE_ALIGNED (seg));
  l = seg->lookaside;
  GM_STRUCT_UNLOCK (*l);

  /* perform previously delayed free. */

  if (l->delayed_free_entry)
    _gm_lookaside_free (l->delayed_free_entry);

  /* delay the current free */

  l->delayed_free_entry = free_entry;

  GM_STRUCT_LOCK (*l);
  GM_RETURN_NOTHING ();
}

/**
 
   gm_create_lookaside() returns a newly created lookaside list to be used
             to allocate blocks of ENTRY_LEN bytes.  MIN_ENTRY_CNT entries
             are preallocated.
 
   @retval gm_lookaside Handle to the lookaside table.
 
   @param entry_len \b (IN)
   @param min_entry_cnt \b (IN)
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT struct gm_lookaside *
gm_create_lookaside (gm_size_t entry_len, gm_size_t min_entry_cnt)
{
  struct gm_lookaside *ret;

  GM_CALLED ();

  /* Initialize the lookaside package, if needed. */

  if (_gm_lookaside_cnt == 0)
    {
      gm_always_assert (GM_PAGE_LEN != 0);
      if (_gm_lookaside_init (&_gm_internal_lookaside,
			      sizeof (struct gm_lookaside),
			      0) != GM_SUCCESS)
	{
	  goto abort_with_nothing;
	}
    }

  /* Allocate the lookaside */

  ret = gm_lookaside_alloc (&_gm_internal_lookaside);
  if (!ret)
    goto abort_with_internal_lookaside;
  _gm_lookaside_cnt++;

  /* Initialize the user's lookaside table. */

  if (_gm_lookaside_init (ret, entry_len, min_entry_cnt) != GM_SUCCESS)
    goto abort_with_lookaside;

  GM_PRINT (GM_PRINT_LEVEL >= 10, ("There are %lx lookasides after create.\n",
				   _gm_lookaside_cnt));
  GM_RETURN_PTR (ret);

abort_with_lookaside:
  gm_lookaside_free (ret);
  _gm_lookaside_cnt--;
abort_with_internal_lookaside:
  if (_gm_lookaside_cnt == 0)
    _gm_lookaside_finalize (&_gm_internal_lookaside);
abort_with_nothing:
  GM_PRINT (GM_PRINT_LEVEL >= 10,
	    ("GM_create_lookaside FAILED, cnt = 0x%lx\n", _gm_lookaside_cnt));
  GM_RETURN_PTR (0);
}

/**
 
   gm_destroy_lookaside() frees a lookaside list and all associated
            resources, including any buffers currently allocated from
            the lookaside list.
 
   @param l \b (IN) Handle to the lookaside table.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT void
gm_destroy_lookaside (struct gm_lookaside *l)
{
  GM_CALLED ();

  if (!l)
    {
      GM_WARN (("caller tried to free NULL lookaside.\n"));
      GM_RETURN_NOTHING ();
    }

  /* destroy the user's lookaside. */

  GM_STRUCT_UNLOCK (*l);
  _gm_lookaside_finalize (l);
  gm_lookaside_free (l);
  _gm_lookaside_cnt--;

  GM_PRINT (GM_PRINT_LEVEL >= 10,
	    ("There are %lx lookasides after destroy.\n", _gm_lookaside_cnt));

  /* If all lookasides have been destroyed, destroy the internal
     lookaside list. */

  if (_gm_lookaside_cnt == 0)
    {
      GM_PRINT (GM_PRINT_LEVEL >= 10, ("finalizing lookaside support\n"));
      _gm_lookaside_finalize (&_gm_internal_lookaside);
    }

  GM_STRUCT_LOCK (*l);
  GM_RETURN_NOTHING ();
}


/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
