/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

/* This file implements memory leak detection support for GM.  Because
   it debugs the GM memory allocators, it must not use the
   instrumented memory allocator interfaces.  Consequently, it does
   not use the documented GM memory allocator interfaces, using only
   the uninstrumented _gm_malloc() and _gm_free() functions for memory
   allocation.  This file implements its own hash table and simple
   lookaside list to for low overhead debugging while avoiding similar
   instrumented support in the GM API. */

#include "gm.h"
#include "gm_call_trace.h"
#include "gm_debug_malloc.h"
#include "gm_malloc_debug.h"

/** A record holding a pointer free function that an allocation should be
    used to free the allocation, and the top of the call stack when
    the memory was allocated.  This allows us to report when the wrong
    function is used to free the memory, and to report the call stack
    of the allocation in this case or if a memory leak is detected. */
struct gm_mdebug
{
  char *free_function;
  char *caller[5];
};

static unsigned long user_alloc_cnt;
static unsigned long internal_alloc_cnt;

/****************************************************************
 * Simple lookaside allocator for mdebug records.
 ****************************************************************/

/** Union to allow chaining of free mdebug records, without requiring
    an extra field in the mdebug record. */
union gm_free_mdebug
{
  union gm_free_mdebug *next_free;
  struct gm_mdebug mdebug;
};

/** Holds a groups of preallocated malloc debugging records.  We
    allocated these records in chunks to reduce the average allocation
    overhead, and for storage density.  We allocate 2^N-1 records at a
    time because many malloc() implementations use memory most
    efficiently for allocations just smaller than 2^M bytes long. */
struct preallocated_record_chunk 
{
  struct preallocated_record_chunk *next;
  union gm_free_mdebug record[63];
};

union gm_free_mdebug *free_records;
struct preallocated_record_chunk *preallocated_record_chunks;

static gm_inline
void
free_record (struct gm_mdebug *record)
{
  union gm_free_mdebug *f;
  f = (union gm_free_mdebug *) record;
  f->next_free = free_records;
  free_records = f;
}

static
gm_status_t
preallocate_more_records (void)
{
  struct preallocated_record_chunk *c;
  unsigned long i;

  GM_CALLED ();
  c = _gm_malloc (sizeof (struct preallocated_record_chunk));
  if (!c)
    {
      GM_RETURN (GM_OUT_OF_MEMORY);
    }
  internal_alloc_cnt++;
  c->next = preallocated_record_chunks;
  preallocated_record_chunks = c;
  for (i=0; i<GM_NUM_ELEM(c->record); i++)
    {
      free_record (&c->record[i].mdebug);
    }
  GM_RETURN (GM_SUCCESS);
}

static
struct gm_mdebug *
alloc_record (void)
{
  union gm_free_mdebug *f;

  GM_CALLED ();
  
  /* If possible, return a previously allocated record */

  f = free_records;
  if (f)
    {
      free_records = f->next_free;
      GM_RETURN (&f->mdebug);
    }

  /* preallocate more records */
  
  if (preallocate_more_records () != GM_SUCCESS)
    GM_RETURN (0);

  /* return a preallocated record */
  
  f = free_records;
  gm_assert (f);
  free_records = f->next_free;
  GM_RETURN (&f->mdebug);
}

static
void
finalize_all_records (void)
{
  struct preallocated_record_chunk *c, *next;

  GM_CALLED ();
  c = preallocated_record_chunks;
  preallocated_record_chunks = 0;
  while (c)
    {
      next = c->next;
      _gm_free (c);
      internal_alloc_cnt--;
      c = next;
    }
  GM_RETURN_NOTHING();
}

/****************************************************************
 * Hash table for tracking allocations.
 ****************************************************************/

/** An entry in a custom hash table used to track memory allocations. */
struct hash_entry 
{
  void *ptr;
  struct gm_mdebug *record;
};

static unsigned long hash_max_index;
static unsigned long hash_population;
static struct hash_entry *hash_entry;

static gm_status_t hash_grow (void);

static gm_status_t
hash_insert (void *ptr, struct gm_mdebug *record)
{
  gm_status_t status;
  int loop_cnt = 0;
  unsigned long pos;
  GM_VAR_MAY_BE_UNUSED (loop_cnt);

  GM_CALLED ();
  if (ptr == 0)
    {
      GM_RETURN_STATUS (GM_INVALID_PARAMETER);
    }
  
  /* grow the hash table if it is more than half full */
  
  if (hash_population >= hash_max_index/2)
    {
      status = hash_grow ();
      if (status != GM_SUCCESS)
	GM_RETURN (status);
    }

  /* find the bin */

  gm_assert (GM_POWER_OF_TWO (hash_max_index + 1));
  pos = gm_crc (&ptr, sizeof (ptr)) & hash_max_index;
  while (hash_entry[pos].ptr)
    {
      pos = (pos + 1) & hash_max_index;
      gm_assert (loop_cnt++ < 1000);
    }

  /* record the entry */
  
  hash_entry[pos].ptr = ptr;
  hash_entry[pos].record = record;
  ++hash_population;
  GM_RETURN (GM_SUCCESS);
}

static struct gm_mdebug *
hash_find (void *ptr)
{
  unsigned long pos;

  GM_CALLED ();
  gm_assert (ptr);
  
  if (!hash_entry)
    {
      GM_RETURN (0);
    }
  
  /* find the corresponding entry... */
  
  pos = gm_crc (&ptr, sizeof (ptr)) & hash_max_index;
  while (hash_entry[pos].ptr != ptr)
    {
      if (hash_entry[pos].ptr == 0)
	GM_RETURN (0);
      pos = (pos + 1) & hash_max_index;
    }
  GM_RETURN (hash_entry[pos].record);
}

static void
hash_remove (void *ptr)
{
  unsigned long pos;

  GM_CALLED ();
  gm_assert (ptr);
  
  /* find the corresponding entry... */
  
  pos = gm_crc (&ptr, sizeof (ptr)) & hash_max_index;
  while (hash_entry[pos].ptr != ptr)
    {
      if (hash_entry[pos].ptr == 0)
	GM_RETURN_NOTHING ();
      pos = (pos + 1) & hash_max_index;
    }
  
  /* ...and remove it. */
  
  hash_entry[pos].ptr = 0;
  free_record (hash_entry[pos].record);
  --hash_population;
  
  /* Remove and reinsert all subsequent hash entries to remove any bubble
     that was introduced */
  
  pos = (pos + 1) & hash_max_index;
  while ((ptr = hash_entry[pos].ptr) != 0)
    {
      gm_status_t status;

      hash_entry[pos].ptr = 0;
      --hash_population;
      status = hash_insert (ptr, hash_entry[pos].record);
      gm_assert (status == GM_SUCCESS);
      pos = (pos + 1) & hash_max_index;
    }
  
  GM_RETURN_NOTHING ();
}

static gm_status_t
hash_grow (void)
{
  gm_size_t byte_cnt;
  gm_status_t status;
  unsigned long old_hash_max_index;
  unsigned long old_hash_population;
  struct hash_entry *old_hash_entry;

  GM_CALLED ();
  
  old_hash_max_index = hash_max_index;
  old_hash_population = hash_population;
  old_hash_entry = hash_entry;
  if (hash_max_index == 0)
    {
      hash_max_index = 127;
      gm_assert (hash_population == 0);
    }
  else
    {
      hash_max_index += hash_max_index + 1;
    }
  
  byte_cnt = (hash_max_index + 1) * sizeof (struct hash_entry);
  hash_entry = _gm_malloc (byte_cnt);
  if (!hash_entry)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_nothing;
    }
  internal_alloc_cnt++;
  gm_bzero (hash_entry, byte_cnt);
  if (old_hash_entry)
    {
      unsigned long i;
      
      hash_population = 0;
      for (i=0; i<=old_hash_max_index; i++)
	{
	  if (old_hash_entry[i].ptr)
	    {
	      hash_insert (old_hash_entry[i].ptr, old_hash_entry[i].record);
	    }
	}
      _gm_free (old_hash_entry);
      internal_alloc_cnt--;
    }
  GM_RETURN_STATUS (GM_SUCCESS);
  
 abort_with_nothing:
  hash_max_index = old_hash_max_index;
  hash_population = old_hash_population;
  hash_entry = old_hash_entry;
  GM_RETURN_STATUS (status);
}

static void
hash_reset (void)
{
  GM_CALLED ();
  
  if (hash_entry == 0)
    GM_RETURN_NOTHING ();
  _gm_free (hash_entry);
  internal_alloc_cnt--;
  hash_entry = 0;
  hash_max_index = 0;
  hash_population = 0;

  finalize_all_records ();

  GM_RETURN_NOTHING ();
}

static void report (void *ptr, struct gm_mdebug *record);

static void
hash_report (void)
{
  unsigned long i;

  GM_CALLED ();

  if (user_alloc_cnt == 0)
    {
      GM_PRINT (GM_DEBUG_MALLOC, ("No memory leak detected.\n"));
    }
  else
    {
      GM_WARN (("%lu memory leaks detected.\n", user_alloc_cnt));
    }
  

  if (hash_entry && user_alloc_cnt)
    {
      for (i=0; i<=hash_max_index; i++)
	{
	  void *ptr = hash_entry[i].ptr;
	  if (ptr)
	    {
	      report (ptr, hash_entry[i].record);
	    }
	}
    }
  GM_RETURN_NOTHING ();
}

/****************************************************************
 * Internal interface for reportint memory leaks
 ****************************************************************/

/* chain of function descriptions built by GM_CALLED*() */

struct gm_function_description *gm_first_function_description;

/* Return a text description of the closest function entry either
   before or after _pc.  Returns NULL if none can be found. */

static void
report (void *ptr, struct gm_mdebug *record)
{
  unsigned long i;

  GM_CALLED ();
  
  GM_NOTE (("Block not freed: 0x%p allocated by %s()\n",
	    ptr, record->caller[0]));

  /*
   * report stack trace for the leaked allocation
   */

  /* Call trace according to gm_call_trace.h, which records the
     function name of each function that has a GM_CALLED*() macro at
     its start.  The plus side of this mechanism is that it works on
     all architectures.  The down side is that it may skip some calls
     in the call stack that are not instrumented. */

  _GM_NOTE (("According to the best effort of gm_call_trace.h,\n"));
  _GM_NOTE (("the allocation's call stack appears to be:\n"));
  for (i=0; i<GM_NUM_ELEM(record->caller); i++)
    {
      if (record->caller[i] == 0)
	{
	  break;
	}
      _GM_NOTE (("    %s\n", record->caller[i]));
    }
  _GM_NOTE (("    ....\n"));

  GM_RETURN_NOTHING ();
}

void
_gm_report_memory_leaks (void)
{
  GM_CALLED ();
  hash_report ();
  hash_reset ();
  if (internal_alloc_cnt)
    {
      GM_NOTE (("gm_malloc_debug leaked %ld internal allocations.\n",
		internal_alloc_cnt));
    }
  GM_RETURN_NOTHING ();
}

/****************************************************************
 * Function to determine if ptr was allocated via a GM interface
 ****************************************************************/

int
_gm_allocated (void *ptr)
{
  GM_CALLED ();
  GM_RETURN (hash_find (ptr) ? 1 : 0);
}

/****************************************************************
 * External interface
 ****************************************************************/

GM_ENTRY_POINT
gm_status_t
_gm_mdebug_record_ptr (void *ptr, char *free_func, char *return_addr)
{
  unsigned int i;
  gm_status_t status;
  struct gm_mdebug *record;
  GM_PARAMETER_MAY_BE_UNUSED (return_addr);
  
  GM_CALLED_WITH_ARGS (("%p, \"%s\", \"%s\"", ptr, free_func, return_addr));
  gm_assert (free_func);

  record = alloc_record ();
  if (!record)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_nothing;
    }
  
  record->free_function = free_func;

  /* copy the stack trace for this malloc */
  for (i=0; i<GM_NUM_ELEM (record->caller); i++)
    {
      record->caller[i] = GM_CALLER (i);
    }

  status = hash_insert (ptr, record);
  if (status != GM_SUCCESS)
    {
      goto abort_with_record;
    }
  user_alloc_cnt++;
  GM_RETURN_STATUS (GM_SUCCESS);

 abort_with_record:
  free_record (record);
 abort_with_nothing:
  GM_RETURN_STATUS (status);
}

/* Verify that an appropriate pointer is being freed. */

GM_ENTRY_POINT
void
_gm_mdebug_remove_ptr (void *client_ptr, char *free_func, char *caller)
{
  struct gm_mdebug *record;
  GM_PARAMETER_MAY_BE_UNUSED (caller);
  
  GM_CALLED_WITH_ARGS (("%p, \"%s\", \"%s\"", client_ptr, free_func, caller));

  /* Find any record of the allocation */

  record = hash_find (client_ptr);
  
  /* Report an error if the client pointer was not allocated */
  
  if (!record)
    {
      GM_WARN (("Free function %s called from %s tried to free "
		"unallocated pointer %p,\n",
		free_func, caller, client_ptr));
      GM_RETURN_NOTHING ();	/* just in case panic does not panic */
    }
  
  /* Report frees of allocated pointers with the wrong free function */

  if (record->free_function != free_func
      && gm_strcmp (record->free_function, free_func) != 0)
    {
      _GM_WARN (("%s() called from %s() tried to free %p,\n",
		  free_func, caller, client_ptr));
      GM_WARN (("    which was allocated by %s() and should have been "
		 "freed by %s().\n",
		 record->caller[0], record->free_function));
      GM_RETURN_NOTHING ();	/* just in case panic does no panic */
    }

  /* Remove or invalidate the record of the allocation */

  hash_remove (client_ptr);
  user_alloc_cnt--;
  GM_RETURN_NOTHING ();
}
