/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2002 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_call_trace.h"
#include "gm_internal.h"
#include "gm_internal_funcs.h"
#include "gm_io.h"
#include "gm_ptr_hash.h"

#if GM_KERNEL
#include "gm_pio.h"
#include "gm_undocumented_funcs.h"
#endif

#define GM_DEBUG_NODE_ID_TO_GLOBAL_ID 0

#if GM_DEBUG_NODE_ID_TO_GLOBAL_ID
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

#if !GM_KERNEL

/* Internal user-level node_id->global_id translation function.  Slow,
   but does not require the cache. */

static
gm_status_t
_gm_node_id_to_global_id (gm_port_t *port,
			  unsigned int node_id,
			  unsigned int *global_id)
{
  GM_CALLED_WITH_ARGS (("%p,0x%x,%p", port, node_id, global_id));
  {
    gm_status_t status;
    gm_u32_t in_out;
    
    /* Have the driver perform the translation. */

    in_out = node_id;
    status
      = _gm_user_ioctl (port,
			GM_NODE_ID_TO_GLOBAL_ID,
			&in_out,
			sizeof (in_out));
    if (status != GM_SUCCESS)
      {
	GM_RETURN_STATUS (status);
      }
    *global_id = in_out;
    GM_RETURN_STATUS (GM_SUCCESS);
  }
}

/* Make sure that all translations are cached.  This should be called
   whenever a translation is found to not be cached to ensure the
   cache is up-to-date before assuming the translation really failed. */

gm_status_t
_gm_update_global_id_caches (gm_port_t *port)
{
  gm_status_t status;
  unsigned int first_uncached_node_id;
  unsigned int max_node_id_in_use;

  status = gm_max_node_id_in_use (port, &max_node_id_in_use);
  if (status != GM_SUCCESS)
    {
      goto abort_with_nothing;
    }
  first_uncached_node_id = port->global_id_cache.first_uncached_node_id;
  while (first_uncached_node_id <= max_node_id_in_use)
    {
      unsigned int global_id;

      /* Find the global ID for the uncached node ID. */
      
      status
	= _gm_node_id_to_global_id (port, first_uncached_node_id, &global_id);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_nothing;
	}

      /* Insert the global_id->node_id translation */
	 
      gm_assert (port->global_id_cache.hash);
      status
	= gm_ptr_hash_insert (port->global_id_cache.hash,
			      (void *) (gm_size_t) global_id,
			      (void *) (gm_size_t) first_uncached_node_id);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_nothing;
	}

      /* Insert the node_id->global_id translation. */
      
      gm_assert (port->global_id_cache.array);
      port->global_id_cache.array[first_uncached_node_id] = global_id;
      port->global_id_cache.first_uncached_node_id = ++first_uncached_node_id;
    }
  return GM_SUCCESS;
  
 abort_with_nothing:
  return status;
}
#endif

/**
   gm_node_id_to_global_id() converts a GM connection ID into a GM
   node ID.

   A GM connection ID is a small integer uniquely identifying the remote host,
   but which is valid only only for this port.  Connections are numbered from
   0 to gm_max_node_id(port).

   A GM node ID is a global unique ID uniquely identifying a GM board.
   GM node ID's are sparsely packed.  That is, they are not sequentially
   numbered.
   
   @retval GM_SUCCESS Operation completed successfully.
 
   @param port \b (IN) the GM port associated with the specific GM ID
 
   @param node_id \b (IN) the connection ID to translate

   @param global_id \b (OUT) a pointer to storage for the GM node ID
   corresponding to node_id
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

gm_status_t
gm_node_id_to_global_id (gm_port_t *port,
			 unsigned int node_id,
			 unsigned int *global_id)
{
  GM_CALLED_WITH_ARGS (("%p,0x%x,%p", port, node_id, global_id));
#if GM_KERNEL
  {
    gm_instance_state_t *is;
    gm_port_state_t *ps;
    unsigned int cached_global_id;
    
    gm_assert (port);
    ps = port->kernel_port_state;
    gm_assert (ps);
    is = ps->instance;
    gm_assert (is);

    if (node_id >= is->max_node_id)
      {
	GM_RETURN_STATUS (GM_INVALID_PARAMETER);
      }
    cached_global_id = is->global_id_array[node_id];
    if (cached_global_id == 0)
      {
	GM_RETURN_STATUS (GM_INVALID_PARAMETER);
      }
    *global_id = cached_global_id;
    GM_RETURN_STATUS (GM_SUCCESS);
  }
#else  /* !GM_KERNEL */
  {
    gm_status_t status;
  
    /* Make sure the node ID translation is cached. */
  
    if (node_id >= port->global_id_cache.first_uncached_node_id)
      {
	status = _gm_update_global_id_caches (port);
	if (status != GM_SUCCESS)
	  {
	    GM_RETURN_STATUS (status);
	  }
      }
    if (node_id >= port->global_id_cache.first_uncached_node_id)
      {
	GM_RETURN_STATUS (GM_INVALID_PARAMETER);
      }

    /* Do a simple table lookup to translate the cached node ID to a
       global ID. */

    *global_id = port->global_id_cache.array[node_id];
    GM_RETURN_STATUS (GM_SUCCESS);
  }
#endif
}
