/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_node_id_to_host_name.c This file contains the GM API
   functions gm_node_id_to_host_name() and
   gm_node_id_to_host_name_ex().
*/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_debug_yp.h"
#include "gm_internal.h"
#include "gm_lanai_command.h"

#define GM_DEBUG_NODE_ID_TO_HOST_NAME 0

#if GM_DEBUG_NODE_ID_TO_HOST_NAME || GM_DEBUG_YP
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/**
   This function is deprecated.
   
   gm_node_id_to_host_name() returns a pointer to the host name of
   the host containing the network interface card with GM node
   id node_id. The name referenced by the returned pointer is only
   valid until the next GM API call.
 
   @retval host_name The host name of the host containing the GM node
                     id node_id.
 
   @param port \b (IN) The GM port associated with the network
                interface card.
 
   @param node_id \b (IN) The GM node id associated with the
                network interface card.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
char *
gm_node_id_to_host_name (gm_port_t *port, unsigned int node_id)
{
  static char name[GM_MAX_HOST_NAME_LEN+1];
  gm_status_t status;
  
  status
    = (gm_node_id_to_host_name_ex
       (port, 0, node_id, (char (*)[GM_MAX_HOST_NAME_LEN+1]) name));
  if (status != GM_SUCCESS)
    {
      return 0;
    }
  return name;
}

/**
   Store at *name the host name of the host containing the network
   interface card with GM node id node_id.  The buffer pointed to by
   name must have at least GM_MAX_HOST_NAME_LEN+1 bytes of storage.
 
   @retval host_name The host name of the host containing the GM node
                     id node_id.
 
   @param timeout_usecs \b (IN) The maximum length of time (in
   		microseconds) to query, or 0 to use the default (long)
   		timeout.
 
   @param port \b (IN) The GM port associated with the network
                interface card.
 
   @param node_id \b (IN) The GM node id associated with the
                network interface card.

   @param name \b (OUT) Where to store the pointer to the host name.

   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
gm_status_t
gm_node_id_to_host_name_ex (gm_port_t *port,
			    unsigned int timeout_usecs,
			    unsigned int node_id,
			    char (*name)[GM_MAX_HOST_NAME_LEN+1])
{
  gm_status_t status;

  GM_CALLED_WITH_ARGS (("%p,%d,%d,%p", port, timeout_usecs, node_id, name));

#if GM_KERNEL
  {
    gm_port_state_t *ps;

    gm_assert (port);
    ps = port->kernel_port_state;
    gm_assert (ps);
    
    /* Translate the name using the yellow pages.  We are looking for
       the "gm.hostname" value of node_id. */

    status = gm_lanai_command_yp (ps, timeout_usecs, "gm.hostname", "",
				  &node_id, name);
  }
#else
  {
    union
    {
      struct 
      {
	unsigned int timeout_usecs;
	unsigned int node_id;
      } in;
      char out[GM_MAX_HOST_NAME_LEN+1];
    } in_out;

    in_out.in.node_id = node_id;
    in_out.in.timeout_usecs = timeout_usecs;
    status
      = (_gm_user_ioctl
	 (port, GM_NODE_ID_TO_HOST_NAME, &in_out, sizeof (in_out)));
    gm_strncpy ((char *) name, in_out.out, sizeof (*name));
  }
#endif
  if (status == GM_SUCCESS)
    {
      _GM_PRINT (GM_DEBUG_YP, ("name = \"%s\"\n", name));
    }
  GM_RETURN_STATUS (status);
}
