/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/**
   @file gm_on_exit.c
   This file contains the GM API function gm_on_exit().
*/

#include "gm.h"
#include "gm_call_trace.h"
#include "gm_config.h"
#include "gm_internal_funcs.h"

#if HAVE_ON_EXIT && !GM_KERNEL && STDC_HEADERS
#include <stdlib.h>
int on_exit(void (*function)(int , void *), void *arg);
#elif HAVE_ATEXIT && !GM_KERNEL && STDC_HEADERS
#include <stdlib.h>
int atexit(void (*function)(void));
#endif

/****************************************************************
 * Typedefs
 ****************************************************************/

/** List element storing the details of a callback that should be called
    upon exit. */
typedef struct gm_on_exit_record 
{
  struct gm_on_exit_record *next;
  gm_on_exit_callback_t callback;
  void *arg;
} gm_on_exit_record_t;

/****************************************************************
 * Vars
 ****************************************************************/

static gm_on_exit_record_t *_gm_first_on_exit_record;
static int initialized;

/****************************************************************
 * Functions
 ****************************************************************/

/* This function causes all gm_on_exit() registered callbacks
   to be called in the reverse order they were registered. */

#if HAVE_ON_EXIT && !GM_KERNEL
static
void
on_exit_callback (int status, void *arg)
{
  _gm_perform_on_exit_callbacks (status == 0 ? GM_SUCCESS : GM_FAILURE);
}
#elif HAVE_ATEXIT && !GM_KERNEL
static
void
atexit_callback (void)
{
  _gm_perform_on_exit_callbacks (GM_FAILURE);
}
#endif /* HAVE_ATEXIT */

#if !GM_KERNEL || GM_OS_VXWORKS
static gm_inline
gm_status_t
gm_init_on_exit (void)
{
  gm_status_t status;
  if (initialized)
    {
      return GM_SUCCESS;
    }

  /* Initialize GM to ensure that GM is not finalized until after all the
     gm_on_exit () callbacks have been called */

  status = gm_init ();
  if (status != GM_SUCCESS)
    {
      return status;
    }
  
#if HAVE_ON_EXIT && !GM_KERNEL
  if (0 != on_exit (on_exit_callback, NULL))
    {
      return GM_FAILURE;
    }
#elif HAVE_ATEXIT && !GM_KERNEL
  if (0 != atexit (atexit_callback))
    {
      return GM_FAILURE;
    }
#endif /* HAVE_ATEXIT && !GM_KERNEL */

  initialized = 1;
  return GM_SUCCESS;
}

/**
   gm_on_exit() is like Linux on_exit().  This function registers
   a callback so that
     `CALLBACK(STATUS,ARG)' is called when the program exits.  Callbacks are
     called in the reverse of the order of registration.  This function is
     also somewhat similar to BSD `atexit()'.
 
   Call the callbacks in the
   reverse order registered inside gm_exit(), passing GM exit status
   and registered argument to the callback.
 
   @retval GM_SUCCESS Operation completed successfully.
   @retval GM_OUT_OF_MEMORY
 
   @param callback
 
   @param arg
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
gm_status_t
gm_on_exit (gm_on_exit_callback_t callback, void *arg)
{
  gm_on_exit_record_t *record;
  gm_status_t status;

  GM_CALLED ();
  status = gm_init_on_exit ();
  if (status != GM_SUCCESS)
    {
      goto abort_with_nothing;
    }
  record = (gm_on_exit_record_t *) gm_malloc (sizeof (gm_on_exit_record_t));
  if (!record)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_init;
    }
  record->next = _gm_first_on_exit_record;
  record->callback = callback;
  record->arg = arg;
  _gm_first_on_exit_record = record;
  GM_RETURN_STATUS (GM_SUCCESS);

 abort_with_init:
 abort_with_nothing:
  GM_RETURN_STATUS (status);
}

/****************
 * gm_on_exit() support
 *
 * This is here instead of in gm_on_exit.c to prevent drawing the
 * whole of gm_on_exit.o into applications that don't use gm_on_exit().
 ****************/

void
_gm_perform_on_exit_callbacks (gm_status_t status)
{
  gm_on_exit_record_t *next;

  if (!initialized)
    return;
  initialized = 0;
  
  while (_gm_first_on_exit_record)
    {
      (*_gm_first_on_exit_record->callback) (status,
					     _gm_first_on_exit_record->arg);
      next = _gm_first_on_exit_record->next;
      gm_free (_gm_first_on_exit_record);
      _gm_first_on_exit_record = next;
    }

  /* gm_init_on_exit() called gm_init() to prevent GM finalization until
     after all the gm_on_exit records have been freed.  This is necessary to
     prevent these buffers from being reported as leaked. */
  
  gm_finalize ();
}
#endif /* !GM_KERNEL || GM_OS_VXWORKS */
