/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_open.c
   The file containing the GM API function gm_open().
*/

#include "gm_call_trace.h"
#include "gm_compiler.h"
#include "gm_debug.h"
#include "gm_debug_open.h"
#include "gm_internal.h"
#include "gm_internal_funcs.h"
#include "gm_enable_ethernet.h"
#include "gm_enable_security.h"
#include "gm_enable_trace.h"
#include "gm_ptr_hash.h"

#if GM_DEBUG_OPEN
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif /* GM_DEBUG_OPEN */

#if GM_KERNEL
#include "gm_enable_security.h"
#endif

#if !GM_KERNEL
				/* open() */
#  ifdef HAVE_SYS_TYPES_H
#    include <sys/types.h>
#  endif

#  ifdef HAVE_SYS_ERRNO_H
#    include <sys/errno.h>
#    include <errno.h>
#  endif
				/* open() */
#  ifdef HAVE_SYS_STAT_H
#    include <sys/stat.h>
#  endif
#  ifdef HAVE_FCNTL_H
#    include <fcntl.h>
#  endif

				/* sprintf() */
#  ifdef STDC_HEADERS
#    include <errno.h>
#    include <stdio.h>
#    include <string.h>
#  endif


				/* close() */
#  ifdef HAVE_UNISTD_H
#    include <unistd.h>
#  endif
#  if GM_OS_SOLARIS
int close (int);
#  endif
#endif /* !GM_KERNEL */

/************************************************************************
 * Driver entry points
 ************************************************************************/

/************************************************************************
 * gm_open family of calls
 ************************************************************************/

#if !GM_KERNEL
#  ifdef _WIN32
#include <winsock.h>
#    define GM_DEV_TEMPLATE "\\\\.\\gm%x"
#    define GM_PRIV_DEV_TEMPLATE "\\\\.\\gmp%x"
WSADATA gm_wsadata;
int gm_wsadata_initialized = 0;
#  else
#    define GM_DEV_TEMPLATE "/dev/gm%d"
#    define GM_PRIV_DEV_TEMPLATE "/dev/gmp%d"
#    define GM_DEV_TEMPLATE_ALT "/dev/myricom/gm%d"
#    define GM_PRIV_DEV_TEMPLATE_ALT "/dev/myricom/gmp%d"
#  endif
#endif

#if HAVE_SYS_ERRNO_H && !GM_KERNEL
/* This logic should be in a new libgm module. */
#define GM_STATUS_FROM_ERRNO(status) do					\
{									\
  switch (errno)							\
    {									\
    case EBUSY:								\
      status = GM_BUSY;							\
      break;								\
    case ENODEV:							\
    case ENXIO:								\
      status = GM_NO_SUCH_DEVICE;					\
      break;								\
    case ENOMEM:							\
      status = GM_OUT_OF_MEMORY;					\
      break;								\
    case EPERM:								\
      status = GM_PERMISSION_DENIED;					\
      break;								\
    case EACCES:							\
      status = GM_ACCESS_DENIED;					\
      break;								\
    case ENOENT:							\
      status = GM_DEV_NOT_FOUND;					\
      break;								\
    default:								\
      status = GM_UNTRANSLATED_SYSTEM_ERROR;				\
    }									\
} while (0)
#endif

#if GM_OS_IRIX && !GM_KERNEL
#if GM_USER_maxstern
/* display the special ioctl args used by IRIX */
static void
display_setpn_args (gm_irix_set_port_num_arg_t *the_args, char *the_comment)
{
  GM_PRINT (GM_DEBUG_OPEN, ("set_port_num_arg struct %s:\n", the_comment));

  if (!the_args)
    {
      _GM_PRINT (GM_DEBUG_OPEN, ("    Args pointer is VOID\n") );
    }
  else
    {
      _GM_PRINT (GM_DEBUG_OPEN, ("    Args pointer is 0x%p\n", the_args));
      _GM_PRINT (GM_DEBUG_OPEN, ("    port_id   = %d\n", the_args->port_id));
      _GM_PRINT (GM_DEBUG_OPEN, ("    name_size = %d\n", the_args->name_size));
      _GM_PRINT (GM_DEBUG_OPEN, ("    name      = %s\n",
                              ((the_args->name) ? the_args->name : "<NULL>")));
    }
}
#endif
#endif /* GM_OS_IRIX */

/* Open a LANai port and return a pointer to the gm_internal_port_t structure.

   The gm_internal_port_t structure is assembled by allocating an
   aligned buffer to store the structure and then using mmap() to map
   the appropriate parts of the the /dev/gm* device into this
   structure. */

static gm_status_t
_gm_open (gm_port_t **port_p, unsigned unit, unsigned port_id,
	  const char *client_type, enum gm_api_version api_version,
	  int privileged)
{
  gm_status_t status = GM_SUCCESS;
  gm_port_t *port;
  unsigned int i;
  int supported = 0;
  
#if GM_KERNEL
  gm_arch_minor_t minor;
  gm_port_state_t *ps;
  gm_instance_state_t *is;
#else  /* !GM_KERNEL */
#  if defined WIN32
  HANDLE temp_fd;
#  endif
#if GM_OS_IRIX
  gm_irix_set_port_num_arg_t irix_pn_args;
  gm_status_t                irix_ioctl_status;
#endif /* GM_OS_IRIX */
  char *template, dev[40];
#endif /* !GM_KERNEL */
  GM_PARAMETER_MAY_BE_UNUSED (client_type);

  GM_CALLED_WITH_ARGS (("%p,%d,%d,%s,%d", port_p, unit, port_id, client_type,
			api_version));
  GM_PRINT(GM_DEBUG_OPEN,("Trying to open unit %d, port %d\n", unit, port_id));

  /************
   * Simple Checks
   ************/

  /* Use a switch statement to check the API version so that missing
     cases will be reported by picky compilers if any part of the enum
     is missing.  This also allows for simple support of discontiguous
     version numbering, such as when the major version number
     changes. */
  switch (api_version)
    {
    case (enum gm_api_version) GM_API_VERSION_1_0: supported = 1; break;
    case (enum gm_api_version) GM_API_VERSION_1_1: supported = 1; break;
    case (enum gm_api_version) GM_API_VERSION_1_2: supported = 1; break;
    case (enum gm_api_version) GM_API_VERSION_1_3: supported = 1; break;
    case (enum gm_api_version) GM_API_VERSION_1_4: supported = 1; break;
    case (enum gm_api_version) GM_API_VERSION_1_5: supported = 1; break;
    case (enum gm_api_version) GM_API_VERSION_1_6: supported = 1; break;
    case (enum gm_api_version) GM_API_VERSION_2_0: supported = 1; break;
    case (enum gm_api_version) GM_API_VERSION_2_0_6: supported = 1; break;
    default:
      GM_PRINT (GM_DEBUG_OPEN,
		("funky api_version is %x", api_version));
      break;
    }
    
  if (!supported)
    {
      GM_NOTE (("API version %x.%x not supported by %s.\n",
		api_version >> 16, api_version & 0xff, _gm_version));
      status = GM_INVALID_PARAMETER;
      goto abort_with_nothing;
    }
  
  /************
   * Initialization
   ************/

  status = gm_init ();
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("gm_init() failed in _gm_open\n"));
      goto abort_with_nothing;
    }
  
  /************
   * Allocate storage for the port
   ************/

  /* allocate storage for the user port */
  port = (gm_port_t *) gm_malloc (sizeof (*port));
  if (!port)
    {
      GM_NOTE (("Could not allocate storage for user port (in kernel).\n"));
      status = GM_OUT_OF_MEMORY;
      goto abort_with_init;
    }
  GM_PRINT (GM_DEBUG_OPEN, ("Allocated Port at %p for unit %d, port %d\n",
                            port, unit, port_id));
  gm_bzero (port, sizeof (*port));

  /************
   * Open the device (arrange for ioctl-based API functions to work)
   ************/

#if GM_KERNEL
  /* In the kernel, the ioctl-based API functions don't use IOCTLs, but
     rather access the kernel internals via the port->kernel_port_state
     field.  So, we just need to create initialize this field with
     a new gm_port_state and set its port number to cause it to
     become fully-initialized. */
  
  is = gm_instance_for_id (unit);
  if (!is)
    {
      if (!GM_OS_VXWORKS) 
	GM_NOTE (("_gm_open: no instance for unit = %d\n",unit));
      status = GM_INTERNAL_ERROR;
      goto abort_with_port;
    }

  /* allocate storage for the kernel port_state */

  GM_PRINT (GM_DEBUG_OPEN, ("Allocating minor node in the kernel.\n"));
  status = gm_minor_alloc (is, &minor);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not allocate storage for kernel port_state\n"));
      goto abort_with_port;
    }
  ps = port->kernel_port_state = gm_minor_get_port_state (minor);
  gm_assert (ps);
  ps->privileged = 1;
  ps->is_a_kernel_port = 1;
  
  /* open the port state */

  GM_PRINT (GM_DEBUG_OPEN, ("opening port state in the kernel.\n"));
  status = gm_port_state_open (ps, port_id);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not open port state in kernel.\n"));
      goto abort_with_minor;
    }
  GM_PRINT (GM_DEBUG_OPEN,
            ("[%d]: opened port state for minor %d, port %d in the kernel.\n",
              unit, minor, port_id));
#else /* !GM_KERNEL */

  /****************************************************************
   * User version
   ****************************************************************/

  /* In user land, we open a new device to get (indirect) access to a
     kernel gm_port_state that allows the api functions to work.

     On unix, we get a new device by opening a clone device and
     setting the port number.

     On NT, we open a pseudo-clone device and set the port number,
     causing the device to create the device we really wanted.  We
     then close the pseudo-clone device and open the device we wanted
     to open in the first place. */

  if (privileged)
    {
      GM_PRINT (GM_DEBUG_SECURITY, ("Using privileged device\n"));
      template = GM_PRIV_DEV_TEMPLATE;
    }
  else
    {
      GM_PRINT (GM_DEBUG_SECURITY, ("Using nonprivileged device\n"));
      template = GM_DEV_TEMPLATE;
    }
  sprintf (dev, template, unit);

  GM_PRINT (GM_DEBUG_OPEN, ("about to try to open unit %d port %d as %s\n",
	                    unit, port_id, dev));

  /****************
   * Open the clone device
   ****************/
  
#ifdef WIN32
#define GM_INVALID_FD INVALID_HANDLE_VALUE
  sprintf (dev, template, unit);
  GM_PRINT (GM_DEBUG_OPEN, ("Trying to open device \"%s.\"\n", dev));
  port->fd = CreateFile (dev, (GENERIC_READ | GENERIC_WRITE),
			 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_SYSTEM,
			 NULL);
#elif GM_OS_VXWORKS
  GM_NOTE (("VxWorks - no user open defined\n"));
  status = GM_PERMISSION_DENIED;
  goto abort_with_port;
#else  /* default open style */
  port->fd = open (dev, O_RDWR);
#endif /* default open style */

#ifdef STDC_HEADERS
  GM_PRINT (GM_DEBUG_OPEN, ("open() returned errno %d\n", errno));
#endif

  /****
   * Verify the open
   ****/

#ifndef GM_INVALID_FD
#define GM_INVALID_FD (-1)
#endif
  
  if (port->fd == GM_INVALID_FD)
    {
      GM_PRINT (GM_DEBUG_OPEN, ("Could not open device \"%s\"\n", dev));
#if HAVE_SYS_ERRNO_H
      GM_STATUS_FROM_ERRNO(status);
#elif GM_OS_WIN2K
      switch (GetLastError()) {
      case ERROR_FILE_NOT_FOUND:
        status = GM_NO_SUCH_DEVICE;
        break;
      default:
        status = GM_CLONE_BUSY;
        break;
      }
#else      
      status = GM_CLONE_BUSY;
#endif
      
#if GM_OS_LINUX
      /* try to open alternate device name */
      if (privileged)
	{
	  template = GM_PRIV_DEV_TEMPLATE_ALT;
	}
      else
	{
	  template = GM_DEV_TEMPLATE_ALT;
	}
      sprintf (dev, template, unit);
      
      GM_PRINT (GM_DEBUG_OPEN, 
		("about to try to open unit %d port %d as alternate %s\n",
		 unit, port_id, dev));
      
      port->fd = open (dev, O_RDWR);
      
      /* check if alternate device works */
      if (port->fd == GM_INVALID_FD)
	{
	  GM_PRINT (GM_DEBUG_OPEN, ("Could not open device \"%s\"\n", dev));
	  goto abort_with_port;
	}
      else
	{
	  status = GM_SUCCESS;
	}
#else
      goto abort_with_port;
#endif
    }

  /****************
   * ???
   ****************/

#if GM_OS_MACOSX || GM_OS_OSF1 || GM_OS_FREEBSD || GM_OS_HPUX
  {
    int minor_id=-1;
    
    GM_PRINT (GM_DEBUG_OPEN,("dev %s\n", dev));
    if (ioctl (port->fd, GM_GET_DEV, &minor_id) == -1)
      {
	GM_NOTE (("Could not get_dev ioctl rv=-1  minor = %d errno = %d\n",
		     minor_id, errno));
	perror("GET_DEV failed");
	status = GM_INTERNAL_ERROR;
	goto abort_with_open_device;
      }
    close(port->fd);
    GM_PRINT (GM_DEBUG_OPEN,("new minor_id %d\n", minor_id));
    sprintf (dev, "/dev/gm%d", minor_id);
#if 0
    switch (minor_id) {
    case 2:
      sprintf (dev, "/dev/gm%d", 1);
      break;
    case 3:
      sprintf (dev, "/dev/gmp%d", 1);
      break;

    case 4:
      sprintf (dev, "/dev/gm%d", 2);
      break;
    case 5:
      sprintf (dev, "/dev/gmp%d", 2);
      break;	
    case 6:
      sprintf (dev, "/dev/gm%d", 3);
      break;
    case 7:
      sprintf (dev, "/dev/gmp%d", 3);
      break;
    default:
      break;
    }
#endif
    port->fd = open(dev, O_RDWR);


    if (port->fd == -1)
      {
	GM_PRINT (GM_DEBUG_OPEN,
		  ("Round 2 Could not open device \"%s\" %d\n", dev, errno));
#if HAVE_SYS_ERRNO_H
	GM_STATUS_FROM_ERRNO(status);
#else      
	status = GM_BUSY;
#endif
	goto abort_with_port;
      } 
  }
#endif
  
  
  

  /****************
   * Set the port number
   ****************/

#if GM_OS_IRIX
/* special use of port-number-set to get new device name to open */

/* Do we need to do this check?  Oh, well, it can't hurt... ////// */
  if (port_id >= GM_NUM_PORTS)
    {
      GM_NOTE (("port_id %d larger than maximum %d\n",
                  port_id, GM_NUM_PORTS-1) );
      status = GM_NO_SUCH_DEVICE;
      goto abort_with_port;
    }

  irix_pn_args.port_id = port_id;
  irix_pn_args.name_size = GM_IRIX_OPEN_NAME_MAX_SZ;
  irix_pn_args.name = gm_malloc(GM_IRIX_OPEN_NAME_MAX_SZ+1);
  if (!irix_pn_args.name)
    {
       GM_PRINT (GM_DEBUG_OPEN,
		 ("gm_malloc failed to get irix_pn_args.name\n"));
       status = GM_BUSY;
       goto abort_with_port;
    }

  strcpy(irix_pn_args.name, "");

#if GM_USER_maxstern
  display_setpn_args(&irix_pn_args, "before _gm_user_ioctl");
#endif
  irix_ioctl_status = _gm_user_ioctl(port, GM_SET_PORT_NUM,
				     &irix_pn_args, sizeof(irix_pn_args));
  if (irix_ioctl_status != GM_SUCCESS)
    {
      if ((GM_DEBUG) || (GM_DEBUG_OPEN))
          gm_perror ("Special irix ioctl call failed with error",
                     irix_ioctl_status);
      status = GM_BUSY;
      goto abort_with_port;
    }
#if GM_USER_maxstern
  display_setpn_args(&irix_pn_args, "after _gm_user_ioctl");
#endif

  close(port->fd);
  port->fd = open(irix_pn_args.name, O_RDWR);
  if (port->fd == -1)
    {
      GM_PRINT (GM_DEBUG_OPEN,
		("Could not open device \"%s\"\n", irix_pn_args.name));
#if HAVE_SYS_ERRNO_H
      GM_STATUS_FROM_ERRNO(status);
#else      
      status = GM_BUSY;
#endif
      goto abort_with_port;
    }

#else  /* !GM_OS_IRIX */

  /* set the port number.  Must be done before most other IOCTLs or MMAPs. */

  GM_PRINT (GM_DEBUG_OPEN,  ("Setting port number to %d.\n", port_id));
  status = _gm_set_port_num (port, port_id);
  if (status != GM_SUCCESS)
    {
      /* GM_NOTE (("Failed to set the port num for the opened port.\n")); */
      goto abort_with_open_device;
    }
  GM_PRINT (GM_DEBUG_OPEN,  ("Did set the port number.\n"));
#endif /* !GM_OS_IRIX */


  /* Print a warning and exit if the installed driver may not be 
     compatible with this version of libgm. */
  
  {
    const char *kernel_build_id;

    kernel_build_id = _gm_get_kernel_build_id (port);
    if (!kernel_build_id)
      {
	GM_WARN (("The GM user library linked with this program may not\n"));
	_GM_WARN (("be compatible with the installed driver\n"));
	_GM_WARN (("Driver is too old to determine its build ID\n"));
	_GM_WARN (("User library build ID is \"%s\"\n", _gm_build_id));
	status = GM_INCOMPATIBLE_LIB_AND_DRIVER;
        goto abort_with_open_device;
      }
    else 
      {
        /* For maintainers, require that the library and driver are
           from the same build.  For everyone else, require only that
           the the GM versions match. */

        if ((GM_ENABLE_MAINTAINER_MODE && GM_USER_glenn
             && strncmp (kernel_build_id,_gm_build_id, strlen(_gm_build_id)))
            || strncmp (kernel_build_id,_gm_version, strlen(_gm_version)))
          {
            status = GM_INCOMPATIBLE_LIB_AND_DRIVER;
            goto abort_with_open_device;
          }
      }
  }

  /****************
   * Open the cloned device, if needed.
   ****************/

#  if defined WIN32
  /* Under NT, close the clone device and open the port device. */
  
  GM_PRINT (GM_DEBUG_OPEN,
	    ("Closing the clone device and opening the cloned device.\n"));
  temp_fd = port->fd;
  sprintf (dev, "\\\\.\\gm%x-%x", unit, port_id);
  GM_PRINT (GM_DEBUG_OPEN, ("Opening the cloned device \"%s.\"\n", dev));
  port->fd = CreateFile (dev, (GENERIC_READ | GENERIC_WRITE),
			 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_SYSTEM,
			 NULL);
  if (port->fd == INVALID_HANDLE_VALUE)
    {
      GM_NOTE (("Could not open port device \"%s\".\n", dev));
      _GM_NOTE (("    GetLastError() returned 0x%x.\n", GetLastError()));
      status = GM_BUSY;
      goto abort_with_open_device;
    }
  CloseHandle (temp_fd);
  GM_PRINT (GM_DEBUG_OPEN, ("Succeeded in opening the cloned device.\n"));
#  endif/* WIN32 */
#endif /* !GM_KERNEL */

  /****************
   * Device Mapping
   ****************/

  /****
   * Determine the page length
   ****/

  if (!GM_PAGE_LEN)
    {
#if GM_OS_VXWORKS
      GM_NOTE (("gm_open: Don't know the page length.\n"));
      _GM_NOTE (("gm_open: Setting page length = 4096\n"));
      GM_PAGE_LEN = 4096;
#else
      status = GM_INTERNAL_ERROR;
      goto abort_with_open_device;
#endif
    }

  /****
   * Alloc mappings storage
   ****/

  GM_PRINT (GM_DEBUG_OPEN, ("Allocating space for mappings struct.\n"));
  port->mappings = gm_malloc (sizeof (gm_mapping_specs_t));
  if (!port->mappings) {
    GM_PRINT (GM_DEBUG_OPEN, ("whoops!  gm_malloc failed\n"));
    status = GM_OUT_OF_MEMORY;
    goto abort_with_open_device;
  }

  /****
   * Get mapping specs
   ****/

  GM_PRINT (GM_DEBUG_OPEN, ("Getting mapping specs\n"));
  status = _gm_get_mapping_specs (port, port->mappings);
  if (status != GM_SUCCESS)
    {
      GM_PRINT (GM_DEBUG_OPEN,
		("_gm_get_mapping_specs returned bad status\n"));
      goto abort_with_mappings_mem;
    }

  /****
   * Map the real time clock
   ****/

  GM_PRINT (GM_DEBUG_OPEN, ("Mapping RTC\n"));
  status = _gm_mmap (port,
		     port->mappings->RTC.offset,
		     port->mappings->RTC.len,
		     GM_MAP_READ,
		     (void **) &port->first_sram_page);
  if (status != GM_SUCCESS)
    {
      GM_PRINT (GM_DEBUG_OPEN, ("gm_open: couldn't _gm_mmap the RTC page\n"));
      goto abort_with_mappings_mem;
    }
  GM_PRINT (GM_DEBUG_OPEN, ("RTC at %p.\n", port->first_sram_page));
  gm_assert (port->recv_queue_start != (void *)-1);
  if (!GM_PAGE_ALIGNED (port->first_sram_page))
    {
      status = GM_INTERNAL_ERROR;
      goto abort_with_RTC_mapping;
    }

#if GM_ENABLE_TRACE
  /* we need to mmap the special registers to access the "real" RTC */
  if (_gm_mmap (port,
		port->mappings->special_regs.offset,
		port->mappings->special_regs.len,
		GM_MAP_READ,
		(void**)&port->lanai_special_regs) != GM_SUCCESS)
    goto abort_with_RTC_mapping;
#endif

  /* Why don't we treat this port like any other? */
  if (port_id != GM_DAEMON_PORT_ID)
    {
      /****
       * Map the send queue
       ****/

      status = _gm_mmap (port,
			 port->mappings->send_queue.offset,
			 port->mappings->send_queue.len,
			 GM_MAP_READ | GM_MAP_WRITE,
			 (void **) &port->lanai);
      if (status != GM_SUCCESS)
	goto abort_with_real_RTC_mapping;
      gm_assert (port->lanai != 0 && port->lanai != (void *) -1);
      gm_bzero (port->lanai, port->mappings->send_queue.len);

      /****
       * Create and register the receive queue
       ****/
      
      /* alloc space for the receive queue, which must be aligned on a
	 sizeof(gm_recv_queue_slot_t) boundary.  This involves
	 alocating a bit extra and aligning after the allocation.  */
	
      port->recv_queue_allocation
	= gm_dma_calloc (port,
			 GM_NUM_RECV_QUEUE_SLOTS+1,
			 sizeof (gm_recv_queue_slot_t));
      if (!port->recv_queue_allocation)
	{
	  status = GM_OUT_OF_MEMORY;
	  goto abort_with_send_queue_mapping;
	}
      GM_PRINT (GM_DEBUG_OPEN,
		("recv queue allocation at %p\n",
		 port->recv_queue_allocation));
      port->recv_queue_start
	= (void *) GM_ROUNDUP (up, port->recv_queue_allocation,
			       sizeof (gm_recv_queue_slot_t));
      GM_PRINT (GM_DEBUG_OPEN,
		("recv queue starts at %p\n", port->recv_queue_start));
	
      /* tell the driver where the receive queue is */

      status = gm_register_recv_queue (port,
				       (gm_up_t) port->recv_queue_start);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_recv_queue_memory;
	}
    }

  /****************
   * Constant caching
   ****************/

  status = gm_max_node_id (port, &port->max_node_id);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not get max node ID.\n"));
      goto abort_with_registered_receive_queue;
    }

  status = gm_get_node_id (port, &port->this_node_id);
  if (status != GM_SUCCESS)
    {
      if (status == GM_NODE_ID_NOT_YET_SET)
	port->this_node_id = GM_NO_SUCH_NODE_ID;
      else
	{
	  GM_NOTE (("Could not get node id.\n"));
	  goto abort_with_registered_receive_queue;
	}
    }

  status = gm_get_unique_board_id (port, port->unique_board_id);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not get unique board id.\n"));
      goto abort_with_registered_receive_queue;
    }

  {
    gm_pid_t my_pid;

    my_pid = gm_getpid ();
    status = _gm_set_opener_pid (port,my_pid);
    if (status != GM_SUCCESS)
      {
        GM_NOTE (("Could not set the 'opener_pid'.\n"));
      }
  }

  /****************
   * node_id<->global_id cache initialization
   ****************/

  /* Skip node 0, which is GM_NO_SUCH_NODE_ID, so it is never cached. */

  port->global_id_cache.first_uncached_node_id = 1;

  /* Allocate the node_id->global_id array */
  
  port->global_id_cache.array
    = gm_calloc (port->max_node_id + 1,
		 sizeof (port->global_id_cache.array[0]));
  if (port->global_id_cache.array == 0)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_registered_receive_queue;
    }

  /* Allocate the global_id->node_id array */
  
  port->global_id_cache.hash = gm_create_ptr_hash (port->max_node_id + 1);
  if (port->global_id_cache.hash == 0)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_global_id_cache_array;
    }
  
  /************
   * Finish initializing port.
   ************/

  port->api_version = api_version;
  /* port->kernel_port_state set above */
  /* port->this_node_id set above */
  /* port->max_node_id set above */

  /*** Initializes pointers to LANai-resident queues ***/

  port->send_queue_start = &port->lanai->send_token_queue[0];
  port->send_queue_limit = (&port->lanai->send_token_queue
			    [GM_NUM_SEND_QUEUE_SLOTS]);

  port->first_free_send_queue_slot_token = &port->send_queue_slot_token[0];
  for (i = 0; i < GM_NUM_ELEM (port->send_queue_slot_token) - 1; i++)
    {
      port->send_queue_slot_token[i].next
	= &port->send_queue_slot_token[i+1];
    }
  port->last_free_send_queue_slot_token = &port->send_queue_slot_token[i];
  port->last_free_send_queue_slot_token->next = 0;
  
  port->recv_token_queue_start = &port->lanai->recv_token_queue[0];
  port->recv_token_queue_slot = port->recv_token_queue_start;
  port->recv_token_queue_limit = (&port->lanai->recv_token_queue
				  [GM_NUM_RECV_TOKEN_QUEUE_SLOTS]);
  
  /* Initialize pointers to host-memory-resident queues. */

  /* recv_queue_start set above */
  port->recv_queue_slot = port->recv_queue_start;
  port->recv_queue_limit = &port->recv_queue_start[GM_NUM_RECV_QUEUE_SLOTS];

  /* port->num_zones = 0 */
  /* port->dma_zone = {0} */
  /* alloced_dma_mem_len = 0 */
  port->unit = unit;
  port->id = port_id;
  /* gm_set_alarm.c gets 1 send queue slot for setting alarms and
     another for flushing */
  port->alarm_token_cnt = 2;
  
  /* max_used_id; */
  /* _recv_tokens[GM_NUM_RECV_TOKENS + 1] = {0} */
  /* lanai_globals = 0 */
  /* lanai_special_regs = 0; */
  /* fd set above */
  /* mappings set above */
  /* sleep_pending = 0 */
  /* trash = 0 */
  /* unique_board_id set above */
  /* first_set_alarm = 0 */
  /* first_unset_alarm = 0 */
  /* lanai_alarm_pending = 0 */
  /* lanai_alarm_flush_pending = 0 */
  /* ignore_next_lanai_alarm = 0 */

  /* commit */

  _gm_open_ports_insert (port);
  *port_p = port;

  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_global_id_cache_array:
  gm_free (port->global_id_cache.array);
abort_with_registered_receive_queue:
  /* freeing the memory automatically renders the receive queue as
     deregistered. */
abort_with_recv_queue_memory:
  /* only do this conditionally, since the mapping might have been skipped */
  if (port->recv_queue_allocation)
    {
      gm_dma_free (port, (void *) port->recv_queue_allocation);
      port->recv_queue_allocation = 0;
    }
abort_with_send_queue_mapping:
  /* only do this conditionally, since the mapping might have been skipped */
  if (port->send_queue_start)
    _gm_munmap (port, (void *)&port->send_queue_start,
		(unsigned long) (port->mappings->send_queue.len));
abort_with_real_RTC_mapping:
#if GM_ENABLE_TRACE
  if (port->lanai_special_regs)
    _gm_munmap (port, (void *)&port->lanai_special_regs,
		(unsigned long) port->mappings->special_regs.len);
#endif
abort_with_RTC_mapping:
  if (port->first_sram_page)
    _gm_munmap (port, (void *)port->first_sram_page,
		(unsigned long) (port->mappings->RTC.len));
abort_with_mappings_mem:
  gm_free (port->mappings);
abort_with_open_device:
#if GM_KERNEL
  gm_port_state_close (ps);
abort_with_minor:
  gm_minor_free (minor);
#else 
#  ifdef _WIN32
  if (port->fd != INVALID_HANDLE_VALUE)
    CloseHandle (port->fd);
#  else
  close (port->fd);
#  endif
#endif
abort_with_port:
  gm_free (port);
abort_with_init:
  gm_finalize ();
abort_with_nothing:
  GM_RETURN_STATUS (status);
}

/**
   gm_open() opens a GM port \b port_p for LANai interface
   \b unit, a pointer to the port's state at \b *port_p.  The pointer
   must be passed to all subsequent functions that operate on the
   opened port.  \b port_id is a null-terminated ASCII string that
   is used to identify the port client for debugging (and potentially
   other) purposes; pass in the name of your program.
   Note that unit and port numbers start at 0, and that ports 0 (internal
   use) and 1 (mapper) and 3 (ethernet emulation) are reserved, so clients
   should use port 2 and ports 4-7.
 
   gm_open() is to be called by clients other than the daemon and
   mapper.
 
   @retval GM_SUCCESS Operation completed successfully.
   @retval GM_INVALID_PARAMETER Invalid parameter passed.
   @retval GM_BUSY Could not open device.
   @retval GM_NO_SUCH_DEVICE port_id > GM_NUM_PORTS - 1.
   @retval GM_INCOMPATIBLE_LIB_AND_DRIVER The GM user library linked
            with this program may not be compatible with the installed
            driver.
   @retval GM_OUT_OF_MEMORY Could not allocate storage for user port.

 
   @param port_p \b (IN) Pointer to the handle of the GM port.
 
   @param unit \b (IN) The device for the Myrinet interface.
          = 0 if device is myri0.
 
   @param port_id \b (OUT) The id of the GM port that is opened.
          = 2, 4, 5, 6, or 7
            (Ports 0, 1, and 3 are for privileged use only.)
 
   @param client_type \b (IN) Unused.
 
   @param api_version \b (IN) GM_API_VERSION
          = GM_API_VERSION as defined in gm.h
 
   @see gm_close gm_init gm_abort gm_finalize gm_exit
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT gm_status_t
gm_open (gm_port_t **port_p, unsigned unit, unsigned port_id,
	  const char *client_type, enum gm_api_version api_version)
{
  gm_status_t status;

  GM_CALLED ();

  /* Try opening the device with privileges */
  
  status = _gm_open (port_p, unit, port_id, client_type, api_version, 1);

  /* If we could not open with privileges, try without privileges */
  
  if (status == GM_ACCESS_DENIED)
    {
      status = _gm_open (port_p, unit, port_id, client_type, api_version, 0);
    }
  
  GM_RETURN_STATUS (status);
}

#if !GM_KERNEL
#ifdef _WIN32
/* Include an entry point if we are building a WIN32 DLL. */
BOOL WINAPI
gm_entry_point(HINSTANCE hinstDLL,  /* handle to DLL module */
	       DWORD fdwReason,     /* reason for calling function */
	       LPVOID lpReserved )  /* reserved */
{
  switch (fdwReason)  
    { 
    case DLL_PROCESS_ATTACH:
    case DLL_PROCESS_DETACH:
      return TRUE;

    default:
      return FALSE;
    }
}
#endif

#if 0 && GM_USER_loic && !GM_KERNEL
/* to be able to use gm_pseudo_global_clone with gdb, referencing it here forces its inclusion in the binary */
void gm_pseudo_global_clone();
static void(*dummy_func)()=gm_pseudo_global_clone;
#endif

#endif
