/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2002 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

/* This file implements a simple growable hash table optimized to
   minimize cache misses.  Both the key and data are pointers, and the
   keys should be unique to optimize performance.  The table will work
   for non-unique keys, but with degraded performance. */

#include "gm.h"
#include "gm_bit_twiddling.h"
#include "gm_cache_line.h"
#include "gm_call_trace.h"
#include "gm_crc32.h"
#include "gm_stbar.h"
#include "gm_ptr_hash.h"

/* The max number of entries allowed in a hash table with bin_cnt bins. */
#define GM_PTR_HASH_MAX_CNT(bin_cnt) ((bin_cnt)/2)

#define GM_SELECT_NINE 1
#define GM_SELECT_DPHIL 2
#define GM_SELECT_CRC_TABLE 3
#define GM_SELECT_CRC321 4
#define GM_SELECT_SHIFTY 5
#define GM_SELECT_CRC32 6

/* Which hashing function should we use? */
#define GM_PTR_HASH_SELECTION GM_SELECT_CRC32

#define GM_PTR_HASH_FUNC_DESCRIPTION GM_SUB_N_STRINGIFY(GM_PTR_HASH_SELECTION)

typedef struct gm_ptr_hash_bin 
{
  void *key;
  void *data;
} gm_ptr_hash_bin_t;

struct gm_ptr_hash
{
  /* Put crc32 table at the beginning so it will be cache line aligned. */
#if GM_PTR_HASH_SELECTION == GM_SELECT_CRC_TABLE
  gm_u32_t crc32[GM_CACHE_LINE_LEN / sizeof (gm_u32_t)];
#endif
  gm_ptr_hash_bin_t *bin;
  gm_size_t bin_cnt;
  gm_size_t cnt;
#if GM_PTR_HASH_SELECTION == GM_SELECT_NINE
  unsigned char bin_cnt_shift;
#endif
};

#define GM_PTR_HASH_BINS_PER_CACHE_LINE \
(GM_CACHE_LINE_LEN / sizeof (gm_ptr_hash_bin_t))

/* Grow the hash table. */

static gm_status_t gm_ptr_hash_grow (struct gm_ptr_hash *hash);

/* Slow (bit at a time) CRC32 over a pointer */

static inline
gm_u32_t
slow_crc32 (void *key)
{
  gm_u32_t crc32;
  gm_u32_t in;
  int i;
  
  /* Compute the CRC-32 of the key */
  
  crc32 = (gm_u32_t) ((gm_size_t) key);
#if GM_SIZEOF_VOID_P == 8
  in = (gm_u32_t) ((gm_size_t) key >> 32);
  for (i=32; i; i--)
    {
      crc32 = ((crc32 << 1)
	       ^ (((crc32 >> 31) & 1) ? GM_CRC32_POLY : 0)
	       ^ ((in >> 31) & 1));
      in <<= 1;
    }
#endif
  for (i=32; i; i--)
    {
      crc32 = ((crc32 << 1)
	       ^ (((crc32 >> 31) & 1) ? GM_CRC32_POLY : 0));
    }
  return crc32;
}

/* Fast table driven CRC32 using a table the same size as a cache line. */

#if GM_PTR_HASH_SELECTION == GM_SELECT_CRC_TABLE
static inline
gm_u32_t
fast_crc32 (struct gm_ptr_hash *hash, void *key)
{
  gm_u32_t in;
  gm_u32_t crc32;
  gm_u32_t mask;
  gm_u32_t u32;

  gm_assert (GM_POWER_OF_TWO (GM_NUM_ELEM (hash->crc32)));
  mask = GM_NUM_ELEM (hash->crc32) - 1;
  
#if GM_SIZEOF_VOID_P == 8
  crc32 = (gm_u32_t) (((gm_size_t) key) >> 32);
  in = (gm_u32_t) ((gm_size_t) key);
  for (u32 = 0xffffffffUL;
       u32;
       u32 *= GM_NUM_ELEM (hash->crc32))
    {
      crc32
	= (/* shift crc32 left N bits */
	   crc32 * GM_NUM_ELEM (hash->crc32)
	   /* xor in top N bits of input */
	   ^ ((((in >> 16) * GM_NUM_ELEM (hash->crc32)) >> 16) & mask)
	   /* use top N bits of old crc32 to select table value */
	   ^ hash->crc32[(((crc32 >> 16) * GM_NUM_ELEM (hash->crc32)) >> 16)
			 & mask]);
      /* shift up input by N bits */
      in *= GM_NUM_ELEM (hash->crc32);
    }
#elif GM_SIZEOF_VOID_P == 4
  crc32 = (gm_u32_t) ((gm_size_t) key);
#else
#error
#endif
  for (u32 = 0xffffffffUL;
       u32;
       u32 *= GM_NUM_ELEM (hash->crc32))
    {
      crc32
	= (/* shift crc32 left N bits */
	   crc32 * GM_NUM_ELEM (hash->crc32) /* shift crc32 left */
	   /* use top N bits of old crc32 to select table value */
	   ^ hash->crc32[(((crc32 >> 16) * GM_NUM_ELEM (hash->crc32)) >> 16)
			 & mask]);
    }
  return crc32;
}
#endif

/* Create a pointer hash table. */

struct gm_ptr_hash *
gm_create_ptr_hash (gm_size_t min_cnt)
{
  struct gm_ptr_hash *hash;

  GM_CALLED ();
  
  /* Allocate the hash table */
  
  hash = gm_page_alloc ();
  if (!hash)
    {
      GM_RETURN_PTR (0);
    }
  gm_bzero (hash, sizeof (*hash));
  
  /* Allocate the hash table bins */

  do
    {
      if (gm_ptr_hash_grow (hash) != GM_SUCCESS)
	{
	  goto abort_with_hash;
	}
    }
  while (GM_PTR_HASH_MAX_CNT (hash->bin_cnt) < min_cnt);
  
#if GM_PTR_HASH_SELECTION == GM_SELECT_CRC_TABLE
  /* Initialize the crc32 table for the hash table. */
  
  {
    gm_size_t i;

    for (i=0; i<GM_NUM_ELEM(hash->crc32); i++)
      {
	hash->crc32[i] = slow_crc32 ((void *) i);
      }
  }
#endif

  GM_RETURN_PTR (hash);
  
 abort_with_hash:
  gm_free (hash);
  GM_RETURN_PTR (0);
}

/* Destroy the pointer hash table. */

void
gm_destroy_ptr_hash (struct gm_ptr_hash *hash)
{
  GM_CALLED_WITH_ARGS (("%p", hash));
  gm_free_pages (hash->bin, hash->bin_cnt * sizeof (hash->bin[0]));
  gm_page_free (hash);
  GM_RETURN_NOTHING ();
}

/* Return the bin number where a search should start for a key */

static inline
gm_size_t
gm_ptr_hash_hash (struct gm_ptr_hash *hash, void *key, gm_size_t bin_cnt)
{
  gm_size_t bin_num;
  
  GM_CALLED_WITH_ARGS (("%p,%lu", key, (unsigned long) bin_cnt));
  
#if GM_PTR_HASH_SELECTION == GM_SELECT_NINE
  {
    unsigned char shift;
    gm_size_t in;

    /* Multiply by nine */
    bin_num = (gm_size_t) key * 9;

    /* Determine the number of bits we will ultimately keep, so we can
       pack all the entropy into these bits. */
    
    shift = (hash->bin_cnt_shift
	     - GM_SLOW_LOG2_ROUNDUP (GM_PTR_HASH_BINS_PER_CACHE_LINE));
    
    /* Pack all the entropy down in the bottom bits.  We need to mask
       the first shift to ensure that it is logical, not arithmetic,
       else the loop may never terminate. */

    /* in = bin_num shifted right arithmetically by "shift" bits */
    in = ((bin_num >> shift) & (1 << (8 * sizeof (key) - shift)) - 1);
    while (in)
      {
 	bin_num ^= in;
	in >>= shift;
      }
  }
#elif GM_PTR_HASH_SELECTION == GM_SELECT_DPHIL
  {
    gm_size_t in;
    gm_u32_t hash0 = 0x12a3fe2d, hash1 = 0x37abe8f9;
    int i;

    in = (gm_size_t) key;
    for (i=0; i<sizeof(in); i++)
      {
	gm_u32_t hash;

	hash = hash1 + (hash0 ^ (in & 0xff * 7152373));
	in >>= 8;
	if (hash & 0x80000000)
	  {
	    hash -= 0x7fffffff;
	  }
	hash1 = hash0;
	hash0 = hash;
      }
    bin_num = (gm_size_t) hash0;
    if (GM_SIZEOF_VOID_P == 8)
      {
	bin_num |= (((gm_size_t) hash1) << 31) << 1;
      }
  }
#elif GM_PTR_HASH_SELECTION == GM_SELECT_CRC_TABLE
  bin_num = (gm_size_t) fast_crc32 (hash, key);
#elif GM_PTR_HASH_SELECTION == GM_SELECT_CRC321
  bin_num = (gm_size_t) slow_crc32 (key);
#elif GM_PTR_HASH_SELECTION == GM_SELECT_SHIFTY
  bin_num = (gm_size_t) key;
  bin_num ^= bin_num << 1;
  bin_num ^= bin_num << 2;
  bin_num ^= bin_num << 4;
  bin_num ^= bin_num << 8;
  bin_num ^= bin_num << 16;
  if (GM_SIZEOF_VOID_P == 8)
    {
      bin_num ^= (bin_num << 31) << 1;
    }
#elif GM_PTR_HASH_SELECTION == GM_SELECT_CRC32
  {
    gm_u32_t crc;

    crc = 0xffffffff;
    if (GM_SIZEOF_VOID_P == 8)
      {
	gm_crc32_u32 ((gm_u32_t) ((((gm_size_t) key) >> 31) >> 1), &crc);
      }
    gm_crc32_u32 ((gm_u32_t) ((gm_size_t) key), &crc);
    bin_num = (gm_size_t) ~crc;
  }
#else
#error
#endif
  
  gm_assert (GM_POWER_OF_TWO (bin_cnt));
  
  /* Hash to a cache line, not a bin. */
  bin_num *= GM_PTR_HASH_BINS_PER_CACHE_LINE;

  /* Wrap the hashed bin number, which is almost surely too large. */
  bin_num &= bin_cnt-1;

  GM_RETURN ((gm_size_t) bin_num);
}

gm_status_t
gm_ptr_hash_insert (struct gm_ptr_hash *hash, void *key, void *data)
{
  gm_ptr_hash_bin_t *hash__bin;
  gm_size_t bin_num;
  gm_size_t hash__bin_cnt;
  
  GM_CALLED_WITH_ARGS (("%p,%p,%p", hash, key, data));
  
  /* Grow the hash table if needed. */
  
  if (GM_PTR_HASH_MAX_CNT (hash->bin_cnt) <= hash->cnt)
    {
      gm_status_t status;

      status = gm_ptr_hash_grow (hash);
      if (status != GM_SUCCESS)
	{
	  GM_RETURN_STATUS (status);
	}
    }
  hash->cnt += 1;
  
  /* Scan for an empty bin. */
  
  bin_num = gm_ptr_hash_hash (hash, key, hash->bin_cnt);
  hash__bin = hash->bin;
  hash__bin_cnt = hash->bin_cnt;
  while (hash__bin[bin_num].key)
    {
      bin_num = (bin_num + 1) & (hash__bin_cnt - 1);
    }
  
  /* Store the data in the bin. */
  
  hash__bin[bin_num].data = data;
  GM_STBAR ();
  hash__bin[bin_num].key = key;
  GM_RETURN_STATUS (GM_SUCCESS);
}

void *
gm_ptr_hash_remove (struct gm_ptr_hash *hash, void *key)
{
  gm_size_t bin_num;
  gm_size_t hash__bin_cnt;
  gm_size_t hash__cnt;
  gm_ptr_hash_bin_t *hash__bin;
  void *data;
  
  GM_CALLED_WITH_ARGS (("%p,%p", hash, key));
  
  /* Find the bin holding the entry */

  bin_num = gm_ptr_hash_hash (hash, key, hash->bin_cnt);
  hash__bin = hash->bin;
  hash__bin_cnt = hash->bin_cnt;
  while (hash__bin[bin_num].key != key)
    {
      if (!hash__bin[bin_num].key)
	{
	  /* no match in cluster */
	  GM_RETURN_PTR (0);
	}
      bin_num = (bin_num + 1) & (hash__bin_cnt - 1);
    }

  /* Remove the entry */

  hash__cnt = hash->cnt;
  data = hash__bin[bin_num].data;
  hash__bin[bin_num].key = 0;
  hash->cnt = --hash__cnt;
  
  /* Remove and reinsert all subsequent entries in the cluster.  This
     is required since we introduced a bubble into the cluster, which
     might interfere with entries stored in higher bins that hash to
     lower bins (but could not be stored there due to collisions).  */
  
  bin_num = (bin_num + 1) & (hash__bin_cnt - 1);
  while (hash__bin[bin_num].key)
    {
      void *k, *d;
      
      k = hash__bin[bin_num].key;
      d = hash__bin[bin_num].data;
      hash__bin[bin_num].key = 0;
      hash->cnt = hash__cnt - 1;
      gm_ptr_hash_insert (hash, k, d);
      bin_num = (bin_num + 1) & (hash__bin_cnt - 1);
    }

  GM_RETURN_PTR (data);
}

/* Find an entry in the hash table. */

void *
gm_ptr_hash_find (struct gm_ptr_hash *hash, void *key)
{
  gm_ptr_hash_bin_t *hash__bin;
  gm_size_t bin_num;
  gm_size_t hash__bin_cnt;
  
  GM_CALLED_WITH_ARGS (("%p,%p", hash, key));
  
  /* Find the bin holding the entry */
  
  bin_num = gm_ptr_hash_hash (hash, key, hash->bin_cnt);
  hash__bin = hash->bin;
  hash__bin_cnt = hash->bin_cnt;
  while (hash__bin[bin_num].key != key)
    {
      if (!hash__bin[bin_num].key)
	{
	  /* no match in cluster */
	  GM_RETURN_PTR (0);
	}
      bin_num = (bin_num + 1) & (hash__bin_cnt - 1);
    }

  /* Return the data. */

  GM_RETURN_PTR (hash__bin[bin_num].data);
}

static gm_status_t
gm_ptr_hash_grow (struct gm_ptr_hash *hash)
{
  gm_ptr_hash_bin_t *new_bin;
  gm_ptr_hash_bin_t *old_bin;
  gm_size_t i;
  gm_size_t new_cnt;
  gm_size_t old_cnt;
  gm_size_t old_bin_cnt;
  gm_status_t status;

  GM_CALLED_WITH_ARGS (("%p", hash));

  old_bin = hash->bin;
  old_cnt = hash->bin_cnt;
  new_cnt = 2 * old_cnt;
  if (!old_bin)
    {
      /* Make the first allocation be 1 page. */
      gm_assert (GM_POWER_OF_TWO (sizeof (hash->bin[0])));
      new_cnt = GM_PAGE_LEN / sizeof (hash->bin[0]);
#if GM_PTR_HASH_SELECTION == GM_SELECT_NINE
      /* Determine log_2 (new_cnt) */
      for (hash->bin_cnt_shift = 1;
	   1UL << hash->bin_cnt_shift != new_cnt;
	   hash->bin_cnt_shift++)
	;
#endif
    }
  else if (old_cnt - 1 == 0xffffffff)
    {
      /* The hash table would grow too large for the hashing function */
      status = GM_INTERNAL_ERROR;
      goto abort_with_nothing;
    }
  
  /* Allocate a new set of page aligned bins for the table.  It is
     critical that the bins be at least cache line aligned. */
  
  new_bin = gm_alloc_pages (new_cnt * sizeof (hash->bin[0]));
  if (!new_bin)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_nothing;
    }

  /* Clear the new bins. */
  
  for (i=0; i<new_cnt; i++)
    {
      new_bin[i].key = 0;
    }

  /* Save some state we are about to clobber. */
  
  old_bin_cnt = hash->bin_cnt;
      
  /* Update state so we can use gm_ptr_hash_insert() to insert
     into the new bins. */
  
  hash->cnt = 0;
  hash->bin = new_bin;
  hash->bin_cnt = new_cnt;

  /* If there are old bins. */
  
  if (old_bin)
    {
      gm_size_t i;

      /* Copy the the entries to the new table. */

      for (i=0; i<old_cnt; i++)
	{
	  void *old_key;

	  old_key = old_bin[i].key;
	  if (old_key)
	    {
	      gm_ptr_hash_insert (hash, old_key, old_bin[i].data);
	    }
	}

      /* Free the replaced bins. */

      gm_free_pages (old_bin, old_bin_cnt * sizeof (old_bin[0]));
    }

  GM_RETURN_STATUS (GM_SUCCESS);

 abort_with_nothing:
  GM_RETURN_STATUS (status);
}

const char *
gm_ptr_hash_func_description (void)
{
  return GM_PTR_HASH_FUNC_DESCRIPTION;
}
