/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_put.c
   This file contains the GM API function
   gm_put().
*/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_enable_put.h"
#include "gm_internal.h"
#include "gm_send_queue.h"

#if GM_DEBUG_PUT
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/**
   gm_put() transfers the \b len bytes at
   \b source_buffer to \b target_port_id on \b target_node_id with
   priority \b priority and stores the data at the remote virtual
   memory address \b target_buffer.  Call callback(port,context,status)
   when the send completes or fails, with \b status indicating the
   status of the send.  The order of the transfer is preserved relative
   to messages of the same priority sent using gm _send() or
   gm_send_to_peer().
 
   @param p \b (IN) The GM port on the source/sender GM node from
               which the communication is being sent.
   @param source_buffer \b (IN) Address of the send buffer.
   @param target_buffer \b (OUT) Address of the receive buffer.
   @param len \b (IN) The length in bytes of the buffer to be sent.
   @param priority \b (IN) The priority of the data being sent.
   @param target_node_id \b (IN) The GM node to which the data is
                      being sent.
   @param target_port_id \b (IN) The GM port on the destination GM node
                      to which the message is being sent.
   @param callback \b (IN) The function called when the send is complete.
 
   @param context \b (IN) Pointer to an integer or to a structure
                           that is passed to the callback function.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
void
gm_put (gm_port_t * p,
	void *source_buffer,
	gm_remote_ptr_t target_buffer,
	gm_size_t len,
	enum gm_priority priority,
	unsigned int target_node_id,
	unsigned int target_port_id,
	gm_send_completion_callback_t callback,
	void *context)
{
  GM_CALLED_WITH_ARGS (("%p, %p, "GM_U64_TMPL", %lu, %d, %u, %u, %p, %p",
			p, source_buffer, GM_U64_ARG (target_buffer), len,
			priority, target_node_id, target_port_id, callback,
			context));
  
  gm_assert (p);
  gm_assert (source_buffer);
  gm_assert (target_buffer);

  if (GM_DEBUG_PUT)
    {
      GM_PRINT (1, ("gm_put() data:\n"));
      gm_hex_dump (source_buffer, len);
    }
  
#if 0
  if (!GM_ALIGNED (source_buffer, GM_DMA_GRANULARITY)
      || !GM_ALIGNED (target_buffer, GM_DMA_GRANULARITY)
      || !GM_ALIGNED (len, GM_DMA_GRANULARITY))
    {
      GM_WARN (("this function only works for 8-byte-aligned pointers"
		" and lengths\n"));
      _GM_WARN (("    NOT sending message\n\n"));
      GM_RETURN_NOTHING ();
    }
#endif

  GM_SIMULATE_SIMULATE ( );

  /* Pass send token to LANai.  Care must be taken to flush the message to
     memory before flushing the token, and the token should be flushed to
     memory immediately. */

  {
    struct gm_put_get_send_event volatile *se;
    struct gm_put_get_send_event batch_write;

    se = GM_SEND_QUEUE_SLOT (p, callback, context, put_get);
    
    if (GM_STRUCT_WRITE_COMBINING)
      {
	batch_write.remote_buffer = gm_hton_rp (target_buffer);
	batch_write.local_buffer = gm_hton_up ((gm_up_t) source_buffer);
	batch_write.length = gm_hton_u32 ((gm_u32_t) len);
	batch_write.remote_node_id = gm_hton_u16 ((gm_u16_t) target_node_id);
	batch_write.remote_subport_id
	  = gm_hton_u8 ((gm_u8_t) GM_SUBPORT (priority, target_port_id));
	batch_write.type = gm_hton_u8 ((gm_u8_t) GM_PUT_SEND_EVENT);
	GM_COPY_TO_IO_SPACE (*se, batch_write);
      }
    else
      {
	se->remote_buffer = gm_hton_rp (target_buffer);
	se->local_buffer = gm_hton_up ((gm_up_t) source_buffer);
	se->length = gm_hton_u32 ((gm_u32_t) len);
	se->remote_node_id = gm_hton_u16 ((gm_u16_t) target_node_id);
	se->remote_subport_id
	  = gm_hton_u8 ((gm_u8_t) GM_SUBPORT (priority, target_port_id));
	GM_STBAR ();
	se->type = gm_hton_u8 (GM_PUT_SEND_EVENT);
      }
    GM_FLUSH_SEND_EVENT (p, se);
  }
  GM_RETURN_NOTHING ();
}
