/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_register.c
   This file contains the GM API function gm_register_memory().
*/

#include "gm_call_trace.h"
#include "gm_compiler.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_io.h"
#include "gm_debug_mem_register.h"
#include "gm_lanai_command.h"

void
_gm_register_memory_mutex_enter ()
{
  /* NEED_MUTEX */
  return;
}

void
_gm_register_memory_mutex_exit ()
{
  /* NEED_MUTEX */
  return;
}

/**
   gm_register_memory_ex() registers \b length bytes of user virtual
   memory starting at \b ptr for DMA transfers, associating the memory
   with port virtual address \b pvma.  The memory is locked down (made
   nonpageable) and DMAs on the region of memory are enabled.  Memory
   may be registered multiple times.  Memory may be deregistered using
   matching calls to gm_deregister_memory().  Note that memory
   registration is an expensive operation relative to sending and
   receiving packets, so you should use persistent memory
   registrations wherever possible.  Also note that memory
   registration is not supported on Solaris 2.7 and earlier due to
   operating system limitations.

   Note that \b pvma must be used in all subsequent GM API calls to refer
   to the registered memory region.

   gm_register_memory(p,ptr,len) is equivalent to
   gm_register_memory_ex(p,ptr,len,ptr).
   
   @retval GM_SUCCESS Operation completed successfully.
   @retval GM_FAILURE
   @retval GM_PERMISSION_DENIED
   @retval GM_INVALID_PARAMETER
   @retval GM_OUT_OF_MEMORY
 
   @param p \b (IN) Handle to the GM port.
 
   @param ptr \b (IN) The address of the memory region to be
                      registered.
 
   @param length \b (IN)  The length in bytes of the memory region
                      to be registered.
 
   @param pvma \b (IN)  The port virtual memory address with which to
                      associate this region.
 
   @see gm_register_memory
   @see gm_deregister_memory
 
   @author Glenn Brown
 
   @version GM_API_VERSION_2_0_6
 
*/

GM_ENTRY_POINT
gm_status_t
gm_register_memory_ex (gm_port_t *p, void *_ptr, gm_size_t length, void *_pvma)
{
  gm_status_t status;
  gm_up_t ptr;
  gm_up_t pvma;

  GM_CALLED_WITH_ARGS (("%p,%p,"GM_U64_TMPL",%p",
			p, _ptr, GM_U64_ARG (length), _pvma));
    
  ptr = (gm_up_t) (gm_size_t) _ptr;
  pvma = (gm_up_t) (gm_size_t) _pvma;
  
#if !GM_KERNEL && !GM_CAN_REGISTER_MEMORY
  GM_PARAMETER_MAY_BE_UNUSED (p);
  GM_PARAMETER_MAY_BE_UNUSED (ptr);
  GM_PARAMETER_MAY_BE_UNUSED (length);
  GM_PARAMETER_MAY_BE_UNUSED (pvma);

  status = GM_NOT_SUPPORTED_ON_ARCH;
  
#elif GM_KERNEL
  {
    gm_instance_state_t *is;
    gm_page_hash_table_t *hash;
    gm_port_state_t *ps;
    gm_size_t offset;
    unsigned int port_id;
    
    gm_assert (p);
    ps = p->kernel_port_state;
    gm_assert (ps);
    is = ps->instance;
    gm_assert (is);
    hash = &is->page_hash;
    
    /* Round start of registration down to a page boundary. */

    length += GM_PAGE_OFFSET (pvma);
    ptr -= GM_PAGE_OFFSET (pvma);
    pvma -= GM_PAGE_OFFSET (pvma);

    /* Round end of registration up to a page boundary. */

    length = GM_PAGE_ROUNDUP (size, length);
    
    /* Check input parameters. */
    
    if (pvma+length < pvma
	|| ptr+length < ptr
	|| !GM_PAGE_ALIGNED (ptr)
	|| length == 0
	|| pvma == 0)
      {
	GM_RETURN_STATUS (GM_INVALID_PARAMETER);
      }

    port_id = ps->id;
    if (port_id == GM_DAEMON_PORT_ID)	/* port for information only */
      {
	GM_RETURN_STATUS (GM_PERMISSION_DENIED);
      }
    
    /* Register the memory one page at a time.  Extreme care must
       be taken to wrap all accesses to the user-supplied buffer
       in exception handlers in case the user has passed a bad
       buffer. */

    gm_arch_mutex_enter (&hash->sync);
    for (offset=0; offset<length; offset += GM_PAGE_LEN)
      {
	/* Register a page. */

	GM_PRINT (GM_DEBUG_MEM_REGISTER,
		  ("Adding mapping for unit %d - call "
		   "gm_add_mapping_to_page_table (\n"
		   "    %p, %p, %d, 0)\n", is->id,
		   is, pvma + offset, ps->id));

	status = gm_add_mapping_to_page_table (ps,
					       ptr + offset,
					       pvma + offset,
					       0);
	if (status != GM_SUCCESS)
	  {
	    status = GM_INVALID_PARAMETER;
	    GM_PRINT (GM_DEBUG_MEM_REGISTER,
		      ("gm_add_mapping_to_page_table failed"
		       " len=%ld (0x%lx) bytes?\n",
		       length, length));
	    goto gm_register_loop__abort;
	  }
      }

    gm_arch_mutex_exit (&hash->sync);
    GM_RETURN_STATUS (GM_SUCCESS);

  gm_register_loop__abort:
    /* loic: we need to pause, the Lanai may have loaded some
       entries into its cache */
    gm_pause_lanai (is);
    while (offset > 0)
      {
	offset -= GM_PAGE_LEN;
	gm_dereference_mapping (ps, pvma + offset);
      }
    gm_unpause_lanai (is);
    gm_lanai_command_deregister (is, port_id, pvma, length);
    gm_dereference_mapping_commit (is, ps);
    gm_arch_mutex_exit (&hash->sync);
    /* fall through */
  }
#else
  {
    gm_u32_t klen;
  
    if (sizeof (length) > 4)
      {
	gm_always_assert (length <= 0xFFFFFFFFUL);
      }
    klen = (gm_u32_t) length;

    _gm_register_memory_mutex_enter ();

    /* Register the memory via one or more ioctls. */

    GM_PRINT (GM_DEBUG_MEM_REGISTER,
	      ("Registering 0x%lx bytes at address %p.\n", length, ptr));

    if (GM_OS_OSF1 || GM_OS_LINUX || GM_OS_FREEBSD || GM_OS_IRIX || GM_OS_HPUX ||
	GM_OS_WIN2K || GM_OS_MACOSX || GM_OS_AIX)
      {
	struct gm_off_len_uvma_pvma s;

	/* On these OS's, we can register by passing in the UVMA to be
	   registers in an ioctl to the kernel. */

	s.offset = 0;
	s.len = length;
	s.uvma = ptr;
	s.pvma = pvma;
      
	status = _gm_user_ioctl (p, GM_REGISTER_MEMORY_BY_STRUCT, &s,
				 sizeof (s));
	if (status != GM_SUCCESS)
	  {
	    GM_PRINT (GM_DEBUG_MEM_REGISTER,
		      ("Could not register memory (see kernel log) error = %d.\n",
		       status));
	    goto abort_with_mutex;
	  }
      }
    else if (GM_OS_SOLARIS)
      {
	struct gm_len_pvma s;
      
	s.len = length;
	s.pvma = pvma;

	/* On this architecture, we can only register I/O buffers (such
	   as ioctl()/read()/write() buffers).  We cannot just pass a
	   UVMA to the kernel and have the driver pin the corresponding
	   memory.  To avoid altering the buffer to be registered, we
	   pass the registration parameters in a separate ioctl. */

	status = _gm_user_ioctl (p, GM_SET_REGISTER_MEMORY_LENGTH_AND_PVMA,
				 &s, sizeof (s));
	if (status != GM_SUCCESS)
	  {
	    GM_NOTE (("Could not set registration length and handle.\n"));
	    goto abort_with_mutex;
	  }
	
	/* register the memory by passing the buffer to be registered to
	   a special ioctl. */

	status = _gm_user_ioctl (p, GM_REGISTER_MEMORY, (void *) (gm_size_t) ptr,
				 length);
	if (status != GM_SUCCESS)
	  {
	    GM_PRINT (GM_DEBUG_MEM_REGISTER,
		      ("Could not register memory (see kernel log).\n"));
	    goto abort_with_mutex;
	  }
      }
    else
      {
	/* This should never happen. */
	status = GM_NOT_SUPPORTED_ON_ARCH;
      }
  
  abort_with_mutex:
    _gm_register_memory_mutex_exit ();
    /* fall through */
  }
#endif
  
  GM_RETURN_STATUS (status);
}

#if GM_DEBUG_MEM_REGISTER
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/**
   gm_register_memory() registers \b length bytes of user virtual
   memory starting at \b ptr for DMA transfers.  The memory is locked
   down (made nonpageable) and DMAs on the region of memory are
   enabled.  Memory may be registered multiple times.  Memory may be
   deregistered using matching calls to gm_deregister_memory().  Note
   that memory registration is an expensive operation relative to sending
   and receiving packets, so you should use persistent memory registrations
   wherever possible.  Also note that memory registration is not supported
   on Solaris 2.7 and earlier due to operating system limitations.
 
   @retval GM_SUCCESS Operation completed successfully.
   @retval GM_FAILURE
   @retval GM_PERMISSION_DENIED
   @retval GM_INVALID_PARAMETER
   @retval GM_OUT_OF_MEMORY
 
   @param p \b (IN) Handle to the GM port.
 
   @param ptr \b (OUT) The address of the memory location to be
                      registered.
 
   @param length \b (IN)  The length in bytes of the memory location
                      to be registered.
 
   @see gm_deregister_memory
   @see gm_register_memory_ex
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT gm_status_t
gm_register_memory (gm_port_t *p, void *ptr, gm_size_t length)
{
  gm_status_t status;
  
  GM_CALLED_WITH_ARGS (("%p,%p,"GM_U64_TMPL, p, ptr, GM_U64_ARG (length)));
  status = gm_register_memory_ex (p, ptr, length, ptr);
  GM_RETURN_STATUS (status);
}



/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
