/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_call_trace.h"
#include "gm_debug_reg_recv_queue.h"
#include "gm_internal.h"
#include "gm_io.h"
#include "gm_page_hash.h"
#if GM_KERNEL
#include "gm_pio.h"
#endif
#include "gm_register_recv_queue.h"

#if GM_DEBUG_REGISTER_RECV_QUEUE
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

#if GM_KERNEL
void
gm_deregister_recv_queue (gm_port_state_t *ps)
{
  gm_dp_t bogus;
  gm_instance_state_t *is;
  gm_up_t user_vma;
  unsigned int port_id;
  unsigned int slot;

  GM_CALLED ();
  
  GM_PRINT (GM_DEBUG_REGISTER_RECV_QUEUE,
	    ("Deregistering receive queue for port %d\n", ps->id));
  
  is = ps->instance;
  gm_assert (is);
  port_id = ps->id;

  /* Catch the case where the recv queue is not registered. */
  
  if (!ps->recv_queue_page_hash)
    {
      GM_RETURN_NOTHING ();
    }
  
  /* Make all future receive queue DMAs go into the page reserved for
     bogus RDMAs. */
  
  bogus = gm_read_lanai_global_dp (is, page_hash.bogus_rdma_ptr);
  for (slot=0; slot<GM_NUM_RECV_QUEUE_SLOTS; slot++)
    {
      gm_write_lanai_global_dp
	(is, port[port_id].recv_queue_slot_dma_addr[slot], bogus);
    }

  /* remove the recv queue pages for this port from the
     recv_queue_page_hash */
  
  user_vma
    = gm_read_lanai_global_up (is, port[port_id].recv_queue_slot_host_addr[0]);
  
  GM_PRINT (GM_DEBUG_REGISTER_RECV_QUEUE,
	    ("Receive queue UVA was 0x%p\n", user_vma));
  
  {
    gm_up_t recv_queue_page;

    for (recv_queue_page = user_vma - GM_PAGE_OFFSET (user_vma);
	 recv_queue_page < user_vma + (GM_NUM_RECV_QUEUE_SLOTS
				       * sizeof (gm_recv_queue_slot_t));
	 recv_queue_page += GM_PAGE_LEN)
      {
	gm_hash_remove (ps->recv_queue_page_hash, &recv_queue_page);
      }
  }

  /* Free the recv queue registration monitoring resources. */

  gm_destroy_hash (ps->recv_queue_page_hash);

  /* Mark the recv queue as not registered. */
  
  ps->recv_queue_page_hash = 0;

  GM_RETURN_NOTHING ();
}

void
gm_catch_recv_queue_deregistration (gm_port_state_t *ps, gm_up_t dereg_addr)
{
  GM_CALLED ();
  
  if (ps->recv_queue_page_hash
      && gm_hash_find (ps->recv_queue_page_hash, &dereg_addr))
    {
      gm_deregister_recv_queue (ps);
    }
  
  GM_RETURN_NOTHING ();
}

#endif /* GM_KERNEL */

/* Internal funtion to allow the user gm_open() code to tell the
   kernel the location of the receive queue in user virtual memory.
   The user library will have allocated this space with
   gm_dma_malloc(), so each of the pages will be in the GM page table,
   allowing the kernel to translate the user virtual memory address to
   a DMA address safely, simply, and in a device-independent
   fashion.

   The user need never deregisters the receive queue, since the driver
   must automatically detect when the queue is freed or deregistered,
   or if the port is closed. */
   
gm_status_t
gm_register_recv_queue (gm_port_t *port, gm_up_t user_vma)
{
  gm_status_t status = GM_SUCCESS;
  
  GM_CALLED_WITH_ARGS (("%p, %lx", port, (long) user_vma));

#if GM_KERNEL
  {
    gm_instance_state_t *is;
    gm_port_state_t *ps;
    gm_up_t recv_queue_page;
    gm_up_t up;
    unsigned int port_id;
    unsigned int slot;

    gm_assert (GM_POWER_OF_TWO (sizeof (gm_recv_queue_slot_t)));
    if (!GM_ALIGNED (user_vma, sizeof (gm_recv_queue_slot_t))
	|| !GM_ALIGNED (user_vma, GM_RDMA_GRANULARITY))
      {
	GM_WARN(("gm_register_recv_queue: address 0x%lx not aligned\n",
		(unsigned long)user_vma));
	GM_RETURN_STATUS (GM_INVALID_PARAMETER);
      }
    
    ps = port->kernel_port_state;
    gm_assert (ps);
    is = ps->instance;
    gm_assert (is);
    port_id = ps->id;

    /* The receive queue is already registered. */
    
    if (ps->recv_queue_page_hash)
      {
	GM_RETURN_STATUS (GM_BUSY);
      }

    /* Create a hash table for recording which pages correspond to the
       port's receive queue so we can later detect if these pages are
       deregistered and invalidate the receive queue in that case. */
    
    ps->recv_queue_page_hash = gm_create_hash (_gm_hash_compare_ups,
					       _gm_hash_hash_up,
					       sizeof (gm_up_t), 0, 0, 0);
    if (!ps->recv_queue_page_hash)
      {
	GM_NOTE (("Could not create recv_queue_page_hash\n"));
	status = GM_OUT_OF_MEMORY;
	goto abort_with_nothing;
      }
    
    GM_PRINT (GM_DEBUG_REGISTER_RECV_QUEUE,
	      ("Registering receive queue at UVA 0x%p for port %d\n",
	       user_vma, ps->id));

    /* Record which pages are in the receive queue. */

    for (recv_queue_page = user_vma - GM_PAGE_OFFSET (user_vma);
	 recv_queue_page < user_vma + (GM_NUM_RECV_QUEUE_SLOTS
				       * sizeof (gm_recv_queue_slot_t));
	 recv_queue_page += GM_PAGE_LEN)
      {
	gm_dp_t dp;
	
	dp = gm_dma_addr_for_mapping (ps->instance, recv_queue_page, ps->id);
	GM_PRINT (GM_DEBUG_REGISTER_RECV_QUEUE >= 2,
		  ("recv queue page at UVMA "GM_U64_TMPL" and"
		   " DMAA "GM_U64_TMPL"\n",
		   GM_U64_ARG (recv_queue_page), GM_U64_ARG (dp)));
	status = gm_hash_insert (ps->recv_queue_page_hash, &recv_queue_page,
				 ps);
	if (status != GM_SUCCESS)
	  {
	    GM_PRINT (GM_DEBUG_REGISTER_RECV_QUEUE,
		      ("Failed to insert recv queue page in hash table.\n"));
	    goto abort_with_recorded_pages;
	  }
      }
    
    /* register each slot */

    GM_PRINT (GM_DEBUG_REGISTER_RECV_QUEUE,
	      ("Computing DMA addrs for receive queue slots.\n"));
    for (slot=0; slot<GM_NUM_RECV_QUEUE_SLOTS; slot++)
      {
	gm_dp_t dp;
	
	up = user_vma + slot * sizeof (gm_recv_queue_slot_t);
	dp = gm_dma_addr_for_mapping (is, up, port_id);
	if (!dp)
	  {
	    GM_WARN(("gm_dma_addr_for_mapping failed slot=%d\n", slot));
	    status = GM_INVALID_PARAMETER;
	    goto abort_with_slot_table_writes;
	  }
      	gm_write_lanai_global_up
	  (is, port[port_id].recv_queue_slot_host_addr[slot], up);
	gm_write_lanai_global_dp
	  (is, port[port_id].recv_queue_slot_dma_addr[slot], dp);
      }

    GM_RETURN_STATUS (GM_SUCCESS);

  abort_with_slot_table_writes:
    {
      gm_dp_t bogus;
      
      bogus = gm_read_lanai_global_dp (is, page_hash.bogus_rdma_ptr);
      for (slot=0; slot<GM_NUM_RECV_QUEUE_SLOTS; slot++)
	{
	  gm_write_lanai_global_dp
	    (is, port[port_id].recv_queue_slot_dma_addr[slot], bogus);
	}
    }
  abort_with_recorded_pages:
    for (recv_queue_page = user_vma - GM_PAGE_OFFSET (user_vma);
	 recv_queue_page < user_vma + (GM_NUM_RECV_QUEUE_SLOTS
				       * sizeof (gm_recv_queue_slot_t));
	 recv_queue_page += GM_PAGE_LEN)
      {
	gm_hash_remove (ps->recv_queue_page_hash, &recv_queue_page);
      }
  abort_with_nothing:
    gm_destroy_hash (ps->recv_queue_page_hash);
    ps->recv_queue_page_hash = 0;
  }
#else  /* !GM_KERNEL */
  {
    status = _gm_user_ioctl (port, GM_REGISTER_RECV_QUEUE,
			     &user_vma, sizeof (user_vma));
  }
#endif /* !GM_KERNEL */

  GM_RETURN_STATUS (status);
}
