/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1999 by Myricom, Inc.                                   *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_set_alarm.c
   This file includes source for the user-level alarm API calls
   gm_initialize_alarm(), gm_set_alarm(), and gm_cancel_alarm().
 
   The alarm API allows the GM client to schedule a callback function
   to be called after a delay, specified in microseconds.  An unbounded
   number of alarms may be set, although alarm overhead increases linearly
   in the number of set alarms, and the client must provide storage for
   each set alarm.
 
   These source for all the functions is included here, because if the
   user needs one, the user needs them all.
 
   This code is a bit tricky, since the LANai provides only one alarm
   per port.  This alarm is used to time the first pending alarm, but
   if any new alarm is set, then the LANai alarm must be flushed to
   find out what time it is and potentially to allow the alarm to be
   rescheduled for an earlier time.  Therefore, unless no other alarm
   is set, alarms are set by adding them to an "unset alarm" queue,
   flushing any pending alarm, and then adding the alarms to the
   sorted alarm queue only once the pending LANai alarm has been
   flushed or triggered.
 
*/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_cmp64.h"
#include "gm_internal.h"
#include "gm_send_queue.h"
#include "gm_set_alarm.h"

#define GM_DEBUG_ALARMS 0

#if GM_DEBUG_ALARMS
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/****************************************************************
 * LANai alarm layering
 ****************************************************************
 There is only one lanai-side alarm per port, on which we layer all GM
 alarms.  This is done by setting the LANai alarm to go off at the
 same time as the first scheduled GM alarm.  If a GM alarm arrives is
 scheduled that goes of before the first LANai alarm, then we flush
 the lanai alarm and reset it.

 Note that we are careful only to call alarm callbacks after receiving
 GM_ALARM_EVENTs, as documented in the GM API. */



static
void
alarm_event_sent (struct gm_port *p, void *context, gm_status_t status)
{
  GM_CALLED_WITH_ARGS (("%p,%p,%d", p, context, status));
  
  GM_PARAMETER_MAY_BE_UNUSED (context);
  GM_PARAMETER_MAY_BE_UNUSED (status);

  gm_assert (status == GM_SUCCESS);
  p->alarm_token_cnt++;
  _reschedule_lanai_alarm_if_needed (p);

  GM_RETURN_NOTHING ();
}

/****************
 * Set a lanai alarm, knowing that no other lanai alarm is set.
 ****************/

static void
_set_lanai_alarm (gm_port_t *port, gm_u64_t deadline)
{
  struct gm_set_alarm_send_event volatile *se;
  struct gm_set_alarm_send_event batch_write;

  GM_CALLED_WITH_ARGS (("%p,"GM_U64_TMPL, port, GM_U64_ARG (deadline)));
  
  gm_assert (port->lanai_alarm_state == LANAI_ALARM_IDLE);

  /* Schedule the one and only alarm */  

  se = GM_SEND_QUEUE_SLOT (port, alarm_event_sent, 0, set_alarm);
  if (GM_STRUCT_WRITE_COMBINING)
    {
      batch_write.usecs = gm_hton_u64 (deadline);
      batch_write.type = gm_hton_u8 (GM_SET_ALARM_EVENT);
      GM_COPY_TO_IO_SPACE (*se, batch_write);
    }
  else
    {
      se->usecs = gm_hton_u64 (deadline);
      GM_STBAR ();
      gm_assert (!gm_ntoh_u8 (se->type));
      se->type = gm_hton_u8 (GM_SET_ALARM_EVENT);
    }
  GM_FLUSH_SEND_EVENT (port, se);

  /* update state and return */

  port->lanai_alarm_time = deadline;
  port->lanai_alarm_state = LANAI_ALARM_SET;
  port->alarm_token_cnt--;

  GM_RETURN_NOTHING ();
}

/************
 * Schedule a LANai GM_ALARM_EVENT.
 ************/

void
_reschedule_lanai_alarm_if_needed (gm_port_t *port)
{
  GM_CALLED_WITH_ARGS (("%p", port));
  
  /* verify we have something to reschedule */

  if (!port->first_alarm)
    GM_RETURN_NOTHING ();
  
  /* schedule the alarm if the alarm is idle.  Otherwise, flush the
     alarm and reschedule once the alarm is idle, if needed. */

  if (port->alarm_token_cnt)
    {
      switch (port->lanai_alarm_state)
	{
	case LANAI_ALARM_IDLE:
	  _set_lanai_alarm (port, port->first_alarm->deadline);
	  break;

	case LANAI_ALARM_SET:

	  /* if the new first alarm is earlier than the currently
	     scheduled lanai alarm, flush the current lanai alarm so we
	     can set an earlier lanai alarm. */
	 
	  if (gm_cmp64 (port->first_alarm->deadline, port->lanai_alarm_time)
	      < 0)
	    {
	      gm_flush_alarm (port);
	    }
	  
	  break;

	default:
	  gm_abort ();
	  break;
      
	case LANAI_ALARM_FLUSHING:
	  /* the alarm will be reschedule after the flush, which has
	     already been requested. */
	  break;
	}
    }
  GM_RETURN_NOTHING ();
}

/************
 * Handle an alarm event from the LANai.  This is called by
 * gm_unknown().
 ************/

gm_status_t
_gm_handle_alarm (gm_port_t *port)
{
  gm_alarm_t **ap, *a, *port__first_alarm, *next;
  gm_u64_t rtc;

  GM_CALLED_WITH_ARGS (("%p", port));

  port__first_alarm = port->first_alarm;
  rtc = gm_ticks (port);

  /* If no alarm has expired, there is nothing to do. */

  if (port__first_alarm && gm_cmp64 (rtc, port__first_alarm->deadline) >=0)
    {
      /* remove all expired alarms from the alarm list.  This must be
	 done before calling any alarm handler, because the alarm
	 handlers may modify this list. */
  
      for (ap = &port->first_alarm; *ap; ap = &(*ap)->next)
	{
	  if (gm_cmp64 (rtc, (*ap)->deadline) < 0)
	    break;
	}
      gm_assert (ap != &port->first_alarm);
      port->first_alarm = *ap;
      *ap = 0;

      /* perform the callbacks for all alarms that have expired */

      for (a = port__first_alarm; a; a = next)
	{
	  next = a->next;
	  a->state = GM_ALARM_FREE;
	  a->next = 0;
	  a->callback (a->context);
	}
    }
  
  /* schedule next alarm check */

  gm_assert (port->lanai_alarm_state == LANAI_ALARM_SET
	     || port->lanai_alarm_state == LANAI_ALARM_FLUSHING);
  port->lanai_alarm_state = LANAI_ALARM_IDLE;
  _reschedule_lanai_alarm_if_needed (port);

  GM_RETURN (GM_SUCCESS);
}

/************
 * Handle a flush event from the LANai.  This is called by
 * gm_unknown().
 ************/

gm_status_t
_gm_handle_flushed_alarm (gm_port_t *port)
{
  /* do not perform callbacks here, because the GM API docs state that
     callbacks are performed only after GM_ALARM_EVENTs. */

  GM_CALLED ();
  
  gm_assert (port->lanai_alarm_state == LANAI_ALARM_FLUSHING);
  port->alarm_token_cnt++;
  port->lanai_alarm_state = LANAI_ALARM_IDLE;
  _reschedule_lanai_alarm_if_needed (port);

  GM_RETURN (GM_SUCCESS);
}

/****************************************************************
 * API entry points
 ****************************************************************/

/************
 * Cancel an alarm.
 ************
 Alarms are cancelled by simply removing them from the alarm list so
 that when the next GM_ALARM_EVENT arrives from the lanai
 _gm_handle_alarm() will not find the alarm. */

/**
   gm_cancel_alarm() cancels a scheduled alarm, or does nothing if an
   alarm is not scheduled.  Alarms are cancelled by simply
   removing them from the alarm list so that when the next GM_ALARM_EVENT
   arrives from the LANai _gm_handle_alarm() will not find the alarm.
 
   @param my_alarm \b (IN) Alarm to be cancelled.
 
   @see gm_initialize_alarm gm_set_alarm gm_flush_alarm
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
void
gm_cancel_alarm (gm_alarm_t *my_alarm)
{
  gm_alarm_t **a;

  GM_CALLED_WITH_ARGS (("%p", my_alarm));

  if (my_alarm->state == GM_ALARM_FREE)
    goto done;
  gm_assert (my_alarm->port);
  for (a = &my_alarm->port->first_alarm; *a; a = &(*a)->next)
    {
      if (*a == my_alarm)
	{
	  *a = my_alarm->next;
	  my_alarm->state = GM_ALARM_FREE;
	  goto done;
	}
    }
  GM_WARN (("Inconsistent alarm state detected.\n"));

 done:
  GM_RETURN_NOTHING ();
}

/************
 * Initialize user-allocated storage for an alarm.
 ************/

/**
   gm_initialize_alarm() initializes a client-allocated \b gm_alarm_t
   structure for use with
   gm_set_alarm().  This function should be called after the structure is
   allocated but before a pointer to it is passed to gm_set_alarm() or
   gm_cancel_alarm().
 
   @param my_alarm \b (IN) Alarm to be initialized.
 
   @see gm_cancel_alarm gm_set_alarm gm_flush_alarm
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
void
gm_initialize_alarm (gm_alarm_t *my_alarm)
{
  GM_CALLED ();
  my_alarm->state = GM_ALARM_FREE;
  GM_RETURN_NOTHING ();
}

/************
 * Set an alarm, which may already be pending.  If it is pending, it
 * is rescheduled.  The user supplies MY_ALARM, a pointer to storage for
 * the alarm state, which has been initialized with gm_init_alarm_my_alarm().
 * When the alarm occurs, CALLBACK (CONTEXT) is() called.
 ************/

/**
   gm_set_alarm() sets an alarm, which may already be pending.  If it is
   pending, it is rescheduled.  The user supplies MY_ALARM, a pointer to
   storage for the alarm state, which has been initialized with
   gm_init_alarm_my_alarm().  When the alarm occurs, CALLBACK(CONTEXT)
   is called.
 
 
   gm_set_alarm() schedules \b CALLBACK(CONTEXT) to be called after \b USEC
   microseconds (or
   later), or reschedule the alarm if it has already been scheduled and has
   not yet triggered.  CALLBACK must be non-NULL.  CONTEXT is treated as
   an opaque pointer by GM, and will be passed as the single parameter to
   the client-supplied CALLBACK function.
 
   GM clients will also be able to take advantage of the fact that an
   application is guaranteed to receive a single \b GM_ALARM_EVENT for each call   to a client-supplied callback, with the corresponding callback occurring
   during the call to gm_unknown() that processes that alarm.  This means that
   a case statement like the following in the client's event loops can be used
   to significantly reduce the overhead of polling for any effect of a client
   supplied alarm callback:
 
   @verbatim
     case GM_ALARM_EVENT:
       gm_unknown (event);
 
          poll for the effect of alarm callbacks only here
 
       break;
   @endverbatim
 
   @param port \b (IN) A pointer to the GM Port.
 
   @param my_alarm \b (IN) Alarm to be initialized.
 
   @param usecs
 
   @param callback
 
   @param context
 
   @see gm_cancel_alarm gm_initialize_alarm gm_flush_alarm
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
void
gm_set_alarm (gm_port_t *port, gm_alarm_t *my_alarm, gm_u64_t usecs,
	      void (*callback) (void *), void *context)
{
  gm_alarm_t **a;

  GM_CALLED_WITH_ARGS (("*,*,%d,*,*", (int) usecs));
  
  /* Guarantee that the alarm is not pending */

  if (my_alarm->state != GM_ALARM_FREE)
    gm_cancel_alarm (my_alarm);
      
  /* Store the alarm state */

  my_alarm->state = GM_ALARM_SET;
  my_alarm->port = port;
  my_alarm->callback = callback;
  my_alarm->context = context;
  my_alarm->deadline = gm_ticks (port) + 2 * usecs; /* ticks are .5us each */

  /* insert the alarm in the list of alarms */

  for (a = &port->first_alarm; *a; a = &(*a)->next)
    {
      if (gm_cmp64 (my_alarm->deadline, (*a)->deadline) < 0)
	break;
    }
  my_alarm->next = *a;
  *a = my_alarm;
  
  /* arrange for a lanai alarm at an appropriate time. */

  _reschedule_lanai_alarm_if_needed (port);
  
  GM_PRINT (GM_DEBUG_ALARMS, ("gm_set_alarm() returning\n"));
  GM_RETURN_NOTHING ();
}
