/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1999 by Myricom, Inc.                                   *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_ticks.c
   This file contains the GM API function gm_ticks().
*/

#include "gm_internal.h"

/**
   gm_ticks() returns a 64-bit extended version of the LANai real time
   clock (RTC).  For implementation reasons, the granularity of 
   gm_ticks() is 50 microseconds at the application level.
 
   @retval gm_u64_t
 
   @param port \b (IN) The handle to the GM port.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT gm_u64_t
gm_ticks (struct gm_port *port)
{
  gm_u32_t low, high, old_high;
  gm_u64_t both;
  
  gm_assert (port->first_sram_page);
    
  /* The LANai writes the RTC64 value as write_high, write_low.
     The Host reads the value either atomically or as read_low read_high.
     Therefore, the possible orders of these operations is

     read_low read_high write_high write_low
     read_low write_high read_high write_low
     read_low write_high write_low read_high
     write_high read_low write_low read_high
     write_high write_low read_low read_high

     In all of these cases, if there is an inconsistency, the
     read_high gets the new value and read_low get the old value.
     Therefore, if we can detect when read_high gets a new value, then
     we can detect when an inconsistency may have occurred.

     Noting that RTC64 is a counter that is incremented once every
     .5usec, we see that even if there is a read order consistency, it
     only gives us a bad value if we read it when the high part
     changes (once every 30 minutes or so), and we can guarantee that
     we will be able to read the value twice in the time it takes for
     the high part to wrap around (about 2 centuries).

     Fortunately, we can catch the once every 30 minutes inconsistency
     by confirming that the high part did not change since the last
     read.  In this case, we know that the low part and high part that
     we read are consistent.  Otherwise, we simply try again. */
  
  old_high = port->gm_ticks_old_high;
  while (1)
    {
      /* The I/O mapped memory addrs are always read order of
	 increasing address, and we want to read both halves in a
	 single I/O operation.  Therefore we store the high and low
	 parts in a sort of "Wrong-Endian" order in the LANai so we can
	 perform a 64-bit read.  Here we extract the parts. */

      both = gm_ntoh_u64 (port->first_sram_page->rtc.whole);
      low = (gm_u32_t) (both >> 32); /* really! */
      high = (gm_u32_t) both;	/* really! */
      if (old_high == high)
	{
	  /* There is no chance of an inconsistency */
	  break;
	}
      /* The reads may have been inconsistent.  Try again.  */

      old_high = high;
    }
  port->gm_ticks_old_high = high;
  return ((gm_u64_t) high << 32) + low;
}

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End: */
