/*
  mi.c
  
  This is the shim between GM and the mapper.
  Functions beginning with mi_* are the interface to the mapper (mapper interface).
  see "mi.h" (included in "lx.h").sverbose
*/

#include <stdarg.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <time.h>
#include "gm_internal.h"
#include "lx.h"

#define MAX_PRINTLINE 200

extern const char*map_filename;
extern const char*out_filename;
extern const char*daemon_pid_filename;
extern int mi_verbose_flag;
extern int mi_is_demon;
extern FILE*out_fp;

static int last_alarm_hook_count;
static int alarm_hook_count;

static void alarm_hook (void*p)
{
  *(int*)p = 1;
  alarm_hook_count++;
}

void mi_println_verbose_function (const char*fmt,...)
{
  va_list args;
  char s [MAX_PRINTLINE];

  if (!out_fp && daemon_pid_filename)
    return;
  
  va_start (args, fmt);
  vsprintf (s, fmt, args);
  va_end (args);

  fprintf (out_fp ? out_fp : stdout, "%s\n", s);
  if (out_fp)
    fflush (out_fp);
}

void mi_send (lx_t*lx, void*route, unsigned route_length, void*p, unsigned length, void*context)
{
  buffer_t*t;
  
  insist (route_length <= LX_ROUTE_SIZE);
  
  if (length <= LX_SMALL_MTU)
  {
    t = lx->mi.smalls [lx->mi.small_count % NUM_SMALLS];
    lx->mi.small_count++;
  }
  else
  {
    insist (length <= LX_BIG_MTU);
    
    t = lx->mi.bigs [lx->mi.big_count % LX_NUM_BIGS];
    lx->mi.big_count++;
  }
  
  t->context = context;
  memcpy (t->p, route, route_length);
  memcpy (t->p + route_length, p, length);
  
  _gm_raw_send (lx->mi.gm_port, t->p, length + route_length, route_length);
  
  except:;
}

void mi_set_alarm (lx_t*lx, unsigned microseconds)
{
  if (microseconds)
  {
    last_alarm_hook_count = alarm_hook_count;
    gm_set_alarm (lx->mi.gm_port, &(lx->mi.gm_alarm), microseconds, alarm_hook, &(lx->mi.alarmed));
  }
  else
  {
    last_alarm_hook_count = 0;
    gm_cancel_alarm (&(lx->mi.gm_alarm));
  }
}

void mi_set_map_version (struct lx_t*lx, unsigned char mac_address [6], unsigned map_version, unsigned num_hosts, int done_everywhere, int valid)
{
  gm_status_t status;
  struct gm_mapper_state ms;
  
  insist (lx && mac_address);
  
  memcpy (ms.mapper_mac, mac_address, sizeof (ms.mapper_mac));
  ms.map_version = map_version;
  ms.num_hosts = num_hosts;
  ms.network_configured = done_everywhere;
  ms.routes_valid = valid;
  status = _gm_set_mapper_state (lx->mi.gm_port, &ms);
  insist (status == GM_SUCCESS);
  except:;
}

void mi_set_route_begin (struct lx_t*lx)
{
  
}

void mi_set_route_end (struct lx_t*lx)
{
}

void mi_set_route (lx_t*lx, int index, unsigned char mac_address [6], int host_type, int host_index, signed char*route, unsigned length)
{
  int r;
  signed char hops [LX_ROUTE_SIZE];
  
  mi_memcpy (hops, route, length);
  lx_physical_route (hops, length);
  
  if ((r = _gm_set_route (lx->mi.gm_port, mac_address, length, hops)) != GM_SUCCESS)
    gm_perror ("_gm_set_route failed", r);
}

void mi_get_address (lx_t*lx, unsigned char mac_address [6])
{
  memcpy (mac_address, lx->mi.gm_address, 6);
}

int mi_wait (lx_t*lx, void*message, unsigned*length, void*contexts [], int early)
{
  gm_up_n_t *k;
  
  insist (message && length && contexts);
  
  for (;;)
  {
    gm_recv_event_t*event;

    event = gm_blocking_receive_no_spin (lx->mi.gm_port);

    switch (GM_RECV_EVENT_TYPE (event))
    {
      case GM_SENT_EVENT:
	*length = 0;
	for (k = (gm_up_n_t *) gm_ntohp (event->sent.message_list); gm_ntohp (*k); k++)
	{
	  buffer_t*t = (buffer_t*) (((char*)(gm_ntohp (*k))) - sizeof (char*));
	  contexts [*length] = t->context;
	  (*length)++;

	}
	return LX_SEND_DONE;
      case GM_RAW_RECV_EVENT:

	*length = gm_ntoh_u32 (event->recv.length);

	if (*length <= LX_BIG_MTU)
	  memcpy ((char*) message, (char*) gm_ntohp (event->recv.buffer), *length);

	_gm_provide_raw_receive_buffer (lx->mi.gm_port, gm_ntohp (event->recv.buffer));
	insist (*length <= LX_BIG_MTU);

	return LX_RECEIVE;

      case GM_FAST_RECV_EVENT:
      case GM_FAST_HIGH_RECV_EVENT:
      case GM_FAST_PEER_RECV_EVENT:
      case GM_FAST_HIGH_PEER_RECV_EVENT:
      case GM_RECV_EVENT:
      case GM_HIGH_RECV_EVENT:
      case GM_PEER_RECV_EVENT:
      case GM_HIGH_PEER_RECV_EVENT:
	insist (0);
	break;
      case GM_NO_RECV_EVENT:
	break;
      case GM_ALARM_EVENT:
	lx->mi.alarmed = 0;
	gm_unknown (lx->mi.gm_port, event);
	if (lx->mi.alarmed)
	{
	  last_alarm_hook_count = 0;
	  lx->mi.alarmed = 0;
	  return LX_ALARM;
	}
	break;
      case GM_BAD_SEND_DETECTED_EVENT:
      case GM_SEND_TOKEN_VIOLATION_EVENT:
	insist (0);
      
      default:
	gm_unknown (lx->mi.gm_port, event);
	break;
    }
  }
  except:
  return 0;
}

/*fix broken GM definitions of htonl*/

mi_u32_t mi_htonl (mi_u32_t x)
{
  gm_u32_n_t k;
  k = gm_hton_u32 (x);
  return GM_N (k);
}

mi_u16_t mi_htons (mi_u16_t x)
{
  gm_u16_n_t k;
  k = gm_hton_u16 (x);
  return GM_N (k);
}

void mi_fail (void)
{
}
unsigned mi_time (lx_t*lx)
{
  return gm_ticks (lx->mi.gm_port);
}

static FILE*map_fp;
void mi_dump_map (lx_t*lx, lx_map_t*m)
{
  if (map_filename && (map_fp = fopen (map_filename, "w")))
  {
    lx_map_print (lx, m);
    fclose (map_fp);
    map_fp = 0;
  }
}
void mi_println_map (lx_t*lx, const char*fmt,...)
{
  va_list args;
  char s [MAX_PRINTLINE];
  
  va_start (args, fmt);
  vsprintf (s, fmt, args);
  va_end (args);

  if (map_fp)
    fprintf (map_fp, "%s\n", s);
}

int mi_rand (void)
{  
  int a = rand ();
  a = (a >> 9);
  return a;
}
void mi_go (lx_t*lx, int unit)
{
  gm_status_t r;
  int i;
#if GM_OS_WIN2K
  char name[GM_MAX_HOST_NAME_LEN+1];
  char name2[MAX_COMPUTERNAME_LENGTH+1];
  DWORD name2_len;
#endif

  lx->mi.alarmed = 0;
  lx->mi.small_count = lx->mi.big_count = 0;

  lx->host_type = LX_GM_HOST_TYPE;
  lx->num_routes = 1;

  lx->die = 0;

  /* Allocate GM resources */

  if ((r = _gm_mapper_open (&(lx->mi.gm_port), unit, (enum gm_api_version) GM_API_VERSION)) != GM_SUCCESS)
  {
    gm_perror ("gm_mapper_port open failed", r);
    exit (1);
  }

#if GM_OS_WIN2K
  if (gm_get_host_name (lx->mi.gm_port, name) == GM_SUCCESS)
  {
    name2_len = MAX_COMPUTERNAME_LENGTH+1;
    if ((strncmp(name, "localhost", strlen("localhost")) == 0)
	&& GetComputerName(name2, &name2_len))
    {
      if (unit > 0) {
	sprintf(name, "%s:%d", name2, unit);
      }
      else {
	strcpy(name, name2);
      }
      _gm_set_host_name (lx->mi.gm_port, &name);
      gm_get_host_name (lx->mi.gm_port, name);
      printf("[%s]\n", name);
    }
  }
#endif


  gm_get_unique_board_id (lx->mi.gm_port, lx->mi.gm_address);

  for (i = 0; i < NUM_SMALLS; i++)
  {
    if (!(lx->mi.smalls [i] = (buffer_t*) gm_dma_malloc (lx->mi.gm_port, sizeof (buffer_t) + LX_ROUTE_SIZE)))
    {
      gm_perror ("gm_dma_malloc failed", 0);
      exit (1);
    }
  }
  for (i = 0; i < LX_NUM_BIGS; i++)
  {
    if (!(lx->mi.bigs [i] = (buffer_t*) gm_dma_malloc (lx->mi.gm_port, sizeof (buffer_t) + LX_BIG_MTU - LX_SMALL_MTU + LX_ROUTE_SIZE)))
    {
      gm_perror ("gm_dma_malloc failed", 0);
      exit (1);
    }
  }
  for (i = 0; i < NUM_RECEIVES; i++)
  {
    if (!(lx->mi.receives [i] = gm_dma_malloc (lx->mi.gm_port, LX_BIG_MTU + LX_ROUTE_SIZE)))
    {
      gm_perror ("gm_dma_malloc failed", 0);
      exit (1);
    }
    _gm_provide_raw_receive_buffer (lx->mi.gm_port, lx->mi.receives [i]);
  }

  gm_initialize_alarm (&(lx->mi.gm_alarm));
  srand (gm_ticks (lx->mi.gm_port));
  
  /* Map forever */  

  lx_mapper_main (lx);
}

/*
  relax GM indent
  Local Variables:
  c-file-style:""
  End:
*/
