#include <stdarg.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <time.h>
#include "gm_internal.h"
#include "lx.h"

/****************
 * Autoconfiguration
 ****************/

#if GM_BUILDING_GM
#include "gm_auto_config.h"
#elif HAVE_CONFIG_H
#include "config.h"
#endif

#define CAN_BE_A_DAEMON (HAVE_SYS_TYPES_H && HAVE_UNISTD_H && HAVE_FCNTL_H)

#if CAN_BE_A_DAEMON

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

int getdtablesize(void);
int open (const char *pathname, int flags, ...);
mode_t umask (mode_t mask);
pid_t setsid (void);

static
int demonize (const char*filename)
{ 
  int pid;
  int pid_file;
  int fd;
  char buf[16];

  if (!filename)
    return 0;
	
  /* Create a new thread, so the parent can exit immediately. */

  pid = fork ();
  if (pid < 0)
  {
    printf ("Failed to spawn daemon.\n");
    /* Error spawning daemon. */
    return 0;
  }
  else if (pid > 0)
  {
    /* Parent exits successfully. */
    exit (0);
  }

  /* Create a process group, so we won't be killed when our parent
     process group exits. */

  setsid ();		/* always succeeds for child */

  /* Write the process ID to a file, taking care that there is
     no other mapper running and holding a lock on this file. */
      
  pid_file = open (filename, O_RDWR | O_CREAT, 0644);
  if (pid_file < 0)
  {
    printf ("could not open %s:", filename);
    perror ("");
    return 0;
  }
  if (lockf (pid_file, F_TLOCK, 0) < 0)
  {
    /* We cannot lock the file, so there must be another mapper
       running. */
    printf ("could not lock %s:", filename);
    perror ("");
    return 0;
  }
  sprintf (buf, "%d\n", getpid ());
  if (write (pid_file, buf, strlen (buf)) == -1)
  {
    printf ("could not write PID to %s:", filename);
    perror ("");
    return 0;
  }
  
  /* Close any open files that might be attached to the controlling
     terminal.  Not doing this is know to prevent the mapper from
     being spawned via ssh. */

  if ((fd = open("/dev/null", O_RDWR, 0)) != -1)
  {
    (void)dup2(fd, STDIN_FILENO);
    (void)dup2(fd, STDOUT_FILENO);
    (void)dup2(fd, STDERR_FILENO);
    if (fd > 2)
    {
      (void)close(fd);
    }
  }
  return 1;
}

#endif /* CAN_DAEMON */

/****************
 * End of autoconfiguration
 ****************/

const char*map_filename;
const char*out_filename;
const char*daemon_pid_filename;
int mi_verbose_flag;
int mi_is_demon;
FILE*out_fp;

static void signal_handler (int sig)
{
  if ((mi_verbose_flag = !mi_verbose_flag))
  {
    if (out_filename)
      out_fp = fopen (out_filename, "w");
  }
  else if (out_fp && out_filename)
  {
    fclose (out_fp);
    out_fp = 0;
  }
}

/* Return nonzero iff arg matches prefix.  Store a pointer to the remaining
   arg at match, if match is nonzero. */

static int match (const char *arg, const char *prefix, const char **match)
{
  gm_size_t prefix_len;

  prefix_len = strlen (prefix);
  if (!strncmp (arg, prefix, prefix_len))
  {
    if (match)
      *match = arg + prefix_len;
    return 1;
  }
  return 0;
}

static int short_match (const char *arg, const char *prefix, const char **match,
			int *argc, char **argv[])
{
  gm_size_t prefix_len;

  prefix_len = strlen (prefix);
  if (!strncmp (arg, prefix, prefix_len))
  {
    if (*argc > 1)
    {
      (*argc)--;
      (*argv)++;
      *match = **argv;
      return 1;
    }
  }
  return 0;
}

/* Print program usage information. */

static
void
usage (const char *program_name)
{
  printf
    ("usage: %s [options]\n"
     "valid options include:\n"
     "  --daemon-pid-file=<path>\n"     "  --num-passes=<integer>\n"
     "    number of route computation passes (default 8).\n"

     "    Run as a daemon, storing daemon PID at <path>.\n"
     "  --help\n"
     "    print this message and exit\n"
     "  --level=<level>\n"
     "    mapper level (default 1, 0 for never map).\n"
     "  --num-passes=<integer>\n"
     "    number of route computation passes (default 8).\n"
     "  --map-file=<path>\n"
     "    Store map files at <path>.\n"
     "  --printeur\n"
     "    debugging feature. accept network printfs.\n"
     "  --thrash\n"
     "    another debugging feature. keep remapping.\n"
     "  --extra-hosts=<count>\n"
     "    a debugging feature. make artificially inflated map.\n"
     "  --pause\n"
     "    for mute find_xbars: pause all other mappers\n"
     "  --map-once\n"
     "    terminate after mapping is complete everywhere.\n"
     "  --map-incrementally\n"
     "    re-map incrementally.\n"
     "  -B <Board>,\n" 
     "  -b <board>,\n" 
     "  --unit=<board>\n"
     "    board number.\n"
     "  --non-clos\n"
     "    for safe routing on non-clos networks.\n"
     "  --no-xbar-check\n"
     "    skip xbar check.\n"
     "  -v,\n"
     "  --verbose\n"
     "    verbose output.\n"
     "  --verbose-file=<path>\n"
     "    write verbose output to <path>\n",
     program_name);
}

struct lx_t lx;

int main (int argc, char*argv[])
{
  int unit = 0;
  const char*s;
  char*program_name = argv [0];
  int level = 1;
  int clos = 1;
  int check_xbars = 1;
  int map_once = 0;
  int printeur = 0;
  int extra_hosts = 0;
  int service = 0;
  int map_incrementally = 0;
  int num_passes = 16;
  int thrash = 0;
  int pause = 0;
  
  insist (GM_NUM_SEND_TOKENS >= NUM_SMALLS + LX_NUM_BIGS);
  

  /* Parse command-line parameters. */

  for (argc--, argv++; argc; argc--, argv++)
  {
    
    if (CAN_BE_A_DAEMON && match (*argv, "--daemon-pid-file=", &daemon_pid_filename))
    {
      printf ("Storing daemon PID in file \"%s\"\n", daemon_pid_filename);
      continue;
    }
    if (!strcmp (*argv, "--help"))
    {
      usage (program_name);
      exit (0);
    }
    if (match (*argv, "--map-file=", &map_filename))
      continue;
    if (match (*argv, "--verbose-file=", &out_filename))
      continue;
    if (match (*argv, "--unit=", &s)
	|| short_match (*argv, "-B", &s, &argc, &argv)
	|| short_match (*argv, "-b", &s, &argc, &argv))
    {
      unit = atoi (s);
      continue;
    }
    if (match (*argv, "--level=", &s))
    {
      level = atoi (s);
      continue;
    }
    if (match (*argv, "--num-passes=", &s))
    {
      if ((num_passes = atoi (s)) < 1 || num_passes > 16)
      {
	usage (program_name);
	exit (0);
      }
      continue;
    }
    if (match (*argv, "--extra-hosts=", &s))
    {
      extra_hosts = atoi (s);
      continue;
    }
    if (!strcmp (*argv, "--non-clos"))
    {
      clos = 0;
      continue;
    }
    if (!strcmp (*argv, "--thrash"))
    {
      thrash = 1;
      continue;
    }
    if (!strcmp (*argv, "--no-xbar-check"))
    {
      check_xbars = 0;
      continue;
    }
    if (!strcmp (*argv, "--map-once"))
    {
      map_once = 1;
      continue;
    }
    if (!strcmp (*argv, "--pause"))
    {
      pause = 1;
      continue;
    }
    if (!strcmp (*argv, "--map-incrementally"))
    {
      map_incrementally = 1;
      continue;
    }
    if (!strcmp (*argv, "--printeur"))
    {
      printeur = 1;
      continue;
    }
    if (!strcmp (*argv, "--verbose") || !strcmp (*argv, "-v"))
    {
      mi_verbose_flag = 1;
      continue;
    }
    if (!strcmp (*argv, "--service"))
    {
      service = 1;
      continue;
    }

    usage (program_name);
    printf ("Unrecognized option: \"%s\"\n", *argv);
    exit (1);
  }  

  if (mi_verbose_flag && !daemon_pid_filename && out_filename)
  {
    mi_verbose_flag = 0;
    signal_handler (0);
  }
  
#if CAN_BE_A_DAEMON
  /* Become a daemon, if necessary. */
  if (daemon_pid_filename)
  {
    if (!demonize (daemon_pid_filename))
      return 1;
  }
#endif

#if !GM_OS_WIN2K
  signal (SIGHUP, signal_handler);
#endif

#if GM_OS_WIN2K
  if (service)
    windows_service ();
  else
  {
#endif

    lx.level = level;
    lx.clos = clos;
    lx.check_xbars = check_xbars;
    lx.map_once = map_once;
    lx.pause = pause;
    lx.map_incrementally = map_incrementally;
    lx.printeur = printeur;
    lx.extra_hosts = extra_hosts;
    lx.num_passes = num_passes;
    lx.thrash = thrash;
    mi_go (&lx, unit);
#if GM_OS_WIN2K
  }
#endif
  return 0;
  except: return 1;
}
