/*
  sm_Packet.c
  map tools
  finucane@myri.com (David Finucane)
*/

#include <string.h>
#include <stdlib.h>

#include "insist.h"
#include "sm_Callable.h"
#include "sm_Packet.h"
#include "sm_Port.h"
#include "sm_Node.h"
#include "sm_Host.h"
#include "sm_Simulator.h"
#include "mt_Message.h"
#include "mt_htonl.h"

int sm_Packet::numSent;
int sm_Packet::numReceived;

void sm_Packet::clearCounters()
{
  numSent = 0;
  numReceived = 0;
}

sm_Packet::sm_Packet (mt_Route*route, char*p, int length, int extraLength, sm_Time startTime, int letMeDie)
{
  this->myMessage = 0;

  insist (route && p);
  insist (length + extraLength > 0);
  //insist (!((length + extraLength) % 4));

  this->letMeDie = letMeDie;
  dead = 0;
  mark = 0;
  
  route->copyOut (bytes);
  byteLength = route->getLength ();
  byteIndex = 0;
  messageLength = length;
  this->length = byteLength + messageLength + extraLength;
  
  if (messageLength <= MAX_LENGTH)
    message = buffer;
  else
  {
    myMessage = message = (char*) malloc (messageLength);
    insistp (this->message, ("sm_Packet::sm_Packet: alloc failed"));
  }
  insist (message);
  insist (!((long) message % sizeof (int)));

  memcpy (message, p, length);
  
  headPort = tailPort = oldTail = 0;
  headPosition = 0;
  tailPosition = 0;
  next = prev = nextBlocked = 0;
  this->startTime = lastTime = startTime;
  blocked = 1;
  blockedPort = 0;
  
  id = numSent++;
  latency = 0;
  exception: return;
}

sm_Packet::~sm_Packet ()
{
  if (myMessage) 
    free (myMessage);
}

void sm_Packet::received(sm_Time when)
{
  latency = when - startTime;
  numReceived++;
}

char*sm_Packet::getMessage ()
{
  insist (this);
  return message;
  exception: return 0;
}

int sm_Packet::getMessageLength ()
{
  insist (this);
  return messageLength;
  exception: return 0;
}

int sm_Packet::getLatency ()
{
  return latency;
}

int sm_Packet::getNumReceived ()
{
  return numReceived;
}
int sm_Packet::getNumSent ()
{
  return numSent;
}

int sm_Packet::getId ()
{
  return id;
}

sm_Port*sm_Packet::getNextHeadPort ()
{
  insist (this);
  insist (!dead);
  insist (headPort);

  sm_Node*n;
  n = (sm_Node*) headPort->getNode ();
  insist (n);

  sm_Port*p;
  p = headPort->getNext ();
  if (p && p->getNode () != n)
    return p;

  if (byteLength)
  {
    insist (byteIndex <= byteLength);
    if (byteIndex == byteLength)
      return 0;
    return n->getOutput (headPort->getPort () + bytes [byteIndex]);
  }
  else return n->getOutput (headPort->getPort ());
  exception: return 0;
}



void sm_Packet::call (sm_EventList*list, int type, void*)
{
  sm_Time ct = list->getCurrentTime ();
  //blocked = 0;
  insist (ct >= lastTime);
  update (list);

  exception: return;
}  
int sm_Packet::getTailPosition ()
{
  return tailPosition;
}
void sm_Packet::setHeadPort (sm_Port*p)
{
  headPort = p;
}
void sm_Packet::setTailPort (sm_Port*p)
{
  tailPort = p;
}
int sm_Packet::getLength ()
{
  return length;
}
sm_Port*sm_Packet::getHeadPort ()
{
  return headPort;
}
sm_Port*sm_Packet::getTailPort ()
{
  return tailPort;
}
void sm_Packet::setHeadPosition (int p)
{
  headPosition = p;
}
void sm_Packet::setTailPosition (int p)
{
  tailPosition = p;
}
void sm_Packet::setStartTime (sm_Time time)
{
  startTime = time;
}
sm_Time sm_Packet::getStartTime ()
{
  return startTime;
}

void sm_Packet::dump (FILE *fp)
{
  if (mark == PRINT_MARK)
    return;

  sm_Port*nextPort = getNextHeadPort ();
  
  mark = PRINT_MARK;
  
  fprintf (fp, "packet %d %s:\n", id, blocked ? "(blocked)" : "");
  fprintf (fp, "tail location: node %s %s port %d position %d\n",
	   tailPort->getNode ()->getName (),
	   tailPort == tailPort->getNode()->getOutput (tailPort->getPort ()) ? "output" : "input",
	   tailPort->getPort (),
	   tailPosition);
  fprintf (fp, "head location: node %s %s port %d position %d\n",
	   headPort->getNode ()->getName (),
	   headPort == headPort->getNode()->getOutput (headPort->getPort ()) ? "output" : "input",
	   headPort->getPort (),
	   headPosition);

  fprintf (fp, "next: %d\n", next ? next->id : -1);
  if (nextPort)
  {
    sm_Packet*k;
    
    fprintf (fp, "on next port: %d\n", (k = nextPort->getPacket ()) ? k->id : -1);
    
    fprintf (fp, "waiting on next port:\n");
    for (sm_Packet*k = nextPort->head; k; k = k->nextBlocked)
      fprintf (fp, "packet %d %s:\n", k->id, k->blocked ? "(blocked)" : "");

    fprintf (fp, "next port %s %s port %d\n",
	     nextPort->getNode ()->getName (),
	     nextPort == nextPort->getNode()->getOutput (nextPort->getPort ()) ? "output" : "input",
	     nextPort->getPort ());
  }
  
  fprintf (fp, "message length: %d\n",messageLength);
  fprintf (fp, "route: ");
  
  for (int i = 0; i < byteLength; i++)
    fprintf (fp, "%d ", bytes [i]);

  fprintf (fp, "\nbody:\n");

  for (int i = 0; i < messageLength; i++)
    fprintf (fp, "%2.2x%s", message [i] & 0xff, i && (i % 8 == 0) ? "\n" : " ");
  
  fprintf (fp,"\n\n");
}

void sm_Packet::dumpChain (FILE *fp, int detailed)
{
  insist (this);
  
  if (mark == PRINT_MARK)
    return;

  for (sm_Packet*p = this; p && !p->mark; p = p->next)
  {
    if (detailed)
      p->dump (fp);
    else
      fprintf (fp, "%d ", p->id);
   
    p->mark = PRINT_MARK;
    
    sm_Port*o;
    sm_Packet*k;

    if (!p->next && (o = p->getNextHeadPort ()) && (k = o->getPacket ()))
      k->dumpChain (fp, detailed);
  }
  
  for (sm_Packet*p = this; p && p->mark; p = p->next)
    p->mark = UNMARKED;

  exception:;
}

int sm_Packet::addBlocked (sm_EventList*list, sm_Port*p)
{
  insist (p);
  insist (list);
  sm_Packet*ot;
  
  removeAllEvents (list);

  insist (!blocked);
  blocked = 1;
  insist (this != p->getBlocked ());
  insist (p == getNextHeadPort ());
  
  perr ("blocked message %d on port (%x) %s  (behind %d)",
	  this->id, p, p->getNode()->getName(), p->getBlocked()->id);

  if (this == p->getPacket ())
    return 0;
  
  insist (!dead);
  ot = p->tail;
  
  insist (!blockedPort);
  blockedPort = p;
  
  if (p->tail)
    p->tail->nextBlocked = this;
  else p->head = this;
  p->tail = this;
  p->tail->nextBlocked = 0;
  insist (p->head);
  
  next = 0;
  insist (waitingToCross ());
  
  if (isDeadlocked (list))
  {
    fprintf (stdout, "deadlock detected\n");
    dumpChain (stdout, 0);
    fprintf (stdout, "\n");
    
    if (ot)
      ot->nextBlocked = 0;
    else
      p->head = p->tail = 0;
    blockedPort = 0;
    return 0;
  }
  
  return 1;
  exception: return 0;
}

int sm_Packet::slamIntoNext (sm_EventList*list, int td)
{
  insist (list && next);
  insist (td);
  insist (next->tailPort == headPort && next->tailPosition > headPosition);
  
  int s;
  s = next->tailPosition - 1 - headPosition;
  
  if (s < td)
  {
    td = s;
    if ((blocked = next->blocked) && isDeadlocked (list))
    {
      sm_Packet*t = next;
      fprintf (stdout, "deadlock detected\n");
      dumpChain (stdout, 0);
      fprintf (stdout, "\n");

      dead = 1;
      dropDeadlocked (list);
      //insist (!t->isDeadlocked (list));
      
      return 0;
    }
  }
  
  headPosition += td;
  return td;
  exception: return 0;
}

int sm_Packet::slamIntoEnd (sm_EventList*, int td)
{
  insist (!next);  
 
  int s;
  s = headPort->getSize () - 1 - headPosition;
  insist (s >= td);

  headPosition += td;
  return td;  
  exception: return 0;
}

int sm_Packet::cross (sm_EventList *list)
{
  insist (headPosition == headPort->getSize () - 1);
  insist (!dead);
  
  sm_Port*p;
  p = getNextHeadPort ();

  insistp (p || letMeDie, ("sm_Packet::cross: packet died"));

  if (!p || p->getDisabled ())
  {
    if (p)
      perr ("disabled port %x", headPort);

    perr ("message %d dropped off of port %x", id, headPort);

    drop (list);
    return 0;
  }
  
  if (p->getBlocked () && p->getBlocked () != this)
  {
    if (!addBlocked (list, p))
    {
      perr ("dropping deadlocked message %d of port %x", id, headPort);

      dropDeadlocked (list);
      //insist (!p->getBlocked ()->isDeadlocked (list));
      
      return 0;
    }
    insist (waitingToCross ());
    return 0;
  }

  if (p->getBlocked ())
  {
    perr ("removed message %d from block list", id);
    p->removeBlocked ();
  }
    
  blocked = 0;
  
  perr ("message %d crossed into port %x", id, p);
  
  if ((next = p->getPacket ()))
  {
    insist (next != this);
    insist (!next->dead);
    insist (next->getTailPort () == p);
    next->prev = this;
  }
  
  insist (prev != this);
  
  p->setPacket (this);  

  if (p->getNode () == headPort->getNode ())
  {
    byteIndex++;
    length--;
    insist (!headPort->getNext ());
    headPort->setNext (p);
  }
  headPort = p;
  headPosition = 0;

  list->reportMove ();  

  return 1;
  exception: return 0;
}

int sm_Packet::inch (sm_EventList*, int td)
{
  insist (this);
  insist (!dead);

  if (td == 0)
    return 0;

  int s;
  s = tailPort->getSize () - 1 - tailPosition;
  if (s < td)
  {
    sm_Port*p = tailPort->getNext ();
    insist (p);
    insist (p || dead);
    if (!p)
    {
      insist (0);
      tailPort = 0;
      tailPosition = 0;
      return 0;
    }
    
    if (p->getNode () == tailPort->getNode ())
    {
      perr ("message %d tail crossed from port %x to port %x (clearing next).", id, tailPort, p);
      tailPort->setNext (0);
    }
    
    tailPort = p;
    tailPosition = td - s - 1;
    return s;
  }
  tailPosition += td;
  return td;
  exception: return 0;
}

void sm_Packet::reschedule (sm_EventList*list)
{
  insist (list);
  insist (!blocked);
  insist (!dead);

  if (list->getSimulator()->isAborted ())
    return;
  
  removeAllEvents(list);

  sm_Time ct;
  ct = list->getCurrentTime ();
  lastTime = ct;
  
  perr ("rescheduling message %d", id);
  
  int hd;
  hd = headPort->getSize () - 1 - headPosition;
  int td;
  td = tailPosition == 0 ? 1 : tailPort->getSize () - tailPosition;

  perr ("%d resc hd = %d, td = %d", id, hd, td);
  
  insist (!dead);
  if (!td || !hd)
    lastTime = list->getCurrentTime () - 1;

  blocked = 0;

  addEvent (list, td < hd ? td : hd, sm_EventList::MOVE, 0);
  perr ("message %d reschedule for %lld", id, (sm_Time) (ct + (td < hd ? td : hd)));
  exception: return;
}

void sm_Packet::update (sm_EventList*list)
{
  insist (list);
  sm_Time ct;
  ct = list->getCurrentTime();
  insist (ct >= lastTime);
  insist (!blocked);

  if (list->getSimulator()->isAborted ())
    return;

  insist (!dead);
  
  insist (headPort && tailPort);
  
  int td;
  td = ct - lastTime;
  //insist (td);
  
  if (td == 0)
  {
    perr ("zero time update for %d %d", id);
    return;
  }
  lastTime = ct;

  perr ("message %d update at time %lld head = %d, tail = %d, headPort = %x, tailPort = %x, (%d, %d)",
	  id, ct, headPosition, tailPosition, dead ? 0 : headPort, tailPort,
	  dead ? 0 : headPort->getSize(), tailPort->getSize());
  
  if (next && !next->blocked)
  {
    perr ("message %d is pushing message %d", id, next->id);
    insist (!dead);
    next->update (list);
    insist (!next || !next->dead);
  }
  
  int ntd;      
  
  if (dead)
    ntd = td;
  else if (headPosition == headPort->getSize () - 1)
    ntd = cross (list);
  else if (next)
    ntd = slamIntoNext (list, td);
  else
    ntd = slamIntoEnd (list, td);  

  if (dead)
  {
    perr ("deleting message %d", id);
    insist (!blockedPort);
    delete this;
    return;
  }
  
  sm_Port*ot;
  ot = tailPort;
  int tm;
  tm = ntd == 0 ? 0 : inch (list, ntd);
  
  if (ot != tailPort)
  {
    insist (tailPosition == 0);
    insist (!dead);
    ot->addEvent (list, 0, sm_EventList::DEPARTURE, (void*) (length + byteLength - byteIndex));
    
    perr ("message %d detected crossing. prev message is %d", id, !prev ? -1 : prev->id);
    if (ot->getPacket() == this)
      ot->setPacket (0);
    
    insist (ot->getNode () != tailPort->getNode () || !ot->getNext ());
    
    if (prev)
    {
      perr ("message %d disengaged from message %d in crossing", id, prev->id);
      insist (prev->next == this);
      prev->next = 0;
      prev->blocked = 0;
      prev->lastTime = list->getCurrentTime ();
      prev->addEvent (list, 1, sm_EventList::MOVE, 0);
      prev = 0;
    }
    oldTail = ot;
  }
  
  if (tm && tailPosition == 1)
  {
    insist (tailPort->getPacket() == this);

    if (oldTail->getNode ()->isHost () ||  oldTail->getNode()->getNodeType () == mt_Node::CLOUD)
      ((sm_Callable*)oldTail->getNode ())->addEvent(list, 0, sm_EventList::DEPARTURE, this);
    
    if (tailPort->getNode ()->isHost () || tailPort->getNode()->getNodeType () == mt_Node::CLOUD)
    {
      tailPort->setPacket (0);
      insist (!prev);
      
      perr ("message %d added arrival event for node %s", id, tailPort->getNode()->getName ());
      
      ((sm_Callable*)tailPort->getNode ())->addEvent(list, 0, sm_EventList::ARRIVAL, this);
      tailPort->unblock (list);
      return;
    }
    
    tailPort->unblock (list);
  }  

  if (!blocked)
    reschedule (list);
  else
  {
    insist (next && next->prev == this || prev && prev->next == this || waitingToCross ());
    removeAllEvents(list);
  }
  
  perr ("%d movement: head = %d, tail = %d, headPort = %x, tailPort = %x, %s %s",
	  id, headPosition, tailPosition, headPort ? headPort : 0, tailPort,
	  blocked ? "(blocked)" : "(not blocked)", dead ? "dead" : "alive");
  exception: return;
}

int sm_Packet::isGoodForHost ()
{
  insist (this);
  return byteIndex == byteLength;
  exception: return 0;
}

int sm_Packet::canDie ()
{
  insist (this);
  return letMeDie;
  exception: return 0;
}

sm_Packet*sm_Packet::isDeadlocked (sm_EventList*list)
{
  sm_Time ct = list->getCurrentTime ();
  sm_Time td = ct - startTime;
  sm_Packet*deadlocked = 0;

  for (sm_Packet*p = next; p; p = p->next)
  {
    insist (!dead);

    if (!p->blocked)
      break;
    
    if (p->mark == DEADLOCK_MARK)
    {
      deadlocked = p;
      break;
    }
    else p->mark = DEADLOCK_MARK;

    if (!p->next)
    {
      sm_Port*o = p->getNextHeadPort ();
      insist (o);
      sm_Packet*k;
      if ((k = o->getPacket ()) && k == o->getBlocked () && (deadlocked = k->isDeadlocked (list)))
	break;
    }
  }
  
  for (sm_Packet*p = next; p; p = p->next)
  {
    if (!p->blocked || !p->mark)
      break;

    p->mark = UNMARKED;
  }
#if 0
  if (td > (sm_Time) sm_EventList::MICROS_PER_TICK * 1000 * 1000)
    return 1;
#endif
  return deadlocked;
  exception:
  return 0;
}


int sm_Packet::dropDeadlocked (sm_EventList*list)
{
  sm_Packet*t = isDeadlocked (list);
  insist (t);
  t->drop (list);
#if 0
  
  if (!t)
  {
    sm_Port*o = getNextHeadPort ();
    insist (o);
    t = o->getPacket ();
  }
  
  drop (list);

  if (t && t->isDeadlocked (list))
  {
    t->dropDeadlocked (list);
    delete t;
  }
  #endif
  exception:
  return 0;
}

int sm_Packet::getMark ()
{
  insist (this);
  return mark;
  exception: return 0;
}

void sm_Packet::setMark (int mark)
{
  insist (this);
  this->mark = mark;
  exception: return;
}

int sm_Packet::drop (sm_EventList*list)
{  
  insist (list);
  insist (headPort && tailPort);
  
  removeAllEvents (list);

  if (next)
  {
    insist (next->tailPort == headPort);
    insist (next->blocked);
    insist (next->prev == this);

    next->prev = 0;

    if (headPort->getPacket () == this)
    {
      headPort->setPacket (next);
      headPort->unblock (list);
    }
  }
  if (prev)
  {
    insist (prev->headPort == tailPort);
    insist (tailPort->getPacket () != this);

    if (headPort == tailPort && next)
    {
      prev->next = next;
      next->prev = prev;
    }
    else prev->next = 0;
    prev->blocked = 0;
    prev->reschedule (list);
  }
  next = prev = 0;

  while (tailPort)
  {
    sm_Port*p;

    if ((p = tailPort->getNext ()) && p->getNode () == tailPort->getNode ())
      tailPort->setNext (0);

    if (tailPort->getPacket () == this)
    {
      tailPort->setPacket (0);
      tailPort->unblock (list);
    }
    tailPort = tailPort == headPort ? 0 : p;
  }

  dead = 1;
  exception: return 0;
}

int sm_Packet::waitingToCross ()
{
  insist (this);
  sm_Port*p;
  p = getNextHeadPort ();

  if (!p) return 0;

  for (sm_Packet*k = p->head; k; k = k->nextBlocked)
    if (k == this)
      return 1;

  exception: return 0;
  
}

void sm_Packet::check ()
{
  insist (this);
  exception:;
}
