/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997, 1998, 1999 by Myricom, Inc.                 *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* author: glenn@myri.com */

/* NOTE: If you change the operating characteristics or the flags
 *       used by this program, please notify Max so he can change the
 *       Myrinet Handbook document.
 */

#if (defined __STDC__) || (defined WIN32)
#if GM_OS_VXWORKS
#include "vxWorks.h"
#endif  /* GM_OS_VXWORKS */
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include <time.h>
#if (!defined WIN32) && (defined HAVE_SYS_TIME_H)
#include <sys/time.h>
#endif

#include <math.h>
#include <sys/types.h>
#ifndef WIN32
#include <sys/wait.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "gm.h"
#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_trace.h"
#include "stats.h"

#define GM_DEBUG_ALLSIZE 0
/* Set this to 0 for no alarm, or to the number of seconds between alarms */
#define GM_DEBUG_ALLSIZE_ALARM 0
#define GM_DEBUG_ALLSIZE_RECV_Q 0
#define GM_LONG_PACKET_ERROR_PRINT 1

#define TIME_CALIBRATION 1e6

#define PRINTF(a) GM_PRINT (GM_DEBUG_ALLSIZE, a)

/* test parameters */

#ifndef WIN32

#define LEN_INC 8
#define MAX_TEST_LEN(size) (gm_max_length_for_size (size))
#define MAX_TEST_SIZE 12
#define MIN_TEST_LEN 1
#define MIN_TEST_SIZE 12
#define RECV_TOKENS_OF_SIZE 20
#define SEND_BUFFERS 8

#else /* WIN32 */

/* WIN32 */
#define MIN_TEST_SIZE 12
#define MAX_TEST_SIZE 12
#define MIN_TEST_LEN 1
#define MAX_TEST_LEN(size) (gm_max_length_for_size (size))
#define LEN_INC	8
#define SEND_BUFFERS 3
#define RECV_TOKENS_OF_SIZE 6

#endif

/* Possible useful test for setting the number of receive tokens to 0
   to test behaviour when the receiver does not provide receive tokens
   fast enough. */

#if 0
#undef RECV_TOKENS_OF_SIZE
#define RECV_TOKENS_OF_SIZE 1
#endif

#define PAGE_OFFSET(ptr) ((gm_u32_t)((gm_size_t) (ptr)) & (GM_PAGE_LEN - 1))
#define PAGE_REMAINING(ptr) (GM_PAGE_LEN - PAGE_OFFSET (ptr))
#define PAGE_ALIGNED(ptr) (PAGE_MISALIGNMENT (ptr) == 0)

/****************************************************************
 * Typedefs
 ****************************************************************/

enum {
  NO,
  YES,
  MAYBE
};

enum send_mode {
  RELIABLE_MODE,
  DATAGRAM_MODE,
  MAYBE_DATAGRAM_MODE,
  GET_MODE,
  PUT_MODE,
  ETHERNET_MODE
};

enum { GM_ARBITRARY_RECV_BUFFER_TAG = 42 };

/****************
 * messages
 ****************/

enum message_type
{
  TYPE_BLANK = 0,
  
  TYPE_LEN_MASK = 0xf,
  TYPE_REPLY = 0x10,
  TYPE_VERIFY = 0x20,
  
  TYPE_SEND_MODE_MASK = 0xc0,
  TYPE_SEND_MODE_NONPEER = 0x00,
  TYPE_SEND_MODE_PEER = 0x40,
  TYPE_SEND_MODE_DATAGRAM = 0x80,
  TYPE_SEND_MODE_DATAGRAM_OR_RELIABLE = 0xc0
};

#define MESSAGE_LEN(m) ((m)->type & TYPE_LEN_MASK			\
			? (m)->type & TYPE_LEN_MASK			\
			: gm_ntoh_u32 ((m)->large.length))

typedef union message
{
  gm_u8_t type;
  struct small_message {
    gm_u8_n_t type;
    gm_u8_n_t data[1];
  } small;
  struct large_message {
    gm_u8_n_t type;
    gm_u8_n_t pad[3];
    gm_u32_n_t length;
    gm_u8_n_t data[1];
  } large;
} message_t;

typedef struct message_ref
{
  struct message_ref *next;
  struct message_ref *prev_created;
  char *original_ptr;
  message_t *message;
  int size;
  char is_free;
} message_ref_t;

/****************
 * other
 ****************/

struct test;			/* forward declaration to placate gcc */

typedef void send_func_t (struct test *t,
			  struct message_ref *ref,
			  unsigned int size,
			  unsigned long len,
			  enum message_type basic_type,
			  unsigned int target_node_id,
			  unsigned int target_port_id);
static send_func_t _send_abort;
static send_func_t _send_datagram;
static send_func_t _send_datagram_or_reliable;
static send_func_t _send_to_nonpeer;
static send_func_t _send_to_peer;

typedef struct enqueued_send
{
  struct enqueued_send *next;
  struct test *test;
  message_ref_t *message_ref;
  unsigned int size;
  unsigned long len;
  enum message_type basic_type;
  unsigned int target_node_id;
  unsigned int target_port_id;
  send_func_t *send;
  void (*fill) (struct test *t, gm_u8_n_t *buf, unsigned int len, char seed);
} enqueued_send_t;

typedef void recv_func_t (struct test *t, struct gm_port *p,
			  gm_recv_event_t *e, send_func_t send);
static recv_func_t recv_only;
static recv_func_t recv_and_verify;
static recv_func_t recv_and_reply;
static recv_func_t recv_and_reply_and_verify;
static recv_func_t recv_slave;

typedef struct test
{
  /* NOTES:
   * The general principle is that options are processed left to right on the
   * command line, and later options override earlier ones.  In a few cases,
   * options have multiple relationships.  Thus, --both-ways is only meaningful
   * with --bandwidth and --bidirectional, so we allow it to IMPLY them.
   * Thus, the following set of flags:
   *   --both-ways --latency
   * will result in --both-ways being OFF; otoh, the following:
   *   --latency --both-ways
   * will result in --latency being OFF (i.e., --bandwith ON).
   */

  char *points;
  char *remote_host_name;
  char *title;
  char *ylabel;
  enum gm_priority priority;
  int datagrams;
  int do_copy;
  int fast;
  int fork_child;
  int fork_parent;
  int get;
  int hysteresis;
  int rand_seed;
  int randomize;
  int reverse;
  int send_mode;
  int stats_socket;
  int timestamps;
  char *time_string;
  int version;
  struct stats *stats;
  unsigned int both_ways;	/* overrides latency and unidirectional */
  unsigned int count_per_len;
  unsigned int density;
  unsigned int enable_nack_down_flag;
  unsigned int enable_nack_down_set;
  unsigned int exit_on_error;
  unsigned int length_inc;
  unsigned int max_len;
  unsigned int max_size;
  unsigned int min_len;
  unsigned int min_sends_per_point;
  unsigned int min_size;
  unsigned int offset;
  unsigned int outstanding_send_cnt;
  unsigned int ping_pong;	/* else spray */
  unsigned int port_id;
  unsigned int report_bytes_per_sec; /* else report Mbyte/sec*/
  unsigned int report_latency;	/* else report bandwidth */
  unsigned int send_cnt;
  unsigned int recv_cnt;
  unsigned int target_node_id;
  unsigned int target_port_id;
  unsigned int timer;
  unsigned int verify;
  unsigned long min_usecs_per_point;
  
  message_ref_t *free_send_message_refs[GM_NUM_SIZES];
  /* keep track of the memory allocated so we can free it later*/
  message_ref_t *last_created_message_ref;
  
  FILE *data_file;
  struct 
  {
    FILE *file;
    char *plot_string;
    char *basename;
  } gnuplot;
  
  enqueued_send_t *first_enqueued_send;
  enqueued_send_t *last_enqueued_send;

  struct gm_lookaside *send_record_lookaside;
  struct gm_lookaside *enqueued_send_lookaside;

  unsigned int this_node_id;
  struct gm_port *port;
  int board;
  struct length_record
  {
    struct length_record *next;
    unsigned long length;
  } *length_stack;
  enum message_type basic_type;
  
  /****
   * handles
   ****/

  send_func_t *_send;
  void (*_sent) (struct gm_port *p, void *c);
  void (*fill_buffer) (struct test *t, gm_u8_n_t *buf, unsigned int len,
		       char seed);
  gm_status_t (*build_length_stack) (struct test *t, unsigned int);
  recv_func_t *recv;
  void (*start_sending) (struct test *, unsigned int s, unsigned long l);
  gm_status_t (*test_do) (struct test *t, unsigned int size);
  int (*verify_buffer) (struct test *t, gm_u8_n_t *buf, unsigned int len);
  void (*record_point) (struct test *t, unsigned long len, unsigned int cnt,
			gm_u64_t time);
  void (*report) (struct test *t, unsigned long len);
  gm_recv_event_t *(*gm_receive) (struct gm_port *port);
  void (*provide_recv_buffer_with_tag)
       (struct gm_port *p, void *ptr, unsigned size, unsigned priority,
	unsigned int tag);
} test_t;

typedef struct send_record
{
  struct send_record *next;
  test_t *test;
  message_ref_t *message_ref;
  unsigned int size;
  unsigned long len;
} send_record_t;

/****************************************************************
 * global data
 ****************************************************************/

char remote_host_parm[256];	/* arg of --remote-host or -h  */
char output_file_parm[256];	/* arg of --output or -o       */


/****************************************************************
 * utility functios
 ****************************************************************/

#if !GM_OS_VXWORKS
/* Microsoft compiler can't convert u64 to double */
static double
u64_to_double (gm_u64_t u64)
{
#ifdef WIN32
  return 2.0 * (double) ((gm_s64_t) (u64 >> 1)) + (int) (u64 & 0x01);
#else
  return (double) u64;
#endif
}
#endif

/****************************************************************
 * Auto-cleanup support functions
 ****************************************************************/

/* like gm_on_exit(), only it cleans up and exits on failure */

static void
on_exit_must_succeed (gm_on_exit_callback_t callback, void *arg)
{
  gm_status_t status;

  GM_CALLED ();
  status = gm_on_exit (callback, arg);
  if (status != GM_SUCCESS)
    {
      (*callback) (status, arg);
      gm_exit (status);
    }
  GM_RETURN_NOTHING ();
}

/* gm_on_exit() callback to close a file */

static void
fclose_callback (gm_status_t status, FILE *file)
{
  fclose (file);
}

/* gm_on_exit() callback to close a stats socket */

static void
close_stats_socket_callback (gm_status_t status, int *fd_ptr)
{
  if (GM_ENABLE_STATS_SOCKET)
    {
      stats_socket_close (*fd_ptr);
    }
}

/****************************************************************
 * Message References
 *
 * Message refs are created at init time and kept in a sort of
 * lookaside list, from which they can be alloc's and to which they
 * can be freed.
 ****************************************************************/

static message_ref_t *
_create_message_ref (test_t *t, unsigned int size, unsigned long buf_len)
{
  message_t *message;
  message_ref_t *ref;
  struct gm_port *port = t->port;
  unsigned long len;
  
  ref = gm_calloc (1, sizeof (message_ref_t));
  if (!ref)
    {
      fprintf (stderr, "memory allocation failed\n");
      goto abort_with_nothing;
    }

  /* Allocate enough memory to guarantee that it will include in it a
     buffer with the correct page offset */

  len = buf_len + GM_PAGE_LEN;
  ref->original_ptr = gm_dma_calloc (port, 1, len);
  if (!ref->original_ptr)
    {
      fprintf (stderr, "out of DMAable memory\n");
      goto abort_with_ref;
    }

  /* Compute message to be have page offset t->offset within the buffer */
  
  message = (message_t *) (((gm_size_t) ref->original_ptr
			    - PAGE_OFFSET (ref->original_ptr)
			    + t->offset));
  if ((gm_size_t) message < (gm_size_t) ref->original_ptr)
    {
      message = (message_t *) ((gm_size_t) message + GM_PAGE_LEN);
    }
  ref->message = message;
  
  ref->size = size;
  gm_always_assert (GM_DMA_ALIGNED (message));

  /* record the ref so we can deallocate it later */
  ref->prev_created = t->last_created_message_ref;
  t->last_created_message_ref = ref;
  
  return ref;

 abort_with_ref:
  gm_free (ref);
 abort_with_nothing:
  return 0;
}

static message_ref_t *
create_message_ref (test_t *t, unsigned int size)
{
  if (size == GM_ETHERNET_TAG_SIZE)
    {
      return _create_message_ref (t, GM_ETHERNET_TAG_SIZE, GM_ETHERNET_MTU);
    }
  else
    {
      gm_assert (size < 32);
      return _create_message_ref (t, size, gm_max_length_for_size (size));
    }
}

static void
destroy_all_message_refs (test_t *t)
{
  message_ref_t *ref;
  
  while (t->last_created_message_ref)
    {
      /* splice the reference out of the created list */
      
      ref = t->last_created_message_ref;
      t->last_created_message_ref = ref->prev_created;

      /* destory the referenced message and the reference */
      
      gm_dma_free (t->port, ref->original_ptr);
      gm_free (ref);
    }
}

static message_ref_t *
alloc_message_ref (test_t *t, unsigned int size)
{
  message_ref_t *ret;

  gm_assert (size < GM_NUM_ELEM (t->free_send_message_refs));
  ret = t->free_send_message_refs[size];
  if (ret)
    {
      t->free_send_message_refs[size] = ret->next;
      ret->next = 0;
      if (GM_DEBUG_ALLSIZE)
	{
	  gm_assert (ret->is_free);
	  ret->is_free = 0;
	}
    }
  return ret;
}

static void
free_message_ref (test_t *t, message_ref_t *ref)
{
  unsigned int size;
  message_ref_t **mrp;

  /* After a message is sent, add it to the free send buffer reference
     list */

  gm_assert (ref);
  gm_assert (t);
  size = ref->size;
  if (GM_DEBUG_ALLSIZE)
    {
      gm_always_assert (ref);
      gm_always_assert (!ref->is_free);
      ref->is_free = 1;
    }
  mrp = &t->free_send_message_refs[size];
  ref->next = *mrp;
  *mrp = ref;
}

/****************************************************************
 * Packet counts.
 ****************************************************************/

/* HACK: These are globals so handle_sigint can access them. */

static unsigned long total_queued;
static unsigned long total_sent;
static unsigned long total_recvd;

/* extra "status" parameter allows this to be used as a gm_on_exit()
   callback */

static void
print_packet_counts (gm_status_t status, void *ignored)
{
  GM_PARAMETER_MAY_BE_UNUSED (ignored);
  
  fprintf (stderr,
	   "\nTotal queued   = %lu.\n"
	     "Total sent     = %lu.\n"
	     "Total received = %lu.\n",
	   total_queued, total_sent, total_recvd);
}

/****************************************************************
 * Interrupt handling.
 ****************************************************************/

static int caught_sigint = 0;

static void sigint_exit (void)
{
  fprintf (stderr, "*** program interrupted by user ***\n");
  gm_exit (GM_INTERRUPTED);
}

static void
handle_sigint (int i)
{
  GM_PARAMETER_MAY_BE_UNUSED (i);
  caught_sigint = 1;
}

/****************************************************************
 * Command line processing
 ****************************************************************/

#define TP(a) fprintf (stderr, a)

static void
print_synopsis(void)
{
  TP ("\n");
  TP ("Myricom 'gm_allsize' verifies performance and correctness of\n");
  TP ("message-passing between specific ports on specific interfaces.\n");
  TP ("It operates between specified hosts, or (by default) in loopback\n");
  TP ("mode.  It optionally outputs data in a format ready for plotting by\n");
  TP ("the gnuplot utility program.\n");
}


static void
print_options(void)
{
  TP ("\nusage: gm_allsize [options]\n");

  TP ("\nGeneral Operation Mode Options:\n");
  TP ("-b, --bidirectional      [on]  bidirectional (ping pong) test\n");
  TP ("    --blocking           [off] use blocking receives\n");
  TP ("    --blocking-no-spin   [off] use immediately-blocking receives\n");
  TP ("    --both-ways          [off] stream data in both directions for BW\n");
  TP ("    --copy               [off] copy on send and receive side \n");
  TP ("    --datagrams          [off] always use datagram sends\n");
  TP ("    --datagrams=maybe    [off] randomly use datagram sends\n");
  TP ("    --enable-nack-down=NUM [off] control the sending of nack-down messages\n");
  TP ("-e, --exit-on-error      [off] exit when an error is detected\n");
  TP ("-f, --fast               [off] perform faster packet verification, if any\n");
  TP ("    --fork-child         [off] fork, then run as the child\n");
  TP ("    --fork-parent        [off] fork, then run as the parent\n");
  TP ("    --ignore-errors      [on]  continue when an error is detected\n");
  /* The '--no-versionstamp' option is non-orthogonal; e.g., there is no
   * '--no-timestamp' option.  This is intentional.  '--no-versionstamp' is
   * needed because gmperf.pl now emits '--versionstamp' by default, and we
   * want to enable users to at least *force* its prior default behavior.
   * OTOH, we don't want to go to the effort -- and clutter -- of making
   * negative counterparts to *every* positive gm_allsize option.            */
  TP ("    --no-versionstamp    [on]  suppress the GM-version-stamp from output\n");
  TP ("    --nonblocking        [on]  use nonblocking receives\n");
  TP ("    --offset=NUM         [0]   offset of message in send/recv buffer\n");
  TP ("    --quiet, --silent    [off] run test without reporting data points\n");
  if (GM_ENABLE_STATS_SOCKET)
    {
      TP ("    --socket-output=HOST:PORT\n"
	  "                         [none] output to remote host, not stdout\n");
    }
  TP ("    --timestamp          [off] timestamp the printed and gnuplot output\n");
  TP ("-u, --unidirectional     [off] unidirectional (spray) test\n");
  TP ("-v, --verify             [off] verify contents of all messages (slow)\n");
  TP ("    --versionstamp       [off] GM-version-stamp the printed and gnuplot output\n");

  TP ("\nDevice Control and Selection:\n");
  TP ("-B, --board=NUM         [0]   board to use in this machine\n");
  TP ("-p, --port=NUM          [2]   specifies local port number\n");
  TP ("-h, --remote-host=NAME  []    remote host (running in slave mode)\n");
  TP ("    --remote-port=NUM   [2]   specifies port number of remote slave\n");
  TP ("    --slave             [off] run in slave mode\n");

  TP ("\nData Set Generation:\n");
  TP ("-bw,  --bandwidth               [off] report bandwidth, not latency\n");
  TP ("      --bytes-per-second        [off] bandwidth in bytes/sec, not Mbytes/sec\n");
  TP ("      --count-per-length=NUM    [1]   number of messages to send per length\n");
  TP ("      --density=NUM             [3]   density of samples for geometric plot\n");
  TP ("-g,   --geometric               [off] geometric length progression\n");
  TP ("      --hysteresis              [off] perform up-down hysteresis check\n");
  TP ("-lat, --latency                 [on]  report latency, not bandwidth\n");
  TP ("      --length-increment=NUM    [1]   length increment\n");
  TP ("      --linear                  [on]  linear length progression\n");
  TP ("      --max-length=NUM          [0]   maximum length to test\n");
  TP ("      --max-size=NUM            [12]  maximum size to test\n");
  TP ("      --min-length=NUM          [1]   minimum length to test\n");
  TP ("      --min-sends-per-point=NUM [100] minimum # of sends per point\n");
  TP ("      --min-size=NUM            [12]  minimum size to test\n");
  TP ("      --min-usecs-per-point=NUM [5000] minimum test time in usecs\n");
  TP ("      --random-lengths          [off] random length progression\n");
  TP ("      --randomize               [off] shuffle the test lengths\n");
  TP ("      --reverse                 [off] perform test in reverse order\n");
  TP ("-s,   --size=NUM                [12]  alias for --min-size=NUM --max-size=NUM\n");

  TP ("\nGnuplot Options:\n");
  TP ("    --dots              [off] use dots instead of points in gnuplots\n");
  TP ("    --max-plot          [off] plot maximum value\n");
  TP ("    --min-plot          [off] plot minimum value\n");
  TP ("-o, --output=PATH[.gpl] []    output-file pathname.  If the name ends in \".gpl\",\n");
  TP ("                              generate gnuplot files\n");
  TP ("    --scatter-plot      [on]  scatter-plot all measured values\n");
  TP ("    --stddev-plot       [off] plot mean and standard deviation with errorbars\n");

  TP ("\nInformative Options:\n");
  TP ("    --help        output this usage information, then exit\n");
  TP ("\n");
  TP ("  By default, packets are sent to the local host.\n");
  fprintf (stderr, 
      "  GM sizes must be from %d to 31.\n",
	   gm_min_message_size ((struct gm_port *)0));
  TP ("\n");

  /* The following options are not displayed because they are deprecated.
   * Nevertheless, gm_allsize still honors them.
   *
   * --bps			Alias for --bytes-per-second
   * -c				Alias for --count-per-length
   * -l				Alias for --length-increment
   * --max-len			Alias for --max-length
   * --min-len			Alias for --min-length
   *
   * The following option was parsed, but had no effect; all support
   * has been removed as of GM-1.4.1pre10:
   * -q, -q -q			quiet and very quiet
   */

#undef TP

} /* print_options() */


static void
#ifdef va_dcl
usage (fmt, va_alist)
     const char *fmt;
     va_dcl
#else
usage (const char *fmt, ...)
#endif
{
  va_list ap;

  print_options();

/* Now print the caller's diagnostic */
#if (defined __STDC__) || (defined WIN32)
  va_start (ap, fmt);
#else
  va_start (ap);
#endif
  vfprintf (stderr, fmt, ap);
  va_end (ap);

  gm_exit (GM_INVALID_PARAMETER);

} /* usage() */

/****************************************************************
 * Packet content verification
 ****************************************************************/

static
char
random_seed (void)
{
  char ret;
  do
    {
      ret = gm_rand() & 0xff;
    }
  while (ret == 0);
  return ret;
}

/* don't fill or verify */

static void
fill_buffer_not (test_t *t, gm_u8_n_t *buf, unsigned int len, char seed)
{
  GM_CALLED ();
  if (t->do_copy)
    {
      char *copybuf;
      copybuf = gm_malloc (len);
      if (copybuf) 
	{
	  gm_bcopy(buf,copybuf,len);
	  gm_free (copybuf);
	}
    }

  GM_RETURN_NOTHING ();
}

static int
verify_buffer_not (test_t *t, gm_u8_n_t *buf, unsigned int len)
{
  GM_CALLED ();
  if (t->do_copy)
    {
      char *copybuf;
      copybuf = gm_malloc (len);
      if (copybuf) 
        {
          gm_bcopy(buf,copybuf,len);
          gm_free (copybuf);
        }
    }

  GM_RETURN (1);
}

/* simple buffer fill and verify routines */

static void
fill_buffer_fast (test_t *t, gm_u8_n_t *buf, unsigned int len, char seed)
{
  unsigned int i;

  GM_CALLED ();
  
  buf[0] = gm_hton_u8 (seed);

  for (i = 1; i < len; i++)
    buf[i] = gm_hton_u8 ((gm_u8_t) (i & 0xFF));

  if (t->do_copy)
    {
      char *copybuf;
      copybuf = gm_malloc (len);
      if (copybuf) 
        {
          gm_bcopy(buf,copybuf,len);
          gm_free (copybuf);
        }
    }

  GM_RETURN_NOTHING ();
}

static int
verify_buffer_fast (test_t *t, gm_u8_n_t *buf, unsigned int len)
{
  unsigned int i;
  unsigned char val;

  GM_CALLED ();
  for (i = 1; i < len; i++)
    {
      val = i&0xFF; /*seed;*/
      if ((gm_ntoh_u8 (buf[i]) & 0xff) != (val & 0xff))
	{
	  printf ("data error; at pos %d expected 0x%x and got "
		  "0x%x\n"
		  "            memory location %p\n",
		  i, val, gm_ntoh_u8 (buf[i]), &buf[i]);

	  if ( !GM_LONG_PACKET_ERROR_PRINT )
	    {
	      if ((i+8) < len)
		{
		  printf ("            next 8 bytes"
			  " %02x %02x %02x %02x %02x %02x %02x %02x \n",
			  gm_ntoh_u8 (buf[i+1]), gm_ntoh_u8 (buf[i+2]),
			  gm_ntoh_u8 (buf[i+3]), gm_ntoh_u8 (buf[i+4]),
			  gm_ntoh_u8 (buf[i+5]), gm_ntoh_u8 (buf[i+6]),
			  gm_ntoh_u8 (buf[i+7]), gm_ntoh_u8 (buf[i+8]));
		}
	      GM_RETURN (1);
	    }
	  else /* GM_LONG_PACKET_ERROR_PRINT */
	    {
	      unsigned int j;
	      
	      /* if we got an error, print out the entire buffer */
	      printf ("Received:\n");
	      for (j = 1; j < len; j++)
		{
		  printf ("%02x ", gm_ntoh_u8 (buf[j]));
		  if( (j % 16) == 0 ) printf("\n");
		}
	      printf ("\nExpected:\n");
	      for (j = 1; j < len; j++)
		{
		  printf("%02x ", (unsigned char) (j & 0xff));
		  if( (j % 16) == 0 ) printf("\n");
		}
	      printf ("\n");
	      GM_RETURN (1);	
	    }
	  break;
	}
    }

  if (t->do_copy)
    {
      char *copybuf;
      copybuf = gm_malloc (len);
      if (copybuf) 
        {
          gm_bcopy(buf,copybuf,len);
          gm_free (copybuf);
        }
    }
  
  GM_RETURN (0);
}

static void
fill_buffer_thorough (test_t *t, gm_u8_n_t *buf, unsigned int len, char seed)
{
  unsigned int i;

  GM_CALLED ();
  
  gm_srand (seed);

  buf[0] = gm_hton_u8 (seed);
  
  for (i = 1; i < len; i++)
    {
      buf[i] = gm_hton_u8 ((gm_u8_t) (gm_rand_mod (255) + 1));
    }
  
  if (t->do_copy)
    {
      char *copybuf;
      copybuf = gm_malloc (len);
      if (copybuf) 
        {
          gm_bcopy(buf,copybuf,len);
          gm_free (copybuf);
        }
    }

  GM_RETURN_NOTHING ();
}

static int
verify_buffer_thorough (test_t *t, gm_u8_n_t *buf, unsigned int len)
{
  unsigned int i = 0, rv=0;
  char val;
  char *pattern;
  char seed = gm_ntoh_u8 (buf[0]);

  GM_CALLED ();
  
  gm_srand (seed);

  pattern = gm_malloc (len);
  if (!pattern)
    {
      printf ("verify_buffer_thorough:\n"
	      "Can't gm_malloc a len = %d buffer for the pattern\n", len);
      GM_RETURN (1);
    }

  for (i = 1; i < len; i++)
    pattern[i] = gm_rand_mod (255) + 1;

  for (i = 1; i < len; i++)
    {
      val = pattern[i];
      if ((gm_ntoh_u8 (buf[i]) & 0xff) != (val & 0xff))
	{
	  rv = 1;
	  printf ("data error; msg %lu at pos %d of %d expected 0x%x"
		  " and got 0x%x\n"
		  "            memory location %p\n",
		  total_recvd, i, len, val&0xff, gm_ntoh_u8 (buf[i]) & 0xff,
		  &buf[i]);

	  if (!GM_LONG_PACKET_ERROR_PRINT)
	    {
	      if ((i+8) < len)
		{
		  printf ("            next 8 bytes"
			  " %02x %02x %02x %02x %02x %02x %02x %02x \n",
			  gm_ntoh_u8 (buf[i+1]), gm_ntoh_u8 (buf[i+2]),
			  gm_ntoh_u8 (buf[i+3]), gm_ntoh_u8 (buf[i+4]),
			  gm_ntoh_u8 (buf[i+5]), gm_ntoh_u8 (buf[i+6]),
			  gm_ntoh_u8 (buf[i+7]), gm_ntoh_u8 (buf[i+8]));
		  printf ("        expected 8 bytes"
			  " %02x %02x %02x %02x %02x %02x %02x %02x \n",
			  (gm_u8_t) pattern[i+1], (gm_u8_t) pattern[i+2],
			  (gm_u8_t) pattern[i+3], (gm_u8_t) pattern[i+4],
			  (gm_u8_t) pattern[i+5], (gm_u8_t) pattern[i+6],
			  (gm_u8_t) pattern[i+7], (gm_u8_t) pattern[i+8]);
		}
	    }
	  else /* GM_LONG_PACKET_ERROR_PRINT */
	    {
	      unsigned int j;
	      
	      /* if we got an error, print out the entire buffer, flagging the
	         discrepencies. */
	      printf ("Received:\n");
	      for (j = 0; j < len; j++)
		{
		  printf (gm_ntoh_u8 (buf[j]) != (unsigned char) pattern[j]
			  ? "%02x'"
			  : "%02x ",
			  gm_ntoh_u8 (buf[j]));
		  if( ((j+1) % 16) == 0 ) printf("\n");
		}
	      printf ("\nExpected:\n");
	      for (j = 0; j < len; j++)
		{
		  printf (gm_ntoh_u8 (buf[j]) != (unsigned char) pattern[j]
			  ? "%02x'"
			  : "%02x ",
			  (unsigned char) pattern[j]);
		  if( ((j+1) % 16) == 0 ) printf("\n");
		}
	      printf ("\n");
	    }
	  break;
	}
    }

  if (t->do_copy)
    {
      char *copybuf;
      copybuf = gm_malloc (len);
      if (copybuf) 
        {
          gm_bcopy(buf,copybuf,len);
          gm_free (copybuf);
        }
    }

  if (pattern)
    gm_free (pattern);
  GM_RETURN (rv);
}

static void
fill_message (test_t *t, message_t *m, unsigned long len,
	      enum message_type basic_type)
{
  switch (len)
    {
    case 0:
      fprintf (stderr, "cannot fill zero-length packets");
      if (t->exit_on_error)
	{
	  gm_exit (GM_INTERNAL_ERROR);
	}
      break;

    case 1:
      m->small.type = gm_hton_u8 ((gm_u8_t) (basic_type + 1));
      break;

    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
      m->small.type
	= gm_hton_u8 ((gm_u8_t) (basic_type + (unsigned char) len));
      t->fill_buffer (t, m->small.data,
		      len - GM_OFFSETOF (message_t, small.data),
		      random_seed ());
      break;

    case 8:
      m->large.type = gm_hton_u8 ((gm_u8_t) (basic_type + 8));
      t->fill_buffer (t, m->large.pad, sizeof (m->large.pad),
		      (char)gm_rand ());
      m->large.length = gm_hton_u32 (8);
      break;

    default:
      m->large.type = gm_hton_u8 ((gm_u8_t) basic_type);
      t->fill_buffer (t, m->large.pad, sizeof (m->large.pad),
		      (char)gm_rand ());
      m->large.length = gm_hton_u32 (len);
      t->fill_buffer (t, m->large.data, len, (char)gm_rand ());
      if (GM_DEBUG_ALLSIZE && (MESSAGE_LEN (m) != len))
	{
	  fprintf
		(stderr,
		"Packet before send msg len 0x%x does not match length request"
		" 0x%lx\n",
		MESSAGE_LEN (m), len);
	}
      else 
	{
/*
	  fprintf
		(stderr,
		"Packet before send msg len OK 0x%x "
		" 0x%lx\n", MESSAGE_LEN (m), len);
*/
	}
      break;
    }
}

/* BAD: should check for legit types here and make sure the length passed
   in here is form the event, not the packet. */

static int
verify_message (test_t *t, message_t *m, unsigned long len)
{
  int error = 0;

  if (MESSAGE_LEN (m) != len)
    {
      fprintf
	(stderr,
	 "Packet message length 0x%x does not match length reported"
	 " by GM 0x%lx\n",
	 MESSAGE_LEN (m), len);
      gm_hex_dump (m, len);
      return 1;
    }
  
  switch (MESSAGE_LEN (m))
    {
    case 0:
    case 1:
    case 2:
      break;

    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
      error = t->verify_buffer (t, m->small.data,
				len - GM_OFFSETOF (message_t, small.data));
      break;

    case 8:
      error = t->verify_buffer (t, m->large.pad, sizeof (m->large.pad));
      break;

    default:
      error = (t->verify_buffer (t, m->large.pad, sizeof(m->large.pad))
	       || t->verify_buffer (t, m->large.data,
				    len - GM_OFFSETOF (message_t,
						       large.data)));
      break;
    }
  if (error && t->exit_on_error)
    {
      gm_exit (GM_FAILURE);
    }
  return error;
}

/****************************************************************
 * Gnuplot support
 ****************************************************************/

#define GNUPLOT_EXT ".gpl"

static
int
#ifdef va_dcl
gnuplot_printf (t, fmt, va_alist)
     test_t *t;
     const char *fmt;
     va_dcl
#else
gnuplot_printf (test_t *t, const char *fmt, ...)
#endif
{
  int ret;
  va_list ap;

  if (t->gnuplot.file == 0)
    return 0;
  
#if (defined __STDC__) || (defined WIN32)
  va_start (ap, fmt);
#else
  va_start (ap);
#endif

  ret = vfprintf (t->gnuplot.file, fmt, ap);
  va_end (ap);
  return ret;
}

static
FILE *
gnuplot_open (test_t *t, const char *filename)
{
  size_t basename_len;

  GM_CALLED ();
  
  /* open the file for output */
  
  t->gnuplot.file = fopen (filename, "w");
  if (!t->gnuplot.file)
    {
      fprintf (stderr, "Could not open gnuplot file \"%s\""
	       " for writing", filename);
      perror (" ");
      gm_exit (GM_FAILURE);
    }

  /* build the basename for the filename */
  
  basename_len = strlen (filename) - strlen (GNUPLOT_EXT);
  t->gnuplot.basename = gm_malloc (basename_len + 1);
  gm_always_assert (t->gnuplot.basename);
  strncpy (t->gnuplot.basename, filename, basename_len);
  t->gnuplot.basename[basename_len] = 0;

  GM_RETURN (t->gnuplot.file);
}

static
void
gnuplot_start_size (test_t *t, unsigned int size, int reversed)
{
  char *filename, *new_plot_string;
  const char *_reversed = "_reversed";

  GM_CALLED ();
  
  if (!t->gnuplot.file)
    GM_RETURN_NOTHING ();
  
  gm_always_assert (size <= 99);
  filename = gm_malloc (strlen (t->gnuplot.basename) + 1 + 2
			+ strlen (_reversed));
  sprintf (filename, "%s_%u%s", t->gnuplot.basename, size,
	   reversed ? _reversed : "");
  t->data_file = fopen (filename, "w");
  if (!t->data_file)
    {
      fprintf (stderr, "could not open data output file \"%s\"",
	       filename);
      perror (" ");
      gm_exit (GM_FAILURE);
    }
  
  if (t->gnuplot.plot_string)
    {
      new_plot_string = (char *) gm_malloc (strlen (t->gnuplot.plot_string)
					    + 3 + strlen (filename) + 2);
      gm_always_assert (new_plot_string);
      sprintf (new_plot_string, "%s, \"%s\"", t->gnuplot.plot_string,
	       filename);
      gm_free (t->gnuplot.plot_string);
    }
  else
    {
      const char *plot = "plot \"";
      
      new_plot_string = (char *) gm_malloc (strlen (plot)
					    + strlen (filename)
					    + 2);
      sprintf (new_plot_string, "%s%s\"", plot, filename);
    }
  t->gnuplot.plot_string = new_plot_string;
  
  gm_free (filename);
  fprintf (t->data_file, "# output for gnuplot\n");
  GM_RETURN_NOTHING ();
}

static void report_scatter (test_t *t, unsigned long len);
static void report_stddev (test_t *t, unsigned long len);
static void record_point_bandwidth (test_t *t, unsigned long len,
				    unsigned int cnt, gm_u64_t time);

static
void
gnuplot_close (gm_status_t status, test_t *t)
{
  GM_PARAMETER_MAY_BE_UNUSED (status);
  
  if (!t->gnuplot.file)
    return;

  if (t->record_point == record_point_bandwidth)
    {
      fprintf (t->gnuplot.file, "set nologscale y\n");
    }
  fprintf (t->gnuplot.file, "set terminal x11\n");
  /* fprintf (t->gnuplot.file, "set output STDOUT\n"); */
  fprintf (t->gnuplot.file, "set grid\n");
  if ((t->timestamps) && (t->version))
    fprintf (t->gnuplot.file,
	     "set title \"%s -- GM %s\\n%s %s\"\n",
	     t->time_string, _gm_build_id, t->title, t->ylabel);
  else if (t->timestamps)
    fprintf (t->gnuplot.file,
	     "set title \"%s\\n%s %s\"\n",
	     t->time_string, t->title, t->ylabel);
  else if (t->version)
    fprintf (t->gnuplot.file,
	     "set title \"GM %s\\n%s %s\"\n",
	     _gm_build_id, t->title, t->ylabel);
  else
    fprintf (t->gnuplot.file, "set title \"%s %s\"\n", t->title, t->ylabel);
  fprintf (t->gnuplot.file, "set xlabel \"Message Length (bytes)\"\n");
  fprintf (t->gnuplot.file, "set ylabel \"%s\"\n", t->ylabel);
  fprintf (t->gnuplot.file, "set key below\n");
  if (t->report == report_stddev)
    fprintf (t->gnuplot.file, "set data style yerrorbars\n");
  else
    fprintf (t->gnuplot.file, "set data style %s\n", t->points);

  /* output command to plot all files */

  if (t->gnuplot.plot_string)
    {
      fputs (t->gnuplot.plot_string, t->gnuplot.file);
      gm_free (t->gnuplot.plot_string);
      t->gnuplot.plot_string = 0;
    }
  
  gnuplot_printf (t, "\npause -1 \"Hit return to continue\"\n");

  /* close the gnuplot file */

  fflush (t->gnuplot.file);
  fclose (t->gnuplot.file);
  if (t->data_file != NULL)
    fflush (t->data_file);
  t->gnuplot.file = 0;
}


/****************************************************************
 * data output
 ****************************************************************/

#if 0
/* function to print both to stdout and to any open output file. */
/* This function and its use excited some ppc-linux bugs. 3/20/3000 feldy */
static
int
#ifdef va_dcl
printf_both (t, fmt, va_alist)
     test_t *t;
     const char *fmt;
     va_dcl
#else
printf_both (test_t *t, const char *fmt, ...)
#endif
{
  int ret;
  va_list ap;

#if (defined __STDC__) || (defined WIN32)
  va_start (ap, fmt);
#else
  va_start (ap);
#endif

  if (t->data_file)
    {
      vfprintf (t->data_file, fmt, ap);
      fflush (t->out);
    }
  ret = vfprintf (stdout, fmt, ap);
  fflush (stdout);
  gm_sleep (0);
  va_end (ap);
  return ret;
}
#endif

/****************
 * record_point_...
 ****************/

/* handlers for recording the output of a single timed test */

static
void
record_point_latency (test_t *t, unsigned long len, unsigned int cnt,
		      gm_u64_t time)
{
  /*GM_PARAMETER_MAY_BE_UNUSED (len);*/
  
#if GM_OS_VXWORKS
  /* vxworks can't do most floating point or 64 bit math
     so we do a little magic to give vxworks meaningful numbers */
  const unsigned scale_factor = 100;
  unsigned long ltime;
  
  ltime = (unsigned long) time;
#endif /* GM_OS_VXWORKS */
  
  if (t->ping_pong && (t->target_node_id != t->this_node_id))
    {
      static int noted;
      
      if (!noted)
	{
	  printf ("# (reporting half round trip time)\n");
	  noted = 1;
	}
#if GM_OS_VXWORKS
      stats_append (t->stats,
		    (double)(scale_factor * ltime / cnt / 2) / scale_factor);
#else /*!GM_OS_VXWORKS*/
      stats_append (t->stats, u64_to_double (time) / (double) (cnt) / 2.0);
#endif /*!GM_OS_VXWORKS*/
      
    }
  else
    {
#if GM_OS_VXWORKS
      stats_append (t->stats,
		    (double)(scale_factor * ltime / cnt ) / scale_factor
		    );
#else /*!GM_OS_VXWORKS*/
      stats_append (t->stats, u64_to_double (time) / (double) (cnt));
#endif /*!GM_OS_VXWORKS*/
    }
}

static
void
record_point_bandwidth (test_t *t, unsigned long len, unsigned int cnt,
			gm_u64_t time)
{
#if GM_OS_VXWORKS
  /* vxworks can't do most floating point or 64 bit math
     so we do a little magic to give vxworks meaningful numbers */
  unsigned long ltime;	
  const unsigned long scale_factor = 10;

  ltime = (unsigned long) time;
  
  stats_append (t->stats,
		(double)( scale_factor * cnt * len / ltime ) / scale_factor );
#else /* !GM_OS_VXWORKS */  
  if (time == 0)
    time = 1;

  if (t->report_bytes_per_sec)
    {
      stats_append (t->stats,
		    ((double)1e6 *  (double) (cnt) * (double) (len) / 
		     u64_to_double (time)));
    }
  else 
    {
      stats_append (t->stats,
		    ((double) (cnt) * (double) (len) / u64_to_double (time)));
    }
#endif
}
 
/****************
 * report_...
 ****************/


/* handlers for reporting a summary of earlier record data points */

static void
report_min (test_t *t, unsigned long len)
{
  if (t->data_file)
    {
      fprintf(t->data_file, "%lu %f\n", len, stats_get_min (t->stats));
    }
  fprintf(stdout, "%lu %f\n", len, stats_get_min (t->stats));
  if (t->stats_socket)
    {
      stats_socket_printf (t->stats_socket, "%lu %f\n", len,
			   stats_get_min (t->stats));
    }
}

static void
report_max (test_t *t, unsigned long len)
{
  if (t->data_file)
    {
      fprintf(t->data_file, "%lu %f\n", len, stats_get_max (t->stats));
    }
  fprintf(stdout, "%lu %f\n", len, stats_get_max (t->stats));
  if (t->stats_socket)
    {
      stats_socket_printf (t->stats_socket, "%lu %f\n", len,
			   stats_get_max (t->stats));
    }
}

static void
report_stddev (test_t *t, unsigned long len)
{
  if (t->data_file)
    {
      fprintf(t->data_file, "%lu %f %f\n", len, stats_get_mean (t->stats),
	      stats_get_standard_deviation (t->stats));
    }
  fprintf(stdout, "%lu %f %f\n", len, stats_get_mean (t->stats),
	  stats_get_standard_deviation (t->stats));
  if (t->stats_socket)
    {
      stats_socket_printf (t->stats_socket, "%lu %f %f\n", len,
			   stats_get_mean (t->stats),
			   stats_get_standard_deviation (t->stats));
    }
}

static void
report_scatter (test_t *t, unsigned long len)
{
  char fmt[20];
  sprintf (fmt, "%lu %%.2f\n", len);
  stats_output_points (stdout, fmt, t->stats);

  if (t->stats_socket)
    {
      stats_socket_output_points (t->stats_socket, fmt, t->stats);
    }
  
  if (t->data_file)
    stats_output_points (t->data_file, fmt, t->stats);
}

static void
report_nothing (test_t *t, unsigned long len)
{
  GM_PARAMETER_MAY_BE_UNUSED (t);
  GM_PARAMETER_MAY_BE_UNUSED (len);
  
  return;
}

/****************
 * label_output
 ****************/

/* function to output a comment describing the data that follows
   in the output file, and by the same token to specify title and ylabel
   strings for the gnuplot (if any) */

static
void
label_output (test_t *t, unsigned int size, int reversed)
{
  GM_PARAMETER_MAY_BE_UNUSED (reversed);

  GM_CALLED ();
  if (t->report == report_nothing)
    GM_RETURN_NOTHING ();
  
  if (t->this_node_id == t->target_node_id)
    {
      if (t->ping_pong)
	{
	  if (t->record_point == record_point_bandwidth)
	    {
	      t->title = "Reliable Loopback Ping-Pong";
	      if (t->report_bytes_per_sec)
		t->ylabel = "Bandwidth (Byte/s)";
	      else
		t->ylabel = "Bandwidth (MByte/s)";
	    }
	  else
	    {
	      t->title = "Reliable Loopback";
	      t->ylabel = "Latency (usec)";
	    }
	}
      else
	{
	  if (t->record_point == record_point_bandwidth)
	    {
	      t->title = "Reliable Loopback";
	      if (t->report_bytes_per_sec)
		t->ylabel = "Bandwidth (Byte/s)";
	      else
		t->ylabel = "Bandwidth (MByte/s)";
	    }
	  else
	    {
	      t->title = "Reliable Loopback";
	      t->ylabel = "Gap (usec)";
	    }
	}
    }
  else
    {
      if (t->ping_pong)
	{
	  if (t->record_point == record_point_bandwidth)
	    {
	      if (t->both_ways)
		t->title = "Reliable Point-to-Point Bidirectional Summed";
	      else
		t->title = "Reliable Point-to-Point Ping-Pong";

	      if (t->report_bytes_per_sec)
		t->ylabel = "Bandwidth (Byte/s)";
	      else
		t->ylabel = "Bandwidth (MByte/s)";
	    }
	  else
	    {
	      t->title = "Reliable Point-To-Point Half-Round-Trip";
	      t->ylabel = "Latency (usec)";
	    }
	}
      else
	{
	  if (t->record_point == record_point_bandwidth)
	    {
	      t->title = "Reliable Point-To-Point Unidirectional";
	      if (t->report_bytes_per_sec)
		t->ylabel = "Bandwidth (Byte/s)";
	      else
		t->ylabel = "Bandwidth (MByte/s)";
	    }
	  else
	    {
	      t->title = "Reliable Point-To-Point";
	      t->ylabel = "Gap (usec)";
	    }
	}
    }

  if (t->timestamps)
    {
      time_t uct;

      uct = time(NULL);
      t->time_string = asctime(localtime(&uct));
      /* overwrite the final NL character */
      if (t->time_string[strlen(t->time_string) - 1] == '\n')
	t->time_string[strlen(t->time_string) - 1] = '\0';
      if (t->data_file)
	fprintf(t->data_file, "# Test beginning %s\n", t->time_string);
      fprintf(stdout, "# Test beginning %s\n", t->time_string);
    }
  if (t->version)
    {
      if (t->data_file)
	fprintf(t->data_file, "# GM build ID is \"%s.\"\n", _gm_build_id);
      fprintf(stdout, "# GM build ID is \"%s.\"\n", _gm_build_id);
    }
  if (t->data_file)
    fprintf(t->data_file, "# %s %s\n", t->title, t->ylabel);
  fprintf(stdout, "# %s %s\n", t->title, t->ylabel);

  if (t->data_file)
    fprintf(t->data_file, "# size %d\n", size);
  fprintf(stdout, "# size %d\n", size);
  GM_RETURN_NOTHING ();
}

/****************************************************************
 * Receive buffer provision functions
 ****************************************************************/

/* Special function to provide ethernet receive buffers.  Normally,
   the buffer is simply passed to the LANai with
   gm_provide_receive_buffer_with_tag(), but for ethernet buffers we
   must build a scatter list for the receive buffer and pass that to
   _gm_provide_ethernet_scatter_list() instead. */

static void
reprovide_ethernet_scatter_list (struct gm_port *p, void *ptr, unsigned size,
				 unsigned priority, unsigned int tag)
{
  gm_ethernet_scatter_list_t list;
  gm_ethernet_segment_descriptor_n_t *from_seg;
  gm_ethernet_segment_descriptor_t *to_seg;
  
  to_seg = &list[0];
  for (from_seg = (gm_ethernet_segment_descriptor_n_t *) ptr;
       gm_ntoh_dp (from_seg->ptr);
       from_seg++)
    {
      to_seg->ptr = gm_ntoh_dp (from_seg->ptr);
      to_seg->len = gm_ntoh_u32 (from_seg->len);
      to_seg++;
    }

  gm_assert (GM_ETHERNET_MTU / GM_PAGE_LEN < 3);
  gm_ethernet_provide_scatter_list_with_tag (p, to_seg - &list[0], list,
					     GM_ARBITRARY_RECV_BUFFER_TAG);
}

static void
provide_ethernet_receive_buffer (struct gm_port *p, void *ptr, unsigned size,
				 unsigned priority, unsigned int tag)
{
  gm_dp_t dma_addr;
  gm_ethernet_scatter_list_t scatter_list;
  gm_ethernet_segment_descriptor_t *segment = &scatter_list[0];
  gm_status_t status;
  unsigned int remaining = GM_ETHERNET_MTU, page_remaining;

  GM_CALLED_WITH_ARGS (("%p, %p, %u, %u, %u", p, ptr, size, priority, tag));
  
  GM_PARAMETER_MAY_BE_UNUSED (priority);
  GM_PARAMETER_MAY_BE_UNUSED (tag);

  gm_assert (size == GM_ETHERNET_TAG_SIZE);

  /* build the scatter list into an array */

  remaining = GM_ETHERNET_MTU;
  page_remaining = PAGE_REMAINING (ptr);
  while (remaining)
    {
      status = gm_dma_addr (p, (gm_up_t) ptr, &dma_addr);
      gm_always_assert (status == GM_SUCCESS);
      segment->ptr = dma_addr;
      if (remaining <= page_remaining)
	{
	  segment->len = remaining;
	  segment++;
	  break;
	}
      segment->len = page_remaining;
      gm_assert (PAGE_OFFSET (segment->ptr + segment->len) == 0);
      segment++;
      remaining -= page_remaining;
      ptr = (void *) ((gm_up_t) ptr + page_remaining);
      page_remaining = GM_PAGE_LEN;
    }

  /* provide the scatter list to GM */

  gm_ethernet_provide_scatter_list_with_tag
    (p, segment - &scatter_list[0], &scatter_list[0],
     GM_ARBITRARY_RECV_BUFFER_TAG);

  GM_RETURN_NOTHING ();
}

/****************************************************************
 * recv handlers
 ****************************************************************/

static void
recv_only (test_t *t, struct gm_port *port, gm_recv_event_t *e,
	   send_func_t send)
{
  void *buffer;
  unsigned int size;
  GM_PARAMETER_MAY_BE_UNUSED (t);
  GM_PARAMETER_MAY_BE_UNUSED (send);

  GM_CALLED ();

  gm_assert (gm_ntoh_u8 (e->recv.tag) == GM_ARBITRARY_RECV_BUFFER_TAG);
  
  buffer = gm_ntohp (e->recv.buffer);
  size = gm_ntoh_u8 (e->recv.size);
  t->provide_recv_buffer_with_tag
    (port, buffer, size, GM_LOW_PRIORITY, GM_ARBITRARY_RECV_BUFFER_TAG);
  total_recvd++;
  t->recv_cnt++;

  GM_RETURN_NOTHING ();
}

static void
recv_and_verify (test_t *t, struct gm_port *port, gm_recv_event_t *e,
		 send_func_t send)
{
  void *message;
  int rv = 0;

  GM_CALLED ();
  
  GM_PARAMETER_MAY_BE_UNUSED (send);
  
  message = gm_ntohp (e->recv.buffer);
  rv = verify_message (t, message, gm_ntoh_u32 (e->recv.length));
  if (rv && t->exit_on_error)
    gm_exit (GM_FAILURE);

  recv_only (t, port, e, _send_abort);
  
  GM_RETURN_NOTHING ();
}

static void
send (test_t *t, unsigned int size,
      unsigned long len, enum message_type basic_type,
      unsigned int target_node_id,
      unsigned int target_port_id,
      send_func_t send_func);

static void
recv_and_reply (test_t *t, struct gm_port *port, gm_recv_event_t *e,
		send_func_t send_func)
{
  void (*fill_buffer_save) (struct test *, gm_u8_n_t *, unsigned int , char );

  GM_CALLED ();
  
  /* disable buffer filling for the reply, since the default
     "t->fill_buffer" does fill the buffer. */
  
  fill_buffer_save = t->fill_buffer;
  t->fill_buffer = fill_buffer_not;
  send (t, gm_ntoh_u8 (e->recv.size), gm_ntoh_u32 (e->recv.length),
	t->basic_type, gm_ntoh_u16 (e->recv.sender_node_id),
	gm_ntoh_u8 (e->recv.sender_port_id),
	send_func);
  t->fill_buffer = fill_buffer_save;
  
  recv_only (t, port, e, _send_abort);

  GM_RETURN_NOTHING ();
}

static void
recv_and_reply_and_verify (test_t *t, struct gm_port *port, gm_recv_event_t *e,
			   send_func_t send_func)
{
  unsigned int len;
  
  GM_CALLED ();
  
  len = gm_ntoh_u32 (e->recv.length);
  
  recv_and_verify (t, port, e, _send_abort);
  
  send (t, gm_ntoh_u8 (e->recv.size), len,
	t->basic_type, gm_ntoh_u16 (e->recv.sender_node_id),
	gm_ntoh_u8 (e->recv.sender_port_id), send_func);

  GM_RETURN_NOTHING ();
}

static void
recv_slave (test_t *t, struct gm_port *port, gm_recv_event_t *e,
	    send_func_t ignored_send_func)
{
  message_t *message;
  enum message_type type;
  GM_PARAMETER_MAY_BE_UNUSED (ignored_send_func);
  
  GM_CALLED ();
  
  PRINTF (("slave received following event\n"
	   "  message = %p\n"
	   "  buffer = %p\n"
	   "  length = %u\n"
	   "  sender_node_id = %u\n"
	   "  tag %u\n"
	   "  size %u\n"
	   "  sender_port_id = %u\n",
	   gm_ntohp (e->recv.message),
	   gm_ntohp (e->recv.buffer),
	   gm_ntoh_u32 (e->recv.length),
	   (unsigned int) gm_ntoh_u16 (e->recv.sender_node_id),
	   (unsigned int) gm_ntoh_u8 (e->recv.tag),
	   (unsigned int) gm_ntoh_u8 (e->recv.size),
	   (unsigned int) gm_ntoh_u8 (e->recv.sender_port_id)));
  
  message = (message_t *) gm_ntohp (e->recv.buffer);
  type
    = (enum message_type) (message->type
			   & (TYPE_SEND_MODE_MASK | TYPE_REPLY | TYPE_VERIFY));
  PRINTF (("type is 0x%x\n", (int) type));
  
  switch (type)
    {
    case TYPE_SEND_MODE_NONPEER:
    case TYPE_SEND_MODE_PEER:
    case TYPE_SEND_MODE_DATAGRAM:
    case TYPE_SEND_MODE_DATAGRAM_OR_RELIABLE:
      recv_only (t, port, e, _send_abort);
      break;

      /* reply */
      
    case TYPE_REPLY + TYPE_SEND_MODE_NONPEER:
      recv_and_reply (t, port, e, _send_to_nonpeer);
      break;
    case TYPE_REPLY + TYPE_SEND_MODE_PEER:
      recv_and_reply (t, port, e, _send_to_peer);
      break;
    case TYPE_REPLY + TYPE_SEND_MODE_DATAGRAM:
      recv_and_reply (t, port, e, _send_datagram);
      break;
    case TYPE_REPLY + TYPE_SEND_MODE_DATAGRAM_OR_RELIABLE:
      recv_and_reply (t, port, e, _send_datagram_or_reliable);
      break;

      /* verify */

    case TYPE_VERIFY + TYPE_SEND_MODE_NONPEER:
    case TYPE_VERIFY + TYPE_SEND_MODE_PEER:
    case TYPE_VERIFY + TYPE_SEND_MODE_DATAGRAM:
    case TYPE_VERIFY + TYPE_SEND_MODE_DATAGRAM_OR_RELIABLE:
      recv_and_verify (t, port, e, _send_abort);
      break;

      /* reply and verify */

    case TYPE_REPLY + TYPE_VERIFY + TYPE_SEND_MODE_NONPEER:
      recv_and_reply_and_verify (t, port, e, _send_to_nonpeer);
      break;
      
    case TYPE_REPLY + TYPE_VERIFY + TYPE_SEND_MODE_PEER:
      recv_and_reply_and_verify (t, port, e, _send_to_peer);
      break;
      
    case TYPE_REPLY + TYPE_VERIFY + TYPE_SEND_MODE_DATAGRAM:
      recv_and_reply_and_verify (t, port, e, _send_datagram);
      break;

    case TYPE_REPLY + TYPE_VERIFY + TYPE_SEND_MODE_DATAGRAM_OR_RELIABLE:
      recv_and_reply_and_verify (t, port, e, _send_datagram_or_reliable);
      break;

    default:
      { 
	int i;
	unsigned char *cptr = (unsigned char *)message;
	printf ("receive packet of unrecognized type %u (0x%x)\n",
		(unsigned int) type, (unsigned int) type);
        for (i=0; i<16; i++) 
	  {
	    printf("%02x ",*cptr++);
	  }
	printf("\n");
      }
      break;
    }

  GM_RETURN_NOTHING ();
}

/****************************************************************
 * send completion handlers
 ****************************************************************/

void free_send_token (test_t *t);

/* handle send completion by either using the send token for an
   enqueued send or recycling it. */

static void
_sent_dont_send (struct gm_port *port, void *context)
{
  test_t *t;
  message_ref_t *ref;
  send_record_t *rec;
  GM_PARAMETER_MAY_BE_UNUSED (port);

  GM_CALLED ();
  
  /* free the resources for this send */
  
  rec = (send_record_t *) context;
  t = rec->test;
  ref = rec->message_ref;
  gm_assert (ref);
  gm_lookaside_free (rec);
  free_message_ref (t, ref);

  GM_RETURN_NOTHING ();
}

/* For each completed send in a unidirectional test, reenqueue the send */

static void
_sent__send_again (struct gm_port *port, void *context)
{
  send_record_t *rec;
  test_t *t;
  message_ref_t *ref;
  unsigned int size, target_node_id, target_port_id;
  unsigned long len;
  enum message_type basic_type;
  GM_PARAMETER_MAY_BE_UNUSED (port);

  GM_CALLED ();
  
  rec = (send_record_t *) context;
  t = rec->test;

  /* free the message ref so the send below will probably have the
     resources it needs */
  
  ref = rec->message_ref;
  gm_assert (ref);
  free_message_ref (t, ref);

  /* send another message */
  
  size = rec->size;
  len = rec->len;
  basic_type = t->basic_type;
  target_node_id = t->target_node_id;
  target_port_id = t->target_port_id;
  send (t, size, len, basic_type, target_node_id, target_port_id,
	t->_send);

  /* Free the send record */
  
  gm_lookaside_free (rec);

  GM_RETURN_NOTHING ();
}

/* send completion handler */

static void
sent (struct gm_port *port, void *context, gm_status_t status)
{
  send_record_t *rec;
  test_t *t;

  GM_CALLED ();
  
  rec = (send_record_t *) context;
  t = rec->test;
  if (status != GM_SUCCESS)
    {
      gm_perror ("gm_allsize: send failed", status);
      if (t->exit_on_error)
	{
	  gm_exit (status);
	}
      fprintf(stderr,"gm_allsize: not exiting after a send error\n");
      fflush(stderr);
    }
  free_send_token (t);
  t->_sent (port, context);
  total_sent++;
  t->outstanding_send_cnt--;

  GM_RETURN_NOTHING ();
}

/****************************************************************
 * send functions
 ****************************************************************/

static void
_send_ethernet (struct test *t, struct message_ref *ref, unsigned int size,
		unsigned long len,
		enum message_type basic_type,
		unsigned target_node_id,
		unsigned target_port_id)
{
  unsigned char uid[6];
  gm_status_t status;
  struct gm_port *port;
  message_t *m;
  send_record_t *rec;
  gm_ethernet_segment_descriptor_t seg[1];
  GM_PARAMETER_MAY_BE_UNUSED (basic_type);

  port = t->port;
  m = ref->message;
  rec = (send_record_t *) gm_lookaside_alloc (t->send_record_lookaside);
  gm_assert (rec);

  /* Since we are about to violate security, make sure we do so safely: */

  if ((((gm_up_t) m) ^ ((gm_up_t) m + len)) >= GM_PAGE_LEN)
    {
      status = GM_INVALID_PARAMETER;
      gm_perror ("Ethernet send crosses a page boundary", status);
      gm_exit (status);
    }
  status = gm_dma_addr (port, (gm_up_t) m, &seg[0].ptr);
  if (status != GM_SUCCESS)
    {
      gm_perror ("Could not get DMA addr for user pointer", status);
      gm_exit (status);
    }
  
  seg[0].len = len;
  status = gm_node_id_to_unique_id (port, target_node_id, (char *) uid);
  if (status != GM_SUCCESS)
    {
      gm_perror ("Could not determine ethernet addr for target", status);
      gm_exit (status);
    }
#if 0
  gm_ethernet_send_with_callback (port, 1, seg, uid, sent, rec);
#else
  gm_ethernet_broadcast_with_callback (port, 1, seg, sent, rec);
#endif
  
  rec->test = t;
  rec->message_ref = ref;
  rec->size = size;
  rec->len = len;
}

static void
_send_put (struct test *t, struct message_ref *ref, unsigned int size,
	   unsigned long len,
	   enum message_type basic_type,
	   unsigned target_node_id,
	   unsigned target_port_id)
{
  struct gm_port *port;
  message_t *m;
  unsigned int priority;
  send_record_t *rec;
  GM_PARAMETER_MAY_BE_UNUSED (basic_type);
  
  port = t->port;
  m = ref->message;
  priority = t->priority;
  rec = (send_record_t *) gm_lookaside_alloc (t->send_record_lookaside);
  gm_assert (rec);
  gm_put (port, m, (gm_remote_ptr_t) ((gm_size_t) m), len, GM_LOW_PRIORITY,
	  target_node_id, target_port_id, sent, rec);
  rec->test = t;
  rec->message_ref = ref;
  rec->size = size;
  rec->len = len;
}

static void
_send_get (struct test *t, struct message_ref *ref, unsigned int size,
	   unsigned long len,
	   enum message_type basic_type,
	   unsigned target_node_id,
	   unsigned target_port_id)
{
  struct gm_port *port;
  message_t *m;
  unsigned int priority;
  send_record_t *rec;
  GM_PARAMETER_MAY_BE_UNUSED (basic_type);
  
  port = t->port;
  m = ref->message;
  priority = t->priority;
  rec = (send_record_t *) gm_lookaside_alloc (t->send_record_lookaside);
  gm_assert (rec);
  gm_get (port, (gm_remote_ptr_t) ((gm_size_t) m), m, len, GM_LOW_PRIORITY,
	  target_node_id, target_port_id, sent, rec);
  rec->test = t;
  rec->message_ref = ref;
  rec->size = size;
  rec->len = len;
}

static void
_send_datagram (struct test *t, struct message_ref *ref, unsigned int size,
		unsigned long len,
		enum message_type basic_type,
		unsigned target_node_id,
		unsigned target_port_id)
{
  struct gm_port *port;
  message_t *m;
  unsigned int priority;
  send_record_t *rec;
  GM_PARAMETER_MAY_BE_UNUSED (basic_type);
  
  port = t->port;
  m = ref->message;
  priority = t->priority;
  rec = (send_record_t *) gm_lookaside_alloc (t->send_record_lookaside);
  gm_assert (rec);
  if (len <= 4096)
    {
      gm_datagram_send (port, m, size, len, priority, target_node_id,
			target_port_id, sent, rec);
    }
  else
    {
      gm_send_with_callback (port, m, size, len, priority, target_node_id,
			     target_port_id, sent, rec);
    }
  rec->test = t;
  rec->message_ref = ref;
  rec->size = size;
  rec->len = len;
}

static void
_send_datagram_or_reliable (struct test *t,
			    struct message_ref *ref,
			    unsigned int size,
			    unsigned long len,
			    enum message_type basic_type,
			    unsigned target_node_id,
			    unsigned target_port_id)
{
  if (gm_rand_mod (2))
    _send_to_nonpeer (t, ref, size, len, basic_type, target_node_id,
		      target_port_id);
  else
    _send_datagram (t, ref, size, len, basic_type, target_node_id,
		    target_port_id);
}

static void
_send_to_nonpeer (test_t *t, message_ref_t *ref, unsigned int size,
		  unsigned long len,
		  enum message_type basic_type,
		  unsigned target_node_id,
		  unsigned target_port_id)
{
  struct gm_port *port;
  message_t *m;
  unsigned int priority;
  send_record_t *rec;
  GM_PARAMETER_MAY_BE_UNUSED (basic_type);
  
  port = t->port;
  m = ref->message;
  priority = t->priority;
  rec = (send_record_t *) gm_lookaside_alloc (t->send_record_lookaside);
  gm_assert (rec);
  gm_send_with_callback (port, m, size, len, priority, target_node_id,
			 target_port_id, sent, rec);
  rec->test = t;
  rec->message_ref = ref;
  rec->size = size;
  rec->len = len;
}

static void
_send_to_peer (test_t *t, message_ref_t *ref, unsigned int size,
	       unsigned long len,
	       enum message_type basic_type,
	       unsigned int target_node_id,
	       unsigned int target_port_id)
{
  struct gm_port *port;
  message_t *m;
  unsigned int priority;
  send_record_t *rec;
  GM_PARAMETER_MAY_BE_UNUSED (basic_type);
  GM_PARAMETER_MAY_BE_UNUSED (target_port_id);
  
  port = t->port;
  m = ref->message;
  priority = t->priority;
  rec = (send_record_t *) gm_lookaside_alloc (t->send_record_lookaside);
  gm_assert (rec);
  if (GM_DEBUG_ALLSIZE && (MESSAGE_LEN (m) != len))
    {
      fprintf
	(stderr,
	 "Packet send message length 0x%x does not match length requested"
	 " 0x%lx\n",
	 MESSAGE_LEN (m), len);
    }
  gm_send_to_peer_with_callback (port, m, size, len, priority, target_node_id,
				 sent, rec);
  
  rec->test = t;
  rec->message_ref = ref;
  rec->size = size;
  rec->len = len;
}

static void
_send_abort (test_t *t, message_ref_t *ref, unsigned int size,
	     unsigned long len,
	     enum message_type message_type,
	     unsigned int target_node_id,
	     unsigned int target_port_id)
{
  GM_PARAMETER_MAY_BE_UNUSED (t);
  GM_PARAMETER_MAY_BE_UNUSED (ref);
  GM_PARAMETER_MAY_BE_UNUSED (size);
  GM_PARAMETER_MAY_BE_UNUSED (len);
  GM_PARAMETER_MAY_BE_UNUSED (message_type);
  GM_PARAMETER_MAY_BE_UNUSED (target_node_id);
  GM_PARAMETER_MAY_BE_UNUSED (target_port_id);
  
  /* This function should never be called */
  GM_ABORT ();
}

/* send a message if sufficient resources are available */

static
void
make_send_progress (test_t *t)
{
  message_ref_t *ref;
  enqueued_send_t *first;

  /* send as many packets as possible */

 again:
  first = t->first_enqueued_send;
  if (!first)
    {
      goto abort_with_nothing;
    }
      
  ref = alloc_message_ref (t, first->size);
  if (!ref)
    {
      goto abort_with_nothing;
    }

  if (!gm_alloc_send_token (t->port, t->priority))
    {
      goto abort_with_ref;
    }
      
  (*first->fill) (t, (gm_u8_n_t *) ref->message, first->len,
		  (char) first->basic_type);
  (*first->send) (t, ref, first->size, first->len,
		  first->basic_type, first->target_node_id,
		  first->target_port_id);
  t->first_enqueued_send = first->next;
  gm_lookaside_free (first);
  goto again;
  
 abort_with_ref:
  free_message_ref (t, ref);
 abort_with_nothing:
  return;
}

/* send function */

static void
send (test_t *t, unsigned int size, unsigned long len,
      enum message_type basic_type, unsigned int target_node_id,
      unsigned int target_port_id, send_func_t send_func)
{
  enqueued_send_t *first;
  message_ref_t *ref;

  GM_CALLED_WITH_ARGS (("%p, %u, %lu, %u, %u, %u)\n",
			t, size, len, (int) basic_type, target_node_id,
			target_port_id));
  
  t->outstanding_send_cnt++;
  t->send_cnt++;

  /* Attempt to send the packet now, else abort. */

  first = t->first_enqueued_send;
  if (first)
    {
      /* earlier sends are pending */
      goto abort_with_nothing;
    }

  ref = alloc_message_ref (t, size);
  if (!ref)
    {
      /* Need to wait for a send to complete to free a send buffer */
      goto abort_with_nothing;
    }

  if (!gm_alloc_send_token (t->port, t->priority))
    {
      /* Need to wait for a send to complete to free a send token */
      goto abort_with_ref;
    }

  fill_message (t, ref->message, len, basic_type);
  (*send_func) (t, ref, size, len, basic_type, target_node_id,
		target_port_id);
  total_queued++;
  GM_RETURN_NOTHING ();

  /* If the attempt, above, to send the packet now aborted, then free
     any resources allocated for the send and enqueue the packet to be
     sent later. */
  
 abort_with_ref:
  free_message_ref (t, ref);
 abort_with_nothing:
  {
    enqueued_send_t *es;

    es = gm_lookaside_alloc (t->enqueued_send_lookaside);
    gm_assert (es);
    if (first)
      t->last_enqueued_send->next = es;
    else
      t->first_enqueued_send = es;
    t->last_enqueued_send = es;
    es->next = 0;
      
    es->test = t;
    es->size = size;
    es->len = len;
    es->basic_type = basic_type;
    es->target_node_id = target_node_id;
    es->target_port_id = target_port_id;
    es->send = send_func;
    es->fill = t->fill_buffer;
  }
  total_queued++;
  make_send_progress (t);
  GM_RETURN_NOTHING ();
}

void
free_send_token (test_t *t)
{
  gm_free_send_token (t->port, t->priority);
  make_send_progress (t);
}

/****************************************************************
 * send start functions
 ****************************************************************/

static void
start_sending_many (test_t *t, unsigned int size, unsigned long len)
{
  message_ref_t *ref;
  
  while ((ref = alloc_message_ref (t, size)) != 0)
    {
      free_message_ref (t, ref);
      send (t, size, len, t->basic_type, t->target_node_id,
	    t->target_port_id, t->_send);
    }
}

static void
start_sending_one (test_t *t, unsigned int size, unsigned long send_len)
{
  send (t, size, send_len, t->basic_type, t->target_node_id,
	t->target_port_id, t->_send);
}

/****************************************************************
 * length test
 ****************************************************************/

static void
test_size_len (test_t *t, unsigned int size, unsigned long send_len)
{
  gm_u64_t time = 0;
  unsigned int caught;
  unsigned int need_to_catch;
  struct gm_port *port = t->port;
  unsigned long total_send_cnt;
  void (*saved_sent) (struct gm_port *, void *);
  recv_func_t *saved_recv;

  /* The goal here is to send for at least 50*100 microseconds (100
     times the resolution of gm_ticks()) and to send at least 100
     packets by default. */

  GM_CALLED ();
  for (total_send_cnt = t->min_sends_per_point;
       time < t->min_usecs_per_point;
       total_send_cnt *= 2)
    {
      gm_u64_t start_ticks;

      t->send_cnt = 0;
      t->recv_cnt = 0;
      gm_always_assert (t->outstanding_send_cnt == 0);
      start_ticks = gm_ticks (port);

      /* start sending */

      t->start_sending (t, size, send_len);

      /* send the appropriate number of packets */

      while (t->send_cnt < total_send_cnt)
	{
	  gm_recv_event_t *e;

	  e = t->gm_receive (port);
	  switch (gm_ntoh_u8 (e->recv.type))
	    {
	    case GM_ETHERNET_RECV_EVENT:
	    case GM_HIGH_RECV_EVENT:
	    case GM_RECV_EVENT:
	      t->recv (t, port, e, t->_send);
	      break;

	    default:
	      gm_unknown (t->port, e);
	      break;
	    }

	  if (caught_sigint)
	    {
	      sigint_exit ();
	    }

	}

      /* wait for all sends to complete and catch any ping-pong
	 packet or packets. */

      saved_sent = t->_sent;
      t->_sent = _sent_dont_send;
      saved_recv = t->recv;
      t->recv = recv_only;
      need_to_catch = ((t->send_mode == GET_MODE
			|| t->send_mode == PUT_MODE) ? 0 :
		       t->ping_pong ? (t->send_cnt - t->recv_cnt) : 0);
      caught = 0;
      while (t->outstanding_send_cnt || caught < need_to_catch)
	{
	  gm_recv_event_t *e;

	  e = t->gm_receive (port);
	  switch (gm_ntoh_u8 (e->recv.type))
	    {
	    case GM_ETHERNET_RECV_EVENT:
	    case GM_HIGH_RECV_EVENT:
	    case GM_RECV_EVENT:
	      recv_only (t, port, e, 0);
	      caught++;
	      break;
	    case GM_NO_RECV_EVENT:
	      break;
	    default:
	      gm_unknown (port, e);
	      break;
	    }

	  if (caught_sigint)
	    {
	      sigint_exit ();
	    }

	}

      t->_sent = saved_sent;
      t->recv = saved_recv;

      /* compute time */

      time = (gm_ticks (port) - start_ticks) / 2;
    }

  if (t->both_ways)
    t->record_point (t, send_len, t->send_cnt+t->recv_cnt, time);
  else 
    t->record_point (t, send_len, t->send_cnt, time);
  GM_RETURN_NOTHING ();
}

/****************************************************************
 *
 ****************************************************************/

static void
do_test_size_len (test_t *t, unsigned int size, unsigned long len)
{
  unsigned int cnt;

  GM_CALLED ();
  stats_clear (t->stats);
  for (cnt = t->count_per_len; cnt; cnt--)
    {
      test_size_len (t, size, len);
    }
  t->report (t, len);
  GM_RETURN_NOTHING ();
}

/****************************************************************
 * build length stack
 ****************************************************************/

/****************
 * length stack
 ****************/

static gm_status_t
push_len (struct length_record **rp, unsigned long l)
{
  struct length_record *r;

  GM_CALLED ();
  gm_always_assert (l != 0);
  r = (struct length_record *) gm_malloc (sizeof (*r));
  if (!r)
    {
      GM_RETURN_STATUS (GM_OUT_OF_MEMORY);
    }
  r->next = *rp;
  r->length = l;
  *rp = r;
  GM_RETURN_STATUS (GM_SUCCESS);
}

static unsigned long
pop_len (struct length_record **rp)
{
  struct length_record *r;
  unsigned long l;

  r = *rp;
  if (!r)
    return 0;
  *rp = r->next;
  l = r->length;
  gm_free (r);
  return l;
}

static void
reverse_stack (struct length_record **rp)
{
  struct length_record *from, *to, *tmp;

  from = *rp;
  to = 0;
  while (from)
    {
      tmp = from->next;
      from->next = to;
      to = from;
      from = tmp;
    }
  *rp = to;
}

static void
free_length_stack (struct length_record **rp)
{
  while (*rp)
    {
      pop_len (rp);
    }
}

/****************
 * builders
 ****************/

static unsigned long
max_build_len (test_t *t, unsigned int size)
{
  unsigned long ret;
  
  if (size == GM_ETHERNET_TAG_SIZE)
    {
      ret = GM_ETHERNET_MTU;
    }
  else
    {
      ret = gm_max_length_for_size (size);
    }
  return ret <= t->max_len ? ret : t->max_len;
}

static gm_status_t
build_linear_lengths (test_t *t, unsigned int size)
{
  unsigned long l;
  unsigned long min, max;
  gm_status_t status;

  GM_CALLED ();
  min = t->min_len > 1 ? t->min_len : 1;
  max = max_build_len (t, size);
  
  for (l = min; l <= max; l += t->length_inc)
    {
      status = push_len (&t->length_stack, l);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_stack;
	}
    }
  reverse_stack (&t->length_stack);

  gnuplot_printf (t, "set nologscale\n");

  GM_RETURN_STATUS (GM_SUCCESS);

 abort_with_stack:
  free_length_stack (&t->length_stack);
  GM_RETURN_STATUS (status);
}

static gm_status_t
build_geometric_lengths (test_t *t, unsigned int size)
{
  gm_status_t status;
  unsigned long l, incr;
  unsigned long min, max;

  GM_CALLED ();
  min = t->min_len > 1 ? t->min_len : 1;
  max = max_build_len (t, size);

  l = min;
  while (l <= max)
    {
      status = push_len (&t->length_stack, l);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_stack;
	}
      
      /* compute next test length */

      incr = l >> t->density;
      if (incr != 0)
	l += incr;
      else
	l += 1;
    }
  reverse_stack (&t->length_stack);

  gnuplot_printf (t, "set logscale\n");

  GM_RETURN_STATUS (GM_SUCCESS);

 abort_with_stack:
  free_length_stack (&t->length_stack);
  GM_RETURN_STATUS (status);
}

static gm_status_t
build_random_lengths (test_t *t, unsigned int size)
{
  gm_status_t status;
  unsigned long l;
  unsigned long iterations;
  unsigned long min, max;

  GM_CALLED ();
  min = t->min_len > 1 ? t->min_len : 1;
  max = max_build_len (t, size);

  for (iterations = (1 << size) / t->length_inc;
       iterations;
       iterations--)
    {
      l = gm_rand_mod (max - min + 1) + min;
      status = push_len (&t->length_stack, l);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_stack;
	}
    }
  gnuplot_printf (t, "set nologscale\n");
  GM_RETURN_STATUS (GM_SUCCESS);

 abort_with_stack:
  free_length_stack (&t->length_stack);
  GM_RETURN_STATUS (status);
}

/****************
 * lengths_... modifications
 ****************/

static gm_status_t
lengths_reverse (test_t *t)
{
  GM_CALLED ();
  reverse_stack (&t->length_stack);
  GM_RETURN_STATUS (GM_SUCCESS);
}

/* append reversed-order test */

static gm_status_t
lengths_add_hysteresis (test_t *t)
{
  gm_status_t status;
  struct length_record *new = 0, *lr = 0, *last_lr;

  GM_CALLED ();
  last_lr = GM_STRUCT_CONTAINING (struct length_record, next, &lr);
  for (lr = t->length_stack; lr; lr = lr->next)
    {
      status = push_len (&new, lr->length);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_stack;
	}
      last_lr = lr;
    }
  last_lr->next = new;
  GM_RETURN_STATUS (GM_SUCCESS);

 abort_with_stack:
  free_length_stack (&new);
  GM_RETURN_STATUS (status);
}

/* randomize the length stack using a O(N) randomization algorith. */

static gm_status_t
lengths_randomize (test_t *t)
{
  gm_status_t status;
  struct gm_hash *h;
  struct length_record *lr, *randomized_lengths = 0;
  unsigned long c;

  GM_CALLED ();
  
  /* create hash table to act as an array in which to store the lengths */

  h = gm_create_hash (gm_hash_compare_longs, gm_hash_hash_long,
		      sizeof (unsigned long), sizeof (unsigned long), 0, 0);
  if (!h)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_nothing;
    }
  
  /* put unrandomized lengths into a hash table and count them. */
      
  c = 0;
  for (lr = t->length_stack; lr; lr = lr->next)
    {
      status = gm_hash_insert (h, &c, &lr->length);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_hash;
	}
      c++;
    }

  /* repeatedly copy half the lengths out of the hash table onto
     a stack and repack the hash table. */

  while (1)
    {
      unsigned long from, to, limit, len, *len_ptr;

      if (c == 0)
	break;

      /* randomly extract half the entries */
      
      limit = c;
      while (c > limit/2)
	{
	  do
	    {
	      from = gm_rand_mod (limit+1);
	      len_ptr = (unsigned long *) gm_hash_remove (h, &from);
	    }
	  while (!len_ptr);
	  len = *len_ptr;
	  status = push_len (&randomized_lengths, len);
	  if (status != GM_SUCCESS)
	    {
	      goto abort_with_stack;
	    }
	  c--;
	}

      /* repack the hash table */

      to = 0;
      for (from = 0; from <= limit; from++)
	{
	  len_ptr = (unsigned long *) gm_hash_remove (h, &from);
	  if (!len_ptr)
	    continue;
	  len = *len_ptr;
	  
	  status = gm_hash_insert (h, &to, &len);
	  if (status != GM_SUCCESS)
	    {
	      goto abort_with_stack;
	    }
	  to++;
	}
      
      gm_assert (to == c);
    }

  /* replace the old length stack with the new one */
  
  free_length_stack (&t->length_stack);
  t->length_stack = randomized_lengths;

  /* clean up and return */
  
  gm_destroy_hash (h);
  GM_RETURN_STATUS (GM_SUCCESS);
  
 abort_with_stack:
  free_length_stack (&randomized_lengths);
 abort_with_hash:
  gm_destroy_hash (h);
 abort_with_nothing:
  GM_RETURN_STATUS (status);
}

/****************************************************************
 * test_do
 ****************************************************************/

static gm_status_t
test_do_normal (test_t *t, unsigned int size)
{
  gm_status_t status;
  unsigned long len;

  GM_CALLED ();
  
  /* build list of test lengths */
  
  status = t->build_length_stack (t, size);
  if (status != GM_SUCCESS)
    {
      goto abort_with_nothing;
    }
  
  /* massage the list of test lengths */

  if (t->reverse)
    {
      status = lengths_reverse (t);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_length_stack;
	}
    }
  if (t->randomize)
    {
      status = lengths_randomize (t);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_length_stack;
	}
    }
  if (t->hysteresis)
    {
      status = lengths_add_hysteresis (t);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_length_stack;
	}
    }

  /* perform the test */
  
  while ((len = pop_len (&t->length_stack)) != 0)
    {
      do_test_size_len (t, size, len);
      if (caught_sigint)
	{
	  sigint_exit ();
	}
    }

  GM_RETURN_STATUS (GM_SUCCESS);

 abort_with_length_stack:
  free_length_stack (&t->length_stack);
 abort_with_nothing:
  GM_RETURN_STATUS (status);
}

/* Act as a slave, like the netperf server */

static gm_status_t
test_do_slave (test_t *t, unsigned int size)
{
  struct gm_port *port;
  GM_PARAMETER_MAY_BE_UNUSED (size);
  GM_CALLED ();
  
  port = t->port;
  t->_sent = _sent_dont_send;
  t->_send = _send_abort;
  while (1)
    {
      gm_recv_event_t *e;

      e = t->gm_receive (port);
      switch (gm_ntoh_u8 (e->recv.type))
	{
	case GM_ETHERNET_RECV_EVENT:
	case GM_HIGH_RECV_EVENT:
	case GM_RECV_EVENT:
	  recv_slave (t, port, e, _send_abort);
	  break;

	default:
	  gm_unknown (t->port, e);
	  /* fall through */
	case GM_NO_RECV_EVENT:
	  break;
	}
      
      if (caught_sigint)
	{
	  sigint_exit ();
	}
    }
  GM_RETURN_STATUS (GM_SUCCESS);
}

/****************************************************************
 *  Create a new test state
 ****************************************************************/

static test_t *
new_test_state (void)
{
  test_t *t;
  int i;

  GM_CALLED ();
  if (gm_init () != GM_SUCCESS)
    {
      goto abort_with_nothing;
    }

  strcpy (output_file_parm, "");

  t = (test_t *) gm_calloc (1, sizeof (*t));
  if (!t)
    {
      goto abort_with_init;
    }
  
  /* initialize globals here for VxWorks compat. */

  t->build_length_stack = build_linear_lengths;
  t->count_per_len = 1;
  t->data_file = NULL;
  t->density = 3;
  t->enqueued_send_lookaside
    = gm_create_lookaside (sizeof (enqueued_send_t), 1);
  if (!t->enqueued_send_lookaside)
    {
      goto abort_with_tee;
    }
  t->exit_on_error = 1;
  t->fast = 0;
  for (i=0; i<32; i++)
    t->free_send_message_refs[i] = 0;
  t->gm_receive = gm_receive;
  t->last_created_message_ref = 0;
  t->length_inc = 1;
  t->max_len = ~0L;
  t->max_size = MAX_TEST_SIZE;
  t->min_len = 0;
  t->min_sends_per_point = 100;
  t->min_size = MIN_TEST_SIZE;
  t->min_usecs_per_point = 5000;
  t->offset = 0;
  t->ping_pong = 1;
  t->both_ways = 0;
  t->points = "points";
  t->port_id = 2;
  t->priority = GM_LOW_PRIORITY;
  t->rand_seed = 42;
  t->record_point = record_point_latency;
  t->remote_host_name = 0;
  t->report = report_scatter;
  t->send_record_lookaside = gm_create_lookaside (sizeof (send_record_t), 1);
  if (!t->send_record_lookaside)
    {
      goto abort_with_enqueued_send_lookaside;
    }
  t->stats_socket = 0;
  t->target_node_id = 0;
  t->target_port_id = 2;
  t->test_do = test_do_normal;
  t->timer = 0;
  t->timestamps = 0;
  t->time_string = "";
  t->verify = 0;
  
  GM_RETURN (t);

 abort_with_enqueued_send_lookaside:
  gm_destroy_lookaside (t->enqueued_send_lookaside);
 abort_with_tee:
  gm_free (t);
 abort_with_init:
  gm_finalize ();
 abort_with_nothing:
  GM_RETURN (0);
}

/* extra "status" parameter for gm_on_exit() compatibility. */

static void
free_test_state (gm_status_t status, test_t *t)
{
  GM_PARAMETER_MAY_BE_UNUSED (status);
  
  gm_destroy_lookaside (t->send_record_lookaside);
  gm_destroy_lookaside (t->enqueued_send_lookaside);
  gm_free (t);
  gm_finalize ();
}

/****************************************************************
 * argument parsing
 ****************************************************************/

char output_host [2500];

  /* Parsing macros; up to three actions are supported for each flag    */
  /* The 'F' macro is for one-form flags with no argument (parameter)   */
#define F(name,action1,action2,action3) \
 if (strcmp (argv[i], name) == 0) { action1; action2; action3; continue; }
  /* The 'F2' macro is for two-form flags with no argument (parameter)   */
#define F2(name1,name2,action1,action2,action3) \
 if ((strcmp (argv[i], name1) == 0) || (strcmp (argv[i], name2) == 0)) \
 { action1; action2; action3; continue; }


static void
parse_args (test_t *t, int argc, char *argv[])
{
/* NOTE: If you change the operating characteristics or the flags
 *       used by this program, please notify Max so he can change the
 *       Myrinet Handbook document.
 */
  int i;
  int output_port;

  GM_CALLED ();
  
  for (i = 1; i < argc; i++)
    {
      /* Dispose of the --help flag first                                */
      if (strcmp (argv[i], "--help") == 0)
	{
	  print_synopsis();
	  print_options();
	  gm_exit (GM_SUCCESS);
	}

      /* Here are the flags with no argument.                                */
      /* Please keep short form, if any, with its long form equivalent using */
      /* the 'F2' macro, but keep the long form options in alpha sequence.   */

      F2("--bandwidth", "-bw", t->record_point = record_point_bandwidth,;,;);
      F ("--bps",              t->report_bytes_per_sec = 1,;,;); /* DEPRECATED */
      F2("--bidirectional","-b", t->ping_pong = 1,;,;);
      F ("--blocking",         t->gm_receive = gm_blocking_receive,;,;);
      F ("--blocking-no-spin", t->gm_receive = gm_blocking_receive_no_spin,;,;);
      F ("--both-ways",        t->both_ways =1,
			       /* override --latency & --unidirectional  */
			       t->record_point = record_point_bandwidth,
			       t->ping_pong = 1);
      F ("--bytes-per-second", t->report_bytes_per_sec = 1,;,;);
      F ("--copy",             t->do_copy = 1,;,;);
      F ("--datagrams=maybe",  t->send_mode = MAYBE_DATAGRAM_MODE,;,;);
      F ("--datagrams",        t->send_mode = DATAGRAM_MODE,;,;);
      F ("--dots",             t->points = "dots",;,;);
      F ("--ethernet",         t->send_mode = ETHERNET_MODE,;,;);
      F2("--exit-on-error","-e", t->exit_on_error = 1,;,;);
      F2("--fast","-f",        t->fast = 1,;,;);
      F ("--fork-child",       t->fork_child  = 1,;,;);
      F ("--fork-parent",      t->fork_parent = 1,;,;);
      F2("--geometric", "-g",  t->build_length_stack =
						    build_geometric_lengths,;,;);
      F ("--get",              t->send_mode = GET_MODE,;,;);
      F ("--hysteresis",       t->hysteresis = 1,;,;);
      F ("--ignore-errors",    t->exit_on_error = 0,;,;);
      F2("--latency", "-lat",  t->record_point = record_point_latency,
			       /* override --both-ways */
			       t->both_ways = 0,;);
      F ("--linear",           t->build_length_stack = build_linear_lengths,;,;);
      F ("--max-plot",         t->report = report_max,;,;);
      F ("--min-plot",         t->report = report_min,;,;);
      F ("--no-versionstamp",  t->version = 0,;,;);
      F ("--nonblocking",      t->gm_receive = gm_receive,;,;);
      F ("--random-lengths",   t->build_length_stack = build_random_lengths,;,;);
      F ("--put",              t->send_mode = PUT_MODE,;,;);
      F ("--quiet",            t->report = report_nothing,;,;);
      F ("--randomize",        t->randomize = 1,;,;);
      F ("--reverse",          t->reverse = 1,;,;);
      F ("--scatter-plot",     t->report = report_scatter,;,;);
      F ("--silent",           t->report = report_nothing,;,;);
      F ("--slave",            t->test_do = test_do_slave,;,;);
      F ("--stddev-plot",      t->report = report_stddev,;,;);
      F ("--timestamp",        t->timestamps = 1,;,;);
      F2("--unidirectional", "-u",
			       t->ping_pong = 0,
			       /* override --both-ways */
			       t->both_ways = 0,;);
      F2("--verify", "-v",     t->verify = 1,;,;);
      F ("--versionstamp",     t->version = 1,;,;);

#undef F2
#undef F

      /* The check_opt macro is for long-form options with numeric        */
      /* arguments and simple validation tests.                           */
#define check_opt(prefix, var, test)					\
	  if (sscanf (argv[i], "--" prefix "=%i", &var) == 1)		\
	    {								\
	      if (!(test))						\
		{							\
		  usage ("bad value for option \"%s\"\n", argv[i]);	\
		}							\
	      continue;							\
	    }

      /* Here are the flags with arguments, sequenced as above */

      check_opt ("board", t->board, t->board >= 0);
      if (strcmp (argv[i], "-B") == 0)
	{
	  if (i + 1 >= argc)
	      usage ("Board number expected after '-B'.\n");

	  t->board = atoi (argv[++i]);
	  if (t->board < 0)
	      usage ("bad board number: %d\n", t->board);

	  continue;
	}

      check_opt ("count-per-length", t->count_per_len, t->count_per_len > 0);
      if (strcmp (argv[i], "-c") == 0)
	{
	  if (i + 1 >= argc)
	      usage ("Count expected after '-c'.\n");

	  t->count_per_len = atoi (argv[++i]);
	  if (t->count_per_len < 1)
	      usage ("Bad count: %d\n", t->count_per_len);

	  continue;
	}

      if (sscanf (argv[i], "--density=%u", &t->density) == 1) continue;

      if (sscanf (argv[i], "--enable-nack-down=%u",
		  &t->enable_nack_down_flag) == 1)
	{
	  t->enable_nack_down_set = 1;
	  continue;
	}

      check_opt ("length-increment", t->length_inc, t->length_inc > 0);
      if (strcmp (argv[i], "-l") == 0)
        {
          if (i + 1 >= argc)
              usage ("length increment expected after '-l'.\n");

          t->length_inc = atoi (argv[++i]);
          if ((t->length_inc <= 0))
              usage ("bad length increment: %d\n", t->length_inc);

          continue;
        }
      check_opt ("max-len",    t->max_len, 1);                 /* DEPRECATED */
      check_opt ("max-length", t->max_len, 1);

      check_opt ("max-size",   t->max_size, t->max_size < 32);

      check_opt ("min-len",    t->min_len, 1);                 /* DEPRECATED */
      check_opt ("min-length", t->min_len, 1);

      if (sscanf (argv[i], "--min-sends-per-point=%i",
		  &t->min_sends_per_point) == 1) continue;

      /* Note that the parameter to gm_min_message_size() is a dummy */
      check_opt ("min-size",   t->min_size,
		 t->min_size >= gm_min_message_size ((struct gm_port *)0));

      if (sscanf (argv[i], "--min-usecs-per-point=%li",
		  &t->min_usecs_per_point) == 1) continue;

      if (sscanf (argv[i], "--offset=%i", &t->offset) == 1) 
	{
	  if (t->offset & 0x7)
	      usage ("offset must be a multiple of 8\n");

	  if (t->offset >= GM_PAGE_LEN)
	      usage ("offset must be less than GM_PAGE_LEN\n");

	  continue;
	}

      /* Note that the parameter to gm_num_ports() is a dummy */
      /* Note that t->port_id is unsigned ... no compare to 0 */
      check_opt ("port", t->port_id,
		 (t->port_id < gm_num_ports ((struct gm_port *)0)));
      if (strcmp (argv[i], "-p") == 0)
	{
	  if (i + 1 >= argc)
	      usage ("Port number expected after '-p'.\n");

	  t->port_id = atoi (argv[++i]);
	  /* Note that the parameter to gm_num_ports() is a dummy */
	  if (t->port_id >= gm_num_ports ((struct gm_port *)0))
	      usage ("bad port number: %d\n", t->port_id);

	  continue;
	}

      if (sscanf (argv[i], "--remote-host=%s", remote_host_parm) == 1)
	{
	  remote_host_parm[sizeof(remote_host_parm) -1] = '\0';
	  t->remote_host_name = remote_host_parm;
	  continue;
	}
      if (strcmp (argv[i], "-h") == 0)
	{
	  if (i + 1 >= argc)
	      usage ("Host name expected after '-h'.\n");

	  strncpy (remote_host_parm, argv[++i], sizeof(remote_host_parm));
	  remote_host_parm[sizeof(remote_host_parm) -1] = '\0';
	  t->remote_host_name = remote_host_parm;
	  continue;
	}

      /* Note that the parameter to gm_num_ports() is a dummy */
      /* Note that t->target_port_id is unsigned ... no compare to 0 */
      check_opt ("remote-port", t->target_port_id,
		 (t->target_port_id < gm_num_ports ((struct gm_port *)0)));
      if (strcmp (argv[i], "-d") == 0)
	{
	  if (i + 1 >= argc)
	      usage ("Destination port number expected after '-d'.\n");

	  t->target_port_id = atoi (argv[++i]);
	  /* Note that the parameter to gm_num_ports() is a dummy */
	  if (t->target_port_id >= gm_num_ports ((struct gm_port *)0))
	      usage ("bad destination port number: %d\n",
			      t->target_port_id);

	  continue;
	}

      if (sscanf (argv[i], "--size=%i", &t->min_size) == 1)
	{
	  t->max_size = t->min_size;
	  /* Note that the parameter to gm_min_message_size() is a dummy */
	  if (!(t->min_size >= gm_min_message_size ((struct gm_port *)0)
		&& t->max_size < 32))
	      usage ("bad value for option \"%s\"\n", argv[i]);

	  continue;
	}
      if (strcmp (argv[i], "-s") == 0)
        {
          if (i + 1 >= argc)
              usage ("size expected after '-s'.\n");

          t->min_size = t->max_size = atoi (argv[++i]);
	  /* Note that the parameter to gm_min_message_size() is a dummy */
	  if (!(t->min_size >= gm_min_message_size ((struct gm_port *)0)
		&& t->max_size < 32))
	      usage ("bad value for size: %d", t->min_size);

          continue;
        }

      if (GM_ENABLE_STATS_SOCKET)
	{
	  if (sscanf (argv[i], "--socket-output=%[^:]:%d", output_host,
		      &output_port)
	      == 2)
	    {
	      t->stats_socket = stats_socket_open (output_host, output_port);
	      if (!t->stats_socket)
		gm_exit (GM_FAILURE);
	      on_exit_must_succeed (((gm_on_exit_callback_t)
				     close_stats_socket_callback),
				    &t->stats_socket);
	      continue;
	    }
	}

      if (sscanf (argv[i], "--output=%s", output_file_parm) == 1)
	{
	  output_file_parm[sizeof(output_file_parm) -1] = '\0';
	  continue;
	}
      if (strcmp (argv[i], "-o") == 0)
	{
	  if (i + 1 >= argc)
	      usage ("file name expected after '-o'.\n");

	  strncpy (output_file_parm, argv[++i], sizeof(output_file_parm));
	  output_file_parm[sizeof(output_file_parm) -1] = '\0';
	  continue;
	}

#undef check_opt

      usage ("unknown arg: '%s'\n", argv[i]); /* Function does not return */

    }

  /* Open the output file, if specified         */
  if (strlen(output_file_parm))
    {
      if (strcmp ("-", output_file_parm) == 0)
	{
	  t->data_file = stdout;
	}
      else if (strlen (output_file_parm) > strlen (GNUPLOT_EXT)
	       && !strcmp (GNUPLOT_EXT ,
			   (output_file_parm + strlen (output_file_parm)
			    - strlen (GNUPLOT_EXT))))
	{
	  t->gnuplot.file = gnuplot_open (t, output_file_parm);
	  if (t->gnuplot.file == 0)
	    gm_exit (GM_FAILURE);
	  on_exit_must_succeed ((gm_on_exit_callback_t) gnuplot_close, t);
	}
      else
	{
	  t->data_file = fopen (output_file_parm, "w");
	  if (t->data_file == NULL)
	    {
	      fprintf (stderr, "Could not open output file \"%s\".\n",
		       output_file_parm);
	      perror ("could not open file");
	      gm_exit (GM_FAILURE);
	    }
	  on_exit_must_succeed ((gm_on_exit_callback_t) fclose_callback,
				t->data_file);
	}
    }

  if (t->min_size > t->max_size)
      usage ("min_size(%d) > max_size(%d) - you need to adjust one"
		      " or the other\n", t->min_size, t->max_size);
  /* Function usage() does not return */

  /* let the user know what the parameters are */

  printf ("# Opening board %d  port %d. Sending to port %d.\n",
	  t->board, t->port_id, t->target_port_id);

  if (t->verify)
    {
      printf ("# Verififying contents of all messages.\n");
      if (t->exit_on_error)
	{
	  printf ("# Will exit immediately if we see an error.\n");
	}
      else
	{
	  printf ("# Will NOT exit immediately if we see an error.\n");
	}
    }
  else
    {
      printf ("# Not verifying contents of messages.\n");
    }
  if (t->do_copy)
    {
      printf ("# Copying messages before send and after receive.\n");
    }

  GM_RETURN_NOTHING ();
}

/****************************************************************
 * Support for GM_DEBUG_ALLSIZE_ALARM
 *
 * Optional alarm to periodically report progress.
 ****************************************************************/

static
gm_alarm_t _my_alarm;

static
void
my_alarm (void *context)
{
  test_t *t;

  t = (test_t *) context;
  print_packet_counts (GM_SUCCESS, 0);
  fprintf (stderr, "# outstanding_send_cnt = %u\n", t->outstanding_send_cnt);
  if (GM_DEBUG_ALLSIZE_RECV_Q)
    _gm_dump_receive_queue (t->port);
  gm_set_alarm (t->port, &_my_alarm,
		(unsigned int)(GM_DEBUG_ALLSIZE_ALARM * TIME_CALIBRATION),
		my_alarm, context);
}

/****************************************************************
 * Test Initialization
 ****************************************************************/

static gm_status_t
test_init (test_t *t)
{
  int i;
  unsigned int min_size, size, max_size;
  gm_status_t status;

  GM_CALLED ();
  
  /* initialize GM */

  t->port = NULL;
  status = gm_open (&t->port, t->board, t->port_id, "gm_allsize",
		    (enum gm_api_version) GM_API_VERSION);
  if (status != GM_SUCCESS)
    {
      gm_perror ("could not open port", status);
      goto abort_with_nothing;
    }

  if (t->enable_nack_down_set)
    {
      fprintf (stderr, "Setting enable_nack_down to %d\n",
	       t->enable_nack_down_flag);
      gm_set_enable_nack_down (t->port, t->enable_nack_down_flag);
    }

  /* optionally set an alarm to periodically report packet counts */

  if (GM_DEBUG_ALLSIZE_ALARM)
    {
      gm_initialize_alarm (&_my_alarm);
      gm_set_alarm (t->port, &_my_alarm, (gm_u64_t) 1000000UL, my_alarm,
		    (void *) t);
    }
  
  status = gm_set_acceptable_sizes (t->port, t->priority,
				    ((1 << (t->max_size + 1)) - 1)
				    ^ ((1 << t->min_size) - 1));
  if (status != GM_SUCCESS)
    {
      gm_perror ("could not set GM acceptable sizes", status);
      goto abort_with_open_port;
    }

  /* Set the remote node ID to this node ID */

  status = gm_get_node_id (t->port, &t->this_node_id);
  if (status != GM_SUCCESS)
    {
      printf ("Error: could not determine node id.\n");
      goto abort_with_open_port;
    }

  /* compute the remote node ID from the remote host name, if any, else
     assume loopback. */

  if (t->remote_host_name)
    {
      t->target_node_id = gm_host_name_to_node_id (t->port,
						   t->remote_host_name);
      if (t->target_node_id == GM_NO_SUCH_NODE_ID)
	{
	  printf ("Error: host '%s' not found.\n", t->remote_host_name);
	  status = GM_INVALID_PARAMETER;
	  goto abort_with_open_port;
	}

      printf ("Remote GM node id %s (%d).\n",
	      t->remote_host_name, t->target_node_id);
    }
  else
    {
      gm_assert (t->target_node_id == 0);
      t->target_node_id = t->this_node_id;
      if (t->both_ways)
	{
	  printf ("--both-ways is meaningless for a loopback test;\n"
		      "results may not be as expected\n");
	  t->both_ways = 0;		/* meaningless for loopback tests */
	  /* Any overrides caused by --both_ways in parse_args() will stick */
	}
    }

  /* let GM manage the send tokens */

  gm_free_send_tokens (t->port, GM_LOW_PRIORITY,
		       gm_num_send_tokens (t->port) / 2);
  gm_free_send_tokens (t->port, GM_HIGH_PRIORITY,
		       gm_num_send_tokens (t->port) / 2);

  t->provide_recv_buffer_with_tag = (t->send_mode == ETHERNET_MODE
				     ? provide_ethernet_receive_buffer
				     : gm_provide_receive_buffer_with_tag);
  
  /* setup recv tokens */

  min_size = t->min_size;
  max_size = t->max_size;
  if (t->send_mode == ETHERNET_MODE)
    {
      max_size = min_size = GM_ETHERNET_TAG_SIZE;
    }
  for (size=min_size; size<=max_size; size++)
    {
      /* allocate as many as we want - GM is not keeping track of them */

      int send_buffers = SEND_BUFFERS;

      /* total over all sizes must be <= GM_NUM_RECV_TOKENS */

      int recv_tokens_of_size = RECV_TOKENS_OF_SIZE;

      /* allocate receive buffers */

      for (i = 0; i < recv_tokens_of_size; i++)
	{
	  message_ref_t *ref;

	  ref = create_message_ref (t, size);
	  if (!ref)
	    {
	      status = GM_OUT_OF_MEMORY;
	      goto abort_with_message_refs;
	    }
	  t->provide_recv_buffer_with_tag (t->port, ref->message, size,
					   t->priority,
					   GM_ARBITRARY_RECV_BUFFER_TAG);
	}

      /* allocate send buffers.  Alloc twice as many to allow for recv
         handlers allocating another ref before freeing the one used,
         and also allocate extra to allow for send completion
         handlers that have not yet been called. */

      for (i = 0; i < send_buffers; i++)
	{
	  message_ref_t *ref;

	  ref = create_message_ref (t, size);
	  if (!ref)
	    {
	      status = GM_OUT_OF_MEMORY;
	      goto abort_with_message_refs;
	    }
	  free_message_ref (t, ref);
	}
    }
  if (t->send_mode == ETHERNET_MODE)
    {
      t->provide_recv_buffer_with_tag = reprovide_ethernet_scatter_list;
    }
  
  t->stats = stats_alloc ();
  if (!t->stats)
    {
      status = GM_OUT_OF_MEMORY;
      gm_perror ("Could not initialize statistics package.", status);
      goto abort_with_message_refs;
    }

  if (t->min_len < MIN_TEST_LEN)
    t->min_len = MIN_TEST_LEN;

  /* compute basic type of sends */

  t->basic_type = TYPE_BLANK;
  if (t->ping_pong)
    t->basic_type |= TYPE_REPLY;
  if (t->verify)
    t->basic_type |= TYPE_VERIFY;
  if (t->send_mode == DATAGRAM_MODE)
    t->basic_type |= TYPE_SEND_MODE_DATAGRAM;
  if (t->send_mode == MAYBE_DATAGRAM_MODE)
    t->basic_type |= TYPE_SEND_MODE_DATAGRAM_OR_RELIABLE;
  else if (t->target_port_id == t->port_id)
    t->basic_type |= TYPE_SEND_MODE_PEER;
  else
    t->basic_type |= TYPE_SEND_MODE_NONPEER;

  /****************
   * Setup the handlers.
   ****************/

  t->_send = (t->send_mode == ETHERNET_MODE ? _send_ethernet :
	      t->send_mode == GET_MODE ? _send_get :
	      t->send_mode == PUT_MODE ? _send_put :
	      t->send_mode == MAYBE_DATAGRAM_MODE ? _send_datagram_or_reliable :
	      t->send_mode == DATAGRAM_MODE ? _send_datagram :
	      t->target_port_id == t->port_id ? _send_to_peer :
	      _send_to_nonpeer);
  t->_sent = ((t->send_mode == GET_MODE
	       || t->send_mode == PUT_MODE) ? _sent__send_again :
	      t->ping_pong ? _sent_dont_send :
	      _sent__send_again);
  /* HACK: fill_buffer will be called by a slave only if it should fill */
  t->fill_buffer = (t->verify || t->test_do == test_do_slave
		    ? (t->fast ? fill_buffer_fast : fill_buffer_thorough)
		    : fill_buffer_not);
  /* t->build_length_stack set in parse_args() */
  t->recv = (t->ping_pong
	     ? (t->verify ? recv_and_reply_and_verify : recv_and_reply)
	     : (t->verify ? recv_and_verify : recv_only));
  if (t->ping_pong)
    {
      if (t->both_ways)
	t->start_sending = start_sending_many;
      else 
	t->start_sending = start_sending_one;
    }
  else
    {
      t->start_sending = start_sending_many;
    }
  
  /* t->test_do set in parse_args() */
  t->verify_buffer = (t->verify
		      ? (t->fast ? verify_buffer_fast : verify_buffer_thorough)
		      : verify_buffer_not);
  /* t->report set in parse_args() */

  GM_RETURN (GM_SUCCESS);

 abort_with_message_refs:
  destroy_all_message_refs (t);
 abort_with_open_port:
  gm_close (t->port);
 abort_with_nothing:
  GM_RETURN (status);
}

static void
test_finalize (gm_status_t status, test_t *t)
{
  GM_PARAMETER_MAY_BE_UNUSED (status);

  free_length_stack (&t->length_stack);
  stats_free (t->stats);
  destroy_all_message_refs(t);
  gm_close (t->port);
}

static void
dump_args (test_t *t, int argc, char *argv[])
{
  int i;

  GM_CALLED ();
  if (t->data_file)
    {
      fprintf(t->data_file, "#");
      for (i=0; i<argc; i++)
	fprintf(t->data_file, " %s", argv[i]);
      fprintf(t->data_file, "\n");
    }

  fprintf(stdout, "#");
  for (i=0; i<argc; i++)
    fprintf(stdout, " %s", argv[i]);
  fprintf(stdout, "\n");
  GM_RETURN_NOTHING ();
}

/****************************************************************
 * main entry point
 ****************************************************************/

static int
gm_allsize (int argc, char *argv[])
{
  test_t *t;
  unsigned int min_size, size, max_size;
  gm_status_t status;

  GM_CALLED ();

  /* clear globals */
  total_queued = total_recvd = total_sent = 0;

  /* arrange to report results on exit. */

  on_exit_must_succeed (print_packet_counts, 0);
  
  /* allocate test state */
  
  t = new_test_state ();	/* sets defaults */
  if (!t)
    {
      gm_exit (GM_OUT_OF_MEMORY);
    }
  on_exit_must_succeed ((gm_on_exit_callback_t) free_test_state, t);

  /* parse the command line arguments.  This will exit on an error, but
     that is OK, since we use gm_on_exit() to clean up. */
  
  parse_args (t, argc, argv);	/* override defaults */


  /* initialize the test using the information gathered in parse_args */
  
  status = test_init (t);		/* get ready to test */
  if (status != GM_SUCCESS)
    {
      gm_exit (status);
    }
  on_exit_must_succeed ((gm_on_exit_callback_t) test_finalize, t);

  /* catch control-C signals and clean up as best we can */

  signal (SIGINT, handle_sigint);

#if !defined (WIN32) && !defined (GM_OS_VXWORKS)
  if (t->fork_parent || t->fork_child)
    {
      volatile pid_t fork_rv;
      int status;


      fork_rv = fork();
      if ((int)fork_rv < 0) 
	{
	   printf("Fork failed\n");
	   gm_exit (GM_FAILURE);
	}

      if ((int)fork_rv && t->fork_child)
	{
	  /* I'm the parent and only child should stay up */
	  wait(&status);
	  _exit(WEXITSTATUS(status));
	}
      else if (!(int)fork_rv && t->fork_parent) 
      	{
	  /* I'm the child and only the parent should stay alive */
	  _exit(0);
	}

      printf("%s %s enabled\n",t->fork_parent?"fork_parent":"",
		t->fork_child?"fork_child":"");
      if (fork_rv) 
      	printf("Parent is running the test\n");
      else
      	printf("Child is running the test\n");
    }
#endif

  /* perform the tests */

  min_size = t->min_size;
  max_size = t->max_size;
  if (t->send_mode == ETHERNET_MODE)
    {
      min_size = max_size = GM_ETHERNET_TAG_SIZE;
    }
  for (size=min_size; size<=max_size; size++)
    {
      gnuplot_start_size (t, size, 0);
      label_output (t, size, 0);
      dump_args (t, argc, argv);
      status = t->test_do (t, size); /* do test */
      if (status != GM_SUCCESS)
	{
	  gm_exit (status);
	}
    }

  /* done */

  gm_exit (GM_SUCCESS);

  /* never get here */
  
  GM_RETURN (0);		/* retval to prevent compiler warning */
}

/****************************************************************
 * Entry point
 ****************************************************************/

#define GM_MAIN gm_allsize
#include "gm_main.h"

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
