/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1999 by Myricom, Inc.                                   *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* author: glenn@myri.com */

/* NOTE: If you change the operating characteristics or the flags
 *       used by this program, please notify Max so he can change the
 *       Myrinet Handbook document.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#if defined __STDC__ || defined WIN32
#if GM_OS_VXWORKS
#include "vxWorks.h"
#endif  /* GM_OS_VXWOKRS */
#include <stdarg.h>
#else
#include <varargs.h>
#endif

/* This program uses undocumented software interfaces that are not
   intended for use in code written outside Myricom.  Hence the
   following undocumented includes. */

#include "gm_config.h"
#include "gm_internal.h"
#include "gm_lanai.h"
#include "myriInterface.h"
#include "myriProm.h"

/****************************************************************
 * global data
 ****************************************************************/

int show_route_in_hex  = 1;
int specific_board_num = -1;
int quit_on_error      = 1;     /* 0 => keep going after open error       */
int info_level         = 1;	/* 1 => verbose; 0 => terse; -1 => silent */
int verify = 0;
int wait_for_mapper = 0;
int yp = 1;

/****************************************************************
 * subroutines
 ****************************************************************/

static void
route_byte_string (unsigned char byte, char *route_string)
{
  int my_byte = (int) byte;

  if ((my_byte >= 0x80) && (my_byte <= 0x8F))
    {
      sprintf (route_string, "%d", my_byte & ~0x80);
      return;
    }

  if ((my_byte >= 0xB0) && (my_byte <= 0xBF))
    {
      /* negative route */
      sprintf (route_string, "%d", (my_byte - 0xC0));
      return;
    }

  sprintf (route_string, "%02x", my_byte);
}

#define CHECK_SIZING(type) do {						    \
  if (!GM_POWER_OF_TWO (sizeof (struct type))				    \
      && sizeof (struct type) != sizeof (struct type ## _fields))	    \
    {									    \
      printf								    \
	("sizeof " #type " struct is neither a power of two\n"		    \
	 "nor as small as possible, reducing performance and scalabity\n"); \
    }									    \
} while (0)



/****************************************************************
 * Command line processing
 ****************************************************************/

#define TP(a) fprintf (stderr, a)

static void
print_synopsis(void)
{
  TP ("\n");
  TP ("Myricom 'gm_board_info' displays information about the Myrinet\n");
  TP ("interfaces on this host.  If the mapper has been run, it also\n");
  TP ("displays information about the other interfaces reachable from\n");
  TP ("these interfaces on the Myrinet, and the routes to them.\n");
}


static void
print_options (void)
{
  TP ("\nusage: gm_board_info [options]\n\n");

  TP ("-B, --board=NUM      [all boards] display this specific board\n");
  TP ("    --decimal-routes [off]        show routes in decimal instead of hex\n");
  TP ("    --help                        output this usage information, then exit\n");
  TP ("-k  --keep-going     [off]        keep going even if open error\n");
  TP ("    --quiet          [off]        show only abbreviated output\n");
  TP ("    --silent         [off]        show errors only\n");
  TP ("    --no-yp          [off]        don't use yp (for resolving hostnames)\n");
  TP ("    --verify         [off]        verify hostnames resolution\n");
  TP ("    --wait           [off]        wait for the mapper to configure\n"
      "                                  the network before reporting.\n");

  printf ("\n");

#undef TP
}


static void
#ifdef va_dcl
usage (fmt, va_alist)
     const char *fmt;
     va_dcl
#else
usage (const char *fmt, ...)
#endif
{
  va_list ap;

  print_options();

/* Now print the caller's diagnostic */
#if defined __STDC__ || defined WIN32
  va_start (ap, fmt);
#else
  va_start (ap);
#endif
  vfprintf (stderr, fmt, ap);
  va_end (ap);

} /* usage() */


static void
parse_args (int argc, char *argv[])
{
/* NOTE: If you change the operating characteristics or the flags
 *       used by this program, please notify Max so he can change the
 *       Myrinet Handbook document.
 */
  int i;

  /* The 'F' macro is for all flags with no argument (parameter)     */
#define F(name,action) if (strcmp (argv[i], name) == 0) { action; continue; }

  for (i = 1; i < argc; i++)
    {
      /* Dispose of the --help flag first                                */
      if (strcmp (argv[i], "--help") == 0)
	{
	  print_synopsis();
	  print_options();
	  gm_exit (GM_SUCCESS);
	}

      /* Here are the flags with no argument.                            */
      /* Please keep short form, if any, after its long form equivalent, */
      /* but keep the long form options in alpha sequence.               */
      F ("--decimal-routes", show_route_in_hex = 0);
      F ("-d"              , show_route_in_hex = 0);       /* Deprecated */
      F ("--keep-going"    , quit_on_error     = 0);
      F ("-k"              , quit_on_error     = 0);
      F ("--no-yp", yp = 0);
      F ("--quiet"         , info_level        = 0);
      F ("-t"              , info_level        = 0);       /* Deprecated */
      F ("--silent"        , info_level        = -1);
      F ("--verify", verify = 1);
      F ("--wait", wait_for_mapper = 1);

#undef F

      /* The check_opt macro is for long-form options with numeric        */
      /* arguments and simple validation tests.                           */
#define check_opt(prefix, var, test)					\
	  if (sscanf (argv[i], "--" prefix "=%i", &var) == 1)		\
	    {								\
	      if (!(test))						\
		{							\
		  usage ("bad value for option \"%s\"\n", argv[i]);	\
		  gm_exit (GM_FAILURE);					\
		}							\
	      continue;							\
	    }

      /* Here are the flags with arguments, sequenced as above */

      check_opt ("board", specific_board_num, specific_board_num >= 0);
      if (strcmp (argv[i], "-B") == 0)
	{
	  if (i + 1 >= argc)
	    {
	      usage ("Board number expected after '-B'.\n");
	      gm_exit (GM_FAILURE);
	    }
	  specific_board_num = atoi (argv[++i]);
	  if (specific_board_num < 0)
	    {
	      usage ("bad board number: %d\n", specific_board_num);
	      gm_exit (GM_FAILURE);
	    }
	  continue;
	}

#undef check_opt

      usage ("unknown arg: '%s'\n", argv[i]);
      gm_exit (GM_FAILURE);
    }
}



/****************************************************************
 * produce the actual output
 ****************************************************************/

static int
verbose_details_1(gm_port_t *port)
{
  gm_myrinet_eeprom_t eeprom;
  gm_status_t status;


  /* Copy the eeprom */
  status = _gm_get_eeprom (port, &eeprom);
  if (status != GM_SUCCESS)
    {
      gm_perror ("gm_board_info: error reading eeprom", status);
      return 1;
    }

  /* print the eeprom */

  if (eeprom.lanai_cpu_version >= 0xA00)
    {
      char board_config_strings[256];
      char *ptr;
      
      printf ("  lanai_cpu_version = 0x%04x ",
	      (unsigned int) eeprom.lanai_cpu_version);
      printf ("(LANai%d.%d)\n",(unsigned int) eeprom.lanai_cpu_version>>8,
	      (unsigned int)(eeprom.lanai_cpu_version&0xFF));
      printf ("  lanai_sram_size   = 0x%08x (%dK bytes)\n",
	      (unsigned int) eeprom.lanai_sram_size,
	      (unsigned int) eeprom.lanai_sram_size >> 10);

      /* Print the board config strings, which contain the rest of the
	 relevant board details. */
      
      status = _gm_get_board_config_strings (port, &board_config_strings);
      if (status != GM_SUCCESS)
	{
	  return 1;
	}
      printf ("ROM settings:\n");
      for (ptr = board_config_strings;
	   *ptr;
	   ptr += gm_strlen (ptr) + 1)
	{
	  printf ("  %s\n", ptr);
	}
    }
  else
    {
  printf ("  lanai_clockval    = 0x%08x\n",
	  (unsigned int) eeprom.lanai_clockval);
  printf ("  lanai_cpu_version = 0x%04x ",
	  (unsigned int) eeprom.lanai_cpu_version);
  printf ("(LANai%d.%d)\n",(unsigned int) eeprom.lanai_cpu_version>>8,
      (unsigned int)(eeprom.lanai_cpu_version&0xFF));
  printf ("  lanai_board_id    = %02x:%02x:%02x:%02x:%02x:%02x\n",
	  eeprom.lanai_board_id[0] & 0xff,
	  eeprom.lanai_board_id[1] & 0xff,
	  eeprom.lanai_board_id[2] & 0xff,
	  eeprom.lanai_board_id[3] & 0xff,
	  eeprom.lanai_board_id[4] & 0xff,
	  eeprom.lanai_board_id[5] & 0xff);
  printf ("  lanai_sram_size   = 0x%08x (%dK bytes)\n",
	  (unsigned int) eeprom.lanai_sram_size,
	  (unsigned int) eeprom.lanai_sram_size >> 10);

  if (eeprom.lanai_cpu_version < 0x0900)
    {
      printf ("  fpga_version      = \"%s\"\n", eeprom.fpga_version);
      printf ("  more_version      = \"%s\"\n", eeprom.more_version);
    }

  if (!((eeprom.max_lanai_speed == 0) ||
	(eeprom.max_lanai_speed == 0xFFFF)))
    printf ("  max_lanai_speed   = %d MHz\n", eeprom.max_lanai_speed);

  printf ("  product_code      = %d\n",
	  (unsigned int) eeprom.product_code);
  printf ("  serial_number     = %d\n",
	  (unsigned int) eeprom.serial_number);
  if (PRODUCT_CODE_OK (eeprom.product_code))
    {

      printf ("    (should be labeled: \"");
      if (strchr(boards[eeprom.product_code].name,'%'))
      	{
	  printf (boards[eeprom.product_code].name, eeprom.serial_number);
	}
      else 
	{
	  printf (boards[eeprom.product_code].name);
	  printf ("-%05d",eeprom.serial_number);
	}

      printf ("\")\n");
    }
  else
    {
      if (isprint (eeprom.board_label[0])
	  && isprint (eeprom.board_label[1]))
	{
	  char boardname[32];
	  char numberstring[8];
	  gm_bzero (boardname, 32);
	  sprintf(numberstring,"%05d",eeprom.serial_number);
	  strncpy ((char *) boardname, (char *) eeprom.board_label,
		   31);
	  if (strstr(boardname,numberstring))
	    {
	      printf ("    (should be labeled: \"%s\")\n", boardname);
	    }
	  else 
	    {
	      printf ("    (should be labeled: \"%s-%s\")\n",
			boardname,numberstring);
	    }
	}
      else
	{
	  printf ("    (Unknown board label)\n");
	}
    }
    }
  
  /* print the LANai time */

  {
    gm_u32_t high, low;
    gm_u64_t minutes;

    gm_u64_t ticks = gm_ticks (port);
    ticks = gm_ticks (port);
    high = (gm_u32_t) (ticks >> 32);
    low = (gm_u32_t) ticks;
    minutes = (ticks >> 27);	/* should be (ticks/60/2000000) */
    printf ("LANai time is 0x%0x%08x ticks,"
	    " or about %d minutes since reset.\n",
	    high, low, (int) minutes);
  }
  return (0);

} /* verbose_details_1() */


static void
verbose_details_2(gm_port_t *port)
{
  gm_pid_t opener_pids[GM_NUM_PORTS+1];
  int i;
  gm_pid_t my_pid = gm_getpid ();

  for (i = 0; i < GM_NUM_PORTS; i++)
    {
      opener_pids[i] = 0;
    }
  if (_gm_get_opener_pids (port, opener_pids) == GM_SUCCESS)
    {
      printf ("Port: Status  PID\n");
      for (i = 0; i < GM_NUM_PORTS; i++)
        {
	  if (opener_pids[i])
	    {
	      printf ("%4d: %6s %5d  ", i, "BUSY", opener_pids[i]);
	      if (opener_pids[i] == my_pid)
	        printf ("%s", "(this process [gm_board_info])");
	      if (opener_pids[i] == 1)
	        printf ("%s", "(ip-over-GM)");

	      printf ("\n");
	    }
/*
	  else
	    printf ("%4d: %6s\n", i, "free");
*/
	}
    }
}


static gm_status_t
open_board_num(int the_board_num, gm_port_t **the_port)
{
  /* Try to open this board num, using each possible port in turn */
  /* Returns gm_open() status; returns GM_BUSY only if all ports busy, */
  /* and GM_PERMISSION_DENIED only if this error occurs on the final port */

  int port_num;
  gm_status_t status = GM_FAILURE;

  /* Scan for the first not-busy port, or report error */
  for (port_num = 0; port_num < GM_NUM_PORTS; port_num++)
    {
      status = gm_open (the_port, the_board_num, port_num, "gm_board_info",
			(enum gm_api_version) GM_API_VERSION);
      if ((status != GM_BUSY) && (status != GM_PERMISSION_DENIED))
	{
	  break;
	}
    }
  return (status);
}

/****************
 * Error recording & reporting
 *
 * We record any non-fatal errors encountered in 
 ****************/

/* A chainable error report */

struct error_record
{
  struct error_record *next;
  char buf[80];
} *first_error_record;

/* Record a string (up to 80 bytes) describing an error */

#define RECORD_ERROR(string) do {					\
  static struct error_record record;					\
									\
  gm_always_assert (gm_strlen (string) < sizeof (record.buf));		\
  strcpy (record.buf, string);						\
  record.next = first_error_record;					\
  first_error_record = &record;						\
} while (0);

/* Print all recorded error messages. */

static
void
report_errors (void)
{
  while (first_error_record)
    {
      fputs ("*** ", stdout);
      puts (first_error_record->buf);
      first_error_record = first_error_record->next;
    }
}

static int
one_board_info(int the_board_num, gm_port_t *the_port)
{
  char error_report[80];
  char host_name_buf[GM_MAX_HOST_NAME_LEN+1];
  gm_status_t  status;
  int          found_routes = 0;
  struct gm_mapper_state mapper_state;
  unsigned int page_hash_cache_size=0;
  unsigned int node_id,
               max_node_id,
               max_node_id_in_use = 0;
  unsigned int this_node = 0;
  gm_u8_t my_name[GM_MAX_HOST_NAME_LEN + 1] = { 0 };
  
  if (info_level >= 0)
    {
      /* WARNING: this print is parsed by scripts/etc/init.d/gm, and
	 should not be casually modified. */
      printf ("\n\nBoard number %d:\n", the_board_num);
      if (info_level > 0)
	{
	  if (verbose_details_1(the_port) != 0)
	    {
	      gm_close (the_port);
	      return 1;
	    }
	}
    }

  /* Get the mapper state. */

  status = _gm_get_mapper_state (the_port, &mapper_state);
  if (status != GM_SUCCESS)
    {
      fprintf (stderr, "Could not get mapper state.\n");
      gm_exit (status);
    }

  /* Optionally, wait for the mapper to configure the network. */

  if (wait_for_mapper)
    {
      printf ("Waiting for network configuration.\n");
      while (!mapper_state.network_configured)
	{
	  gm_sleep (1);
	  status = _gm_get_mapper_state (the_port, &mapper_state);
	  if (status != GM_SUCCESS)
	    {
	      fprintf (stderr, "Could not get mapper state.\n");
	      gm_exit (status);
	    }
	}
    }

  /* Print the mapper state */

  printf ("Mapper is %02x:%02x:%02x:%02x:%02x:%02x.\n",
	  mapper_state.mapper_mac[0],
	  mapper_state.mapper_mac[1],
	  mapper_state.mapper_mac[2],
	  mapper_state.mapper_mac[3],
	  mapper_state.mapper_mac[4],
	  mapper_state.mapper_mac[5]);
  printf ("Map version is %u.\n", mapper_state.map_version);
  printf ("%d hosts.\n", mapper_state.num_hosts);
  printf ("Network is %sfully configured.\n",
	  mapper_state.network_configured ? "" : "NOT ");

  /* print the routes to the other nodes in the network */

  if (gm_get_node_id (the_port, &this_node) != GM_SUCCESS)
    fprintf (stderr, "Could not determine this node ID.\n");
  my_name[0] = 0;
  (void) gm_get_host_name (the_port, (char *) my_name);
  if (info_level >= 0)
    printf ("This node is \"%s\"\n", my_name);

  if (this_node == GM_NO_SUCH_NODE_ID)
    {
      printf ("   *** Node ID not set, mapper not yet run?\n");
    }

  status = gm_max_node_id (the_port, &max_node_id);
  if (status != GM_SUCCESS)
    {
      fprintf (stderr, "Could not determine max node ID.\n");
      gm_exit (status);
    }

  if (gm_max_node_id_in_use (the_port, &max_node_id_in_use) != GM_SUCCESS)
    {
      fprintf (stderr,
	       "\n******** Could not determine max_node_id_in_use. *****\n");
      max_node_id_in_use = 16;
      fprintf (stderr,
	       "******** Will only search for routes up to id = %d\n\n",
	       max_node_id_in_use);
    }

  _gm_get_page_hash_cache_size(the_port, &page_hash_cache_size);
  if (info_level >= 0)
    {
      printf ("Board has room for %d ports,  %d nodes/routes,"
	      "  %d cache entries\n"
	      "          Port token cnt: send=%d, recv=%d\n",
	      gm_num_ports(the_port), max_node_id, page_hash_cache_size+1,
	      gm_num_send_tokens(the_port), gm_num_receive_tokens(the_port));

      if (info_level > 0)
	verbose_details_2(the_port);

      printf ("Route table for this node follows:\n");
      printf
	("gmID MAC Address                                 gmName Route\n");
      printf
	("---- ----------------- --------------------------------"
	 " ---------------------\n");

      for (node_id = 0; node_id <= max_node_id_in_use; node_id++)
	{
	  char unique[6];
	  const char *host_name;

	  /* Report any errors from the previous iteration. */
	  report_errors ();
	  
	  /* Get and print the node IDs */
	  status = gm_node_id_to_unique_id (the_port, node_id, unique);
	  if (status == GM_INVALID_PARAMETER)
	    {
	      continue;
	    }
	  else if (status != GM_SUCCESS)
	    {
	      printf ("%4d ??:??:??:??:??:?? ", node_id);
	    }
	  else
	    {
	      found_routes = 1;
	      printf ("%4d %02x:%02x:%02x:%02x:%02x:%02x ",
		      (int) node_id,
		      unique[0] & 0xff, unique[1] & 0xff,
		      unique[2] & 0xff, unique[3] & 0xff,
		      unique[4] & 0xff, unique[5] & 0xff);

	      /* Verify the reverse (MAC->node_id) translation. */
	      
	      if (verify)
		{
		  unsigned int check;

		  status = gm_unique_id_to_node_id (the_port, unique, &check);
		  if (status != GM_SUCCESS)
		    {
		      sprintf (error_report, "Could not verify unique mac id.");
		      RECORD_ERROR (error_report);
		    }
		  else if (check != node_id)
		    {
		      sprintf
			(error_report,
			 "node_id->unique_id->node_id is 0x%x.", (int) check);
		      RECORD_ERROR (error_report);
		    }
		}
	    }

	  /* print the host name in a 32 character field, portably. */

	  if (yp)
	    {
	      status
		= (gm_node_id_to_host_name_ex
		   (the_port, 100000, node_id, &host_name_buf));
	      if (status != GM_SUCCESS)
		{
		  sprintf (error_report, "Could not get host name: %s",
			   gm_strerror (status));
		  RECORD_ERROR (error_report);
		  host_name = 0;
		}
	      else
		{
		  host_name = &host_name_buf[0];
		}
	    }
	  else
	    {
	      host_name="[requires YP]";
	    }
	  
	  {
	    const char *n;

	    n = host_name ? host_name : "????";
	    printf ("                                %s"
		    + strlen (n) % 32, n);
	  }

	  /* Verify the reverse (host_name->node_id) lookup. */
	  
	  if (host_name && yp && verify)
	    {
	      unsigned int check;

	      status
		= (gm_host_name_to_node_id_ex
		   (the_port, 100000, host_name, &check));
	      if (status != GM_SUCCESS)
		{
		  sprintf (error_report, "Could not verify host name id:");
		  RECORD_ERROR (error_report);
		  RECORD_ERROR (gm_strerror (status));
		}
	      else if (check != node_id)
		{
		  sprintf
		    (error_report,
		     "node_id->host_name->node_id is 0x%x.",
		     (int) check);
		  RECORD_ERROR (error_report);
		}
	    }
	  
	  /* Print the route to the node, if any. */

	  {
	    gm_u8_t route_buf[GM_MAX_NETWORK_DIAMETER];
	    unsigned int route_len = GM_MAX_NETWORK_DIAMETER;
	    int i;

	    for (i = 0; i < GM_MAX_NETWORK_DIAMETER; i++)
	      route_buf[i] = -1;

	    status = _gm_get_route (the_port, node_id, (char *) route_buf,
				    &route_len);
	    if (status != GM_SUCCESS)
	      {
		sprintf (error_report,
			 "Could not get route for node %d (error 0x%x).",
			 (int) node_id, (int) status);
		RECORD_ERROR (error_report);
		continue;
	      }

	    if (route_len && (route_len != GM_NO_ROUTE))
	      {
		int j = 0;
		char bytestr[8];

		while (route_len--)
		  {
		    if (show_route_in_hex)
		      {
			printf (" %02x", route_buf[j++]);
		      }
		    else
		      {
			route_byte_string (route_buf[j++], bytestr);
			printf (" %s", bytestr);
		      }
		  }
	      }
	  }

	  if (node_id == this_node)
	    printf (" (this node)");

	  if (!gm_memcmp (unique, mapper_state.mapper_mac, sizeof (unique)))
	    {
	      printf (" (mapper)");
	    }
	  
	  printf ("\n");
	}
      /* Report any errors in the last iteration. */
      report_errors ();
      
      if (!found_routes)
	printf ("   *** No routes found *** \n");
    }

  /* close the port */
  gm_close (the_port);

  return 0;

}


static int
gm_board_info (int argc, char *argv[])
{
  int         board_num;
  gm_port_t   *port;
  gm_status_t status;

  /* The following define specifies how many consecutive board numbers which do
   * not identify boards on this host will indicate that we are "done."
   */
#define MAX_BOARD_GAP 4

  parse_args(argc, argv);

  /* Explicitly initialize GM to suppress excessive debugging messages
     when opening and closing ports. */
  
  status = gm_init();
  if (status != GM_SUCCESS)
    {
      gm_exit (status);
    }
  
  if (info_level >= 0)
    printf ("GM build ID is \"%s.\"\n", _gm_build_id);

  /* Check some performance-related assumptions that do not affect
     correct operation of GM. */

  CHECK_SIZING (gm_port_protected_lanai_side);
  CHECK_SIZING (gm_connection);

  if (specific_board_num != -1)
    {
      /* Open the specified Myrinet interface and print its description */
      status = (open_board_num(specific_board_num, &port));
      switch (status)
	{
	case GM_SUCCESS:
	  one_board_info(specific_board_num, port);
	  break;

	case GM_BUSY:
	  printf ("Cannot open board %d; all ports are busy\n",
		  specific_board_num);
	  gm_exit (status);
	  break;		/* not reached */

	default:
	  printf ("gm_board_info: error %d opening board %d\n",
		  status, specific_board_num);
	  gm_perror ("gm_board_info", status);
	  gm_exit (status);
	}
    }

  else /* Scan for all Myrinet interfaces and print their descriptions */

    { /* If no open errors, the for loop will terminate after MAX_BOARD_GAP
       * consecutive board numbers have returned GM_DEV_NOT_FOUND.  If
       * (!quit_on_error), the behavior is the same.  If (quit_on_error), the
       * for loop will terminate on the first error that is not GM_BUSY.  For
       * the purpose of this discussion, GM_DEV_NOT_FOUND is not an error.
       */
      gm_status_t first_error = GM_SUCCESS; /* GM_SUCCESS ==> no error */
      int found = 0;
      int last_good_board_num = 0;

      for (board_num = 0; ; board_num++)
	{

	  status = open_board_num(board_num, &port);
	  switch (status)
	    {
	    case GM_SUCCESS:
	      last_good_board_num = board_num;
	      one_board_info(board_num, port);
	      found++;
	      break;

	    case GM_DEV_NOT_FOUND:
	    case GM_NO_SUCH_DEVICE:
	      break;

	    case GM_BUSY:
	    default:
	      if (status == GM_BUSY)
		{
		  last_good_board_num = board_num;
		  printf ("Cannot open board %d; all ports are busy\n",
			  board_num);
		  found++;
		}
	      else
		{
		  printf ("gm_board_info: error %d opening board %d\n",
			  status, board_num);
		  gm_perror ("gm_board_info", status);
		}

	      if (quit_on_error)
		{
		  gm_exit (status);
		}
	      else
		{
		  if (first_error == GM_SUCCESS)
		    {
		      first_error = status;
		    }
		}
	    } /* switch */

	  if ((board_num - last_good_board_num) > MAX_BOARD_GAP)
	    {
	      if (!found)
		{
		  printf ("No boards found\n");
		  gm_exit (GM_UNATTACHED);
		}
	      break;
	    }

	} /* for */

      if (first_error != GM_SUCCESS)
	{
	  gm_perror ("*** gm_board_info encountered errors", first_error);
	}

      gm_exit (first_error);

    } /* else Scan */

  /* Not reached, but return something to prevent compiler warnings. */
  
  return 0;
}
#undef MAX_BOARD_GAP


/****************************************************************
 * Entry point
 ****************************************************************/

#define GM_MAIN gm_board_info
#include "gm_main.h"

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
