/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2003 by Myricom, Inc.					 *
 * All rights reserved.  See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* This program tests for a bug in GM cluster maintenance where the
   firmware is reading from a page hash table collision cluster while
   that cluster is being updated to remove an entry.  To do this, this
   test allocates buffers that all hash to the same bin (assuming the
   driver is freshly loaded so GM is using only a single page of PTEs
   as the page hash) and that cross a page boundary.  It then
   dynamically registers and deregisters these buffers as they are
   used for sends, trying to catch PTE problems. */
   
/* author: glenn@myri.com */

#include <stdio.h>

#include "gm.h"
#include "gm_page_hash.h"

/* Total number of colliding buffers to use. */
   
#define NBUFFS 16

/* Total number of packets to send. */

#define ITERS (1<<20)

/* Size to use for GM sends */

#define SIZE 12

#define PORT_ID 2

#if 0
#define PRINTF(x) printf x
#else
#define PRINTF(x)
#endif

/****************
 * Buff manipulation
 ****************/

/* A buff used as a send or receive buffer. */

struct buff
{
  struct buff *next;
  gm_u32_t registered;		/* boolean */
  gm_u32_t expected;
  gm_u32_t pad[1000];		/* ensures more time to lose race */
  /* --- This will be a page boundary. --- */
  gm_u32_t got;
};

static struct buff *send_buffs;	/* free send buffers */
static struct buff *recv_buffs;	/* free recv buffers */
static unsigned int send_tokens; /* available GM send tokens */

/* Fill a buff with test values. */

static void
buff_fill (struct buff *buff)
{
  static gm_u32_t i;

  buff->expected = i;
  buff->got = i;
  
  i++;
}

/* Return 1 if a buff is properly filled with test values. */

static int
buff_filled (struct buff *buff)
{
  return (buff->expected == buff->got);
}

/* Make the buff look unfilled */

static void
buff_unfill (struct buff *buff)
{
  buff->expected = 0;
  buff->got = 1;
}

/****************
 * Colliding buffs support (buffs that hash to the same bin)
 ****************/

/* The jumbo allocation holding a few useful buffs, but mostly holding
   pages that are never touched. */
static char *allocated_pages;

/* Deregister each of the buffs and zero the list. */

static
void
_free_colliding_buffs (struct gm_port *port, struct buff **l)
{
  struct buff *next;
  
  while (*l)
    {
      next = (*l)->next;
      if ((*l)->registered)
	{
	  gm_deregister_memory (port, *l, GM_PAGE_LEN);
	}
      *l = next;
    }
}

/* Free the allocated colliding buffs. */

static
void
free_colliding_buffs (struct gm_port *port)
{
  _free_colliding_buffs (port, &send_buffs);
  _free_colliding_buffs (port, &recv_buffs);
  gm_free (allocated_pages);
}

/* Allocate send and receive buffers that collide in the hash table,
   assuming that the hash table only has one page of PTEs.  (This is
   true if the GM driver has been freshly loaded and has the mapper
   did not register tons of memory. */

static
gm_status_t
alloc_colliding_buffs (struct gm_port *port)
{
  char *aligned_pages;
  struct buff *buff;
  struct buff *useless = 0;
  unsigned int buff_cnt = 0;		/* useful buffs found */
  unsigned int port_id = gm_get_port_id (port);
  unsigned long alloc_len;
  unsigned long offset;
  unsigned long mask = (GM_PAGE_LEN / 16) - 1;

  /* Allocate enough buffs that we expect NBUFFS to be useful.  We use
     one big allocation because most of the allocation to avoid
     touching most of the pages. */

  for (alloc_len = NBUFFS*(GM_PAGE_LEN/16)*GM_PAGE_LEN;
       buff_cnt < NBUFFS;
       alloc_len += alloc_len/4)
    {
      if (allocated_pages)
	gm_free (allocated_pages);
      send_buffs = 0;
      buff_cnt=0;

      /* Allocate alloc_len pages, plus an extra page so we can
	 round up to a page boundary, plus an extra page so we
	 can round up so the buffers cross a page boundary. */
      
      allocated_pages = gm_malloc (alloc_len + 2*GM_PAGE_LEN);
      if (!allocated_pages)
	{
	  goto abort_with_nothing;
	}

      /* Align the allocation on a page boundary. */
      
      aligned_pages = (allocated_pages
		       - ((gm_size_t) allocated_pages & (GM_PAGE_LEN-1))
		       + GM_PAGE_LEN);
      for (offset=0; offset<=alloc_len-GM_PAGE_LEN; offset+=GM_PAGE_LEN)
	{
	  /* Consider buffers that start near the end of the page such
	     that the "got" field of the buffer is on the other side
	     of a page boundary. */
	  
	  buff = ((struct buff *)
		  &aligned_pages[offset
				 + GM_PAGE_LEN
				 - GM_OFFSETOF (struct buff, got)]);
	  if ((GM_HASH_PAGE_PORT (((gm_up_t) (gm_size_t) aligned_pages
				  + offset + port_id))
	       & mask)
	      == 0)
	    {
	      /* This buff collides.  Save it. */

	      PRINTF (("buff %p\n", buff));
	      buff->next = send_buffs;
	      send_buffs = buff;
	      buff_cnt++;
	    }
	}
    }

  /* Discard any extra buffs */
  
  while (buff_cnt > NBUFFS)
    {
      send_buffs = send_buffs->next;
      buff_cnt--;
    }
  
  /* Initialize the buffs */

  for (buff=send_buffs; buff; buff=buff->next)
    {
      buff->registered = 0;
      buff_fill (buff);
    }
  
  /* Move half of the useable buffs into the recv_buffs list. */

  while (buff_cnt >= 2)
    {
      struct buff *moving;

      moving = send_buffs;
      send_buffs = moving->next;
      moving->next = recv_buffs;
      recv_buffs = moving;
      
      buff_cnt -= 2;
    }

  return GM_SUCCESS;

 abort_with_nothing:
  return GM_OUT_OF_MEMORY;
}

/****************
 * Random number generation
 ****************/

/* Simulate a coin toss */

static
int
heads (void)
{
  return gm_rand () & 32;
}

/****************
 * Core program
 ****************/

/* Process completed sends. */

static unsigned int sending;

static
void
send_done (struct gm_port *port, void *context, gm_status_t status)
{
  struct buff *buff;

  gm_always_assert (status == GM_SUCCESS);
  
  buff = (struct buff *) context;

  /* Possibly verify the send buffer */
  
  gm_always_assert (buff_filled (buff));

  /* deregister the buff */

  if (heads ())
    {
      PRINTF (("deregistering %p\n", buff));
      gm_deregister_memory (port, buff, GM_PAGE_LEN);
      buff->registered = 0;
    }
  
  /* Free the send buffer */
  
  buff->next = send_buffs;
  send_buffs = buff;

  /* Record that it is safe to send again. */

  send_tokens++;
  sending--;

  PRINTF (("sent %p (%u outstanding)\n", buff, sending));
}

/* Send and receive from dynamically registered buffs. */

static
void
collide (struct gm_port *port)
{
  gm_status_t status = GM_SUCCESS;
  unsigned long recvs = ITERS;
  unsigned long sends = ITERS;
  unsigned int receiving = 0;

  /* Send until done. */
  
  while (recvs)
    {
      struct buff *buff;

      /* Possibly post a receive. */
      
      if (recv_buffs && recvs && heads ())
	{
	  /* Register the buff if needed. */
	  
	  buff = recv_buffs;
	  if (!buff->registered)
	    {
	      PRINTF (("registering %p\n", buff));
	      status = gm_register_memory (port, buff, GM_PAGE_LEN);
	      gm_always_assert (status == GM_SUCCESS);
	      buff->registered = 1;
	    }

	  /* Remove the buff from the free list. */
	     
	  recv_buffs = buff->next;

	  /* Mark the receive buffer as empty. */

	  buff_unfill (buff);

	  /* Post the receive */
	  
	  gm_always_assert (status == GM_SUCCESS);
	  gm_provide_receive_buffer (port, buff, SIZE, GM_LOW_PRIORITY);
	  receiving++;

	  PRINTF (("posted recv %p (%u outstanding)\n", buff, receiving));
	}

      /* Possibly send something. */
      
      if (send_buffs && sends && send_tokens && heads())
	{
	  /* Register the buff if needed. */

	  buff = send_buffs;
	  if (!buff->registered)
	    {
	      PRINTF (("registering %p\n", buff));
	      status = gm_register_memory (port, buff, GM_PAGE_LEN);
	      gm_always_assert (status == GM_SUCCESS);
	      buff->registered = 1;
	    }

	  /* Remove the buff from the free list. */
	  
	  send_buffs = buff->next;

	  /* Reserve a send token. */
	  
	  send_tokens--;
	  
	  /* Send the buff */
	  
	  gm_send_with_callback (port, buff, SIZE, sizeof (struct buff),
				 GM_LOW_PRIORITY, 1, PORT_ID,
				 send_done, buff);
	  sending++;
	  PRINTF (("posted send %p (%u outstanding)\n", buff, sending));
	}

      /* Possible receive something, but only if a send and receive
	 are posted. */

      if (sending && receiving && heads ())
	{
	  gm_recv_event_t *e;
	  
	  e = gm_blocking_receive_no_spin (port);
	  switch (gm_ntoh_u8 (e->recv.type))
	    {
	    case GM_RECV_EVENT:
	      
	      buff = gm_ntohp (e->recv.buffer);
	      
	      /* verify the contents of the receive. */

	      gm_always_assert (buff_filled (buff));
	      
	      /* Record the receive completion. */
	      
	      receiving--;
	      recvs--;
	      if (GM_POWER_OF_TWO (ITERS-recvs))
		{
		  printf ("%lu receives completed\n", ITERS-recvs);
		}
	      
	      /* Free the recv buff. */
	      
	      buff->next = recv_buffs;
	      recv_buffs = buff;
	      
	      PRINTF (("received %p (%u outstanding)\n", buff, receiving));
	      break;

	    default:
	      gm_unknown (port, e);
	      break;
	    }
	}
    }
}

int
main (int argc, char *argv[])
{
  gm_status_t status;
  struct gm_port *port;

  /* Open a GM port. */
  
  status = gm_open (&port, 0, PORT_ID, argv[0], GM_API_VERSION);
  if (status != GM_SUCCESS)
    {
      goto abort_with_nothing;
    }
  send_tokens = gm_num_send_tokens (port);
  
  /* Allocate buffs that all hash to the same PTE. */
  
  status = alloc_colliding_buffs (port);
  if (status != GM_SUCCESS)
    {
      goto abort_with_open;
    }

  /* Exercise the hash table with sends, registrations, and
     deregistrations. */
  
  collide (port);

  free_colliding_buffs (port);
  
  /* fall through */
  
 abort_with_open:
  gm_close (port);
 abort_with_nothing:
  if (status == GM_SUCCESS)
    {
      printf ("SUCCESS\n");
    }
  else
    {
      printf ("FAILED: %s", gm_strerror (status));
    }
  gm_exit (status);
  return 1;
}
