/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1999 by Myricom, Inc.                                   *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* Description
 *
 *   gm_debug always displays its estimate of DMA rates.  Then, if
 *   GM_ENABLE_ERROR_COUNTERS was set at build time, it will display some
 *   LANai counters; the display will be longer if GM_ENABLE_DEBUG_COUNTERS
 *   was set.  Following this, any additional output as specified by the
 *   flags listed in function usage() will be generated.
 */

#if defined __STDC__ || defined WIN32
#if GM_OS_VXWORKS
#include "vxWorks.h"
#endif  /* GM_OS_VXWOKRS */
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

/* This program uses undocumented software interfaces that are not
   intended for use in code written outside Myricom.  Hence the
   following undocumented includes. */

#include "gm_internal.h"
#include "gm_debug_counters.h"
#include "gm_enable_debug_counters.h"
#include "gm_enable_error_counters.h"
#include "gm_enable_fork_system.h"
#include "gm_enable_log_dispatch.h"
#include "gm_enable_packet_counters.h"
#include "gm_error_counters.h"
#include "gm_dma_perf.h"
#if GM_CPU_mips
#include "gm_arch_debug.h"
#endif
#include "gm_lanai.h"


/* Following macros needed by GM_ERROR_COUNTERS and GM_DEBUG_COUNTERS */
#define GM_ERROR_CNT(name,desc) _P_NAMED ("", name, name ## _error_cnt);
#define GM_DEBUG_CNT(name,desc) _P_NAMED ("", name, name ## _debug_cnt);

#define _P_NAMED(prefix, name, fld) do {								\
	printf (prefix "  %-40s %12u\n",									\
            #name, gm_ntoh_u32 (globals->fld));							\
  } while (0)

#define P(prefix, fld) _P_NAMED (prefix, fld, fld)

/* this 1024 was hard-coded in drivers/gm.c */
/* don't really want to get all the data from a card with 2Meg+ */
#define MY_MAX_NODE	(3*1024)

/****************************************************************
 * global data
 ****************************************************************/

int board_num = 0,
    port_num  = 0;
	int do_abort           = 0,
	    do_print_dma_rates = 1,
        do_print_info      = 0,
        do_print_log       = 0,
        do_print_lanai_cts = 1,
        do_repeat          = 0,
		do_timestamp       = 0,
	    do_dump            = 0;
	int sleep_secs     = 0;
	int max_connection = 0;
	int one_connection = 0;

#if	GM_OS_AIX

struct gm_lanai_globals *globals = NULL;

#else	/* GM_OS_AIX */

unsigned int space[(sizeof(struct gm_lanai_globals) +
		       ((MY_MAX_NODE+1) * sizeof(struct gm_connection))) / sizeof(int)];

struct gm_lanai_globals *globals = (struct gm_lanai_globals *) space;

#endif	/* GM_OS_AIX */

gm_u64_t current_rtc;

gm_port_t *port;

/****************************************************************
 * Command line processing
 ****************************************************************/

#define TP(a) fprintf (stderr, a)

static void
print_synopsis(void)
{
  TP ("\n");
  TP ("Myricom 'gm_debug' displays diagnostic statistics about the\n");
  TP ("operation of the Myrinet interface.  This program is used almost\n");
  TP ("exclusively for problem diagnosis.\n");
}


static void
print_options (void)
{
  TP ("\nusage: gm_debug [--board=board] [--port=port] <other options>\n");

  TP ("\nDevice Control and Selection:\n");
  TP ("-B, --board=NUM  [0]   board number to debug\n");
  TP ("-p, --port=NUM   [0]   port number to use\n");

  TP ("\nInformation Display Selection:\n");
  TP ("    --connection=NUM   [off] dump specific connection NUM\n");
  TP ("-C, --connections=NUM  [off] dump all connections from 1 through NUM\n");
  TP ("-i, --internal-info    [off] display internal debugging information\n");
  TP ("    --lanai-dispatches [off] display LANai dispatch information (requires\n");
  TP ("                             MCP built with GM_ENABLE_LOG_DISPATCH\n");
  TP ("-L, --no-counters      [off] omit standard display of LANai counters\n");

  TP ("\nLooping and Time-stamping:\n");
  TP ("-R, --repeat     [off] repeat all selected information in a loop\n");
  TP ("    --sleep=NUM  [0]   sleep NUM seconds between each repetition\n");
  TP ("    --timestamp  [off] display time of day before each repetition\n");

  TP ("\nInformative Options:\n");
  TP ("    --abort       instead of normal termination, call 'abort()', which\n");
  TP ("                  should produce a dump ('core' file)\n");
  TP ("    --help        output usage and options information, then exit\n");
  printf ("\n");

#undef TP
}


static void
#ifdef va_dcl
usage (fmt, va_alist)
     const char *fmt;
     va_dcl
#else
usage (const char *fmt, ...)
#endif
{
  va_list ap;

  print_options();

/* Now print the caller's diagnostic */
#if defined __STDC__ || defined WIN32
  va_start (ap, fmt);
#else
  va_start (ap);
#endif
  vfprintf (stderr, fmt, ap);
  va_end (ap);

  gm_exit (GM_INVALID_PARAMETER);

} /* usage() */


static void
parse_args (int argc, char *argv[])
{
/* NOTE: If you change the operating characteristics or the flags
 *       used by this program, please notify Max so he can change the
 *       Myrinet Handbook document.
 */
  int i;

  /* The 'F' macro is for all flags with no argument (parameter)     */
#define F(name,action) if (strcmp (argv[i], name) == 0) { action; continue; }

  for (i = 1; i < argc; i++)
  {
      /* Dispose of the --help flag first                                */
      if (strcmp (argv[i], "--help") == 0)
	  {
		  print_synopsis();
		  print_options();
		  gm_exit (GM_SUCCESS);
	  }

      /* Here are the flags with no argument.                            */
      /* Please keep short form, if any, after its long form equivalent, */
	  /* but keep the long form options in alpha sequence.               */

	  F ("--abort"        , do_abort           = 1);
	  F ("--dump"         , do_dump            = 1);
	  F ("-d"             , do_dump            = 1);
	  F ("--internal-info", do_print_info      = 1);
	  F ("-i"             , do_print_info      = 1);
	  if ((strcmp(argv[i], "--lanai-dispatches") == 0) ||
		  (strcmp(argv[i], "-l")                 == 0)    )    /* Deprecated */
	  {
		  if (GM_ENABLE_LOG_DISPATCH)
			  do_print_log = 1;
		  else
			  printf("flag %s ignored because LANai is not logging dispatches\n",
					 argv[i]);
		  continue;
	  }
	  F ("--no-counters"  , do_print_lanai_cts = 0);
	  F ("-L"             , do_print_lanai_cts = 0);
	  F ("--repeat"       , do_repeat          = 1);
	  F ("-R"             , do_repeat          = 1);
	  F ("--timestamp"    , do_timestamp       = 1);
	  F ("-t"             , do_timestamp       = 1);      /* "-t" Deprecated */
#undef F

      /* The check_opt macro is for long-form options with numeric        */
      /* arguments and simple validation tests.                           */
#define check_opt(prefix, var, test)							\
	  if (sscanf (argv[i], "--" prefix "=%i", &var) == 1)		\
	  {															\
	      if (!(test))											\
		  {														\
			  usage ("bad value for option \"%s\"\n", argv[i]);	\
		  }														\
	      continue;												\
	  }

      /* Here are the flags with arguments, sequenced as above */

      check_opt ("board", board_num, board_num >= 0);
      if (strcmp (argv[i], "-B") == 0)
	  {
		  if (i + 1 >= argc)
			  usage ("Board number expected after '-B'\n");

		  board_num = atoi (argv[++i]);
		  if (board_num < 0)
			  usage ("bad board number: %d\n", board_num);

		  continue;
	  }

      check_opt ("connection", one_connection, one_connection > 0);
      if (strcmp (argv[i], "-c") == 0)                         /* Deprecated */
	  {
		  if (i + 1 >= argc)
			  usage ("Connection number expected after '-c'\n");

		  one_connection = atoi (argv[++i]);
		  if (one_connection <= 0)
			  usage ("bad 'one connection' number: %d\n", one_connection);

		  continue;
	  }

      check_opt ("connections", max_connection, max_connection > 0);
      if (strcmp (argv[i], "-C") == 0)
	  {
		  if (i + 1 >= argc)
			  usage ("Max connection number expected after '-C'\n");

		  max_connection = atoi (argv[++i]);
		  if (max_connection <= 0)
			  usage ("bad 'max connection' number: %d\n", max_connection);

		  continue;
	  }

      /* Note that at present the parameter to gm_num_ports is a dummy */
      check_opt ("port", port_num,
				 (port_num >= 0 && port_num < gm_num_ports ((gm_port_t *)0)));
      if (strcmp (argv[i], "-p") == 0)
	  {
		  if (i + 1 >= argc)
			  usage ("Port number expected after '-p'.\n");

		  port_num = atoi (argv[++i]);
		  if (port_num < 0 || port_num >= gm_num_ports ((gm_port_t *)0))
			  usage ("bad port number: %d\n", port_num);

		  continue;
	  }

      check_opt ("sleep", sleep_secs, sleep_secs >= 0);
      if (strcmp (argv[i], "-s") == 0)                         /* Deprecated */
	  {
		  if (i + 1 >= argc)
			  usage ("Sleep seconds expected after '-s'\n");

		  sleep_secs = atoi (argv[++i]);
		  if (sleep_secs < 0)
			  usage ("bad 'sleep seconds' number: %d\n", sleep_secs);

		  continue;
	  }
#undef check_opt

      usage ("unknown arg: '%s'\n", argv[i]);    /* Function does not return */

  }
}



/****************************************************************
 * produce the actual output
 ****************************************************************/

static void
print_debug_info(void)
{
#define P_CPU(name)             \
    if (GM_CPU_##name) {         \
	   printf (" %s\n", #name); \
       cpu_printed = 1; }


  int cpu_printed = 0;


  printf ("GM version is %s,\n build ID is \"%s.\"\n",
          _gm_get_version (), _gm_get_build_id ());
  printf ("Built for");
  P_CPU(alpha);
  P_CPU(hppa);
  P_CPU(ia64);
  P_CPU(lanai);
  P_CPU(mips);
  P_CPU(powerpc);
  P_CPU(sparc);
  P_CPU(sparc64);
  P_CPU(x86);
  if (!cpu_printed)
     printf(" *** unknown ***\n");

#if GM_CPU_mips
  if (GM_IRIX_FORCE_DMA_CMD)
     printf("WARNING: driver forces use of PCIIO_DMA_CMD for all DMA areas\n");
  if (GM_IRIX_FORCE_DMA_DATA)
     printf("WARNING: driver forces use of PCIIO_DMA_DATA for all DMA areas\n");
#endif

#ifdef GM_DEBUG
  printf ("GM_DEBUG = %d\n", GM_DEBUG);
#else
  printf ("GM_DEBUG is undefined\n");
#endif
  printf ("Print level is %d\n", GM_PRINT_LEVEL);
  printf ("Sized types:\n"
		  "    gm_s64_t  %d\n"
		  "    gm_up_t   %d\n"
		  "    gm_dp_t   %d\n"
		  "    gm_remote_ptr_t   %d\n"
		  "    gm_size_t %d\n"
		  "    void *    %d\n"
		  "    long      %d\n",
		  (int)sizeof(gm_s64_t),  (int)sizeof(gm_up_t), (int)sizeof(gm_dp_t),
		  (int)sizeof(gm_remote_ptr_t),
		  (int)sizeof(gm_size_t), (int)sizeof(void *),  (int)sizeof(long));
  printf ("Some defines:\n"
		  "    GM_SIZEOF_VOID_P     %d\n"
		  "    GM_SIZEOF_UP_T       %d\n"
		  "    GM_SIZEOF_DP_T       %d\n"
		  "    GM_NUM_PORTS         %d\n"
		  "    GM_DMA_GRANULARITY   %d\n"
		  "    GM_RDMA_GRANULARITY  %d\n"
		  "    GM_IP_MTU            %d\n",
		  GM_SIZEOF_VOID_P, GM_SIZEOF_UP_T,     GM_SIZEOF_DP_T,
		  GM_NUM_PORTS,     GM_DMA_GRANULARITY, GM_RDMA_GRANULARITY,
		  GM_IP_MTU);
  printf ("Some structure sizes:\n"
		  "    gm_port_protected_lanai_side_t  %d\n"
		  "    gm_pte_t                        %d\n"
		  "    gm_page_hash_cache_t            %d\n",
		  (int)sizeof(gm_port_protected_lanai_side_t),
		  (int)sizeof(gm_pte_t),
		  (int)sizeof(gm_page_hash_cache_t));
  printf ("    gm_cached_pte_t                 %d\n"
		  "    gm_recv_queue_slot_t            %d\n"
		  "    gm_recv_event_t                 %d\n"
		  "    struct gm_lanai_globals         %d (0x%x)\n"
		  "    struct gm_connection            %d\n",
		  (int)sizeof(gm_cached_pte_t),
		  (int)sizeof(gm_recv_queue_slot_t),
		  (int)sizeof(gm_recv_event_t),
		  (int)sizeof(struct gm_lanai_globals),
		  (int)sizeof(struct gm_lanai_globals),
		  (int)sizeof(struct gm_connection));

#ifdef GM_SUPPORT_L4
  printf ("GM_SUPPORT_L4 = %d\n", GM_SUPPORT_L4);
#else
  printf ("GM_SUPPORT_L4 is undefined\n");
#endif

#ifdef GM_SUPPORT_L7
  printf ("GM_SUPPORT_L7 = %d\n", GM_SUPPORT_L7);
#else
  printf ("GM_SUPPORT_L7 is undefined\n");
#endif

#ifdef GM_SUPPORT_L8
  printf ("GM_SUPPORT_L8 = %d\n", GM_SUPPORT_L8);
#else
  printf ("GM_SUPPORT_L8 is undefined\n");
#endif

#ifdef GM_SUPPORT_L9
  printf ("GM_SUPPORT_L9 = %d\n", GM_SUPPORT_L9);
#else
  printf ("GM_SUPPORT_L9 is undefined\n");
#endif

#ifdef GM_SUPPORT_PCI_REV_1
  printf ("GM_SUPPORT_PCI_REV_1 = %d\n", GM_SUPPORT_PCI_REV_1);
#else
  printf ("GM_SUPPORT_PCI_REV_1 is undefined\n");
#endif

#ifdef GM_SUPPORT_PCI_REV_2
  printf ("GM_SUPPORT_PCI_REV_2 = %d\n", GM_SUPPORT_PCI_REV_2);
#else
  printf ("GM_SUPPORT_PCI_REV_2 is undefined\n");
#endif

#ifdef GM_SUPPORT_PCI_REV_3
  printf ("GM_SUPPORT_PCI_REV_3 = %d\n", GM_SUPPORT_PCI_REV_3);
#else
  printf ("GM_SUPPORT_PCI_REV_3 is undefined\n");
#endif

#ifdef GM_SUPPORT_SBUS
  printf ("GM_SUPPORT_SBUS = %d\n", GM_SUPPORT_SBUS);
#else
  printf ("GM_SUPPORT_SBUS is undefined\n");
#endif

#ifdef GM_ENABLE_DEBUG_COUNTERS
  printf ("GM_ENABLE_DEBUG_COUNTERS = %d\n", GM_ENABLE_DEBUG_COUNTERS);
#else
  printf ("GM_ENABLE_DEBUG_COUNTERS is undefined\n");
#endif

#ifdef GM_ENABLE_ERROR_COUNTERS
  printf ("GM_ENABLE_ERROR_COUNTERS = %d\n", GM_ENABLE_ERROR_COUNTERS);
#else
  printf ("GM_ENABLE_ERROR_COUNTERS is undefined\n");
#endif

#ifdef GM_ENABLE_FORK_SYSTEM
  printf ("GM_ENABLE_FORK_SYSTEM = %d\n", GM_ENABLE_FORK_SYSTEM);
#else
  printf ("GM_ENABLE_FORK_SYSTEM is undefined\n");
#endif

#ifdef GM_ENABLE_LOG_DISPATCH
  printf ("GM_ENABLE_LOG_DISPATCH = %d\n", GM_ENABLE_LOG_DISPATCH);
#else
  printf ("GM_ENABLE_LOG_DISPATCH is undefined\n");
#endif

#ifdef GM_ENABLE_PACKET_COUNTERS
  printf ("GM_ENABLE_PACKET_COUNTERS = %d\n", GM_ENABLE_PACKET_COUNTERS);
#else
  printf ("GM_ENABLE_PACKET_COUNTERS is undefined\n");
#endif

  printf ("page size is %d\n", (int)(GM_PAGE_LEN));

  printf ("Some derived values:\n"
          "    GM_RECV_QUEUE_SLOTS_PER_PAGE %d\n"
          "    GM_NUM_RECV_QUEUE_SLOTS      %d\n",
		  (int)(GM_RECV_QUEUE_SLOTS_PER_PAGE),
          (GM_NUM_RECV_QUEUE_SLOTS));
  printf ("    GM_NUM_RECV_TOKENS           %d\n"
          "    GM_NUM_SEND_TOKENS           %d\n",
		  (GM_NUM_RECV_TOKENS),
          (GM_NUM_SEND_TOKENS));
}


static void
print_dma_rates(void)
{
  unsigned long rate1=0,
                rate2=0;
  int           i;
  gm_u32_t      bus_width,
		        bus_rate;

  for (i = 0; i < (GM_DMA_PERF_ALTERNATIVE_TIMING ? 4 : 1); i++) {
	rate1 = rate2 = 0;

	/* the multiplication by 2 is because the RTC runs at 2MHz */
	if (gm_ntoh_u32(globals->e2l_time[i])) {
		rate1 = (GM_PAGE_LEN * GM_DMA_PERF_NUM_PAGES * 2
				 / (gm_ntoh_u32(globals->e2l_time[i])));
	}
	if (gm_ntoh_u32(globals->l2e_time[i])) {
		rate2 = (GM_PAGE_LEN * GM_DMA_PERF_NUM_PAGES * 2
				 / (gm_ntoh_u32(globals->l2e_time[i])));
	}
	bus_width = gm_ntoh_u32(globals->bus_width);
	bus_rate = gm_ntoh_u32(globals->bus_rate)? gm_ntoh_u32(globals->bus_rate):33;

	if (bus_width == 0) {
		/* attempt to infer the width from the bandwidth and frequency */
		if (__GM_MAX(rate1, rate2) / bus_rate > 4) {
			bus_width = 64;
		}
	}

	if (bus_width) {
		printf ("GM-2 DMA rate for %lu Byte pages (%dbit / %dMHz bus)\n",
				GM_PAGE_LEN, bus_width, bus_rate);
				
	} else {
		printf ("GM-2 DMA rate for %lu Byte pages (%dMHz bus)\n",
				GM_PAGE_LEN, bus_rate);
	}

    switch (i) {
	case 0:
		printf ("Timing %d pages.\n", GM_DMA_PERF_NUM_PAGES);
		break;
	case 1:
		printf ("2nd: sdma a page, rdma a page %d times\n",
				GM_DMA_PERF_NUM_PAGES);
		break;
	case 2:
		printf ("3rd: sdma and rdma to/from alternating pages (coarse grain)\n");
		break;
	case 3:
		printf ("4th: sdma and rdma to/from alternating pages (fine grain)\n");
		break;
    default:
		printf ("CASE %d is BOGUS ******\n", i);
	}

	printf ("\tbus_read  (send) = %lu MBytes/s\n"
		    "\tbus_write (recv) = %lu MBytes/s\n\n",
			rate1, rate2);
  } /* for i */
}



static void
print_log(void)
{
	gm_lp_t log_end;
	gm_lp_t log_slot;
	int index;
	int prev_time_known = 0;
	int start_slot;
	
	if (!GM_ENABLE_LOG_DISPATCH)
		return;
	
	for (index=0; index<GM_LOG_LEN; index++) {
		if (!gm_ntoh_lp (globals->log[index].message)) {
			index++;
			break;
		}
	}
	if (index>=GM_LOG_LEN) {
		index = 0;
	}
	printf ("record_log (count) = %d\n",
			gm_ntoh_s32 (globals->record_log));
	log_end = gm_ntoh_lp (globals->log_end);
	gm_always_assert (log_end);
	log_slot = gm_ntoh_lp(globals->log_slot);
	gm_always_assert (log_slot);
	printf ("log_end = 0x%x, log_slot = 0x%x\n", log_end, log_slot);
	start_slot = (GM_LOG_LEN
				  - ((log_end - log_slot) / sizeof (globals->log[0])));
	printf("start_slot = 0x%x\n", start_slot);
	index = start_slot;
	do {
		gm_status_t status;
		gm_s32_t new_time;
		gm_s32_t prev_time;
		
		if (gm_ntoh_lp (globals->log[index].message) != 0) {
			char str[GM_MAX_FIRMWARE_STRING_LEN+1];
			gm_s32_t firmware_time;
			gm_size_t length;
			
			status = (_gm_get_firmware_string
					  (port, gm_ntoh_lp (globals->log[index].message),
					   str, &length));
			firmware_time = gm_ntoh_s32 (globals->log[index].time);
			if (status == GM_SUCCESS) {
				str[GM_MAX_FIRMWARE_STRING_LEN] = 0;
				new_time = gm_ntoh_s32 (globals->log[index].time);
				if (!prev_time_known) {
					prev_time = new_time;
					prev_time_known = 1;
				}
				printf ("%10d %s\n", new_time - prev_time, str);
				prev_time = new_time;
			}
			else {
				gm_perror ("firmware string error", status);
			}
		}
		if (++index >= GM_LOG_LEN) {
			index = 0;
		}
	} while (index != start_slot);

	printf("dispatch counts:\n");

	{
		int j;
		for (j=0; j<GM_DISPATCH_MAX_NUM; j++) {
			if (gm_ntoh_lp (globals->dispatch_seen[j]))
			{
				char str[GM_MAX_FIRMWARE_STRING_LEN+1];
				gm_size_t length;
				gm_status_t status;

				/* Try to get the firmware string describing the dispatch */

				status
					= (_gm_get_firmware_string
					   (port,
						gm_ntoh_lp (globals->dispatch_seen[j]),
						str, &length));

				/* Print the best description we can. */
				
				printf ("%3d: %9d %s\n", j,
						gm_ntoh_u32 (globals->dispatch_cnt[j]),
						status == GM_SUCCESS ? str : "???");
			}
		}
	}
}



static void
print_lanai_cts()
{
    printf("First two words of LANai globals: 0x%x, 0x%x\n",
           gm_ntoh_u32 (((gm_u32_n_t *) globals)[0]),
           gm_ntoh_u32 (((gm_u32_n_t *) globals)[1]) );

	printf("Some counters from LANai\n"
		   "\n");
	if (GM_ENABLE_PACKET_COUNTERS) {
		printf("netsend_cnt      %u\n", gm_ntoh_u32 (globals->netsend_cnt));
		printf("netrecv_cnt      %u\n", gm_ntoh_u32 (globals->netrecv_cnt));
	}

	if (GM_ENABLE_ERROR_COUNTERS) {
		printf("error counters\n");
		GM_ERROR_COUNTERS
		printf("\n");
		printf("       packet_sexno %04x %04x\n", 
			   gm_ntoh_s16 (globals->packet_sexno.parts.sesno),
			   gm_ntoh_s16 (globals->packet_sexno.parts.seqno));
		printf("       ack_sexno %04x %04x\n", 
			   gm_ntoh_s16 (globals->ack_sexno.parts.sesno),
			   gm_ntoh_s16 (globals->ack_sexno.parts.seqno));
		printf("\n"
			   "\n");
	}

	if (GM_ENABLE_DEBUG_COUNTERS) {
		printf("debug counters\n");
		GM_DEBUG_COUNTERS
	}

	printf("port counters\n");
	P("", port[2].active_subport_cnt);
}



static void
dump_connection(struct gm_connection *c, int number)
{
	if (c) {
		printf("globals->connection[%d]\n",number);
		printf("\tcrcs = 0x%x\n",gm_ntoh_u8(c->probable_crc_error_cnt));
		printf("\tbadcrc      = 0x%x\n",gm_ntoh_u32(c->bad_crc_cnt));
		/*  printf("\tack_pending = 0x%x\n",gm_ntoh_u8(c->ack_pending)); */
		printf("\thas_route   = 0x%x\n",gm_ntoh_u8(c->has_route));
		printf("\tsend_sexno  = %04x %04x\n",
		       gm_ntoh_s16(c->send_sexno.parts.sesno),
		       gm_ntoh_s16(c->send_sexno.parts.seqno));
		printf("\tack.sexno   = %04x %04x\n",
			   gm_ntoh_s16(c->ack_packet.sexno.parts.sesno),
			   gm_ntoh_s16(c->ack_packet.sexno.parts.seqno));
		printf("\tclose_sexno = %04x %04x\n",
			   gm_ntoh_s16(c->close_sexno.parts.sesno),
			   gm_ntoh_s16(c->close_sexno.parts.seqno));
		printf("\topen_time   = 0x%lx\n",(long) gm_ntoh_s64(c->open_time));
		printf("\tclose_time  = 0x%lx\n",(long) gm_ntoh_s64(c->close_time));
		printf("\tcurrentRTC  = 0x%lx\n",(long) current_rtc);
		printf("\tack.type subtype = %04x %04x\n",
			   gm_ntoh_u16(c->ack_packet.type),
			   gm_ntoh_u16(c->ack_packet.subtype));
		printf("\tack.target  = 0x%x   ",
			   gm_ntoh_u32(c->ack_packet.target_global_id));
		printf("\tack.sender  = 0x%x\n",
			   gm_ntoh_u32(c->ack_packet.sender_global_id));
	}
}

static gm_status_t
dump_sram (gm_port_t *port)
{
  gm_myrinet_eeprom_t eeprom;
  gm_size_t sram_size;
  gm_status_t status;
  FILE *wrfp;
  int i, numwords;
  gm_u32_n_t *words;
  char *sram;


  /* Copy the eeprom */
  status = _gm_get_eeprom (port, &eeprom);
  if (status != GM_SUCCESS) {
      gm_perror ("gm_debug: error reading eeprom", status);
      return status;
  }
  sram_size = (gm_size_t) eeprom.lanai_sram_size;
  sram = gm_calloc (sram_size, 1);
  if (sram == NULL) {
      gm_perror ("gm_debug: error allocating dump buffer", status);
      return status;
  }

  status = _gm_user_ioctl (port, GM_DUMP_MCP, sram, sram_size);
  if (status != GM_SUCCESS) {
      gm_perror ("gm_debug: error reading eeprom", status);
      return status;
  }

  wrfp = fopen("sram.dump", "w");
  if (wrfp == NULL) {
	  perror("can't open sram.dump");
	  exit(1);
  }

  numwords = sram_size / sizeof (gm_u32_t);
  fprintf (wrfp, "0x0\n"); /* HACK, lxgdb wants ISR here */
  fprintf (wrfp, "0x%x\n", numwords);
  words = (gm_u32_n_t *)sram;
  for (i = 0; i < numwords; i++)
	  fprintf (wrfp, "0x%x\n", gm_ntoh_u32 (words[i]));


  return GM_SUCCESS;
}

extern int atoi (const char *);

static int
gm_debug (int argc, char **argv)
{
	int globals_len;
	gm_status_t status;
    time_t uct;
	char   *time_string;

/*  Parse run-time flags */

	parse_args(argc, argv);

    printf("Opening board %d, port %d\n", board_num, port_num);
	fflush(stdout);				/* just in case we hang in the open */

	status = gm_open (&port, board_num, port_num, "gm_debug",
					  (enum gm_api_version) GM_API_VERSION);
	if (status != GM_SUCCESS) {
		printf    ("gm_debug: _gm_open board %d port %d failed:\n",
                   board_num, port_num);
		gm_perror ("          Error status", status);
		gm_exit(status);
	}

    /* This needs to wait until after open, otherwise some globals like
     * GM_PAGE_LEN won't be set.
     */
	if (do_print_info)
		print_debug_info();

	globals_len = sizeof(struct gm_lanai_globals) +
	    (MY_MAX_NODE * sizeof(struct gm_connection));

#if	GM_OS_AIX
	if ((globals = malloc (globals_len)) == NULL) {
		printf ("gm_debug: error in malloc\n");
		gm_exit (GM_FAILURE);
	}
#endif	/* GM_OS_AIX */

	if (do_repeat) {
		printf("\nWill run in a loop\n");
	}

	do {
		if (do_timestamp) {
			uct = time(NULL);
			time_string = asctime(localtime(&uct));
			printf("\n---> %s", time_string);
		}

	  	current_rtc = gm_ticks(port);
		status = _gm_get_globals(port, (gm_u8_t *) globals, globals_len);
		if (status != GM_SUCCESS) {
			printf("gm_debug: _gm_get_globals failed\n");
			gm_exit(status);
		}

		/* now we have a copy of the lanai_globals */
		/* use GDB? */

		if (do_print_dma_rates)
		    print_dma_rates();

		if (do_print_log)
			print_log();

        if (do_print_lanai_cts)
			print_lanai_cts();

		if (max_connection) {
			int j;
			for (j=1; j<=max_connection; j++) {
				dump_connection(&globals->connection.array[j],j);
			}
		}
		if (one_connection) {
				dump_connection(&globals->connection.array[one_connection],
								one_connection);
		}

/* ********************************************************************* */
/* =========> *//* THIS IS A GOOD LINE FOR A BREAKPOINT */
/* ********************************************************************* */

		if (sleep_secs) {
		  printf("Going to sleep for %d seconds\n", sleep_secs);
		  gm_sleep (sleep_secs);
		}

	} while (do_repeat);

	if (do_abort) {
		printf("do_abort is set - will now 'abort' this program"
               " to create a core file\n\n");
		fflush(stdout);
		abort();
	}

	if (do_dump) {
		printf ("saving SRAM contents to file sram.dump\n");
		status = dump_sram (port);
		if (status != GM_SUCCESS)
			printf ("failed to save SRAM");
	}
	gm_close(port);

	return(0);
}

#define GM_MAIN gm_debug
#include "gm_main.h"

/*
  This comment holds special emacs settings for this file.

  Local Variables:
  mode:c
  c-file-style:"bsd"
  tab-width:4
  End:
*/
