/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 2003 by Myricom, Inc.                                   *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gm.h"
#include "gm_io.h"
#include "gm_undocumented_funcs.h"


/* Return nonzero iff arg matches prefix.  Store a pointer to the remaining
   arg at match, if match is nonzero. */

static int 
match (const char *arg, const char *prefix, const char **match)
{
  gm_size_t prefix_len;

  prefix_len = gm_strlen (prefix);
  if (!gm_strncmp (arg, prefix, prefix_len))
    {
      if (match)
	*match = arg + prefix_len;
      return 1;
    }
  return 0;
}

static int 
short_match (const char *arg, const char *prefix, const char **match,
	     int *argc, char **argv[])
{
  gm_size_t prefix_len;
  
  prefix_len = gm_strlen (prefix);
  if (!gm_strncmp (arg, prefix, prefix_len))
    {
      if (*argc > 0)
	{
	  (*argc)--;
	  (*argv)++;
	  *match = **argv;
	  return 1;
	}
    }
  return 0;
}

static void
usage (const char *progname)
{
  printf ("usage: %s [options]\n", progname);
  printf ("valid options include:\n");
  printf ("  -B <unit>, (default unit = 0)\n" );
  printf ("  --unit=<unit> (default unit = 0) \n");
  printf ("  -i <Receive IRQ Credit Increment>, \n");
  printf ("  --credit-increment=<Receive IRQ Credit Increment>\n");
  printf ("    Set the number of credits each receive interrupt uses \n");
  printf ("  -m <Maximum Receive IRQ Credits>, \n");
  printf ("  --max-credits=<Maximum Receive IRQ Credits>\n");
  printf ("    Set the maximum amount of receive interrupt credits \n");
  printf ("  -x <Maximum Transmits Until IRQ>, \n");
  printf ("  --max-xmits=<Maximum Transmits Until IRQ>\n");
  printf ("    Set the maximum number of transmits which may complete \n");
  printf ("    before a transmit complete interrupt is generated\n");
  printf ("  -s, \n");
  printf ("  --silent\n");
  printf ("    Do not print the current settings\n");
  printf ("  -h, \n");
  printf ("  --help\n");
  printf ("    This message is displayed\n");
}

static int
gm_ether_tune (int argc, char *argv[])
{
  gm_status_t status;
  int display;
  struct gm_port* port;
  unsigned int unit;
  static struct gm_lanai_globals globals;
  const char *s;
  const char *progname;
  gm_u32_t credit_incr, max_credits, max_xmits;

  unit = credit_incr= max_credits = max_xmits = 0;
  display = 1;
  progname = argv[0];

  for (argc--, argv++;
       argc;
       argc--, argv++)
    {
      if (match (*argv, "--unit=", &s)
	  || short_match (*argv, "-B", &s, &argc, &argv))
	{
	  unit = atoi (s);
	  continue;
	}

      if (match (*argv, "--credit-increment=", &s)
	  || short_match (*argv, "-i", &s, &argc, &argv))
	{
	  credit_incr = atoi (s);
	  continue;
	}

      if (match (*argv, "--max-credits=", &s)
	  || short_match (*argv, "-m", &s, &argc, &argv))
	{
	  max_credits = atoi (s);
	  continue;
	}

      if (match (*argv, "--max-xmits=", &s)
	  || short_match (*argv, "-x", &s, &argc, &argv))
	{
	  max_xmits = atoi (s);
	  continue;
	}

      if (match (*argv, "--unit=", &s)
	  || short_match (*argv, "-B", &s, &argc, &argv))
	{
	  unit = atoi (s);
	  continue;
	}

      if (!gm_strcmp (*argv, "--silent")
	  || !gm_strcmp (*argv, "-s"))
	{
	  display = 0;
	  continue;
	}
      if (!gm_strcmp (*argv, "--help")
	  || !gm_strcmp (*argv, "-h"))
	{
	  usage (progname);
	  exit (0);
	}
      usage (progname);
      exit (1);
    }
  memset (&globals, 0, sizeof (globals));
  if ((status = gm_open (&port, unit, 0, "gm_ether_tune",
	       (enum gm_api_version) GM_API_VERSION)
       != GM_SUCCESS))
    {
      gm_perror ("Couldn't open device", status);
      exit (1);
    }

  if (credit_incr)
    {
      status = _gm_user_ioctl (port, GM_ETHERNET_SET_CREDIT_INCR, &credit_incr,
			       sizeof (credit_incr));
      if (status != GM_SUCCESS)
	{
	  gm_perror ("Failed to set credit increment", status);
	  exit (1);
	}
    }
  
  if (max_credits)
    {
      status = _gm_user_ioctl (port, GM_ETHERNET_SET_MAX_CREDITS, &max_credits,
			       sizeof (max_credits));
      if (status != GM_SUCCESS)
	{
	  gm_perror ("Failed to set max credits", status);
	  exit (1);
	}
    }
  
  if (max_xmits)
    {
      status = _gm_user_ioctl (port, GM_ETHERNET_SET_MAX_SDMAS, &max_xmits,
			       sizeof (max_xmits));
      if (status != GM_SUCCESS)
	{
	  gm_perror ("Failed to set max transmits", status);
	  exit (1);
	}
    }

  status = _gm_user_ioctl (port, GM_ETHERNET_GET_PARAMS, &globals.ethernet.wake,
			   sizeof (globals.ethernet.wake));
  if (status != GM_SUCCESS)
    {
      gm_perror ("GM_ETHERNET_GET_PARAMS failed", status);
      exit (EXIT_FAILURE);
    }

  if (display)
    {
      printf ("GM Ethernet Emulation Interrupt Coalescing settings for Board %d:\n", 
	      unit);
      printf ("-----------------------------------------------------------------\n");
      printf ("Receive IRQ Credits Available\t= %d\n", 
	      gm_ntoh_u8 (globals.ethernet.wake.credits));
      printf ("Receive IRQ Credit Increment\t= %d\n", 
	      gm_ntoh_u8 (globals.ethernet.wake.credit_incr));
      printf ("Maximum Receive IRQ Credits\t= %d\n", 
	      gm_ntoh_u8 (globals.ethernet.wake.max_credits));
      printf ("Transmits Until IRQ\t\t= %d\n", 
	      gm_ntoh_u16 (globals.ethernet.wake.sdmas_until_wake));
      printf ("Maximum Transmits Until IRQ\t= %d\n", 
	      gm_ntoh_u16 (globals.ethernet.wake.max_sdmas_until_wake));
    }
  gm_close(port);
  
  return EXIT_SUCCESS;
}

#define GM_MAIN gm_ether_tune
#include "gm_main.h"

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
