/*
 * Find the cut-off length for FAST messages
 * Author: Max Stern (Myricom)
 *
 * Syntax: gm_get_fast_length [<local_boardnum>]
 *
 *         The default for <local_boardnum> is 0.
 *
 * Description:
 *
 *   This program sends itself messages in loopback mode, increasing the
 *   message length from 1 to a set limit (GM_GFL_LIMIT).  It notes the
 *   cut-off point at which the received messages cease being FAST messages,
 *   and also verifies that the changeover is monotonic.
 * */

#include "gm.h"

#define GM_GFL_PORT_NUM 2
#define GM_GFL_LIMIT 257
#define GM_GFL_PRIORITY GM_LOW_PRIORITY
#define GM_GFL_SIZE gm_min_size_for_length(GM_GFL_LIMIT)
#define GM_GFL_BUFFER_COUNT 2
#define GM_GFL_BUFFER_LENGTH gm_max_length_for_size(GM_GFL_SIZE)
#define GM_GFL_DEBUG 1

struct lengths_s
{
  int max_length_for_fast;
  int min_length_for_nonfast;
};


void my_send_callback (struct gm_port *port,
		       void *context,
		       gm_status_t status);
int receive_stuff(struct gm_port *the_port,
		  struct lengths_s *the_lengths,
		  int *expected);


int
main (int argc, char **argv)
{
  static struct gm_port *my_port = 0;
  static void *out_buffer = 0, *in_buffer = 0;
  int my_boardnum = 0;		/* Default boardnum is 0 */
  unsigned int my_node_id;
  int expected_messages = 2;
  gm_status_t gm_status;
/* expected_callbacks makes sure we don't exit until we have handled all
   the callbacks we expect.                                              */
  static int expected_callbacks = 0;
  int msg_length;
  struct lengths_s lengths;


  lengths.max_length_for_fast = 0;
  lengths.min_length_for_nonfast = GM_GFL_LIMIT + 1;;
  if (argc > 2)
    {
      gm_printf ("USAGE: gm_get_fast_length [<local_boardnum>]\n");
      gm_exit (GM_FAILURE);
    }
  if (argc > 1)
    {
      my_boardnum = atoi (argv[1]);
    }

  gm_status = gm_open (&my_port, my_boardnum,
		       GM_GFL_PORT_NUM, "gm_get_fast_length",
		       GM_API_VERSION_1_1);
  if (gm_status == GM_SUCCESS)
    {
      gm_printf ("Opened board %d port %d\n",
		 my_boardnum, GM_GFL_PORT_NUM);
    }
  else
    {
      gm_perror ("Couldn't open GM port", gm_status);
      goto abort_with_nothing;
    }

  gm_get_node_id (my_port, &my_node_id);
#if GM_GFL_DEBUG
  gm_printf ("My node ID is %d\n", my_node_id);
#endif

  out_buffer =
    gm_dma_calloc (my_port,
		   GM_GFL_BUFFER_COUNT,
		   GM_GFL_BUFFER_LENGTH);
  if (out_buffer == 0)
    {
      gm_printf ("Couldn't allocate out_buffer\n");
      goto abort_with_open_port;
    }
#if GM_GFL_DEBUG
  gm_printf ("Allocated %d out_buffers of size %d, length %d\n",
	     GM_GFL_BUFFER_COUNT, GM_GFL_SIZE, GM_GFL_BUFFER_LENGTH);
#endif

  in_buffer =
    gm_dma_calloc (my_port,
		   GM_GFL_BUFFER_COUNT, 
		   GM_GFL_BUFFER_LENGTH);
  if (in_buffer == 0)
    {
      gm_printf ("Couldn't allocate in_buffer\n");
      goto abort_with_out_buffer;
    }
#if GM_GFL_DEBUG
  gm_printf ("Allocated %d in_buffers of size %d, length %d\n",
	     GM_GFL_BUFFER_COUNT, GM_GFL_SIZE, GM_GFL_BUFFER_LENGTH);
#endif

  gm_free_send_tokens (my_port, 
		       GM_GFL_PRIORITY, 
		       gm_num_send_tokens (my_port));

  gm_provide_receive_buffer (my_port,
			     in_buffer,
			     GM_GFL_SIZE,
			     GM_GFL_PRIORITY);

  for (msg_length = 1; msg_length <= GM_GFL_LIMIT; msg_length++)
    {
      gm_send_to_peer_with_callback (my_port,
				     out_buffer,
				     GM_GFL_SIZE,
				     msg_length,
				     GM_GFL_PRIORITY,
				     my_node_id,
				     my_send_callback,
				     &expected_callbacks);
      expected_callbacks++;
      expected_messages++;

      while ((expected_callbacks > 0) && (expected_messages > 0))
	{
	  if (receive_stuff(my_port, &lengths, &expected_messages))
	    {
	      /* non-zero return means some error occurred */
	      goto abort_with_in_buffer;
	    }
	}

    }

  gm_printf ("gm_get_fast_length completed successfully\n");
  gm_printf ("Message lengths from 1 to %d were tested\n", GM_GFL_LIMIT);
  gm_printf ("Maximum FAST length was %d; minimum non-FAST length was %d\n",
	     lengths.max_length_for_fast, lengths.min_length_for_nonfast);
  if (lengths.max_length_for_fast >= lengths.min_length_for_nonfast)
    {
      gm_printf ("**** WARNING --- ANOMALOUS RESULTS ****\n");
    }

  gm_close (my_port);
  gm_exit (GM_SUCCESS);

 abort_with_in_buffer:
  gm_free (in_buffer);
 abort_with_out_buffer:
  gm_free (out_buffer);
 abort_with_open_port:
  gm_close (my_port);
 abort_with_nothing:
  gm_exit (GM_FAILURE);
}


int receive_stuff(struct gm_port *the_port,
		  struct lengths_s *the_lengths,
		  int *expected)
{
  gm_recv_event_t *event;
  unsigned int event_type;
  unsigned int size;
  int received_length;
  void *buffer;


  event = gm_receive (the_port);

  event_type = (gm_ntoh_u8 (event->recv.type));
  switch (event_type)
    {
    case GM_PEER_RECV_EVENT:
    case GM_FAST_PEER_RECV_EVENT:
      if (*expected <= 0)
	{
	  gm_printf ("**** ERROR: unexpected message received ****\n");
	  return (1);
	}
      (*expected)--;
      received_length = gm_ntoh_u32 (event->recv.length);
      if (event_type == GM_PEER_RECV_EVENT)
	{
	  if (received_length < the_lengths->min_length_for_nonfast)
	    {
	      the_lengths->min_length_for_nonfast = received_length;
	    }
	}
      else /* GM_FAST_PEER_RECV_EVENT */
	{
	  if (received_length > the_lengths->max_length_for_fast)
	    {
	      the_lengths->max_length_for_fast = received_length;
	    }
	}
      /* Return the buffer for reuse */
      buffer = gm_ntohp (event->recv.buffer);
      size = gm_ntoh_u8 (event->recv.size);
      gm_provide_receive_buffer (the_port,
				 buffer, 
				 size, 
				 GM_GFL_PRIORITY);
      break;

    case GM_NO_RECV_EVENT:
      break;

    default:
      gm_unknown (the_port, event);	/* gm_unknown calls the callback */
    }

  return (0);
}


void
my_send_callback (struct gm_port *port, void *context, gm_status_t status)
{
     /* Description:

	This function is called inside gm_unknown() when there is a callback
	ready to be processed.  It tells us that a send has completed, either
	successfully or with error.
     */

  (*(int *)context)--;

  if (status != GM_SUCCESS)
    {
      if (status != GM_SEND_DROPPED)
	{
	  gm_perror ("send completed with error", status);
	}
    }
}
