/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997, 1998, 1999 by Myricom, Inc.                 *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/*
  gm_stress.c
  all-to-all test for GM. Run it first on host with lowest GM ID.
  finucane@myri.com
  gcc -g -o gm_stress gm_stress.c ~/gm/binary/.gm_uninstalled_libs/lib/.libs/libgm.a -I ~/gm/include
  
*/

/*
Specify a list of hosts with the -f option else it's loopback.
The list should be the same (including order) for all hosts.
If you don't specify a list the test is loopback.

You can use a little rsh script to launch it on a cluster:

racksaver1% /ufs/arpa/finucane/gm/tests# cat rsh_stress
#!/bin/bash

for i in `cat $1`
do
ssh $i $2 $3 $4 $5 $6 $7 $8 &
done
racksaver1% /ufs/arpa/finucane/gm/tests# 

racksaver1% /ufs/arpa/finucane/gm/tests# cat hosts
racksaver1
racksaver2
racksaver3
racksaver4
racksaver5
racksaver6
racksaver7

racksaver1% /ufs/arpa/finucane/gm/tests# ./rsh_stress hosts /ufs/finucane_home/gm/tests/gm_stress -f /ufs/finucane_home/gm/tests/hosts
racksaver1% /ufs/arpa/finucane/gm/tests# 7 hosts.
7 hosts.
7 hosts.
7 hosts.
7 hosts.
7 hosts.
7 hosts.

sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 93 / 92 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 93 / 90 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 93 / 93 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 88 / 89 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 88 / 91 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 88 / 86 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 93 / 94 MB/s


It works like this: Every host sends a given number of random sized
packets out of a big chunk of gm_dma_malloc memory to each host
randomly and terminates when there is an error or when it has received
the given number of packets from each host, (including itself). It
uses gm_zone.c, which is part of libgm, for malloc and free from the
send heap.  If any host terminates without error then it means the
test has succeeded globally. Otherwise the host that detected the
error will terminate first, and then the others after sends fail. The
test prints out a host count before the test starts.

It sends to self and will work in loopback but not point-to-point,
where there is no route-to-self.

Because of congestion at the receiver, send timeouts can occur. If the
test fails with the "exceeded --max-resends" error on any node,
increase --max-resends.  Increasing receive buffering with
--min-receives is also an option.

The datagrams option adds some percentage of uncounted datagrams to
the stream. 100 percent is legal but will never terminate.

The put option makes the test send puts.

Since you require an interrupt for *each* messages in gm_stress (using
gm_blocking_receive_no_spin) if the host is not fast enough to handle
interrupts, you will start to see NODE_UNREACHABLE errors, you will
start to see IP timeouts (because IP does require interrupt for each
packet) and you may eventually put the port in a funky state. The
work-around for this is to try --non-blocking.


*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#define GM_STRONG_TYPES 0
#include "gm.h"

#define insist(e) do{if(!(e)){printf ("assertion failed. %s:%d (%s)\n", __FILE__,__LINE__, #e); goto exception;}}while (0)
#define pinsis(e, p) do{if(!(e)){printf p; printf ("\n"); goto exception;}}while (0)
#define gcheck(r, p) do{if(r != GM_SUCCESS){gm_perror (p, (enum gm_status) r); goto exception;}}while (0)
#define ZONE_FACTOR 4

#ifndef htonl
#define htonl gm_hton_u32
#endif

#ifndef htonll
#define htonll gm_hton_u64
#endif

enum
{
  BARRIER,
  BUFFER_LIST,
  DATA,
  DATAGRAM,
  PUT,
  CABOOSE,
  PUT_ACK
};

typedef struct message_t
{
  gm_u32_t crc;
  gm_u32_t sequence;
  gm_u32_t from;
  gm_u32_t to;
  gm_u32_t type;
  gm_u32_t length;
  gm_u64_t put_buffer;
}
message_t;

typedef struct buffer_list_t
{
  message_t m;
  gm_u64_t buffers [1];
} buffer_list_t;


typedef struct host_t
{
  unsigned int id;
  char hostname[GM_MAX_HOST_NAME_LEN + 1];
  gm_u64_t bytes_sent;
  gm_u64_t bytes_received;
  int dropping;
  unsigned messages_completed;
  unsigned messages_sent;
  unsigned messages_resent;
  unsigned messages_received;
  unsigned datagrams_sent;
  unsigned datagrams_received;
  int pending_puts;
  int acks_needed;
  int puts_sent;
  int puts_received;
  gm_u64_t put_buffer;
}
host_t;

static int malloc_bytes;
static int failed;
static unsigned my_id;
static int my_index;
static int num_hosts;
static host_t *hosts;
static int send_tokens;
static int max_send_tokens;
static char *send_buffer;
static struct gm_zone *zone;
static unsigned long size_mask;
static unsigned biggest;
static int num_hostnames;
static char **hostnames;
static char *filename;
static int checksum;
static int datagram_percent;
static int unit = 0;
static int port_id = 2;
static int max_sends = 1024;
static int max_size = 19;
static unsigned max_length;
static int min_receives = 4;
static int alarmed = 0;
static gm_alarm_t alarm;
static int barrier_seconds = 60 * 2;
static int max_resends = 30;
static int same_size = 0;

static int initialized = 0;
static int opened = 0;
static int started = 0;
static int non_blocking = 0;
static int num_puts = 0;
static int puts_received = 0;
static int barriers_received = 0;
static int buffer_lists_received = 0;
static int zone_length;

static void careful_exit (struct gm_port *port, int r)
{
  int i;

  if (hosts)
    free (hosts);

  if (zone)
    gm_zone_destroy_zone (zone);

  if (hostnames)
  {
    for (i = 0; i < num_hostnames; i++)
      if (hostnames[i])
	free (hostnames[i]);
    free (hostnames);
  }
  if (port && opened)
    gm_close (port);

  if (initialized)
    gm_finalize ();
  exit (r);
}

static void usage (char *option)
{
  if (option)
    printf ("unknown option %s\n", option);

  printf ("options:\n");

  printf ("-u, --unit=NUM            [0],      myrinet board number.\n");
  printf ("-p, --port=NUM            [2],      GM port number.\n");
  printf ("-n, --num-sends=NUM       [1024],   sends per host.\n");
  printf ("-m, --max-size=NUM        [19],     max send \"size\" (in log2 units).\n");
  printf ("-f, --filename=FILENAME   [none],   file of hostnames, one per line\n");
  printf ("-c, --checksum            [off],    compute checksums.\n");
  printf ("-r, --min-receives=NUM    [4],      minimum number of receive buffers per size.\n");
  printf ("-b, --barrier-seconds=NUM [120],    timeout for barrier.\n");
  printf ("-x, --max-resends=NUM     [0],      max resends per host destination for send timeouts.\n");
  printf ("-s, --same-size           [off],    make all messages the same size (--max-size).\n");
  printf ("-d, --datagram=NUM        [0],      percent datagrams.\n");
  printf ("-o, --non-blocking        [off],    use gm_receive instead of gm_blocking_receive_no_spin\n");
  printf ("-t, --num-puts            [0],      test gm_put (probably broken)\n");
  
  careful_exit (0, 1);
}

static int option (int i, int argc, char **argv, char *format, char *letter, int *v,
		   int v_min, int v_max)
{
  int r = 0;
  insist (argv && format && letter && v);

  if ((r = sscanf (argv[i], format, v)) == 1 || (i + 1 < argc && !strcmp (argv[i], letter)))
  {
    if (!r)
      *v = atoi (argv[i + 1]);
    if (*v < v_min || *v > v_max)
    {
      printf ("%d is not between %d and %d\n", *v, v_min, v_max);
      usage (0);
    }
    return r ? 1 : 2;
  }
  exception:return 0;
}

static void sig_handler (int sig)
{
  failed = 1;
}

static int compare (const void *a, const void *b)
{
  return strcmp (*(char **) a, *(char **) b);
}

static int read_hostnames (char *filename, int num_hostnames, char **hostnames)
{
  FILE *fp;
  int count = 0;
  char *s, *e;
  char buffer[GM_MAX_HOST_NAME_LEN + 1];

  insist (filename);
  insist (num_hostnames >= 0);
  insist (!num_hostnames || hostnames);

  fp = fopen (filename, "r");
  pinsis (fp != 0, ("couldn't open \"%s\".", filename));

  while ((s = fgets (buffer, GM_MAX_HOST_NAME_LEN, fp)))
  {
    while (isspace (*s))
      s++;
    for (e = s + strlen (s) - 1; e >= s && isspace (*e); e--)
      *e = 0;
    if (*s)
    {
      if (num_hostnames && hostnames)
      {
	insist (count < num_hostnames);
	insist (hostnames[count]);
	strcpy (hostnames[count], s);
      }
      count++;
    }
  }

  pinsis (count, ("no hostnames in \"%s\".", filename));
  insist (!num_hostnames || count == num_hostnames);
  fclose (fp);

  return count;
  exception:return 0;
}

static int dump (char*p, int length)
{
#define MAX_DUMP 200
  char buffer [MAX_DUMP * 3];
  char*s = buffer;
  int i;
  
  if (length > MAX_DUMP ) length = MAX_DUMP;
  
  for (i = 0; i < length; i++)
    s += sprintf
      (s, "%2.2x%s", ((unsigned char*)p) [i],
       i  && (i + 1) % 16 == 0 ? "\n" : i && i % 2 == 1  ? " " : "");
  printf ("%d:\n%s\n\n", length, buffer);
}

static int create_hostnames (char *filename)
{
  int i;

  insist (filename);

  if ((num_hostnames = read_hostnames (filename, 0, 0)) <= 0)
    return 0;

  hostnames = (char **) calloc (num_hostnames, sizeof (char *));
  pinsis (hostnames, ("malloc failed for hostnames"));

  for (i = 0; i < num_hostnames; i++)
  {
    hostnames[i] = (char *) malloc (GM_MAX_HOST_NAME_LEN);
    pinsis (hostnames[i], ("malloc failed for hostname %i", i));
  }

  i = read_hostnames (filename, num_hostnames, hostnames);
  insist (i == num_hostnames);
  qsort (hostnames, num_hostnames, sizeof (char *), compare);
  return 1;

  exception:
  if (hostnames)
    free (hostnames);
  hostnames = 0;
  return 0;
}

static int get_mac_address (char*s, char*mac_address)
{
  int i;
  int u [6];

  insist (s && mac_address);
  
  if (sscanf (s, "%x:%x:%x:%x:%x:%x", u+0, u+1, u+2, u+3, u+4, u+5, u+6) == 6)
  {
    for (i = 0; i < 6; i++)
      mac_address [i] = (char) u [i];
    return 1;
  }
  exception: return 0;
}

static int get_hosts (struct gm_port *port, host_t ** hosts)
{
  int num_hosts, i;
  gm_status_t r;
  char mac_address [6];
  
  insist (port && hosts);

  num_hosts = num_hostnames ? num_hostnames : 1;
  *hosts = (host_t *) calloc (num_hosts, sizeof (host_t));
  insist (*hosts);
  
  for (i = 0; i < num_hostnames; i++)
  {
    r = get_mac_address (hostnames [i], mac_address) ?
      gm_unique_id_to_node_id (port, mac_address, &(*hosts)[i].id) :
      gm_host_name_to_node_id_ex (port, barrier_seconds * 1000 * 1000, hostnames [i], &(*hosts)[i].id);
    
    gcheck (r, "gm_node_id lookup failed");
    strncpy ((*hosts)[i].hostname, hostnames [i], GM_MAX_HOST_NAME_LEN);
  }
  if (!num_hostnames)
  {
    insist (my_id);
    (*hosts)[0].id = my_id;

    r = gm_node_id_to_host_name_ex (port, barrier_seconds * 1000 * 1000, (*hosts)[0].id, &(*hosts) [0].hostname);
    gcheck (r, "gm_host_name_to_node_id_ex failed");
  }

  return num_hosts;

  exception:return 0;
}

static char*create_buffers (struct gm_port *port, unsigned max_length, int min_receives, int*send_buffer_length)
{
  int r, i, j, t, over;
  char *p, *sp;

  insist (port);
  insist (send_buffer_length);

  /*
    first alloc send buffer because recv alloc handles
    malloc fail better
  */
  
  *send_buffer_length = max_length * ZONE_FACTOR;  
  sp = (char *) gm_dma_malloc (port, *send_buffer_length);
  pinsis (sp, ("gm_dma_malloc failed"));
  malloc_bytes +=  *send_buffer_length;
  
  /*
    allocate min_receives recv buffers for each size.
    or try to, at least.
  */

  t = 0;
  for (j = 0; j < min_receives; j++)
  {
    for (i = gm_min_message_size (port); i <= gm_min_size_for_length (max_length); i++)
    {
      insist (j || t < gm_num_receive_tokens (port));
      if (t >= gm_num_receive_tokens (port))
	break;

      if (!(p = (char *) gm_dma_malloc (port, gm_max_length_for_size (i))) && j)
	break;
      
      pinsis (p, ("gm_dma_malloc failed"));
      malloc_bytes +=  gm_max_length_for_size (i);
      
      gm_provide_receive_buffer_with_tag (port, p, i, GM_LOW_PRIORITY, 0);
      t++;
      size_mask |= (1 << i);
    }
  }
  
  insist (size_mask);

  /*
    now go back and use up the rest of the receive tokens.
    half goes to the biggest size,
    half of the remaining half goes to the second to the biggest size.
    etc.
  */
  
  over = gm_num_receive_tokens (port) - t;
  /*printf ("over is %d\n", over); */
  
  for (i = gm_min_size_for_length (max_length); over > 2 && i >= gm_min_message_size (port); i--)
  {
    over /= 2;
    insist (over);

    /*
      if there were so many recv tokens left that allocating buffers for them all
      at this size exhausts our dma-able memory, try allocating less and less until
      dma_malloc works. this throws away some recv tokens but i don't care.
    */
 
    while (over && !(p = (char *) gm_dma_malloc (port, gm_max_length_for_size (i) * over)))
      over /= 2;

    malloc_bytes +=  gm_max_length_for_size (i) * over;
    
    if (!over)
      break;

    /*printf ("size %d, %d extra recv buffers\n", i, over);*/
    
    for (j = 0; j < over; j++)
    {
      insist (t < gm_num_receive_tokens (port));
      gm_provide_receive_buffer_with_tag (port, p + j * gm_max_length_for_size (i), i, GM_LOW_PRIORITY, 0);
      t++;
    }
  }
  
  r = gm_set_acceptable_sizes (port, GM_LOW_PRIORITY, size_mask);
  gcheck (r, "gm_set_acceptable_sizes failed");

  return sp;

  exception:return 0;
}

static char*create_put_buffers (struct gm_port *port, int*send_buffer_length)
{
  int r, i, num_buffers, size, length, rounded;
  char *p;

  insist (port);
  insist (send_buffer_length);
  insist (num_hosts > 0);
  num_buffers = gm_num_receive_tokens (port) - 2;
  if (num_buffers > num_hosts * num_puts)
    num_buffers = num_hosts * num_puts;

  p = gm_dma_malloc (port, num_hosts * max_length * num_puts);
  pinsis (p, ("gm_dma_malloc failed"));
  malloc_bytes += num_hosts * max_length * num_puts;
    
  memset (p, 0, num_hosts * max_length * num_puts);
  hosts [my_index].put_buffer = (gm_u64_t) (gm_size_t) p;
  
  length = sizeof (buffer_list_t) + (num_hosts - 1) * sizeof (gm_u64_t);
  size = gm_min_size_for_length (length);
  rounded = 1 << size;
  insist (rounded >= length);
  size_mask |= (1 << size);
  
  p = (char *) gm_dma_malloc (port, rounded);
  pinsis (p, ("gm_dma_malloc failed"));
  malloc_bytes += rounded;
  
  gm_provide_receive_buffer_with_tag (port, p, size, GM_LOW_PRIORITY, 0);
  
  length = sizeof (message_t);
  size = gm_min_size_for_length (length);
  rounded = 1 << size;
  insist (rounded >= length);
  size_mask |= (1 << size);
  
  for (i = 0; i < num_buffers; i++)
  {
    p = (char *) gm_dma_malloc (port, rounded);
    pinsis (p, ("gm_dma_malloc failed"));
    malloc_bytes += rounded;

    gm_provide_receive_buffer_with_tag (port, p, size, GM_LOW_PRIORITY, 0);
  }

  r = gm_set_acceptable_sizes (port, GM_LOW_PRIORITY, size_mask);
  gcheck (r, "gm_set_acceptable_sizes failed");
  
  length = max_length * ZONE_FACTOR + sizeof (buffer_list_t) + (num_hosts - 1) * sizeof (gm_u64_t);
  size = gm_min_size_for_length (length);
  rounded = 1 << size;
  insist (rounded >= length);
  *send_buffer_length = rounded;
  p = (char *) gm_dma_malloc (port, rounded);
  pinsis (p, ("gm_dma_malloc failed"));
  malloc_bytes += rounded;
  
  return p;
  exception:return 0;
}

static void alarm_callback (void*c)
{
  alarmed = 1;
}

static int handle_packet (void *p, int length)
{
  message_t *m = (message_t *) p;
  buffer_list_t*b = (buffer_list_t*) p;
  message_t*pb = 0;
  int i, from, to, sequence, type, p_len;

  insist (p && hosts);
  insist (num_hosts > 0);
  insist (my_index >= 0 || my_index < num_hosts);

  pinsis (length >= sizeof (message_t), ("bad length"));
  pinsis (!checksum || gm_crc ((char *) m + sizeof (unsigned), length - sizeof (unsigned)) == htonl (m->crc), ("bad crc"));

  to = htonl (m->to);
  from = htonl (m->from);
  sequence = htonl (m->sequence);
  type = htonl (m->type);

  pinsis (to == my_index, ("bad destination"));
  pinsis (from >= 0 && from < num_hosts, ("bad source"));
  
  switch (type)
  {
    case DATA:
      pinsis (sequence == hosts[from].messages_received + 1,
	      ("bad sequence %d from %d, expected %d", sequence, from,
	       hosts[from].messages_received + 1));

      hosts[from].messages_received++;
      hosts[from].bytes_received += length;
      break;
    case CABOOSE:
      insist (num_puts);
      pb = (message_t*) (((char*) (gm_size_t) hosts [my_index].put_buffer) + (from * num_puts + hosts [from].puts_received % num_puts) * max_length);
  
      printf ("put from %d completed in buffer %x\n", from, pb);
      
      p_len = htonl (pb->length);
      pinsis (p_len >= sizeof (message_t), ("short put length %d", p_len));
      pinsis (p_len <= max_length, ("long put length"));
      hosts[from].bytes_received += p_len;

      pinsis (!checksum || gm_crc ((char*) pb + sizeof (unsigned), p_len - sizeof (unsigned)) == htonl (pb->crc), ("bad put crc"));
      pinsis (m->to == pb->to, ("bad put to %d", htonl (pb->to)));
      pinsis (m->from == pb->from, ("bad put from"));
      pinsis (m->sequence == pb->sequence, ("bad put sequence"));

      pinsis (hosts [from].acks_needed < num_puts, ("overhasty put from %d", from));
      hosts[from].acks_needed++;
      hosts [from].puts_received++;
      hosts [from].messages_received++;
      
      break;
    case PUT_ACK:
      /*printf ("got ack from %d\n", from);*/
      pinsis (hosts [from].pending_puts > 0, ("unexpected put ack from node %d", from));
      hosts [from].pending_puts--;
      break;
    case DATAGRAM:
      hosts[from].datagrams_received++;
      break;
    case BARRIER:
      if (num_puts)
      {
	if (!m->put_buffer, ("null put buffer"));
	if (!my_index)
	{
	  pinsis (!hosts [from].put_buffer, ("wayward put buffer"));
	  hosts [from].put_buffer = htonll (m->put_buffer);
	}
      }
      barriers_received++;
      break;
    case BUFFER_LIST:
      pinsis (!buffer_lists_received, ("wayward buffer list"));
      pinsis (my_index, ("wayward buffer list"));
      pinsis (length >= sizeof (buffer_list_t) + (num_hosts - 1) * sizeof (gm_u64_t), ("bad length"));

      for (i = 0; i < num_hosts; i++)
      {
	pinsis (i != my_index || htonll (b->buffers [i]) == hosts [i].put_buffer, ("wrong buffer %lld", htonll (b->buffers [i])));
	pinsis (!num_puts || b->buffers [i], ("null put buffer"));
	hosts [i].put_buffer = htonll (b->buffers [i]);
      }
      buffer_lists_received++;
      break;
    default:
      insist (0);
  }  

  return 1;
  exception:
  if (pb)
    dump ((char*)pb, 200);
  return 0;
}

static int epilogue (time_t start_time)
{
  int i, resends = 0;
  gm_u64_t bytes_sent = 0;
  gm_u64_t bytes_received = 0;
  time_t stop_time = time (0);

  insist (hosts && num_hosts > 0 && my_index >= 0 && my_index < num_hosts);

  for (i = 0; i < num_hosts; i++)
  {
    insist (hosts[i].messages_sent == max_sends);
    insist (hosts[i].messages_completed == max_sends);
    insist (hosts[i].messages_received == max_sends);

    bytes_received += hosts[i].bytes_received;
    bytes_sent += hosts[i].bytes_sent;
    resends += hosts[i].messages_resent;
  }
  printf ("sent / received %d messages per host\n", max_sends);
  if (datagram_percent)
    printf ("not including datagrams\n");
  printf ("in lengths %s %d bytes.\n", same_size ? "of" : "up to", biggest);

  if (stop_time > start_time)
    printf ("at a rate of %u / %u MB/s\n",
	    (unsigned) (bytes_sent / (1024 * 1024) / (gm_u64_t) difftime (stop_time, start_time)),
	    (unsigned) (bytes_received / (1024 * 1024) / (gm_u64_t) difftime (stop_time, start_time)));

  if (resends)
    printf ("there were %d resends\n", resends);
  
  return 1;
  exception:return 0;
}

static void send_callback (struct gm_port *port, void *context, gm_status_t status);

static void drop_callback (struct gm_port *port, void *context, gm_status_t status)
{
  message_t *m = (message_t *) context;
  int length, to;

  insist (port && zone && context);

  to = htonl (m->to);
  length = htonl (m->length);
  
  insist (to >= 0 && to < num_hosts);
  insist (length >= sizeof (message_t) && length <= max_length + sizeof (message_t));
  
  insist (hosts[to].hostname && *hosts[to].hostname);
  insist (hosts [to].dropping);
  hosts [to].messages_resent++;
  
  gm_send_to_peer_with_callback (port, m, gm_min_size_for_length (length), length, GM_LOW_PRIORITY, hosts[to].id, send_callback, m);
  exception:;
}

void send_callback (struct gm_port *port, void *context, gm_status_t status)
{
  char s[GM_MAX_HOST_NAME_LEN * 2];
  message_t *m = (message_t *) context;
  int to, type;
  
  insist (port && zone && context);

  to = htonl (m->to);
  insist (to >= 0 && to < num_hosts);
  insist (hosts[to].hostname && *hosts[to].hostname);
  type = htonl (m->type);
  
  switch (status)
  {
    case GM_SEND_DROPPED:
      drop_callback (port, context, status);
      break;
    case GM_SUCCESS:
      gm_zone_free (zone, context);
      if (type == DATA || type == CABOOSE)
      {
	hosts [to].messages_completed++;
	insist (hosts [to].messages_completed <= hosts [to].messages_sent);
	if (hosts [to].messages_completed == hosts [to].messages_sent)
	  hosts [to].dropping = 0;
      }
      send_tokens++;
      break;
    case GM_SEND_TARGET_NODE_UNREACHABLE:
      failed = 1;
    case GM_SEND_TARGET_PORT_CLOSED:
      if (started)
	failed = 1;
      sprintf (s, "%s: send to %s failed", hosts [my_index].hostname, hosts[to].hostname);
      gm_perror (s, status);
    case GM_SEND_TIMED_OUT:
      if (!failed && max_resends && hosts [to].messages_resent >= max_resends)
      {
	printf ("%s: send to %s failed: exceeded --max-resends=%d\n", hosts [my_index].hostname, hosts[to].hostname, max_resends);
	failed = 1;
      }
      hosts [to].dropping = 1;
      if (!failed)
	gm_drop_sends (port, GM_LOW_PRIORITY, hosts[to].id, port_id, drop_callback, context);
      break;

    default:
      insist (0);
  }
  exception:;
}

static int receive (struct gm_port *port)
{
  gm_recv_event_t *e;

  insist (port);

  if (failed)
    return -1;

  e = non_blocking ? gm_receive (port) : gm_blocking_receive_no_spin (port);
  switch (gm_ntoh_u8 (e->recv.type))
  {
    case GM_RECV_EVENT:
      if (!handle_packet
	  (gm_ntohp (e->recv.buffer), gm_ntoh_u32 (e->recv.length)))
	return failed = -1;
      gm_provide_receive_buffer_with_tag (port, gm_ntohp (e->recv.buffer), gm_ntoh_u8 (e->recv.size), GM_LOW_PRIORITY, 0);
      return 1;
    case GM_ALARM_EVENT:
      alarmed = 0;
      gm_unknown (port, e);
      if (alarmed)
	return failed = -1;
    default:
      gm_unknown (port, e);
      break;
  }
  exception:return 0;
}

static int send_packet (struct gm_port*port, int type, int to)
{
  message_t*m;
  int length;

  insist (port);
  insist (num_hosts > 0);
  insist (to >= 0 && to < num_hosts);
  insist (!num_puts || hosts [my_index].put_buffer);
  
  while (!(m = (message_t *) gm_zone_malloc (zone, sizeof (*m))))
    if (receive (port) < 0)
      return 0;

  insist (m && (char *) m >= send_buffer && (char *) m < send_buffer + zone_length);  

  m->from = htonl (my_index);
  m->to = htonl (to);
  m->sequence = 0;
  m->type = htonl (type);
  m->put_buffer = htonll (hosts [my_index].put_buffer);
  
  while (!send_tokens)
    if (receive (port) < 0)
      return 0;

  length = sizeof (message_t);
  m->length = htonl (length);

  if (checksum)
    m->crc = htonl (gm_crc ((char *) m + sizeof (unsigned), sizeof (message_t) - sizeof (unsigned)));
  
  send_tokens--;
  gm_send_to_peer_with_callback (port, m, gm_min_size_for_length (length), length, GM_LOW_PRIORITY, hosts[to].id, send_callback, m);

  return 1;
  exception:return 0;

}


static int send_put_acks (struct gm_port*port)
{
  int i;
  
  insist (port);
  
  for (i = 0; i < num_hosts; i++)
  {
    for (;hosts[i].acks_needed; hosts[i].acks_needed--)
      if (!send_packet (port, PUT_ACK, i))
	return 0;
  }
  return 1;
  exception: return 0;
}


static int send_put (struct gm_port *port, int to, int num_bytes)
{
  message_t*pm, *m;
  int i, length;
  char*p;
  gm_u64_t dp;
  
  insist (port);
  insist (num_hosts > 0);
  insist (num_bytes >= 0);
  insist (to >= 0 && to < num_hosts);

  while (hosts [to].pending_puts == num_puts)
    if (receive (port) < 0 || !send_put_acks (port))
      return 0;
  
  while (!(pm = (message_t *) gm_zone_malloc (zone, sizeof (*pm) + num_bytes)))
    if (receive (port) < 0)
      return 0;

  while (!(m = (message_t *) gm_zone_malloc (zone, sizeof (*m))))
    if (receive (port) < 0)
      return 0;

  insist (pm && (char *) pm >= send_buffer && (char *) pm < send_buffer + zone_length);
  insist (m && (char *) m >= send_buffer && (char *) m < send_buffer + zone_length);
  
  insist (hosts[to].messages_sent < max_sends);
  hosts[to].messages_sent++;

  m->from = pm->from = htonl (my_index);
  m->to = pm->to = htonl (to);
  m->sequence = pm->sequence = htonl (hosts[to].messages_sent);
  m->type = htonl (CABOOSE);
  pm->type = htonl (PUT);

  while (send_tokens <= 1)
    if (receive (port) < 0)
      return 0;

  length = num_bytes + sizeof (*pm);
  pm->length = htonl (length);
  m->length = htonl (sizeof (message_t));
  hosts[to].bytes_sent += length;

  if (length > biggest)
    biggest = length;

  if (checksum)
  {
    char*p = (char*) pm + sizeof (message_t);
    for (i = 0; i < num_bytes; i++)
      p [i] = i;
    
    pm->crc = htonl (gm_crc ((char *) pm + sizeof (unsigned), length - sizeof (unsigned)));
    m->crc = htonl (gm_crc ((char *) m + sizeof (unsigned), sizeof (message_t) - sizeof (unsigned)));
  }

  send_tokens -= 2;
  hosts [to].pending_puts++;

  dp = hosts [to].put_buffer + (my_index * num_puts + hosts [to].puts_sent++ % num_puts) * max_length;
  /*printf ("sending put to %d, buffer is %x\n", to, (void*) (gm_size_t) dp);*/
  
  gm_put (port, pm, dp, length, GM_LOW_PRIORITY, hosts[to].id, port_id, send_callback, pm);
  gm_send_to_peer_with_callback (port, m, gm_min_size_for_length (sizeof (*m)), sizeof (*m), GM_LOW_PRIORITY, hosts[to].id, send_callback, m);
  
  return 1;
  exception:return 0;

}

static int send_buffer_list (struct gm_port*port, int to)
{
  buffer_list_t*m;
  int i, size, length;

  insist (port);
  insist (num_hosts > 0);
  insist (to >= 0 && to < num_hosts);
  insist (my_index == 0);

  length = sizeof (buffer_list_t) + (num_hosts - 1) * sizeof (gm_u64_t);
  
  while (!(m = (buffer_list_t*) gm_zone_malloc (zone, length)))
    if (receive (port) < 0)
      return 0;

  insist (m && (char *) m >= send_buffer && (char *) m < send_buffer + zone_length);  

  m->m.from = htonl (my_index);
  m->m.to = htonl (to);
  m->m.sequence = 0;
  m->m.type = htonl (BUFFER_LIST);
  m->m.length = htonl (length);
  
  for (i = 0; i < num_hosts; i++)
  {
    insist (hosts [i].put_buffer);
    m->buffers [i] = htonll (hosts [i].put_buffer);
  }  
  
  while (!send_tokens)
    if (receive (port) < 0)
      return 0;

  if (checksum)
    m->m.crc = htonl (gm_crc ((char *) m + sizeof (unsigned), length - sizeof (unsigned)));

  send_tokens--;
  gm_send_to_peer_with_callback (port, m, gm_min_size_for_length (length), length, GM_LOW_PRIORITY, hosts[to].id, send_callback, m);

  return 1;
  exception:return 0;

}

static int send (struct gm_port *port, int to, int num_bytes, int datagram)
{
  message_t *m;
  int size;
  int length;

  insist (port);
  insist (num_hosts > 0);
  insist (num_bytes >= 0);
  insist (to >= 0 && to < num_hosts);

  while (!(m = (message_t *) gm_zone_malloc (zone, sizeof (*m) + num_bytes)))
    if (receive (port) < 0)
      return 0;

  insist (m && (char *) m >= send_buffer && (char *) m < send_buffer + zone_length);
  insist (hosts[to].messages_sent < max_sends);

  if (datagram)
    hosts[to].datagrams_sent++;
  else
    hosts[to].messages_sent++;

  m->from = htonl (my_index);
  m->to = htonl (to);
  m->sequence = htonl (hosts[to].messages_sent);
  m->type = datagram ? htonl (DATAGRAM) : htonl (DATA);

  while (send_tokens < 0)
    if (receive (port) < 0)
      return 0;

  length = num_bytes + sizeof (*m);
  m->length = htonl (length);
  size = gm_min_size_for_length (length);

  hosts[to].bytes_sent += length;

  if (length > biggest)
    biggest = length;

  if (checksum)
    m->crc = htonl (gm_crc ((char *) m + sizeof (unsigned), length - sizeof (unsigned)));

  if (size < gm_min_message_size (port))
    size = gm_min_message_size (port);

  insist (size_mask & (1 << size));

  send_tokens--;
  if (datagram)
  {
    insist (length <= gm_mtu (port));
    gm_datagram_send (port, m, size, length, GM_LOW_PRIORITY, hosts[to].id,
		      port_id, send_callback, m);
  }
  else
    gm_send_to_peer_with_callback (port, m, size, length, GM_LOW_PRIORITY,
				   hosts[to].id, send_callback, m);

  return 1;
  exception:return 0;
}


static int barrier (struct gm_port *port)
{
  int i, j;

  insist (port && num_hosts > 0 && hosts);
  insist (my_id > 0);
  insist (my_index >= 0);
  insist (send_buffer);
  insist (barrier_seconds > 0 && barrier_seconds <= 60 * 60);
  
  gm_set_alarm (port, &alarm, barrier_seconds * 1024 * 1024, alarm_callback, 0);
  alarmed = 0;
  
  if (my_id == hosts[0].id)
  {
    while (!failed && !alarmed && barriers_received < num_hosts - 1)
      if (receive (port) < 0)
	return 0;

    if (alarmed || failed)
    {
      printf ("barrier failed for:\n");
      
      for (j = 1; j < num_hosts; j++)
      {
	if (!hosts [j].messages_received)
	  printf ("%s\n", hosts[j].hostname);
      }
      return 0;
    }
    
    for (i = 1; i < num_hosts; i++)
      if (num_puts && !send_buffer_list (port, i) || !send_packet (port, BARRIER, i))
	return 0; 
  }
  else
  {
    if (!send_packet (port, BARRIER, 0))
      return 0;
    
    while (!failed && !alarmed && !barriers_received && receive (port) <= 0);
  }

  if (!alarmed)
    gm_cancel_alarm (&alarm);
  
  return !failed && !alarmed;
  
  exception:return 0;
}

int get_random (struct gm_port *port, int *to, int *length, int *datagram)
{
  int i, h;

  insist (to && length && datagram);

  h = (rand () >> 13) % num_hosts;
  insist (h >= 0 && h < num_hosts);

  if (datagram_percent == 100)
    *datagram = 1;
  else if (datagram_percent == 0)
    *datagram = 0;
  else
    *datagram = rand () < RAND_MAX / 100 * datagram_percent;  

  for (;;)
  {
    for (i = 0; i < num_hosts && !hosts[i].dropping; i++);
    if (i == num_hosts)
      break;
    while (hosts [i].dropping)
      if (receive (port) < 0)
	return 0;
  }
  
  for (i = 0; i < num_hosts && hosts[(h + i) % num_hosts].messages_sent == max_sends; i++);
  
  if (i == num_hosts)
  {
    for (;;)
    {
      for (i = 0; i < num_hosts &&  hosts[i].messages_received == max_sends && !hosts [i].pending_puts && !hosts [i].acks_needed; i++);
      if (i == num_hosts)
      {
	while (send_tokens != max_send_tokens && receive (port) >= 0);
	return 0;
      }
      if (receive (port) < 0 || !send_put_acks (port))
	return 0;
    }
  }

  *to = (h + i) % num_hosts;
  insist (*to >= 0 && *to < num_hosts);
  insist (hosts[*to].messages_sent < max_sends);
  insist (!hosts[*to].dropping);

  insist (max_length >= sizeof (message_t));

  if (same_size)
    *length = max_length;
  else
    *length =
      max_length > sizeof (message_t) ?
      rand () %
      ((*datagram ? gm_mtu (port) : max_length) + 1 - sizeof (message_t)) : 0;

  insist (*length >= 0 && *length <= max_length);
  insist (!*datagram || *length <= gm_mtu (port) - sizeof (message_t));
  insist (num_puts || size_mask & (1 << gm_min_size_for_length (*length)));

  return 1;
  exception:return 0;
}

#ifdef GM_BUILDING_GM
static int
gm_stress (int argc, char **argv)
#else
     int main  (int argc, char **argv)
#endif
{
  struct gm_port *port = 0;
  int i, r, to, size, datagram;
  void *m;
  time_t start_time;

  for (i = 1; i < argc; i++)
  {
    if (!strcmp (argv[i], "--help") || !strcmp (argv[i], "-h"))
      usage (0);

    if ((r = option (i, argc, argv, "--unit=%i", "-u", &unit, 0, 16))
	|| (r = option (i, argc, argv, "--port=%i", "-p", &port_id, 0, 16))
	|| (r = option (i, argc, argv, "--num-sends=%i", "-n", &max_sends, 1, 1024 * 1024 * 1024))
	|| (r = option (i, argc, argv, "--num-puts=%i", "-t", &num_puts, 0, 100))
	|| (r = option (i, argc, argv, "--max-size=%i", "-m", &max_size, 4, 29))
	|| (r = option (i, argc, argv, "--min-receives=%i", "-r", &min_receives, 1, 1024))
	|| (r = option (i, argc, argv, "--barrier-seconds=%i", "-b", &barrier_seconds, 1, 60 * 60))
	|| (r = option (i, argc, argv, "--datagram=%i", "-d", &datagram_percent, 0, 100))
	|| (r = option (i, argc, argv, "--max-resends=%i", "-x", &max_resends, 0, 1024 * 1024 * 1024)))
    {
      i += r - 1;
      continue;
    }

    if (!strcmp (argv[i], "--non-blocking") || !strcmp (argv[i], "-o"))
      non_blocking = 1;
    else if (!strcmp (argv[i], "--checksum") || !strcmp (argv[i], "-c"))
      checksum = 1;
    else if (!strcmp (argv[i], "--same-size") || !strcmp (argv[i], "-s"))
      same_size = 1;
    else if (i < argc - 1 && !strcmp (argv[i], "-f"))
      filename = argv[++i];
    else if ((filename = strstr (argv[i], "--filename=")))
      filename += strlen ("--filename=");
    else
      usage (argv[i]);
  }

  if (filename && !create_hostnames (filename))
    careful_exit (port, 1);

  max_length = 2 << max_size;
  insist (max_length >= sizeof (message_t));

  initialized = gm_init ();
  gcheck (initialized, "gm_init failed");

  r =
    gm_open (&port, unit, port_id, "gm_stress",
	     (enum gm_api_version) GM_API_VERSION);
  gcheck (r, "gm_open failed");
  opened = 1;

  gm_initialize_alarm (&alarm);
  
  r = gm_get_node_id (port, &my_id);
  gcheck (r, "gm_get_node_id failed");

  if ((num_hosts = get_hosts (port, &hosts)) <= 0)
    careful_exit (port, 1);

  for (my_index = 0; my_index < num_hosts && hosts [my_index].id != my_id;
       my_index++);
  
  send_buffer = num_puts ? create_put_buffers (port, &zone_length) : create_buffers (port, max_length, min_receives, &zone_length);
  insist (send_buffer);

  zone = gm_zone_create_zone (send_buffer, zone_length);
  insist (zone);

  m = gm_zone_malloc (zone, max_length);
  insist (m);
  gm_zone_free (zone, m);

  max_send_tokens = send_tokens = gm_num_send_tokens (port) - 1;
  insist (send_tokens > 0);

  r = barrier (port);
  pinsis (r, ("barrier failed"));
  started = 1;
  
  srand (time (&start_time));
  signal (SIGINT, sig_handler);

  printf ("%d host%s (%d MB pinned).\n", num_hosts, num_hosts != 1 ? "s" : "", malloc_bytes / (1024 * 1024));
  fflush (stdout);

  while (get_random (port, &to, &size, &datagram) && (num_puts ? send_put (port, to, size) : send (port, to, size, datagram)));

  if (!failed)
    epilogue (start_time);

  careful_exit (port, failed);

  exception:
  careful_exit (port, 1);
  return 1;
}

#ifdef GM_BUILDING_GM
#define GM_MAIN gm_stress
#include "gm_main.h"
#endif

/*
  disable GM indentation virus
  please leave this in, so I don't have
  to keep putting it back.
  
  Local Variables:
  c-file-style:""
  End:
*/

  
