/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

/* This fairly simple programs is used to automatically generate
   conservative Makefile dependency lists for the source files
   specified on stdin.  Unlike other similar programs, this program is
   designed to generate a single dependency file including dependencies
   for all architectures.

   The program works as follows: The user supplies on STDIN a list of
   all source files within the project, including all header files.
   The program scans these files for include lines, and then outputs a
   dependency list, indicating that each C source file depends on all
   header files with the same basename.  This conservative strategy
   works quite well for the GM source tree.

   This program has some GM-specific features.  It treats
   the mcp/, drivers/, and libgm/ directories specially. */

#if HAVE_CONFIG_H
#if GM_BUILDING_GM
#include "gm_config.h"
#else
#include "config.h"
#endif
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>		/* malloc */
#include <string.h>
#include <stdio.h>
#endif

#if defined HAVE_UNISTD_H
#include <unistd.h>		/* mmap, lseek */
#endif

#ifdef HAVE_MMAP
#ifdef HAVE_SYS_MMAN_H
#include <sys/mman.h>		/* mmap */
#endif
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>		/* lseek */
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#if GM_OS_WIN2K
#include <io.h>
#define open _open
#define read _read
#define lseek _lseek
#define close _close
#define ssize_t int
#define getpagesize() 4096
#endif

#if STDC_HEADERS
# include <string.h>
#else
# if !HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif

/****************************************************************
 * Error codes
 ****************************************************************/

enum my_status {
  MY_SUCCESS,
  MY_FAILURE,
  MY_OUT_OF_MEMORY,
  MY_INVALID_PARAMETER
};

typedef enum my_status my_status_t;

static
const char *
my_strerror (my_status_t s)
{
  switch (s)
    {
    case MY_SUCCESS:
      return "succeeded";
    case MY_FAILURE:
      return "failed";
    case MY_OUT_OF_MEMORY:
      return "out of memory";
    default:
      return "bogus error code";
    }
}

static
void
my_perror (const char *prefix, my_status_t s)
{
  fprintf (stderr, "%s: %s", prefix, my_strerror (s));
}

/****************************************************************
 * hash table support
 *
 * There is no ANSI hash table support, and we don't want to drag in
 * the many files required of GM hash tables, so we include a minimal
 * hash table implementation here.
 ****************************************************************/

struct my_hash 
{
  struct bin 
  {
    char *key;
    void *data;
  } *bin;
  unsigned long bin_cnt;
  unsigned long population;
};

/* Create a hash table. */

static
struct my_hash *
my_create_hash (void)
{
  struct my_hash *h;

  h = calloc (1, sizeof (struct my_hash));
  if (!h)
    {
      return 0;
    }
  h->bin = calloc (32, sizeof (struct bin));
  if (!h->bin)
    {
      free (h);
      return 0;
    }
  h->bin_cnt = 32;
  return h;
}

/* Destroy a hash table. */

static
void
my_destroy_hash (struct my_hash *h)
{
  free (h->bin);
  free (h);
}

/* Compute a hash value for a key. */

static
unsigned long
_my_hash (char *key)
{
  unsigned long ret;
  
  for (ret=0; *key; key++)
    {
      ret = ret * 9 + (unsigned long) *key;
    }
  return ret;
}

/* find a key in the hash table. */

static
void *
my_hash_find (struct my_hash *h, char *key)
{
  unsigned long bin;

  for (bin = _my_hash (key) & (h->bin_cnt-1);
       h->bin[bin].key;
       bin = (bin + 1) & (h->bin_cnt-1))
    {
      if (!strcmp (key, h->bin[bin].key))
	{
	  return h->bin[bin].data;
	}
    }
  return 0;
}

static my_status_t _my_hash_grow (struct my_hash *h);

/* Insert a {key,value} pair in the hash table. */

static
my_status_t
my_hash_insert (struct my_hash *h, char *key, void *data)
{
  unsigned long bin;
  my_status_t status;

  /* Grow the hash table, if needed. */
  
  if (h->bin_cnt <= 2 * h->population)
    {
      status = _my_hash_grow (h);
      if (status != MY_SUCCESS)
	{
	  return status;
	}
    }

  /* Find an empty bin. */
  
  for (bin = _my_hash (key) & (h->bin_cnt-1);
       h->bin[bin].key;
       bin = (bin + 1) & (h->bin_cnt-1))
    ;

  /* Store the entry there. */

  h->bin[bin].key = key;
  h->bin[bin].data = data;
  h->population++;
  
  return MY_SUCCESS;
}

/* Double the number of bins in the hash table, preserving the data. */

static
my_status_t
_my_hash_grow (struct my_hash *h)
{
  unsigned long old_bin_cnt;
  struct bin *old_bin;

  /* Record the old bin information. */
  
  old_bin = h->bin;
  old_bin_cnt = h->bin_cnt;
  
  /* Make the hash table have twice as many bins, but be empty. */
      
  h->bin = calloc (2 * old_bin_cnt, sizeof (struct bin));
  if (!h->bin)
    {
      h->bin = old_bin;
      return MY_OUT_OF_MEMORY;
    }
  h->bin_cnt = 2 * old_bin_cnt;
  h->population = 0;
  
  /* Move the entries from the old to new bins. */
  
  {
    unsigned long i;
    
    i=old_bin_cnt;
    while (i--)
      {
	if (old_bin[i].key)
	  {
	    /* Since we grew the table, we know that the new hash table
	       has enough bins that it will not grow during this
	       insert. */
	    my_hash_insert (h, old_bin[i].key, old_bin[i].data);
	  }
      }
  }
  
  /* Free the old bins. */
  
  free (old_bin);
  return MY_SUCCESS;
}

/* Remove an entry from the hash table. */

static
void *
my_hash_remove (struct my_hash *h, char *key)
{
  unsigned long bin;
  void *ret = 0;
  
  /* Find the bin, and remove the entry, possibly leaving a hole in
     the cluster. */
  
  for (bin = _my_hash (key) & (h->bin_cnt-1);
       h->bin[bin].key;
       bin = (bin + 1) & (h->bin_cnt-1))
    {
      if (!strcmp (key, h->bin[bin].key))
	{
	  ret = h->bin[bin].data;
	  h->bin[bin].key = 0;
	  h->population--;
	  break;
	}
    }

  /* Fail if no match was found. */
  
  if (!ret)
    {
      return 0;
    }
  
  /* Rehash the any entries after the removed entry, since they
     may have collided with the entry we just removed. */

  for (bin = (bin + 1) & (h->bin_cnt-1);
       h->bin[bin].key;
       bin = (bin + 1) & (h->bin_cnt-1))
    {
      char *newkey;
      void *data;

      newkey = h->bin[bin].key;
      data = h->bin[bin].data;
      h->bin[bin].key = 0;
      h->population--;
      /* Since we just removed the entry, we can reinsert it without
	 risking the table growing. */
      my_hash_insert (h, newkey, data);
    }

  return ret;
}

/****************************************************************
 * Structures
 ****************************************************************/

/* Element in a list of files with a given basename. */

struct file_ref
{
  struct file_ref *next;
  struct file *file;
};

/* Element in a list of basenames of files included by a scanned file. */

struct basename_ref
{
  struct basename_ref *next;
  char *basename;
};

/* Element in a list of scanned files. */

struct file
{
  struct file *next;
  char *fullname;
  struct basename_ref *basename_refs;
};

/****************************************************************
 * globals
 ****************************************************************/

struct my_hash *basename_to_file_refs;

/* Boolean indicating whether the file suffixes should be doubled for
   file preprocessing. */

int double_suffix;

/****************************************************************
 * portability functions
 ****************************************************************/

/****************
 * strdup() emulation
 ****************/

static
char *
my_strdup (char *in)
{
  char *dupl;
  
  dupl = malloc (strlen (in) + 1);
  if (!dupl)
    {
      return 0;
    }
  strcpy (dupl, in);
  return dupl;
}

/****************
 * mmap() emulation
 ****************/

#ifndef HAVE_MMAP

#ifndef MAP_PRIVATE
#define MAP_PRIVATE 1
#endif

#ifndef PROT_READ
#define PROT_READ 1
#endif

static void *
my_mmap (void *addr, size_t len, int prot, int flags, int fd, off_t off)
{
  off_t pos;
  char *buf;

  /* save current position */

  pos = lseek (fd, 0, SEEK_CUR);
  if (pos == -1)
    {
      perror ("seek failed");
      goto abort_with_nothing;
    }

  /* seek to start of the region to mmap. */

  {
    off_t result;
    result = lseek (fd, off, SEEK_SET);
    if (result == -1)
      {
	perror ("seek failed");
	goto abort_with_nothing;
      }
  }

  /* allocate a buffer for the data */

  buf = (void *) malloc (len + 1);
  if (!buf)
    {
      perror ("could not malloc for my_mmap()");
      goto abort_with_pos;
    }

  /* read the requested data into the buffer and null terminate */

  {
    ssize_t result;

    result = read (fd, buf, len);
    if (result == -1)
      {
	perror ("could not read in file.\n");
	goto abort_with_buf;
      }
    buf[len] = 0;
  }

  /* reset the file position. */

  pos = lseek (fd, pos, SEEK_SET);
  if (pos == -1)
    {
      perror ("could not reset pos in my_mmap()");
      goto abort_with_buf;
    }

  return buf;

abort_with_buf:
  free (buf);
abort_with_pos:
  lseek (fd, pos, SEEK_SET);
abort_with_nothing:
  return (void *) -1;
}


int
my_munmap (void *ptr, size_t len)
{
  free (ptr);
  return 0;
}

#else /* HAVE_MMAP */

static void *
my_mmap (void *addr, size_t len, int prot, int flags, int fd, off_t off)
{
  return mmap (addr, len, prot, flags, fd, off);
}

static int
my_munmap (void *ptr, size_t len)
{
  return munmap (ptr, len);
}

#endif /* HAVE_MMAP */

/****************************************************************
 * Utility functions
 ****************************************************************/

static char *
my_basename (char *fullname)
{
  char *b;

  for (b = fullname + strlen (fullname) - 1; b > fullname; b--)
    {
      if (*b == '/')
	return b + 1;
    }
  return fullname;
}

static char *
suffix (char *fullname)
{
  char *b;

  for (b = fullname + strlen (fullname) - 1; b > fullname; b--)
    {
      if (*b == '.')
	return b;
    }
  return fullname + strlen (fullname);	/* no suffix */
}

static char *
_rindex (const char *s, int c)
{
  const char *ret = 0;

  while (*s)
    {
      if (*s == c)
	ret = s;
      s++;
    }
  return (char *) ret;
}

/****************************************************************
 * Header output functions
 ****************************************************************/

static my_status_t
output_headers1 (struct file *f, struct my_hash *exclude)
{
  struct basename_ref *br;
  my_status_t status;

  if (my_hash_find (exclude, f->fullname))
    return MY_SUCCESS;

  fputs ("\\\n\t", stdout);
  fputs (f->fullname, stdout);
  if (double_suffix)
    {
      fputs (suffix (f->fullname), stdout);
    }
  
  status = my_hash_insert (exclude, f->fullname, f);
  if (status != MY_SUCCESS)
    {
      fprintf (stderr, "Could not insert file in hash table.\n");
      return status;
    }

  for (br = f->basename_refs; br; br = br->next)
    {
      struct file_ref *fr;

      for (fr = my_hash_find (basename_to_file_refs, br->basename);
	   fr; fr = fr->next)
	{
	  status = output_headers1 (fr->file, exclude);
	  if (status != MY_SUCCESS)
	    return status;
	}
    }

  return MY_SUCCESS;
}

static my_status_t
output_headers (struct file *f)
{
  struct basename_ref *br;
  struct my_hash *exclude_hash;
  my_status_t status;

  /* allocate resources, recursively output headers, and free resources */

  exclude_hash = my_create_hash ();
  if (!exclude_hash)
    {
      fprintf (stderr, "Could not create hash table.\n");
      return MY_OUT_OF_MEMORY;
    }
  status = my_hash_insert (exclude_hash, f->fullname, f);
  if (status != MY_SUCCESS)
    {
      goto abort_with_hash;
    }
  for (br = f->basename_refs; br; br = br->next)
    {
      struct file_ref *fr;

      for (fr = my_hash_find (basename_to_file_refs, br->basename);
	   fr; fr = fr->next)
	{
	  status = output_headers1 (fr->file, exclude_hash);
	  if (status != MY_SUCCESS)
	    {
	      goto abort_with_hash;
	    }
	}
    }
  my_destroy_hash (exclude_hash);

  return MY_SUCCESS;

 abort_with_hash:
  my_destroy_hash (exclude_hash);
  return status;
}

/****************************************************************
 * Scanning
 ****************************************************************/

static my_status_t
notice_file_basename (struct file *f)
{
  char *b;
  struct file_ref *fr, *first;
  my_status_t status;

  b = my_basename (f->fullname);

  /* allocate storage for the file reference */

  fr = (struct file_ref *) calloc (1, sizeof (struct file_ref));
  if (!fr)
    {
      status = MY_OUT_OF_MEMORY;
      goto abort_with_nothing;
    }
  
  fr->file = f;

  /* add the file reference to the list of file references for the
     basename */

  first = my_hash_find (basename_to_file_refs, b);
  if (first)
    {
      fr->next = first->next;
      first->next = fr;
    }
  else
    {
      status = my_hash_insert (basename_to_file_refs, b, fr);
      if (status != MY_SUCCESS)
	goto abort_with_fr;
    }
  return MY_SUCCESS;

abort_with_fr:
  free (fr);
abort_with_nothing:
  return status;
}

static
void
forget_file_basename (struct file *f)
{
  struct file_ref *fr, *next;
  
  fr = my_hash_remove (basename_to_file_refs, my_basename (f->fullname));
  while (fr)
    {
      next = fr->next;
      free (fr);
      fr = next;
    }
}

/* record the fact that file F includes a file named BASENAME. */

static my_status_t
record_basename_ref (struct file *f, char *basename)
{
  my_status_t status;
  struct basename_ref *br;

  br = (struct basename_ref *) malloc (sizeof (struct basename_ref));
  if (!br)
    {
      status = MY_OUT_OF_MEMORY;
      goto abort_with_nothing;
    }
  br->basename = my_strdup (basename);
  if (!br->basename)
    {
      status = MY_OUT_OF_MEMORY;
      goto abort_with_br;
    }
  br->next = f->basename_refs;
  f->basename_refs = br;
  return MY_SUCCESS;

abort_with_br:
  free (br);
abort_with_nothing:
  return status;
}

/* Splice out a corresponding basename ref */
  
static void
forget_basename_ref (struct file *f, char *basename)
{
  struct basename_ref **br;

  for (br=&f->basename_refs; *br; br=&((*br)->next))
    {
      if (strcmp ((*br)->basename, basename) == 0)
	{
	  struct basename_ref *next;

	  next = (*br)->next;
	  free ((*br)->basename);
	  free (*br);
	  *br = next;
	  break;
	}
    }
}

/* Build a structure describing the file and all of the local includes
   that it references. */

static struct file *
scan (char *fullname)
{
  int fd;
  void *mm;
  char *cp, c, *tmp, *mark;
  struct file *f;
  long length;
  int err;

  /****************
   * Create the file record
   ****************/

  /* alloc */

  f = (struct file *) calloc (1, sizeof (struct file));
  if (!f)
    {
      perror ("could not malloc space for file");
      goto abort_with_nothing;
    }

  /* record the full relative path to the file */

  while (strncmp (fullname, "./", 2) == 0)
    fullname += 2;
  f->fullname = my_strdup (fullname);
  if (!f->fullname)
    goto abort_with_f;

  /****************
   * record that basename maps to the file
   ****************/

  if (notice_file_basename (f) != MY_SUCCESS)
    goto abort_with_fullname;

  /****************
   * Scan the file for local include file references.
   ****************/

  fd = open (fullname, O_RDONLY);
  if (fd == -1)
    {
      perror ("could not open file");
      goto abort_with_fullname;
    }

  /* determine the length of the file. */

  length = lseek (fd, 0, SEEK_END);
  if (length == -1)
    {
      perror ("could not determine file length");
      goto abort_with_open;
    }

  /* round length up to a page boundary. */

  length = (length + getpagesize () - 1) & ~(getpagesize () - 1);
  if (!length)
    goto munmap_done;

  /* map the file. */

  mm = my_mmap (0, length, PROT_READ, MAP_PRIVATE, fd, 0);
  if (mm == 0 || mm == (void *) -1)
    {
      perror ("could not map file");
      goto abort_with_open;
    }

#define NEXT(cnt) do							\
  {									\
    cp += cnt;								\
    if ((char *) cp - (char *) mm >= length)				\
      goto scan_done;							\
    c = *cp;								\
    if (!c)								\
      goto scan_done;							\
  } while (0)

  cp = mm;
  c = *cp;
  while (1)
    {
      /* In this loop outside the NEXT() macro, cp points to the
         current character, and c is a copy of that character. */

      /* match "#"' */

      if (c != '#')
	goto next_line;
      NEXT (1);

      /* eat whitespace */

      while (c == ' ' || c == '\t')
	NEXT (1);

      /* match "include" */

      if (strncmp (cp, "include", 7) != 0)
	goto next_line;
      NEXT (7);

      /* match whitespace */

      if (c != ' ' && c != '\t')
	goto next_line;

      /* eat extra whitespace */

      while (c == ' ' || c == '\t')
	NEXT (1);

      /* match "\"" */

      if (c != '"' && c != '<')
	goto next_line;
      NEXT (1);

      /* eat and record filename.  (eat first to catch EOF) */

      mark = cp;
      while (c != '"' && c != '>')
	NEXT (1);
      tmp = (char *) malloc (cp - mark + 1);
      if (!tmp)
	goto abort_with_my_mmap;
      memcpy (tmp, mark, cp - mark);
      tmp[cp - mark] = 0;
      if (record_basename_ref (f, my_basename (tmp)) != MY_SUCCESS)
	{
	  free (tmp);
	  goto abort_with_my_mmap;
	}
      free (tmp);

      /* find start of next line. */

    next_line:
      while (c != '\n')
	NEXT (1);
      NEXT (1);
    }
scan_done:

  err = my_munmap (mm, length);
  if (err)
    {
      perror ("could not munmap");
      goto abort_with_open;
    }
munmap_done:
  err = close (fd);
  if (err)
    {
      perror ("could not close");
      goto abort_with_fullname;
    }
  return f;

abort_with_my_mmap:
  my_munmap (mm, length);
abort_with_open:
  close (fd);
abort_with_fullname:
  free (f->fullname);
abort_with_f:
  free (f);
abort_with_nothing:
  return 0;
}

static
void
unscan (struct file *f)
{
  /* free all the basename refs */

  while (f->basename_refs)
    {
      forget_basename_ref (f, f->basename_refs->basename);
    }
  forget_file_basename (f);
  free (f->fullname);
  free (f);
}

/****************************************************************
 * Extra files not included in the source release
 ****************************************************************/

struct file *generated_files;

static my_status_t
notice_generated_file (char *fullname)
{
  my_status_t status;
  struct file *f;

  f = (struct file *) calloc (1, sizeof (struct file));
  if (!f)
    {
      status = MY_OUT_OF_MEMORY;
      my_perror ("could not malloc space for file", status);
      goto abort_with_nothing;
    }

  /* record the full relative_path to the file */

  while (strncmp (fullname, "./", 2) == 0)
    fullname += 2;
  f->fullname = my_strdup (fullname);
  if (!f->fullname)
    {
      status = MY_OUT_OF_MEMORY;
      my_perror ("could not strdup", status);
      goto abort_with_f;
    }

  /* notice the basename */

  status = notice_file_basename (f);
  if (status != MY_SUCCESS)
    {
      my_perror ("could not notice file basename", status);
      goto abort_with_fullname;
    }

  f->next = generated_files;
  generated_files = f;
  return MY_SUCCESS;

abort_with_fullname:
  free (f->fullname);
abort_with_f:
  free (f);
abort_with_nothing:
  return status;
}

static
void
forget_generated_file (struct file *f)
{
  forget_file_basename (f);
  free (f->fullname);
  free (f);
}

static void
forget_generated_files (void)
{
  struct file *f;

  while (generated_files)
    {
      f = generated_files;
      generated_files = f->next;
      forget_generated_file (f);
    }
  return ;
}

static my_status_t
notice_generated_files (char **f)
{
  my_status_t status;
  
  do
    {
      status = notice_generated_file (*f);
      if (status != MY_SUCCESS)
	{
	  forget_generated_files ();
	  return status;
	}
    }
  while (*++f);

  return MY_SUCCESS;
}

/****************************************************************
 * main program
 ****************************************************************/

int
main (int argc, char *argv[])
{
  const char *program_name;
  struct file *file, *scanned_files = 0;
  char fullname[256];
  my_status_t status = MY_SUCCESS;

  /* Process command-line parameters */

  program_name = argv[0];
  for (argc--, argv++; argc; argc--, argv++)
    {
      if (argv[0][0] != '-')
	{
	  /* Done processing options. */
	  break;
	}
      if (!strcmp ("--double-suffix", *argv))
	{
	  double_suffix = 1;
	  continue;
	}
      /* No matching argument template. */
      fprintf
	(stderr,
	 "Syntax:\n\t%s [--double-suffix] [generated_file ...]\n",
	 program_name);
      exit (MY_INVALID_PARAMETER);
    }
  /* argv is now the list of generated files. */
  
  /* initialize globals */

  basename_to_file_refs = my_create_hash ();
  if (!basename_to_file_refs)
    {
      status = MY_OUT_OF_MEMORY;
      fprintf (stderr,
	       "Could not create basename_to_file_refs hash table.\n");
      goto abort_with_hash;
    }

  /* scan each file and record its includes */

  while (fgets (fullname, 256, stdin))
    {
      int len;

      /* ignore empty lines */

      if (fullname[0] == 0 || fullname[0] == '\n')
	continue;

      /* strip trailing newline, if any */

      len = strlen (fullname);
      if (fullname[len - 1] == '\n')
	fullname[len - 1] = 0;

      /* scan the file for include files */

      file = scan (fullname);
      if (!file)
	{
	  status = MY_FAILURE;
	  goto abort_with_scanned_files;
	}
      file->next = scanned_files;
      scanned_files = file;
    }

  /* also notice the files that may need to be generated. */

  status = notice_generated_files (argv);
  if (status != MY_SUCCESS)
    {
      goto abort_with_scanned_files;
    }
  
  /* print the dependency list for each .c* file */

  for (file = scanned_files; file; file = file->next)
    {
      char *f, *s, save;

      f = file->fullname;
      s = suffix (f);

      /* output only .c* file dependencies */

      if (strncmp (s, ".c", 2) != 0)
	continue;

      /* create a timestamp to represent the dependencies by proxy.
	 This dramatically reduces the memory of Gnu make when it
	 reads the output. */

      fputs (f, stdout);
      if (double_suffix)
	{
	  fputs (suffix (f), stdout);
	}
      fputs (":", stdout);

#if 0
      /* ouput dependency on the preprocessed version of the file. */

      fputs (f, stdout);
      fputs (suffix (f), stdout);
#endif
  
      /* output dependencies on the timestamps of all the included headers */

      status = output_headers (file);
      if (status != MY_SUCCESS)
	{
	  fprintf (stderr, "failed to output headers.\n");
	  goto abort_having_noticed_generated_files;
	}

      /* output a rule to touch the timestamp. */

      fputs ("\n", stdout);
    }

 abort_having_noticed_generated_files:
  forget_generated_files ();
 abort_with_scanned_files:
  while (scanned_files)
    {
      /* splice out a scanned file */
      
      file = scanned_files;
      scanned_files = file->next;

      unscan (file);
    }
 abort_with_hash:
  my_destroy_hash (basename_to_file_refs);
  fflush (stdout);
  exit (status);
  return status;
}

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
