/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* This file implements platform-dependent support for LANai
   interfaces.  This file relies heavily on the platform-dependent
   functions declared in gm_impl.h, and implemented in the
   architecture's gm_arch.c file. */


#if !GM_KERNEL
#error GM_KERNEL not defined in gm_instance.c
#endif

#include "gm_call_trace.h"
#include "gm_cpp.h"
#include "gm_debug.h"
#include "gm_debug_board_init.h"
#include "gm_debug_host_dma.h"
#include "gm_debug_instance_state.h"
#include "gm_debug_lanai_dma.h"
#include "gm_debug_sync.h"
#include "gm_enable_debug_counters.h"
#include "gm_enable_firmware.h"
#include "gm_enable_force_64bit_mode.h"
#include "gm_enable_prototype_nic_support.h"
#include "gm_instance.h"
#include "gm_internal.h"
#include "gm_pio.h"
#include "gm_ptr_hash.h"
#include "zlib.h"

#define GM_LX(x) (31 - ((x) ^ 7))

/****************************************************************
 * Forward declarations
 ****************************************************************/
#if GM_SUPPORT_PCI_REV_3
static gm_status_t gm_update_clockval_rev3 (gm_instance_state_t * is);
#endif /* GM_SUPPORT_PCI_REV_3 */
	  
/************************************************************************
 * Package initialization/finalization
 ************************************************************************/

/* convert CSPI_MAP26xx to use 0/1 instead of def/undef */
#ifdef CSPI_MAP26xx
#  undef CSPI_MAP26xx
#  define CSPI_MAP26xx 1
#else
#  define CSPI_MAP26xx 0
#endif

#if CSPI_MAP26xx
#if GM_OS_LINUX
#include <asm/cspi.h>
#elif GM_OS_VXWORKS
#include "boardCsr.h"
#else
#error GM does not support this OS for CSPI 26xx/27xx/28xx boards
#endif /* GM_OS... */
#endif /* CSPI_MAP26xx */

static unsigned long gm_init_instance_cnt;
static struct gm_hash *gm_instance_hash;

static gm_status_t
gm_init_instance_support (void)
{
  GM_CALLED ();
#if !GM_ETHERBOOT
  if (!gm_init_instance_cnt)
    {
      gm_instance_hash = gm_create_hash (gm_hash_compare_ints,
					 gm_hash_hash_int,
					 sizeof (unsigned int), 0, 1, 0);
      if (!gm_instance_hash)
	return GM_FAILURE;
    }
#endif
  gm_init_instance_cnt++;
  GM_RETURN_STATUS (GM_SUCCESS);
}

static void
gm_finalize_instance_support (void)
{
  GM_CALLED ();
  gm_assert (gm_init_instance_cnt > 0);
  if (!--gm_init_instance_cnt)
    {
#if !GM_ETHERBOOT
      gm_destroy_hash (gm_instance_hash);
#endif
    }
  GM_RETURN_NOTHING ();
}

/**********************************************************************
 * Interface register abstraction
 **********************************************************************/

/* default do-nothing handler */

static void
gm_oo_set_abort (gm_instance_state_t * is, gm_u32_t val)
{
  GM_PARAMETER_MAY_BE_UNUSED (is);
  GM_PARAMETER_MAY_BE_UNUSED (val);

  GM_WARN (
	   ("Internal driver error: Uninitialized dynamic handler called.\n"));
  gm_abort ();
}

static gm_u32_t
gm_oo_get_abort (gm_instance_state_t * is)
{
  GM_PARAMETER_MAY_BE_UNUSED (is);

  GM_WARN (
	   ("Internal driver error: Uninitialized dynamic handler called.\n"));
  gm_abort ();
  return (0);
}

static
void
gm_oo_able_abort (gm_instance_state_t *is)
{
  GM_PARAMETER_MAY_BE_UNUSED (is);

  GM_WARN (
	   ("Internal driver error: Uninitialized dynamic handler called.\n"));
  gm_abort ();
}

static
int
gm_oo_interrupting_abort (gm_instance_state_t *is)
{
  GM_PARAMETER_MAY_BE_UNUSED (is);
  GM_WARN
    (("Internal driver error: Uninitialized dynamic handler called.\n"));
  gm_abort ();
  return 0;
}

static
void
gm_oo_claim_interrupt_abort (gm_instance_state_t *is)
{
  GM_PARAMETER_MAY_BE_UNUSED (is);
  GM_WARN
    (("Internal driver error: Uninitialized dynamic handler called.\n"));
  gm_abort ();
}

/****************
 * Version checking functions
 ****************/

/* These functions test for recognized version numbers.  The should
   only accept version numbers that are known to work. */

static int
gm_is_l4 (gm_instance_state_t * is)
{
  switch (is->lanai.eeprom.lanai_cpu_version)
    {
    case 0x400:
      return GM_ENABLE_PROTOTYPE_NIC_SUPPORT;

    case 0x401:
    case 0x402:
    case 0x403:
      return 1;

    default:
      break;			/* for some really picky compilers */
    }
  return 0;
}

static int
gm_is_l5 (gm_instance_state_t * is)
{
  switch (is->lanai.eeprom.lanai_cpu_version)
    {
    case 0x500:
    case 0x501:
    case 0x502:
    case 0x503:
    case 0x504:
    case 0x505:
    case 0x506:
      return 1;

    default:
      break;			/* for some really picky compilers */
    }
  return 0;
}

static int
gm_is_l6 (gm_instance_state_t * is)
{
  switch (is->lanai.eeprom.lanai_cpu_version)
    {
    case 0x601:
    case 0x602:
      return 1;

    default:
      break;			/* for some really picky compilers */
    }
  return 0;
}

static int
gm_is_l7 (gm_instance_state_t * is)
{
  switch (is->lanai.eeprom.lanai_cpu_version)
    {
    case 0x700:		/* This was missing ???  - Max 2000/06/07 */
    case 0x701:
    case 0x702:
    case 0x703:
    case 0x704:
    case 0x705:
      return 1;

    default:
      break;			/* for some really picky compilers */
    }
  return 0;
}

static int
gm_is_l8 (gm_instance_state_t * is)
{
  switch (is->lanai.eeprom.lanai_cpu_version)
    {
    case 0x800:
    case 0x801:
      return GM_ENABLE_PROTOTYPE_NIC_SUPPORT;

    case 0x802:
      return 1;

    default:
      break;			/* for some really picky compilers */
    }
  return 0;
}

int
gm_is_l9 (gm_instance_state_t * is)
{
  switch (is->lanai.eeprom.lanai_cpu_version)
    {
    case 0x900:
    case 0x901:
    case 0x902:
    case 0x903:
    case 0x904:
    case 0x905:
      return 1;

    default:
      break;			/* for some really picky compilers */
    }
  return 0;
}

int
gm_is_lX (gm_instance_state_t *is)
{
  GM_CALLED ();
  GM_RETURN_INT
    (GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4
     || GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5);
}

/********************
 * LANai-version-dependent functions
 ********************/

/****
 * LANai4
 ****/

static void
gm_set_EIMR_l4 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l4.EIMR, val);
}

static void
gm_set_ISR_l4 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l4.ISR, val);
}

static gm_u32_t
gm_get_EIMR_l4 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l4.EIMR);
}

static gm_u32_t
gm_get_ISR_l4 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l4.ISR);
}

static void
gm_enable_interrupts_l4 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l4.EIMR, GM_HOST_SIG_BIT);
}

static void
gm_disable_interrupts_l4 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l4.EIMR, 0);
}

static int
gm_interrupting_l4 (gm_instance_state_t *is)
{
  return gm_read_lanai_special_reg_u32 (is, l4.ISR) & GM_HOST_SIG_BIT;
}

static void
gm_claim_interrupt_l4 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l4.ISR, GM_HOST_SIG_BIT);
}

/****
 * LANai5
 ****/

static void
gm_set_EIMR_l5 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l5.EIMR, val);
}

static void
gm_set_ISR_l5 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l5.ISR, val);
}

static gm_u32_t
gm_get_EIMR_l5 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l5.EIMR);
}

static gm_u32_t
gm_get_ISR_l5 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l5.ISR);
}

static void
gm_enable_interrupts_l5 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l5.EIMR, GM_HOST_SIG_BIT);
}

static void
gm_disable_interrupts_l5 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l5.EIMR, 0);
}

static int
gm_interrupting_l5 (gm_instance_state_t *is)
{
  return gm_read_lanai_special_reg_u32 (is, l5.ISR) & GM_HOST_SIG_BIT;
}

static void
gm_claim_interrupt_l5 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l5.ISR, GM_HOST_SIG_BIT);
}

/****
 * LANai6
 ****/

static void
gm_set_EIMR_l6 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l6.EIMR, val);
}

static void
gm_set_ISR_l6 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l6.ISR, val);
}

static gm_u32_t
gm_get_EIMR_l6 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l6.EIMR);
}

static gm_u32_t
gm_get_ISR_l6 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l6.ISR);
}

static void
gm_enable_interrupts_l6 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l6.EIMR, GM_HOST_SIG_BIT);
}

static void
gm_disable_interrupts_l6 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l6.EIMR, 0);
}

static int
gm_interrupting_l6 (gm_instance_state_t *is)
{
  return gm_read_lanai_special_reg_u32 (is, l6.ISR) & GM_HOST_SIG_BIT;
}

static void
gm_claim_interrupt_l6 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l6.ISR, GM_HOST_SIG_BIT);
}

/****
 * LANai7
 ****/

static void
gm_set_EIMR_l7 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l7.EIMR, val);
}

static void
gm_set_ISR_l7 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l7.ISR, val);
}

static gm_u32_t
gm_get_EIMR_l7 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l7.EIMR);
}

static gm_u32_t
gm_get_ISR_l7 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l7.ISR);
}

static void
gm_enable_interrupts_l7 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l7.EIMR, GM_HOST_SIG_BIT);
}

static void
gm_disable_interrupts_l7 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l7.EIMR, 0);
}

static int
gm_interrupting_l7 (gm_instance_state_t *is)
{
  return gm_read_lanai_special_reg_u32 (is, l7.ISR) & GM_HOST_SIG_BIT;
}

static void
gm_claim_interrupt_l7 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l7.ISR, GM_HOST_SIG_BIT);
}

/****
 * LANai8
 ****/

static void
gm_set_EIMR_l8 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l8.EIMR, val);
}

static void
gm_set_ISR_l8 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l8.ISR, val);
}

static gm_u32_t
gm_get_EIMR_l8 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l8.EIMR);
}

static gm_u32_t
gm_get_ISR_l8 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l8.ISR);
}

static void
gm_enable_interrupts_l8 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l8.EIMR, GM_HOST_SIG_BIT);
}

static void
gm_disable_interrupts_l8 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l8.EIMR, 0);
}

static int
gm_interrupting_l8 (gm_instance_state_t *is)
{
  return gm_read_lanai_special_reg_u32 (is, l8.ISR) & GM_HOST_SIG_BIT;
}

static void
gm_claim_interrupt_l8 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l8.ISR, GM_HOST_SIG_BIT);
}

/****
 * LANai9
 ****/

static void
gm_set_EIMR_l9 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l9.EIMR, val);
}

static void
gm_set_ISR_l9 (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, l9.ISR, val);
}

static gm_u32_t
gm_get_EIMR_l9 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l9.EIMR);
}

static gm_u32_t
gm_get_ISR_l9 (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, l9.ISR);
}

static void
gm_enable_interrupts_l9 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l9.EIMR, GM_HOST_SIG_BIT);
}

static void
gm_disable_interrupts_l9 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l9.EIMR, 0);
}

static int
gm_interrupting_l9 (gm_instance_state_t *is)
{
  return gm_read_lanai_special_reg_u32 (is, l9.ISR) & GM_HOST_SIG_BIT;
}


/****
 * Slower gm_interrupting using EIMR for use by drivers which want to
 * test interrupt delivery.  
 *
 * This function is exported for debugging use in arch-specific code
 ****/
int
gm_test_interrupting_pre_lx (gm_instance_state_t *is)
{
  gm_u32_t eisr, eimr;

  eisr = is->get_ISR (is);
  eimr = is->get_EIMR (is);
  return (int) (eisr & eimr);
}

static void
gm_claim_interrupt_l9 (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, l9.ISR, GM_HOST_SIG_BIT);
}

/****
 * LANaiX
 ****/

#define GM_DEBUG_LX 0

static void
gm_set_EIMR_lX (gm_instance_state_t * is, gm_u32_t val)
{
  /* The LANaiX does not need to have its EIMR set. */
}

static void
gm_set_ISR_lX (gm_instance_state_t * is, gm_u32_t val)
{
  gm_write_lanai_special_reg_u32 (is, lX.ISR, val);
}

static gm_u32_t
gm_get_EIMR_lX (gm_instance_state_t * is)
{
  return GM_LX_REQ_ACK_0;
}

static gm_u32_t
gm_get_ISR_lX (gm_instance_state_t * is)
{
  return gm_read_lanai_special_reg_u32 (is, lX.ISR);
}

static void
gm_enable_interrupts_lX (gm_instance_state_t *is)
{
  is->lanai.control_regs[0x10] = gm_hton_u32 (1 << GM_LX (27));
}

static void
gm_disable_interrupts_lX (gm_instance_state_t *is)
{
  is->lanai.control_regs[0x10] = gm_hton_u32 (1 << GM_LX (26));
}

static int
gm_interrupting_lX (gm_instance_state_t *is)
{
  return (is->msi_enabled 
	  || (gm_read_lanai_special_reg_u32 (is, lX.ISR) & GM_LX_REQ_ACK_0));
}

static void
gm_claim_interrupt_lX (gm_instance_state_t *is)
{
  gm_write_lanai_special_reg_u32 (is, lX.ISR, GM_LX_REQ_ACK_0);
}

/****************
 * MSI-cabable
 ****************/

#define GM_DEBUG_MSI 0

/* Setup message signalled interrupts for the card if needed. */

gm_status_t
gm_setup_capabilities_if_needed (gm_instance_state_t *is)
{
  gm_status_t status = GM_SUCCESS;
  
  GM_CALLED ();
  
  /* Detect and handle the various board types. */

  is->msi_enabled = 0;
  if (gm_is_lX (is))
    {
      gm_u8_t cap;
      gm_u16_t pci_status;
      gm_u32_t addr_high, addr_low;
      gm_u16_t data;
      int checked_MSI = 0;
      int set_read_count = 0;
		  
      /* Fetch the status register so we can confirm that the board
	 has a capabilities list. */

      status
	= (gm_arch_read_pci_config_16
	   (is, GM_OFFSETOF (gm_pci_config_t, Status), &pci_status));
      if (status != GM_SUCCESS)
	{
	  goto LX_done;
	}

      /* Confirm that the board has a capabilities list. */

      if (!GM_PCI_STATUS_CAPABILITIES_LIST (pci_status))
	{
	  GM_WARN (("LANaiX board does not have new capabilities.\n"));
	  _GM_WARN (("(status = 0x04%x)\n", pci_status));
	  status = GM_INTERNAL_ERROR;
	  goto LX_done;
	}
      
      /* Fetch the offset of the first capability structure */
      
      status
	= (gm_arch_read_pci_config_8
	   (is, GM_OFFSETOF (gm_pci_config_t, Capabilities_Pointer), &cap));
      if (status != GM_SUCCESS)
	{
	  GM_WARN (("Could not read Capabilities pointer.\n"));
	  goto LX_done;
	}
      cap = GM_PCI_CAP_POINTER_CLEANUP (cap);

      /* Scan the linked list of capabilities structures for a message
	 signalled interrupts capability. */
      
      GM_PRINT (GM_DEBUG_MSI, ("Scanning capabilities list\n"));
      while (cap)
	{
	  gm_u8_t cap_id;
	  
	  _GM_PRINT (GM_DEBUG_MSI, ("cap = 0x%x\n", (int) cap));
	  
	  /* Read the capability type */
	  
	  status 
	    = (gm_arch_read_pci_config_8
	       (is,
		cap + GM_OFFSETOF (gm_pci_capability_t, common.Capability_ID),
		&cap_id));
	  if (status != GM_SUCCESS)
	    {
	      GM_WARN (("Could not read Capability ID\n"));
	      goto LX_done;
	    }
	  
	  /* Check if we have found the capability we are searching for. */

	  if (cap_id == GM_PCI_CAP_MESSAGE_SIGNALLED_INTERRUPTS)
	    {
	      gm_u16_t message_control;

	      _GM_PRINT (GM_DEBUG_MSI, ("Found an MSI capability\n"));
	      
	      /* We found the board's MSI capability. */

	      /* Read the message_control field from the capability struct,
	         which has some bits we need. */

	      _GM_PRINT (GM_DEBUG_MSI, ("Reading Message_Control.\n"));
	      status
		= (gm_arch_read_pci_config_16
		   (is,
		    cap + GM_OFFSETOF (gm_pci_capability_t,
					msi.Message_Control),
		    &message_control));
	      if (status != GM_SUCCESS)
		{
		  GM_WARN (("Could not read msi.Message_Control\n"));
		  goto LX_done;
		}
	      
	      /* Confirm that the board is 64-bit MSI capable. */

	      _GM_PRINT (GM_DEBUG_MSI, ("Confirming 64-bit capable.\n"));
	      if (!GM_PCI_CAP_MESSAGE_CONTROL_64_BIT_ADDRESS_CAPABLE
		  (message_control))
		{
		  GM_WARN (("board is not 64-bit address capable.\n"));
		  status = GM_INTERNAL_ERROR;
		  goto LX_done;
		}

	      /* Determine if the OS enabled MSI. */

	      _GM_PRINT (GM_DEBUG_MSI, ("Checking MSI enable.\n"));
	      if (GM_PCI_CAP_MESSAGE_CONTROL_MSI_ENABLE (message_control))
		{
		  /* The OS told the card to use MSI, so tell the firmware
		     the details. */

		  /* Read the MSI settings from PCI config space. */

		  status
		    = (gm_arch_read_pci_config_32
		       (is,
			cap + GM_OFFSETOF (gm_pci_capability_t,
					   msi64.Message_Address),
			&addr_low));
		  if (status != GM_SUCCESS)
		    {
		      GM_WARN (("Error reading msi64.MessageAddress\n"));
		      goto LX_done;
		    }
		  
		  status
		    = (gm_arch_read_pci_config_32
		       (is,
			cap + GM_OFFSETOF (gm_pci_capability_t,
					   msi64.Message_Upper_Address),
			&addr_high));
		  if (status != GM_SUCCESS)
		    {
		      GM_WARN
			(("Error reading msi64.Message_Upper_Address\n"));
		      goto LX_done;
		    }
		  
		  status
		    = (gm_arch_read_pci_config_16
		       (is,
			cap + GM_OFFSETOF (gm_pci_capability_t,
					   msi64.Message_Data),
			&data));
		  if (status != GM_SUCCESS)
		    {
		      GM_WARN (("Error reading msi64.Message_Data\n"));
		      goto LX_done;
		    }
		  
		  /* Write these MSI settings to the NIC globals. */
		  
		  {
		    gm_dp_t dp;

		    dp = ((gm_dp_t) addr_high << 32) + addr_low;
		    gm_write_lanai_global_dp (is, msi.addr, dp);
		  }
		  gm_write_lanai_global_u16 (is, msi.data, data);

		  /* Enable MSI. */

		  _GM_PRINT (GM_DEBUG_MSI, ("Enabling MSI\n"));
		  
		  GM_STBAR ();
		  gm_write_lanai_global_u32 (is, msi.enabled, 1);
		  is->msi_enabled = 1;
		}
	      else
		{
		  /* The OS told the card not to use MSI. */

		  _GM_PRINT (GM_DEBUG_MSI, ("OS disabled MSI\n"));
		  gm_assert (is->msi_enabled == 0);
		}

	      /* Record that checked for MSI. */
	      
	      checked_MSI = 1;
	    }
	  else if (cap_id == GM_PCI_CAP_PCI_X && GM_PCI_X_FORCE_LARGE_READ)
	    {
	      gm_u16_t command;

	      _GM_PRINT (GM_DEBUG_MSI, ("Found a PCI-X capability\n"));
	      
	      /* Read the PCI-X command value */

	      status
		= (gm_arch_read_pci_config_16
		   (is,
		    cap + GM_OFFSETOF (gm_pci_capability_t, pci_x.Command),
		    &command));
	      if (status != GM_SUCCESS)
		{
		  GM_WARN (("Could not read pci_x.Command\n"));
		  goto LX_done;
		}
	      
	      /* Set the Maximum memory read byte count to 4096 */

	      command
		= ((command
		    & ~GM_PCI_CAP_PCI_X_COMMAND_MAX_MEM_READ_BYTE_CNT_MASK)
		   | GM_PCI_CAP_PCI_X_COMMAND_MAX_MEM_READ_BYTE_CNT_4096);
	      
	      /* Write back the PCI-X command value */

	      status
		= (gm_arch_write_pci_config_16
		   (is,
		    cap + GM_OFFSETOF (gm_pci_capability_t, pci_x.Command),
		    command));
	      if (status != GM_SUCCESS)
		{
		  GM_WARN (("Could not write pci_x.Command\n"));
		  goto LX_done;
		}

	      /* Record that we set the maximum read count. */

	      set_read_count = 1;
	    }

	  /* Check if we completed all the capability handling required. */

	  if (checked_MSI
	      && (!GM_PCI_X_FORCE_LARGE_READ || set_read_count))
	    {
	      gm_assert (status == GM_SUCCESS);
	      goto LX_done;
	    }
	  
	  /* The capability was not an MSI capability, so advance to the
	     next capability. */
	  
	  status
	    = (gm_arch_read_pci_config_8
	       (is,
		cap + GM_OFFSETOF (gm_pci_capability_t, common.Next_Pointer),
		&cap));
	  if (status != GM_SUCCESS)
	    {
	      GM_WARN (("Error reading capability's Next_Pointer\n"));
	      goto LX_done;
	    }

	  cap = GM_PCI_CAP_POINTER_CLEANUP (cap);
	}
      GM_WARN (("NIC is missing a PCI capability.\n"));
      
      /* We did not find a MSI capability for an LX card.  This should
	 never happen. */
      status = GM_INTERNAL_ERROR;

    LX_done:
      /* Print debug info */
  
      if (GM_DEBUG_MSI)
	{
	  if (status == GM_SUCCESS)
	    {
	      if (is->msi_enabled)
		{
		  GM_PRINT (1, ("Using MSI, addr=0x%08x%08x, data=0x%04x\n",
				addr_high, addr_low, data));
		}
	      else
		{
		  GM_PRINT (1, ("Not using MSI\n"));
		}
	    }
	  else
	    {
	      gm_perror ("not using MSI", status);
	    }
	}
  
      goto done;
    }
  
  /* Unrecognized card. */

  is->msi_enabled = 0;
  
 done:

  /* Return */
  
  GM_RETURN_STATUS (status);
}

int
gm_instance_using_msi (gm_instance_state_t *is)
{
  return is->msi_enabled;
}

/****************
 * 
 ****************/

void
gm_set_EIMR (gm_instance_state_t * is, gm_u32_t val)
{
  is->_set_EIMR (is, val);
  /* is->pausible = val;  moved this functionality to gm_enable/disable interrupts */
}

#if GM_SUPPORT_PCI
/**********************************************************************
 * PCI register access.
 **********************************************************************/

static gm_status_t
gm_read_pci_cache_line_size (gm_instance_state_t * is, gm_u8_t * val)
{
  return gm_arch_read_pci_config_8 (is,
				    GM_OFFSETOF (gm_pci_config_t,
						 Cache_Line_Size), val);
}

static gm_status_t
gm_read_pci_config_command (gm_instance_state_t * is, gm_u16_t * command)
{
  return gm_arch_read_pci_config_16 (is,
				     GM_OFFSETOF (gm_pci_config_t,
						  Command), command);
}

static gm_status_t
gm_write_pci_config_command (gm_instance_state_t * is, gm_u16_t value)
{
  return gm_arch_write_pci_config_16 (is,
				      GM_OFFSETOF (gm_pci_config_t, Command),
				      value);
}

static gm_status_t
gm_read_pci_config_revision_id (gm_instance_state_t * is, gm_u8_t * value)
{
  return gm_arch_read_pci_config_8 (is,
				    GM_OFFSETOF (gm_pci_config_t,
						 Revision_ID), value);
}

static gm_status_t
gm_read_pci_config_status (gm_instance_state_t * is, gm_u16_t * status)
{
  return gm_arch_read_pci_config_16 (is,
				     GM_OFFSETOF (gm_pci_config_t,
						  Status), status);
}

static gm_status_t
gm_write_pci_config_status (gm_instance_state_t * is, gm_u16_t value)
{
  return gm_arch_write_pci_config_16 (is,
				      GM_OFFSETOF (gm_pci_config_t, Status),
				      value);
}


#endif /* GM_SUPPORT_PCI */
/**********************************************************************
 * LANai register setting functions  (not performance critical)
 **********************************************************************/

#define GM_HACKED_STS 0

static gm_status_t
gm_setup_DMA_engine (gm_instance_state_t * is)
{
  /* Verify that this function was not called out-of-order */

  if (!is->lanai.globals)
    return GM_FAILURE;

#if CSPI_MAP26xx
#if GM_SUPPORT_L5
  if (gm_is_l5(is))
    {
      /* not sure we need to do anything */
      /* Set burst size to 32 bytes (the cache line size) and PCI
         compatibility. */
      gm_write_lanai_special_reg_u32 (is, l5.BURST, 0x5);
      GM_PRINT (1, ("Finished DMA setup for cspi lanai 5\n"));
      return GM_SUCCESS;
    }
#endif /* GM_SUPPORT_L5 */
#if GM_SUPPORT_L9
  if (gm_is_l9(is))
    {
      CSPI_REG4 *descriptors;
      gm_lp_t pointer;

      
	/* Tell the LANai where to find the DMA descriptor for
	   "channel 0" */
      descriptors = (CSPI_REG4 *) ((gm_u8_t *) is->lanai.control_regs
				 + DMA0_REG_OFFSET);
      pointer = (gm_lp_t) ((gm_u8_t *) is->lanai.globals
			   + GM_OFFSETOF (gm_lanai_globals_t,
					  dma_descriptor) -
			   (gm_u8_t *) is->lanai.sram);
      descriptors[0].reg = pointer;
      GM_PRINT (1,("DMA descriptor at LANai address 0x%x.\n", pointer));


      /* Tell the LANai where to find the DMA descriptor for
	 "channel 2" */
      
      pointer = (gm_lp_t) ((gm_u8_t *) is->lanai.globals
			   + GM_OFFSETOF (gm_lanai_globals_t,
					  user_dma_descriptor[0])
			   - (gm_u8_t *) is->lanai.sram);
      descriptors[2].reg = pointer;
      GM_PRINT (1,("DMA descriptor at LANai address 0x%x.\n", pointer));

      GM_PRINT (1,("Finished DMA setup for cspi lanai 9\n"));
      return GM_SUCCESS;
    }
#endif /* GM_SUPPORT_L9 */
#endif /* CSPI_MAP26xx */
  
#if GM_SUPPORT_SBUS
  if (GM_SUPPORT_SBUS && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      if (gm_is_l4 (is))
	{
	  if (GM_HACKED_STS)
	    {
	      gm_write_lanai_special_reg_u32 (is, l4.DMA_STS, 0);
	      GM_PRINT (GM_PRINT_LEVEL >= 4,
			("gm_setup_DMA_engine: GM_HACKED_STS 'sts'"
			 " set to 0 would have been 0x%x.\n",
			 gm_arch_dma_region_status (&is->page_hash.
						    segment[0].dma_region)));
	    }
	  else
	    {
	      gm_write_lanai_special_reg_u32
		(is, l4.DMA_STS,
		 gm_arch_dma_region_status (&is->page_hash.
					    segment[0].dma_region));
	      GM_PRINT (GM_PRINT_LEVEL >= 4,
			("gm_setup_DMA_engine: DMA 'sts' set to 0x%x.\n",
			 gm_arch_dma_region_status (&is->page_hash.
						    segment[0].dma_region)));
	    }
	  return GM_SUCCESS;
	}
    }
#endif /* GM_SUPPORT_SBUS */
#if GM_SUPPORT_PCI
  
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if (GM_SUPPORT_PCI_REV_1 && is->ifc.pci.config.Revision_ID == 1)
	{
	  if (gm_is_l4 (is))
	    {
	      /* DMA_STS is always 0xf for L4 PCI */
	      gm_write_lanai_special_reg_u32 (is, l4.DMA_STS, 0xf);
	      return GM_SUCCESS;
	    }
	}
      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  if (gm_is_l5 (is))
	    {
	      gm_lp_t *descriptors;
	      gm_lp_t pointer;

	      /* This sets the burst size to 32 bytes (the
	         cache line size) and PCI compatibility. */
	      gm_write_lanai_special_reg_u32 (is, l5.BURST, 0x5);

	      /* Tell the LANai where to find the DMA descriptor for
	         "channel 0" */

	      descriptors = (gm_lp_t *) ((gm_u8_t *) is->board_base
					 + 0x00800100);
	      pointer = (gm_lp_t) ((gm_u8_t *) is->lanai.globals
				   + GM_OFFSETOF (gm_lanai_globals_t,
						  sync_dma.descriptor[0]) -
				   (gm_u8_t *) is->lanai.sram);
	      descriptors[0] = gm_htopci_u32 (pointer);
	      GM_PRINT (GM_PRINT_LEVEL >= 4,
			("DMA descriptor at LANai address 0x%x.\n", pointer));
	      /* Tell the LANai where to find the DMA descriptor for
	         "channel 2" */

	      pointer = (gm_lp_t) ((gm_u8_t *) is->lanai.globals
				   + GM_OFFSETOF (gm_lanai_globals_t,
						  user_dma_descriptor[0])
				   - (gm_u8_t *) is->lanai.sram);
	      descriptors[2] = gm_htopci_u32 (pointer);
	      return GM_SUCCESS;
	    }
	  if (gm_is_l6 (is))
	    {
	      gm_lp_t *descriptors;
	      gm_lp_t pointer;

	      /* Tell the LANai where to find the DMA descriptor for
	         "channel 0" */

	      descriptors = (gm_lp_t *) ((gm_u8_t *) is->board_base
					 + 0x00800100);
	      pointer = (gm_lp_t) ((gm_u8_t *) is->lanai.globals
				   + GM_OFFSETOF (gm_lanai_globals_t,
						  sync_dma.descriptor[0]) -
				   (gm_u8_t *) is->lanai.sram);
	      descriptors[0] = gm_htopci_u32 (pointer);
	      GM_PRINT (GM_PRINT_LEVEL >= 4,
			("DMA descriptor at LANai address 0x%x.\n", pointer));

	      /* Tell the LANai where to find the DMA descriptor for
	         "channel 2" */

	      pointer = (gm_lp_t) ((gm_u8_t *) is->lanai.globals
				   + GM_OFFSETOF (gm_lanai_globals_t,
						  user_dma_descriptor[0])
				   - (gm_u8_t *) is->lanai.sram);
	      descriptors[2] = gm_htopci_u32 (pointer);
	      return GM_SUCCESS;
	    }
	  if (gm_is_l7 (is) || gm_is_l8 (is) || gm_is_l9 (is))
	    {
	      gm_lp_t *descriptors;
	      gm_lp_t pointer;

	      /* Tell the LANai where to find the DMA descriptor for
	         "channel 0" */

	      /* If you modify this code... ie to use more than just
		 DMA channel 0, then please make the corresponding
		 changes to the CSPI lanai 9 specific section earlier
	         in this function */
	      
	      /* The "magic number" in this assignment is the byte
	         offset of the bank of pointers for DMA and Doorbell
	         as described in the M2M-PCI64X specification: */

	      descriptors = (gm_lp_t *) ((gm_u8_t *) is->board_base
					 + 0x00800100);
	      pointer = (gm_lp_t) ((gm_u8_t *) is->lanai.globals
				   + GM_OFFSETOF (gm_lanai_globals_t,
						  sync_dma.descriptor[0]) -
				   (gm_u8_t *) is->lanai.sram);
	      descriptors[0] = gm_htopci_u32 (pointer);
	      GM_PRINT (GM_PRINT_LEVEL >= 4,
			("DMA descriptor at LANai address 0x%x.\n", pointer));

	      /* Tell the LANai where to find the DMA descriptor for
	         "channel 2" */

	      pointer = (gm_lp_t) ((gm_u8_t *) is->lanai.globals
				   + GM_OFFSETOF (gm_lanai_globals_t,
						  user_dma_descriptor[0])
				   - (gm_u8_t *) is->lanai.sram);
	      descriptors[2] = gm_htopci_u32 (pointer);
	      return GM_SUCCESS;
	    }
	}
      if ((GM_SUPPORT_PCI_REV_4 && (is->ifc.pci.config.Revision_ID == 4))
	  || (GM_SUPPORT_PCI_REV_5 && (is->ifc.pci.config.Revision_ID == 5)))
	{
	  if (gm_is_lX (is))
	    {
	      /* No DMA initialization is required */
	      return GM_SUCCESS;
	    }
	}
    }
#endif /* GM_SUPPORT_PCI */
  return GM_FAILURE;
}

static gm_status_t
gm_set_clock_val (gm_instance_state_t * is)
{
  GM_CALLED ();
  
#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI_REV_3 && is->ifc.pci.config.Revision_ID == 3)
    {
      unsigned int clockval = 0;

      /* use a 1x clockval initially */
      clockval = is->lanai.eeprom.lanai_clockval;
      clockval = clockval & (unsigned int) 0xFFFFFF8F;
      GM_PRINT (GM_DEBUG_BOARD_INIT,
		("gm_set_clock_val: PCI rev3 board - using a 1x clock"
		 " clockval = 0x%x\n", clockval));
      
      gm_write_lanai_special_reg_u32 (is, l7.CLOCK, clockval);
      goto done;
    }
#endif /* GM_SUPPORT_PCI */
  if (gm_is_l4 (is))
    {
      gm_write_lanai_special_reg_u32 (is, l4.clock_val,
				      is->lanai.eeprom.lanai_clockval);
      goto done;
    }

  if (gm_is_l5 (is))
    {
      gm_write_lanai_special_reg_u32 (is, l5.clock_val,
				      is->lanai.eeprom.lanai_clockval);
      goto done;
    }

  if (gm_is_l6 (is))
    {
      gm_write_lanai_special_reg_u32 (is, l6.CLOCK,
				      is->lanai.eeprom.lanai_clockval);
      goto done;
    }

  if (gm_is_l7 (is))
    {
      gm_write_lanai_special_reg_u32 (is, l7.CLOCK,
				      is->lanai.eeprom.lanai_clockval);
      goto done;
    }

  if (gm_is_l8 (is))
    {
      gm_write_lanai_special_reg_u32 (is, l8.CLOCK,
				      is->lanai.eeprom.lanai_clockval);
      goto done;
    }

  if (gm_is_l9 (is))
    {
      gm_write_lanai_special_reg_u32 (is, l9.CLOCK,
				      is->lanai.eeprom.lanai_clockval);
      goto done;
    }
  if (gm_is_lX (is))
    {
      /* The L10 ROM sets the clock in the LX. */
      goto done;
    }

  GM_RETURN_STATUS (GM_FAILURE);

done:
  gm_arch_spin (is, 100000);	/* seems to need long delay. */
  GM_RETURN_STATUS (GM_SUCCESS);
}


/* Stop the runaway phase lock loop circuitry in revision 2 boards.  This
   should be done after taking the board out of reset, but before reading
   the EEPROM. */

static gm_status_t
gm_stop_runaway_PLL (gm_instance_state_t * is)
{
  GM_CALLED ();
  
#if GM_SUPPORT_PCI
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Stopping runaway PLL, if needed.\n"));

  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  /* HACK: Must assume that the clock_val register is at the same
	     location as for a LANai 5. */
	  gm_write_lanai_special_reg_u32 (is, l5.clock_val, 0x00000080);
	  /* Pause for at least 10ms to allow LANai PLL to synchronize */
	  gm_arch_spin (is, 20000);
	  GM_RETURN_STATUS (GM_SUCCESS);
	}
      else if ((GM_SUPPORT_PCI_REV_4
		&& ((is->ifc.pci.config.Revision_ID == 4)))
	       || (GM_SUPPORT_PCI_REV_5
		   && (is->ifc.pci.config.Revision_ID == 5)))
	{
	  /* Nothing to do. */
	  GM_RETURN_STATUS (GM_SUCCESS);
	}
    }

  GM_WARN (("Could not stop runaway PLL.\n"));
  GM_RETURN_STATUS (GM_FAILURE);
#else /* !GM_SUPPORT_PCI */
  GM_RETURN_STATUS (GM_SUCCESS);
#endif /* !GM_SUPPORT_PCI */
}

static gm_status_t
gm_set_globals_ptr (gm_instance_state_t * is, gm_lp_t val)
{
  gm_write_first_sram_page (is, globals, val, lp);
  return GM_SUCCESS;
}

static gm_status_t
gm_set_lanai_scratch_register (gm_instance_state_t * is, gm_lp_t val)
{
  gm_write_first_sram_page (is, scratch, val, lp);
  return GM_SUCCESS;
}

static gm_status_t
gm_set_VERSION (gm_instance_state_t * is)
{
  if (gm_is_l4 (is))
    {
      switch (is->lanai.eeprom.lanai_cpu_version)
	{
	case 0x400:
	  gm_write_lanai_special_reg_u32 (is, l4.VERSION, 0);
	  goto done;
	case 0x401:
	case 0x402:
	case 0x403:
	  gm_write_lanai_special_reg_u32 (is, l4.VERSION, 3);
	  goto done;
	}
    }

  GM_NOTE (("Don't know VERSION value to use for LANai %x.%x\n",
	    is->lanai.eeprom.lanai_cpu_version >> 8,
	    is->lanai.eeprom.lanai_cpu_version & 0xff));
  return GM_FAILURE;

done:
  gm_arch_spin (is, 20000);
  return GM_SUCCESS;
}

static gm_status_t
gm_set_WINDOW (gm_instance_state_t * is)
{
  if (gm_is_l5 (is))
    {
      gm_write_lanai_special_reg_u32 (is, l5.WINDOW, 3);
      gm_arch_spin (is, 20000);
      return GM_SUCCESS;
    }

  GM_NOTE (("Don't know WINDOW value to use for LANai 0x%x.%x\n",
	    is->lanai.eeprom.lanai_cpu_version >> 8,
	    is->lanai.eeprom.lanai_cpu_version & 0xff));
  return GM_FAILURE;
}

/**********************************************************************
 * LANai register readout functions  (not performance critical)
 **********************************************************************/

static gm_status_t
gm_get_globals_ptr (gm_instance_state_t * is, gm_u32_t * val)
{
  *val = gm_read_first_sram_page (is, globals, lp);
  return GM_SUCCESS;
}

static gm_status_t
gm_get_lanai_scratch_register (gm_instance_state_t * is, gm_u32_t * val)
{
  *val = gm_read_first_sram_page (is, scratch, lp);
  return GM_SUCCESS;
}

static gm_status_t
gm_get_RTC (gm_instance_state_t * is, gm_s32_t * val)
{
  if (gm_is_l4 (is))
    {
      *val = gm_read_lanai_special_reg_s32 (is, l4.RTC);
      return GM_SUCCESS;
    }

  if (gm_is_l5 (is))
    {
      *val = gm_read_lanai_special_reg_s32 (is, l5.RTC);
      return GM_SUCCESS;
    }

  if (gm_is_l6 (is))
    {
      *val = gm_read_lanai_special_reg_s32 (is, l6.RTC);
      return GM_SUCCESS;
    }

  if (gm_is_l7 (is))
    {
      *val = gm_read_lanai_special_reg_s32 (is, l7.RTC);
      return GM_SUCCESS;
    }

  if (gm_is_l8 (is))
    {
      *val = gm_read_lanai_special_reg_s32 (is, l8.RTC);
      return GM_SUCCESS;
    }

  if (gm_is_l9 (is))
    {
      *val = gm_read_lanai_special_reg_s32 (is, l9.RTC);
      return GM_SUCCESS;
    }

  if (gm_is_lX (is))
    {
      *val = gm_read_lanai_special_reg_s32 (is, lX.RTC);
      return GM_SUCCESS;
    }

  return GM_FAILURE;
}

/**********************************************************************
 * LANai microcontroller functions
 **********************************************************************/

static gm_status_t
gm_disable_beats_l6_pci2 (gm_instance_state_t * is)
{
  int base_addr;

  /* Get the address of the status/control array. */

  if (__gm_cop_wakeup (is))
    goto l6_abort;
  if (__gm_cop_send (is, 0x1f))
    goto l6_abort;
  if (__gm_cop_send (is, 0x00))
    goto l6_abort;
  if ((base_addr = __gm_cop_receive (is)) < 0)
    goto l6_abort;
  if (__gm_cop_end (is))
    goto l6_abort;

  /* write 0x0c to DModeReg (offset 0x10) */

  if (__gm_cop_wakeup (is))
    goto l6_abort;
  if (__gm_cop_send (is, 0x20))
    goto l6_abort;
  if (__gm_cop_send (is, base_addr + 0x10))
    goto l6_abort;
  if (__gm_cop_send (is, 0x0c))
    goto l6_abort;
  if (__gm_cop_end (is))
    goto l6_abort;

  return GM_SUCCESS;

l6_abort:
  GM_WARN (("Could not disable beats.\n"));
  return GM_FAILURE;
}

static gm_status_t
gm_init_uc (gm_instance_state_t * is)
{
  GM_CALLED ();
  
  if (gm_is_l4 (is))
    {
      GM_RETURN_STATUS (GM_SUCCESS);
    }

  if (gm_is_l5 (is))
    {
      GM_RETURN_STATUS (GM_SUCCESS);
    }

  if (gm_is_l6 (is))
    {
#if GM_SUPPORT_PCI
      if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
	{
	  if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	      && ((is->ifc.pci.config.Revision_ID == 2) ||
		  (is->ifc.pci.config.Revision_ID == 3)))
	    {
	      GM_RETURN_STATUS (gm_disable_beats_l6_pci2 (is));
	    }
	}
#endif /* GM_SUPPORT_PCI */
    }

  if (gm_is_l7 (is) || gm_is_l8 (is) || gm_is_l9 (is) || gm_is_lX (is))
    {
      GM_RETURN_STATUS (GM_SUCCESS);
    }

  GM_RETURN_STATUS (GM_FAILURE);
}

/* If we have a cspi board, use the functions specific to that board
   instead of the PCIDMA chip based ones */
#if CSPI_MAP26xx
#include "cspi-26xx.c"
#else

/**********************************************************************
 * Control register access.
 **********************************************************************/

/****
 * LANai control register debugging
 ****/

#define GM_DEBUG_CONTROL_REG 0

static gm_status_t
gm_check_control_register_value (gm_instance_state_t * is)
{
#if 0				/* disabled by feldy --Glenn */
  GM_PARAMETER_MAY_BE_UNUSED (is);
  return GM_SUCCESS;
#else
  gm_u32_t crv;			/* control register value */
  GM_PARAMETER_MAY_BE_UNUSED (is);

  if (!is)
    {
      GM_NOTE (("internal error: is is NULL\n"));
      return GM_FAILURE;
    }

  if (GM_SUPPORT_SBUS && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      crv = gm_ntoh_u16 (*(gm_u16_n_t *) is->lanai.control_regs);
      if ((crv >> 15 & 1) == (crv >> 14 & 1))
	goto error;
      if ((crv >> 13 & 1) == (crv >> 12 & 1))
	goto error;
      return GM_SUCCESS;
    }

  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      crv = gm_ntoh_u32 (is->lanai.control_regs[0]);
      if (GM_SUPPORT_PCI_REV_1 && is->ifc.pci.config.Revision_ID == 1)
	{
	  if ((crv >> 31 & 1) == (crv >> 30 & 1))
	    goto error;
	  if ((crv >> 29 & 1) == (crv >> 28 & 1))
	    goto error;
	  return GM_SUCCESS;
	}

      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  return GM_SUCCESS;
	}

      if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	  || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
	{
	  crv = gm_ntoh_u32 (is->lanai.control_regs[0x10]);
	  GM_PRINT
	    (GM_DEBUG_CONTROL_REG, ("control register value = 0x%x\n", crv));
	  if ((crv >> GM_LX (31) & 1) == (crv >> GM_LX (30) & 1))
	    goto error;
	  if ((crv >> GM_LX (29) & 1) == (crv >> GM_LX (28) & 1))
	    goto error;
	  if ((crv >> GM_LX (27) & 1) == (crv >> GM_LX (26) & 1))
	    goto error;
	  if ((crv >> GM_LX (25) & 1) == (crv >> GM_LX (24) & 1))
	    goto error;
	  return GM_SUCCESS;
	}
    }

  return GM_FAILURE;

error:
  GM_NOTE (("Bogus control register value detected: 0x%x\n", crv));
  return GM_FAILURE;
#endif
}



/****
 * LANai reset control
 ****/

gm_status_t
gm_lanai_reset_off (gm_instance_state_t * is)
{
  if (gm_check_control_register_value (is) != GM_SUCCESS)
    return GM_FAILURE;

  if (GM_SUPPORT_SBUS && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      GM_PRINT (GM_PRINT_LEVEL >= 1, ("Turning off reset.\n"));
      GM_PRINT (GM_PRINT_LEVEL >= 1,
		("control_reg = 0x%04x before deasserting reset\n",
		 (int) *(gm_u16_t *) is->lanai.control_regs));
      gm_always_assert (!GM_NO_PARTWORD_PIO_WRITE);
      *(gm_u16_n_t *) is->lanai.control_regs = gm_hton_u16 (1 << 15);
      GM_PRINT (GM_PRINT_LEVEL >= 1,
		("control_reg = 0x%04x after deasserting reset.\n",
		 (int) *(gm_u16_t *) is->lanai.control_regs));
      GM_STBAR ();
      goto done;
    }

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if (GM_SUPPORT_PCI_REV_1 && is->ifc.pci.config.Revision_ID == 1)
	{
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 7);
	  GM_STBAR ();
	  goto done;
	}

      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 7);
	  GM_STBAR ();
	  goto done;
	}

      if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	  || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
	{
	  is->lanai.control_regs[0x10] = gm_hton_u32 (1 << GM_LX (30));
	  GM_STBAR ();
	  goto done;
	}
    }
#endif
  
  return GM_FAILURE;

done:
  /* Allow the LANai time to come out of reset before any of its
     registers is accessed. */
  gm_arch_spin (is, 20000);
  return gm_check_control_register_value (is);
}

gm_status_t
gm_lanai_reset_on (gm_instance_state_t * is)
{
  if (gm_check_control_register_value (is) != GM_SUCCESS)
    return GM_FAILURE;

  if (GM_SUPPORT_SBUS && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      gm_always_assert (!GM_NO_PARTWORD_PIO_WRITE);
      *(gm_u16_n_t *) is->lanai.control_regs = gm_hton_u16 (1 << 14);
      GM_STBAR ();
      goto done;
    }

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if (GM_SUPPORT_PCI_REV_1 && is->ifc.pci.config.Revision_ID == 1)
	{
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 6);
	  GM_STBAR ();
	  goto done;
	}
 
      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 6);
	  GM_STBAR ();
	  goto done;
	}

      if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	  || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
	{
	  is->lanai.control_regs[0x10] = gm_hton_u32 (1 << GM_LX (31));
	  GM_STBAR ();
	  goto done;
	}
    }
#endif /* GM_SUPPORT_PCI */

  return GM_FAILURE;

 done:
  return gm_check_control_register_value (is);
}

/****
 * Ebus reset control
 ****/

/* Return GM_SUCCESS if operation was performed, or operation is not
   applicable. */

gm_status_t
gm_ebus_reset_off (gm_instance_state_t * is)
{
  GM_CALLED ();
  
  if (gm_check_control_register_value (is) != GM_SUCCESS)
    GM_RETURN_STATUS (GM_FAILURE);

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 5);
	  GM_STBAR ();
	  goto done;
	}

      if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	  || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
	{
	  /* There is no ebus reset on the LX */
	  goto done;
	}
    }
#endif /*GM_SUPPORT_PCI*/
  
  GM_RETURN_STATUS (GM_FAILURE);

 done:
  GM_RETURN_STATUS (gm_check_control_register_value (is));
}

gm_status_t
gm_ebus_reset_on (gm_instance_state_t * is)
{
  if (gm_check_control_register_value (is) != GM_SUCCESS)
    return GM_FAILURE;

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 4);
	  GM_STBAR ();
	  goto done;
	}

      if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	  || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
	{
	  /* There is no ebus reset on the LX */
	  goto done;
	}
    }
#endif /* GM_SUPPORT_PCI */
  
  return GM_FAILURE;

 done:
  return gm_check_control_register_value (is);
}

/****
 * board reset control
 ****/

/* These functions must function correctly before the EEPROM is read,
   so they don't depend on any part of the EEPROM, except the "bus_type"
   field, which is set by the caller.

   Return GM_SUCCESS of operation was performed, or operation is not
   applicable. */

gm_status_t
gm_board_reset_off (gm_instance_state_t * is)
{
  GM_CALLED ();
  
  if (gm_check_control_register_value (is) != GM_SUCCESS)
    GM_RETURN_STATUS (GM_FAILURE);

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 3);
	  GM_STBAR ();
	  goto done;
	}
      else if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	       || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
	{
	  /* Turn off the board MEM reset, so the eeprom can be read. */
	  is->lanai.control_regs[0x10] = gm_hton_u32 (1 << GM_LX (28));
	  GM_STBAR ();
	  goto done;
	}
    }
#endif /*GM_SUPPORT_PCI*/
  
  GM_RETURN_STATUS (GM_FAILURE);

 done:
  GM_RETURN_STATUS (gm_check_control_register_value (is));
}

gm_status_t
gm_board_reset_on (gm_instance_state_t * is)
{
  if (gm_check_control_register_value (is) != GM_SUCCESS)
    return GM_FAILURE;

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 2);
	  GM_STBAR ();
	  goto done;
	}
      else if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	       || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
	{
	  /* Turn on the board MEM reset. */
	  is->lanai.control_regs[0x10] = gm_hton_u32 (1 << GM_LX (29));
	  GM_STBAR ();
	  goto done;
	}
    }
#endif /* GM_SUPPORT_PCI */
  
  return GM_FAILURE;

 done:
  return gm_check_control_register_value (is);
}


/* we've discovered some slots don't initialize correctly, force
   64-bit mode or 32-bit mode */

static gm_status_t
gm_board_force_64or32bit_mode (gm_instance_state_t * is, int which, int *readback)
{
  gm_u32_t bit_shift=0;

  GM_CALLED ();
  
  *readback = 0;

  if (gm_check_control_register_value (is) != GM_SUCCESS)
    return GM_FAILURE;

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  /* bit 07 force 64-bit mode */
	  /* bit 06 force 64-bit mode */
	  if (which==64)
	     bit_shift = 31;
	  if (which==32)
	     bit_shift = 30;
          if (!bit_shift) 
	    GM_RETURN_STATUS (GM_FAILURE);

	  GM_PRINT(1,("Forcing %d-bit PCI bus width\n",which));

	  is->lanai.control_regs[0]
	    	= gm_hton_u32 ((gm_u32_t) (((gm_u32_t) 1) << bit_shift));
	  GM_STBAR ();
	  if (gm_ntoh_u32 (is->lanai.control_regs[0])
	      & (gm_u32_t) (((gm_u32_t) 1) << bit_shift))
	    {
	      *readback = 1;
	    }
	  GM_RETURN_STATUS (GM_SUCCESS);
	}
    }
#endif /* GM_SUPPORT_PCI */
  GM_RETURN_STATUS (GM_FAILURE);
}

/* we've discovered some slots don't initialize correctly check it */
/* This function is exported for debugging use in arch-specific code */
gm_status_t
gm_read_correct_sequence_bit (gm_instance_state_t * is, int *correct)
{
  *correct = 0;

  GM_CALLED ();
  
  if (gm_check_control_register_value (is) != GM_SUCCESS)
    GM_RETURN_STATUS (GM_FAILURE);

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if ((is->ifc.pci.config.Revision_ID == 2) ||
	  (is->ifc.pci.config.Revision_ID == 3))
	{
	  /* bit 14 correct RST/REQ64 sequence */
	  if (gm_ntoh_u32 (is->lanai.control_regs[0]) & (1 << 22))
	    {
	      *correct = 1;
	      GM_RETURN_STATUS (GM_SUCCESS);
	    }
	  GM_RETURN_STATUS (GM_SUCCESS);
	}
    }
#endif /* GM_SUPPORT_PCI */
  GM_RETURN_STATUS (GM_FAILURE);
}

/****
 * Interrupt generation control
 ****/

gm_status_t
gm_initialize_interrupts (gm_instance_state_t * is)
{
  gm_status_t status;

  GM_CALLED ();

  if (gm_check_control_register_value (is) != GM_SUCCESS)
    {
      GM_WARN (("gm_check_control_register_value failed.\n"));
      return GM_FAILURE;
    }

  GM_PRINT (GM_PRINT_LEVEL >= 9, ("gm_initialize_interrupts called\n"));

  if (status = gm_arch_dma_region_alloc 
      (is, &is->interrupt_region,  sizeof (union gm_interrupt_descriptor),
       GM_ARCH_DMA_READ | GM_ARCH_DMA_CONSISTENT, NULL, is) != GM_SUCCESS)
    {
      GM_WARN (("Failed to allocate DMA region for host interrupt status\n"));
      return status;
    }
  
  is->interrupt = (union gm_interrupt_descriptor *)
    gm_arch_dma_region_kernel_addr (&is->interrupt_region);
  is->interrupt->type = gm_hton_u32 (GM_NO_INTERRUPT);
  is->interrupt->last_byte.ready = gm_hton_u8 (0);
  gm_write_lanai_global_dp (is, interrupt_dp, 
			    gm_arch_dma_region_dma_addr (&is->interrupt_region));
			    
  /* Mask any interrupt glitch during HW init. */
  is->disable_interrupts (is);
  GM_STBAR ();
  
  /* Make sure the host only ever gets host interrupts. */

  gm_arch_spin (is, 20000);

  status = GM_FAILURE;  /* fail by default */

  if (GM_SUPPORT_SBUS && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      gm_assert (!GM_NO_PARTWORD_PIO_WRITE);
      *(gm_u16_n_t *) is->lanai.control_regs = gm_hton_u16 (1 << 13);
      GM_STBAR ();
      is->pausible = 1;
      status = GM_SUCCESS;
    }

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if (GM_SUPPORT_PCI_REV_1 && is->ifc.pci.config.Revision_ID == 1)
	{
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 5);
	  GM_STBAR ();
	  is->pausible = 1;
	  status = GM_SUCCESS;
	}

      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  /* is->lanai.control_regs[0]=1<<25; */
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 1);
	  GM_STBAR ();
	  is->pausible = 1;
	  is->enable_interrupts (is);
	  status = GM_SUCCESS;
	}
      
      if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	  || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
	{
	  gm_enable_interrupts_lX (is);
	  is->pausible = 1;
	  status = GM_SUCCESS;
	}
    }
#endif /* GM_SUPPORT_PCI */

  if (status == GM_SUCCESS)
    {
      /* With the LANai out of reset, but before it can possibly
	 generate a legitimate interrupt, clear any interrupt left
	 over from HW reset. */

      if (is->interrupting (is))
	{
	  GM_NOTE (("claiming premature interrupt for board %d\n", is->id));
	  is->claim_interrupt (is);
	  GM_STBAR ();
	}

      /* Now that HW init is done, unmask interrupts. */

      is->enable_interrupts (is);
      GM_STBAR ();
  
      /* Allow the lanai to interrupt us */
      gm_write_lanai_global_u32 (is, interrupt.type, GM_NO_INTERRUPT);
      GM_STBAR ();
    }
  else
    {
      is->interrupt = NULL;
      gm_arch_dma_region_free (&is->interrupt_region);
    }
  
  GM_RETURN_STATUS (status);
}

gm_status_t
gm_finalize_interrupts (gm_instance_state_t * is)
{
  GM_CALLED ();

  if (gm_check_control_register_value (is) != GM_SUCCESS)
    return GM_FAILURE;
  gm_check_pci_config_space (is);
  
  gm_set_EIMR (is, 0);
  gm_check_pci_config_space (is);

  if (GM_SUPPORT_SBUS && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      gm_assert (!GM_NO_PARTWORD_PIO_WRITE);
      *(gm_u16_n_t *) is->lanai.control_regs = gm_hton_u16 (1 << 12);
      GM_STBAR ();
      is->pausible = 0;
      goto done;
    }

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if (GM_SUPPORT_PCI_REV_1 && is->ifc.pci.config.Revision_ID == 1)
	{
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 4);
	  GM_STBAR ();
	  is->pausible = 0;
	  goto done;
	}

      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  is->lanai.control_regs[0] = gm_hton_u32 (1 << 0);
	  GM_STBAR ();
	  is->pausible = 0;
	  goto done;
	}
      
      if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	  || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
	{
	  gm_disable_interrupts_lX (is);
	  goto done;
	}
    }
#endif /* GM_SUPPORT_PCI */

  GM_RETURN_STATUS (GM_FAILURE);


 done:

  /* Spin long enough for any interrupts in progress to complete. */
  
  gm_arch_spin (is, 20000);
  GM_STBAR ();
  
  /* Free the descriptor only after interrupts have been disabled;
     And make certain to only free the region once  */
  if (is->interrupt != NULL)
    {
      is->interrupt = NULL;
      gm_arch_dma_region_free (&is->interrupt_region);
    }
  GM_RETURN_STATUS (GM_SUCCESS);  
}

/****
 * Reading the control reg for printing/debugging purposes.  Note
 * that host byte order is used.
 ****/

static gm_u32_t
gm_get_control_reg (gm_instance_state_t * is)
{
  if (GM_SUPPORT_SBUS && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    return 0;

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if (GM_SUPPORT_PCI_REV_1 && is->ifc.pci.config.Revision_ID == 1)
	{
	  return gm_ntoh_u32 (is->lanai.control_regs[0]);
	}

      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  return gm_ntoh_u32 (is->lanai.control_regs[0]);
	}
    }
#endif /* GM_SUPPORT_PCI */
  
  return GM_FAILURE;
}

#endif /* !CSPI_MAP26xx */

/**********************************************************************/

static gm_status_t
gm_instance_init_functions (gm_instance_state_t * is)
{
  /* By default, initialize handlers to abort if called.  The function
     pointers will be replaced with appropriate functions as enough
     information becomes available to do so. */

  is->_set_EIMR = gm_oo_set_abort;
  is->set_ISR = gm_oo_set_abort;
  is->get_EIMR = gm_oo_get_abort;
  is->get_ISR = gm_oo_get_abort;
  is->enable_interrupts = gm_oo_able_abort;
  is->disable_interrupts = gm_oo_able_abort;
  is->interrupting = gm_oo_interrupting_abort;
  is->claim_interrupt = gm_oo_claim_interrupt_abort;
  
  return GM_SUCCESS;
}

#if GM_SUPPORT_PCI
/* Set flag if board and host are both capable of 64-bit DMA addressing.
 *
 * NOTE:  This is the CAPABILITY flag.  Whether the driver will actually
 *        USE 64-bit DMA addressing (aka A64 DMA) will be decided by the
 *        platform-specific code.
 * NOTE:  For the PCI case, this test is currently based on the fact that the
 *        board is L7 or greater, and that GM_SIZEOF_DP_T is 8.  The
 *        implication is that GM_SIZEOF_DP_T will represent the capability of
 *        the host.  Be sure either to preserve this association, or change
 *        the design.
 *     
 *        As of today, only IRIX hosts support 8-byte (also called A64) DMA.
 *                                                             - Max 00/03/21
 */
static gm_status_t
gm_set_64_bit_flag(gm_instance_state_t * is)
{
   if (GM_SUPPORT_PCI && (is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI))
   {
      if (is->ifc.pci.config.Revision_ID >= 2)
      {
         if (GM_SIZEOF_DP_T == 8)
	 {
            is->flags |= GM_INSTANCE_64BIT_DMA_ADDR_OK;
            GM_PRINT(1, ("Unit %d is 64-bit-DMA-addr-capable\n", is->id));
	 }
         else
            GM_WARN(("Unit %d is 64-bit-DMA-addr-capable, but "
                     "GM_SIZEOF_DP_T = %d\n", is->id, GM_SIZEOF_DP_T));
      }
      else
	 GM_PRINT(1, ("Using A32 DMA for unit %d\n", is->id));
   }
   return GM_SUCCESS;
}

#endif /* GM_SUPPORT_PCI */

/* Initialize the board-specific functions for the interface.  Returns
   GM_SUCCESS on success. This should be called as soon as the EEPROM
   contents have been copied into the LANai. */

static gm_status_t
gm_init_eeprom_dependent_functions (gm_instance_state_t * is)
{
  gm_assert (is);

  if (gm_is_l4 (is))
    {
      is->_set_EIMR = gm_set_EIMR_l4;
      is->set_ISR = gm_set_ISR_l4;
      is->get_EIMR = gm_get_EIMR_l4;
      is->get_ISR = gm_get_ISR_l4;
      is->enable_interrupts = gm_enable_interrupts_l4;
      is->disable_interrupts = gm_disable_interrupts_l4;
      is->interrupting = gm_interrupting_l4;
      is->claim_interrupt = gm_claim_interrupt_l4;
      return GM_SUCCESS;
    }

  if (gm_is_l5 (is))
    {
      is->_set_EIMR = gm_set_EIMR_l5;
      is->set_ISR = gm_set_ISR_l5;
      is->get_EIMR = gm_get_EIMR_l5;
      is->get_ISR = gm_get_ISR_l5;
      is->enable_interrupts = gm_enable_interrupts_l5;
      is->disable_interrupts = gm_disable_interrupts_l5;
      is->interrupting = gm_interrupting_l5;
      is->claim_interrupt = gm_claim_interrupt_l5;
      return GM_SUCCESS;
    }

  if (gm_is_l6 (is))
    {
      is->_set_EIMR = gm_set_EIMR_l6;
      is->set_ISR = gm_set_ISR_l6;
      is->get_EIMR = gm_get_EIMR_l6;
      is->get_ISR = gm_get_ISR_l6;
      is->enable_interrupts = gm_enable_interrupts_l6;
      is->disable_interrupts = gm_disable_interrupts_l6;
      is->interrupting = gm_interrupting_l6;
      is->claim_interrupt = gm_claim_interrupt_l6;
      return GM_SUCCESS;
    }

  if (gm_is_l7 (is))
    {
      is->_set_EIMR = gm_set_EIMR_l7;
      is->set_ISR = gm_set_ISR_l7;
      is->get_EIMR = gm_get_EIMR_l7;
      is->get_ISR = gm_get_ISR_l7;
      is->enable_interrupts = gm_enable_interrupts_l7;
      is->disable_interrupts = gm_disable_interrupts_l7;
      is->interrupting = gm_interrupting_l7;
      is->claim_interrupt = gm_claim_interrupt_l7;
      return GM_SUCCESS;
    }

  if (gm_is_l8 (is))
    {
      is->_set_EIMR = gm_set_EIMR_l8;
      is->set_ISR = gm_set_ISR_l8;
      is->get_EIMR = gm_get_EIMR_l8;
      is->get_ISR = gm_get_ISR_l8;
      is->enable_interrupts = gm_enable_interrupts_l8;
      is->disable_interrupts = gm_disable_interrupts_l8;
      is->interrupting = gm_interrupting_l8;
      is->claim_interrupt = gm_claim_interrupt_l8;
      return GM_SUCCESS;
    }

  if (gm_is_l9 (is))
    {
      is->_set_EIMR = gm_set_EIMR_l9;
      is->set_ISR = gm_set_ISR_l9;
      is->get_EIMR = gm_get_EIMR_l9;
      is->get_ISR = gm_get_ISR_l9;
      is->enable_interrupts = gm_enable_interrupts_l9;
      is->disable_interrupts = gm_disable_interrupts_l9;
      is->interrupting = gm_interrupting_l9;
      is->claim_interrupt = gm_claim_interrupt_l9;
      return GM_SUCCESS;
    }

  if (gm_is_lX (is))
    {
      is->_set_EIMR = gm_set_EIMR_lX;
      is->set_ISR = gm_set_ISR_lX;
      is->get_EIMR = gm_get_EIMR_lX;
      is->get_ISR = gm_get_ISR_lX;
      is->enable_interrupts = gm_enable_interrupts_lX;
      is->disable_interrupts = gm_disable_interrupts_lX;
      is->interrupting = gm_interrupting_lX;
      is->claim_interrupt = gm_claim_interrupt_lX;
      return GM_SUCCESS;
    }
  
  /* No match, so fail */

  GM_NOTE (("gm_init_eeprom_dependent_functions: lanai 0x%x not supported\n",
	    is->lanai.eeprom.lanai_cpu_version));

  return GM_FAILURE;
}

/**********************************************************************
 * Memory-mapped resource location interface
 **********************************************************************/

/* gm_resource_location (IS, TYPE, RL) stores the resouce location of
   the resouce of type TYPE for the instance IS at RL.

   The implementation is inefficient in order to allow a tabular code
   structure that obviously yields correct memory layouts described
   below. This slightly inefficient but easily maintained
   implementation was chosen to simplify code maintenance in the
   future when there are even more board layouts.

   The Revision 1 PCI board layout is as follows:

   +-------------------+
   | memory            | 512K (used if memory <= 512K)
   +-------------------+ (0x00080000)
   | eeprom            | 256K
   +-------------------+ (0x000c0000)
   | special registers | 128K
   +-------------------+ (0x000e0000)
   | control registers | 128K
   +-------------------+ (0x00100000)
   | memory 2          | 1M (used if memory > 512K)
   +-------------------+


   The Revision 2 PCI board layout is as follows:

   +-------------------+ (0)
   | memory            | 
   +-------------------+ (0x00800040)
   | control registers | 
   +-------------------+ (0x00804000)
   | special registers | 
   +-------------------+ (0x00A00000)
   | eeprom            | 
   +-------------------+


   Sbus layout is as follows: (EEPROM is not memory-mapped)

 Byte offset   What                    Length
 -----------   -------------           ------
  0x00000000   Control Space (FPGA)    256*1024 Bytes ( 256KBytes)
  0x00040000   LANai Special Regs      256*1024 Bytes ( 256KBytes)
  0x00080000   LANai SRAM Memory      1024*1024 Bytes (1024KBytes)
  0x00180000   End of Board Memory            TOTAL   (2048KBytes)



CSPI MAP-26xx
 Byte offset   What                    Length
 -----------   -------------           ------
  0x40000000   LANai SRAM Memory      8*1024*1024 Bytes (8MBytes)
  0x48000000   LANai Special Regs        256*1024 Bytes ( 256KBytes)
  0xf0000000   PPCLAN Board Regs              256 Bytes (  256Bytes)
  0xffffff00   EEPROM                         128 Bytes (  128Bytes)

  The Revision 4 PCI-X board layout is as follows:

   +-------------------+ (0)
   | memory            | (eeprom is at the top of memory)
   +-------------------+ (0x00800000)
   | control registers | 
   +-------------------+ (0x00800100)
   :                   :
   +-------------------+ (0x00C00000)
   | special registers | 
   +-------------------+ (0x00C01000)

   The 512KB eeprom is accessed through the ROM BAR PCI configuration reg.
  
 */

#define GM_DEBUG_RESOURCE_LOCATION 0

static gm_status_t
gm_resource_location (gm_instance_state_t * is, enum gm_resource_type type,
		      gm_resource_location_t * rl)
{
  gm_resource_location_t eeprom;
  gm_resource_location_t memory;
  gm_resource_location_t control;
  gm_resource_location_t specials;
  gm_status_t status;

  GM_CALLED_WITH_ARGS (("%p,%d,%p", is, type, rl));
  
  GM_PRINT
    (GM_DEBUG_RESOURCE_LOCATION, ("Looking for resource of type %d\n", type));

  gm_assert (rl);
#if !GM_ETHERBOOT
  if (!GM_PAGE_LEN)
    {
      GM_WARN (("GM_PAGE_LEN not set in call to "__GM_FUNCTION__"().\n"));
      GM_RETURN_STATUS (GM_INTERNAL_ERROR);
    }
#endif

  /* Compute all locations.  It's OK to build illegitimate location
     information if IS is not sufficiently initialized, because the
     caller should never actually request that such descriptors be
     returned to it. */

  if (GM_SUPPORT_SBUS && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      control.offset = 0x00000000;
      specials.offset = 0x00040000;
      memory.offset = 0x00080000;

      control.length = 256 * 1024;
      specials.length = 256 * 10324;
      memory.length = is->lanai.eeprom.lanai_sram_size;
      goto match;
    }
#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if (GM_SUPPORT_PCI_REV_1 && is->ifc.pci.config.Revision_ID == 1)
	{
	  memory.offset = 1024 * (0);
	  eeprom.offset = 1024 * (512);
	  specials.offset = 1024 * (512 + 256);
	  control.offset = 1024 * (512 + 256 + 128);
	  if (is->lanai.eeprom.lanai_sram_size >= 1024 * 512)
	    memory.offset = 1024 * (512 + 256 + 128 + 128);

	  memory.length = is->lanai.eeprom.lanai_sram_size;
	  eeprom.length = 256 * 1024;
	  specials.length = 128 * 1024;
	  control.length = 128 * 1024;
	  goto match;
	}

      if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  memory.offset = 0;
	  control.offset = 0x00800040;
	  specials.offset = 0x00804000;
	  eeprom.offset = 0x00A00000;

	  memory.length = is->lanai.eeprom.lanai_sram_size;
	  control.length = 0x4000;
	  specials.length = 1024 * 16;
	  eeprom.length = 1024 * 1024 * 4;
	  goto match;
	}

      if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	  || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
	{
	  memory.offset = 0;
	  control.offset = 0x800000;
	  specials.offset = 0x00C00000;
	  status
	    = (gm_arch_read_pci_config_32
	       (is,
		(GM_OFFSETOF
		 (gm_pci_config_t,
		  Expansion_ROM_Base_Address)),
		&eeprom.offset));
	  if (status != GM_SUCCESS)
	    {
	      GM_WARN (("Could not read location of expansion ROM\n"));
	      goto abort_with_nothing;
	    }

	  memory.length = is->lanai.eeprom.lanai_sram_size;
	  control.length = 0x100;
	  specials.length = 0x1000;
	  eeprom.length = sizeof (gm_myrinet_eeprom_t);
	  goto match;
	}
    }
#endif /* GM_SUPPORT_PCI */
  
#if CSPI_MAP26xx
  {
    /* should probably replace these with the defines from
       <asm/cspi.h> (linux) or boardCsr.h (vxWorks) */
    memory.offset   = 0x40000000;
    control.offset  = 0xf0000000;
    specials.offset = 0x48000000;
    eeprom.offset   = 0xffffff00;

    memory.length   = is->lanai.eeprom.lanai_sram_size;
    control.length  = 0x00004000;
    specials.length = 0x00004000;
    eeprom.length   = 0x00000100;
    goto match;
  }
#endif

  /* It is an error for the caller to ask for an unmatched type. */
  GM_RETURN_STATUS (GM_INTERNAL_ERROR);
  
 abort_with_nothing:
  GM_RETURN_STATUS (status);
  
match:

  /* Return the requested location in the user-supplied buffer. */

  switch (type)
    {
    case GM_MEMORY_REGION:
      *rl = memory;
      break;
    case GM_EEPROM_REGION:
      *rl = eeprom;
      break;
    case GM_SPECIAL_REGS_REGION:
      *rl = specials;
      break;
    case GM_CONTROL_REGS_REGION:
      *rl = control;
      break;
    }

  GM_PRINT
    (GM_DEBUG_RESOURCE_LOCATION,
     ("Found resource at 0x%x of length 0x%x.\n", rl->offset, rl->length));

  /* Check for bogus returns, which should never happen if the
     instance state passed to this function was sufficiently
     initialized by the caller. */

  if (rl->length == 0)
    {
      GM_WARN (("Zero-length resource detected.\n"));
      GM_RETURN_STATUS (GM_INTERNAL_ERROR);
    }
  
  GM_RETURN_STATUS (GM_SUCCESS);
}

/* Map part of the LANai PCI interface into system space.  This
   generic function is used below to map the SRAM, special registers,
   and control registers, among other things.

   Returns pointer to mapped region, or NULL on failure.
 */

#define GM_DEBUG_MAP_BOARD_REGION 0

static void *
gm_map_board_region (gm_instance_state_t * is, enum gm_resource_type type)
{
  gm_resource_location_t rl;
  void *ret;

  GM_CALLED ();
  
  if (CSPI_MAP26xx)
    is->board_base = 0;
  else
    gm_assert (is->board_base);
  
  /* Determine the location of the resource to map */
  if (gm_resource_location (is, type, &rl) != GM_SUCCESS)
    GM_RETURN_PTR (0);

#if CSPI_MAP26xx && GM_OS_LINUX
  /* For most systems, we map one large space once, and we use
     this function to differentiate the different regions in
     that one large space.  But unfortunately, CSPI boards have
     their LANAI chip soldered on, and have their resources
     scattered throughout the address space.  Therefore we have to
     map the individual resources here */
  /* Addendum, the eeprom area is pre-mapped to the
     same area as the physical address by the linux kernel so we
     don't do it here*/
  if (type != GM_EEPROM_REGION)
    {
      void *mapped_region;
      gm_arch_map_io_space( is, rl.offset, rl.length, &mapped_region );
      rl.offset = (gm_u32_t) mapped_region;
    }
#endif

  GM_PRINT (GM_DEBUG_MAP_BOARD_REGION,
	    ("Mapping board resource at 0x%x of length 0x%x.\n", rl.offset,
	     rl.length));

  /* Verify the board has been mapped. */

  if (!CSPI_MAP26xx && !is->board_base)
    {
      GM_WARN (("internal error: board base not set.\n"));
      GM_RETURN_PTR (0);
    }

  /* Compute the return value */

  ret = (void *) ((char *) is->board_base + rl.offset);

  /* Print debugging view of the control registers */
  
  if (GM_DEBUG_BOARD_INIT && type == GM_CONTROL_REGS_REGION)
    {
      GM_PRINT (1, ("\n"));
      gm_hex_dump (ret, GM_MIN (rl.length, 256));
    }
  
  /* Return pointer to the requested resource. */

  GM_RETURN_PTR (ret);
}

/* Unmap a region mapped above */

static void
gm_unmap_board_region (gm_instance_state_t * is, enum gm_resource_type type,
		       void *ptr)
{
  GM_PARAMETER_MAY_BE_UNUSED (is);
  GM_PARAMETER_MAY_BE_UNUSED (type);
  GM_PARAMETER_MAY_BE_UNUSED (ptr);

#if CSPI_MAP26xx && GM_OS_LINUX
{  gm_resource_location_t rl;

  if( type != GM_EEPROM_REGION && type != GM_CONTROL_REGS_REGION )
    {
      
      if (gm_resource_location (is, type, &rl) != GM_SUCCESS)
	return;
      gm_arch_unmap_io_space( is, rl.offset, rl.length, ptr );
    }
} 
#endif /* CSPI_MAP26xx */
}

#define GM_PRINT_EEPROM (GM_USER_gallatin == 0)

static void
gm_print_eeprom (gm_instance_state_t * is)
{
  gm_myrinet_eeprom_t *ee;
  char board_label[sizeof (ee->board_label) + 1];

  ee = &is->lanai.eeprom;

  /* Print EEPROM contents */

  GM_PRINT (GM_PRINT_EEPROM,
            ("Contents of Myrinet EEPROM:\n"));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     lanai_clockval = 0x%08x\n", ee->lanai_clockval));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     lanai_cpu_version = 0x%04x\n", ee->lanai_cpu_version));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     lanai_board_id = %02x:%02x:%02x:%02x:%02x:%02x\n",
  		 ee->lanai_board_id[0], ee->lanai_board_id[1],
  		 ee->lanai_board_id[2], ee->lanai_board_id[3],
  		 ee->lanai_board_id[4], ee->lanai_board_id[5]));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     lanai_sram_size = 0x%08x (%dK)\n",
  		 ee->lanai_sram_size,
  		 (ee->lanai_sram_size >> 10) & 0x003fffff));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     fpga_version = \"%s\"\n", ee->fpga_version));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     more_version = \"%s\"\n", ee->more_version));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     delay_line_value = 0x%04x\n", ee->delay_line_value));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     max_lanai_speed = 0x%04x\n", ee->max_lanai_speed));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     board_type = 0x%04x\n", ee->board_type));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     bus_type = 0x%04x\n", ee->bus_type));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     product_code = 0x%04x\n", ee->product_code));
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     serial_number = 0x%08x\n", ee->serial_number));

  gm_strncpy (board_label, (char *) ee->board_label,
	      sizeof (ee->board_label));
  board_label[sizeof (ee->board_label)] = 0;
  _GM_PRINT (GM_PRINT_EEPROM,
             ("     board_label = \"%s\"\n", board_label));
}

/* Convert the eeprom from network byte order to host byte order,
   and fix buggy PCI eeproms */

static void
gm_fixup_eeprom (gm_instance_state_t * is)
{
  gm_myrinet_eeprom_t *ee;
  gm_myrinet_eeprom_n_t *from;

  ee = &is->lanai.eeprom;
  from = (gm_myrinet_eeprom_n_t *) & is->lanai.eeprom;

  /* fix the byte ordering of the eeprom to be in host order. */

  ee->lanai_clockval = gm_ntoh_u32 (from->lanai_clockval);
  ee->lanai_cpu_version = gm_ntoh_u16 (from->lanai_cpu_version);
  ee->lanai_sram_size = gm_ntoh_u32 (from->lanai_sram_size);
  ee->board_type = gm_ntoh_u16 (from->board_type);
  ee->bus_type = gm_ntoh_u16 (from->bus_type);
  ee->product_code = gm_ntoh_u16 (from->product_code);
  ee->serial_number = gm_ntoh_u32 (from->serial_number);
  ee->max_lanai_speed = gm_ntoh_u16 (from->max_lanai_speed);

  /* Null terminate board label in old eeproms that don't have a board
     label. */

  if (!gm_isprint (ee->board_label[0]))
    {
      ee->board_label[0] = 0;
    }

  /* fix buggy eeproms */

  if (!ee->lanai_clockval)
    ee->lanai_clockval = 0x11371137;

  if (GM_ENABLE_PROTOTYPE_NIC_SUPPORT && gm_is_l5 (is))
    {
      /* HACK for playing with prototype LANai5 PCI interfaces. */
      GM_WARN (("Overriding EEPROM for EXPERIMENTAL reasons.\n"));
      ee->lanai_clockval = 0x04278280;	/* 30MHz, single clocked */
      ee->lanai_clockval = 0x04278290;	/* 45MHz, single clocked */
      ee->lanai_clockval = 0x04278291;	/* 45MHz, double clocked */
      ee->lanai_clockval = 0x07098691;	/* 45MHz, double clocked, custom */
      ee->lanai_clockval = 0x04278281;	/* 30MHz, double clocked */
      GM_PRINT (GM_PRINT_LEVEL >= 1,
		("Using clock value 0x%08x.\n", ee->lanai_clockval));
      ee->lanai_sram_size = 1024 * 1024;
    }
}

/* Convert a hex character [0-9a-fA-F] to a value from 0 to 15. */
#define GM_ATOX(c) \
((c)>='0'&&(c)<='9'?(c)-'0':(c)>='a'&&(c)<='f'?(c)-'a'+10:(c)-'A'+10)

#define GM_DEBUG_COPY_EEPROM 0

/* Function to copy the eeprom from any type of interface into the
   instance state */

gm_status_t
gm_copy_eeprom (gm_instance_state_t * is)
{
  enum gm_bus_type known_type;
  gm_myrinet_eeprom_t *ee;	/* mapped eeprom */

  GM_CALLED_EX (GM_DEBUG_COPY_EEPROM);
  
  known_type = (enum gm_bus_type) is->lanai.eeprom.bus_type;
  gm_assert (known_type != GM_MYRINET_BUS_UNKNOWN);

#if GM_SUPPORT_SBUS
  /* Call special code for sbus interfaces, since eeprom mapping is
     not supported for them */
  if (is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      if (gm_arch_sbus_copy_eeprom (is) != GM_SUCCESS)
	goto abort_with_known_type;
      goto copied_eeprom;
    }
#endif

#if CSPI_MAP26xx
  {
    unsigned long i;
    /* map, copy, unmap the eerpm */
    ee = (gm_myrinet_eeprom_t *) gm_map_board_region(is, GM_EEPROM_REGION);
    
    if (ee)
      {	  
	for (i = 0; i < sizeof (gm_myrinet_eeprom_t) / 4; i++)
	  ((gm_u32_t *) & is->lanai.eeprom)[i] = ((gm_u32_t *) ee)[i];
      }
    
    
    is->lanai.eeprom.bus_type = GM_MYRINET_BUS_MAP26xx;
    is->lanai.eeprom.board_type = GM_MYRINET_BOARD_TYPE_MAP26xx;
    gm_unmap_board_region( is, GM_EEPROM_REGION, ee );
  }
  goto copied_eeprom;
#endif
  
  /****************
   * PCI
   ****************/

#if GM_SUPPORT_PCI
  
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      GM_PRINT (GM_DEBUG_COPY_EEPROM, ("Found a PCI card.\n"));
      
      /****
       * Rev 4
       ****/
      
      if (is->ifc.pci.config.Revision_ID == 4
	  || is->ifc.pci.config.Revision_ID == 5)
	{
	  int found_mac_addr = 0;
	  int found_serial_number = 0;
	  unsigned int spec_offset;
	  
	  GM_PRINT (GM_DEBUG_COPY_EEPROM, ("Found a revision 4 card.\n"));

	  /* Build a minimal fake eeprom. */

	  gm_bzero (&is->lanai.eeprom, sizeof (is->lanai.eeprom));

	  /* Rev4 cards all have LX processors. */
	  
	  is->lanai.eeprom.lanai_cpu_version = 0xA00; /* LX */
	  
	  /* Compute the SRAM size of the board's DMA_CONFIG register,
	     which points to the PCI config space backing store at the top
	     of memory.  Below this is the 256 byte specification string,
	     and below that is usable memory. */
	  
	  spec_offset
	    = ((gm_read_lanai_special_reg_u32 (is, lX.DMA_CONFIG)
		& 0x0FFFFC00)
	       - 256		/* string specs */
	       );
	  _GM_PRINT
	    (GM_DEBUG_COPY_EEPROM, ("specs at ofset %x\n", spec_offset));
	  
	  gm_assert (GM_PAGE_LEN);
	  is->lanai.eeprom.lanai_sram_size
	    = GM_PAGE_ROUND_DOWN (u32, spec_offset);
	  _GM_PRINT
	    (GM_DEBUG_COPY_EEPROM,
	     ("sram_size = %x\n", is->lanai.eeprom.lanai_sram_size));
	  
	  /* Read card details from from the spec space. */
	  
	  {
	    char *c;
	    char specs[256];

	    /* Copy the specification strings into a buffer. */
	    
	    gm_bcopy (((char *) is->board_base + spec_offset), specs, 256);
	    
	    GM_PRINT (GM_DEBUG_COPY_EEPROM, ("EEPROM specs:\n"));

	    /* Process each specification string. */
	    
	    for (c = specs;
		 *c;
		 c += gm_strlen(c) + 1)
	      {
		_GM_PRINT (GM_DEBUG_COPY_EEPROM, ("%s\n", c));
		
		/* Parse MAC address. */
		
		if (!gm_strncmp (c, "MAC=", 4))
		  {
		    gm_u8_t *mac_addr;

		    mac_addr = is->lanai.eeprom.lanai_board_id;
		    gm_bzero (mac_addr, 6);
		    for (c+=4; *c; c++)
		      {
			switch (*c)
			  {
			  case ':':
			    mac_addr++;
			  default:
			    *mac_addr *= 0x10;
			    *mac_addr += GM_ATOX (*c);
			  }
		      }
		    
		    mac_addr = is->lanai.eeprom.lanai_board_id;
		    _GM_PRINT
		      (GM_DEBUG_COPY_EEPROM,
		       ("MAC=%x:%x:%x:%x:%x:%x\n",
			(int) mac_addr[0], (int) mac_addr[1],
			(int) mac_addr[2], (int) mac_addr[3],
			(int) mac_addr[4], (int) mac_addr[5]));
		    found_mac_addr = 1;
		  }

		/* Parse serial number. */
		
		else if (!gm_strncmp (c, "SN=", 3))
		  {
		    unsigned int sn;
		    
		    sn = 0;
		    for (c+=3; *c; c++)
		      {
			sn*=10;
			sn += (int) (*c - '0');
		      }
		    _GM_PRINT (GM_DEBUG_COPY_EEPROM, ("SN=%d\n", sn));
		    is->lanai.eeprom.serial_number = sn;
		    found_serial_number = 1;
		  }

		/* Parse version */

		else if (!strncmp (c, "PV=", 3))
		  {
		    gm_strncpy ((char *)(is->lanai.eeprom.fpga_version),
				c+3,
				sizeof(is->lanai.eeprom.fpga_version)
				+ sizeof(is->lanai.eeprom.more_version));
		    is->lanai.eeprom.more_version
		      [GM_NUM_ELEM (is->lanai.eeprom.more_version)-1]
		      = 0;
		    _GM_PRINT
		      (GM_DEBUG_COPY_EEPROM,
		       ("PV=%s\n", is->lanai.eeprom.fpga_version));
		  }

		else
		  {
		    _GM_PRINT
		      (GM_DEBUG_COPY_EEPROM,
		       ("Unrecognized spec: %s\n", c));
		  }
	      }

	    /* Make sure we found everything. */
	    
	    if (!found_mac_addr
		|| !found_serial_number)
	      {
		GM_NOTE (("Invalid PCI Revision 4 eeprom.\n"));
		GM_RETURN_STATUS (GM_UNSUPPORTED_DEVICE);
	      }
	  }
	  goto copied_eeprom;
	}
      
      /* map, copy, and upmap the eeprom. */
      GM_PRINT (GM_DEBUG_COPY_EEPROM, ("Mapping eeprom.\n"));
      ee = (gm_myrinet_eeprom_t *) gm_map_board_region (is, GM_EEPROM_REGION);
      if (ee)
	{
	  /****
	   * Rev 1
	   ****/

	  if (GM_SUPPORT_PCI_REV_1 && is->ifc.pci.config.Revision_ID == 1)
	    {
	      unsigned int ui;

	      GM_PRINT (GM_DEBUG_COPY_EEPROM, ("Reading PCI Rev 1 eeprom.\n"));

	      /* Copy eeprom as 32-bit ints */
	      for (ui = 0; ui < sizeof (gm_myrinet_eeprom_t) / 4; ui++)
		((gm_u32_t *) & is->lanai.eeprom)[ui] = ((gm_u32_t *) ee)[ui];
	    }

	  /****
	   * Revs 2, 3
	   ****/

	  if ((GM_SUPPORT_PCI_REV_2 || GM_SUPPORT_PCI_REV_3)
	      && ((is->ifc.pci.config.Revision_ID == 2) ||
		  (is->ifc.pci.config.Revision_ID == 3)))
	    {
	      unsigned int i, quicklogic = 1;
	      gm_u32_n_t *ee_word_ptr;

	      ee_word_ptr = (gm_u32_n_t *) ee;
	      for (i = 0; i < 16; i++)
		{
		  if ((gm_ntoh_u32 (ee_word_ptr[i]) >> 24 & 0xFF) != i)
		    {
		      quicklogic = 0;
		      break;
		    }
		}

	      if (!quicklogic)
		{
		  unsigned int ui;

		  GM_PRINT (GM_DEBUG_COPY_EEPROM,
			    ("Reading normal-format eeprom.\n"));

		  /* Copy eeprom as 32-bit ints */
		  for (ui = 0; ui < sizeof (gm_myrinet_eeprom_t) / 4; ui++)
		    {
		      ((gm_u32_t *) & is->lanai.eeprom)[ui]
			= ((gm_u32_t *) ee)[ui];
		    }
		}
	      else
		{
		  unsigned int ui;

		  GM_INFO (("Reading QuickLogic-format eeprom.\n"));

		  /* Copy eeprom in hacked-up quicklogic fashion (one byte
		     per word) */
		  for (ui = 0; ui < sizeof (gm_myrinet_eeprom_t); ui++)
		    {
		      GM_PRINT (GM_DEBUG_COPY_EEPROM,
				("raw word[%d] = 0x%08x\n", ui,
				 ((gm_u32_t *) ee)[ui]));
		      ((gm_u8_t *) & is->lanai.eeprom)[ui] =
			((gm_u8_t *) ee)[4 * ui + 3];
		    }
		  if ((is->lanai.eeprom.lanai_cpu_version < 0x0500) ||
		      (is->lanai.eeprom.lanai_cpu_version > 0x0FFF))
		    {
		      ee_word_ptr = (gm_u32_n_t *) & is->lanai.eeprom;
		      GM_WARN (("EEPROM is corrupt.\n"
				"    %x %x %x %x\n",
				gm_ntoh_u32 (ee_word_ptr[0]),
				gm_ntoh_u32 (ee_word_ptr[1]),
				gm_ntoh_u32 (ee_word_ptr[2]),
				gm_ntoh_u32 (ee_word_ptr[3])));
		      goto abort_with_mapping;
		    }
		}
	    }

	  gm_unmap_board_region (is, GM_EEPROM_REGION, ee);

	  gm_fixup_eeprom (is);
	  goto copied_eeprom;
	}
    }
#endif /* GM_ENABLE_PCI */

  GM_RETURN_STATUS (GM_FAILURE);

copied_eeprom:
  gm_print_eeprom (is);

  /* Handle interfaces that report their bus type incorrectly. (sigh) */
  is->lanai.eeprom.bus_type = known_type;

  /* Now that we have the contents of the interface EEPROM, initialize
     the dynamic functions for the instance, which depend on the
     interface type. */

  if (gm_init_eeprom_dependent_functions (is) != GM_SUCCESS)
    {
      GM_WARN (("Could not init dynamic instance functions.\n"));
      GM_RETURN_STATUS (GM_FAILURE);
    }
  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_mapping:
  gm_unmap_board_region (is, GM_EEPROM_REGION, ee);
#if GM_SUPPORT_SBUS
abort_with_known_type:
#endif
  is->lanai.eeprom.bus_type = known_type;	/* restore */
  GM_RETURN_STATUS (GM_FAILURE);
}

/************************************************************************
 * Board mapping/unmapping
 ************************************************************************/

/************
 * utility functions to assist mapping.
 ************/

#if GM_SUPPORT_PCI

static gm_status_t
gm_enable_PCI_config_command_bit (gm_instance_state_t * is, gm_u16_t value)
{
  gm_u16_t command;
  gm_status_t status;

  GM_CALLED ();
  
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Enabling PCI command bit 0x%x\n", value));
  status = gm_read_pci_config_command (is, &command);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not read PCI command register.\n"));
      GM_RETURN_STATUS (status);
    }

  command |= value;
  status = gm_write_pci_config_command (is, command);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not write PCI command register.\n"));
      GM_RETURN_STATUS (status);
    }

  /* Pause for at least 10ms */
  gm_arch_spin (is, 15000);

  status = gm_read_pci_config_command (is, &command);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not read back PCI command register.\n"));
      GM_RETURN_STATUS (status);
    }

  if ((command & value) != value)
    {
      GM_NOTE (
	       ("gm_enable_PCI_config_command_bit couldn't set bit = 0x%x\n",
		value));
    }

  GM_RETURN_STATUS (GM_SUCCESS);
}

static gm_status_t
gm_disable_PCI_config_command_bit (gm_instance_state_t * is, gm_u16_t value)
{
  gm_u16_t command;

  GM_CALLED ();
  
/* This function should be using gm_read_pci_config_command() and
   gm_write_pci_config_command()                  - Max 2000/0606  */
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Disable PCI command bit 0x%x\n", value));
  if (gm_arch_read_pci_config_16 (is, GM_OFFSETOF (gm_pci_config_t, Command),
				  &command) != GM_SUCCESS)
    {
      GM_NOTE (("Could not read PCI command register.\n"));
      GM_RETURN_STATUS (GM_FAILURE);
    }
  command &= ~(value);
  if (gm_arch_write_pci_config_16 (is, GM_OFFSETOF (gm_pci_config_t, Command),
				   command) != GM_SUCCESS)
    {
      GM_NOTE (("Could not write PCI command register.\n"));
      GM_RETURN_STATUS (GM_FAILURE);
    }
  /* Pause for at least 10ms */
  gm_arch_spin (is, 15000);
  GM_RETURN_STATUS (GM_SUCCESS);
}

/* clear a bit in the PCI config status register by writing it */

static gm_status_t
gm_clear_PCI_config_status_bit (gm_instance_state_t * is, gm_u16_t value)
{
  gm_status_t status;
  gm_u16_t pci_status;

  GM_CALLED ();
  
  status = gm_read_pci_config_status (is, &pci_status);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not read PCI status register.\n"));
      GM_RETURN_STATUS (status);
    }

  /* only clear the bit we are interested in */
  pci_status &= value;

  if (pci_status == 0)
    {
      /* bit of interest is not set, so we have nothing to do */
      GM_RETURN_STATUS (status);
    }

  status = gm_write_pci_config_status (is, pci_status);
  
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not write PCI status register.\n"));
    }
  
  GM_RETURN_STATUS (status);
}
  

#endif /* GM_SUPPORT_PCI */

/************
 * Utility function to perform tests on mappings.
 ************/

/* the SRAM test failed, but it is helpful to know if it was a PCI
   error or an SRAM failure.  LANai >=7 do parity checks that can
   determine this */

static void
gm_diagnose_sram_failure(gm_instance_state_t *is)
{
  gm_u16_t pci_status;
  gm_status_t read_status;
  gm_u32_t lanai_parity_bit;

  /* This value is defined in lanai{7,8,9}_def.h, but we can't
     include those files because they conflict with each other.
     Luckily it is the same value for all three. */
  lanai_parity_bit = (1<<15);

  if ( gm_is_l7 (is) || gm_is_l8 (is) || gm_is_l9 (is) )
    {
#if GM_SUPPORT_PCI
      read_status = gm_read_pci_config_status (is, &pci_status);
      if (read_status != GM_SUCCESS)
	{
	  GM_NOTE (("Could not read PCI status register.\n"));
	  return;
	}
      else if (pci_status & GM_PCI_STATUS_PERR)
	{
	  GM_NOTE (("PCI parity error detected.\n"));
	  return;
	}
      else
#endif
	  if (is->get_ISR (is) & lanai_parity_bit)
	{
	  GM_NOTE (("SRAM parity error detected: memory failure.\n"));
	  return;
	}
      else
	{
	  GM_NOTE (("Memory corruption detected. "
		    "Possibly host bridge or memory failure.\n"));
	  _GM_NOTE (("Check to make sure card and possibly riser "
		    "are properly seated.\n"));
	  return;
	}
    }
}

/* test a region of RAM to make sure it is readable, writable, and
   is comprised of unique memory locations. */

static gm_status_t
gm_test_and_clear_sram (gm_instance_state_t * is,
			volatile gm_u8_n_t * orig_ptr, unsigned int orig_len)
{
  unsigned int offset, len = orig_len / 4;
  volatile gm_u32_n_t *ptr = (gm_u32_n_t *) orig_ptr;
  gm_u32_t tmp;
  
  if (((gm_size_t) orig_ptr & 3) != 0)
    return GM_FAILURE;

  /* test for stuck bits... */

  /* set the even bits and clear the odd bits. */

  gm_always_assert (((gm_size_t) orig_ptr & 3) == 0);
  gm_always_assert ((orig_len & 3) == 0);
  for (offset = 0; offset < len; offset++)
    __gm_kton_u32 (is, &ptr[offset], 0xaaaaaaaa);
  for (offset = 0; offset < len; offset++)
    {
      tmp = __gm_ntok_u32 (is, &ptr[offset]);
      if (tmp != 0xaaaaaaaa)
	{
	  GM_NOTE (("SRAM test failed at 0x%08x.\n",
		    offset * sizeof (ptr[0])));
	  _GM_PRINT (GM_DEBUG_BOARD_INIT,
		     ("Got 0x%08x, re-try 0x%08x, expected 0x%08x.\n",
		      tmp, __gm_ntok_u32 (is, &ptr[offset]), 0xaaaaaaaa));
	  gm_diagnose_sram_failure (is);
	  return GM_FAILURE;
	}
    }

  /* ... and vice versa */

  for (offset = 0; offset < len; offset++)
    __gm_kton_u32 (is, &ptr[offset], 0x55555555);
  for (offset = 0; offset < len; offset++)
    {
      tmp = __gm_ntok_u32 (is, &ptr[offset]);
      if ( tmp != 0x55555555)
	{
	  GM_NOTE (("SRAM test failed at 0x%08x.\n",
		    offset * sizeof (ptr[0])));
	  _GM_PRINT (GM_DEBUG_BOARD_INIT,
		     ("Got 0x%08x, re-try 0x%08x, expected 0x%08x.\n",
		      tmp, __gm_ntok_u32(is, &ptr[offset]), 0x55555555));
	  gm_diagnose_sram_failure (is);
	  return GM_FAILURE;
	}
    }

#if 0
  /* all zeros */
  for (offset = 0; offset < len; offset++)
    __gm_kton_u32 (is, &ptr[offset], 0x00000000);
  for (offset = 0; offset < len; offset++)
    {
      tmp = __gm_ntok_u32 (is, &ptr[offset]);
      if (tmp != 0x00000000)
	{
	  GM_NOTE (("SRAM test failed at 0x%08x.\n",
		    offset * sizeof (ptr[0])));
	  _GM_NOTE (("Got 0x%08x, re-try 0x%08x, expected 0x%08x.\n", tmp,
		     __gm_ntok_u32 (is, &ptr[offset]), 0x00000000));
	  return GM_FAILURE;
	}
    }

  /* all ones */
  for (offset = 0; offset < len; offset++)
    __gm_kton_u32 (is, &ptr[offset], (unsigned int)0xFFFFFFFF);
  for (offset = 0; offset < len; offset++)
    {
      tmp = __gm_ntok_u32 (is, &ptr[offset]);
      if (tmp != (unsigned int)0xFFFFFFFF)
	{
	  GM_NOTE (("SRAM test failed at 0x%08x.\n",
		    offset * sizeof (ptr[0])));
	  _GM_NOTE (("Got 0x%08x, re-try 0x%08x, expected 0x%08x.\n", tmp,
		     __gm_ntok_u32 (is, &ptr[offset]), (unsigned int)0xFFFFFFFF));
	  return GM_FAILURE;
	}
    }

#endif 

  /* test for mirroring by writing unique identifiers into each 32-bit
     word. */

#if 0
#error Does not use PIO functions --Glenn
  /* align the bottom of the buffer on a 32-bit boundary. */
  offset = (unsigned long) ptr & 3;
  if (offset)
    {
      ptr = (volatile unsigned char *) ((char *) ptr - offset + 4);
      len = len + offset - 4;
    }

  /* align the top of the buffer on a 32-bit boundary. */
  offset = len & 3;
  if (offset)
    len -= offset;
#endif

  /* Write unique IDs */
  for (offset = 0; offset < len; offset++)
    __gm_kton_u32 (is, &ptr[offset], offset);

  /* Verify unique IDs */
  for (offset = 0; offset < len; offset++)
    {
      if (__gm_ntok_u32 (is, &ptr[offset]) != offset)
	{
	  GM_NOTE
	    (("SRAM test failed at 0x%08x read: 0x%08x expected: 0x%08x.\n",
	      offset * sizeof (ptr[0]), __gm_ntok_u32(is, &ptr[offset]),
	      offset ));
	  return GM_FAILURE;
	}
    }

  /* All tests passed.  Clear memory and return. */

  __gm_kton_bzero (is, orig_ptr, orig_len);
  return GM_SUCCESS;
}

static gm_status_t
gm_test_RMP (gm_instance_state_t * is)
{
  gm_u32_t val;

  /* Test writing and reading RMP */

  if (gm_set_lanai_scratch_register (is, 0xcafebab8) != GM_SUCCESS)
    {
      GM_WARN (("Could not set RMP to cafebabe.\n"));
      goto abort_with_nothing;
    }
  if (gm_get_lanai_scratch_register (is, &val) != GM_SUCCESS)
    {
      GM_WARN (("Could not read RMP (cafebab8).\n"));
      goto abort_with_nothing;
    }
  if (val != 0xcafebab8)
    {
      GM_WARN (("RMP = 0x%x (!= 0xcafebab8)\n", val));
      goto abort_with_nothing;
    }

  /* Clear RMP */

  if (gm_set_lanai_scratch_register (is, 0) != GM_SUCCESS)
    {
      GM_WARN (("Could not set RMP to 0.\n"));
      goto abort_with_nothing;
    }
  if (gm_get_lanai_scratch_register (is, &val) != GM_SUCCESS)
    {
      GM_WARN (("Could not read RMP (0).\n"));
      goto abort_with_nothing;
    }
  if (val != 0)
    {
      GM_WARN (("RMP = 0x%x (!= 0)\n", val));
      goto abort_with_nothing;
    }

  return GM_SUCCESS;

abort_with_nothing:
  return GM_FAILURE;
}

/* Put the interface into a known state - requires control regs to be
   mapped. */
static void
gm_reset_everything (gm_instance_state_t * is)
{
  GM_CALLED ();
  
  gm_lanai_reset_on (is);
  gm_ebus_reset_on (is);
  gm_board_reset_on (is);
  gm_arch_spin (is, 20000);

  GM_RETURN_NOTHING ();
}

/****************************************************************
 * Map the various parts of the LANai into system memory.
 ****************************************************************/

#define GM_DEBUG_MAP_IO_SPACES 0

#if GM_DEBUG_MAP_IO_SPACES
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/* Due to hardware variations, the steps needed to detect and map
   various boards varies significantly from board revision to board
   revision.  Therefore, we implement one version of
   gm_map_io_spaces() for each board revision. */

/****************
 * sbus
 ****************/

static gm_status_t
gm_map_io_spaces_sbus (gm_instance_state_t * is)
{
  GM_PRINT (GM_PRINT_LEVEL >= 4,
            ("Mapping LANai[%d] io spaces.\n", is->id));

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("mapping SBUS board.\n"));

  /* Map the entire board */

  is->board_span = 0x00180000;
  if (gm_arch_map_io_space (is, 0, is->board_span, &is->board_base)
      != GM_SUCCESS)
    goto abort_with_nothing;

  /* Map the control regs. */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Mapping control registers.\n"));
  is->lanai.control_regs
    = (gm_u32_n_t *) gm_map_board_region (is, GM_CONTROL_REGS_REGION);
  if (!is->lanai.control_regs)
    {
      GM_WARN (("Could not map control_regs\n"));
      goto abort_with_mapped_board;
    }
  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("Control regs at %p.\n", is->lanai.control_regs));
#define GM_PRINT_CONTROL_REG() GM_PRINT (GM_PRINT_LEVEL >= 4, ("is->lanai.control_regs[0] = 0x%04x.\n", ((gm_u16_t *)is->lanai.control_regs)[0]))
  GM_PRINT_CONTROL_REG ();

  /* Force the interface into reset, since the previously loaded
     driver may have left it in a messy state. */
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Resetting LANai.\n"));
  gm_reset_everything (is);

  GM_PRINT_CONTROL_REG ();

  /* Copy the EEPROM from the LANai, mapping it temporarily to do so.
     This must be done before mapping the SRAM, since the SRAM region
     location depends on the contents of the EEPROM.

     Also, gm_copy_eeprom() intializes all the eeprom-dependent
     instance function pointers, so the EEPROM should be copied before
     any of these functions is invoked (other than
     is->board_reset_off() above. */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Copying board EEPROM\n"));
  if (gm_copy_eeprom (is) != GM_SUCCESS)
    {
      GM_WARN (("Could not copy EEPROM.\n"));
      goto abort_with_control_regs;
    }

  /* We don't set the delay_line value, so GM won't work on the very
     very earliest LANai4 boards.  If we set it, this would be the
     place. */

  /* Map the special registers */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("  Mapping special registers.\n"));
  is->lanai.special_regs
    = ((gm_lanai_special_registers_t *)
       gm_map_board_region (is, GM_SPECIAL_REGS_REGION));
  if (!is->lanai.special_regs)
    {
      GM_WARN (("Could not map special_regs\n"));
      goto abort_with_eeprom;
    }

  /* After mapping the special regs, set the clock value */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Setting LANai clock_val to 0x%08x.\n",
				  is->lanai.eeprom.lanai_clockval));
  if (gm_set_clock_val (is) != GM_SUCCESS)
    goto abort_with_special_regs;

  /* Map and clear the SRAM. */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("  Mapping SRAM (0x%x bytes).\n",
				  is->lanai.eeprom.lanai_sram_size));
  is->lanai.sram = gm_map_board_region (is, GM_MEMORY_REGION);
  if (is->lanai.sram == NULL)
    {
      GM_WARN (("Could not map SRAM\n"));
      goto abort_with_special_regs;
    }

  return GM_SUCCESS;

abort_with_special_regs:
  gm_unmap_board_region (is, GM_SPECIAL_REGS_REGION, is->lanai.special_regs);
abort_with_eeprom:
abort_with_control_regs:
  gm_unmap_board_region (is, GM_CONTROL_REGS_REGION,
			 (gm_u32_t *) is->lanai.control_regs);
abort_with_mapped_board:
abort_with_nothing:
  return GM_FAILURE;
}

static void
gm_unmap_io_spaces_sbus (gm_instance_state_t * is)
{
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Unmapping IO spaces.\n"));
  gm_unmap_board_region (is, GM_MEMORY_REGION, (char *) is->lanai.sram);
  gm_unmap_board_region (is, GM_SPECIAL_REGS_REGION, is->lanai.special_regs);
  gm_unmap_board_region (is, GM_CONTROL_REGS_REGION,
			 (gm_u32_t *) is->lanai.control_regs);
  gm_arch_unmap_io_space (is, 0, is->board_span, &is->board_base);
}

/****************
 * map26xx
 ****************/

static gm_status_t
gm_map_io_spaces_map26xx (gm_instance_state_t * is)
{
  GM_INFO (("Mapping CSPI MAP 26xx board.\n"));

  /* Determine the board span */

  /* CSPI 26xx dont use pci 'boards' so set board_base to 0 */
  is->board_base = 0;
  
  /* Map the control regs first, since we must take the board out of reset
     before accessing the EEPROM to prevent bus errors. */

  GM_INFO (("Mapping control registers.\n"));
  is->lanai.control_regs
    = (gm_u32_n_t *) gm_map_board_region (is, GM_CONTROL_REGS_REGION);
  if (!is->lanai.control_regs)
    {
      GM_WARN (("Could not map control_regs\n"));
      goto abort_with_mapped_board;
    }
  GM_INFO (("Mapped control registers at %p.\n",
	    is->lanai.control_regs));
  
  /* Force the interface into reset, since the previously loaded
     driver may have left it in a messy state. */
  gm_reset_everything (is);

  /* Copy the EEPROM from the LANai, mapping it temporarily to do so.
     This must be done before mapping the SRAM, since the SRAM region
     location depends on the contents of the EEPROM.

     Also, gm_copy_eeprom() intializes all the eeprom-dependent
     instance function pointers, so the EEPROM should be copied before
     any of these functions is invoked (other than
     is->board_reset_off() above. */

  if (gm_copy_eeprom (is) != GM_SUCCESS)
    {
      GM_WARN (("Could not copy EEPROM.\n"));
      goto abort_with_control_regs;
    }

  /* We don't set the delay_line value, so GM won't work on the very
     very earliest LANai4 boards.  If we set it, this would be the
     place. */

  /* Map the special registers */

  GM_INFO (("Mapping special registers.\n"));
  is->lanai.special_regs
    = ((gm_lanai_special_registers_t *)
       gm_map_board_region (is, GM_SPECIAL_REGS_REGION));
  if (!is->lanai.special_regs)
    {
      GM_WARN (("Could not map special_regs\n"));
      goto abort_with_eeprom;
    }
  GM_INFO (("Mapped special registers at %p.\n",
	    is->lanai.special_regs));
  
  /* After mapping the special regs, set the clock value */

  GM_INFO (("Setting LANai clock_val to 0x%08x.\n",
	    is->lanai.eeprom.lanai_clockval));
  if (gm_set_clock_val (is) != GM_SUCCESS)
    goto abort_with_special_regs;
  
  (void) gm_get_control_reg;
  GM_INFO (("control reg value is 0x%x.\n",
	    gm_get_control_reg (is)));

#define SAB
#ifdef SAB
  /* Now that we have mapped the special regs, but before we access
     any memory accessed through the LANai EBUS, we need to take the
     ebus out of reset. We also make sure the LANai is in reset before
     we clear its memory below.  */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Taking EBUS out of reset. (0x%08x)\n",
				  gm_get_control_reg (is)));
  if (gm_ebus_reset_off (is) != GM_SUCCESS)
    {
      GM_NOTE (("Could not take EBUS out of reset.\n"));
      goto abort_with_eeprom;
    }

  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("control reg value is 0x%x.\n", gm_get_control_reg (is)));
#endif /* SAB */
#undef SAB

  /* Map and clear the SRAM. */

  GM_INFO (("Mapping SRAM (0x%x bytes).\n",
	    is->lanai.eeprom.lanai_sram_size));
  is->lanai.sram = gm_map_board_region (is, GM_MEMORY_REGION);
  if (is->lanai.sram == NULL)
    {
      GM_WARN (("Could not map SRAM\n"));
      goto abort_with_special_regs;
    }
  GM_INFO (("Mapped SRAM at %p.\n", is->lanai.sram));
  GM_INFO (("Clearing LANai SRAM from %p to %p.\n",
	    is->lanai.sram,
	    (char *) is->lanai.sram + is->lanai.eeprom.lanai_sram_size));
  __gm_kton_bzero (is, is->lanai.sram, is->lanai.eeprom.lanai_sram_size);
  GM_INFO (("Cleared LANai[%d] SRAM\n", is->id));

  return GM_SUCCESS;

abort_with_special_regs:
  gm_unmap_board_region (is, GM_SPECIAL_REGS_REGION, is->lanai.special_regs);
abort_with_eeprom:
abort_with_control_regs:
  gm_unmap_board_region (is, GM_CONTROL_REGS_REGION,
			 (void *) is->lanai.control_regs);
abort_with_mapped_board:
  gm_arch_unmap_io_space (is, 0, is->board_span, &is->board_base);
  return GM_FAILURE;
}

static void
gm_unmap_io_spaces_map26xx (gm_instance_state_t * is)
{
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Unmapping IO spaces (CSPI MAP26xx).\n"));
  gm_unmap_board_region (is, GM_MEMORY_REGION, (void *) is->lanai.sram);
  gm_unmap_board_region (is, GM_SPECIAL_REGS_REGION,
			 (void *) is->lanai.special_regs);
  gm_unmap_board_region (is, GM_CONTROL_REGS_REGION,
			 (void *) is->lanai.control_regs);
  gm_arch_unmap_io_space (is, 0, is->board_span, &is->board_base);
}

#if GM_SUPPORT_PCI

/****************
 * PCI
 ****************/

/* Determine the PCI board span */

#define GM_DEBUG_PCI_BOARD_SPAN 0

gm_u32_t
gm_pci_board_span (gm_instance_state_t * is)
{
  gm_u32_t bits;
  gm_u32_t ret;

  GM_CALLED ();
  
  if (is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      gm_u32_t save;

      if (gm_arch_read_pci_config_32
	  (is, GM_OFFSETOF (gm_pci_config_t,
			    Base_Addresses_Registers[0]), &save)
	  != GM_SUCCESS)
	{
	  GM_RETURN (0);
	}
      GM_PRINT
	(GM_DEBUG, ("LANai PCI offset is 0x%x.\n", save & 0xfffffff0UL));
      if (gm_arch_write_pci_config_32
	  (is, GM_OFFSETOF (gm_pci_config_t, Base_Addresses_Registers[0]),
	   0xffffffff) != GM_SUCCESS)
	{
	  GM_RETURN (0);
	}
      if (gm_arch_read_pci_config_32
	  (is, GM_OFFSETOF (gm_pci_config_t,
			    Base_Addresses_Registers[0]), &bits)
	  != GM_SUCCESS)
	{
	  GM_RETURN (0);
	}
      if (gm_arch_write_pci_config_32
	  (is, GM_OFFSETOF (gm_pci_config_t,
			    Base_Addresses_Registers[0]), save) != GM_SUCCESS)
	{
	  GM_RETURN (0);
	}
      
      ret = ~(bits & 0xfffffff0) + 1;
      GM_PRINT (GM_DEBUG_PCI_BOARD_SPAN, ("PCI board span = 0x%x\n", ret));
      GM_RETURN (ret);
    }

  GM_RETURN (0);
}

static gm_status_t
gm_map_io_spaces_pci_rev_1 (gm_instance_state_t * is)
{
  GM_CALLED ();
  
  GM_PRINT (GM_PRINT_LEVEL >= 4,
            ("Mapping LANai[%d] io spaces.\n", is->id));

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("mapping PCI Rev 1 board.\n"));

  /* Determine the board span */

  is->board_span = gm_pci_board_span (is);
  if (!is->board_span)
    {
      GM_NOTE (("Could not determine board span.\n"));
      goto abort_with_nothing;
    }

  /* Map the entire board */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("  Mapping entire board.\n"));
  if (gm_arch_map_io_space (is, 0, is->board_span, &is->board_base)
      != GM_SUCCESS)
    {
      GM_NOTE (("Could not map entire LANai interface into kernel.\n"));
      goto abort_with_nothing;
    }

  /* Map the control regs first, since we must take the board out of reset
     before accessing the EEPROM to prevent bus errors. */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("  Mapping control registers.\n"));
  is->lanai.control_regs
    = (gm_u32_n_t *) gm_map_board_region (is, GM_CONTROL_REGS_REGION);
  if (!is->lanai.control_regs)
    {
      GM_NOTE (("Could not map control_regs\n"));
      goto abort_with_mapped_board;
    }

  /* Force the interface into reset, since the previously loaded
     driver may have left it in a messy state. */
  gm_reset_everything (is);

  /* Copy the EEPROM from the LANai, mapping it temporarily to do so.
     This must be done before mapping the SRAM, since the SRAM region
     location depends on the contents of the EEPROM.

     Also, gm_copy_eeprom() intializes all the eeprom-dependent
     instance function pointers, so the EEPROM should be copied before
     any of these functions is invoked (other than
     is->board_reset_off() above. */

  if (gm_copy_eeprom (is) != GM_SUCCESS)
    {
      GM_NOTE (("Could not copy EEPROM.\n"));
      goto abort_with_control_regs;
    }

  /* We don't set the delay_line value, so GM won't work on the very
     very earliest LANai4 boards.  If we set it, this would be the
     place. */

  /* Map the special registers */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("  Mapping special registers.\n"));
  is->lanai.special_regs
    = ((gm_lanai_special_registers_t *)
       gm_map_board_region (is, GM_SPECIAL_REGS_REGION));
  if (!is->lanai.special_regs)
    {
      GM_NOTE (("Could not map special_regs\n"));
      goto abort_with_eeprom;
    }

  /* After mapping the special regs, set the clock value */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Setting LANai clock_val to 0x%08x.\n",
				  is->lanai.eeprom.lanai_clockval));
  if (gm_set_clock_val (is) != GM_SUCCESS)
    {
      GM_NOTE (("Could not set lanai clock\n"));
      goto abort_with_special_regs;
    }

  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("control reg value is 0x%x.\n", gm_get_control_reg (is)));

  /* Map and clear the SRAM. */

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("  Mapping SRAM (0x%x bytes).\n",
				  is->lanai.eeprom.lanai_sram_size));
  is->lanai.sram = gm_map_board_region (is, GM_MEMORY_REGION);
  if (is->lanai.sram == NULL)
    {
      GM_WARN (("Could not map SRAM\n"));
      goto abort_with_special_regs;
    }

  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Clearing LANai SRAM from %p to %p.\n",
				  is->lanai.sram,
				  (char *) is->lanai.sram +
				  is->lanai.eeprom.lanai_sram_size));
  __gm_kton_bzero (is, is->lanai.sram, is->lanai.eeprom.lanai_sram_size);
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Cleared LANai[%d] SRAM\n", is->id));

  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_special_regs:
  gm_unmap_board_region (is, GM_SPECIAL_REGS_REGION, is->lanai.special_regs);
abort_with_eeprom:
abort_with_control_regs:
  gm_unmap_board_region (is, GM_CONTROL_REGS_REGION,
			 (void *) is->lanai.control_regs);
abort_with_mapped_board:
  gm_arch_unmap_io_space (is, 0, is->board_span, &is->board_base);
abort_with_nothing:
  GM_RETURN_STATUS (GM_FAILURE);
}

static void
gm_unmap_io_spaces_pci_rev_1 (gm_instance_state_t * is)
{
  GM_PRINT (GM_PRINT_LEVEL >= 4, ("Unmapping IO spaces.\n"));
  gm_unmap_board_region (is, GM_MEMORY_REGION, (void *) is->lanai.sram);
  gm_unmap_board_region (is, GM_SPECIAL_REGS_REGION,
			 (void *) is->lanai.special_regs);
  gm_unmap_board_region (is, GM_CONTROL_REGS_REGION,
			 (void *) is->lanai.control_regs);
  gm_arch_unmap_io_space (is, 0, is->board_span, &is->board_base);
}

#define GM_DEBUG_PCI_CONFIG_SPACE 0

/* Utility function to help debug config space backing store
   corruption. */

void
gm_check_pci_config_space (gm_instance_state_t *is)
{
  gm_u32_t old, new = 0;
  
  if (GM_DEBUG_PCI_CONFIG_SPACE && gm_is_lX (is))
    {
      gm_u32_t config_space[128/4];
      int i;
      
      /* Determine the initial location of the config space, if needed */

      if (is->lanai.special_regs)
	{
	  new = gm_read_lanai_special_reg_u32 (is, lX.DMA_CONFIG)
	    & 0x0ffffc00;
	}
      old = is->lx_pci_config_sram_backing_offset;
      if (!old)
	{
	  old = new;
	}

      /* Warn if the config space moved */
      
      if (old != new)
	{
	  GM_WARN (("DMA_CONFIG moved for 0x%x to 0x%x\n", old, new));
	}

      /* Dump the config space. */
	
      GM_PRINT (1, ("DMA_CONFIG = 0x%x (masked)\n", old));
      gm_bzero (config_space, sizeof (config_space));
      for (i=0; i<GM_NUM_ELEM(config_space); i++)
	{
	  gm_status_t status;
	  
	  status
	    = gm_arch_read_pci_config_32 (is, 
					  i * sizeof (config_space[i]),
					  &config_space[i]);
	  if (status != GM_SUCCESS)
	    {
	      GM_WARN (("Error reading config space in debug code.\n"));
	      break;
	    }
	}
      _gm_hex_dump ((char *) config_space, sizeof (config_space));
    }
}

static gm_status_t
gm_map_io_spaces_pci_rev_2 (gm_instance_state_t * is)
{
  GM_CALLED ();
  
  GM_PRINT (GM_DEBUG_MAP_IO_SPACES,
            ("Mapping LANai[%d] io spaces.\n", is->id));

  GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("mapping PCI Rev 2 (or similar) board.\n"));

  /* Determine the board span */

  is->board_span = gm_pci_board_span (is);
  if (!is->board_span)
    {
      GM_NOTE (("Could not determine board span.\n"));
      goto abort_with_nothing;
    }

  /* Map the entire board */

  GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("  Mapping entire board.\n"));
  if (gm_arch_map_io_space (is, 0, is->board_span, &is->board_base)
      != GM_SUCCESS)
    {
      GM_NOTE (("pci_rev2: Could NOT map board into kernel (span = 0x%x)\n",
		is->board_span));
      goto abort_with_nothing;
    }

  /* Disable busmaster DMA until the board is taken out of ebus
     reset because these boards can cause bus glitches */

  if (gm_disable_PCI_config_command_bit (is, GM_PCI_COMMAND_MASTER) !=
      GM_SUCCESS)
    {
      GM_NOTE (("Could not disable PCI bus master DMA.\n"));
      goto abort_with_mapped_board;
    }

  /* Make sure that PERR is cleared. */

  if (gm_clear_PCI_config_status_bit (is, GM_PCI_STATUS_PERR) !=
      GM_SUCCESS)
    {
      GM_NOTE (("Could not clear PERR bit in PCI config space.\n"));
      /* Some machines don't let you set bits in the status register,
	 so we don't treat this as a fatal error.		     */
    }

  /* Map the control regs first, since we must take the board out of reset
     before accessing the EEPROM to prevent bus errors. */

  GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("  Mapping control registers.\n"));
  is->lanai.control_regs
    = (gm_u32_n_t *) gm_map_board_region (is, GM_CONTROL_REGS_REGION);
  if (!is->lanai.control_regs)
    {
      GM_WARN (("Could not map control_regs\n"));
      goto abort_with_mapped_board;
    }

  /* Force the interface into reset, since the previously loaded
     driver may have left it in a messy state. */
  gm_reset_everything (is);

  /* Take the board out of reset, so we can get at the eeprom. */

  GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("Taking board out of reset.\n"));
  if (gm_board_reset_off (is) != GM_SUCCESS)
    {
      GM_NOTE (("Could not take interface out of reset.\n"));
      goto abort_with_control_regs;
    }

  /* Map the special regs before reading the EEPROM so we can stop the
     runaway PLL (via the clock_val register) before reading the
     eeprom */

  GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("  Mapping special registers.\n"));
  is->lanai.special_regs
    = ((gm_lanai_special_registers_t *)
       gm_map_board_region (is, GM_SPECIAL_REGS_REGION));
  if (!is->lanai.special_regs)
    {
      GM_WARN (("Could not map special_regs\n"));
      goto abort_with_board_reset_off;
    }

  /* Make sure the PLL on the board is not running away, draining too much
     power to read the EEPROM. */

  if (gm_stop_runaway_PLL (is) != GM_SUCCESS)
    {
      GM_WARN (("Could not stop runaway PLL\n"));
      goto abort_with_special_regs;
    }

  gm_check_pci_config_space (is);
  
  /* Copy the EEPROM from the LANai, mapping it temporarily to do so.
     This must be done before mapping the SRAM, since the SRAM region
     location depends on the contents of the EEPROM.

     Also, gm_copy_eeprom() intializes all the eeprom-dependent
     instance function pointers, so the EEPROM should be copied before
     any of these functions is invoked (other than
     is->board_reset_off() above. */

  if (gm_copy_eeprom (is) != GM_SUCCESS)
    {
      GM_WARN (("Could not copy EEPROM.\n"));
      goto abort_with_special_regs;
    }

  /* Set the clock value */
  if (gm_set_clock_val (is) != GM_SUCCESS)
    {
      GM_WARN (("Could not set clock value.\n"));
      goto abort_with_eeprom;
    }
  
  /* Reenable PCI bus master DMA */

  GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("Reenabling PCI bus master DMA.\n"));
  if (gm_enable_PCI_config_command_bit (is, GM_PCI_COMMAND_MASTER) !=
      GM_SUCCESS)
    goto abort_with_eeprom;

  if (is->ifc.pci.config.Revision_ID == 2
      || is->ifc.pci.config.Revision_ID == 3)
    {
      int mode64 = 0;
      gm_read_correct_sequence_bit (is, &mode64);
      if (mode64)
	{
	  GM_PRINT (GM_DEBUG_BOARD_INIT,
		    ("64-bit slot initialized correctly.\n"));
	  is->bus_width = 64;

	  if (GM_ENABLE_FORCE_32BIT_MODE)
	    {
	      int readback = 0;
	      gm_board_force_64or32bit_mode (is, 32, &readback);
	      if (readback)
		{
		  GM_INFO (("PCI slot FORCED to 32-bit mode\n"));
		}
	      else
		{
		  GM_INFO (("unable to FORCE PCI slot to 32-bit mode"
			    " - 64-bit mode enabled\n"));
		}
	    }
	}
      else
	{
	  GM_INFO (("32-bit slot detected.\n"));
	  is->bus_width = 32;

	  if (GM_ENABLE_FORCE_64BIT_MODE)
	    {
	      int readback = 0;
	      gm_board_force_64or32bit_mode (is, 64, &readback);
	      if (readback)
		{
		  GM_INFO (("PCI slot FORCED to 64-bit mode\n"));
		}
	      else
		{
		  GM_INFO (("unable to FORCE PCI slot to 64-bit mode"
			    " - 32-bit mode enabled\n"));
		}
	    }
	}
    }
  else if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	   || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
    {
      /* according to Wen, we cannot tell the width of the PCI
	 slot on a rev 4 board */
      is->bus_width = 0;
    }


  /* Now that we have mapped the special regs, but before we access
     any memory accessed through the LANai EBUS, we need to take the
     ebus out of reset. We also make sure the LANai is in reset before
     we clear its memory below.  */

  GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("Taking EBUS out of reset. (0x%08x)\n",
				  gm_get_control_reg (is)));
  if (gm_ebus_reset_off (is) != GM_SUCCESS)
    {
      GM_NOTE (("Could not take EBUS out of reset.\n"));
      goto abort_with_eeprom;
    }

  gm_arch_spin (is, 20000);	/* BAD: Should not be needed. */

  GM_PRINT (GM_DEBUG_MAP_IO_SPACES,
	    ("control reg value is 0x%x.\n", gm_get_control_reg (is)));

  /* Initialize the L6 microcontroller, if any. */

  if (gm_init_uc (is) != GM_SUCCESS)
    {
      GM_NOTE (("Could not initialize microcontroller.\n"));
      goto abort_with_ebus_reset_off;
    }

  /* Map the SRAM */

  GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("  Mapping SRAM (0x%x bytes).\n",
				  is->lanai.eeprom.lanai_sram_size));
  is->lanai.sram = gm_map_board_region (is, GM_MEMORY_REGION);
  if (is->lanai.sram == NULL)
    {
      GM_WARN (("Could not map SRAM\n"));
      goto abort_with_ebus_reset_off;
    }
  gm_check_pci_config_space (is);
  
  /* Backup the SRAM contents on LX so they can be restored just
     before unmapping the board.  This ensures that as long as the GM
     driver unloads cleanly, the PCI config space and EEPROM state,
     which is buffered in SRAM in the LANaiX, will be restored so
     the GM driver can be loaded again. */
  
  if (gm_is_lX (is))
    {
      is->lanai.sram_backup = gm_malloc (is->lanai.eeprom.lanai_sram_size);
      if (!is->lanai.sram_backup)
	{
	  GM_WARN (("Could not allocate LANai SRAM backup buffer\n"));
	  goto abort_with_ebus_reset_off;
	}
      gm_bcopy ((void *)is->lanai.sram,
		is->lanai.sram_backup,
		is->lanai.eeprom.lanai_sram_size);
    }

  /* Clear the LANai SRAM */
  
  GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("Clearing LANai SRAM from %p to %p.\n",
				  is->lanai.sram,
				  (char *) is->lanai.sram +
				  is->lanai.eeprom.lanai_sram_size));
  __gm_kton_bzero (is, (void *) is->lanai.sram,
		   is->lanai.eeprom.lanai_sram_size);
  GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("Cleared LANai[%d] SRAM\n", is->id));

  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_ebus_reset_off:
  gm_ebus_reset_on (is);
abort_with_eeprom:
abort_with_special_regs:
  gm_unmap_board_region (is, GM_SPECIAL_REGS_REGION, is->lanai.special_regs);
abort_with_board_reset_off:
  gm_board_reset_on (is);
abort_with_control_regs:
  gm_unmap_board_region (is, GM_CONTROL_REGS_REGION,
			 (void *) is->lanai.control_regs);
abort_with_mapped_board:
  gm_arch_unmap_io_space (is, 0, is->board_span, &is->board_base);
abort_with_nothing:
  GM_RETURN_STATUS (GM_FAILURE);
}

static void
gm_unmap_io_spaces_pci_rev_2 (gm_instance_state_t * is)
{
  GM_CALLED ();
  
  gm_check_pci_config_space (is);
  
  GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("Unmapping IO spaces.\n"));
  if (is->lanai.sram_backup)
    {
      gm_bcopy (is->lanai.sram_backup, (void *) is->lanai.sram,
		is->lanai.eeprom.lanai_sram_size);
      gm_free (is->lanai.sram_backup);
      is->lanai.sram_backup = 0;
    }
  
  gm_check_pci_config_space (is);
  
  gm_unmap_board_region (is, GM_MEMORY_REGION, (void *) is->lanai.sram);
  gm_ebus_reset_on (is);
  gm_unmap_board_region (is, GM_SPECIAL_REGS_REGION, is->lanai.special_regs);
  gm_board_reset_on (is);
  gm_unmap_board_region (is, GM_CONTROL_REGS_REGION,
			 (void *) is->lanai.control_regs);
  gm_arch_unmap_io_space (is, 0, is->board_span, &is->board_base);
  /* Hack: we know that gm_unmap_board_region does nothing, so we can delay this until now. */
  
  GM_RETURN_NOTHING ();
}

static void
gm_verify_pci_command_bit (gm_instance_state_t * is, gm_u16_t bit,
			   const char *s)
{
  gm_u16_t command;

  gm_read_pci_config_command (is, &command);
  if (!(command & bit))
    {
      if (gm_enable_PCI_config_command_bit (is, bit) != GM_SUCCESS)
	{
	  GM_WARN (("PCI config %s NOT set by BIOS and could not be set\n",
		    s));
	}
      else
	{
	  gm_read_pci_config_command (is, &command);
	  if (!(command & bit))
	    {
	      GM_WARN (("PCI config %s NOT set by BIOS and could not be set\n",
			s));
	    }
	}
    }
  else
    {
      GM_PRINT (GM_PRINT_LEVEL >= 1, ("PCI config %s is already set\n", s));
    }
}

#endif /* GM_SUPPORT_PCI */

/****************************************************************
 * Generic board mapping/unmapping.
 ****************************************************************/

gm_status_t
gm_map_io_spaces (gm_instance_state_t * is)
{
  GM_CALLED_WITH_ARGS (("%p", is));
  
  if (GM_SUPPORT_SBUS && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("mapping sbus\n"));
      GM_RETURN_STATUS (gm_map_io_spaces_sbus (is));
    }

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      GM_PRINT (GM_DEBUG_MAP_IO_SPACES,
		("PCI_REV %d\n", is->ifc.pci.config.Revision_ID));

      gm_verify_pci_command_bit (is, GM_PCI_COMMAND_INVALIDATE,
				 "memory-write-and-invalidate");
      gm_verify_pci_command_bit (is, GM_PCI_COMMAND_MASTER, "master");
      gm_verify_pci_command_bit (is, GM_PCI_COMMAND_MEMORY, "memory");

      if (GM_SUPPORT_PCI_REV_1 && is->ifc.pci.config.Revision_ID == 1)
	{
	  GM_RETURN_STATUS (gm_map_io_spaces_pci_rev_1 (is));
	}

      if ((GM_SUPPORT_PCI_REV_2
	   || GM_SUPPORT_PCI_REV_3)
	  && ((is->ifc.pci.config.Revision_ID == 2) ||
	      (is->ifc.pci.config.Revision_ID == 3)))
	{
	  GM_RETURN_STATUS (gm_map_io_spaces_pci_rev_2 (is));
	}
      
      if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	  || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
	{
	  /* These boards are so similar to the rev. 2 boards that we
	     can initialize them with the same function. */
	  GM_RETURN_STATUS (gm_map_io_spaces_pci_rev_2 (is));
	}

      GM_NOTE
	(("Unrecognized PCI revision %d.\n", is->ifc.pci.config.Revision_ID));
    }
#endif /* GM_SUPPORT_PCI */
  
  if (CSPI_MAP26xx && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_MAP26xx)
    {
      GM_PRINT (GM_DEBUG_MAP_IO_SPACES, ("mapping CSPI MAP26xx\n"));
      GM_RETURN_STATUS (gm_map_io_spaces_map26xx (is));
    }

  GM_RETURN_STATUS (GM_FAILURE);
}

void
gm_unmap_io_spaces (gm_instance_state_t * is)
{
  GM_CALLED ();

  if (GM_SUPPORT_SBUS && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      gm_unmap_io_spaces_sbus (is);
      GM_RETURN_NOTHING ();
    }

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if (is->ifc.pci.config.Revision_ID == 1)
	{
	  gm_unmap_io_spaces_pci_rev_1 (is);
	  GM_RETURN_NOTHING ();
	}

      if ((GM_SUPPORT_PCI_REV_2 && (is->ifc.pci.config.Revision_ID == 2))
	  || (GM_SUPPORT_PCI_REV_3 && (is->ifc.pci.config.Revision_ID == 3))
	  || (GM_SUPPORT_PCI_REV_4 && (is->ifc.pci.config.Revision_ID == 4))
	  || (GM_SUPPORT_PCI_REV_5 && (is->ifc.pci.config.Revision_ID == 5)))
	{
	  gm_unmap_io_spaces_pci_rev_2 (is);
	  GM_RETURN_NOTHING ();
	}
    }
#endif /*GM_SUPPORT_PCI*/
  
  if (CSPI_MAP26xx && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_MAP26xx)
    {
      gm_unmap_io_spaces_map26xx (is);
      GM_RETURN_NOTHING ();
    }

  GM_RETURN_NOTHING ();
}

#if GM_DEBUG_MAP_IO_SPACES
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 0
#endif

/********************
 * MCP loading
 ********************/

/* Functions required to uncompress the MCP */

static voidpf
gm_zlib_calloc (voidpf opaque, uInt items, uInt size)
{
  void *ret;
  GM_PARAMETER_MAY_BE_UNUSED (opaque);

  GM_CALLED ();
  ret = gm_calloc (items, size);
  GM_RETURN_PTR (ret);
}

static void
gm_zlib_free (voidpf opaque, voidpf addr)
{
  GM_CALLED ();
  GM_PARAMETER_MAY_BE_UNUSED (opaque);
  gm_free (addr);
  GM_RETURN_NOTHING ();
}


/* declare all the MCPs supported by this driver */

#define GM_MCP_DECL(type)						\
extern const char gm_control_program ## type [];			\
extern const unsigned gm_control_program ## type ## _length;		\
extern const unsigned int gm_control_program ## type ## _uncompressed_length

#if GM_SUPPORT_64_BIT_USER_PTRS && GM_SUPPORT_32_BIT_USER_PTRS
#error only one user ptr bit size can be supported by a driver
#endif

#if GM_SUPPORT_64_BIT_USER_PTRS

GM_MCP_DECL (_64b_l5_0k);
GM_MCP_DECL (_64b_l5_16k);
GM_MCP_DECL (_64b_l5_4k);
GM_MCP_DECL (_64b_l5_8k);
GM_MCP_DECL (_64b_l6_0k);
GM_MCP_DECL (_64b_l6_16k);
GM_MCP_DECL (_64b_l6_4k);
GM_MCP_DECL (_64b_l6_8k);
GM_MCP_DECL (_64b_l7_0k);
GM_MCP_DECL (_64b_l7_16k);
GM_MCP_DECL (_64b_l7_4k);
GM_MCP_DECL (_64b_l7_8k);
GM_MCP_DECL (_64b_l8_0k);
GM_MCP_DECL (_64b_l8_16k);
GM_MCP_DECL (_64b_l8_4k);
GM_MCP_DECL (_64b_l8_8k);
GM_MCP_DECL (_64b_l9_0k);
GM_MCP_DECL (_64b_l9_16k);
GM_MCP_DECL (_64b_l9_4k);
GM_MCP_DECL (_64b_l9_8k);
GM_MCP_DECL (_64b_lX_0k);
GM_MCP_DECL (_64b_lX_16k);
GM_MCP_DECL (_64b_lX_4k);
GM_MCP_DECL (_64b_lX_8k);

#define GM_MCP_PARAM(type)						\
  &gm_control_program_64b_##type [0],					\
  &gm_control_program_64b_##type##_length,				\
  &gm_control_program_64b_##type##_uncompressed_length

#elif GM_SUPPORT_32_BIT_USER_PTRS

GM_MCP_DECL (_32b_l5_0k);
GM_MCP_DECL (_32b_l5_16k);
GM_MCP_DECL (_32b_l5_4k);
GM_MCP_DECL (_32b_l5_8k);
GM_MCP_DECL (_32b_l6_0k);
GM_MCP_DECL (_32b_l6_16k);
GM_MCP_DECL (_32b_l6_4k);
GM_MCP_DECL (_32b_l6_8k);
GM_MCP_DECL (_32b_l7_0k);
GM_MCP_DECL (_32b_l7_16k);
GM_MCP_DECL (_32b_l7_4k);
GM_MCP_DECL (_32b_l7_8k);
GM_MCP_DECL (_32b_l8_0k);
GM_MCP_DECL (_32b_l8_16k);
GM_MCP_DECL (_32b_l8_4k);
GM_MCP_DECL (_32b_l8_8k);
GM_MCP_DECL (_32b_l9_0k);
GM_MCP_DECL (_32b_l9_16k);
GM_MCP_DECL (_32b_l9_4k);
GM_MCP_DECL (_32b_l9_8k);
GM_MCP_DECL (_32b_lX_0k);
GM_MCP_DECL (_32b_lX_16k);
GM_MCP_DECL (_32b_lX_4k);
GM_MCP_DECL (_32b_lX_8k);

#define GM_MCP_PARAM(type)						\
  &gm_control_program_32b_##type [0],					\
  &gm_control_program_32b_##type##_length,				\
  &gm_control_program_32b_##type##_uncompressed_length

#else /* GM_SUPPORT_32_BIT_USER_PTRS */
#error None of GM_SUPPORT_*_BIT_USER_PTRS is set
#endif

#if GM_SUPPORT_L7 | GM_SUPPORT_L8 | GM_SUPPORT_L9
GM_MCP_DECL (_lanai_rate);
#endif

/* List of control programs, in order of preference. */

static struct gm_control_program_descriptor
{
  const char *program;
  const unsigned *zlength;
  const unsigned *mcp_length;
  const gm_u32_t required_mem_size;
  const gm_u16_t min_lanai_version;
  const gm_u16_t max_lanai_version;
  const gm_u32_t page_len;
  const gm_u32_t cache_line_size;
  const char *program_name;
}
gm_mcp_descriptor[] =
{
    /* L5 */
#if GM_SUPPORT_L5 && GM_SUPPORT_0K_PAGES
  {
    GM_MCP_PARAM (l5_0k),
      512 * 1024, 0x500, 0x506, 4096, GM_RDMA_GRANULARITY, "L5 0K",}
  ,
#endif
#if GM_SUPPORT_L5 && GM_SUPPORT_4K_PAGES
  {
    GM_MCP_PARAM (l5_4k),
      512 * 1024, 0x500, 0x506, 4096, GM_RDMA_GRANULARITY, "L5 4K",}
  ,
#endif
#if GM_SUPPORT_L5 && GM_SUPPORT_8K_PAGES
  {
    GM_MCP_PARAM (l5_8k),
      512 * 1024, 0x500, 0x506, 8192, GM_RDMA_GRANULARITY, "L5 8K",}
  ,
#endif
#if GM_SUPPORT_L5 && GM_SUPPORT_16K_PAGES
  {
    GM_MCP_PARAM (l5_16k),
      512 * 1024, 0x500, 0x506, 16384, GM_RDMA_GRANULARITY, "L5 16K",}
  ,
#endif
    /* L6 */
#if GM_SUPPORT_L6 && GM_SUPPORT_4K_PAGES
  {
    GM_MCP_PARAM (l6_4k),
      512 * 1024, 0x600, 0x601, 4096, GM_RDMA_GRANULARITY, "L6 4K",}
  ,
#endif
#if GM_SUPPORT_L6 && GM_SUPPORT_8K_PAGES
  {
    GM_MCP_PARAM (l6_8k),
      512 * 1024, 0x600, 0x601, 8192, GM_RDMA_GRANULARITY, "L6 8K",}
  ,
#endif
    /* L7 */
#if GM_SUPPORT_L7 && GM_SUPPORT_0K_PAGES
  {
    GM_MCP_PARAM (l7_0k),
      512 * 1024, 0x700, 0x705, 4096, GM_RDMA_GRANULARITY, "L7 0K",}
  ,
#endif
#if GM_SUPPORT_L7 && GM_SUPPORT_4K_PAGES
  {
    GM_MCP_PARAM (l7_4k),
      1024 * 1024, 0x700, 0x705, 4096, GM_RDMA_GRANULARITY, "L7 4K",}
  ,
#endif
#if GM_SUPPORT_L7 && GM_SUPPORT_8K_PAGES
  {
    GM_MCP_PARAM (l7_8k),
      1024 * 1024, 0x700, 0x705, 8192, GM_RDMA_GRANULARITY, "L7 8K",}
  ,
#endif
#if GM_SUPPORT_L7 && GM_SUPPORT_16K_PAGES
  {
    GM_MCP_PARAM (l7_16k),
      1024 * 1024, 0x700, 0x705, 16384, GM_RDMA_GRANULARITY, "L7 16K",}
  ,
#endif
#if GM_SUPPORT_L8 && GM_SUPPORT_4K_PAGES
  {
    GM_MCP_PARAM (l8_4k),
      512 * 1024, 0x800, 0x802, 4096, GM_RDMA_GRANULARITY, "L8 4K",}
  ,
#endif
#if GM_SUPPORT_L8 && GM_SUPPORT_8K_PAGES
  {
    GM_MCP_PARAM (l8_8k),
      512 * 1024, 0x800, 0x802, 8192, GM_RDMA_GRANULARITY, "L8 8K",}
  ,
#endif
#if GM_SUPPORT_L8 && GM_SUPPORT_16K_PAGES
  {
    GM_MCP_PARAM (l8_16k),
      512 * 1024, 0x800, 0x802, 16384, GM_RDMA_GRANULARITY, "L8 16K",}
  ,
#endif
#if GM_SUPPORT_L9 && GM_SUPPORT_0K_PAGES
  {
    GM_MCP_PARAM (l9_0k),
      1024 * 1024, 0x900, 0x904, 4096, GM_RDMA_GRANULARITY, "L9 0K",}
  ,
#endif
#if GM_SUPPORT_L9 && GM_SUPPORT_4K_PAGES
  {
    GM_MCP_PARAM (l9_4k),
      1024 * 1024, 0x900, 0x904, 4096, GM_RDMA_GRANULARITY, "L9 4K",}
  ,
#endif
#if GM_SUPPORT_L9 && GM_SUPPORT_8K_PAGES
  {
    GM_MCP_PARAM (l9_8k),
      1024 * 1024, 0x900, 0x904, 8192, GM_RDMA_GRANULARITY, "L9 8K",}
  ,
#endif
#if GM_SUPPORT_L9 && GM_SUPPORT_16K_PAGES
  {
    GM_MCP_PARAM (l9_16k),
      1024 * 1024, 0x900, 0x904, 16384, GM_RDMA_GRANULARITY, "L9 16K",}
  ,
#endif
#if GM_SUPPORT_LX && GM_SUPPORT_0K_PAGES
  {
    GM_MCP_PARAM (lX_0k),
      1024 * 1024, 0xA00, 0xA00, 40X6, GM_RDMA_GRANULARITY, "LX 0K",}
  ,
#endif
#if GM_SUPPORT_LX && GM_SUPPORT_4K_PAGES
  {
    GM_MCP_PARAM (lX_4k),
      1024 * 1024, 0xA00, 0xA00, 4096, GM_RDMA_GRANULARITY, "LX 4K",}
  ,
#endif
#if GM_SUPPORT_LX && GM_SUPPORT_8K_PAGES
  {
    GM_MCP_PARAM (lX_8k),
      1024 * 1024, 0xA00, 0xA00, 8192, GM_RDMA_GRANULARITY, "LX 8K",}
  ,
#endif
#if GM_SUPPORT_LX && GM_SUPPORT_16K_PAGES
  {
    GM_MCP_PARAM (lX_16k),
      1024 * 1024, 0xA00, 0xA00, 16384, GM_RDMA_GRANULARITY, "LX 16K",}
  ,
#endif
  {
    0, 0, 0, 0, 0, 0, 0, 0, 0	/* null final array element */
  }
};




/************************************************************************
 * Ethernet address translation table
 ************************************************************************/

/* This driver caches a table of ethernet IDs for each GM ID.  In
   addition, it provides another LANai-writable table of ethernet IDs.
   stale entries are detected by comparing the cached entry with the
   entry in the LANai-writable table before returning values to the
   user.  The cache is updated whenever stale entries are detected. */

static gm_status_t
gm_register_addr_table_piece (void *arg, gm_dp_t dma_addr, gm_u32_t page_num)
{
  gm_instance_state_t *is = arg;

  gm_assert (page_num < GM_MAX_NUM_ADDR_TABLE_PIECES);
  gm_assert (is);
  gm_assert (is->lanai.globals);

  if (((page_num % 16) == 0) || (page_num == GM_MAX_NUM_ADDR_TABLE_PIECES - 1))
     GM_PRINT (GM_PRINT_LEVEL >= 1,
               ("Recording DMA addr 0x%8x for address table page %d.\n",
               dma_addr, page_num));
  else
     GM_PRINT (GM_PRINT_LEVEL >= 7,
               ("Recording DMA addr 0x%8x for address table page %d.\n",
               dma_addr, page_num));
  gm_write_lanai_global_dp (is, ethernet.addr_table_piece[page_num],
                            dma_addr);
  return GM_SUCCESS;
}

static long
compare_unique_ids (void *key1, void *key2)
{
  return gm_memcmp (key1, key2, sizeof (gm_unique_id_64_t));
}

static unsigned long
hash_unique_ids (void *key)
{
  return gm_crc (key, sizeof (gm_unique_id_64_t));
}

/************************************************************************
 * Load the control program
 ************************************************************************/

/* Load the Myrinet Control Program (mcp) into the board's sram
 * and take the board out of reset. 
 *
 * NOTE: Some or all SGI (IRIX) platforms do not set the Cache Line Size
 *       register of the PCI Config space.  In this case, the value
 *       is->cache_line_size will always be 0, and the MCP-selection logic
 *       below will not be influenced by this parameter.  As far as we can
 *       tell, this does not cause us any problems.         - Max 99/04/27
 */

#define GM_DEBUG_MCP_LOADING 0

/* forward decl. */
static gm_status_t gm_update_clockval_rev3 (gm_instance_state_t * is);
static gm_status_t gm_get_pci_speed_rev4 (gm_instance_state_t * is);

#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1

static gm_status_t
gm_instance_load_mcp (gm_instance_state_t * is)
{
  struct gm_control_program_descriptor *d;
  gm_status_t status;

  GM_CALLED_WITH_ARGS (("%p", is));

#if !GM_ETHERBOOT
  if (!GM_PAGE_LEN)
    GM_RETURN_STATUS (GM_INTERNAL_ERROR);
#endif

  GM_PRINT (GM_DEBUG_MCP_LOADING, ("Loading the MCP.\n"));

  /* find the biggest control program that will fit in the LANai SRAM. */

  GM_PRINT (GM_DEBUG_MCP_LOADING,
            ("Finding best control program for this LANai[%d].\n",
            is->id));

  for (d = &gm_mcp_descriptor[0]; d->program; d++)
    {
      GM_PRINT (GM_DEBUG_MCP_LOADING,
                ("check MCP: '%s'  len= %u (0x%x), req= %u (0x%x),\n"
                 "    min_ver= %x, max_ver= %x, page_len = %d, caln= %d\n",
                d->program_name,
		*d->mcp_length,       *d->mcp_length,
	        d->required_mem_size, d->required_mem_size,
	        d->min_lanai_version, d->max_lanai_version,
	        d->page_len,          d->cache_line_size));
      if (d->required_mem_size <= is->lanai.eeprom.lanai_sram_size
	  && *d->mcp_length <= is->lanai.eeprom.lanai_sram_size
	  && d->min_lanai_version <= is->lanai.eeprom.lanai_cpu_version
	  && d->max_lanai_version >= is->lanai.eeprom.lanai_cpu_version
#if GM_ETHERBOOT
	  && d->page_len == 0
#else
	  && d->page_len == GM_PAGE_LEN
#endif
	)

	break;
    }
  if (!d->program)
    {
      GM_WARN (("Could not find matching control program for unit %d.\n",
                is->id));
      _GM_WARN (("LANai = 0x%x  sram_size = 0x%x Bytes\n",
		 is->lanai.eeprom.lanai_cpu_version,
		 is->lanai.eeprom.lanai_sram_size));
      _GM_WARN (("Supported MCPs:\n"));
      for (d = &gm_mcp_descriptor[0]; d->program; d++)
	{
	  GM_INFO (("MCP: '%s' m:%d l:%d v:%x-%x pl:%d cl:%d\n",
		    d->program_name, d->required_mem_size, *d->mcp_length,
		    d->min_lanai_version, d->max_lanai_version, d->page_len,
		    d->cache_line_size));
	}
      status = GM_UNSUPPORTED_DEVICE;
      goto abort_with_nothing;
    }

#if !GM_CPU_x86 && !GM_CPU_alpha && !GM_CPU_powerpc && !GM_CPU_ia64
  /* for most archs (actually probably all), we do not have to
     care about the cache_line_size at the software level */
  if (d->cache_line_size < is->cache_line_size)
    {
      GM_WARN (("PCI Cache line size %d greater than GM granularity(%d)\n",
		is->cache_line_size, d->cache_line_size));
    }
#endif		    
  GM_PRINT (GM_DEBUG_BOARD_INIT,
	    ("Using MCP for unit %d: '%s' m:%d l:%d v:%x-%x pl:%d cl:%d\n",
	     is->id,
	     d->program_name, d->required_mem_size,*d->mcp_length,
	     d->min_lanai_version, d->max_lanai_version,
	     d->page_len, d->cache_line_size));
  GM_INFO (("MCP for unit %d: %s\n", is->id, d->program_name));

#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI_REV_3 && is->ifc.pci.config.Revision_ID == 3)
    {
       if (gm_update_clockval_rev3(is) != GM_SUCCESS)
         {
           GM_NOTE (("Couldn't update the LANai clockval for PCI rev3 board.\n"));
	   goto abort_with_nothing;

         }
       GM_PRINT (GM_DEBUG_BOARD_INIT,
		 ("Updated the clockval for pci rev 3 board\n"));
    }
  if ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
      || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5))
    {
      if (gm_get_pci_speed_rev4(is) != GM_SUCCESS)
	{
	  GM_NOTE (("Couldn't determine PCI bus speed/width on rev4 board.\n"));
	}
    }
#elif CSPI_MAP26xx /* !GM_SUPPORT_PCI && CSPI_MAP26xx */
  /* gm_update_clockval sets bus_rate and bus_width values if we have pci,
     so we should also set them for the cspi boards */
  if ( gm_is_l5(is) )
    {
      is->bus_rate = 33;
      is->bus_width = 64;
    }
  if ( gm_is_l9(is) )
    {
      is->bus_rate = 66;
      is->bus_width = 64;
    }
#endif /* CSPI_MAP26xx */
  
  GM_PRINT (GM_DEBUG_BOARD_INIT,
	    ("Inflating control program.\n"));
  /* Copy the control program into the LANai. */

  {
    /* Load a compressed MCP - see zlib.h for
       details */

    z_stream zs;
    int save_sram_size, rv;
    void *inflate_buffer;

    zs.zalloc = gm_zlib_calloc;
    zs.zfree = gm_zlib_free;
    zs.opaque = 0;

    save_sram_size = is->lanai.eeprom.lanai_sram_size;
    if (GM_NO_PARTWORD_PIO_WRITE)
      {
	/* inflate into kernel memory buffer to avoid partword PIO */

	inflate_buffer = gm_malloc (save_sram_size);
	if (!inflate_buffer)
	  {
	    GM_NOTE (("Could not allocate inflate buffer\n"));
	    goto abort_with_nothing;
	  }
	__gm_kton_bzero (is, inflate_buffer, save_sram_size);
      }
    else
      {
	/* inflate directly into LANai SRAM */

	inflate_buffer = (void *) is->lanai.sram;
      }

    rv = inflateInit (&zs);
    if (rv != Z_OK)
      {
	GM_NOTE (("fatal error in inflateInit(%d)\n", rv));
	goto abort_with_inflate_buffer;
      }

    zs.next_in = (unsigned char *) d->program;
    zs.avail_in = *d->zlength;

    save_sram_size = is->lanai.eeprom.lanai_sram_size;

    zs.next_out = (unsigned char *) inflate_buffer;
    zs.avail_out = save_sram_size;

    GM_PRINT (GM_DEBUG_MCP_LOADING, ("next_in %x %x %x %x %x\n", *zs.next_in,
				     *(zs.next_in + 1), *(zs.next_in + 2),
				     *(zs.next_in + 3), *(zs.next_in + 4)));


    GM_PRINT (GM_DEBUG_MCP_LOADING, ("start to inflate mcp\n"));
    _GM_PRINT (GM_DEBUG_MCP_LOADING,
	       ("   lanai sram space originally    = %d\n", save_sram_size));
    _GM_PRINT (GM_DEBUG_MCP_LOADING,
	       ("   bytes remaining to uncompress  = %d\n", zs.avail_in));
    _GM_PRINT (GM_DEBUG_MCP_LOADING,
	       ("   Available space for decompress = %d\n", zs.avail_out));


    {
      int err;

      if ((err = inflate (&zs, Z_FINISH)) != Z_STREAM_END
	  || zs.avail_in != 0 || zs.avail_out == 0)
	{
	  GM_NOTE (("fatal error in inflate - mcp too big   err %d?\n", err));
	  _GM_NOTE (("   lanai sram space originally    = %d\n",
		     save_sram_size));
	  _GM_NOTE (("   bytes remaining to uncompress  = %d\n",
		     zs.avail_in));
	  _GM_NOTE (("   Available space for decompress = %d\n",
		     zs.avail_out));
	  _GM_NOTE (("   inflate error msg %s\n", zs.msg));
	  _GM_NOTE (("   inflate error char %x\n",
		     *(unsigned char *) zs.next_in));
	  _GM_NOTE (("   inflate error char %x\n",
		     *(unsigned char *) (zs.next_in - 1)));
	  _GM_NOTE (("   inflate error char %x\n",
		     *(unsigned char *) (zs.next_in + 1)));
	  _GM_NOTE (("   inflate error total %ld\n", zs.total_in));
	  goto abort_with_inflate_init;
	}
    }

    if (inflateEnd (&zs) != Z_OK)
      {
	GM_NOTE (("fatal error in inflateEnd\n"));
	goto abort_with_inflate_buffer;
      }


    if (GM_NO_PARTWORD_PIO_WRITE)
      {

	__gm_kton_bcopy (is, inflate_buffer, is->lanai.sram, zs.total_out);
	gm_free (inflate_buffer);
      }

    GM_PRINT (GM_DEBUG_MCP_LOADING, ("Inflated mcp from %ld to %ld bytes.\n",
				     *d->zlength, zs.total_out));

    /* error handling */

    if (0)
      {
      abort_with_inflate_init:
	inflateEnd (&zs);
      abort_with_inflate_buffer:
	gm_free (inflate_buffer);
	goto abort_with_nothing;
      }
  }

  if (gm_read_first_sram_page (is, globals, lp) != 0)
    {
      GM_WARN (("Problem writing LANai SRAM: could not clear globals pointer.\n"));
      goto abort_with_nothing;
    }
  
  /* Take the LANai out of reset. */

  GM_PRINT (GM_DEBUG_MCP_LOADING, ("Taking the LANai out of reset\n"));
  if (gm_lanai_reset_off (is) != GM_SUCCESS)
    {
      GM_WARN (("Could not take LANai out of reset.\n"));
      goto abort_with_nothing;
    }
  is->lanai.running = 1;
  GM_PRINT (GM_DEBUG_MCP_LOADING, ("LANai now out of reset\n"));

  /* Now that the LANai is out of reset, test host access to LANai
     registers. */

  if (gm_test_RMP (is) != GM_SUCCESS)
    {
      GM_WARN (("RMP test FAILED!\n"));
      goto abort_running;
    }

  /* Wait for the MCP to initialize itself */

  GM_PRINT (GM_DEBUG_MCP_LOADING,
	    ("Waiting for the LANai MCP to initialize itself.\n"));
  {
    gm_u32_t globals;
    gm_s32_t start_time, current_time;

    if (gm_get_RTC (is, &start_time) != GM_SUCCESS)
      goto abort_running;

    do
      {
	if (gm_get_globals_ptr (is, &globals) != GM_SUCCESS)
	  goto abort_running;
	if (gm_get_RTC (is, &current_time) != GM_SUCCESS)
	  goto abort_running;

	/* Test for timeout.  This simple comparison works because RTC
	   is initialized to zero during reset */
	if (current_time - start_time > 20 * 1000 * 1000)
	  {
	    GM_NOTE (("[%d]: lanai did not initialize itself in 10 seconds"
		      " (0x%x).\n",
		      is->id, gm_ntoh_u32 (*((gm_u32_n_t *) is->lanai.sram))));
	    GM_PRINT (GM_DEBUG_MCP_LOADING,
		      ("current_time = 0x%x, start_time = 0x%x\n",
		       current_time, start_time));
	    _GM_PRINT
	      (GM_DEBUG_MCP_LOADING,
	       ("globals_ptr = 0x%08x\n", globals));
	    goto abort_running;
	  }
      }
    while (!globals);

    /* Sanity check the globals value. */

    if (globals >= is->lanai.eeprom.lanai_sram_size)
      {
	GM_WARN
	  (("Problem with firmware: globals pointer 0x%x is not in SRAM.\n", globals));
	_GM_PRINT
	  (1, ("re-read globals: 0x%x\n", gm_read_first_sram_page (is, globals, lp)));
	_GM_PRINT
	  (1,
	   ("init status = 0x%x\n", gm_read_first_sram_page (is, init_progress, u32)));
	
	_GM_PRINT (1, ("Start of SRAM contains the following:\n"));
	gm_hex_dump ((void *) is->lanai.sram, sizeof (gm_first_sram_page_t));
	goto abort_running;
      }
    
    /* Record pointer to LANai globals in the kernel address space. */
    is->lanai.globals =
      (gm_lanai_globals_t *) ((char *) is->lanai.sram + globals);
    GM_PRINT (GM_DEBUG_MCP_LOADING,
	      ("LANai globals at LANai address %x.\n", globals));
  }

#define GM_DEBUG_GLOBAL_SIZE 0

  GM_PRINT (GM_DEBUG_GLOBAL_SIZE, ("Globals at 0x%p (SRAM at 0x%p)\n",
				  is->lanai.globals, is->lanai.sram));
  GM_PRINT (GM_DEBUG_GLOBAL_SIZE, ("global magic number is 0x%x.\n",
				  gm_read_lanai_global_u32 (is, magic)));
  _GM_PRINT (GM_DEBUG_GLOBAL_SIZE, ("global length is %d.\n",
				  gm_read_lanai_global_u32 (is, length)));
  _GM_PRINT (GM_DEBUG_GLOBAL_SIZE, ("sizeof(gm_lanai_globals_t) is %d.\n",
				  sizeof (gm_lanai_globals_t)));
  _GM_PRINT (GM_DEBUG_GLOBAL_SIZE, ("global end magic number is %x.\n",
				  gm_read_lanai_global_u32 (is, end_magic)));
  if ((gm_read_lanai_global_u32 (is, magic) != 0xcafebabe)
      || (gm_read_lanai_global_u32 (is, length)
	  != sizeof (gm_lanai_globals_t))
      || (gm_read_lanai_global_u32 (is, end_magic) != 0xcafebabe))
    {
      GM_NOTE (("Myrinet interface is incompatible with installed driver\n"));
      if (gm_read_lanai_global_u32 (is, magic) != 0xcafebabe)
	{
	  _GM_NOTE (("(bad magic number at start)\n"));
	}
      if (gm_read_lanai_global_u32 (is, length)
	  != sizeof (gm_lanai_globals_t))
	{
	  _GM_NOTE (("(bad globals length  lanai is %d  host is %d)\n",
		     gm_read_lanai_global_u32 (is, length),
		     sizeof (gm_lanai_globals_t)));
	}
      if (gm_read_lanai_global_u32 (is, end_magic) != 0xcafebabe)
	{
	  _GM_NOTE (("(bad magic number at end)\n"));
	}
      status = GM_INTERNAL_ERROR;
      goto abort_running;
    }
  
  /* Now that we know where the LANai globals are, set those
     that must be set by the host. */

  GM_PRINT (GM_DEBUG_MCP_LOADING, ("Setting some LANai variables.\n"));

  /* Tell the LANai how to generate interrupts early, in case it needs to
     print to the system console. */

  status = gm_setup_capabilities_if_needed (is);
  if (status != GM_SUCCESS)
    {
      GM_WARN (("Error trying to setup message signalled interrupts\n"));
      gm_perror ("could not set up MSI", status);
      goto abort_running;
    }
  
  /* Tell the LANai the size of its SRAM */

  gm_write_lanai_global_s32
    (is, sram_length, is->lanai.eeprom.lanai_sram_size);

  /* wait for the LANai to compute its max_node_id */

  {
    gm_s32_t start_time, current_time;
    gm_u32_t max_node_id;
      
    if (gm_get_RTC (is, &start_time) != GM_SUCCESS)
      goto abort_running;

    do
      {
	max_node_id = gm_read_lanai_global_u32 (is, connection.max_id);
	if (gm_get_RTC (is, &current_time) != GM_SUCCESS)
	  goto abort_running;
	/* Test for timeout. */
	if (current_time - start_time > 20 * 1000 * 1000)
	  {
	    GM_NOTE (("lanai did not set max_node_id in 10 seconds"
		      " (0x%x).\n",
		      gm_ntoh_u32 (*((gm_u32_n_t *) is->lanai.sram))));
	    goto abort_running;
	  }
      }
    while (!max_node_id);

    if (max_node_id < 16)
      {
	GM_WARN (("Board does not have enough memory (max_node_id = %d).\n",
		  max_node_id));
	_GM_WARN (("The board has %u bytes of memory.\n",
		   is->lanai.eeprom.lanai_sram_size));
	_GM_WARN (("(Perhaps the MCP was compiled with debugging turned on.)"
		   "\n"));
	goto abort_running;
      }

#if 0
#error Warning, limiting max_node_id to 400 for testing purposes
    if (max_node_id > 400)
      max_node_id = 400;	/* Temporary HACK to avoid IRIX problem */
#endif

    /* record the max_node_id */

    is->max_node_id = (gm_u16_t) max_node_id;
    GM_INFO (("Board %u supports %u remote nodes.\n",
	      is->id, (unsigned int) max_node_id));

  }

  /* Set the DMA_STATUS or BURST special register, and also set up the
     DMA chain pointers on the L5+. */

  if (gm_setup_DMA_engine (is) != GM_SUCCESS)
    goto abort_running;

  /* Set the VERSION register if the chip has one. */

  if (gm_is_l4 (is))
    {
      if (gm_set_VERSION (is) != GM_SUCCESS)
	goto abort_running;
    }

  /* Set the WINDOW register if the chip has one. */

  if (gm_is_l5 (is))
    {
      if (gm_set_WINDOW (is) != GM_SUCCESS)
	goto abort_running;
    }

  /* Setup the pointer to the lanai DMA page bitmap. */

  {
    GM_BITMAP_DECL ((*lanai_dma_bitmap), GM_MAX_HOST_DMA_PAGES);
      
    lanai_dma_bitmap
      = &((gm_lanai_globals_t *) is->lanai.globals)->debug.dma_page_bitmap;
#if GM_DMA_PAGE_BITMAP_COPY_NEEDED
    GM_BITMAP_COPY (is->dma_page_bitmap_copy, *lanai_dma_bitmap);
#endif
    is->dma_page_bitmap = lanai_dma_bitmap;
  }

  /* Allocate the page hash table. */

  GM_PRINT (GM_DEBUG_MCP_LOADING, ("Allocating the page hash table.\n"));
  status = gm_alloc_page_hash_table (is);
  if (status != GM_SUCCESS)
    {
      GM_WARN (("Could not allocate page hash table\n"));
      goto abort_running;
    }

  gm_write_lanai_global_u32 (is, bus_rate, (gm_u32_t) (is->bus_rate));
  gm_write_lanai_global_u32 (is, bus_width, (gm_u32_t) (is->bus_width));

  /* Tell the LANai the "ethernet" and GM addresses of the board */
  gm_write_lanai_global_byte_array
    (is, is->lanai.eeprom.lanai_board_id, ethernet.mac_address);
  gm_write_lanai_global_byte_array
    (is, is->lanai.eeprom.lanai_board_id, gm_mac_address);

  /* Tell the LANai the hostname of this machine by initializing the
     mapper reply messages. */

  {
    char tempname[GM_MAX_HOST_NAME_LEN + 8];
#if !GM_ETHERBOOT
    /* loic: change the line above to make the code dependant on the
       OS if you are too lazy to provide gm_arch_gethostname (or in
       a hurry) */
    if (gm_arch_gethostname (tempname, GM_MAX_HOST_NAME_LEN) == GM_SUCCESS)
      {
	if (is->id > 0)
	  {
	    /* try to add id to end of hostname e.g. hostname:1
	       hostname:2 etc. */
	    int i;
	    for (i = 0; i < GM_MAX_HOST_NAME_LEN; i++)
	      {
		if (tempname[i] == 0)
		  {
		    break;
		  }
	      }
	    if (i < GM_MAX_HOST_NAME_LEN - 3)
	      {
		tempname[i] = ':';
		if (is->id < 10)
		  {
		    tempname[i + 1] = ('0' + is->id);
		  }
		else
		  {
 		    tempname[i + 1] = ('a' + (is->id - 10));
		  }
		tempname[i + 2] = 0;
	      }
	  }

	gm_strncpy_to_lanai_globals (is, hostname, tempname);
      }
    else
#endif
      {
	gm_clear_lanai_global (is, hostname);
      }
  }

  /* Tell the LANai where to find the various DMA regions in host
     memory.  The lanai will automatically initialize its ports to use
     these pointers until they are opened.  */

  {
    gm_dp_t r, s;

    r = gm_arch_dma_region_dma_addr (&is->page_hash.bogus_rdma_region);
    s = gm_arch_dma_region_dma_addr (&is->page_hash.bogus_sdma_region);
    gm_notice_dma_addr (is->dma_page_bitmap, r);
    gm_notice_dma_addr (is->dma_page_bitmap, s);
    gm_write_lanai_global_dp (is, page_hash.bogus_sdma_ptr, s);
    gm_write_lanai_global_dp (is, page_hash.bogus_rdma_ptr, r);

    GM_PRINT (GM_TRACE_LANAI_DMA | GM_DEBUG_MCP_LOADING,
	      ("bogus_sdma_ptr = "GM_U64_TMPL"\n", GM_U64_ARG (s)));
    _GM_PRINT (GM_TRACE_LANAI_DMA | GM_DEBUG_MCP_LOADING,
	       ("bogus_rdma_ptr = "GM_U64_TMPL"\n", GM_U64_ARG (r)));
  }

  GM_PRINT (GM_DEBUG_MCP_LOADING,
	    ("Telling the LANai the size of its SRAM is 0x%x\n",
	     is->lanai.eeprom.lanai_sram_size));

#if !GM_ENABLE_VM
  /* If the machine does not support virtual memory, tell the lanai how
     to construct a DMA address from a host virtual memory address. */

  gm_write_lanai_global_dp (is, no_VM.set,
			    gm_arch_no_VM_dma_bits_to_set (is));
  gm_write_lanai_global_dp (is, no_VM.clear,
			    gm_arch_no_VM_dma_bits_to_clear (is));
  
  /*
  gm_kton_dp (is, &is->lanai.globals->no_VM.set,
	      gm_arch_no_VM_dma_bits_to_set (is));
  gm_kton_dp (is, &is->lanai.globals->no_VM.clear,
	      gm_arch_no_VM_dma_bits_to_clear (is));
  */
#endif /* !GM_ENABLE_VM */


  /* Tell the LANai that initialization is done.  Also, mask
     all ISR bits that might cause a spurious interrupt. */

  GM_PRINT (GM_DEBUG_MCP_LOADING,
	    ("Reporting completed initialization to LANai.\n"));
  GM_STBAR ();
  if (gm_set_globals_ptr (is, 0) != GM_SUCCESS)
    {
      GM_NOTE (("Could not set LANai globals pointer.\n"));
      goto abort_with_page_hash;
    }
  {
    gm_u32_t globals;
    if (gm_get_globals_ptr (is, &globals) != GM_SUCCESS)
      {
	GM_NOTE (("Could not get LANai globals pointer.\n"));
	goto abort_with_page_hash;
      }
  }
  
  /* SAB */
  GM_PRINT (GM_DEBUG_MCP_LOADING, ("set_EIMR = 0x%x\n", is->_set_EIMR));
  gm_set_EIMR (is, 0);
  GM_STBAR ();

  /****************
   * Wait for LANai to be completely initialized
   ****************/

  {
    gm_s32_t start_time, current_time;
    gm_u32_t initialized;
      
    gm_arch_spin(is,20000);
    if (gm_get_RTC (is, &start_time) != GM_SUCCESS)
      goto abort_with_page_hash;

    do
      {
	gm_arch_spin(is,10000);
	initialized = gm_read_lanai_global_u32 (is, initialized);
	if (gm_get_RTC (is, &current_time) != GM_SUCCESS)
	  goto abort_with_page_hash;
	/* Test for timeout.  This simple comparison works because RTC
	   is initialized to zero during reset */
	if (current_time - start_time > 20 * 1000 * 1000)
	  {
	    GM_NOTE (("lanai did not initialize in 10 seconds"
		      " (0x%x).\n",
		      gm_ntoh_u32 (*((gm_u32_n_t *) is->lanai.sram))));
	    goto abort_with_page_hash;
	  }
      }
    while (!initialized);
  }

  /****************
   * Read lanai-computed values
   ****************/

  /* report number of page hash cache entries */

  GM_INFO (("Board %u page hash cache has %u bins.\n", is->id,
	    (((unsigned int)
	      gm_read_lanai_global_u32 (is, page_hash.cache.max_index))
	     + 1)));

#if GM_DEBUG
  /* print out the pci performance numbers */
  {
    gm_u32_t e2l_time, l2e_time;

    e2l_time = gm_read_lanai_global_u32 (is, e2l_time[0]);
    l2e_time = gm_read_lanai_global_u32 (is, l2e_time[0]);
    GM_INFO(("e2l_time = %d l2e_time = %d\n", e2l_time, l2e_time));
    GM_INFO(("e2l = %d MB/sec, l2e = %d MB/sec\n",
	     GM_PAGE_LEN * GM_DMA_PERF_NUM_PAGES * 2 / e2l_time,
	     GM_PAGE_LEN * GM_DMA_PERF_NUM_PAGES * 2 / l2e_time));
  }
#endif

  is->node_id_hash = gm_create_ptr_hash (is->max_node_id);
  if (is->node_id_hash == NULL)
    {
      GM_WARN (("Failed to create node id hash\n"));
      goto abort_with_page_hash;
    }

  is->global_id_array
    = gm_calloc (is->max_node_id, sizeof (is->global_id_array[0]));
  if (is->global_id_array == 0)
    {
      GM_WARN (("Failed to create global id array\n"));
      goto abort_with_node_id_hash;
    }
  
  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_node_id_hash:
  gm_destroy_ptr_hash (is->node_id_hash);
abort_with_page_hash:
  gm_disable_lanai (is);
  gm_free_page_hash_table (is);
abort_running:
  is->lanai.running = 0;
  gm_lanai_reset_on (is);
abort_with_nothing:
  GM_RETURN_STATUS (GM_FAILURE);
}

static void
gm_instance_unload_mcp (gm_instance_state_t * is)
{
  int firmware_was_running;

  GM_CALLED ();

  /* Start unloading */

  firmware_was_running = is->lanai.running;
  is->lanai.running = 0;
  gm_check_pci_config_space (is);
  gm_lanai_reset_on (is);

  /* Now that the LANai is in reset, print some debugging info. */
  
  GM_PRINT
    (1,
     ("last firmware handler entered was %s()\n",
      (char *) is->lanai.sram
      + gm_read_lanai_global_lp (is, current_handler)));
  GM_PRINT
    (GM_ENABLE_DEBUG_COUNTERS,
     ("dispatch count = %u\n",
      gm_read_lanai_global_u32 (is, dispatch_cnt_debug_cnt)));
  /* Print a list of all dispatches seen. */
  if (!firmware_was_running && GM_ENABLE_LOG_DISPATCH && GM_DEBUG)
    {
      int i;

      /* Print the dispatch log */
      
      GM_PRINT (GM_ENABLE_LOG_DISPATCH,
		("Dispatch log:\n"));
      {
	gm_lp_t log_slot;
	gm_lp_t log_end;
	unsigned long slot;
	
	log_slot = gm_read_lanai_global_lp (is, log_slot);
	log_end = gm_read_lanai_global_lp (is, log_end);
	slot = (GM_LOG_LEN
		- (log_end - log_slot) / sizeof (gm_firmware_log_entry_t));
	for (i=0; i<GM_LOG_LEN; i++)
	  {
	    gm_lp_t message;

	    message
	      = (gm_read_lanai_global_lp
		 (is, log[(log_slot + i) % GM_LOG_LEN].message));
	    if (message)
	      {
		_GM_PRINT (1, ("%s\n", (char *) is->lanai.sram + message));
	      }
	  }
      }
      
      /* Print the "dispatch_seen" table */
      
      GM_PRINT (GM_ENABLE_LOG_DISPATCH,
		("The following dispatches were seen:\n"));
      for (i=0; i<GM_DISPATCH_MAX_NUM; i++)
	{
	  gm_lp_t lp;

	  lp = gm_read_lanai_global_lp (is, dispatch_seen[i]);
	  if (lp)
	    {
	      gm_u32_t cnt;
	    
	      cnt = gm_read_lanai_global_u32 (is, dispatch_cnt[i]);
	      _GM_PRINT
		(1,
		 ("%4d seen %8u times (%s)\n",
		  (int) i, (unsigned int) cnt, (char *) is->lanai.sram + lp));
	    }
	}

      /* Print the current handler. */
      
      {
	gm_lp_t lp;
	
	lp = gm_read_lanai_global_lp (is, current_handler);
	_GM_PRINT
	  (1,
	   ("current_handler=\"%s\"\n",
	    lp ? (char *)is->lanai.sram + lp : "(NULL)"));
	
	lp = gm_read_lanai_global_lp (is, current_handler_extra);
	_GM_PRINT
	  (1,
	   ("current_handler_extra=\"%s\"\n",
	    lp ? (char *)is->lanai.sram + lp : "(NULL)"));
      }
  }
  _GM_PRINT
    (1,
     ("RTC.whole = 0x%x%08x\n",
      gm_read_first_sram_page (is, rtc.parts.high, u32),
      gm_read_first_sram_page (is, rtc.parts.low, u32)));
  _GM_PRINT (1, ("gm.timeout_time = 0x%x\n",
		 gm_read_lanai_global_s32 (is, timeout_time)));
  if (GM_DEBUG_LX && gm_is_lX (is))
    {
      _GM_PRINT (1, ("ISR = 0x%x\n",
		     gm_read_lanai_special_reg_u32 (is, lX.ISR)));
       _GM_PRINT (1, ("RTC = 0x%x\n",
		     gm_read_lanai_special_reg_s32 (is, lX.RTC)));
      _GM_PRINT (1, ("P0_RECV_COUNT = 0x%x\n",
		     gm_read_lanai_special_reg_u16 (is, lX.P0_RECV_COUNT)));
      _GM_PRINT (1, ("P0_RB = 0x%x\n",
		     gm_read_lanai_special_reg_u32 (is, lX.P0_RB)));
    }
  
  /* Continue unloading */
  
  gm_check_pci_config_space (is);
#if !GM_ETHERBOOT
  gm_free_page_hash_table (is);
#endif
  gm_check_pci_config_space (is);

  GM_RETURN_NOTHING ();
}

#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1


#if 0
void
gm_reset_everything (gm_instance_state_t * is)
{
  gm_lanai_reset_on (is);
  gm_ebus_reset_on (is);
  gm_board_reset_on (is);
  gm_arch_spin (is, 20000);
}
#endif

/* Initialize the interface, mapping all requisite data structures
   into system memory.  Also enables interrupts. */

#define GM_DEBUG_CACHE_LINE_SIZE 0

gm_status_t
gm_instance_init (gm_instance_state_t * is, gm_u32_t unit,
		  enum gm_bus_type type)
{
  gm_status_t status;

  GM_CALLED ();

  if (gm_init () != GM_SUCCESS)
    {
      GM_NOTE (("gm_instance_init: gm_init() failed. Could not init "
		"package.\n"));
      goto abort_with_nothing;
    }


  if (gm_init_instance_support () != GM_SUCCESS)
    {
      GM_NOTE (("Could not initialize instance_state package.\n"));
      goto abort_with_gm_init;
    }

  /* Put the instance in the instance hash table to allow id->instance
     translations to be performed. */


#if !GM_ETHERBOOT
  if (gm_hash_insert (gm_instance_hash, &unit, is) != GM_SUCCESS)
    {
      GM_NOTE (("Could not insert state in hash table.\n"));
      goto abort_with_init;
    }
#endif

  /* Initialize the instance state structure. */

  is->id = unit;
 
  gm_strncpy (is->tag_string, GM_INSTANCE_TAG_STRING, sizeof(is->tag_string));
  (is->tag_string)[sizeof(is->tag_string)-1] = '\0';

#if GM_DMA_PAGE_BITMAP_COPY_NEEDED
  is->dma_page_bitmap = &is->dma_page_bitmap_copy;
  gm_bzero (is->dma_page_bitmap_copy, sizeof (is->dma_page_bitmap_copy));
#else
  is->dma_page_bitmap = 0;
#endif
  /* is->flags = 0; */
  /* is->lanai.* cleared */
  /* is->pause_rqst = 0 */
  /* is->pause_ack = 0 */

  gm_arch_sync_init (&is->command_sync, is);
  is->lanai.eeprom.bus_type = type;	/* HACK */

  /* If this is a pci device, read the revision id and cache line size. */
#if GM_SUPPORT_PCI
  if (GM_SUPPORT_PCI && is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      if (gm_read_pci_config_revision_id (is, &is->ifc.pci.config.Revision_ID)
	  != GM_SUCCESS)
	{
	  GM_WARN (("Could not read PCI board revision ID.\n"));
	  goto abort_with_syncs;
	}
      if (is->ifc.pci.config.Revision_ID == 3)
	{
	  GM_PRINT (GM_DEBUG_BOARD_INIT, ("This is a pci REV 3 board\n"));
	}
      if (gm_read_pci_cache_line_size
	  (is, &is->ifc.pci.config.Cache_Line_Size) != GM_SUCCESS)
	{
	  GM_WARN (("Could not read PCI cache line size.\n"));
	  goto abort_with_syncs;
	}
      is->cache_line_size = 4 * is->ifc.pci.config.Cache_Line_Size;
      if (!GM_POWER_OF_TWO (is->cache_line_size))
	{
	  GM_NOTE (("Cache line size reported by BIOS in PCI "
		    "configuration register (%u) is not a power of 2.  "
		    "Interface disabled.\n", (unsigned int) is->cache_line_size));
	  goto abort_with_syncs;
	}
      GM_PRINT (GM_DEBUG_CACHE_LINE_SIZE,
                ("Cache line size reported in Myrinet card is %d\n",
                 is->cache_line_size));

      /* gm_set_64_bit_flag() depends on is->ifc.pci.config.Revision_ID */
      if (gm_set_64_bit_flag(is) != GM_SUCCESS)
        {
          GM_WARN (("Could not determine board %d DMA address size.\n", unit));
          goto abort_with_syncs;
        }
    }
#endif /* GM_SUPPORT_PCI */
  /* is->ifc.* initialized by caller. */
  if (gm_instance_init_functions (is) != GM_SUCCESS)
    {
      GM_WARN (("Could not initialize instance functions.\n"));
      goto abort_with_syncs;
    }


  /* Now it is safe to touch the interface. */

  /* Map the LANai interface to system memory. This also copies the
     eeprom into the instance state. */

  status = gm_map_io_spaces (is);
  if (status != GM_SUCCESS)
    {
      GM_WARN (("Can't map IO memory to system memory\n"));
      goto abort_with_syncs;
    }

  /* Test and clear the SRAM */

  gm_check_pci_config_space (is);
  GM_PRINT (GM_DEBUG_MAP_IO_SPACES,
	    ("Testing, clearing LANai SRAM from %p to %p.\n", is->lanai.sram,
	     (char *) is->lanai.sram + is->lanai.eeprom.lanai_sram_size));

  status = gm_test_and_clear_sram (is, is->lanai.sram,
				   is->lanai.eeprom.lanai_sram_size);
  if (status != GM_SUCCESS)
    {
      GM_WARN (("sram and register tests failed.\n"));
      goto abort_with_io_spaces;
    }
  gm_check_pci_config_space (is);

  /* Bob,

     Should this really be here? - SAB CSPI

     In gm_map_io_spaces you turn on the lanai chip and ebus for some work.
     Afterwards, you disable ebus for some systems while you leve it on for
     other systems.  If you consistently leave in on in gm_map_io_spaces, then
     you shouldn't need it here.

     We had to leave it on in gm_map_io_spaces so that gm_test_and_clear_sram
     could run. */


  /* Copy the myrinet control program into LANai SRAM. */

  /* CSPI: This reset avoids random hang from boot kernel to loadable
     kernel I'd suspect a soft reset doesn't reset the LANai - that
     could be a problem.  Should we really do lanai_reset_on,
     ebus_reset_on, ebus_reset_off?? */
  gm_check_pci_config_space (is);
  GM_PRINT (GM_DEBUG_MAP_IO_SPACES,
	    ("Extra reset here.\n"));
  gm_lanai_reset_on (is);
  gm_ebus_reset_on (is);
  gm_arch_spin (is, 20000);
  gm_ebus_reset_off (is);
  gm_arch_spin (is, 20000);
  GM_PRINT (GM_DEBUG_MAP_IO_SPACES,
	    ("End of extra reset.\n"));
  gm_check_pci_config_space (is);

  /* Load the mcp into the lanai */

  if (GM_ENABLE_FIRMWARE)
    {
      status = gm_instance_load_mcp (is);
      if (status != GM_SUCCESS)
	{
	  GM_NOTE (("could not load control program.\n"));
	  goto abort_with_io_spaces;
	}
    }
  gm_check_pci_config_space (is);

  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_io_spaces:
  gm_unmap_io_spaces (is);
abort_with_syncs:
  gm_arch_sync_destroy (&is->command_sync);
abort_with_init:
  gm_finalize_instance_support ();
abort_with_gm_init:
  gm_finalize ();
abort_with_nothing:
  GM_RETURN_STATUS (GM_FAILURE);
}

void
gm_instance_finalize (gm_instance_state_t * is)
{
  GM_CALLED ();
  if (GM_ENABLE_FIRMWARE)
    {
      gm_instance_unload_mcp (is);
    }
#if !GM_ETHERBOOT
  if (gm_hash_remove(gm_instance_hash, &is->id) == NULL)
    {
      GM_NOTE(("Error removing instance from hash table for unit %d\n", is->id));
    }
#endif
  gm_unmap_io_spaces (is);
  gm_free (is->global_id_array);
  gm_destroy_ptr_hash(is->node_id_hash);
  gm_arch_sync_destroy (&is->command_sync);
  gm_finalize_instance_support ();
  gm_finalize ();
  GM_RETURN_NOTHING ();
}

#if GM_DEBUG_INSTANCE_STATE
void
gm_dump_instance (gm_instance_state_t * is)
{
  GM_CALLED ();
  GM_PARAMETER_MAY_BE_UNUSED (is);     /* if not GM_DEBUG */

  GM_PRINT (GM_PRINT_LEVEL >= 0, ("* (gm_instance_state_t *) %p = {\n", is));
#define FOO(name,tmpl) _GM_PRINT (GM_PRINT_LEVEL >= 0,			\
				 ("  "#name" = "#tmpl";\n", is->name))
#define FOO2(name) _GM_PRINT (GM_PRINT_LEVEL >= 0, ("  "#name" = ?;\n"))
  FOO (id, %d);
  FOO (flags, %d);
  FOO2 (lanai);
  FOO (lanai.globals, %p);
  FOO (board_base, %p);
  FOO (board_span, %d);
  FOO2 (pause_rqst);
  FOO2 (pause_ack);
  FOO2 (command_sync);
  FOO2 (daemon_rqst_q);
  FOO2 (ifc);
  FOO (_set_EIMR, %p);
  FOO (set_ISR, %p);
  FOO (get_EIMR, %p);
  FOO (get_ISR, %p);
  FOO (max_node_id, %d);
  FOO2 (arch);
  FOO (clone_minor, %d);
  FOO (privileged_clone_minor, %d);
#undef FOO
#undef FOO2
  _GM_PRINT (GM_PRINT_LEVEL >= 0, ("};\n"));

  GM_RETURN_NOTHING ();
}
#endif

#if !GM_ETHERBOOT
gm_instance_state_t *
gm_instance_for_id (unsigned int unit)
{
  return (gm_instance_state_t *) gm_hash_find (gm_instance_hash, &unit);
}
#endif

#if GM_SUPPORT_PCI

#define GM_DEBUG_GET_PCI_SPEED_REV4 0
static gm_status_t
gm_get_pci_speed_rev4 (gm_instance_state_t * is)
{
  gm_u32_t cpuc,rtc1,rtc2,rtc, cpu_freq, pci_clock, pci_freq;
  
  GM_CALLED();
  rtc1 = gm_read_lanai_special_reg_s32 (is, lX.RTC);
  cpuc = gm_read_lanai_special_reg_u32 (is, lX.CPUC);
  rtc2 = gm_read_lanai_special_reg_s32 (is, lX.RTC);

  rtc = (rtc1 + rtc2) / 2;
  GM_PRINT(GM_DEBUG_GET_PCI_SPEED_REV4,
	   (" rtc = %d  cpuc = %d\n",rtc,cpuc));
  if (rtc) {
    cpu_freq = (cpuc*2)/rtc;
    GM_PRINT(GM_DEBUG_GET_PCI_SPEED_REV4,
	     ("Using 1x clockval - measured LANaiX rate as %d MHz\n",
	      cpu_freq));
  } else {
    GM_RETURN_STATUS (GM_FAILURE);    
  }

  pci_clock = (gm_u32_t) gm_read_lanai_special_reg_u16 (is, lX.PCI_CLOCK);
  pci_freq = (pci_clock * cpu_freq) >> 13;

  GM_PRINT(GM_DEBUG_GET_PCI_SPEED_REV4,
	   ("PCI clock = %d, pci_freq %d\n", pci_clock, pci_freq));

  is->bus_rate = pci_freq;

  GM_RETURN_STATUS (GM_SUCCESS);
}

#define GM_DEBUG_UPDATE_CLOCKVAL_REV3 0

static gm_status_t
gm_update_clockval_rev3 (gm_instance_state_t * is)
{
#if GM_SUPPORT_L7 | GM_SUPPORT_L8 | GM_SUPPORT_L9
  unsigned int bus_rate = 0;
  unsigned int clockval = 0;
  unsigned int max_lanai_speed = 0;
  unsigned int multiplier = 0;

  gm_assert (is);

  GM_PRINT (GM_DEBUG_UPDATE_CLOCKVAL_REV3,
	    ("gm_update_clockval_rev3: called\n"));

  /* Disable PCI bus master DMA */

  GM_PRINT (GM_DEBUG_UPDATE_CLOCKVAL_REV3, ("Disable PCI bus master DMA.\n"));
  if (gm_disable_PCI_config_command_bit (is, GM_PCI_COMMAND_MASTER)
      != GM_SUCCESS)
    {
      return (GM_FAILURE);
    }

  gm_lanai_reset_on (is);
  gm_ebus_reset_on (is);
  gm_ebus_reset_off (is);
  gm_arch_spin (is, 25000);

  if (gm_is_l9(is)) 
    {
      gm_u32_t cpuc,rtc1,rtc2,rtc;
      rtc1 = gm_read_lanai_special_reg_s32 (is, l9.RTC);
      cpuc = gm_read_lanai_special_reg_u32 (is, l9.CPUC);
      rtc2 = gm_read_lanai_special_reg_s32 (is, l9.RTC);

      rtc = (rtc1 + rtc2) / 2;
      GM_PRINT(GM_DEBUG_UPDATE_CLOCKVAL_REV3,(" rtc = %d  cpuc = %d\n",rtc,cpuc));
      if (rtc) {
	GM_PRINT(1,("Using 1x clockval - measured LANai9 rate as %d MHz\n",
		    (cpuc*2)/rtc));
	bus_rate = (cpuc*2)/rtc;
      }
      if (bus_rate/2 > 80)
	{
	  GM_NOTE (("PCI clock detected as %d Mhz, gm_stbar does not work"
		    " or something is very wrong\n", bus_rate/2));
	  return GM_FAILURE;
	}
    }

  if (!bus_rate) {
    GM_PRINT (GM_DEBUG_UPDATE_CLOCKVAL_REV3,
	      ("Inflating 'lanai_rate' control program.\n"));
    {
      /* Load a compressed MCP - see zlib.h for details */

      z_stream zs;
      int rv, err;

      zs.zalloc = gm_zlib_calloc;
      zs.zfree = gm_zlib_free;
      zs.opaque = 0;

      rv = inflateInit (&zs);
      if (rv != Z_OK)
	{
	  GM_NOTE (("fatal error in inflateInit(%d)\n", rv));
	  return (GM_FAILURE);
	}
      zs.next_in = (unsigned char *) gm_control_program_lanai_rate;
      zs.avail_in = gm_control_program_lanai_rate_length;
      zs.next_out = (unsigned char *) is->lanai.sram;
      zs.avail_out = is->lanai.eeprom.lanai_sram_size;

      if (((err = inflate (&zs, Z_FINISH)) != Z_STREAM_END)
	  || (zs.avail_in != 0) || (zs.avail_out == 0))
	{
	  GM_NOTE (("fatal error in 'lanai_rate' inflate err=%d?\n", err));
	  return (GM_FAILURE);
	}

      if (inflateEnd (&zs) != Z_OK)
	{
	  GM_NOTE (("fatal error in 'lanai_rate' inflateEnd\n"));
	  return (GM_FAILURE);
	}

      GM_PRINT (GM_DEBUG_UPDATE_CLOCKVAL_REV3,
		("Inflated 'lanai_rate' mcp from %ld to %ld bytes.\n",
		 gm_control_program_lanai_rate_length, zs.total_out));
    }

    GM_PRINT (GM_DEBUG_UPDATE_CLOCKVAL_REV3,
	      ("Taking the LANai out of reset\n"));
    if (gm_lanai_reset_off (is) != GM_SUCCESS)
      {
	GM_WARN (("Could not take LANai out of reset.\n"));
	return (GM_FAILURE);
      }

    {
      int loop_count = 0;
      gm_u32_n_t *iptr = (gm_u32_n_t *) is->lanai.sram;

      /* gm_arch_timed_sleep(&is->command_sync,1); */
      gm_arch_spin (is, 50000);
      while ((loop_count++ < 20) && gm_ntoh_u32 (iptr[0]) != 0xabababab)
	{
	  /* wait */
	  /* gm_arch_timed_sleep(&is->command_sync,1); */
	  gm_arch_spin (is, 100000);
	}

      if (gm_ntoh_u32 (iptr[0]) != 0xabababab)
	{
	  GM_WARN (("LANai 'rate' program never completed\n"));
	  _GM_WARN (("iptrs  0x%x 0x%x 0x%x\n",
		     gm_ntoh_u32 (iptr[0]), gm_ntoh_u32 (iptr[1]),
		     gm_ntoh_u32 (iptr[2])));
	  return (GM_FAILURE);
	}

      bus_rate = gm_ntoh_u32 (iptr[1]);
      GM_PRINT (GM_PRINT_LEVEL >= 0,
		("bus timing took %d iterations\n", loop_count));
    }

  }


  if (bus_rate < 20)
    {
      GM_INFO (("bus_rate (%d) is bad - setting to 66\n", bus_rate));
      bus_rate = 66;
    }

  GM_PRINT (GM_DEBUG_UPDATE_CLOCKVAL_REV3,
	    ("Putting LANai && Ebus back in reset\n"));
  gm_lanai_reset_on (is);
  gm_ebus_reset_on (is);
  gm_arch_spin (is, 20000);

  GM_PRINT (GM_DEBUG_BOARD_INIT,
	    ("gm[%d]: LANai speed is %d MHz\n", is->id, bus_rate));
  if (gm_is_l9(is)) 
    {
      GM_PRINT (GM_DEBUG_BOARD_INIT,
		("gm[%d]: PCI slot speed is %d MHz\n", is->id, bus_rate/2));
      is->bus_rate = bus_rate/2;
    }
  else 
    {
      GM_PRINT (GM_DEBUG_BOARD_INIT,
		("gm[%d]: PCI slot speed is %d MHz\n", is->id, bus_rate));
      is->bus_rate = bus_rate;
    }

  clockval = is->lanai.eeprom.lanai_clockval;
  clockval = clockval & (unsigned int) 0xFFFFFF8F;

  max_lanai_speed = is->lanai.eeprom.max_lanai_speed;
  if ((max_lanai_speed == 0) || (max_lanai_speed == (gm_u16_t) 0xFFFF))
    {
      max_lanai_speed = 67;
    }

  /* use 2x here to allow 1.5 or 2.5 multiplier */
  multiplier = (2 * (max_lanai_speed+1)) / bus_rate;
  switch (multiplier)
    {
    case 1:
    case 2:
      /* 1x clockval */
      clockval = clockval | 0x80;
      break;

    case 3:
      /* 1.5x clockval */
      clockval = clockval | 0x90;
      break;

    case 4:
      /* 2x clockval */
      clockval = clockval | 0xA0;
      break;

    case 5:
      /* 2.5x clockval */
      clockval = clockval | 0xB0;
      break;

    case 6:
      /* 3x clockval */
      clockval = clockval | 0xC0;
      break;

    case 0:
    default:
      clockval = clockval | 0x80;
      GM_NOTE (("**** multiplier is bad (%d) - using 1x clockval ****\n",
		multiplier));
      break;
    }

  GM_INFO (("LANai rate set to %d MHz (max = %d MHz)\n",
	    (bus_rate*multiplier/2), max_lanai_speed));

  gm_write_lanai_special_reg_u32 (is, l7.CLOCK, clockval);
  GM_STBAR ();
  gm_arch_spin (is, 100000);
  GM_PRINT (GM_DEBUG_UPDATE_CLOCKVAL_REV3, ("Ebus reset off.\n"));
  gm_ebus_reset_off (is);
  gm_arch_spin (is, 20000);

  GM_PRINT (GM_DEBUG_UPDATE_CLOCKVAL_REV3, ("Enable PCI bus master DMA.\n"));
  if (gm_enable_PCI_config_command_bit (is, GM_PCI_COMMAND_MASTER)
      != GM_SUCCESS)
    {
      GM_NOTE (("Enable PCI bus master DMA failed"
		" after 'lanai_rate' test\n"));
      return (GM_FAILURE);
    }
#endif /* SUPPORT L7 | L8 | L9 */
  return (GM_SUCCESS);
}
#endif /* GM_SUPPORT_PCI */


/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  c-backslash-column:72
  tab-width:8
  End:
*/
